/* Generated by Cython 0.29.14 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "include/eig_mat.h",
            "include/openmp_wrapper.h",
            "include/typedefs.h"
        ],
        "extra_compile_args": [
            "-fopenmp",
            "-Wno-maybe-uninitialized",
            "-Wno-unused-function",
            "-Wno-cpp",
            "-Wno-shorten-64-to-32",
            "-Wno-unneeded-internal-declaration"
        ],
        "extra_link_args": [
            "-fopenmp"
        ],
        "include_dirs": [
            "/home/sbuchoux/miniconda3/envs/SAV_FATSLiM_Py37/lib/python3.7/site-packages/numpy/core/include",
            "./include",
            "./src",
            ".",
            "./fatslimlib"
        ],
        "name": "fatslimlib.core_base",
        "sources": [
            "fatslimlib/core_base.pyx",
            "src/eig_mat.c",
            "src/typeutil.c"
        ]
    },
    "module_name": "fatslimlib.core_base"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_14"
#define CYTHON_HEX_VERSION 0x001D0EF0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__fatslimlib__core_base
#define __PYX_HAVE_API__fatslimlib__core_base
/* Early includes */
#include "typedefs.h"
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "eig_mat.h"
#include "openmp_wrapper.h"
#include "pythread.h"
#include <stdio.h>
#include "pystate.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "fatslimlib/core_base.pyx",
  "stringsource",
  "fatslimlib/core_base.pxd",
  "fatslimlib/core_analysis.pxd",
};
/* MemviewSliceStruct.proto */
struct __pyx_memoryview_obj;
typedef struct {
  struct __pyx_memoryview_obj *memview;
  char *data;
  Py_ssize_t shape[8];
  Py_ssize_t strides[8];
  Py_ssize_t suboffsets[8];
} __Pyx_memviewslice;
#define __Pyx_MemoryView_Len(m)  (m.shape[0])

/* Atomics.proto */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __pyx_atomic_int_type int
#if CYTHON_ATOMICS && __GNUC__ >= 4 && (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL >= 2)) &&\
                    !defined(__i386__)
    #define __pyx_atomic_incr_aligned(value, lock) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_aligned(value, lock) __sync_fetch_and_sub(value, 1)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER) && 0
    #include <Windows.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type LONG
    #define __pyx_atomic_incr_aligned(value, lock) InterlockedIncrement(value)
    #define __pyx_atomic_decr_aligned(value, lock) InterlockedDecrement(value)
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#elif CYTHON_ATOMICS && (defined(__ICC) || defined(__INTEL_COMPILER)) && 0
    #define __pyx_atomic_incr_aligned(value, lock) _InterlockedIncrement(value)
    #define __pyx_atomic_decr_aligned(value, lock) _InterlockedDecrement(value)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using Intel atomics"
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif
typedef volatile __pyx_atomic_int_type __pyx_atomic_int;
#if CYTHON_ATOMICS
    #define __pyx_add_acquisition_count(memview)\
             __pyx_atomic_incr_aligned(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_atomic_decr_aligned(__pyx_get_slice_count_pointer(memview), memview->lock)
#else
    #define __pyx_add_acquisition_count(memview)\
            __pyx_add_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_sub_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
#endif

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;


/*--- Type declarations ---*/
struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate;
struct __pyx_obj_10fatslimlib_13core_analysis_Membrane;
struct __pyx_obj_10fatslimlib_9core_base_Atom;
struct __pyx_obj_10fatslimlib_9core_base_AtomGroup;
struct __pyx_obj_10fatslimlib_9core_base_PBCBox;
struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup;
struct __pyx_obj_10fatslimlib_9core_base_Topology;
struct __pyx_obj_10fatslimlib_9core_base_TopologyReader;
struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader;
struct __pyx_obj_10fatslimlib_9core_base_IndexReader;
struct __pyx_obj_10fatslimlib_9core_base_Frame;
struct __pyx_obj_10fatslimlib_9core_base_Trajectory;
struct __pyx_array_obj;
struct __pyx_MemviewEnum_obj;
struct __pyx_memoryview_obj;
struct __pyx_memoryviewslice_obj;
struct __pyx_t_10fatslimlib_7core_ns_ns_grid;
struct __pyx_t_10fatslimlib_7core_ns_ns_neighborhood;
struct __pyx_t_10fatslimlib_7core_ns_ns_neighborhood_holder;

/* "core_ns.pxd":28
 * 
 * # Neighbor search related structures and functions
 * cdef struct ns_grid:             # <<<<<<<<<<<<<<
 *     fsl_int size
 *     fsl_int[DIM] ncells
 */
struct __pyx_t_10fatslimlib_7core_ns_ns_grid {
  fsl_int size;
  fsl_int ncells[3];
  real cellsize[3];
  fsl_int *nbeads;
  fsl_int **beadids;
};

/* "core_ns.pxd":35
 *     fsl_int **beadids
 * 
 * cdef struct ns_neighborhood:             # <<<<<<<<<<<<<<
 *     real cutoff
 *     fsl_int allocated_size
 */
struct __pyx_t_10fatslimlib_7core_ns_ns_neighborhood {
  real cutoff;
  fsl_int allocated_size;
  fsl_int size;
  fsl_int *beadids;
};

/* "core_ns.pxd":41
 *     fsl_int *beadids
 * 
 * cdef struct ns_neighborhood_holder:             # <<<<<<<<<<<<<<
 *     fsl_int size
 *     ns_neighborhood **neighborhoods
 */
struct __pyx_t_10fatslimlib_7core_ns_ns_neighborhood_holder {
  fsl_int size;
  struct __pyx_t_10fatslimlib_7core_ns_ns_neighborhood **neighborhoods;
};
struct __pyx_opt_args_10fatslimlib_13core_analysis_9Aggregate_set_neighborhood_cutoff;
struct __pyx_opt_args_10fatslimlib_13core_analysis_9Aggregate_compute_apl;
struct __pyx_opt_args_10fatslimlib_13core_analysis_9Aggregate_fix_apl;
struct __pyx_opt_args_10fatslimlib_13core_analysis_9Aggregate_compute_thickness;
struct __pyx_opt_args_10fatslimlib_13core_analysis_9Aggregate_fix_thickness;
struct __pyx_opt_args_10fatslimlib_13core_analysis_8Membrane_fast_compute_thickness;
struct __pyx_opt_args_10fatslimlib_13core_analysis_8Membrane_fast_compute_apl;

/* "core_analysis.pxd":78
 *     cdef fsl_int fast_size(self) nogil
 *     cdef void compute_neighborhood_normal(self, fsl_int index) nogil
 *     cdef void set_neighborhood_cutoff(self, real cutoff, bint force=*) nogil             # <<<<<<<<<<<<<<
 *     cdef void compute_apl(self, real cutoff, real area_limit=*, bint force=*) nogil except *
 *     cdef real fix_apl(self, fsl_int index, real *tmp_apl, bint onlyfix=*) nogil
 */
struct __pyx_opt_args_10fatslimlib_13core_analysis_9Aggregate_set_neighborhood_cutoff {
  int __pyx_n;
  int force;
};

/* "core_analysis.pxd":79
 *     cdef void compute_neighborhood_normal(self, fsl_int index) nogil
 *     cdef void set_neighborhood_cutoff(self, real cutoff, bint force=*) nogil
 *     cdef void compute_apl(self, real cutoff, real area_limit=*, bint force=*) nogil except *             # <<<<<<<<<<<<<<
 *     cdef real fix_apl(self, fsl_int index, real *tmp_apl, bint onlyfix=*) nogil
 *     cdef void compute_thickness(self, Aggregate other, real interleaflet_cutoff=*, bint force=*) nogil except*
 */
struct __pyx_opt_args_10fatslimlib_13core_analysis_9Aggregate_compute_apl {
  int __pyx_n;
  real area_limit;
  int force;
};

/* "core_analysis.pxd":80
 *     cdef void set_neighborhood_cutoff(self, real cutoff, bint force=*) nogil
 *     cdef void compute_apl(self, real cutoff, real area_limit=*, bint force=*) nogil except *
 *     cdef real fix_apl(self, fsl_int index, real *tmp_apl, bint onlyfix=*) nogil             # <<<<<<<<<<<<<<
 *     cdef void compute_thickness(self, Aggregate other, real interleaflet_cutoff=*, bint force=*) nogil except*
 *     cdef real fix_thickness(self, fsl_int index, real *tmp_thickness, bint onlyfix=*) nogil
 */
struct __pyx_opt_args_10fatslimlib_13core_analysis_9Aggregate_fix_apl {
  int __pyx_n;
  int onlyfix;
};

/* "core_analysis.pxd":81
 *     cdef void compute_apl(self, real cutoff, real area_limit=*, bint force=*) nogil except *
 *     cdef real fix_apl(self, fsl_int index, real *tmp_apl, bint onlyfix=*) nogil
 *     cdef void compute_thickness(self, Aggregate other, real interleaflet_cutoff=*, bint force=*) nogil except*             # <<<<<<<<<<<<<<
 *     cdef real fix_thickness(self, fsl_int index, real *tmp_thickness, bint onlyfix=*) nogil
 * 
 */
struct __pyx_opt_args_10fatslimlib_13core_analysis_9Aggregate_compute_thickness {
  int __pyx_n;
  real interleaflet_cutoff;
  int force;
};

/* "core_analysis.pxd":82
 *     cdef real fix_apl(self, fsl_int index, real *tmp_apl, bint onlyfix=*) nogil
 *     cdef void compute_thickness(self, Aggregate other, real interleaflet_cutoff=*, bint force=*) nogil except*
 *     cdef real fix_thickness(self, fsl_int index, real *tmp_thickness, bint onlyfix=*) nogil             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_10fatslimlib_13core_analysis_9Aggregate_fix_thickness {
  int __pyx_n;
  int onlyfix;
};

/* "core_analysis.pxd":99
 * 
 *     # Analysis methods
 *     cdef void fast_compute_thickness(self, real interleaflet_cutoff=*, bint force=*) nogil except*             # <<<<<<<<<<<<<<
 *     cdef void fast_compute_apl(self, real cutoff=*, real area_limit=*, bint force=*) nogil except*
 */
struct __pyx_opt_args_10fatslimlib_13core_analysis_8Membrane_fast_compute_thickness {
  int __pyx_n;
  real interleaflet_cutoff;
  int force;
};

/* "core_analysis.pxd":100
 *     # Analysis methods
 *     cdef void fast_compute_thickness(self, real interleaflet_cutoff=*, bint force=*) nogil except*
 *     cdef void fast_compute_apl(self, real cutoff=*, real area_limit=*, bint force=*) nogil except*             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_10fatslimlib_13core_analysis_8Membrane_fast_compute_apl {
  int __pyx_n;
  real cutoff;
  real area_limit;
  int force;
};
struct __pyx_t_10fatslimlib_9core_base_cAtom_t;
struct __pyx_t_10fatslimlib_9core_base_cAtomgroup_t;
struct __pyx_t_10fatslimlib_9core_base_cPBCBox_t;
struct __pyx_opt_args_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm;
struct __pyx_opt_args_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm_from_ref;
struct __pyx_t_10fatslimlib_9core_base_topol_atom_t;
struct __pyx_t_10fatslimlib_9core_base_topol_residue_t;
struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_fast_get_normals;
struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_get_aggregates;
struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_get_membranes;
struct __pyx_opt_args_10fatslimlib_9core_base_fast_pbc_xcm_from_ref_c;

/* "fatslimlib/core_base.pxd":32
 * 
 * # Constants
 * cdef enum ForcefieldType:             # <<<<<<<<<<<<<<
 *     fft_notset,
 *     fft_allatom, fft_unified, fft_coarse,
 */
enum __pyx_t_10fatslimlib_9core_base_ForcefieldType {
  __pyx_e_10fatslimlib_9core_base_fft_notset,
  __pyx_e_10fatslimlib_9core_base_fft_allatom,
  __pyx_e_10fatslimlib_9core_base_fft_unified,
  __pyx_e_10fatslimlib_9core_base_fft_coarse,
  __pyx_e_10fatslimlib_9core_base_fft_unknown
};

/* "fatslimlib/core_base.pxd":29
 * 
 * # Types
 * ctypedef char[NAME_SIZE] strname_t             # <<<<<<<<<<<<<<
 * 
 * # Constants
 */
typedef char __pyx_t_10fatslimlib_9core_base_strname_t[6];

/* "fatslimlib/core_base.pxd":53
 * 
 * # Other objects
 * cdef struct cAtom_t:             # <<<<<<<<<<<<<<
 *     rvec coords
 *     fsl_int atomid
 */
struct __pyx_t_10fatslimlib_9core_base_cAtom_t {
  rvec coords;
  fsl_int atomid;
  char *name;
};

/* "fatslimlib/core_base.pxd":64
 *     cdef object __weakref__
 * 
 * cdef struct cAtomgroup_t:             # <<<<<<<<<<<<<<
 *     cAtom_t *atoms
 *     fsl_int n_atoms
 */
struct __pyx_t_10fatslimlib_9core_base_cAtomgroup_t {
  struct __pyx_t_10fatslimlib_9core_base_cAtom_t *atoms;
  fsl_int n_atoms;
  char *name;
  fsl_int groupid;
  rvec xcm;
};

/* "fatslimlib/core_base.pxd":80
 *     cdef fast_append(self, Atom atom)
 * 
 * cdef struct cPBCBox_t:             # <<<<<<<<<<<<<<
 *     matrix     box
 *     rvec       fbox_diag
 */
struct __pyx_t_10fatslimlib_9core_base_cPBCBox_t {
  matrix box;
  rvec fbox_diag;
  rvec hbox_diag;
  rvec mhbox_diag;
  real max_cutoff2;
  fsl_int ntric_vec;
  fsl_int tric_shift[12][3];
  real tric_vec[12][3];
};

/* "fatslimlib/core_base.pxd":104
 *     cdef real fast_leaflet_distance2(self, rvec a, rvec b, rvec normal) nogil
 *     cdef real fast_leaflet_distance(self, rvec a, rvec b, rvec normal) nogil
 *     cdef void fast_pbc_xcm(self, real[:, ::1] coords, rvec xcm, fsl_int limit=*) nogil             # <<<<<<<<<<<<<<
 *     cdef void fast_pbc_xcm_from_ref(self, real[:, ::1] coords, rvec ref, rvec xcm, fsl_int limit=*) nogil
 *     cdef real[:, ::1]fast_put_atoms_in_bbox(self, real[:,::1] coords) nogil
 */
struct __pyx_opt_args_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm {
  int __pyx_n;
  fsl_int limit;
};

/* "fatslimlib/core_base.pxd":105
 *     cdef real fast_leaflet_distance(self, rvec a, rvec b, rvec normal) nogil
 *     cdef void fast_pbc_xcm(self, real[:, ::1] coords, rvec xcm, fsl_int limit=*) nogil
 *     cdef void fast_pbc_xcm_from_ref(self, real[:, ::1] coords, rvec ref, rvec xcm, fsl_int limit=*) nogil             # <<<<<<<<<<<<<<
 *     cdef real[:, ::1]fast_put_atoms_in_bbox(self, real[:,::1] coords) nogil
 * 
 */
struct __pyx_opt_args_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm_from_ref {
  int __pyx_n;
  fsl_int limit;
};

/* "fatslimlib/core_base.pxd":108
 *     cdef real[:, ::1]fast_put_atoms_in_bbox(self, real[:,::1] coords) nogil
 * 
 * cdef struct topol_atom_t:             # <<<<<<<<<<<<<<
 *     fsl_int name_id
 *     fsl_int internal_id
 */
struct __pyx_t_10fatslimlib_9core_base_topol_atom_t {
  fsl_int name_id;
  fsl_int internal_id;
  fsl_int atomid;
  fsl_int resid;
  fsl_int residue_internal_id;
};

/* "fatslimlib/core_base.pxd":115
 *     fsl_int residue_internal_id
 * 
 * cdef struct topol_residue_t:             # <<<<<<<<<<<<<<
 *     fsl_int name_id
 *     fsl_int internal_id
 */
struct __pyx_t_10fatslimlib_9core_base_topol_residue_t {
  fsl_int name_id;
  fsl_int internal_id;
  fsl_int resid;
  fsl_int size;
  fsl_int allocated_size;
  fsl_int *atom_internal_ids;
  fsl_int *atomids;
};

/* "fatslimlib/core_base.pxd":261
 *     cdef list get_lipid_coords_bbox_aslist(self)
 *     cdef real[:, ::1] fast_get_directions(self) nogil except *
 *     cdef real[:, ::1] fast_get_normals(self, real proximity_cutoff=*) nogil             # <<<<<<<<<<<<<<
 *     cpdef list get_aggregates(self, real cutoff=*, bint update=*)
 *     cpdef list get_membranes(self, real cutoff=*, bint update=*)
 */
struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_fast_get_normals {
  int __pyx_n;
  real proximity_cutoff;
};

/* "fatslimlib/core_base.pxd":262
 *     cdef real[:, ::1] fast_get_directions(self) nogil except *
 *     cdef real[:, ::1] fast_get_normals(self, real proximity_cutoff=*) nogil
 *     cpdef list get_aggregates(self, real cutoff=*, bint update=*)             # <<<<<<<<<<<<<<
 *     cpdef list get_membranes(self, real cutoff=*, bint update=*)
 * 
 */
struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_get_aggregates {
  int __pyx_n;
  real cutoff;
  int update;
};

/* "fatslimlib/core_base.pxd":263
 *     cdef real[:, ::1] fast_get_normals(self, real proximity_cutoff=*) nogil
 *     cpdef list get_aggregates(self, real cutoff=*, bint update=*)
 *     cpdef list get_membranes(self, real cutoff=*, bint update=*)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_get_membranes {
  int __pyx_n;
  real cutoff;
  int update;
};

/* "fatslimlib/core_base.pyx":383
 * 
 * 
 * cdef void fast_pbc_xcm_from_ref_c(cPBCBox_t pbcbox, real[:, ::1] coords, rvec ref, rvec xcm,             # <<<<<<<<<<<<<<
 *                                   fsl_int first=0, fsl_int last=NOTSET) nogil:
 *         cdef fsl_int i
 */
struct __pyx_opt_args_10fatslimlib_9core_base_fast_pbc_xcm_from_ref_c {
  int __pyx_n;
  fsl_int first;
  fsl_int last;
};

/* "core_analysis.pxd":29
 * 
 * # Classes
 * cdef class Aggregate:             # <<<<<<<<<<<<<<
 *     # Attributes
 *     cdef Frame frame
 */
struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate {
  PyObject_HEAD
  struct __pyx_vtabstruct_10fatslimlib_13core_analysis_Aggregate *__pyx_vtab;
  struct __pyx_obj_10fatslimlib_9core_base_Frame *frame;
  __Pyx_memviewslice beadids;
  __Pyx_memviewslice resids;
  __Pyx_memviewslice hg_atomids;
  __Pyx_memviewslice lipid_atomids;
  __Pyx_memviewslice coords;
  __Pyx_memviewslice directions;
  __Pyx_memviewslice normals;
  __Pyx_memviewslice neighborhood_normals;
  real neighborhood_cutoff;
  struct __pyx_t_10fatslimlib_7core_ns_ns_neighborhood_holder *neighborhoods;
  rvec xcm;
  rvec avg_normal;
  int is_planar;
  PyObject *lipid_types;
  PyObject *beadids_by_type;
  PyObject *nlipids_by_type;
  real apl_cutoff;
  __Pyx_memviewslice apl_values;
  real apl_min;
  real apl_max;
  real area;
  PyObject *apl_by_types;
  __Pyx_memviewslice area_by_types;
  real apl_avg;
  real thickness_cutoff;
  __Pyx_memviewslice thickness_values;
  real thickness_avg;
  real thickness_min;
  real thickness_max;
};


/* "core_analysis.pxd":85
 * 
 * 
 * cdef class Membrane:             # <<<<<<<<<<<<<<
 *     # Attributes
 *     cdef Frame frame
 */
struct __pyx_obj_10fatslimlib_13core_analysis_Membrane {
  PyObject_HEAD
  struct __pyx_vtabstruct_10fatslimlib_13core_analysis_Membrane *__pyx_vtab;
  struct __pyx_obj_10fatslimlib_9core_base_Frame *frame;
  struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *leaflet1;
  struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *leaflet2;
  fsl_int type;
  real thickness;
  real apl;
};


/* "fatslimlib/core_base.pxd":58
 *     char *name
 * 
 * cdef class Atom:             # <<<<<<<<<<<<<<
 *     # Attributes
 *     cdef cAtom_t c_atom
 */
struct __pyx_obj_10fatslimlib_9core_base_Atom {
  PyObject_HEAD
  struct __pyx_t_10fatslimlib_9core_base_cAtom_t c_atom;
  PyObject *name;
  PyObject *__weakref__;
};


/* "fatslimlib/core_base.pxd":71
 *     rvec xcm
 * 
 * cdef class AtomGroup:             # <<<<<<<<<<<<<<
 *     # Attributes
 *     cdef cAtomgroup_t c_atomgroup
 */
struct __pyx_obj_10fatslimlib_9core_base_AtomGroup {
  PyObject_HEAD
  struct __pyx_vtabstruct_10fatslimlib_9core_base_AtomGroup *__pyx_vtab;
  struct __pyx_t_10fatslimlib_9core_base_cAtomgroup_t c_atomgroup;
  PyObject *__weakref__;
  PyObject *name;
};


/* "fatslimlib/core_base.pxd":90
 *     real[DIM]  tric_vec[MAX_NTRICVEC]
 * 
 * cdef class PBCBox:             # <<<<<<<<<<<<<<
 *     # Attributes
 *     cdef cPBCBox_t c_pbcbox
 */
struct __pyx_obj_10fatslimlib_9core_base_PBCBox {
  PyObject_HEAD
  struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *__pyx_vtab;
  struct __pyx_t_10fatslimlib_9core_base_cPBCBox_t c_pbcbox;
  rvec center;
  rvec bbox_center;
};


/* "fatslimlib/core_base.pxd":124
 *     fsl_int *atomids
 * 
 * cdef class TopologyGroup(object):             # <<<<<<<<<<<<<<
 *     # Attrbutes
 *     cdef readonly Topology parent
 */
struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup {
  PyObject_HEAD
  struct __pyx_vtabstruct_10fatslimlib_9core_base_TopologyGroup *__pyx_vtab;
  struct __pyx_obj_10fatslimlib_9core_base_Topology *parent;
  PyObject *name;
  PyObject *resname;
  fsl_int resid;
  struct __pyx_t_10fatslimlib_9core_base_topol_atom_t *atoms;
  fsl_int size;
  fsl_int allocated_size;
};


/* "fatslimlib/core_base.pxd":142
 * 
 * 
 * cdef class Topology(object):             # <<<<<<<<<<<<<<
 *     # Attributes
 *     cdef strname_t *atomnames
 */
struct __pyx_obj_10fatslimlib_9core_base_Topology {
  PyObject_HEAD
  struct __pyx_vtabstruct_10fatslimlib_9core_base_Topology *__pyx_vtab;
  __pyx_t_10fatslimlib_9core_base_strname_t *atomnames;
  fsl_int atomnames_size;
  fsl_int atomnames_allocated_size;
  __pyx_t_10fatslimlib_9core_base_strname_t *resnames;
  fsl_int resnames_size;
  fsl_int resnames_allocated_size;
  struct __pyx_t_10fatslimlib_9core_base_topol_atom_t *atoms;
  fsl_int atoms_size;
  fsl_int atoms_allocated_size;
  fsl_int *atomids_to_internalids;
  fsl_int atomids_to_internalids_size;
  fsl_int atomids_to_internalids_allocated_size;
  struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *residues;
  fsl_int residues_size;
  fsl_int residues_allocated_size;
  fsl_int *resids_to_internalids;
  fsl_int resids_to_internalids_size;
  fsl_int resids_to_internalids_allocated_size;
};


/* "fatslimlib/core_base.pxd":180
 *     cdef topol_residue_t *fast_get_residue_from_atomid(self, fsl_int atomid) nogil
 * 
 * cdef class TopologyReader(object):             # <<<<<<<<<<<<<<
 *     # Attributes
 *     cdef bytes filename
 */
struct __pyx_obj_10fatslimlib_9core_base_TopologyReader {
  PyObject_HEAD
  struct __pyx_vtabstruct_10fatslimlib_9core_base_TopologyReader *__pyx_vtab;
  PyObject *filename;
  struct __pyx_obj_10fatslimlib_9core_base_Topology *topology;
};


/* "fatslimlib/core_base.pxd":189
 * 
 * 
 * cdef class CoordinateReader(object):             # <<<<<<<<<<<<<<
 *     # Attributes
 *     cdef bytes filename
 */
struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader {
  PyObject_HEAD
  struct __pyx_vtabstruct_10fatslimlib_9core_base_CoordinateReader *__pyx_vtab;
  PyObject *filename;
  fsl_int nframes;
  __Pyx_memviewslice coordinate_offsets;
  __Pyx_memviewslice box_offsets;
  __Pyx_memviewslice timesteps;
};


/* "fatslimlib/core_base.pxd":204
 * 
 * 
 * cdef class IndexReader(object):             # <<<<<<<<<<<<<<
 *     # Attributes
 *     cdef bint loaded
 */
struct __pyx_obj_10fatslimlib_9core_base_IndexReader {
  PyObject_HEAD
  struct __pyx_vtabstruct_10fatslimlib_9core_base_IndexReader *__pyx_vtab;
  int loaded;
  PyObject *filename;
  PyObject *groups;
};


/* "fatslimlib/core_base.pxd":213
 *     cdef fast_load(self)
 * 
 * cdef class Frame:             # <<<<<<<<<<<<<<
 *     # Attributes
 *     cdef fsl_int index
 */
struct __pyx_obj_10fatslimlib_9core_base_Frame {
  PyObject_HEAD
  struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *__pyx_vtab;
  fsl_int index;
  real timestep;
  struct __pyx_obj_10fatslimlib_9core_base_Trajectory *trajectory;
  struct __pyx_obj_10fatslimlib_9core_base_Topology *topology;
  struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *coords_reader;
  struct __pyx_obj_10fatslimlib_9core_base_PBCBox *box;
  fsl_int size;
  __Pyx_memviewslice hg_group_coords;
  __Pyx_memviewslice hg_coords_bbox;
  __Pyx_memviewslice lipid_coords;
  __Pyx_memviewslice lipid_coords_bbox;
  PyObject *lipid_coords_aslist;
  PyObject *lipids_coords_bbox_aslist;
  __Pyx_memviewslice lipid_atomids_offsets;
  __Pyx_memviewslice bead_coords;
  __Pyx_memviewslice bead_coords_bbox;
  __Pyx_memviewslice interacting_group_coords;
  __Pyx_memviewslice interacting_group_coords_bbox;
  real proximity_cutoff;
  struct __pyx_t_10fatslimlib_7core_ns_ns_neighborhood_holder *neighbors;
  __Pyx_memviewslice directions;
  __Pyx_memviewslice normals;
  PyObject *aggregates;
  int aggregates_retrieved;
  real aggregates_cutoff;
  PyObject *membranes;
  int membranes_retrieved;
  real membranes_cutoff;
};


/* "fatslimlib/core_base.pxd":266
 * 
 * 
 * cdef class Trajectory(object):             # <<<<<<<<<<<<<<
 *     # Attributes
 *     cdef bytes bead_group_name
 */
struct __pyx_obj_10fatslimlib_9core_base_Trajectory {
  PyObject_HEAD
  struct __pyx_vtabstruct_10fatslimlib_9core_base_Trajectory *__pyx_vtab;
  PyObject *bead_group_name;
  PyObject *interacting_group_name;
  struct __pyx_obj_10fatslimlib_9core_base_Topology *topology;
  struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *topol_reader;
  struct __pyx_obj_10fatslimlib_9core_base_IndexReader *index_reader;
  struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *coords_reader;
  int be_verbose;
  fsl_int cur_frame;
  int initialized;
  PyObject *timesteps;
  __Pyx_memviewslice hg_group_atomids;
  __Pyx_memviewslice interacting_atomids;
  __Pyx_memviewslice lipid_atomids;
  __Pyx_memviewslice hg_bead_atomids;
  __Pyx_memviewslice lipid_atomids_offsets;
  __Pyx_memviewslice hg_bead_atomids_offsets;
  __Pyx_memviewslice lipid_hg_indices;
  fsl_int forcefield_type;
  real normal_cutoff;
  PyObject *membranes;
  PyObject *aggregates;
};


/* "View.MemoryView":105
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_array_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_array *__pyx_vtab;
  char *data;
  Py_ssize_t len;
  char *format;
  int ndim;
  Py_ssize_t *_shape;
  Py_ssize_t *_strides;
  Py_ssize_t itemsize;
  PyObject *mode;
  PyObject *_format;
  void (*callback_free_data)(void *);
  int free_data;
  int dtype_is_object;
};


/* "View.MemoryView":279
 * 
 * @cname('__pyx_MemviewEnum')
 * cdef class Enum(object):             # <<<<<<<<<<<<<<
 *     cdef object name
 *     def __init__(self, name):
 */
struct __pyx_MemviewEnum_obj {
  PyObject_HEAD
  PyObject *name;
};


/* "View.MemoryView":330
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */
struct __pyx_memoryview_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_memoryview *__pyx_vtab;
  PyObject *obj;
  PyObject *_size;
  PyObject *_array_interface;
  PyThread_type_lock lock;
  __pyx_atomic_int acquisition_count[2];
  __pyx_atomic_int *acquisition_count_aligned_p;
  Py_buffer view;
  int flags;
  int dtype_is_object;
  __Pyx_TypeInfo *typeinfo;
};


/* "View.MemoryView":965
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */
struct __pyx_memoryviewslice_obj {
  struct __pyx_memoryview_obj __pyx_base;
  __Pyx_memviewslice from_slice;
  PyObject *from_object;
  PyObject *(*to_object_func)(char *);
  int (*to_dtype_func)(char *, PyObject *);
};



/* "core_analysis.pxd":29
 * 
 * # Classes
 * cdef class Aggregate:             # <<<<<<<<<<<<<<
 *     # Attributes
 *     cdef Frame frame
 */

struct __pyx_vtabstruct_10fatslimlib_13core_analysis_Aggregate {
  fsl_int (*fast_get_atomid)(struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *, fsl_int);
  fsl_int (*get_atomid)(struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *, fsl_int, int __pyx_skip_dispatch);
  int (*fast_same_restype)(struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *, fsl_int, fsl_int);
  int (*same_restype)(struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *, fsl_int, fsl_int, int __pyx_skip_dispatch);
  PyObject *(*get_resname)(struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *, fsl_int, int __pyx_skip_dispatch);
  fsl_int (*get_resid)(struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *, fsl_int, int __pyx_skip_dispatch);
  void (*refresh_cache)(struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *);
  fsl_int (*fast_size)(struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *);
  void (*compute_neighborhood_normal)(struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *, fsl_int);
  void (*set_neighborhood_cutoff)(struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *, real, struct __pyx_opt_args_10fatslimlib_13core_analysis_9Aggregate_set_neighborhood_cutoff *__pyx_optional_args);
  void (*compute_apl)(struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *, real, struct __pyx_opt_args_10fatslimlib_13core_analysis_9Aggregate_compute_apl *__pyx_optional_args);
  real (*fix_apl)(struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *, fsl_int, real *, struct __pyx_opt_args_10fatslimlib_13core_analysis_9Aggregate_fix_apl *__pyx_optional_args);
  void (*compute_thickness)(struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *, struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *, struct __pyx_opt_args_10fatslimlib_13core_analysis_9Aggregate_compute_thickness *__pyx_optional_args);
  real (*fix_thickness)(struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *, fsl_int, real *, struct __pyx_opt_args_10fatslimlib_13core_analysis_9Aggregate_fix_thickness *__pyx_optional_args);
};
static struct __pyx_vtabstruct_10fatslimlib_13core_analysis_Aggregate *__pyx_vtabptr_10fatslimlib_13core_analysis_Aggregate;


/* "core_analysis.pxd":85
 * 
 * 
 * cdef class Membrane:             # <<<<<<<<<<<<<<
 *     # Attributes
 *     cdef Frame frame
 */

struct __pyx_vtabstruct_10fatslimlib_13core_analysis_Membrane {
  void (*fast_compute_thickness)(struct __pyx_obj_10fatslimlib_13core_analysis_Membrane *, struct __pyx_opt_args_10fatslimlib_13core_analysis_8Membrane_fast_compute_thickness *__pyx_optional_args);
  void (*fast_compute_apl)(struct __pyx_obj_10fatslimlib_13core_analysis_Membrane *, struct __pyx_opt_args_10fatslimlib_13core_analysis_8Membrane_fast_compute_apl *__pyx_optional_args);
};
static struct __pyx_vtabstruct_10fatslimlib_13core_analysis_Membrane *__pyx_vtabptr_10fatslimlib_13core_analysis_Membrane;


/* "fatslimlib/core_base.pyx":323
 *     rvec_normalize(normal)
 * 
 * cdef class AtomGroup(object):             # <<<<<<<<<<<<<<
 * 
 *     # initialization
 */

struct __pyx_vtabstruct_10fatslimlib_9core_base_AtomGroup {
  PyObject *(*fast_append)(struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *, struct __pyx_obj_10fatslimlib_9core_base_Atom *);
};
static struct __pyx_vtabstruct_10fatslimlib_9core_base_AtomGroup *__pyx_vtabptr_10fatslimlib_9core_base_AtomGroup;


/* "fatslimlib/core_base.pyx":420
 * 
 * # noinspection PyNoneFunctionAssignment
 * cdef class PBCBox(object):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, real[:,::1] box):
 */

struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox {
  void (*fast_update)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, __Pyx_memviewslice);
  void (*fast_pbc_dx)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, real *, real *, real *);
  void (*fast_pbc_dx_leaflet)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, real *, real *, real *, real *);
  real (*fast_distance2)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, real *, real *);
  real (*fast_distance)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, real *, real *);
  real (*fast_leaflet_distance2)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, real *, real *, real *);
  real (*fast_leaflet_distance)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, real *, real *, real *);
  void (*fast_pbc_xcm)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, __Pyx_memviewslice, real *, struct __pyx_opt_args_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm *__pyx_optional_args);
  void (*fast_pbc_xcm_from_ref)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, __Pyx_memviewslice, real *, real *, struct __pyx_opt_args_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm_from_ref *__pyx_optional_args);
  __Pyx_memviewslice (*fast_put_atoms_in_bbox)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, __Pyx_memviewslice);
};
static struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *__pyx_vtabptr_10fatslimlib_9core_base_PBCBox;


/* "fatslimlib/core_base.pyx":714
 * 
 * 
 * cdef class TopologyGroup(object):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, Topology parent, str resname="group", fsl_int resid=1):
 */

struct __pyx_vtabstruct_10fatslimlib_9core_base_TopologyGroup {
  int (*fast_contains)(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *, struct __pyx_t_10fatslimlib_9core_base_topol_atom_t);
  void (*add_atom)(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *, struct __pyx_t_10fatslimlib_9core_base_topol_atom_t);
  void (*merge_other_group)(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *, struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *);
  __Pyx_memviewslice (*fast_get_atomids)(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *);
  PyObject *(*group_by_resid)(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *);
};
static struct __pyx_vtabstruct_10fatslimlib_9core_base_TopologyGroup *__pyx_vtabptr_10fatslimlib_9core_base_TopologyGroup;


/* "fatslimlib/core_base.pyx":871
 *     residue.size += 1
 * 
 * cdef class Topology(object):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self):
 */

struct __pyx_vtabstruct_10fatslimlib_9core_base_Topology {
  void (*set_residues_allocation)(struct __pyx_obj_10fatslimlib_9core_base_Topology *, fsl_int);
  void (*set_atoms_allocation)(struct __pyx_obj_10fatslimlib_9core_base_Topology *, fsl_int);
  void (*internal_append)(struct __pyx_obj_10fatslimlib_9core_base_Topology *, fsl_int, PyObject *, PyObject *, fsl_int);
  fsl_int (*fast_get_atom_internal_id)(struct __pyx_obj_10fatslimlib_9core_base_Topology *, fsl_int);
  struct __pyx_t_10fatslimlib_9core_base_topol_atom_t *(*fast_get_atom)(struct __pyx_obj_10fatslimlib_9core_base_Topology *, fsl_int);
  fsl_int (*fast_get_resid_from_atomid)(struct __pyx_obj_10fatslimlib_9core_base_Topology *, fsl_int);
  fsl_int (*fast_get_residue_internal_id)(struct __pyx_obj_10fatslimlib_9core_base_Topology *, fsl_int);
  struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *(*fast_get_residue)(struct __pyx_obj_10fatslimlib_9core_base_Topology *, fsl_int);
  struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *(*fast_get_residue_from_atomid)(struct __pyx_obj_10fatslimlib_9core_base_Topology *, fsl_int);
};
static struct __pyx_vtabstruct_10fatslimlib_9core_base_Topology *__pyx_vtabptr_10fatslimlib_9core_base_Topology;


/* "fatslimlib/core_base.pyx":1211
 *         return self.atoms_size
 * 
 * cdef class TopologyReader(object):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, str filename, bint verbose=True):
 */

struct __pyx_vtabstruct_10fatslimlib_9core_base_TopologyReader {
  PyObject *(*load)(struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *);
};
static struct __pyx_vtabstruct_10fatslimlib_9core_base_TopologyReader *__pyx_vtabptr_10fatslimlib_9core_base_TopologyReader;


/* "fatslimlib/core_base.pyx":1235
 * 
 * 
 * cdef class CoordinateReader(object):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, str filename, bint verbose=True):
 */

struct __pyx_vtabstruct_10fatslimlib_9core_base_CoordinateReader {
  PyObject *(*preload)(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *);
  struct __pyx_obj_10fatslimlib_9core_base_PBCBox *(*load_box)(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *, fsl_int);
  PyObject *(*assert_frame_id)(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *, fsl_int);
  __Pyx_memviewslice (*load_coords)(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *, fsl_int, __Pyx_memviewslice);
};
static struct __pyx_vtabstruct_10fatslimlib_9core_base_CoordinateReader *__pyx_vtabptr_10fatslimlib_9core_base_CoordinateReader;


/* "fatslimlib/core_base.pyx":1265
 * 
 * 
 * cdef class IndexReader(object):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, str filename, bint verbose=True):
 */

struct __pyx_vtabstruct_10fatslimlib_9core_base_IndexReader {
  PyObject *(*fast_load)(struct __pyx_obj_10fatslimlib_9core_base_IndexReader *);
};
static struct __pyx_vtabstruct_10fatslimlib_9core_base_IndexReader *__pyx_vtabptr_10fatslimlib_9core_base_IndexReader;


/* "fatslimlib/core_base.pyx":1299
 * 
 * 
 * cdef class Frame(object):             # <<<<<<<<<<<<<<
 *     def __init__(self, Trajectory trajectory, fsl_int index, real timestep):
 *         self.trajectory = trajectory
 */

struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame {
  fsl_int (*get_forcefield_type)(struct __pyx_obj_10fatslimlib_9core_base_Frame *, int __pyx_skip_dispatch);
  __Pyx_memviewslice (*fast_get_hg_group_coords)(struct __pyx_obj_10fatslimlib_9core_base_Frame *);
  __Pyx_memviewslice (*fast_get_hg_group_coords_bbox)(struct __pyx_obj_10fatslimlib_9core_base_Frame *);
  __Pyx_memviewslice (*fast_get_interacting_group_coords)(struct __pyx_obj_10fatslimlib_9core_base_Frame *);
  __Pyx_memviewslice (*fast_get_interacting_group_coords_bbox)(struct __pyx_obj_10fatslimlib_9core_base_Frame *);
  __Pyx_memviewslice (*fast_get_bead_coords)(struct __pyx_obj_10fatslimlib_9core_base_Frame *);
  __Pyx_memviewslice (*fast_get_bead_coords_bbox)(struct __pyx_obj_10fatslimlib_9core_base_Frame *);
  __Pyx_memviewslice (*fast_get_lipid_coords)(struct __pyx_obj_10fatslimlib_9core_base_Frame *);
  PyObject *(*get_lipid_coords_aslist)(struct __pyx_obj_10fatslimlib_9core_base_Frame *);
  __Pyx_memviewslice (*fast_get_lipid_coords_bbox)(struct __pyx_obj_10fatslimlib_9core_base_Frame *);
  PyObject *(*get_lipid_coords_bbox_aslist)(struct __pyx_obj_10fatslimlib_9core_base_Frame *);
  __Pyx_memviewslice (*fast_get_directions)(struct __pyx_obj_10fatslimlib_9core_base_Frame *);
  __Pyx_memviewslice (*fast_get_normals)(struct __pyx_obj_10fatslimlib_9core_base_Frame *, struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_fast_get_normals *__pyx_optional_args);
  PyObject *(*get_aggregates)(struct __pyx_obj_10fatslimlib_9core_base_Frame *, int __pyx_skip_dispatch, struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_get_aggregates *__pyx_optional_args);
  PyObject *(*get_membranes)(struct __pyx_obj_10fatslimlib_9core_base_Frame *, int __pyx_skip_dispatch, struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_get_membranes *__pyx_optional_args);
};
static struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *__pyx_vtabptr_10fatslimlib_9core_base_Frame;


/* "fatslimlib/core_base.pyx":1714
 * FORCEFIELDTYPE_UNKNOWN = fft_unknown
 * 
 * cdef class Trajectory(object):             # <<<<<<<<<<<<<<
 *     def __init__(self,
 *                  TopologyReader top_reader,
 */

struct __pyx_vtabstruct_10fatslimlib_9core_base_Trajectory {
  PyObject *(*assert_initialization)(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *);
  fsl_int (*get_forcefield_type)(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *, int __pyx_skip_dispatch);
  real (*get_rvdw)(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10fatslimlib_9core_base_Trajectory *__pyx_vtabptr_10fatslimlib_9core_base_Trajectory;


/* "View.MemoryView":105
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_array {
  PyObject *(*get_memview)(struct __pyx_array_obj *);
};
static struct __pyx_vtabstruct_array *__pyx_vtabptr_array;


/* "View.MemoryView":330
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */

struct __pyx_vtabstruct_memoryview {
  char *(*get_item_pointer)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*is_slice)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_slice_assignment)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*setitem_slice_assign_scalar)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_indexed)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*convert_item_to_object)(struct __pyx_memoryview_obj *, char *);
  PyObject *(*assign_item_from_object)(struct __pyx_memoryview_obj *, char *, PyObject *);
};
static struct __pyx_vtabstruct_memoryview *__pyx_vtabptr_memoryview;


/* "View.MemoryView":965
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */

struct __pyx_vtabstruct__memoryviewslice {
  struct __pyx_vtabstruct_memoryview __pyx_base;
};
static struct __pyx_vtabstruct__memoryviewslice *__pyx_vtabptr__memoryviewslice;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyIntFromDouble.proto */
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE PyObject* __Pyx_PyInt_FromDouble(double value);
#else
#define __Pyx_PyInt_FromDouble(value) PyLong_FromDouble(value)
#endif

/* dict_setdefault.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value, int is_safe_type);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_FloorDivideObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_FloorDivideObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceFloorDivide(op1, op2) : PyNumber_FloorDivide(op1, op2))
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* PyObjectFormat.proto */
#if CYTHON_USE_UNICODE_WRITER
static PyObject* __Pyx_PyObject_Format(PyObject* s, PyObject* f);
#else
#define __Pyx_PyObject_Format(s, f) PyObject_Format(s, f)
#endif

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* IncludeStringH.proto */
#include <string.h>

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return __Pyx_PyObject_GetAttrStr(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#else
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* decode_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_bytes(
         PyObject* string, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    return __Pyx_decode_c_bytes(
        PyBytes_AS_STRING(string), PyBytes_GET_SIZE(string),
        start, stop, encoding, errors, decode_func);
}

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* MemviewSliceInit.proto */
#define __Pyx_BUF_MAX_NDIMS %(BUF_MAX_NDIMS)d
#define __Pyx_MEMVIEW_DIRECT   1
#define __Pyx_MEMVIEW_PTR      2
#define __Pyx_MEMVIEW_FULL     4
#define __Pyx_MEMVIEW_CONTIG   8
#define __Pyx_MEMVIEW_STRIDED  16
#define __Pyx_MEMVIEW_FOLLOW   32
#define __Pyx_IS_C_CONTIG 1
#define __Pyx_IS_F_CONTIG 2
static int __Pyx_init_memviewslice(
                struct __pyx_memoryview_obj *memview,
                int ndim,
                __Pyx_memviewslice *memviewslice,
                int memview_is_new_reference);
static CYTHON_INLINE int __pyx_add_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
static CYTHON_INLINE int __pyx_sub_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
#define __pyx_get_slice_count_pointer(memview) (memview->acquisition_count_aligned_p)
#define __pyx_get_slice_count(memview) (*__pyx_get_slice_count_pointer(memview))
#define __PYX_INC_MEMVIEW(slice, have_gil) __Pyx_INC_MEMVIEW(slice, have_gil, __LINE__)
#define __PYX_XDEC_MEMVIEW(slice, have_gil) __Pyx_XDEC_MEMVIEW(slice, have_gil, __LINE__)
static CYTHON_INLINE void __Pyx_INC_MEMVIEW(__Pyx_memviewslice *, int, int);
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *, int, int);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* PyUnicode_Unicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* UnaryNegOverflows.proto */
#define UNARY_NEG_WOULD_OVERFLOW(x)\
        (((x) < 0) & ((unsigned long)(x) == 0-(unsigned long)(x)))

static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *); /*proto*/
/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

/* MemviewSliceIsContig.proto */
static int __pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim);

/* OverlappingSlices.proto */
static int __pyx_slices_overlap(__Pyx_memviewslice *slice1,
                                __Pyx_memviewslice *slice2,
                                int ndim, size_t itemsize);

/* Capsule.proto */
static CYTHON_INLINE PyObject *__pyx_capsule_create(void *p, const char *sig);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_fsl_int(fsl_int value);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_nn_fsl_int(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_nn_fsl_int(const char *itemp, PyObject *obj);

/* IsLittleEndian.proto */
static CYTHON_INLINE int __Pyx_Is_Little_Endian(void);

/* BufferFormatCheck.proto */
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts);
static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type);

/* TypeInfoCompare.proto */
static int __pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b);

/* MemviewSliceValidateAndInit.proto */
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(PyObject *, int writable_flag);

/* MemviewDtypeToObject.proto */
static CYTHON_INLINE PyObject *__pyx_memview_get_nn_real(const char *itemp);
static CYTHON_INLINE int __pyx_memview_set_nn_real(const char *itemp, PyObject *obj);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(PyObject *, int writable_flag);

/* ObjectToMemviewSlice.proto */
static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(PyObject *, int writable_flag);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_10fatslimlib_9core_base_ForcefieldType(enum __pyx_t_10fatslimlib_9core_base_ForcefieldType value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* None.proto */
static CYTHON_INLINE int __Pyx_ErrOccurredWithGIL(void); /* proto */

/* MemviewSliceCopyTemplate.proto */
static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object);

/* CIntFromPy.proto */
static CYTHON_INLINE fsl_int __Pyx_PyInt_As_fsl_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CopyContentsUtility.proto */
#define __pyx_memoryview_copy_slice_d_dc_nn_real_c(slice)\
        __pyx_memoryview_copy_new_contig(&slice, "c", 2,\
                                         sizeof(real), (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT),\
                                         0)

/* CopyContentsUtility.proto */
#define __pyx_memoryview_copy_slice_dcd__nn_real_f(slice)\
        __pyx_memoryview_copy_new_contig(&slice, "fortran", 2,\
                                         sizeof(real), (PyBUF_F_CONTIGUOUS | PyBUF_FORMAT),\
                                         0)

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* VoidPtrExport.proto */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_10fatslimlib_9core_base_9AtomGroup_fast_append(struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *__pyx_v_self, struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_atom); /* proto*/
static void __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_update(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_box); /* proto*/
static void __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_pbc_dx(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, real *__pyx_v_ref, real *__pyx_v_other, real *__pyx_v_dx); /* proto*/
static void __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_pbc_dx_leaflet(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, real *__pyx_v_ref, real *__pyx_v_other, real *__pyx_v_dx, real *__pyx_v_ref_normal); /* proto*/
static real __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_distance2(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, real *__pyx_v_a, real *__pyx_v_b); /* proto*/
static real __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_distance(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, real *__pyx_v_a, real *__pyx_v_b); /* proto*/
static real __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_leaflet_distance2(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, real *__pyx_v_a, real *__pyx_v_b, real *__pyx_v_normal); /* proto*/
static real __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_leaflet_distance(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, real *__pyx_v_a, real *__pyx_v_b, real *__pyx_v_normal); /* proto*/
static void __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_coords, real *__pyx_v_xcm, struct __pyx_opt_args_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm *__pyx_optional_args); /* proto*/
static void __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm_from_ref(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_coords, real *__pyx_v_ref, real *__pyx_v_xcm, struct __pyx_opt_args_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm_from_ref *__pyx_optional_args); /* proto*/
static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_put_atoms_in_bbox(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_coords); /* proto*/
static int __pyx_f_10fatslimlib_9core_base_13TopologyGroup_fast_contains(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self, struct __pyx_t_10fatslimlib_9core_base_topol_atom_t __pyx_v_atom); /* proto*/
static void __pyx_f_10fatslimlib_9core_base_13TopologyGroup_add_atom(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self, struct __pyx_t_10fatslimlib_9core_base_topol_atom_t __pyx_v_atom); /* proto*/
static void __pyx_f_10fatslimlib_9core_base_13TopologyGroup_merge_other_group(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self, struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_other); /* proto*/
static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_13TopologyGroup_fast_get_atomids(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_13TopologyGroup_group_by_resid(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self); /* proto*/
static void __pyx_f_10fatslimlib_9core_base_8Topology_set_atoms_allocation(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_size); /* proto*/
static void __pyx_f_10fatslimlib_9core_base_8Topology_set_residues_allocation(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_size); /* proto*/
static void __pyx_f_10fatslimlib_9core_base_8Topology_internal_append(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_resid, PyObject *__pyx_v_resname, PyObject *__pyx_v_atomname, fsl_int __pyx_v_atomid); /* proto*/
static fsl_int __pyx_f_10fatslimlib_9core_base_8Topology_fast_get_atom_internal_id(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_atomid); /* proto*/
static struct __pyx_t_10fatslimlib_9core_base_topol_atom_t *__pyx_f_10fatslimlib_9core_base_8Topology_fast_get_atom(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_atomid); /* proto*/
static fsl_int __pyx_f_10fatslimlib_9core_base_8Topology_fast_get_resid_from_atomid(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_atomid); /* proto*/
static struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *__pyx_f_10fatslimlib_9core_base_8Topology_fast_get_residue_from_atomid(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_atomid); /* proto*/
static fsl_int __pyx_f_10fatslimlib_9core_base_8Topology_fast_get_residue_internal_id(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_resid); /* proto*/
static struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *__pyx_f_10fatslimlib_9core_base_8Topology_fast_get_residue(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_resid); /* proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_14TopologyReader_load(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_16CoordinateReader_preload(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_self); /* proto*/
static struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_f_10fatslimlib_9core_base_16CoordinateReader_load_box(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_self, CYTHON_UNUSED fsl_int __pyx_v_frame_id); /* proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_16CoordinateReader_assert_frame_id(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_self, fsl_int __pyx_v_frame_id); /* proto*/
static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_16CoordinateReader_load_coords(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_self, CYTHON_UNUSED fsl_int __pyx_v_frame_id, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_atomids); /* proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_11IndexReader_fast_load(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_self); /* proto*/
static fsl_int __pyx_f_10fatslimlib_9core_base_5Frame_get_forcefield_type(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_hg_group_coords(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto*/
static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_hg_group_coords_bbox(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto*/
static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_lipid_coords(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_5Frame_get_lipid_coords_aslist(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto*/
static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_lipid_coords_bbox(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_5Frame_get_lipid_coords_bbox_aslist(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto*/
static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_interacting_group_coords(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto*/
static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_interacting_group_coords_bbox(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto*/
static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_bead_coords(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto*/
static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_bead_coords_bbox(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto*/
static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_directions(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto*/
static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_normals(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self, struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_fast_get_normals *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_5Frame_get_aggregates(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_get_aggregates *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_5Frame_get_membranes(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_get_membranes *__pyx_optional_args); /* proto*/
static fsl_int __pyx_f_10fatslimlib_9core_base_10Trajectory_get_forcefield_type(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static real __pyx_f_10fatslimlib_9core_base_10Trajectory_get_rvdw(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_10Trajectory_assert_initialization(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self); /* proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self); /* proto*/
static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto*/
static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/

/* Module declarations from 'fatslimlib.typedefs' */

/* Module declarations from 'fatslimlib.core_ns' */
static struct __pyx_t_10fatslimlib_7core_ns_ns_neighborhood_holder *(*__pyx_f_10fatslimlib_7core_ns_fast_neighbor_search)(__Pyx_memviewslice, __Pyx_memviewslice, struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, real); /*proto*/
static void (*__pyx_f_10fatslimlib_7core_ns_free_neighborhood_holder)(struct __pyx_t_10fatslimlib_7core_ns_ns_neighborhood_holder *); /*proto*/

/* Module declarations from 'libc.math' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cython.view' */
static struct __pyx_array_obj *__pyx_array_new(PyObject *, Py_ssize_t, char *, char *, char *); /*proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'fatslimlib.core_analysis' */
static PyTypeObject *__pyx_ptype_10fatslimlib_13core_analysis_Aggregate = 0;
static PyTypeObject *__pyx_ptype_10fatslimlib_13core_analysis_Membrane = 0;

/* Module declarations from 'fatslimlib.core_base' */
static PyTypeObject *__pyx_ptype_10fatslimlib_9core_base_Atom = 0;
static PyTypeObject *__pyx_ptype_10fatslimlib_9core_base_AtomGroup = 0;
static PyTypeObject *__pyx_ptype_10fatslimlib_9core_base_PBCBox = 0;
static PyTypeObject *__pyx_ptype_10fatslimlib_9core_base_TopologyGroup = 0;
static PyTypeObject *__pyx_ptype_10fatslimlib_9core_base_Topology = 0;
static PyTypeObject *__pyx_ptype_10fatslimlib_9core_base_TopologyReader = 0;
static PyTypeObject *__pyx_ptype_10fatslimlib_9core_base_CoordinateReader = 0;
static PyTypeObject *__pyx_ptype_10fatslimlib_9core_base_IndexReader = 0;
static PyTypeObject *__pyx_ptype_10fatslimlib_9core_base_Frame = 0;
static PyTypeObject *__pyx_ptype_10fatslimlib_9core_base_Trajectory = 0;
static PyTypeObject *__pyx_array_type = 0;
static PyTypeObject *__pyx_MemviewEnum_type = 0;
static PyTypeObject *__pyx_memoryview_type = 0;
static PyTypeObject *__pyx_memoryviewslice_type = 0;
static fsl_int __pyx_v_10fatslimlib_9core_base_OPENMP_MAX_THREADS;
static fsl_int __pyx_v_10fatslimlib_9core_base_OPENMP_NUM_THREADS;
static PyObject *generic = 0;
static PyObject *strided = 0;
static PyObject *indirect = 0;
static PyObject *contiguous = 0;
static PyObject *indirect_contiguous = 0;
static int __pyx_memoryview_thread_locks_used;
static PyThread_type_lock __pyx_memoryview_thread_locks[8];
static fsl_int __pyx_f_10fatslimlib_9core_base_get_num_threads(int __pyx_skip_dispatch); /*proto*/
static fsl_int __pyx_f_10fatslimlib_9core_base_get_max_threads(int __pyx_skip_dispatch); /*proto*/
static fsl_int __pyx_f_10fatslimlib_9core_base_get_num_procs(int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_set_num_threads(fsl_int, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_10fatslimlib_9core_base_mat_to_memview(real (*)[3], __Pyx_memviewslice); /*proto*/
static void __pyx_f_10fatslimlib_9core_base_rvec_to_memview(real *, __Pyx_memviewslice); /*proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_pretty_delta(PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_pretty_duration(PyObject *, int __pyx_skip_dispatch); /*proto*/
static real __pyx_f_10fatslimlib_9core_base_get_memory_usage(int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_10fatslimlib_9core_base_calculate_lipid_direction(real *, struct __pyx_t_10fatslimlib_9core_base_cPBCBox_t, __Pyx_memviewslice, fsl_int, fsl_int, real *); /*proto*/
static void __pyx_f_10fatslimlib_9core_base_calculate_neighborhood_normal(fsl_int, __Pyx_memviewslice, __Pyx_memviewslice, struct __pyx_t_10fatslimlib_7core_ns_ns_neighborhood *, struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, real *, fsl_int *); /*proto*/
static void __pyx_f_10fatslimlib_9core_base_fast_pbc_dx_c(struct __pyx_t_10fatslimlib_9core_base_cPBCBox_t, real *, real *, real *); /*proto*/
static void __pyx_f_10fatslimlib_9core_base_fast_pbc_xcm_from_ref_c(struct __pyx_t_10fatslimlib_9core_base_cPBCBox_t, __Pyx_memviewslice, real *, real *, struct __pyx_opt_args_10fatslimlib_9core_base_fast_pbc_xcm_from_ref_c *__pyx_optional_args); /*proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_topol_residue_atomids_as_list(struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *); /*proto*/
static fsl_int __pyx_f_10fatslimlib_9core_base_topol_residue_atomid_index(struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *, fsl_int); /*proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_topol_residue_to_dict(struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *, PyObject *); /*proto*/
static void __pyx_f_10fatslimlib_9core_base_topol_residue_append_atom(struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *, struct __pyx_t_10fatslimlib_9core_base_topol_atom_t *); /*proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base___pyx_unpickle_TopologyReader__set_state(struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *, PyObject *); /*proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base___pyx_unpickle_CoordinateReader__set_state(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *, PyObject *); /*proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base___pyx_unpickle_IndexReader__set_state(struct __pyx_obj_10fatslimlib_9core_base_IndexReader *, PyObject *); /*proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base___pyx_unpickle_Trajectory__set_state(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *, PyObject *); /*proto*/
static struct __pyx_array_obj *__pyx_array_new(PyObject *, Py_ssize_t, char *, char *, char *); /*proto*/
static void *__pyx_align_pointer(void *, size_t); /*proto*/
static PyObject *__pyx_memoryview_new(PyObject *, int, int, __Pyx_TypeInfo *); /*proto*/
static CYTHON_INLINE int __pyx_memoryview_check(PyObject *); /*proto*/
static PyObject *_unellipsify(PyObject *, int); /*proto*/
static PyObject *assert_direct_dimensions(Py_ssize_t *, int); /*proto*/
static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *, PyObject *); /*proto*/
static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int, int); /*proto*/
static char *__pyx_pybuffer_index(Py_buffer *, char *, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memslice_transpose(__Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice, int, PyObject *(*)(char *), int (*)(char *, PyObject *), int); /*proto*/
static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *); /*proto*/
static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static Py_ssize_t abs_py_ssize_t(Py_ssize_t); /*proto*/
static char __pyx_get_best_slice_order(__Pyx_memviewslice *, int); /*proto*/
static void _copy_strided_to_strided(char *, Py_ssize_t *, char *, Py_ssize_t *, Py_ssize_t *, Py_ssize_t *, int, size_t); /*proto*/
static void copy_strided_to_strided(__Pyx_memviewslice *, __Pyx_memviewslice *, int, size_t); /*proto*/
static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *, int); /*proto*/
static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *, Py_ssize_t *, Py_ssize_t, int, char); /*proto*/
static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *, __Pyx_memviewslice *, char, int); /*proto*/
static int __pyx_memoryview_err_extents(int, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memoryview_err_dim(PyObject *, char *, int); /*proto*/
static int __pyx_memoryview_err(PyObject *, char *); /*proto*/
static int __pyx_memoryview_copy_contents(__Pyx_memviewslice, __Pyx_memviewslice, int, int, int); /*proto*/
static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *, int, int); /*proto*/
static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *, int, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *, int, size_t, void *, int); /*proto*/
static void __pyx_memoryview__slice_assign_scalar(char *, Py_ssize_t *, Py_ssize_t *, int, size_t, void *); /*proto*/
static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *, PyObject *); /*proto*/
static __Pyx_TypeInfo __Pyx_TypeInfo_nn_fsl_int = { "fsl_int", NULL, sizeof(fsl_int), { 0 }, 0, IS_UNSIGNED(fsl_int) ? 'U' : 'I', IS_UNSIGNED(fsl_int), 0 };
static __Pyx_TypeInfo __Pyx_TypeInfo_nn_real = { "real", NULL, sizeof(real), { 0 }, 0, 'R', 0, 0 };
#define __Pyx_MODULE_NAME "fatslimlib.core_base"
extern int __pyx_module_is_main_fatslimlib__core_base;
int __pyx_module_is_main_fatslimlib__core_base = 0;

/* Implementation of 'fatslimlib.core_base' */
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_open;
static PyObject *__pyx_builtin_print;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_Ellipsis;
static PyObject *__pyx_builtin_id;
static const char __pyx_k_O[] = "O";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_i[] = "%i";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_y[] = "y";
static const char __pyx_k_z[] = "z";
static const char __pyx_k_0f[] = ".0f";
static const char __pyx_k_5f[] = ".5f";
static const char __pyx_k__4[] = " ";
static const char __pyx_k__7[] = "\n";
static const char __pyx_k_id[] = "id";
static const char __pyx_k_ms[] = "ms";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_rb[] = "rb";
static const char __pyx_k_us[] = "us";
static const char __pyx_k__22[] = "";
static const char __pyx_k__25[] = "'... ";
static const char __pyx_k_and[] = "' and '";
static const char __pyx_k_box[] = "box";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_msg[] = "msg";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_unk[] = "unk";
static const char __pyx_k_Atom[] = "Atom";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_load[] = "load";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_pack[] = "pack";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_ASCII[] = "ASCII";
static const char __pyx_k_Frame[] = "Frame";
static const char __pyx_k_VmRSS[] = "VmRSS:";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_atoms[] = "atoms";
static const char __pyx_k_begin[] = "begin";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_empty[] = "empty";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_flush[] = "flush";
static const char __pyx_k_group[] = "group";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_int64[] = "int64";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_resid[] = "resid";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_strip[] = "strip";
static const char __pyx_k_PBCBox[] = "PBCBox";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_arange[] = "arange";
static const char __pyx_k_atomid[] = "atomid";
static const char __pyx_k_cutoff[] = "cutoff";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_natoms[] = "natoms";
static const char __pyx_k_parent[] = "parent";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_stdout[] = "stdout";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_tolist[] = "tolist";
static const char __pyx_k_unpack[] = "unpack";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_0_29_14[] = "0.29.14";
static const char __pyx_k_asarray[] = "asarray";
static const char __pyx_k_atomids[] = "atomids";
static const char __pyx_k_beadids[] = "beadids";
static const char __pyx_k_float64[] = "float64";
static const char __pyx_k_fortran[] = "fortran";
static const char __pyx_k_groupid[] = "groupid";
static const char __pyx_k_memview[] = "memview";
static const char __pyx_k_protein[] = "protein";
static const char __pyx_k_resname[] = "resname";
static const char __pyx_k_verbose[] = "verbose";
static const char __pyx_k_Ellipsis[] = "Ellipsis";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_Topology[] = "Topology";
static const char __pyx_k_basename[] = "basename";
static const char __pyx_k_filename[] = "filename";
static const char __pyx_k_get_rvdw[] = "get_rvdw";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_hg_group[] = "hg_group";
static const char __pyx_k_instance[] = " instance";
static const char __pyx_k_is_not_a[] = " is not a ";
static const char __pyx_k_itemsize[] = "itemsize";
static const char __pyx_k_leaflet1[] = "leaflet1";
static const char __pyx_k_leaflet2[] = "leaflet2";
static const char __pyx_k_pyx_capi[] = "__pyx_capi__";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_residues[] = "_residues";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_timestep[] = "timestep";
static const char __pyx_k_AtomGroup[] = "AtomGroup";
static const char __pyx_k_Done_in_s[] = "Done in %s";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_a_memview[] = "a_memview";
static const char __pyx_k_b_memview[] = "b_memview";
static const char __pyx_k_distance2[] = "distance2";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_nresidues[] = "nresidues";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_Trajectory[] = "Trajectory";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_be_verbose[] = "be_verbose";
static const char __pyx_k_headgroups[] = "headgroups";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_setdefault[] = "setdefault";
static const char __pyx_k_startswith[] = "startswith";
static const char __pyx_k_strip_name[] = "strip_name";
static const char __pyx_k_top_reader[] = "top_reader";
static const char __pyx_k_trajectory[] = "trajectory";
static const char __pyx_k_IndexReader[] = "IndexReader";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_resname_str[] = "resname_str";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_atomname_str[] = "atomname_str";
static const char __pyx_k_index_reader[] = "index_reader";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_TopologyGroup[] = "TopologyGroup";
static const char __pyx_k_coords_reader[] = "coords_reader";
static const char __pyx_k_core_analysis[] = "core_analysis";
static const char __pyx_k_get_membranes[] = "get_membranes";
static const char __pyx_k_pyx_getbuffer[] = "__pyx_getbuffer";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_verbose_print[] = "verbose_print";
static const char __pyx_k_TopologyReader[] = "TopologyReader";
static const char __pyx_k_cython_version[] = "__cython_version__";
static const char __pyx_k_get_aggregates[] = "get_aggregates";
static const char __pyx_k_normal_memview[] = "normal_memview";
static const char __pyx_k_View_MemoryView[] = "View.MemoryView";
static const char __pyx_k_allocate_buffer[] = "allocate_buffer";
static const char __pyx_k_dtype_is_object[] = "dtype_is_object";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_CoordinateReader[] = "CoordinateReader";
static const char __pyx_k_proc_self_status[] = "/proc/self/status";
static const char __pyx_k_interacting_group[] = "interacting_group";
static const char __pyx_k_leaflet_distance2[] = "leaflet_distance2";
static const char __pyx_k_pyx_unpickle_Enum[] = "__pyx_unpickle_Enum";
static const char __pyx_k_resid_not_valid_i[] = "resid not valid: %i";
static const char __pyx_k_OPENMP_MAX_THREADS[] = "OPENMP_MAX_THREADS";
static const char __pyx_k_OPENMP_NUM_THREADS[] = "OPENMP_NUM_THREADS";
static const char __pyx_k_atomid_not_valid_i[] = "atomid not valid: %i";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_retrieve_membranes[] = "retrieve_membranes";
static const char __pyx_k_strided_and_direct[] = "<strided and direct>";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_get_forcefield_type[] = "get_forcefield_type";
static const char __pyx_k_retrieve_aggregates[] = "retrieve_aggregates";
static const char __pyx_k_fatslimlib_core_base[] = "fatslimlib.core_base";
static const char __pyx_k_i_groups_loaded_in_s[] = "%i groups loaded in %s";
static const char __pyx_k_strided_and_indirect[] = "<strided and indirect>";
static const char __pyx_k_FORCEFIELDTYPE_COARSE[] = "FORCEFIELDTYPE_COARSE";
static const char __pyx_k_FORCEFIELDTYPE_NOTSET[] = "FORCEFIELDTYPE_NOTSET";
static const char __pyx_k_Loading_groups_from_s[] = "Loading groups from '%s'... ";
static const char __pyx_k_contiguous_and_direct[] = "<contiguous and direct>";
static const char __pyx_k_Atomid_i_internal_id_i[] = "Atomid #%i -> internal id #%i";
static const char __pyx_k_FORCEFIELDTYPE_ALLATOM[] = "FORCEFIELDTYPE_ALLATOM";
static const char __pyx_k_FORCEFIELDTYPE_UNIFIED[] = "FORCEFIELDTYPE_UNIFIED";
static const char __pyx_k_FORCEFIELDTYPE_UNKNOWN[] = "FORCEFIELDTYPE_UNKNOWN";
static const char __pyx_k_MemoryView_of_r_object[] = "<MemoryView of %r object>";
static const char __pyx_k_Loading_topology_from_s[] = "Loading topology from '%s'... ";
static const char __pyx_k_MemoryView_of_r_at_0x_x[] = "<MemoryView of %r at 0x%x>";
static const char __pyx_k_TopologyGroup_s_i_atoms[] = "TopologyGroup '%s': %i atoms";
static const char __pyx_k_contiguous_and_indirect[] = "<contiguous and indirect>";
static const char __pyx_k_i_frames_preloaded_in_s[] = "%i frames preloaded in %s";
static const char __pyx_k_pyx_unpickle_Trajectory[] = "__pyx_unpickle_Trajectory";
static const char __pyx_k_Cannot_index_with_type_s[] = "Cannot index with type '%s'";
static const char __pyx_k_fatslimlib_core_base_pyx[] = "fatslimlib/core_base.pyx";
static const char __pyx_k_pyx_unpickle_IndexReader[] = "__pyx_unpickle_IndexReader";
static const char __pyx_k_Invalid_shape_in_axis_d_d[] = "Invalid shape in axis %d: %d.";
static const char __pyx_k_Trajectory_not_initialized[] = "Trajectory not initialized";
static const char __pyx_k_itemsize_0_for_cython_array[] = "itemsize <= 0 for cython.array";
static const char __pyx_k_pyx_unpickle_TopologyReader[] = "__pyx_unpickle_TopologyReader";
static const char __pyx_k_Frame_index_error_i_i_frames[] = "Frame index error: %i (%i frames)";
static const char __pyx_k_Preloading_coordinates_from_s[] = "Preloading coordinates from '%s'... ";
static const char __pyx_k_pyx_unpickle_CoordinateReader[] = "__pyx_unpickle_CoordinateReader";
static const char __pyx_k_unable_to_allocate_array_data[] = "unable to allocate array data.";
static const char __pyx_k_strided_and_direct_or_indirect[] = "<strided and direct or indirect>";
static const char __pyx_k_Pickling_of_struct_members_such[] = "Pickling of struct members such as self.c_atom must be explicitly requested with @auto_pickle(True)";
static const char __pyx_k_There_is_probably_something_wro[] = "  There is probably something wrong with your box.";
static const char __pyx_k_WARNING_Found_more_than_d_tricl[] = "\nWARNING: Found more than %d triclinic correction vectors, ignoring some.";
static const char __pyx_k_atomid_not_valid_i_max_atomid_i[] = "atomid not valid: %i (max atomid:%i)";
static const char __pyx_k_i_topology_atoms_from_i_residue[] = "%i topology atoms (from %i residues) loaded in %s";
static const char __pyx_k_Atom_with_atomid_i_is_already_re[] = "Atom with atomid #%i is already registered";
static const char __pyx_k_Atomid_not_registered_in_topolog[] = "Atomid not registered in topology: %i. Does group '%s' from '%s' file contains H?";
static const char __pyx_k_Box_does_not_correspond_to_PBC_x[] = "Box does not correspond to PBC=xyz";
static const char __pyx_k_Box_must_be_a_i_x_i_matrix_shape[] = "Box must be a %i x %i matrix. (shape: %i x %i)";
static const char __pyx_k_Buffer_view_does_not_expose_stri[] = "Buffer view does not expose strides";
static const char __pyx_k_Can_only_create_a_buffer_that_is[] = "Can only create a buffer that is contiguous in memory.";
static const char __pyx_k_Cannot_assign_to_read_only_memor[] = "Cannot assign to read-only memoryview";
static const char __pyx_k_Cannot_create_writable_memory_vi[] = "Cannot create writable memory view from read-only memoryview";
static const char __pyx_k_Could_not_change_number_of_OpenM[] = "Could not change number of OpenMP threads (requested value: %i - actual value: %i)";
static const char __pyx_k_Empty_shape_tuple_for_cython_arr[] = "Empty shape tuple for cython.array";
static const char __pyx_k_Frame_index_out_of_range_i_frame[] = "Frame index out of range (%i frames in trajectory)";
static const char __pyx_k_Incompatible_checksums_s_vs_0x50[] = "Incompatible checksums (%s vs 0x500400c = (box_offsets, coordinate_offsets, filename, nframes, timesteps))";
static const char __pyx_k_Incompatible_checksums_s_vs_0x6a[] = "Incompatible checksums (%s vs 0x6a20261 = (aggregates, be_verbose, bead_group_name, coords_reader, cur_frame, forcefield_type, hg_bead_atomids, hg_bead_atomids_offsets, hg_group_atomids, index_reader, initialized, interacting_atomids, interacting_group_name, lipid_atomids, lipid_atomids_offsets, lipid_hg_indices, membranes, normal_cutoff, timesteps, topol_reader, topology))";
static const char __pyx_k_Incompatible_checksums_s_vs_0x7a[] = "Incompatible checksums (%s vs 0x7ac6dcc = (filename, groups, loaded))";
static const char __pyx_k_Incompatible_checksums_s_vs_0x8d[] = "Incompatible checksums (%s vs 0x8dafc6b = (filename, topology))";
static const char __pyx_k_Incompatible_checksums_s_vs_0xb0[] = "Incompatible checksums (%s vs 0xb068931 = (name))";
static const char __pyx_k_Indirect_dimensions_not_supporte[] = "Indirect dimensions not supported";
static const char __pyx_k_Initializing_trajectory_using_gr[] = "Initializing trajectory using groups: '";
static const char __pyx_k_Invalid_mode_expected_c_or_fortr[] = "Invalid mode, expected 'c' or 'fortran', got %s";
static const char __pyx_k_Out_of_bounds_on_buffer_access_a[] = "Out of bounds on buffer access (axis %d)";
static const char __pyx_k_Trajectory_is_not_initialized_co[] = "Trajectory is not initialized correctly!";
static const char __pyx_k_Unable_to_convert_item_to_object[] = "Unable to convert item to object";
static const char __pyx_k_got_differing_extents_in_dimensi[] = "got differing extents in dimension %d (got %d and %d)";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_self_atomids_to_internalids_self[] = "self.atomids_to_internalids,self.atomnames,self.atoms,self.resids_to_internalids,self.residues,self.resnames cannot be converted to a Python object for pickling";
static const char __pyx_k_self_atoms_cannot_be_converted_t[] = "self.atoms cannot be converted to a Python object for pickling";
static const char __pyx_k_self_neighbors_cannot_be_convert[] = "self.neighbors cannot be converted to a Python object for pickling";
static const char __pyx_k_unable_to_allocate_shape_and_str[] = "unable to allocate shape and strides.";
static const char __pyx_k_Pickling_of_struct_members_such_2[] = "Pickling of struct members such as self.c_pbcbox must be explicitly requested with @auto_pickle(True)";
static PyObject *__pyx_kp_s_0_29_14;
static PyObject *__pyx_kp_u_0f;
static PyObject *__pyx_kp_u_5f;
static PyObject *__pyx_n_s_ASCII;
static PyObject *__pyx_n_s_Atom;
static PyObject *__pyx_n_s_AtomGroup;
static PyObject *__pyx_kp_u_Atom_with_atomid_i_is_already_re;
static PyObject *__pyx_kp_u_Atomid_i_internal_id_i;
static PyObject *__pyx_kp_u_Atomid_not_registered_in_topolog;
static PyObject *__pyx_kp_u_Box_does_not_correspond_to_PBC_x;
static PyObject *__pyx_kp_u_Box_must_be_a_i_x_i_matrix_shape;
static PyObject *__pyx_kp_s_Buffer_view_does_not_expose_stri;
static PyObject *__pyx_kp_s_Can_only_create_a_buffer_that_is;
static PyObject *__pyx_kp_s_Cannot_assign_to_read_only_memor;
static PyObject *__pyx_kp_s_Cannot_create_writable_memory_vi;
static PyObject *__pyx_kp_s_Cannot_index_with_type_s;
static PyObject *__pyx_n_s_CoordinateReader;
static PyObject *__pyx_kp_u_Could_not_change_number_of_OpenM;
static PyObject *__pyx_kp_u_Done_in_s;
static PyObject *__pyx_n_s_Ellipsis;
static PyObject *__pyx_kp_s_Empty_shape_tuple_for_cython_arr;
static PyObject *__pyx_n_s_FORCEFIELDTYPE_ALLATOM;
static PyObject *__pyx_n_s_FORCEFIELDTYPE_COARSE;
static PyObject *__pyx_n_s_FORCEFIELDTYPE_NOTSET;
static PyObject *__pyx_n_s_FORCEFIELDTYPE_UNIFIED;
static PyObject *__pyx_n_s_FORCEFIELDTYPE_UNKNOWN;
static PyObject *__pyx_n_s_Frame;
static PyObject *__pyx_kp_u_Frame_index_error_i_i_frames;
static PyObject *__pyx_kp_u_Frame_index_out_of_range_i_frame;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x50;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x6a;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x7a;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x8d;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xb0;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_n_s_IndexReader;
static PyObject *__pyx_kp_s_Indirect_dimensions_not_supporte;
static PyObject *__pyx_kp_u_Initializing_trajectory_using_gr;
static PyObject *__pyx_kp_s_Invalid_mode_expected_c_or_fortr;
static PyObject *__pyx_kp_s_Invalid_shape_in_axis_d_d;
static PyObject *__pyx_n_s_KeyError;
static PyObject *__pyx_kp_u_Loading_groups_from_s;
static PyObject *__pyx_kp_u_Loading_topology_from_s;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_kp_s_MemoryView_of_r_at_0x_x;
static PyObject *__pyx_kp_s_MemoryView_of_r_object;
static PyObject *__pyx_kp_u_None;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_n_b_O;
static PyObject *__pyx_n_s_OPENMP_MAX_THREADS;
static PyObject *__pyx_n_s_OPENMP_NUM_THREADS;
static PyObject *__pyx_kp_s_Out_of_bounds_on_buffer_access_a;
static PyObject *__pyx_n_s_PBCBox;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_kp_s_Pickling_of_struct_members_such;
static PyObject *__pyx_kp_s_Pickling_of_struct_members_such_2;
static PyObject *__pyx_kp_u_Preloading_coordinates_from_s;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_kp_u_There_is_probably_something_wro;
static PyObject *__pyx_n_s_Topology;
static PyObject *__pyx_n_s_TopologyGroup;
static PyObject *__pyx_kp_u_TopologyGroup_s_i_atoms;
static PyObject *__pyx_n_s_TopologyReader;
static PyObject *__pyx_n_s_Trajectory;
static PyObject *__pyx_kp_u_Trajectory_is_not_initialized_co;
static PyObject *__pyx_kp_u_Trajectory_not_initialized;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Unable_to_convert_item_to_object;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_View_MemoryView;
static PyObject *__pyx_kp_u_VmRSS;
static PyObject *__pyx_kp_u_WARNING_Found_more_than_d_tricl;
static PyObject *__pyx_kp_u__22;
static PyObject *__pyx_kp_u__25;
static PyObject *__pyx_kp_u__4;
static PyObject *__pyx_kp_u__7;
static PyObject *__pyx_n_s_a_memview;
static PyObject *__pyx_n_s_allocate_buffer;
static PyObject *__pyx_kp_u_and;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_arange;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_asarray;
static PyObject *__pyx_n_s_atomid;
static PyObject *__pyx_kp_u_atomid_not_valid_i;
static PyObject *__pyx_kp_u_atomid_not_valid_i_max_atomid_i;
static PyObject *__pyx_n_s_atomids;
static PyObject *__pyx_n_u_atomids;
static PyObject *__pyx_n_s_atomname_str;
static PyObject *__pyx_n_s_atoms;
static PyObject *__pyx_n_s_b_memview;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_basename;
static PyObject *__pyx_n_s_be_verbose;
static PyObject *__pyx_n_s_beadids;
static PyObject *__pyx_n_s_begin;
static PyObject *__pyx_n_s_box;
static PyObject *__pyx_n_u_box;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_u_c;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_kp_s_contiguous_and_direct;
static PyObject *__pyx_kp_s_contiguous_and_indirect;
static PyObject *__pyx_n_s_coords_reader;
static PyObject *__pyx_n_s_core_analysis;
static PyObject *__pyx_n_s_cutoff;
static PyObject *__pyx_n_s_cython_version;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_distance2;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_dtype_is_object;
static PyObject *__pyx_n_s_empty;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_s_fatslimlib_core_base;
static PyObject *__pyx_kp_s_fatslimlib_core_base_pyx;
static PyObject *__pyx_n_s_filename;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_float64;
static PyObject *__pyx_n_s_flush;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_fortran;
static PyObject *__pyx_n_u_fortran;
static PyObject *__pyx_n_s_get_aggregates;
static PyObject *__pyx_n_s_get_forcefield_type;
static PyObject *__pyx_n_s_get_membranes;
static PyObject *__pyx_n_s_get_rvdw;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_kp_s_got_differing_extents_in_dimensi;
static PyObject *__pyx_n_u_group;
static PyObject *__pyx_n_s_groupid;
static PyObject *__pyx_n_u_headgroups;
static PyObject *__pyx_n_s_hg_group;
static PyObject *__pyx_kp_b_i;
static PyObject *__pyx_kp_u_i;
static PyObject *__pyx_kp_u_i_frames_preloaded_in_s;
static PyObject *__pyx_kp_u_i_groups_loaded_in_s;
static PyObject *__pyx_kp_u_i_topology_atoms_from_i_residue;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_index;
static PyObject *__pyx_n_s_index_reader;
static PyObject *__pyx_kp_u_instance;
static PyObject *__pyx_n_s_int64;
static PyObject *__pyx_n_s_interacting_group;
static PyObject *__pyx_kp_u_is_not_a;
static PyObject *__pyx_n_s_itemsize;
static PyObject *__pyx_kp_s_itemsize_0_for_cython_array;
static PyObject *__pyx_n_s_leaflet1;
static PyObject *__pyx_n_s_leaflet2;
static PyObject *__pyx_n_s_leaflet_distance2;
static PyObject *__pyx_n_s_load;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_memview;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_n_u_ms;
static PyObject *__pyx_n_s_msg;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_u_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_natoms;
static PyObject *__pyx_n_s_ndim;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_normal_memview;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_nresidues;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_open;
static PyObject *__pyx_n_s_os;
static PyObject *__pyx_n_s_pack;
static PyObject *__pyx_n_s_parent;
static PyObject *__pyx_n_s_path;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_kp_u_proc_self_status;
static PyObject *__pyx_n_u_protein;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_capi;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_getbuffer;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_CoordinateReader;
static PyObject *__pyx_n_s_pyx_unpickle_Enum;
static PyObject *__pyx_n_s_pyx_unpickle_IndexReader;
static PyObject *__pyx_n_s_pyx_unpickle_TopologyReader;
static PyObject *__pyx_n_s_pyx_unpickle_Trajectory;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_u_rb;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_resid;
static PyObject *__pyx_n_u_resid;
static PyObject *__pyx_kp_u_resid_not_valid_i;
static PyObject *__pyx_n_s_residues;
static PyObject *__pyx_n_s_resname;
static PyObject *__pyx_n_u_resname;
static PyObject *__pyx_n_s_resname_str;
static PyObject *__pyx_n_s_retrieve_aggregates;
static PyObject *__pyx_n_s_retrieve_membranes;
static PyObject *__pyx_n_u_s;
static PyObject *__pyx_kp_s_self_atomids_to_internalids_self;
static PyObject *__pyx_kp_s_self_atoms_cannot_be_converted_t;
static PyObject *__pyx_kp_s_self_neighbors_cannot_be_convert;
static PyObject *__pyx_n_s_setdefault;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_u_size;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_startswith;
static PyObject *__pyx_n_s_stdout;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_kp_s_strided_and_direct;
static PyObject *__pyx_kp_s_strided_and_direct_or_indirect;
static PyObject *__pyx_kp_s_strided_and_indirect;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_strip;
static PyObject *__pyx_n_s_strip_name;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_time;
static PyObject *__pyx_n_s_timestep;
static PyObject *__pyx_n_s_tolist;
static PyObject *__pyx_n_s_top_reader;
static PyObject *__pyx_n_s_trajectory;
static PyObject *__pyx_kp_s_unable_to_allocate_array_data;
static PyObject *__pyx_kp_s_unable_to_allocate_shape_and_str;
static PyObject *__pyx_n_u_unk;
static PyObject *__pyx_n_s_unpack;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_u_us;
static PyObject *__pyx_n_s_verbose;
static PyObject *__pyx_n_s_verbose_print;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_s_y;
static PyObject *__pyx_n_s_z;
static PyObject *__pyx_pf_10fatslimlib_9core_base_get_num_threads(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_2get_max_threads(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_4get_num_procs(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_6set_num_threads(CYTHON_UNUSED PyObject *__pyx_self, fsl_int __pyx_v_num); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_8pretty_delta(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_begin, PyObject *__pyx_v_end); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_10pretty_duration(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_duration); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_12get_memory_usage(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_14verbose_print(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_msg, int __pyx_v_verbose, PyObject *__pyx_v_end); /* proto */
static int __pyx_pf_10fatslimlib_9core_base_4Atom___init__(struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_self, fsl_int __pyx_v_atomid, PyObject *__pyx_v_name, real __pyx_v_x, real __pyx_v_y, real __pyx_v_z); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_4Atom_1x___get__(struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_4Atom_1y___get__(struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_4Atom_1z___get__(struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_4Atom_6atomid___get__(struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_4Atom_4name___get__(struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_4Atom_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_4Atom_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10fatslimlib_9core_base_9AtomGroup___cinit__(struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *__pyx_v_self, PyObject *__pyx_v_name, fsl_int __pyx_v_groupid, PyObject *__pyx_v_atoms); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_9AtomGroup_2append(struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *__pyx_v_self, struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_atom); /* proto */
static Py_ssize_t __pyx_pf_10fatslimlib_9core_base_9AtomGroup_4__len__(struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *__pyx_v_self); /* proto */
static void __pyx_pf_10fatslimlib_9core_base_9AtomGroup_6__dealloc__(struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_9AtomGroup_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_9AtomGroup_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10fatslimlib_9core_base_6PBCBox___init__(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_box); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_2update(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_box); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_3box___get__(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_6center___get__(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_11bbox_center___get__(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_4to_gro(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_6__str__(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_8dx(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_a_memview, __Pyx_memviewslice __pyx_v_b_memview); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_10dx_leaflet(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_a_memview, __Pyx_memviewslice __pyx_v_b_memview, __Pyx_memviewslice __pyx_v_normal_memview); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_12distance2(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_a_memview, __Pyx_memviewslice __pyx_v_b_memview); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_14distance(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_a_memview, __Pyx_memviewslice __pyx_v_b_memview); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_16leaflet_distance2(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_a_memview, __Pyx_memviewslice __pyx_v_b_memview, __Pyx_memviewslice __pyx_v_normal_memview); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_18leaflet_distance(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_a_memview, __Pyx_memviewslice __pyx_v_b_memview, __Pyx_memviewslice __pyx_v_normal_memview); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_20get_xcm(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_coords); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_22asarray(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_24put_atoms_in_bbox(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_coords); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_26__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_28__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10fatslimlib_9core_base_13TopologyGroup___init__(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self, struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_parent, PyObject *__pyx_v_resname, fsl_int __pyx_v_resid); /* proto */
static Py_ssize_t __pyx_pf_10fatslimlib_9core_base_13TopologyGroup_2__len__(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self); /* proto */
static void __pyx_pf_10fatslimlib_9core_base_13TopologyGroup_4__dealloc__(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_13TopologyGroup_7atomids___get__(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_13TopologyGroup_15residue_atomids___get__(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_13TopologyGroup_6__str__(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_13TopologyGroup_6parent___get__(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_13TopologyGroup_5resid___get__(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_13TopologyGroup_4size___get__(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_13TopologyGroup_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_13TopologyGroup_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10fatslimlib_9core_base_8Topology___init__(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self); /* proto */
static void __pyx_pf_10fatslimlib_9core_base_8Topology_2__dealloc__(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_8Topology_4append(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_resid, PyObject *__pyx_v_resname_str, PyObject *__pyx_v_atomname_str, fsl_int __pyx_v_atomid, int __pyx_v_strip_name); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_8Topology_6get_atom(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_atomid); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_8Topology_8get_residue_from_atomid(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_atomid); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_8Topology_10get_residue(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, PyObject *__pyx_v_resid); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_8Topology_9nresidues___get__(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_8Topology_6natoms___get__(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_10fatslimlib_9core_base_8Topology_12__len__(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_8Topology_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_8Topology_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10fatslimlib_9core_base_14TopologyReader___init__(struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *__pyx_v_self, PyObject *__pyx_v_filename, int __pyx_v_verbose); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_14TopologyReader_8topology___get__(struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_14TopologyReader_2__reduce_cython__(struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_14TopologyReader_4__setstate_cython__(struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10fatslimlib_9core_base_16CoordinateReader___init__(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_self, PyObject *__pyx_v_filename, int __pyx_v_verbose); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_16CoordinateReader_7nframes___get__(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_16CoordinateReader_9timesteps___get__(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_16CoordinateReader_2__reduce_cython__(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_16CoordinateReader_4__setstate_cython__(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10fatslimlib_9core_base_11IndexReader___init__(struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_self, PyObject *__pyx_v_filename, int __pyx_v_verbose); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_11IndexReader_2load(struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_10fatslimlib_9core_base_11IndexReader_4__len__(struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_11IndexReader_6__getitem__(struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_11IndexReader_6groups___get__(struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_11IndexReader_8__reduce_cython__(struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_11IndexReader_10__setstate_cython__(struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10fatslimlib_9core_base_5Frame___init__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self, struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_trajectory, fsl_int __pyx_v_index, real __pyx_v_timestep); /* proto */
static void __pyx_pf_10fatslimlib_9core_base_5Frame_2__dealloc__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_10fatslimlib_9core_base_5Frame_4__len__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_6get_forcefield_type(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_15hg_group_coords___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_20hg_group_coords_bbox___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_12lipid_coords___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_17lipid_coords_bbox___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_24interacting_group_coords___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_29interacting_group_coords_bbox___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_11bead_coords___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_16bead_coords_bbox___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_10directions___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_7normals___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_8get_aggregates(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self, real __pyx_v_cutoff, int __pyx_v_update); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_10get_membranes(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self, real __pyx_v_cutoff, int __pyx_v_update); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_10hg_atomids___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_13lipid_atomids___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_5index___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_8timestep___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_10trajectory___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_8topology___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_3box___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_4size___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10fatslimlib_9core_base_10Trajectory___init__(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self, struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *__pyx_v_top_reader, struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_index_reader, struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_coords_reader, int __pyx_v_be_verbose); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_10Trajectory_2initialize(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self, PyObject *__pyx_v_hg_group, PyObject *__pyx_v_interacting_group); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_10Trajectory_4get_forcefield_type(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_10Trajectory_6get_rvdw(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_10Trajectory_8__getitem__(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self, fsl_int __pyx_v_item); /* proto */
static Py_ssize_t __pyx_pf_10fatslimlib_9core_base_10Trajectory_10__len__(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_10Trajectory_12get_topology(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_10Trajectory_9timesteps___get__(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_10Trajectory_14__reduce_cython__(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_10Trajectory_16__setstate_cython__(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_16__pyx_unpickle_TopologyReader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_18__pyx_unpickle_CoordinateReader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_20__pyx_unpickle_IndexReader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10fatslimlib_9core_base_22__pyx_unpickle_Trajectory(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object); /* proto */
static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_10fatslimlib_9core_base_Atom(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10fatslimlib_9core_base_AtomGroup(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10fatslimlib_9core_base_PBCBox(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10fatslimlib_9core_base_TopologyGroup(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10fatslimlib_9core_base_Topology(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10fatslimlib_9core_base_TopologyReader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10fatslimlib_9core_base_CoordinateReader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10fatslimlib_9core_base_IndexReader(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10fatslimlib_9core_base_Frame(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10fatslimlib_9core_base_Trajectory(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_setdefault = {0, &__pyx_n_s_setdefault, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyUnicode_Type_strip = {0, &__pyx_n_s_strip, 0, 0, 0};
static PyObject *__pyx_float_0_0;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_6;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_12;
static PyObject *__pyx_int_83902476;
static PyObject *__pyx_int_111280737;
static PyObject *__pyx_int_128740812;
static PyObject *__pyx_int_148569195;
static PyObject *__pyx_int_184977713;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_int_neg_2;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__42;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__56;
static PyObject *__pyx_tuple__57;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__59;
static PyObject *__pyx_tuple__60;
static PyObject *__pyx_tuple__61;
static PyObject *__pyx_codeobj__47;
static PyObject *__pyx_codeobj__49;
static PyObject *__pyx_codeobj__51;
static PyObject *__pyx_codeobj__53;
static PyObject *__pyx_codeobj__55;
static PyObject *__pyx_codeobj__62;
/* Late includes */

/* "fatslimlib/core_base.pyx":92
 * cdef fsl_int OPENMP_NUM_THREADS = OPENMP_MAX_THREADS
 * 
 * cpdef fsl_int get_num_threads():             # <<<<<<<<<<<<<<
 *     return OPENMP_NUM_THREADS
 * 
 */

static PyObject *__pyx_pw_10fatslimlib_9core_base_1get_num_threads(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static fsl_int __pyx_f_10fatslimlib_9core_base_get_num_threads(CYTHON_UNUSED int __pyx_skip_dispatch) {
  fsl_int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_num_threads", 0);

  /* "fatslimlib/core_base.pyx":93
 * 
 * cpdef fsl_int get_num_threads():
 *     return OPENMP_NUM_THREADS             # <<<<<<<<<<<<<<
 * 
 * cpdef fsl_int get_max_threads():
 */
  __pyx_r = __pyx_v_10fatslimlib_9core_base_OPENMP_NUM_THREADS;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":92
 * cdef fsl_int OPENMP_NUM_THREADS = OPENMP_MAX_THREADS
 * 
 * cpdef fsl_int get_num_threads():             # <<<<<<<<<<<<<<
 *     return OPENMP_NUM_THREADS
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_1get_num_threads(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_1get_num_threads(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_num_threads (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_get_num_threads(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_get_num_threads(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_num_threads", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_f_10fatslimlib_9core_base_get_num_threads(0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.get_num_threads", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":95
 *     return OPENMP_NUM_THREADS
 * 
 * cpdef fsl_int get_max_threads():             # <<<<<<<<<<<<<<
 *     return OPENMP_MAX_THREADS
 * 
 */

static PyObject *__pyx_pw_10fatslimlib_9core_base_3get_max_threads(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static fsl_int __pyx_f_10fatslimlib_9core_base_get_max_threads(CYTHON_UNUSED int __pyx_skip_dispatch) {
  fsl_int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_max_threads", 0);

  /* "fatslimlib/core_base.pyx":96
 * 
 * cpdef fsl_int get_max_threads():
 *     return OPENMP_MAX_THREADS             # <<<<<<<<<<<<<<
 * 
 * cpdef fsl_int get_num_procs():
 */
  __pyx_r = __pyx_v_10fatslimlib_9core_base_OPENMP_MAX_THREADS;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":95
 *     return OPENMP_NUM_THREADS
 * 
 * cpdef fsl_int get_max_threads():             # <<<<<<<<<<<<<<
 *     return OPENMP_MAX_THREADS
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_3get_max_threads(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_3get_max_threads(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_max_threads (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_2get_max_threads(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_2get_max_threads(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_max_threads", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_f_10fatslimlib_9core_base_get_max_threads(0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.get_max_threads", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":98
 *     return OPENMP_MAX_THREADS
 * 
 * cpdef fsl_int get_num_procs():             # <<<<<<<<<<<<<<
 *     return omp_get_num_procs()
 * 
 */

static PyObject *__pyx_pw_10fatslimlib_9core_base_5get_num_procs(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static fsl_int __pyx_f_10fatslimlib_9core_base_get_num_procs(CYTHON_UNUSED int __pyx_skip_dispatch) {
  fsl_int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_num_procs", 0);

  /* "fatslimlib/core_base.pyx":99
 * 
 * cpdef fsl_int get_num_procs():
 *     return omp_get_num_procs()             # <<<<<<<<<<<<<<
 * 
 * cpdef set_num_threads(fsl_int num):
 */
  __pyx_r = omp_get_num_procs();
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":98
 *     return OPENMP_MAX_THREADS
 * 
 * cpdef fsl_int get_num_procs():             # <<<<<<<<<<<<<<
 *     return omp_get_num_procs()
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5get_num_procs(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5get_num_procs(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_num_procs (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_4get_num_procs(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_4get_num_procs(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_num_procs", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_f_10fatslimlib_9core_base_get_num_procs(0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.get_num_procs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":101
 *     return omp_get_num_procs()
 * 
 * cpdef set_num_threads(fsl_int num):             # <<<<<<<<<<<<<<
 *     global OPENMP_NUM_THREADS
 *     if not 0 < num <= OPENMP_MAX_THREADS:
 */

static PyObject *__pyx_pw_10fatslimlib_9core_base_7set_num_threads(PyObject *__pyx_self, PyObject *__pyx_arg_num); /*proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_set_num_threads(fsl_int __pyx_v_num, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("set_num_threads", 0);

  /* "fatslimlib/core_base.pyx":103
 * cpdef set_num_threads(fsl_int num):
 *     global OPENMP_NUM_THREADS
 *     if not 0 < num <= OPENMP_MAX_THREADS:             # <<<<<<<<<<<<<<
 *         num = OPENMP_MAX_THREADS
 * 
 */
  __pyx_t_1 = (0 < __pyx_v_num);
  if (__pyx_t_1) {
    __pyx_t_1 = (__pyx_v_num <= __pyx_v_10fatslimlib_9core_base_OPENMP_MAX_THREADS);
  }
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "fatslimlib/core_base.pyx":104
 *     global OPENMP_NUM_THREADS
 *     if not 0 < num <= OPENMP_MAX_THREADS:
 *         num = OPENMP_MAX_THREADS             # <<<<<<<<<<<<<<
 * 
 *     OPENMP_NUM_THREADS = num
 */
    __pyx_v_num = __pyx_v_10fatslimlib_9core_base_OPENMP_MAX_THREADS;

    /* "fatslimlib/core_base.pyx":103
 * cpdef set_num_threads(fsl_int num):
 *     global OPENMP_NUM_THREADS
 *     if not 0 < num <= OPENMP_MAX_THREADS:             # <<<<<<<<<<<<<<
 *         num = OPENMP_MAX_THREADS
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":106
 *         num = OPENMP_MAX_THREADS
 * 
 *     OPENMP_NUM_THREADS = num             # <<<<<<<<<<<<<<
 * 
 *     if OPENMP_NUM_THREADS != num:
 */
  __pyx_v_10fatslimlib_9core_base_OPENMP_NUM_THREADS = __pyx_v_num;

  /* "fatslimlib/core_base.pyx":108
 *     OPENMP_NUM_THREADS = num
 * 
 *     if OPENMP_NUM_THREADS != num:             # <<<<<<<<<<<<<<
 *         raise RuntimeError("Could not change number of OpenMP threads "
 *                            "(requested value: %i - actual value: %i)" % (num, OPENMP_NUM_THREADS))
 */
  __pyx_t_2 = ((__pyx_v_10fatslimlib_9core_base_OPENMP_NUM_THREADS != __pyx_v_num) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "fatslimlib/core_base.pyx":110
 *     if OPENMP_NUM_THREADS != num:
 *         raise RuntimeError("Could not change number of OpenMP threads "
 *                            "(requested value: %i - actual value: %i)" % (num, OPENMP_NUM_THREADS))             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __Pyx_PyInt_From_fsl_int(__pyx_v_num); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyInt_From_fsl_int(__pyx_v_10fatslimlib_9core_base_OPENMP_NUM_THREADS); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = PyUnicode_Format(__pyx_kp_u_Could_not_change_number_of_OpenM, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "fatslimlib/core_base.pyx":109
 * 
 *     if OPENMP_NUM_THREADS != num:
 *         raise RuntimeError("Could not change number of OpenMP threads "             # <<<<<<<<<<<<<<
 *                            "(requested value: %i - actual value: %i)" % (num, OPENMP_NUM_THREADS))
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 109, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":108
 *     OPENMP_NUM_THREADS = num
 * 
 *     if OPENMP_NUM_THREADS != num:             # <<<<<<<<<<<<<<
 *         raise RuntimeError("Could not change number of OpenMP threads "
 *                            "(requested value: %i - actual value: %i)" % (num, OPENMP_NUM_THREADS))
 */
  }

  /* "fatslimlib/core_base.pyx":101
 *     return omp_get_num_procs()
 * 
 * cpdef set_num_threads(fsl_int num):             # <<<<<<<<<<<<<<
 *     global OPENMP_NUM_THREADS
 *     if not 0 < num <= OPENMP_MAX_THREADS:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.set_num_threads", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_7set_num_threads(PyObject *__pyx_self, PyObject *__pyx_arg_num); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_7set_num_threads(PyObject *__pyx_self, PyObject *__pyx_arg_num) {
  fsl_int __pyx_v_num;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_num_threads (wrapper)", 0);
  assert(__pyx_arg_num); {
    __pyx_v_num = __Pyx_PyInt_As_fsl_int(__pyx_arg_num); if (unlikely((__pyx_v_num == ((fsl_int)-1)) && PyErr_Occurred())) __PYX_ERR(0, 101, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.set_num_threads", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6set_num_threads(__pyx_self, ((fsl_int)__pyx_v_num));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_6set_num_threads(CYTHON_UNUSED PyObject *__pyx_self, fsl_int __pyx_v_num) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("set_num_threads", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10fatslimlib_9core_base_set_num_threads(__pyx_v_num, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.set_num_threads", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":114
 * 
 * 
 * cdef real norm2(real[:] a) nogil:             # <<<<<<<<<<<<<<
 *     return a[XX] * a[XX] + a[YY] * a[YY] + a[ZZ] * a[ZZ]
 * 
 */

static real __pyx_f_10fatslimlib_9core_base_norm2(__Pyx_memviewslice __pyx_v_a) {
  real __pyx_r;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;

  /* "fatslimlib/core_base.pyx":115
 * 
 * cdef real norm2(real[:] a) nogil:
 *     return a[XX] * a[XX] + a[YY] * a[YY] + a[ZZ] * a[ZZ]             # <<<<<<<<<<<<<<
 * 
 * cdef void mat_to_memview(matrix src, real[:, ::1] dest) nogil:
 */
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 1;
  __pyx_t_4 = 1;
  __pyx_t_5 = 2;
  __pyx_t_6 = 2;
  __pyx_r = ((((*((real *) ( /* dim=0 */ (__pyx_v_a.data + __pyx_t_1 * __pyx_v_a.strides[0]) ))) * (*((real *) ( /* dim=0 */ (__pyx_v_a.data + __pyx_t_2 * __pyx_v_a.strides[0]) )))) + ((*((real *) ( /* dim=0 */ (__pyx_v_a.data + __pyx_t_3 * __pyx_v_a.strides[0]) ))) * (*((real *) ( /* dim=0 */ (__pyx_v_a.data + __pyx_t_4 * __pyx_v_a.strides[0]) ))))) + ((*((real *) ( /* dim=0 */ (__pyx_v_a.data + __pyx_t_5 * __pyx_v_a.strides[0]) ))) * (*((real *) ( /* dim=0 */ (__pyx_v_a.data + __pyx_t_6 * __pyx_v_a.strides[0]) )))));
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":114
 * 
 * 
 * cdef real norm2(real[:] a) nogil:             # <<<<<<<<<<<<<<
 *     return a[XX] * a[XX] + a[YY] * a[YY] + a[ZZ] * a[ZZ]
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":117
 *     return a[XX] * a[XX] + a[YY] * a[YY] + a[ZZ] * a[ZZ]
 * 
 * cdef void mat_to_memview(matrix src, real[:, ::1] dest) nogil:             # <<<<<<<<<<<<<<
 *     dest[XX, XX] = src[XX][XX]
 *     dest[XX, YY] = src[XX][YY]
 */

static void __pyx_f_10fatslimlib_9core_base_mat_to_memview(real (*__pyx_v_src)[3], __Pyx_memviewslice __pyx_v_dest) {
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;

  /* "fatslimlib/core_base.pyx":118
 * 
 * cdef void mat_to_memview(matrix src, real[:, ::1] dest) nogil:
 *     dest[XX, XX] = src[XX][XX]             # <<<<<<<<<<<<<<
 *     dest[XX, YY] = src[XX][YY]
 *     dest[XX, ZZ] = src[XX][ZZ]
 */
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  *((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_dest.data + __pyx_t_1 * __pyx_v_dest.strides[0]) )) + __pyx_t_2)) )) = ((__pyx_v_src[0])[0]);

  /* "fatslimlib/core_base.pyx":119
 * cdef void mat_to_memview(matrix src, real[:, ::1] dest) nogil:
 *     dest[XX, XX] = src[XX][XX]
 *     dest[XX, YY] = src[XX][YY]             # <<<<<<<<<<<<<<
 *     dest[XX, ZZ] = src[XX][ZZ]
 *     dest[YY, XX] = src[YY][XX]
 */
  __pyx_t_3 = 0;
  __pyx_t_4 = 1;
  *((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_dest.data + __pyx_t_3 * __pyx_v_dest.strides[0]) )) + __pyx_t_4)) )) = ((__pyx_v_src[0])[1]);

  /* "fatslimlib/core_base.pyx":120
 *     dest[XX, XX] = src[XX][XX]
 *     dest[XX, YY] = src[XX][YY]
 *     dest[XX, ZZ] = src[XX][ZZ]             # <<<<<<<<<<<<<<
 *     dest[YY, XX] = src[YY][XX]
 *     dest[YY, YY] = src[YY][YY]
 */
  __pyx_t_5 = 0;
  __pyx_t_6 = 2;
  *((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_dest.data + __pyx_t_5 * __pyx_v_dest.strides[0]) )) + __pyx_t_6)) )) = ((__pyx_v_src[0])[2]);

  /* "fatslimlib/core_base.pyx":121
 *     dest[XX, YY] = src[XX][YY]
 *     dest[XX, ZZ] = src[XX][ZZ]
 *     dest[YY, XX] = src[YY][XX]             # <<<<<<<<<<<<<<
 *     dest[YY, YY] = src[YY][YY]
 *     dest[YY, ZZ] = src[YY][ZZ]
 */
  __pyx_t_7 = 1;
  __pyx_t_8 = 0;
  *((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_dest.data + __pyx_t_7 * __pyx_v_dest.strides[0]) )) + __pyx_t_8)) )) = ((__pyx_v_src[1])[0]);

  /* "fatslimlib/core_base.pyx":122
 *     dest[XX, ZZ] = src[XX][ZZ]
 *     dest[YY, XX] = src[YY][XX]
 *     dest[YY, YY] = src[YY][YY]             # <<<<<<<<<<<<<<
 *     dest[YY, ZZ] = src[YY][ZZ]
 *     dest[ZZ, XX] = src[ZZ][XX]
 */
  __pyx_t_9 = 1;
  __pyx_t_10 = 1;
  *((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_dest.data + __pyx_t_9 * __pyx_v_dest.strides[0]) )) + __pyx_t_10)) )) = ((__pyx_v_src[1])[1]);

  /* "fatslimlib/core_base.pyx":123
 *     dest[YY, XX] = src[YY][XX]
 *     dest[YY, YY] = src[YY][YY]
 *     dest[YY, ZZ] = src[YY][ZZ]             # <<<<<<<<<<<<<<
 *     dest[ZZ, XX] = src[ZZ][XX]
 *     dest[ZZ, YY] = src[ZZ][YY]
 */
  __pyx_t_11 = 1;
  __pyx_t_12 = 2;
  *((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_dest.data + __pyx_t_11 * __pyx_v_dest.strides[0]) )) + __pyx_t_12)) )) = ((__pyx_v_src[1])[2]);

  /* "fatslimlib/core_base.pyx":124
 *     dest[YY, YY] = src[YY][YY]
 *     dest[YY, ZZ] = src[YY][ZZ]
 *     dest[ZZ, XX] = src[ZZ][XX]             # <<<<<<<<<<<<<<
 *     dest[ZZ, YY] = src[ZZ][YY]
 *     dest[ZZ, ZZ] = src[ZZ][ZZ]
 */
  __pyx_t_13 = 2;
  __pyx_t_14 = 0;
  *((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_dest.data + __pyx_t_13 * __pyx_v_dest.strides[0]) )) + __pyx_t_14)) )) = ((__pyx_v_src[2])[0]);

  /* "fatslimlib/core_base.pyx":125
 *     dest[YY, ZZ] = src[YY][ZZ]
 *     dest[ZZ, XX] = src[ZZ][XX]
 *     dest[ZZ, YY] = src[ZZ][YY]             # <<<<<<<<<<<<<<
 *     dest[ZZ, ZZ] = src[ZZ][ZZ]
 * 
 */
  __pyx_t_15 = 2;
  __pyx_t_16 = 1;
  *((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_dest.data + __pyx_t_15 * __pyx_v_dest.strides[0]) )) + __pyx_t_16)) )) = ((__pyx_v_src[2])[1]);

  /* "fatslimlib/core_base.pyx":126
 *     dest[ZZ, XX] = src[ZZ][XX]
 *     dest[ZZ, YY] = src[ZZ][YY]
 *     dest[ZZ, ZZ] = src[ZZ][ZZ]             # <<<<<<<<<<<<<<
 * 
 * cdef void rvec_to_memview(rvec src, real[:] dest) nogil:
 */
  __pyx_t_17 = 2;
  __pyx_t_18 = 2;
  *((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_dest.data + __pyx_t_17 * __pyx_v_dest.strides[0]) )) + __pyx_t_18)) )) = ((__pyx_v_src[2])[2]);

  /* "fatslimlib/core_base.pyx":117
 *     return a[XX] * a[XX] + a[YY] * a[YY] + a[ZZ] * a[ZZ]
 * 
 * cdef void mat_to_memview(matrix src, real[:, ::1] dest) nogil:             # <<<<<<<<<<<<<<
 *     dest[XX, XX] = src[XX][XX]
 *     dest[XX, YY] = src[XX][YY]
 */

  /* function exit code */
}

/* "fatslimlib/core_base.pyx":128
 *     dest[ZZ, ZZ] = src[ZZ][ZZ]
 * 
 * cdef void rvec_to_memview(rvec src, real[:] dest) nogil:             # <<<<<<<<<<<<<<
 *     dest[XX] = src[XX]
 *     dest[YY] = src[YY]
 */

static void __pyx_f_10fatslimlib_9core_base_rvec_to_memview(real *__pyx_v_src, __Pyx_memviewslice __pyx_v_dest) {
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;

  /* "fatslimlib/core_base.pyx":129
 * 
 * cdef void rvec_to_memview(rvec src, real[:] dest) nogil:
 *     dest[XX] = src[XX]             # <<<<<<<<<<<<<<
 *     dest[YY] = src[YY]
 *     dest[ZZ] = src[ZZ]
 */
  __pyx_t_1 = 0;
  *((real *) ( /* dim=0 */ (__pyx_v_dest.data + __pyx_t_1 * __pyx_v_dest.strides[0]) )) = (__pyx_v_src[0]);

  /* "fatslimlib/core_base.pyx":130
 * cdef void rvec_to_memview(rvec src, real[:] dest) nogil:
 *     dest[XX] = src[XX]
 *     dest[YY] = src[YY]             # <<<<<<<<<<<<<<
 *     dest[ZZ] = src[ZZ]
 * 
 */
  __pyx_t_2 = 1;
  *((real *) ( /* dim=0 */ (__pyx_v_dest.data + __pyx_t_2 * __pyx_v_dest.strides[0]) )) = (__pyx_v_src[1]);

  /* "fatslimlib/core_base.pyx":131
 *     dest[XX] = src[XX]
 *     dest[YY] = src[YY]
 *     dest[ZZ] = src[ZZ]             # <<<<<<<<<<<<<<
 * 
 * cpdef pretty_delta(begin, end):
 */
  __pyx_t_3 = 2;
  *((real *) ( /* dim=0 */ (__pyx_v_dest.data + __pyx_t_3 * __pyx_v_dest.strides[0]) )) = (__pyx_v_src[2]);

  /* "fatslimlib/core_base.pyx":128
 *     dest[ZZ, ZZ] = src[ZZ][ZZ]
 * 
 * cdef void rvec_to_memview(rvec src, real[:] dest) nogil:             # <<<<<<<<<<<<<<
 *     dest[XX] = src[XX]
 *     dest[YY] = src[YY]
 */

  /* function exit code */
}

/* "fatslimlib/core_base.pyx":133
 *     dest[ZZ] = src[ZZ]
 * 
 * cpdef pretty_delta(begin, end):             # <<<<<<<<<<<<<<
 *     delta = end - begin
 * 
 */

static PyObject *__pyx_pw_10fatslimlib_9core_base_9pretty_delta(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_pretty_delta(PyObject *__pyx_v_begin, PyObject *__pyx_v_end, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_delta = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pretty_delta", 0);

  /* "fatslimlib/core_base.pyx":134
 * 
 * cpdef pretty_delta(begin, end):
 *     delta = end - begin             # <<<<<<<<<<<<<<
 * 
 *     return pretty_duration(delta)
 */
  __pyx_t_1 = PyNumber_Subtract(__pyx_v_end, __pyx_v_begin); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_delta = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":136
 *     delta = end - begin
 * 
 *     return pretty_duration(delta)             # <<<<<<<<<<<<<<
 * 
 * cpdef pretty_duration(duration):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10fatslimlib_9core_base_pretty_duration(__pyx_v_delta, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":133
 *     dest[ZZ] = src[ZZ]
 * 
 * cpdef pretty_delta(begin, end):             # <<<<<<<<<<<<<<
 *     delta = end - begin
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.pretty_delta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_delta);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_9pretty_delta(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_9pretty_delta(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_begin = 0;
  PyObject *__pyx_v_end = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pretty_delta (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_begin,&__pyx_n_s_end,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_begin)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("pretty_delta", 1, 2, 2, 1); __PYX_ERR(0, 133, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "pretty_delta") < 0)) __PYX_ERR(0, 133, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_begin = values[0];
    __pyx_v_end = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pretty_delta", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 133, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.pretty_delta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_8pretty_delta(__pyx_self, __pyx_v_begin, __pyx_v_end);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_8pretty_delta(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_begin, PyObject *__pyx_v_end) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pretty_delta", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10fatslimlib_9core_base_pretty_delta(__pyx_v_begin, __pyx_v_end, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.pretty_delta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":138
 *     return pretty_duration(delta)
 * 
 * cpdef pretty_duration(duration):             # <<<<<<<<<<<<<<
 *     if duration != 0:
 *         expo = int(floor(log10(duration)))
 */

static PyObject *__pyx_pw_10fatslimlib_9core_base_11pretty_duration(PyObject *__pyx_self, PyObject *__pyx_v_duration); /*proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_pretty_duration(PyObject *__pyx_v_duration, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_expo = NULL;
  PyObject *__pyx_v_unit = NULL;
  PyObject *__pyx_v_ex = NULL;
  PyObject *__pyx_v_unit_str = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  double __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  Py_ssize_t __pyx_t_8;
  Py_UCS4 __pyx_t_9;
  __Pyx_RefNannySetupContext("pretty_duration", 0);

  /* "fatslimlib/core_base.pyx":139
 * 
 * cpdef pretty_duration(duration):
 *     if duration != 0:             # <<<<<<<<<<<<<<
 *         expo = int(floor(log10(duration)))
 *         unit = {-1 : (3, "ms"), -2:(6, "us")}
 */
  __pyx_t_1 = __Pyx_PyInt_NeObjC(__pyx_v_duration, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "fatslimlib/core_base.pyx":140
 * cpdef pretty_duration(duration):
 *     if duration != 0:
 *         expo = int(floor(log10(duration)))             # <<<<<<<<<<<<<<
 *         unit = {-1 : (3, "ms"), -2:(6, "us")}
 *         ex, unit_str = unit.setdefault(expo // 3, (0, "s"))
 */
    __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_v_duration); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 140, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyInt_FromDouble(floor(log10(__pyx_t_3))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_expo = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "fatslimlib/core_base.pyx":141
 *     if duration != 0:
 *         expo = int(floor(log10(duration)))
 *         unit = {-1 : (3, "ms"), -2:(6, "us")}             # <<<<<<<<<<<<<<
 *         ex, unit_str = unit.setdefault(expo // 3, (0, "s"))
 *     else:
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_int_neg_1, __pyx_tuple_) < 0) __PYX_ERR(0, 141, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_1, __pyx_int_neg_2, __pyx_tuple__2) < 0) __PYX_ERR(0, 141, __pyx_L1_error)
    __pyx_v_unit = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "fatslimlib/core_base.pyx":142
 *         expo = int(floor(log10(duration)))
 *         unit = {-1 : (3, "ms"), -2:(6, "us")}
 *         ex, unit_str = unit.setdefault(expo // 3, (0, "s"))             # <<<<<<<<<<<<<<
 *     else:
 *         ex = 0.0
 */
    __pyx_t_1 = __Pyx_PyInt_FloorDivideObjC(__pyx_v_expo, __pyx_int_3, 3, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyDict_SetDefault(__pyx_v_unit, __pyx_t_1, __pyx_tuple__3, -1L); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 142, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 142, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 142, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_6 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 142, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_7 = Py_TYPE(__pyx_t_6)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_1)) goto __pyx_L4_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_5 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_5)) goto __pyx_L4_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 2) < 0) __PYX_ERR(0, 142, __pyx_L1_error)
      __pyx_t_7 = NULL;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L5_unpacking_done;
      __pyx_L4_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 142, __pyx_L1_error)
      __pyx_L5_unpacking_done:;
    }
    __pyx_v_ex = __pyx_t_1;
    __pyx_t_1 = 0;
    __pyx_v_unit_str = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "fatslimlib/core_base.pyx":139
 * 
 * cpdef pretty_duration(duration):
 *     if duration != 0:             # <<<<<<<<<<<<<<
 *         expo = int(floor(log10(duration)))
 *         unit = {-1 : (3, "ms"), -2:(6, "us")}
 */
    goto __pyx_L3;
  }

  /* "fatslimlib/core_base.pyx":144
 *         ex, unit_str = unit.setdefault(expo // 3, (0, "s"))
 *     else:
 *         ex = 0.0             # <<<<<<<<<<<<<<
 *         unit_str = "us"
 *     return u"%.0f %s" % (duration * 10 ** ex, unit_str)
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_float_0_0);
    __pyx_v_ex = __pyx_float_0_0;

    /* "fatslimlib/core_base.pyx":145
 *     else:
 *         ex = 0.0
 *         unit_str = "us"             # <<<<<<<<<<<<<<
 *     return u"%.0f %s" % (duration * 10 ** ex, unit_str)
 * 
 */
    __Pyx_INCREF(__pyx_n_u_us);
    __pyx_v_unit_str = __pyx_n_u_us;
  }
  __pyx_L3:;

  /* "fatslimlib/core_base.pyx":146
 *         ex = 0.0
 *         unit_str = "us"
 *     return u"%.0f %s" % (duration * 10 ** ex, unit_str)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = 0;
  __pyx_t_9 = 127;
  __pyx_t_5 = PyNumber_Power(__pyx_int_10, __pyx_v_ex, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = PyNumber_Multiply(__pyx_v_duration, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Format(__pyx_t_1, __pyx_kp_u_0f); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_9;
  __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u__4);
  __pyx_t_8 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__4);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_kp_u__4);
  __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_unit_str), __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_9;
  __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_4, 3, __pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":138
 *     return pretty_duration(delta)
 * 
 * cpdef pretty_duration(duration):             # <<<<<<<<<<<<<<
 *     if duration != 0:
 *         expo = int(floor(log10(duration)))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("fatslimlib.core_base.pretty_duration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_expo);
  __Pyx_XDECREF(__pyx_v_unit);
  __Pyx_XDECREF(__pyx_v_ex);
  __Pyx_XDECREF(__pyx_v_unit_str);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_11pretty_duration(PyObject *__pyx_self, PyObject *__pyx_v_duration); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_11pretty_duration(PyObject *__pyx_self, PyObject *__pyx_v_duration) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pretty_duration (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_10pretty_duration(__pyx_self, ((PyObject *)__pyx_v_duration));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_10pretty_duration(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_duration) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pretty_duration", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10fatslimlib_9core_base_pretty_duration(__pyx_v_duration, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.pretty_duration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":149
 * 
 * 
 * cpdef real get_memory_usage():             # <<<<<<<<<<<<<<
 *     memory = -1
 *     with open('/proc/self/status', "rb") as fp:
 */

static PyObject *__pyx_pw_10fatslimlib_9core_base_13get_memory_usage(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static real __pyx_f_10fatslimlib_9core_base_get_memory_usage(CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_memory = NULL;
  PyObject *__pyx_v_fp = NULL;
  PyObject *__pyx_v_line = NULL;
  real __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  real __pyx_t_14;
  __Pyx_RefNannySetupContext("get_memory_usage", 0);

  /* "fatslimlib/core_base.pyx":150
 * 
 * cpdef real get_memory_usage():
 *     memory = -1             # <<<<<<<<<<<<<<
 *     with open('/proc/self/status', "rb") as fp:
 *         for line in fp:
 */
  __Pyx_INCREF(__pyx_int_neg_1);
  __pyx_v_memory = __pyx_int_neg_1;

  /* "fatslimlib/core_base.pyx":151
 * cpdef real get_memory_usage():
 *     memory = -1
 *     with open('/proc/self/status', "rb") as fp:             # <<<<<<<<<<<<<<
 *         for line in fp:
 *             if line.startswith("VmRSS:"):
 */
  /*with:*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_exit); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_enter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 151, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 151, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        /*try:*/ {
          __pyx_v_fp = __pyx_t_4;
          __pyx_t_4 = 0;

          /* "fatslimlib/core_base.pyx":152
 *     memory = -1
 *     with open('/proc/self/status', "rb") as fp:
 *         for line in fp:             # <<<<<<<<<<<<<<
 *             if line.startswith("VmRSS:"):
 *                 memory = int(line.split()[1])
 */
          if (likely(PyList_CheckExact(__pyx_v_fp)) || PyTuple_CheckExact(__pyx_v_fp)) {
            __pyx_t_4 = __pyx_v_fp; __Pyx_INCREF(__pyx_t_4); __pyx_t_9 = 0;
            __pyx_t_10 = NULL;
          } else {
            __pyx_t_9 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_fp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 152, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_10 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 152, __pyx_L7_error)
          }
          for (;;) {
            if (likely(!__pyx_t_10)) {
              if (likely(PyList_CheckExact(__pyx_t_4))) {
                if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_4)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_1 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_9); __Pyx_INCREF(__pyx_t_1); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 152, __pyx_L7_error)
                #else
                __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 152, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_1);
                #endif
              } else {
                if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_9); __Pyx_INCREF(__pyx_t_1); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 152, __pyx_L7_error)
                #else
                __pyx_t_1 = PySequence_ITEM(__pyx_t_4, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 152, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_1);
                #endif
              }
            } else {
              __pyx_t_1 = __pyx_t_10(__pyx_t_4);
              if (unlikely(!__pyx_t_1)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 152, __pyx_L7_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_1);
            }
            __Pyx_XDECREF_SET(__pyx_v_line, __pyx_t_1);
            __pyx_t_1 = 0;

            /* "fatslimlib/core_base.pyx":153
 *     with open('/proc/self/status', "rb") as fp:
 *         for line in fp:
 *             if line.startswith("VmRSS:"):             # <<<<<<<<<<<<<<
 *                 memory = int(line.split()[1])
 *                 break
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_startswith); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 153, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_5 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_kp_u_VmRSS) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u_VmRSS);
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 153, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 153, __pyx_L7_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (__pyx_t_11) {

              /* "fatslimlib/core_base.pyx":154
 *         for line in fp:
 *             if line.startswith("VmRSS:"):
 *                 memory = int(line.split()[1])             # <<<<<<<<<<<<<<
 *                 break
 *     return memory
 */
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_split); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 154, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_5 = NULL;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_3, function);
                }
              }
              __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 154, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF_SET(__pyx_v_memory, __pyx_t_1);
              __pyx_t_1 = 0;

              /* "fatslimlib/core_base.pyx":155
 *             if line.startswith("VmRSS:"):
 *                 memory = int(line.split()[1])
 *                 break             # <<<<<<<<<<<<<<
 *     return memory
 * 
 */
              goto __pyx_L14_break;

              /* "fatslimlib/core_base.pyx":153
 *     with open('/proc/self/status', "rb") as fp:
 *         for line in fp:
 *             if line.startswith("VmRSS:"):             # <<<<<<<<<<<<<<
 *                 memory = int(line.split()[1])
 *                 break
 */
            }

            /* "fatslimlib/core_base.pyx":152
 *     memory = -1
 *     with open('/proc/self/status', "rb") as fp:
 *         for line in fp:             # <<<<<<<<<<<<<<
 *             if line.startswith("VmRSS:"):
 *                 memory = int(line.split()[1])
 */
          }
          __pyx_L14_break:;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "fatslimlib/core_base.pyx":151
 * cpdef real get_memory_usage():
 *     memory = -1
 *     with open('/proc/self/status', "rb") as fp:             # <<<<<<<<<<<<<<
 *         for line in fp:
 *             if line.startswith("VmRSS:"):
 */
        }
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("fatslimlib.core_base.get_memory_usage", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_1, &__pyx_t_3) < 0) __PYX_ERR(0, 151, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_5 = PyTuple_Pack(3, __pyx_t_4, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 151, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 151, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (__pyx_t_11 < 0) __PYX_ERR(0, 151, __pyx_L9_except_error)
          __pyx_t_13 = ((!(__pyx_t_11 != 0)) != 0);
          if (__pyx_t_13) {
            __Pyx_GIVEREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_1);
            __Pyx_XGIVEREF(__pyx_t_3);
            __Pyx_ErrRestoreWithState(__pyx_t_4, __pyx_t_1, __pyx_t_3);
            __pyx_t_4 = 0; __pyx_t_1 = 0; __pyx_t_3 = 0; 
            __PYX_ERR(0, 151, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        goto __pyx_L1_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_2) {
          __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__6, NULL);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 151, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L6:;
    }
    goto __pyx_L19;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    goto __pyx_L1_error;
    __pyx_L19:;
  }

  /* "fatslimlib/core_base.pyx":156
 *                 memory = int(line.split()[1])
 *                 break
 *     return memory             # <<<<<<<<<<<<<<
 * 
 * def verbose_print(msg, bint verbose, end="\n"):
 */
  __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_v_memory); if (unlikely((__pyx_t_14 == ((real)-1)) && PyErr_Occurred())) __PYX_ERR(0, 156, __pyx_L1_error)
  __pyx_r = __pyx_t_14;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":149
 * 
 * 
 * cpdef real get_memory_usage():             # <<<<<<<<<<<<<<
 *     memory = -1
 *     with open('/proc/self/status', "rb") as fp:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("fatslimlib.core_base.get_memory_usage", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_memory);
  __Pyx_XDECREF(__pyx_v_fp);
  __Pyx_XDECREF(__pyx_v_line);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_13get_memory_usage(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_13get_memory_usage(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_memory_usage (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_12get_memory_usage(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_12get_memory_usage(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_memory_usage", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_10fatslimlib_9core_base_get_memory_usage(0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.get_memory_usage", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":158
 *     return memory
 * 
 * def verbose_print(msg, bint verbose, end="\n"):             # <<<<<<<<<<<<<<
 *     if verbose:
 *         print(msg, end=end)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_15verbose_print(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_10fatslimlib_9core_base_15verbose_print = {"verbose_print", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10fatslimlib_9core_base_15verbose_print, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10fatslimlib_9core_base_15verbose_print(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_msg = 0;
  int __pyx_v_verbose;
  PyObject *__pyx_v_end = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("verbose_print (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_msg,&__pyx_n_s_verbose,&__pyx_n_s_end,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)__pyx_kp_u__7);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_msg)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_verbose)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("verbose_print", 0, 2, 3, 1); __PYX_ERR(0, 158, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_end);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "verbose_print") < 0)) __PYX_ERR(0, 158, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_msg = values[0];
    __pyx_v_verbose = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_verbose == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 158, __pyx_L3_error)
    __pyx_v_end = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("verbose_print", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 158, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.verbose_print", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_14verbose_print(__pyx_self, __pyx_v_msg, __pyx_v_verbose, __pyx_v_end);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_14verbose_print(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_msg, int __pyx_v_verbose, PyObject *__pyx_v_end) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("verbose_print", 0);

  /* "fatslimlib/core_base.pyx":159
 * 
 * def verbose_print(msg, bint verbose, end="\n"):
 *     if verbose:             # <<<<<<<<<<<<<<
 *         print(msg, end=end)
 *         sys.stdout.flush()
 */
  __pyx_t_1 = (__pyx_v_verbose != 0);
  if (__pyx_t_1) {

    /* "fatslimlib/core_base.pyx":160
 * def verbose_print(msg, bint verbose, end="\n"):
 *     if verbose:
 *         print(msg, end=end)             # <<<<<<<<<<<<<<
 *         sys.stdout.flush()
 * 
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_msg);
    __Pyx_GIVEREF(__pyx_v_msg);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_msg);
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_end, __pyx_v_end) < 0) __PYX_ERR(0, 160, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "fatslimlib/core_base.pyx":161
 *     if verbose:
 *         print(msg, end=end)
 *         sys.stdout.flush()             # <<<<<<<<<<<<<<
 * 
 * cdef class Atom(object):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sys); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_stdout); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flush); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 161, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "fatslimlib/core_base.pyx":159
 * 
 * def verbose_print(msg, bint verbose, end="\n"):
 *     if verbose:             # <<<<<<<<<<<<<<
 *         print(msg, end=end)
 *         sys.stdout.flush()
 */
  }

  /* "fatslimlib/core_base.pyx":158
 *     return memory
 * 
 * def verbose_print(msg, bint verbose, end="\n"):             # <<<<<<<<<<<<<<
 *     if verbose:
 *         print(msg, end=end)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("fatslimlib.core_base.verbose_print", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":165
 * cdef class Atom(object):
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                   fsl_int atomid=-1,
 *                   str name="unk",
 */

/* Python wrapper */
static int __pyx_pw_10fatslimlib_9core_base_4Atom_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10fatslimlib_9core_base_4Atom_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  fsl_int __pyx_v_atomid;
  PyObject *__pyx_v_name = 0;
  real __pyx_v_x;
  real __pyx_v_y;
  real __pyx_v_z;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_atomid,&__pyx_n_s_name,&__pyx_n_s_x,&__pyx_n_s_y,&__pyx_n_s_z,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[1] = ((PyObject*)__pyx_n_u_unk);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_atomid);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_y);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_z);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 165, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_atomid = __Pyx_PyInt_As_fsl_int(values[0]); if (unlikely((__pyx_v_atomid == ((fsl_int)-1)) && PyErr_Occurred())) __PYX_ERR(0, 166, __pyx_L3_error)
    } else {
      __pyx_v_atomid = ((fsl_int)-1);
    }
    __pyx_v_name = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_x = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_x == ((real)-1)) && PyErr_Occurred())) __PYX_ERR(0, 168, __pyx_L3_error)
    } else {
      __pyx_v_x = ((real)0.0);
    }
    if (values[3]) {
      __pyx_v_y = __pyx_PyFloat_AsDouble(values[3]); if (unlikely((__pyx_v_y == ((real)-1)) && PyErr_Occurred())) __PYX_ERR(0, 169, __pyx_L3_error)
    } else {
      __pyx_v_y = ((real)0.0);
    }
    if (values[4]) {
      __pyx_v_z = __pyx_PyFloat_AsDouble(values[4]); if (unlikely((__pyx_v_z == ((real)-1)) && PyErr_Occurred())) __PYX_ERR(0, 170, __pyx_L3_error)
    } else {
      __pyx_v_z = ((real)0.0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 165, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.Atom.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyUnicode_Type), 1, "name", 1))) __PYX_ERR(0, 167, __pyx_L1_error)
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_4Atom___init__(((struct __pyx_obj_10fatslimlib_9core_base_Atom *)__pyx_v_self), __pyx_v_atomid, __pyx_v_name, __pyx_v_x, __pyx_v_y, __pyx_v_z);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10fatslimlib_9core_base_4Atom___init__(struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_self, fsl_int __pyx_v_atomid, PyObject *__pyx_v_name, real __pyx_v_x, real __pyx_v_y, real __pyx_v_z) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "fatslimlib/core_base.pyx":171
 *                   real y=0,
 *                   real z=0):
 *         self.c_atom.atomid = atomid             # <<<<<<<<<<<<<<
 *         self.name = name.encode() # Used to keep at least one ref of the original Python string
 *         self.c_atom.name = self.name
 */
  __pyx_v_self->c_atom.atomid = __pyx_v_atomid;

  /* "fatslimlib/core_base.pyx":172
 *                   real z=0):
 *         self.c_atom.atomid = atomid
 *         self.name = name.encode() # Used to keep at least one ref of the original Python string             # <<<<<<<<<<<<<<
 *         self.c_atom.name = self.name
 *         self.c_atom.coords[XX] = x
 */
  if (unlikely(__pyx_v_name == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 172, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_name, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":173
 *         self.c_atom.atomid = atomid
 *         self.name = name.encode() # Used to keep at least one ref of the original Python string
 *         self.c_atom.name = self.name             # <<<<<<<<<<<<<<
 *         self.c_atom.coords[XX] = x
 *         self.c_atom.coords[YY] = y
 */
  if (unlikely(__pyx_v_self->name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 173, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_self->name); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 173, __pyx_L1_error)
  __pyx_v_self->c_atom.name = __pyx_t_2;

  /* "fatslimlib/core_base.pyx":174
 *         self.name = name.encode() # Used to keep at least one ref of the original Python string
 *         self.c_atom.name = self.name
 *         self.c_atom.coords[XX] = x             # <<<<<<<<<<<<<<
 *         self.c_atom.coords[YY] = y
 *         self.c_atom.coords[ZZ] = z
 */
  (__pyx_v_self->c_atom.coords[0]) = __pyx_v_x;

  /* "fatslimlib/core_base.pyx":175
 *         self.c_atom.name = self.name
 *         self.c_atom.coords[XX] = x
 *         self.c_atom.coords[YY] = y             # <<<<<<<<<<<<<<
 *         self.c_atom.coords[ZZ] = z
 * 
 */
  (__pyx_v_self->c_atom.coords[1]) = __pyx_v_y;

  /* "fatslimlib/core_base.pyx":176
 *         self.c_atom.coords[XX] = x
 *         self.c_atom.coords[YY] = y
 *         self.c_atom.coords[ZZ] = z             # <<<<<<<<<<<<<<
 * 
 *     # Python API
 */
  (__pyx_v_self->c_atom.coords[2]) = __pyx_v_z;

  /* "fatslimlib/core_base.pyx":165
 * cdef class Atom(object):
 * 
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                   fsl_int atomid=-1,
 *                   str name="unk",
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Atom.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":180
 *     # Python API
 *     property x:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.c_atom.coords[XX]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_4Atom_1x_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_4Atom_1x_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_4Atom_1x___get__(((struct __pyx_obj_10fatslimlib_9core_base_Atom *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_4Atom_1x___get__(struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":181
 *     property x:
 *         def __get__(self):
 *             return self.c_atom.coords[XX]             # <<<<<<<<<<<<<<
 * 
 *     property y:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->c_atom.coords[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":180
 *     # Python API
 *     property x:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.c_atom.coords[XX]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Atom.x.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":184
 * 
 *     property y:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.c_atom.coords[YY]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_4Atom_1y_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_4Atom_1y_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_4Atom_1y___get__(((struct __pyx_obj_10fatslimlib_9core_base_Atom *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_4Atom_1y___get__(struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":185
 *     property y:
 *         def __get__(self):
 *             return self.c_atom.coords[YY]             # <<<<<<<<<<<<<<
 * 
 *     property z:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->c_atom.coords[1])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":184
 * 
 *     property y:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.c_atom.coords[YY]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Atom.y.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":188
 * 
 *     property z:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.c_atom.coords[ZZ]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_4Atom_1z_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_4Atom_1z_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_4Atom_1z___get__(((struct __pyx_obj_10fatslimlib_9core_base_Atom *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_4Atom_1z___get__(struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":189
 *     property z:
 *         def __get__(self):
 *             return self.c_atom.coords[ZZ]             # <<<<<<<<<<<<<<
 * 
 *     property atomid:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->c_atom.coords[2])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":188
 * 
 *     property z:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.c_atom.coords[ZZ]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Atom.z.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":192
 * 
 *     property atomid:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.c_atom.atomid
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_4Atom_6atomid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_4Atom_6atomid_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_4Atom_6atomid___get__(((struct __pyx_obj_10fatslimlib_9core_base_Atom *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_4Atom_6atomid___get__(struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":193
 *     property atomid:
 *         def __get__(self):
 *             return self.c_atom.atomid             # <<<<<<<<<<<<<<
 * 
 *     property name:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_v_self->c_atom.atomid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":192
 * 
 *     property atomid:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.c_atom.atomid
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Atom.atomid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":196
 * 
 *     property name:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return str(self.name.decode())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_4Atom_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_4Atom_4name_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_4Atom_4name___get__(((struct __pyx_obj_10fatslimlib_9core_base_Atom *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_4Atom_4name___get__(struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":197
 *     property name:
 *         def __get__(self):
 *             return str(self.name.decode())             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->name == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "decode");
    __PYX_ERR(0, 197, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_decode_bytes(__pyx_v_self->name, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":196
 * 
 *     property name:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return str(self.name.decode())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("fatslimlib.core_base.Atom.name.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("Pickling of struct members such as self.c_atom must be explicitly requested with @auto_pickle(True)")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_4Atom_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_4Atom_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_4Atom_2__reduce_cython__(((struct __pyx_obj_10fatslimlib_9core_base_Atom *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_4Atom_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("Pickling of struct members such as self.c_atom must be explicitly requested with @auto_pickle(True)")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("Pickling of struct members such as self.c_atom must be explicitly requested with @auto_pickle(True)")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("Pickling of struct members such as self.c_atom must be explicitly requested with @auto_pickle(True)")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Atom.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("Pickling of struct members such as self.c_atom must be explicitly requested with @auto_pickle(True)")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("Pickling of struct members such as self.c_atom must be explicitly requested with @auto_pickle(True)")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_4Atom_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_4Atom_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_4Atom_4__setstate_cython__(((struct __pyx_obj_10fatslimlib_9core_base_Atom *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_4Atom_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("Pickling of struct members such as self.c_atom must be explicitly requested with @auto_pickle(True)")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("Pickling of struct members such as self.c_atom must be explicitly requested with @auto_pickle(True)")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("Pickling of struct members such as self.c_atom must be explicitly requested with @auto_pickle(True)")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("Pickling of struct members such as self.c_atom must be explicitly requested with @auto_pickle(True)")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Atom.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":200
 * 
 * 
 * cdef void calculate_lipid_direction(rvec ref_bead,             # <<<<<<<<<<<<<<
 *                                     cPBCBox_t box,
 *                                     real[:,::1] lipid_coords,
 */

static void __pyx_f_10fatslimlib_9core_base_calculate_lipid_direction(real *__pyx_v_ref_bead, struct __pyx_t_10fatslimlib_9core_base_cPBCBox_t __pyx_v_box, __Pyx_memviewslice __pyx_v_lipid_coords, fsl_int __pyx_v_offset_start, fsl_int __pyx_v_offset_end, real *__pyx_v_direction) {
  rvec __pyx_v_xcm;
  struct __pyx_opt_args_10fatslimlib_9core_base_fast_pbc_xcm_from_ref_c __pyx_t_1;

  /* "fatslimlib/core_base.pyx":212
 * 
 *     # Get lipid center of mass
 *     fast_pbc_xcm_from_ref_c(box, lipid_coords, ref_bead, xcm,             # <<<<<<<<<<<<<<
 *                             offset_start, offset_end)
 * 
 */
  __pyx_t_1.__pyx_n = 2;
  __pyx_t_1.first = __pyx_v_offset_start;
  __pyx_t_1.last = __pyx_v_offset_end;
  __pyx_f_10fatslimlib_9core_base_fast_pbc_xcm_from_ref_c(__pyx_v_box, __pyx_v_lipid_coords, __pyx_v_ref_bead, __pyx_v_xcm, &__pyx_t_1); 

  /* "fatslimlib/core_base.pyx":216
 * 
 *     # Compute direction
 *     fast_pbc_dx_c(box, xcm, ref_bead, direction)             # <<<<<<<<<<<<<<
 * 
 *     # Normalize it:
 */
  __pyx_f_10fatslimlib_9core_base_fast_pbc_dx_c(__pyx_v_box, __pyx_v_xcm, __pyx_v_ref_bead, __pyx_v_direction);

  /* "fatslimlib/core_base.pyx":219
 * 
 *     # Normalize it:
 *     rvec_normalize(direction)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  rvec_normalize(__pyx_v_direction);

  /* "fatslimlib/core_base.pyx":200
 * 
 * 
 * cdef void calculate_lipid_direction(rvec ref_bead,             # <<<<<<<<<<<<<<
 *                                     cPBCBox_t box,
 *                                     real[:,::1] lipid_coords,
 */

  /* function exit code */
}

/* "fatslimlib/core_base.pyx":222
 * 
 * 
 * cdef void calculate_neighborhood_normal(fsl_int refid, real[:, ::1] coords,             # <<<<<<<<<<<<<<
 *                                         real[:,::1] directions,
 *                                         ns_neighborhood *neighborhood,
 */

static void __pyx_f_10fatslimlib_9core_base_calculate_neighborhood_normal(fsl_int __pyx_v_refid, __Pyx_memviewslice __pyx_v_coords, __Pyx_memviewslice __pyx_v_directions, struct __pyx_t_10fatslimlib_7core_ns_ns_neighborhood *__pyx_v_neighborhood, struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_box, real *__pyx_v_normal, fsl_int *__pyx_v_useful_neighbors) {
  fsl_int __pyx_v_i;
  fsl_int __pyx_v_nid;
  fsl_int __pyx_v_neighborhood_size;
  fsl_int __pyx_v_useful_size;
  rvec __pyx_v_dx;
  rvec __pyx_v_xcm;
  rvec __pyx_v_neighborhood_direction;
  rvec __pyx_v_ref_direction;
  matrix __pyx_v_cov_mat;
  rvec __pyx_v_tmp_vec;
  rvec __pyx_v_eig_vals;
  matrix __pyx_v_eig_vecs;
  fsl_int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  fsl_int __pyx_t_11;
  fsl_int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  long __pyx_t_25;
  long __pyx_t_26;

  /* "fatslimlib/core_base.pyx":229
 *                                         fsl_int *useful_neighbors) nogil:
 *     cdef fsl_int i, nid
 *     cdef fsl_int neighborhood_size = neighborhood.size, useful_size = 0             # <<<<<<<<<<<<<<
 *     cdef rvec dx, xcm
 *     cdef rvec neighborhood_direction
 */
  __pyx_t_1 = __pyx_v_neighborhood->size;
  __pyx_v_neighborhood_size = __pyx_t_1;
  __pyx_v_useful_size = 0;

  /* "fatslimlib/core_base.pyx":240
 *     # Don't compute anything if less than 3 neighbors
 *     # Assume the normal as the reverse of the lipid directions
 *     if neighborhood_size < 3:             # <<<<<<<<<<<<<<
 *         normal[XX] = directions[refid, XX]
 *         normal[YY] = directions[refid, YY]
 */
  __pyx_t_2 = ((__pyx_v_neighborhood_size < 3) != 0);
  if (__pyx_t_2) {

    /* "fatslimlib/core_base.pyx":241
 *     # Assume the normal as the reverse of the lipid directions
 *     if neighborhood_size < 3:
 *         normal[XX] = directions[refid, XX]             # <<<<<<<<<<<<<<
 *         normal[YY] = directions[refid, YY]
 *         normal[ZZ] = directions[refid, ZZ]
 */
    __pyx_t_3 = __pyx_v_refid;
    __pyx_t_4 = 0;
    (__pyx_v_normal[0]) = (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_directions.data + __pyx_t_3 * __pyx_v_directions.strides[0]) )) + __pyx_t_4)) )));

    /* "fatslimlib/core_base.pyx":242
 *     if neighborhood_size < 3:
 *         normal[XX] = directions[refid, XX]
 *         normal[YY] = directions[refid, YY]             # <<<<<<<<<<<<<<
 *         normal[ZZ] = directions[refid, ZZ]
 *         return
 */
    __pyx_t_5 = __pyx_v_refid;
    __pyx_t_6 = 1;
    (__pyx_v_normal[1]) = (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_directions.data + __pyx_t_5 * __pyx_v_directions.strides[0]) )) + __pyx_t_6)) )));

    /* "fatslimlib/core_base.pyx":243
 *         normal[XX] = directions[refid, XX]
 *         normal[YY] = directions[refid, YY]
 *         normal[ZZ] = directions[refid, ZZ]             # <<<<<<<<<<<<<<
 *         return
 * 
 */
    __pyx_t_7 = __pyx_v_refid;
    __pyx_t_8 = 2;
    (__pyx_v_normal[2]) = (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_directions.data + __pyx_t_7 * __pyx_v_directions.strides[0]) )) + __pyx_t_8)) )));

    /* "fatslimlib/core_base.pyx":244
 *         normal[YY] = directions[refid, YY]
 *         normal[ZZ] = directions[refid, ZZ]
 *         return             # <<<<<<<<<<<<<<
 * 
 *     # Compute center of masses
 */
    goto __pyx_L0;

    /* "fatslimlib/core_base.pyx":240
 *     # Don't compute anything if less than 3 neighbors
 *     # Assume the normal as the reverse of the lipid directions
 *     if neighborhood_size < 3:             # <<<<<<<<<<<<<<
 *         normal[XX] = directions[refid, XX]
 *         normal[YY] = directions[refid, YY]
 */
  }

  /* "fatslimlib/core_base.pyx":247
 * 
 *     # Compute center of masses
 *     rvec_clear(xcm)             # <<<<<<<<<<<<<<
 *     rvec_clear(neighborhood_direction)
 * 
 */
  rvec_clear(__pyx_v_xcm);

  /* "fatslimlib/core_base.pyx":248
 *     # Compute center of masses
 *     rvec_clear(xcm)
 *     rvec_clear(neighborhood_direction)             # <<<<<<<<<<<<<<
 * 
 *     rvec_copy(&directions[refid, XX], ref_direction)
 */
  rvec_clear(__pyx_v_neighborhood_direction);

  /* "fatslimlib/core_base.pyx":250
 *     rvec_clear(neighborhood_direction)
 * 
 *     rvec_copy(&directions[refid, XX], ref_direction)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(neighborhood_size):
 */
  __pyx_t_9 = __pyx_v_refid;
  __pyx_t_10 = 0;
  rvec_copy((&(*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_directions.data + __pyx_t_9 * __pyx_v_directions.strides[0]) )) + __pyx_t_10)) )))), __pyx_v_ref_direction);

  /* "fatslimlib/core_base.pyx":252
 *     rvec_copy(&directions[refid, XX], ref_direction)
 * 
 *     for i in range(neighborhood_size):             # <<<<<<<<<<<<<<
 *         nid = neighborhood.beadids[i]
 * 
 */
  __pyx_t_1 = __pyx_v_neighborhood_size;
  __pyx_t_11 = __pyx_t_1;
  for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_i = __pyx_t_12;

    /* "fatslimlib/core_base.pyx":253
 * 
 *     for i in range(neighborhood_size):
 *         nid = neighborhood.beadids[i]             # <<<<<<<<<<<<<<
 * 
 *         # Only take into account the neighbors that are oriented toward the same direction (< 90)
 */
    __pyx_v_nid = (__pyx_v_neighborhood->beadids[__pyx_v_i]);

    /* "fatslimlib/core_base.pyx":256
 * 
 *         # Only take into account the neighbors that are oriented toward the same direction (< 90)
 *         if rvec_dprod(ref_direction, &directions[nid, XX]) > 0:             # <<<<<<<<<<<<<<
 *             useful_neighbors[useful_size] = nid
 *             useful_size += 1
 */
    __pyx_t_13 = __pyx_v_nid;
    __pyx_t_14 = 0;
    __pyx_t_2 = ((rvec_dprod(__pyx_v_ref_direction, (&(*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_directions.data + __pyx_t_13 * __pyx_v_directions.strides[0]) )) + __pyx_t_14)) ))))) > 0.0) != 0);
    if (__pyx_t_2) {

      /* "fatslimlib/core_base.pyx":257
 *         # Only take into account the neighbors that are oriented toward the same direction (< 90)
 *         if rvec_dprod(ref_direction, &directions[nid, XX]) > 0:
 *             useful_neighbors[useful_size] = nid             # <<<<<<<<<<<<<<
 *             useful_size += 1
 * 
 */
      (__pyx_v_useful_neighbors[__pyx_v_useful_size]) = __pyx_v_nid;

      /* "fatslimlib/core_base.pyx":258
 *         if rvec_dprod(ref_direction, &directions[nid, XX]) > 0:
 *             useful_neighbors[useful_size] = nid
 *             useful_size += 1             # <<<<<<<<<<<<<<
 * 
 *             # HG xcm
 */
      __pyx_v_useful_size = (__pyx_v_useful_size + 1);

      /* "fatslimlib/core_base.pyx":261
 * 
 *             # HG xcm
 *             box.fast_pbc_dx(&coords[refid, XX], &coords[nid, XX], dx)             # <<<<<<<<<<<<<<
 *             rvec_inc(xcm, dx)
 * 
 */
      __pyx_t_15 = __pyx_v_refid;
      __pyx_t_16 = 0;
      __pyx_t_17 = __pyx_v_nid;
      __pyx_t_18 = 0;
      ((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_box->__pyx_vtab)->fast_pbc_dx(__pyx_v_box, (&(*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_15 * __pyx_v_coords.strides[0]) )) + __pyx_t_16)) )))), (&(*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_17 * __pyx_v_coords.strides[0]) )) + __pyx_t_18)) )))), __pyx_v_dx);

      /* "fatslimlib/core_base.pyx":262
 *             # HG xcm
 *             box.fast_pbc_dx(&coords[refid, XX], &coords[nid, XX], dx)
 *             rvec_inc(xcm, dx)             # <<<<<<<<<<<<<<
 * 
 *             rvec_inc(neighborhood_direction, &directions[nid, XX])
 */
      rvec_inc(__pyx_v_xcm, __pyx_v_dx);

      /* "fatslimlib/core_base.pyx":264
 *             rvec_inc(xcm, dx)
 * 
 *             rvec_inc(neighborhood_direction, &directions[nid, XX])             # <<<<<<<<<<<<<<
 * 
 *     # Don't compute anything if less than 3 useful neighbors
 */
      __pyx_t_19 = __pyx_v_nid;
      __pyx_t_20 = 0;
      rvec_inc(__pyx_v_neighborhood_direction, (&(*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_directions.data + __pyx_t_19 * __pyx_v_directions.strides[0]) )) + __pyx_t_20)) )))));

      /* "fatslimlib/core_base.pyx":256
 * 
 *         # Only take into account the neighbors that are oriented toward the same direction (< 90)
 *         if rvec_dprod(ref_direction, &directions[nid, XX]) > 0:             # <<<<<<<<<<<<<<
 *             useful_neighbors[useful_size] = nid
 *             useful_size += 1
 */
    }
  }

  /* "fatslimlib/core_base.pyx":268
 *     # Don't compute anything if less than 3 useful neighbors
 *     # Instead, set it to notset as the flag for further computations
 *     if useful_size < 3:             # <<<<<<<<<<<<<<
 *         normal[XX] = ref_direction[XX]
 *         normal[YY] = ref_direction[YY]
 */
  __pyx_t_2 = ((__pyx_v_useful_size < 3) != 0);
  if (__pyx_t_2) {

    /* "fatslimlib/core_base.pyx":269
 *     # Instead, set it to notset as the flag for further computations
 *     if useful_size < 3:
 *         normal[XX] = ref_direction[XX]             # <<<<<<<<<<<<<<
 *         normal[YY] = ref_direction[YY]
 *         normal[ZZ] = ref_direction[ZZ]
 */
    (__pyx_v_normal[0]) = (__pyx_v_ref_direction[0]);

    /* "fatslimlib/core_base.pyx":270
 *     if useful_size < 3:
 *         normal[XX] = ref_direction[XX]
 *         normal[YY] = ref_direction[YY]             # <<<<<<<<<<<<<<
 *         normal[ZZ] = ref_direction[ZZ]
 *         return
 */
    (__pyx_v_normal[1]) = (__pyx_v_ref_direction[1]);

    /* "fatslimlib/core_base.pyx":271
 *         normal[XX] = ref_direction[XX]
 *         normal[YY] = ref_direction[YY]
 *         normal[ZZ] = ref_direction[ZZ]             # <<<<<<<<<<<<<<
 *         return
 * 
 */
    (__pyx_v_normal[2]) = (__pyx_v_ref_direction[2]);

    /* "fatslimlib/core_base.pyx":272
 *         normal[YY] = ref_direction[YY]
 *         normal[ZZ] = ref_direction[ZZ]
 *         return             # <<<<<<<<<<<<<<
 * 
 *     rvec_smul(1.0/useful_size, xcm, xcm)
 */
    goto __pyx_L0;

    /* "fatslimlib/core_base.pyx":268
 *     # Don't compute anything if less than 3 useful neighbors
 *     # Instead, set it to notset as the flag for further computations
 *     if useful_size < 3:             # <<<<<<<<<<<<<<
 *         normal[XX] = ref_direction[XX]
 *         normal[YY] = ref_direction[YY]
 */
  }

  /* "fatslimlib/core_base.pyx":274
 *         return
 * 
 *     rvec_smul(1.0/useful_size, xcm, xcm)             # <<<<<<<<<<<<<<
 *     rvec_inc(xcm, &coords[refid, XX])
 *     rvec_smul(1.0/useful_size, neighborhood_direction, neighborhood_direction)
 */
  rvec_smul((1.0 / ((double)__pyx_v_useful_size)), __pyx_v_xcm, __pyx_v_xcm);

  /* "fatslimlib/core_base.pyx":275
 * 
 *     rvec_smul(1.0/useful_size, xcm, xcm)
 *     rvec_inc(xcm, &coords[refid, XX])             # <<<<<<<<<<<<<<
 *     rvec_smul(1.0/useful_size, neighborhood_direction, neighborhood_direction)
 * 
 */
  __pyx_t_21 = __pyx_v_refid;
  __pyx_t_22 = 0;
  rvec_inc(__pyx_v_xcm, (&(*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_21 * __pyx_v_coords.strides[0]) )) + __pyx_t_22)) )))));

  /* "fatslimlib/core_base.pyx":276
 *     rvec_smul(1.0/useful_size, xcm, xcm)
 *     rvec_inc(xcm, &coords[refid, XX])
 *     rvec_smul(1.0/useful_size, neighborhood_direction, neighborhood_direction)             # <<<<<<<<<<<<<<
 * 
 *     # Build covariance matrix
 */
  rvec_smul((1.0 / ((double)__pyx_v_useful_size)), __pyx_v_neighborhood_direction, __pyx_v_neighborhood_direction);

  /* "fatslimlib/core_base.pyx":279
 * 
 *     # Build covariance matrix
 *     mat_clear(cov_mat)             # <<<<<<<<<<<<<<
 *     for i in range(useful_size):
 *         nid = useful_neighbors[i]
 */
  mat_clear(__pyx_v_cov_mat);

  /* "fatslimlib/core_base.pyx":280
 *     # Build covariance matrix
 *     mat_clear(cov_mat)
 *     for i in range(useful_size):             # <<<<<<<<<<<<<<
 *         nid = useful_neighbors[i]
 * 
 */
  __pyx_t_1 = __pyx_v_useful_size;
  __pyx_t_11 = __pyx_t_1;
  for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_i = __pyx_t_12;

    /* "fatslimlib/core_base.pyx":281
 *     mat_clear(cov_mat)
 *     for i in range(useful_size):
 *         nid = useful_neighbors[i]             # <<<<<<<<<<<<<<
 * 
 *         # Retrieve neighbor and its image
 */
    __pyx_v_nid = (__pyx_v_useful_neighbors[__pyx_v_i]);

    /* "fatslimlib/core_base.pyx":284
 * 
 *         # Retrieve neighbor and its image
 *         box.fast_pbc_dx(xcm, &coords[nid, XX], tmp_vec)             # <<<<<<<<<<<<<<
 * 
 *         cov_mat[YY][YY] += tmp_vec[YY] * tmp_vec[YY]
 */
    __pyx_t_23 = __pyx_v_nid;
    __pyx_t_24 = 0;
    ((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_box->__pyx_vtab)->fast_pbc_dx(__pyx_v_box, __pyx_v_xcm, (&(*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_23 * __pyx_v_coords.strides[0]) )) + __pyx_t_24)) )))), __pyx_v_tmp_vec);

    /* "fatslimlib/core_base.pyx":286
 *         box.fast_pbc_dx(xcm, &coords[nid, XX], tmp_vec)
 * 
 *         cov_mat[YY][YY] += tmp_vec[YY] * tmp_vec[YY]             # <<<<<<<<<<<<<<
 *         cov_mat[YY][ZZ] += tmp_vec[YY] * tmp_vec[ZZ]
 *         cov_mat[ZZ][ZZ] += tmp_vec[ZZ] * tmp_vec[ZZ]
 */
    __pyx_t_25 = 1;
    __pyx_t_26 = 1;
    ((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) = (((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) + ((__pyx_v_tmp_vec[1]) * (__pyx_v_tmp_vec[1])));

    /* "fatslimlib/core_base.pyx":287
 * 
 *         cov_mat[YY][YY] += tmp_vec[YY] * tmp_vec[YY]
 *         cov_mat[YY][ZZ] += tmp_vec[YY] * tmp_vec[ZZ]             # <<<<<<<<<<<<<<
 *         cov_mat[ZZ][ZZ] += tmp_vec[ZZ] * tmp_vec[ZZ]
 * 
 */
    __pyx_t_25 = 1;
    __pyx_t_26 = 2;
    ((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) = (((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) + ((__pyx_v_tmp_vec[1]) * (__pyx_v_tmp_vec[2])));

    /* "fatslimlib/core_base.pyx":288
 *         cov_mat[YY][YY] += tmp_vec[YY] * tmp_vec[YY]
 *         cov_mat[YY][ZZ] += tmp_vec[YY] * tmp_vec[ZZ]
 *         cov_mat[ZZ][ZZ] += tmp_vec[ZZ] * tmp_vec[ZZ]             # <<<<<<<<<<<<<<
 * 
 *         rvec_smul(tmp_vec[XX], tmp_vec, tmp_vec)
 */
    __pyx_t_25 = 2;
    __pyx_t_26 = 2;
    ((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) = (((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) + ((__pyx_v_tmp_vec[2]) * (__pyx_v_tmp_vec[2])));

    /* "fatslimlib/core_base.pyx":290
 *         cov_mat[ZZ][ZZ] += tmp_vec[ZZ] * tmp_vec[ZZ]
 * 
 *         rvec_smul(tmp_vec[XX], tmp_vec, tmp_vec)             # <<<<<<<<<<<<<<
 * 
 *         cov_mat[XX][XX] += tmp_vec[XX]
 */
    rvec_smul((__pyx_v_tmp_vec[0]), __pyx_v_tmp_vec, __pyx_v_tmp_vec);

    /* "fatslimlib/core_base.pyx":292
 *         rvec_smul(tmp_vec[XX], tmp_vec, tmp_vec)
 * 
 *         cov_mat[XX][XX] += tmp_vec[XX]             # <<<<<<<<<<<<<<
 *         cov_mat[XX][YY] += tmp_vec[YY]
 *         cov_mat[XX][ZZ] += tmp_vec[ZZ]
 */
    __pyx_t_25 = 0;
    __pyx_t_26 = 0;
    ((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) = (((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) + (__pyx_v_tmp_vec[0]));

    /* "fatslimlib/core_base.pyx":293
 * 
 *         cov_mat[XX][XX] += tmp_vec[XX]
 *         cov_mat[XX][YY] += tmp_vec[YY]             # <<<<<<<<<<<<<<
 *         cov_mat[XX][ZZ] += tmp_vec[ZZ]
 *     cov_mat[YY][XX] = cov_mat[XX][YY]
 */
    __pyx_t_25 = 0;
    __pyx_t_26 = 1;
    ((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) = (((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) + (__pyx_v_tmp_vec[1]));

    /* "fatslimlib/core_base.pyx":294
 *         cov_mat[XX][XX] += tmp_vec[XX]
 *         cov_mat[XX][YY] += tmp_vec[YY]
 *         cov_mat[XX][ZZ] += tmp_vec[ZZ]             # <<<<<<<<<<<<<<
 *     cov_mat[YY][XX] = cov_mat[XX][YY]
 *     cov_mat[ZZ][XX] = cov_mat[XX][ZZ]
 */
    __pyx_t_25 = 0;
    __pyx_t_26 = 2;
    ((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) = (((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) + (__pyx_v_tmp_vec[2]));
  }

  /* "fatslimlib/core_base.pyx":295
 *         cov_mat[XX][YY] += tmp_vec[YY]
 *         cov_mat[XX][ZZ] += tmp_vec[ZZ]
 *     cov_mat[YY][XX] = cov_mat[XX][YY]             # <<<<<<<<<<<<<<
 *     cov_mat[ZZ][XX] = cov_mat[XX][ZZ]
 *     cov_mat[ZZ][YY] = cov_mat[YY][ZZ]
 */
  ((__pyx_v_cov_mat[1])[0]) = ((__pyx_v_cov_mat[0])[1]);

  /* "fatslimlib/core_base.pyx":296
 *         cov_mat[XX][ZZ] += tmp_vec[ZZ]
 *     cov_mat[YY][XX] = cov_mat[XX][YY]
 *     cov_mat[ZZ][XX] = cov_mat[XX][ZZ]             # <<<<<<<<<<<<<<
 *     cov_mat[ZZ][YY] = cov_mat[YY][ZZ]
 * 
 */
  ((__pyx_v_cov_mat[2])[0]) = ((__pyx_v_cov_mat[0])[2]);

  /* "fatslimlib/core_base.pyx":297
 *     cov_mat[YY][XX] = cov_mat[XX][YY]
 *     cov_mat[ZZ][XX] = cov_mat[XX][ZZ]
 *     cov_mat[ZZ][YY] = cov_mat[YY][ZZ]             # <<<<<<<<<<<<<<
 * 
 *     cov_mat[XX][XX] /= useful_size
 */
  ((__pyx_v_cov_mat[2])[1]) = ((__pyx_v_cov_mat[1])[2]);

  /* "fatslimlib/core_base.pyx":299
 *     cov_mat[ZZ][YY] = cov_mat[YY][ZZ]
 * 
 *     cov_mat[XX][XX] /= useful_size             # <<<<<<<<<<<<<<
 *     cov_mat[XX][YY] /= useful_size
 *     cov_mat[XX][ZZ] /= useful_size
 */
  __pyx_t_25 = 0;
  __pyx_t_26 = 0;
  ((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) = (((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) / __pyx_v_useful_size);

  /* "fatslimlib/core_base.pyx":300
 * 
 *     cov_mat[XX][XX] /= useful_size
 *     cov_mat[XX][YY] /= useful_size             # <<<<<<<<<<<<<<
 *     cov_mat[XX][ZZ] /= useful_size
 *     cov_mat[YY][XX] /= useful_size
 */
  __pyx_t_25 = 0;
  __pyx_t_26 = 1;
  ((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) = (((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) / __pyx_v_useful_size);

  /* "fatslimlib/core_base.pyx":301
 *     cov_mat[XX][XX] /= useful_size
 *     cov_mat[XX][YY] /= useful_size
 *     cov_mat[XX][ZZ] /= useful_size             # <<<<<<<<<<<<<<
 *     cov_mat[YY][XX] /= useful_size
 *     cov_mat[YY][YY] /= useful_size
 */
  __pyx_t_25 = 0;
  __pyx_t_26 = 2;
  ((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) = (((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) / __pyx_v_useful_size);

  /* "fatslimlib/core_base.pyx":302
 *     cov_mat[XX][YY] /= useful_size
 *     cov_mat[XX][ZZ] /= useful_size
 *     cov_mat[YY][XX] /= useful_size             # <<<<<<<<<<<<<<
 *     cov_mat[YY][YY] /= useful_size
 *     cov_mat[YY][ZZ] /= useful_size
 */
  __pyx_t_25 = 1;
  __pyx_t_26 = 0;
  ((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) = (((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) / __pyx_v_useful_size);

  /* "fatslimlib/core_base.pyx":303
 *     cov_mat[XX][ZZ] /= useful_size
 *     cov_mat[YY][XX] /= useful_size
 *     cov_mat[YY][YY] /= useful_size             # <<<<<<<<<<<<<<
 *     cov_mat[YY][ZZ] /= useful_size
 *     cov_mat[ZZ][XX] /= useful_size
 */
  __pyx_t_25 = 1;
  __pyx_t_26 = 1;
  ((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) = (((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) / __pyx_v_useful_size);

  /* "fatslimlib/core_base.pyx":304
 *     cov_mat[YY][XX] /= useful_size
 *     cov_mat[YY][YY] /= useful_size
 *     cov_mat[YY][ZZ] /= useful_size             # <<<<<<<<<<<<<<
 *     cov_mat[ZZ][XX] /= useful_size
 *     cov_mat[ZZ][YY] /= useful_size
 */
  __pyx_t_25 = 1;
  __pyx_t_26 = 2;
  ((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) = (((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) / __pyx_v_useful_size);

  /* "fatslimlib/core_base.pyx":305
 *     cov_mat[YY][YY] /= useful_size
 *     cov_mat[YY][ZZ] /= useful_size
 *     cov_mat[ZZ][XX] /= useful_size             # <<<<<<<<<<<<<<
 *     cov_mat[ZZ][YY] /= useful_size
 *     cov_mat[ZZ][ZZ] /= useful_size
 */
  __pyx_t_25 = 2;
  __pyx_t_26 = 0;
  ((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) = (((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) / __pyx_v_useful_size);

  /* "fatslimlib/core_base.pyx":306
 *     cov_mat[YY][ZZ] /= useful_size
 *     cov_mat[ZZ][XX] /= useful_size
 *     cov_mat[ZZ][YY] /= useful_size             # <<<<<<<<<<<<<<
 *     cov_mat[ZZ][ZZ] /= useful_size
 * 
 */
  __pyx_t_25 = 2;
  __pyx_t_26 = 1;
  ((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) = (((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) / __pyx_v_useful_size);

  /* "fatslimlib/core_base.pyx":307
 *     cov_mat[ZZ][XX] /= useful_size
 *     cov_mat[ZZ][YY] /= useful_size
 *     cov_mat[ZZ][ZZ] /= useful_size             # <<<<<<<<<<<<<<
 * 
 *     # Get eigenvalues
 */
  __pyx_t_25 = 2;
  __pyx_t_26 = 2;
  ((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) = (((__pyx_v_cov_mat[__pyx_t_25])[__pyx_t_26]) / __pyx_v_useful_size);

  /* "fatslimlib/core_base.pyx":310
 * 
 *     # Get eigenvalues
 *     rvec_clear(eig_vals)             # <<<<<<<<<<<<<<
 *     mat_clear(eig_vecs)
 *     eigen_33_sym(cov_mat, eig_vecs, eig_vals)
 */
  rvec_clear(__pyx_v_eig_vals);

  /* "fatslimlib/core_base.pyx":311
 *     # Get eigenvalues
 *     rvec_clear(eig_vals)
 *     mat_clear(eig_vecs)             # <<<<<<<<<<<<<<
 *     eigen_33_sym(cov_mat, eig_vecs, eig_vals)
 * 
 */
  mat_clear(__pyx_v_eig_vecs);

  /* "fatslimlib/core_base.pyx":312
 *     rvec_clear(eig_vals)
 *     mat_clear(eig_vecs)
 *     eigen_33_sym(cov_mat, eig_vecs, eig_vals)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  eigen_33_sym(__pyx_v_cov_mat, __pyx_v_eig_vecs, __pyx_v_eig_vals);

  /* "fatslimlib/core_base.pyx":316
 * 
 *     # Retrieve the normal from the chosen eigen vector
 *     rvec_clear(normal)             # <<<<<<<<<<<<<<
 *     if rvec_dprod(eig_vecs[0], &neighborhood_direction[XX]) < 0:
 *         rvec_dec(normal, eig_vecs[0])
 */
  rvec_clear(__pyx_v_normal);

  /* "fatslimlib/core_base.pyx":317
 *     # Retrieve the normal from the chosen eigen vector
 *     rvec_clear(normal)
 *     if rvec_dprod(eig_vecs[0], &neighborhood_direction[XX]) < 0:             # <<<<<<<<<<<<<<
 *         rvec_dec(normal, eig_vecs[0])
 *     else:
 */
  __pyx_t_2 = ((rvec_dprod((__pyx_v_eig_vecs[0]), (&(__pyx_v_neighborhood_direction[0]))) < 0.0) != 0);
  if (__pyx_t_2) {

    /* "fatslimlib/core_base.pyx":318
 *     rvec_clear(normal)
 *     if rvec_dprod(eig_vecs[0], &neighborhood_direction[XX]) < 0:
 *         rvec_dec(normal, eig_vecs[0])             # <<<<<<<<<<<<<<
 *     else:
 *         rvec_inc(normal, eig_vecs[0])
 */
    rvec_dec(__pyx_v_normal, (__pyx_v_eig_vecs[0]));

    /* "fatslimlib/core_base.pyx":317
 *     # Retrieve the normal from the chosen eigen vector
 *     rvec_clear(normal)
 *     if rvec_dprod(eig_vecs[0], &neighborhood_direction[XX]) < 0:             # <<<<<<<<<<<<<<
 *         rvec_dec(normal, eig_vecs[0])
 *     else:
 */
    goto __pyx_L10;
  }

  /* "fatslimlib/core_base.pyx":320
 *         rvec_dec(normal, eig_vecs[0])
 *     else:
 *         rvec_inc(normal, eig_vecs[0])             # <<<<<<<<<<<<<<
 *     rvec_normalize(normal)
 * 
 */
  /*else*/ {
    rvec_inc(__pyx_v_normal, (__pyx_v_eig_vecs[0]));
  }
  __pyx_L10:;

  /* "fatslimlib/core_base.pyx":321
 *     else:
 *         rvec_inc(normal, eig_vecs[0])
 *     rvec_normalize(normal)             # <<<<<<<<<<<<<<
 * 
 * cdef class AtomGroup(object):
 */
  rvec_normalize(__pyx_v_normal);

  /* "fatslimlib/core_base.pyx":222
 * 
 * 
 * cdef void calculate_neighborhood_normal(fsl_int refid, real[:, ::1] coords,             # <<<<<<<<<<<<<<
 *                                         real[:,::1] directions,
 *                                         ns_neighborhood *neighborhood,
 */

  /* function exit code */
  __pyx_L0:;
}

/* "fatslimlib/core_base.pyx":326
 * 
 *     # initialization
 *     def __cinit__(self,             # <<<<<<<<<<<<<<
 *                   str name="group",
 *                   fsl_int groupid=0,
 */

/* Python wrapper */
static int __pyx_pw_10fatslimlib_9core_base_9AtomGroup_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10fatslimlib_9core_base_9AtomGroup_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  fsl_int __pyx_v_groupid;
  PyObject *__pyx_v_atoms = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,&__pyx_n_s_groupid,&__pyx_n_s_atoms,0};
    PyObject* values[3] = {0,0,0};
    values[0] = ((PyObject*)__pyx_n_u_group);

    /* "fatslimlib/core_base.pyx":329
 *                   str name="group",
 *                   fsl_int groupid=0,
 *                   object atoms=()):             # <<<<<<<<<<<<<<
 *         # Store name
 *         self.name = name.encode()
 */
    values[2] = ((PyObject *)__pyx_empty_tuple);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_groupid);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_atoms);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 326, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_name = ((PyObject*)values[0]);
    if (values[1]) {
      __pyx_v_groupid = __Pyx_PyInt_As_fsl_int(values[1]); if (unlikely((__pyx_v_groupid == ((fsl_int)-1)) && PyErr_Occurred())) __PYX_ERR(0, 328, __pyx_L3_error)
    } else {
      __pyx_v_groupid = ((fsl_int)0);
    }
    __pyx_v_atoms = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 326, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.AtomGroup.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyUnicode_Type), 1, "name", 1))) __PYX_ERR(0, 327, __pyx_L1_error)
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_9AtomGroup___cinit__(((struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *)__pyx_v_self), __pyx_v_name, __pyx_v_groupid, __pyx_v_atoms);

  /* "fatslimlib/core_base.pyx":326
 * 
 *     # initialization
 *     def __cinit__(self,             # <<<<<<<<<<<<<<
 *                   str name="group",
 *                   fsl_int groupid=0,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10fatslimlib_9core_base_9AtomGroup___cinit__(struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *__pyx_v_self, PyObject *__pyx_v_name, fsl_int __pyx_v_groupid, PyObject *__pyx_v_atoms) {
  PyObject *__pyx_v_atom = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "fatslimlib/core_base.pyx":331
 *                   object atoms=()):
 *         # Store name
 *         self.name = name.encode()             # <<<<<<<<<<<<<<
 *         self.c_atomgroup.name = self.name
 * 
 */
  if (unlikely(__pyx_v_name == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 331, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_name, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":332
 *         # Store name
 *         self.name = name.encode()
 *         self.c_atomgroup.name = self.name             # <<<<<<<<<<<<<<
 * 
 *         # Store group id
 */
  if (unlikely(__pyx_v_self->name == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 332, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_self->name); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 332, __pyx_L1_error)
  __pyx_v_self->c_atomgroup.name = __pyx_t_2;

  /* "fatslimlib/core_base.pyx":335
 * 
 *         # Store group id
 *         self.c_atomgroup.groupid = groupid             # <<<<<<<<<<<<<<
 * 
 *         # Check and store atoms
 */
  __pyx_v_self->c_atomgroup.groupid = __pyx_v_groupid;

  /* "fatslimlib/core_base.pyx":338
 * 
 *         # Check and store atoms
 *         self.c_atomgroup.n_atoms = 0             # <<<<<<<<<<<<<<
 *         self.c_atomgroup.atoms = NULL
 *         for atom in atoms:
 */
  __pyx_v_self->c_atomgroup.n_atoms = 0;

  /* "fatslimlib/core_base.pyx":339
 *         # Check and store atoms
 *         self.c_atomgroup.n_atoms = 0
 *         self.c_atomgroup.atoms = NULL             # <<<<<<<<<<<<<<
 *         for atom in atoms:
 *             self.append(<Atom> atom)
 */
  __pyx_v_self->c_atomgroup.atoms = NULL;

  /* "fatslimlib/core_base.pyx":340
 *         self.c_atomgroup.n_atoms = 0
 *         self.c_atomgroup.atoms = NULL
 *         for atom in atoms:             # <<<<<<<<<<<<<<
 *             self.append(<Atom> atom)
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_atoms)) || PyTuple_CheckExact(__pyx_v_atoms)) {
    __pyx_t_1 = __pyx_v_atoms; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_atoms); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 340, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 340, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 340, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 340, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 340, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_1);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 340, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_atom, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "fatslimlib/core_base.pyx":341
 *         self.c_atomgroup.atoms = NULL
 *         for atom in atoms:
 *             self.append(<Atom> atom)             # <<<<<<<<<<<<<<
 * 
 *     def append(self, Atom atom):
 */
    __pyx_t_6 = __Pyx_PyObject_Append(((PyObject *)__pyx_v_self), __pyx_v_atom); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 341, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":340
 *         self.c_atomgroup.n_atoms = 0
 *         self.c_atomgroup.atoms = NULL
 *         for atom in atoms:             # <<<<<<<<<<<<<<
 *             self.append(<Atom> atom)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":326
 * 
 *     # initialization
 *     def __cinit__(self,             # <<<<<<<<<<<<<<
 *                   str name="group",
 *                   fsl_int groupid=0,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.AtomGroup.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_atom);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":343
 *             self.append(<Atom> atom)
 * 
 *     def append(self, Atom atom):             # <<<<<<<<<<<<<<
 *         if not isinstance(atom, Atom):
 *             raise TypeError("%s is not a %s instance" % (atom, Atom))
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_9AtomGroup_3append(PyObject *__pyx_v_self, PyObject *__pyx_v_atom); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_9AtomGroup_3append(PyObject *__pyx_v_self, PyObject *__pyx_v_atom) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_atom), __pyx_ptype_10fatslimlib_9core_base_Atom, 1, "atom", 0))) __PYX_ERR(0, 343, __pyx_L1_error)
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_9AtomGroup_2append(((struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *)__pyx_v_self), ((struct __pyx_obj_10fatslimlib_9core_base_Atom *)__pyx_v_atom));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_9AtomGroup_2append(struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *__pyx_v_self, struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_atom) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("append", 0);

  /* "fatslimlib/core_base.pyx":344
 * 
 *     def append(self, Atom atom):
 *         if not isinstance(atom, Atom):             # <<<<<<<<<<<<<<
 *             raise TypeError("%s is not a %s instance" % (atom, Atom))
 *         self.fast_append(atom)
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_atom), __pyx_ptype_10fatslimlib_9core_base_Atom); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "fatslimlib/core_base.pyx":345
 *     def append(self, Atom atom):
 *         if not isinstance(atom, Atom):
 *             raise TypeError("%s is not a %s instance" % (atom, Atom))             # <<<<<<<<<<<<<<
 *         self.fast_append(atom)
 * 
 */
    __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 127;
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(((PyObject *)__pyx_v_atom)), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u_is_not_a);
    __pyx_t_4 += 10;
    __Pyx_GIVEREF(__pyx_kp_u_is_not_a);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_kp_u_is_not_a);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(((PyObject *)__pyx_ptype_10fatslimlib_9core_base_Atom)), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_5) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_5;
    __pyx_t_4 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u_instance);
    __pyx_t_4 += 9;
    __Pyx_GIVEREF(__pyx_kp_u_instance);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_kp_u_instance);
    __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_3, 4, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 345, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":344
 * 
 *     def append(self, Atom atom):
 *         if not isinstance(atom, Atom):             # <<<<<<<<<<<<<<
 *             raise TypeError("%s is not a %s instance" % (atom, Atom))
 *         self.fast_append(atom)
 */
  }

  /* "fatslimlib/core_base.pyx":346
 *         if not isinstance(atom, Atom):
 *             raise TypeError("%s is not a %s instance" % (atom, Atom))
 *         self.fast_append(atom)             # <<<<<<<<<<<<<<
 * 
 *     cdef fast_append(self, Atom atom):
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_AtomGroup *)__pyx_v_self->__pyx_vtab)->fast_append(__pyx_v_self, __pyx_v_atom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":343
 *             self.append(<Atom> atom)
 * 
 *     def append(self, Atom atom):             # <<<<<<<<<<<<<<
 *         if not isinstance(atom, Atom):
 *             raise TypeError("%s is not a %s instance" % (atom, Atom))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("fatslimlib.core_base.AtomGroup.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":348
 *         self.fast_append(atom)
 * 
 *     cdef fast_append(self, Atom atom):             # <<<<<<<<<<<<<<
 *         self.c_atomgroup.n_atoms += 1
 *         self.c_atomgroup.atoms = <cAtom_t *> PyMem_Realloc(<void *> self.c_atomgroup.atoms,
 */

static PyObject *__pyx_f_10fatslimlib_9core_base_9AtomGroup_fast_append(struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *__pyx_v_self, struct __pyx_obj_10fatslimlib_9core_base_Atom *__pyx_v_atom) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  struct __pyx_t_10fatslimlib_9core_base_cAtom_t __pyx_t_2;
  __Pyx_RefNannySetupContext("fast_append", 0);

  /* "fatslimlib/core_base.pyx":349
 * 
 *     cdef fast_append(self, Atom atom):
 *         self.c_atomgroup.n_atoms += 1             # <<<<<<<<<<<<<<
 *         self.c_atomgroup.atoms = <cAtom_t *> PyMem_Realloc(<void *> self.c_atomgroup.atoms,
 *                                              self.c_atomgroup.n_atoms * sizeof(cAtom_t))
 */
  __pyx_v_self->c_atomgroup.n_atoms = (__pyx_v_self->c_atomgroup.n_atoms + 1);

  /* "fatslimlib/core_base.pyx":350
 *     cdef fast_append(self, Atom atom):
 *         self.c_atomgroup.n_atoms += 1
 *         self.c_atomgroup.atoms = <cAtom_t *> PyMem_Realloc(<void *> self.c_atomgroup.atoms,             # <<<<<<<<<<<<<<
 *                                              self.c_atomgroup.n_atoms * sizeof(cAtom_t))
 *         if self.c_atomgroup.atoms == NULL:
 */
  __pyx_v_self->c_atomgroup.atoms = ((struct __pyx_t_10fatslimlib_9core_base_cAtom_t *)PyMem_Realloc(((void *)__pyx_v_self->c_atomgroup.atoms), (__pyx_v_self->c_atomgroup.n_atoms * (sizeof(struct __pyx_t_10fatslimlib_9core_base_cAtom_t)))));

  /* "fatslimlib/core_base.pyx":352
 *         self.c_atomgroup.atoms = <cAtom_t *> PyMem_Realloc(<void *> self.c_atomgroup.atoms,
 *                                              self.c_atomgroup.n_atoms * sizeof(cAtom_t))
 *         if self.c_atomgroup.atoms == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 *         self.c_atomgroup.atoms[self.c_atomgroup.n_atoms - 1] = atom.c_atom
 */
  __pyx_t_1 = ((__pyx_v_self->c_atomgroup.atoms == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "fatslimlib/core_base.pyx":353
 *                                              self.c_atomgroup.n_atoms * sizeof(cAtom_t))
 *         if self.c_atomgroup.atoms == NULL:
 *             raise MemoryError             # <<<<<<<<<<<<<<
 *         self.c_atomgroup.atoms[self.c_atomgroup.n_atoms - 1] = atom.c_atom
 * 
 */
    PyErr_NoMemory(); __PYX_ERR(0, 353, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":352
 *         self.c_atomgroup.atoms = <cAtom_t *> PyMem_Realloc(<void *> self.c_atomgroup.atoms,
 *                                              self.c_atomgroup.n_atoms * sizeof(cAtom_t))
 *         if self.c_atomgroup.atoms == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 *         self.c_atomgroup.atoms[self.c_atomgroup.n_atoms - 1] = atom.c_atom
 */
  }

  /* "fatslimlib/core_base.pyx":354
 *         if self.c_atomgroup.atoms == NULL:
 *             raise MemoryError
 *         self.c_atomgroup.atoms[self.c_atomgroup.n_atoms - 1] = atom.c_atom             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_v_atom->c_atom;
  (__pyx_v_self->c_atomgroup.atoms[(__pyx_v_self->c_atomgroup.n_atoms - 1)]) = __pyx_t_2;

  /* "fatslimlib/core_base.pyx":348
 *         self.fast_append(atom)
 * 
 *     cdef fast_append(self, Atom atom):             # <<<<<<<<<<<<<<
 *         self.c_atomgroup.n_atoms += 1
 *         self.c_atomgroup.atoms = <cAtom_t *> PyMem_Realloc(<void *> self.c_atomgroup.atoms,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.AtomGroup.fast_append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":357
 * 
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.c_atomgroup.n_atoms
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_10fatslimlib_9core_base_9AtomGroup_5__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_10fatslimlib_9core_base_9AtomGroup_5__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_9AtomGroup_4__len__(((struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_10fatslimlib_9core_base_9AtomGroup_4__len__(struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "fatslimlib/core_base.pyx":358
 * 
 *     def __len__(self):
 *         return self.c_atomgroup.n_atoms             # <<<<<<<<<<<<<<
 * 
 *     # C destructor
 */
  __pyx_r = __pyx_v_self->c_atomgroup.n_atoms;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":357
 * 
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.c_atomgroup.n_atoms
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":361
 * 
 *     # C destructor
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self.c_atomgroup.atoms)
 * 
 */

/* Python wrapper */
static void __pyx_pw_10fatslimlib_9core_base_9AtomGroup_7__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10fatslimlib_9core_base_9AtomGroup_7__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_10fatslimlib_9core_base_9AtomGroup_6__dealloc__(((struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_10fatslimlib_9core_base_9AtomGroup_6__dealloc__(struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "fatslimlib/core_base.pyx":362
 *     # C destructor
 *     def __dealloc__(self):
 *         PyMem_Free(self.c_atomgroup.atoms)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  PyMem_Free(__pyx_v_self->c_atomgroup.atoms);

  /* "fatslimlib/core_base.pyx":361
 * 
 *     # C destructor
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self.c_atomgroup.atoms)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_9AtomGroup_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_9AtomGroup_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_9AtomGroup_8__reduce_cython__(((struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_9AtomGroup_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.AtomGroup.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_9AtomGroup_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_9AtomGroup_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_9AtomGroup_10__setstate_cython__(((struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_9AtomGroup_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.AtomGroup.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":365
 * 
 * 
 * cdef void fast_pbc_dx_c(cPBCBox_t pbcbox, rvec ref, rvec other, rvec dx) nogil:             # <<<<<<<<<<<<<<
 *         cdef fsl_int i, j
 *         cdef rvec dx_start, trial
 */

static void __pyx_f_10fatslimlib_9core_base_fast_pbc_dx_c(struct __pyx_t_10fatslimlib_9core_base_cPBCBox_t __pyx_v_pbcbox, real *__pyx_v_ref, real *__pyx_v_other, real *__pyx_v_dx) {
  fsl_int __pyx_v_i;
  fsl_int __pyx_v_j;
  fsl_int __pyx_t_1;
  int __pyx_t_2;
  fsl_int __pyx_t_3;
  fsl_int __pyx_t_4;

  /* "fatslimlib/core_base.pyx":369
 *         cdef rvec dx_start, trial
 * 
 *         for i in range(DIM):             # <<<<<<<<<<<<<<
 *             dx[i] = other[i] - ref[i]
 * 
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 3; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "fatslimlib/core_base.pyx":370
 * 
 *         for i in range(DIM):
 *             dx[i] = other[i] - ref[i]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_dx[__pyx_v_i]) = ((__pyx_v_other[__pyx_v_i]) - (__pyx_v_ref[__pyx_v_i]));
  }

  /* "fatslimlib/core_base.pyx":373
 * 
 * 
 *         for i in range (DIM-1, -1, -1):             # <<<<<<<<<<<<<<
 *             while dx[i] > pbcbox.hbox_diag[i]:
 *                 for j in range (i, -1, -1):
 */
  for (__pyx_t_1 = 2; __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "fatslimlib/core_base.pyx":374
 * 
 *         for i in range (DIM-1, -1, -1):
 *             while dx[i] > pbcbox.hbox_diag[i]:             # <<<<<<<<<<<<<<
 *                 for j in range (i, -1, -1):
 *                     dx[j] -= pbcbox.box[i][j]
 */
    while (1) {
      __pyx_t_2 = (((__pyx_v_dx[__pyx_v_i]) > (__pyx_v_pbcbox.hbox_diag[__pyx_v_i])) != 0);
      if (!__pyx_t_2) break;

      /* "fatslimlib/core_base.pyx":375
 *         for i in range (DIM-1, -1, -1):
 *             while dx[i] > pbcbox.hbox_diag[i]:
 *                 for j in range (i, -1, -1):             # <<<<<<<<<<<<<<
 *                     dx[j] -= pbcbox.box[i][j]
 * 
 */
      for (__pyx_t_3 = __pyx_v_i; __pyx_t_3 > -1; __pyx_t_3-=1) {
        __pyx_v_j = __pyx_t_3;

        /* "fatslimlib/core_base.pyx":376
 *             while dx[i] > pbcbox.hbox_diag[i]:
 *                 for j in range (i, -1, -1):
 *                     dx[j] -= pbcbox.box[i][j]             # <<<<<<<<<<<<<<
 * 
 *             while dx[i] <= pbcbox.mhbox_diag[i]:
 */
        __pyx_t_4 = __pyx_v_j;
        (__pyx_v_dx[__pyx_t_4]) = ((__pyx_v_dx[__pyx_t_4]) - ((__pyx_v_pbcbox.box[__pyx_v_i])[__pyx_v_j]));
      }
    }

    /* "fatslimlib/core_base.pyx":378
 *                     dx[j] -= pbcbox.box[i][j]
 * 
 *             while dx[i] <= pbcbox.mhbox_diag[i]:             # <<<<<<<<<<<<<<
 *                 for j in range (i, -1, -1):
 *                     dx[j] += pbcbox.box[i][j]
 */
    while (1) {
      __pyx_t_2 = (((__pyx_v_dx[__pyx_v_i]) <= (__pyx_v_pbcbox.mhbox_diag[__pyx_v_i])) != 0);
      if (!__pyx_t_2) break;

      /* "fatslimlib/core_base.pyx":379
 * 
 *             while dx[i] <= pbcbox.mhbox_diag[i]:
 *                 for j in range (i, -1, -1):             # <<<<<<<<<<<<<<
 *                     dx[j] += pbcbox.box[i][j]
 * 
 */
      for (__pyx_t_3 = __pyx_v_i; __pyx_t_3 > -1; __pyx_t_3-=1) {
        __pyx_v_j = __pyx_t_3;

        /* "fatslimlib/core_base.pyx":380
 *             while dx[i] <= pbcbox.mhbox_diag[i]:
 *                 for j in range (i, -1, -1):
 *                     dx[j] += pbcbox.box[i][j]             # <<<<<<<<<<<<<<
 * 
 * 
 */
        __pyx_t_4 = __pyx_v_j;
        (__pyx_v_dx[__pyx_t_4]) = ((__pyx_v_dx[__pyx_t_4]) + ((__pyx_v_pbcbox.box[__pyx_v_i])[__pyx_v_j]));
      }
    }
  }

  /* "fatslimlib/core_base.pyx":365
 * 
 * 
 * cdef void fast_pbc_dx_c(cPBCBox_t pbcbox, rvec ref, rvec other, rvec dx) nogil:             # <<<<<<<<<<<<<<
 *         cdef fsl_int i, j
 *         cdef rvec dx_start, trial
 */

  /* function exit code */
}

/* "fatslimlib/core_base.pyx":383
 * 
 * 
 * cdef void fast_pbc_xcm_from_ref_c(cPBCBox_t pbcbox, real[:, ::1] coords, rvec ref, rvec xcm,             # <<<<<<<<<<<<<<
 *                                   fsl_int first=0, fsl_int last=NOTSET) nogil:
 *         cdef fsl_int i
 */

static void __pyx_f_10fatslimlib_9core_base_fast_pbc_xcm_from_ref_c(struct __pyx_t_10fatslimlib_9core_base_cPBCBox_t __pyx_v_pbcbox, __Pyx_memviewslice __pyx_v_coords, real *__pyx_v_ref, real *__pyx_v_xcm, struct __pyx_opt_args_10fatslimlib_9core_base_fast_pbc_xcm_from_ref_c *__pyx_optional_args) {
  fsl_int __pyx_v_first = ((fsl_int)0);
  fsl_int __pyx_v_last = ((fsl_int)-12345);
  fsl_int __pyx_v_i;
  fsl_int __pyx_v_size;
  rvec __pyx_v_dx;
  long __pyx_v_d;
  int __pyx_t_1;
  fsl_int __pyx_t_2;
  fsl_int __pyx_t_3;
  fsl_int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  long __pyx_t_7;
  long __pyx_t_8;
  long __pyx_t_9;
  long __pyx_t_10;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_first = __pyx_optional_args->first;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_last = __pyx_optional_args->last;
      }
    }
  }

  /* "fatslimlib/core_base.pyx":390
 *         cdef real[:,::1] bbox_coords
 * 
 *         if first < 0:             # <<<<<<<<<<<<<<
 *             first = 0
 * 
 */
  __pyx_t_1 = ((__pyx_v_first < 0) != 0);
  if (__pyx_t_1) {

    /* "fatslimlib/core_base.pyx":391
 * 
 *         if first < 0:
 *             first = 0             # <<<<<<<<<<<<<<
 * 
 *         if last == NOTSET:
 */
    __pyx_v_first = 0;

    /* "fatslimlib/core_base.pyx":390
 *         cdef real[:,::1] bbox_coords
 * 
 *         if first < 0:             # <<<<<<<<<<<<<<
 *             first = 0
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":393
 *             first = 0
 * 
 *         if last == NOTSET:             # <<<<<<<<<<<<<<
 *             last = coords.shape[0]
 * 
 */
  __pyx_t_1 = ((__pyx_v_last == -12345L) != 0);
  if (__pyx_t_1) {

    /* "fatslimlib/core_base.pyx":394
 * 
 *         if last == NOTSET:
 *             last = coords.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         size = last - first
 */
    __pyx_v_last = (__pyx_v_coords.shape[0]);

    /* "fatslimlib/core_base.pyx":393
 *             first = 0
 * 
 *         if last == NOTSET:             # <<<<<<<<<<<<<<
 *             last = coords.shape[0]
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":396
 *             last = coords.shape[0]
 * 
 *         size = last - first             # <<<<<<<<<<<<<<
 * 
 *         if size < 1:
 */
  __pyx_v_size = (__pyx_v_last - __pyx_v_first);

  /* "fatslimlib/core_base.pyx":398
 *         size = last - first
 * 
 *         if size < 1:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = ((__pyx_v_size < 1) != 0);
  if (__pyx_t_1) {

    /* "fatslimlib/core_base.pyx":399
 * 
 *         if size < 1:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         rvec_copy(ref, xcm)
 */
    goto __pyx_L0;

    /* "fatslimlib/core_base.pyx":398
 *         size = last - first
 * 
 *         if size < 1:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":401
 *             return
 * 
 *         rvec_copy(ref, xcm)             # <<<<<<<<<<<<<<
 *         rvec_smul(size, xcm, xcm)
 * 
 */
  rvec_copy(__pyx_v_ref, __pyx_v_xcm);

  /* "fatslimlib/core_base.pyx":402
 * 
 *         rvec_copy(ref, xcm)
 *         rvec_smul(size, xcm, xcm)             # <<<<<<<<<<<<<<
 * 
 *         # Step 1: Get XCM
 */
  rvec_smul(__pyx_v_size, __pyx_v_xcm, __pyx_v_xcm);

  /* "fatslimlib/core_base.pyx":405
 * 
 *         # Step 1: Get XCM
 *         for i in range(first, last):             # <<<<<<<<<<<<<<
 *             fast_pbc_dx_c(pbcbox, ref, &coords[i, XX], dx)
 *             rvec_inc(xcm, dx)
 */
  __pyx_t_2 = __pyx_v_last;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = __pyx_v_first; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "fatslimlib/core_base.pyx":406
 *         # Step 1: Get XCM
 *         for i in range(first, last):
 *             fast_pbc_dx_c(pbcbox, ref, &coords[i, XX], dx)             # <<<<<<<<<<<<<<
 *             rvec_inc(xcm, dx)
 *         rvec_smul(1.0/size, xcm, xcm)
 */
    __pyx_t_5 = __pyx_v_i;
    __pyx_t_6 = 0;
    __pyx_f_10fatslimlib_9core_base_fast_pbc_dx_c(__pyx_v_pbcbox, __pyx_v_ref, (&(*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_5 * __pyx_v_coords.strides[0]) )) + __pyx_t_6)) )))), __pyx_v_dx);

    /* "fatslimlib/core_base.pyx":407
 *         for i in range(first, last):
 *             fast_pbc_dx_c(pbcbox, ref, &coords[i, XX], dx)
 *             rvec_inc(xcm, dx)             # <<<<<<<<<<<<<<
 *         rvec_smul(1.0/size, xcm, xcm)
 * 
 */
    rvec_inc(__pyx_v_xcm, __pyx_v_dx);
  }

  /* "fatslimlib/core_base.pyx":408
 *             fast_pbc_dx_c(pbcbox, ref, &coords[i, XX], dx)
 *             rvec_inc(xcm, dx)
 *         rvec_smul(1.0/size, xcm, xcm)             # <<<<<<<<<<<<<<
 * 
 *         # Step 2: Make sure it is inside the brick-shaped box
 */
  rvec_smul((1.0 / ((double)__pyx_v_size)), __pyx_v_xcm, __pyx_v_xcm);

  /* "fatslimlib/core_base.pyx":411
 * 
 *         # Step 2: Make sure it is inside the brick-shaped box
 *         for i in range(DIM - 1, -1, -1):             # <<<<<<<<<<<<<<
 *             while xcm[i] < 0:
 *                 for d in range(i+1):
 */
  for (__pyx_t_2 = 2; __pyx_t_2 > -1; __pyx_t_2-=1) {
    __pyx_v_i = __pyx_t_2;

    /* "fatslimlib/core_base.pyx":412
 *         # Step 2: Make sure it is inside the brick-shaped box
 *         for i in range(DIM - 1, -1, -1):
 *             while xcm[i] < 0:             # <<<<<<<<<<<<<<
 *                 for d in range(i+1):
 *                     xcm[d] += pbcbox.box[i][d]
 */
    while (1) {
      __pyx_t_1 = (((__pyx_v_xcm[__pyx_v_i]) < 0.0) != 0);
      if (!__pyx_t_1) break;

      /* "fatslimlib/core_base.pyx":413
 *         for i in range(DIM - 1, -1, -1):
 *             while xcm[i] < 0:
 *                 for d in range(i+1):             # <<<<<<<<<<<<<<
 *                     xcm[d] += pbcbox.box[i][d]
 *             while xcm[i] >= pbcbox.box[i][i]:
 */
      __pyx_t_7 = (__pyx_v_i + 1);
      __pyx_t_8 = __pyx_t_7;
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
        __pyx_v_d = __pyx_t_9;

        /* "fatslimlib/core_base.pyx":414
 *             while xcm[i] < 0:
 *                 for d in range(i+1):
 *                     xcm[d] += pbcbox.box[i][d]             # <<<<<<<<<<<<<<
 *             while xcm[i] >= pbcbox.box[i][i]:
 *                 for d in range(i+1):
 */
        __pyx_t_10 = __pyx_v_d;
        (__pyx_v_xcm[__pyx_t_10]) = ((__pyx_v_xcm[__pyx_t_10]) + ((__pyx_v_pbcbox.box[__pyx_v_i])[__pyx_v_d]));
      }
    }

    /* "fatslimlib/core_base.pyx":415
 *                 for d in range(i+1):
 *                     xcm[d] += pbcbox.box[i][d]
 *             while xcm[i] >= pbcbox.box[i][i]:             # <<<<<<<<<<<<<<
 *                 for d in range(i+1):
 *                     xcm[d] -= pbcbox.box[i][d]
 */
    while (1) {
      __pyx_t_1 = (((__pyx_v_xcm[__pyx_v_i]) >= ((__pyx_v_pbcbox.box[__pyx_v_i])[__pyx_v_i])) != 0);
      if (!__pyx_t_1) break;

      /* "fatslimlib/core_base.pyx":416
 *                     xcm[d] += pbcbox.box[i][d]
 *             while xcm[i] >= pbcbox.box[i][i]:
 *                 for d in range(i+1):             # <<<<<<<<<<<<<<
 *                     xcm[d] -= pbcbox.box[i][d]
 * 
 */
      __pyx_t_7 = (__pyx_v_i + 1);
      __pyx_t_8 = __pyx_t_7;
      for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
        __pyx_v_d = __pyx_t_9;

        /* "fatslimlib/core_base.pyx":417
 *             while xcm[i] >= pbcbox.box[i][i]:
 *                 for d in range(i+1):
 *                     xcm[d] -= pbcbox.box[i][d]             # <<<<<<<<<<<<<<
 * 
 * # noinspection PyNoneFunctionAssignment
 */
        __pyx_t_10 = __pyx_v_d;
        (__pyx_v_xcm[__pyx_t_10]) = ((__pyx_v_xcm[__pyx_t_10]) - ((__pyx_v_pbcbox.box[__pyx_v_i])[__pyx_v_d]));
      }
    }
  }

  /* "fatslimlib/core_base.pyx":383
 * 
 * 
 * cdef void fast_pbc_xcm_from_ref_c(cPBCBox_t pbcbox, real[:, ::1] coords, rvec ref, rvec xcm,             # <<<<<<<<<<<<<<
 *                                   fsl_int first=0, fsl_int last=NOTSET) nogil:
 *         cdef fsl_int i
 */

  /* function exit code */
  __pyx_L0:;
}

/* "fatslimlib/core_base.pyx":422
 * cdef class PBCBox(object):
 * 
 *     def __init__(self, real[:,::1] box):             # <<<<<<<<<<<<<<
 *         self.update(box)
 * 
 */

/* Python wrapper */
static int __pyx_pw_10fatslimlib_9core_base_6PBCBox_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10fatslimlib_9core_base_6PBCBox_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_box = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_box,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_box)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 422, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_box = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_box.memview)) __PYX_ERR(0, 422, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 422, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6PBCBox___init__(((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_v_self), __pyx_v_box);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10fatslimlib_9core_base_6PBCBox___init__(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_box) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "fatslimlib/core_base.pyx":423
 * 
 *     def __init__(self, real[:,::1] box):
 *         self.update(box)             # <<<<<<<<<<<<<<
 * 
 *     cdef void fast_update(self, real[:,::1] box) nogil:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_update); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_box, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":422
 * cdef class PBCBox(object):
 * 
 *     def __init__(self, real[:,::1] box):             # <<<<<<<<<<<<<<
 *         self.update(box)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_box, 1);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":425
 *         self.update(box)
 * 
 *     cdef void fast_update(self, real[:,::1] box) nogil:             # <<<<<<<<<<<<<<
 *         cdef fsl_int i, j, k, d, jc, kc, shift
 *         cdef real d2old, d2new, d2new_c
 */

static void __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_update(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_box) {
  fsl_int __pyx_v_i;
  fsl_int __pyx_v_j;
  fsl_int __pyx_v_k;
  fsl_int __pyx_v_d;
  fsl_int __pyx_v_shift;
  real __pyx_v_d2old;
  real __pyx_v_d2new;
  real __pyx_v_d2new_c;
  rvec __pyx_v_trial;
  rvec __pyx_v_pos;
  fsl_int __pyx_v_ii;
  fsl_int __pyx_v_jj;
  fsl_int __pyx_v_kk;
  fsl_int *__pyx_v_order;
  int __pyx_v_use;
  real __pyx_v_min_hv2;
  real __pyx_v_min_ss;
  long __pyx_v_dd;
  __Pyx_RefNannyDeclarations
  fsl_int __pyx_t_1[5];
  fsl_int __pyx_t_2;
  fsl_int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  fsl_int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  int __pyx_t_27;
  int __pyx_t_28;
  fsl_int __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  Py_ssize_t __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  Py_ssize_t __pyx_t_33;
  Py_ssize_t __pyx_t_34;
  Py_ssize_t __pyx_t_35;
  real __pyx_t_36;
  real __pyx_t_37;
  real __pyx_t_38;
  long __pyx_t_39;
  Py_ssize_t __pyx_t_40;
  Py_ssize_t __pyx_t_41;
  PyObject *__pyx_t_42 = NULL;
  PyObject *__pyx_t_43 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("fast_update", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "fatslimlib/core_base.pyx":426
 * 
 *     cdef void fast_update(self, real[:,::1] box) nogil:
 *         cdef fsl_int i, j, k, d, jc, kc, shift             # <<<<<<<<<<<<<<
 *         cdef real d2old, d2new, d2new_c
 *         cdef rvec trial, pos
 */
  /*try:*/ {

    /* "fatslimlib/core_base.pyx":430
 *         cdef rvec trial, pos
 *         cdef fsl_int ii, jj ,kk
 *         cdef fsl_int *order = [0, -1, 1, -2, 2]             # <<<<<<<<<<<<<<
 *         cdef bint use
 *         cdef real min_hv2, min_ss
 */
    __pyx_t_1[0] = 0;
    __pyx_t_1[1] = -1;
    __pyx_t_1[2] = 1;
    __pyx_t_1[3] = -2;
    __pyx_t_1[4] = 2;
    __pyx_v_order = __pyx_t_1;

    /* "fatslimlib/core_base.pyx":434
 *         cdef real min_hv2, min_ss
 * 
 *         rvec_clear(self.center)             # <<<<<<<<<<<<<<
 *         # Update matrix
 *         for i in range(DIM):
 */
    rvec_clear(__pyx_v_self->center);

    /* "fatslimlib/core_base.pyx":436
 *         rvec_clear(self.center)
 *         # Update matrix
 *         for i in range(DIM):             # <<<<<<<<<<<<<<
 *             for j in range(DIM):
 *                 self.c_pbcbox.box[i][j] = box[i, j]
 */
    for (__pyx_t_2 = 0; __pyx_t_2 < 3; __pyx_t_2+=1) {
      __pyx_v_i = __pyx_t_2;

      /* "fatslimlib/core_base.pyx":437
 *         # Update matrix
 *         for i in range(DIM):
 *             for j in range(DIM):             # <<<<<<<<<<<<<<
 *                 self.c_pbcbox.box[i][j] = box[i, j]
 *                 self.center[j] += 0.5 * box[i, j]
 */
      for (__pyx_t_3 = 0; __pyx_t_3 < 3; __pyx_t_3+=1) {
        __pyx_v_j = __pyx_t_3;

        /* "fatslimlib/core_base.pyx":438
 *         for i in range(DIM):
 *             for j in range(DIM):
 *                 self.c_pbcbox.box[i][j] = box[i, j]             # <<<<<<<<<<<<<<
 *                 self.center[j] += 0.5 * box[i, j]
 *             self.bbox_center[i] = 0.5 *  box[i, i]
 */
        __pyx_t_4 = __pyx_v_i;
        __pyx_t_5 = __pyx_v_j;
        ((__pyx_v_self->c_pbcbox.box[__pyx_v_i])[__pyx_v_j]) = (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_4 * __pyx_v_box.strides[0]) )) + __pyx_t_5)) )));

        /* "fatslimlib/core_base.pyx":439
 *             for j in range(DIM):
 *                 self.c_pbcbox.box[i][j] = box[i, j]
 *                 self.center[j] += 0.5 * box[i, j]             # <<<<<<<<<<<<<<
 *             self.bbox_center[i] = 0.5 *  box[i, i]
 * 
 */
        __pyx_t_6 = __pyx_v_j;
        __pyx_t_7 = __pyx_v_i;
        __pyx_t_8 = __pyx_v_j;
        (__pyx_v_self->center[__pyx_t_6]) = ((__pyx_v_self->center[__pyx_t_6]) + (0.5 * (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_7 * __pyx_v_box.strides[0]) )) + __pyx_t_8)) )))));
      }

      /* "fatslimlib/core_base.pyx":440
 *                 self.c_pbcbox.box[i][j] = box[i, j]
 *                 self.center[j] += 0.5 * box[i, j]
 *             self.bbox_center[i] = 0.5 *  box[i, i]             # <<<<<<<<<<<<<<
 * 
 *         # Update diagonals
 */
      __pyx_t_9 = __pyx_v_i;
      __pyx_t_10 = __pyx_v_i;
      (__pyx_v_self->bbox_center[__pyx_v_i]) = (0.5 * (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_9 * __pyx_v_box.strides[0]) )) + __pyx_t_10)) ))));
    }

    /* "fatslimlib/core_base.pyx":443
 * 
 *         # Update diagonals
 *         for i in range(DIM):             # <<<<<<<<<<<<<<
 *             self.c_pbcbox.fbox_diag[i]  =  box[i, i]
 *             self.c_pbcbox.hbox_diag[i]  =  self.c_pbcbox.fbox_diag[i] * 0.5
 */
    for (__pyx_t_2 = 0; __pyx_t_2 < 3; __pyx_t_2+=1) {
      __pyx_v_i = __pyx_t_2;

      /* "fatslimlib/core_base.pyx":444
 *         # Update diagonals
 *         for i in range(DIM):
 *             self.c_pbcbox.fbox_diag[i]  =  box[i, i]             # <<<<<<<<<<<<<<
 *             self.c_pbcbox.hbox_diag[i]  =  self.c_pbcbox.fbox_diag[i] * 0.5
 *             self.c_pbcbox.mhbox_diag[i] = - self.c_pbcbox.hbox_diag[i]
 */
      __pyx_t_11 = __pyx_v_i;
      __pyx_t_12 = __pyx_v_i;
      (__pyx_v_self->c_pbcbox.fbox_diag[__pyx_v_i]) = (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_11 * __pyx_v_box.strides[0]) )) + __pyx_t_12)) )));

      /* "fatslimlib/core_base.pyx":445
 *         for i in range(DIM):
 *             self.c_pbcbox.fbox_diag[i]  =  box[i, i]
 *             self.c_pbcbox.hbox_diag[i]  =  self.c_pbcbox.fbox_diag[i] * 0.5             # <<<<<<<<<<<<<<
 *             self.c_pbcbox.mhbox_diag[i] = - self.c_pbcbox.hbox_diag[i]
 * 
 */
      (__pyx_v_self->c_pbcbox.hbox_diag[__pyx_v_i]) = ((__pyx_v_self->c_pbcbox.fbox_diag[__pyx_v_i]) * 0.5);

      /* "fatslimlib/core_base.pyx":446
 *             self.c_pbcbox.fbox_diag[i]  =  box[i, i]
 *             self.c_pbcbox.hbox_diag[i]  =  self.c_pbcbox.fbox_diag[i] * 0.5
 *             self.c_pbcbox.mhbox_diag[i] = - self.c_pbcbox.hbox_diag[i]             # <<<<<<<<<<<<<<
 * 
 *         # Update maximum cutoff
 */
      (__pyx_v_self->c_pbcbox.mhbox_diag[__pyx_v_i]) = (-(__pyx_v_self->c_pbcbox.hbox_diag[__pyx_v_i]));
    }

    /* "fatslimlib/core_base.pyx":452
 *         # Physical limitation of the cut-off
 *         # by half the length of the shortest box vector.
 *         min_hv2 = real_min(0.25 * rvec_norm2(&box[XX, XX]), 0.25 * rvec_norm2(&box[YY, XX]))             # <<<<<<<<<<<<<<
 *         min_hv2 = real_min(min_hv2, 0.25 * rvec_norm2(&box[ZZ, XX]))
 * 
 */
    __pyx_t_13 = 0;
    __pyx_t_14 = 0;
    __pyx_t_15 = 1;
    __pyx_t_16 = 0;
    __pyx_v_min_hv2 = real_min((0.25 * rvec_norm2((&(*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_13 * __pyx_v_box.strides[0]) )) + __pyx_t_14)) )))))), (0.25 * rvec_norm2((&(*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_15 * __pyx_v_box.strides[0]) )) + __pyx_t_16)) )))))));

    /* "fatslimlib/core_base.pyx":453
 *         # by half the length of the shortest box vector.
 *         min_hv2 = real_min(0.25 * rvec_norm2(&box[XX, XX]), 0.25 * rvec_norm2(&box[YY, XX]))
 *         min_hv2 = real_min(min_hv2, 0.25 * rvec_norm2(&box[ZZ, XX]))             # <<<<<<<<<<<<<<
 * 
 *         # Limitation to the smallest diagonal element due to optimizations:
 */
    __pyx_t_17 = 2;
    __pyx_t_18 = 0;
    __pyx_v_min_hv2 = real_min(__pyx_v_min_hv2, (0.25 * rvec_norm2((&(*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_17 * __pyx_v_box.strides[0]) )) + __pyx_t_18)) )))))));

    /* "fatslimlib/core_base.pyx":459
 *         # in the grid search and pbc_dx is a lot faster
 *         # than checking all possible combinations.
 *         min_ss = real_min(box[XX, XX], real_min(box[YY, YY]- real_abs(box[ZZ, YY]), box[ZZ, ZZ]))             # <<<<<<<<<<<<<<
 * 
 *         self.c_pbcbox.max_cutoff2 = real_min(min_hv2, min_ss * min_ss)
 */
    __pyx_t_19 = 0;
    __pyx_t_20 = 0;
    __pyx_t_21 = 1;
    __pyx_t_22 = 1;
    __pyx_t_23 = 2;
    __pyx_t_24 = 1;
    __pyx_t_25 = 2;
    __pyx_t_26 = 2;
    __pyx_v_min_ss = real_min((*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_19 * __pyx_v_box.strides[0]) )) + __pyx_t_20)) ))), real_min(((*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_21 * __pyx_v_box.strides[0]) )) + __pyx_t_22)) ))) - real_abs((*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_23 * __pyx_v_box.strides[0]) )) + __pyx_t_24)) ))))), (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_25 * __pyx_v_box.strides[0]) )) + __pyx_t_26)) )))));

    /* "fatslimlib/core_base.pyx":461
 *         min_ss = real_min(box[XX, XX], real_min(box[YY, YY]- real_abs(box[ZZ, YY]), box[ZZ, ZZ]))
 * 
 *         self.c_pbcbox.max_cutoff2 = real_min(min_hv2, min_ss * min_ss)             # <<<<<<<<<<<<<<
 * 
 *         # Update shift vectors
 */
    __pyx_v_self->c_pbcbox.max_cutoff2 = real_min(__pyx_v_min_hv2, (__pyx_v_min_ss * __pyx_v_min_ss));

    /* "fatslimlib/core_base.pyx":464
 * 
 *         # Update shift vectors
 *         self.c_pbcbox.ntric_vec = 0             # <<<<<<<<<<<<<<
 *         # We will only use single shifts, but we will check a few
 *         # more shifts to see if there is a limiting distance
 */
    __pyx_v_self->c_pbcbox.ntric_vec = 0;

    /* "fatslimlib/core_base.pyx":468
 *         # more shifts to see if there is a limiting distance
 *         # above which we can not be sure of the correct distance.
 *         for kk in range(5):             # <<<<<<<<<<<<<<
 *             k = order[kk]
 * 
 */
    for (__pyx_t_2 = 0; __pyx_t_2 < 5; __pyx_t_2+=1) {
      __pyx_v_kk = __pyx_t_2;

      /* "fatslimlib/core_base.pyx":469
 *         # above which we can not be sure of the correct distance.
 *         for kk in range(5):
 *             k = order[kk]             # <<<<<<<<<<<<<<
 * 
 *             for jj in range(5):
 */
      __pyx_v_k = (__pyx_v_order[__pyx_v_kk]);

      /* "fatslimlib/core_base.pyx":471
 *             k = order[kk]
 * 
 *             for jj in range(5):             # <<<<<<<<<<<<<<
 *                 j = order[jj]
 * 
 */
      for (__pyx_t_3 = 0; __pyx_t_3 < 5; __pyx_t_3+=1) {
        __pyx_v_jj = __pyx_t_3;

        /* "fatslimlib/core_base.pyx":472
 * 
 *             for jj in range(5):
 *                 j = order[jj]             # <<<<<<<<<<<<<<
 * 
 *                 for ii in range(5):
 */
        __pyx_v_j = (__pyx_v_order[__pyx_v_jj]);

        /* "fatslimlib/core_base.pyx":474
 *                 j = order[jj]
 * 
 *                 for ii in range(5):             # <<<<<<<<<<<<<<
 *                     i = order[ii]
 * 
 */
        for (__pyx_t_6 = 0; __pyx_t_6 < 5; __pyx_t_6+=1) {
          __pyx_v_ii = __pyx_t_6;

          /* "fatslimlib/core_base.pyx":475
 * 
 *                 for ii in range(5):
 *                     i = order[ii]             # <<<<<<<<<<<<<<
 * 
 *                     # A shift is only useful when it is trilinic
 */
          __pyx_v_i = (__pyx_v_order[__pyx_v_ii]);

          /* "fatslimlib/core_base.pyx":478
 * 
 *                     # A shift is only useful when it is trilinic
 *                     if j != 0 or k != 0:             # <<<<<<<<<<<<<<
 *                         d2old = 0
 *                         d2new = 0
 */
          __pyx_t_28 = ((__pyx_v_j != 0) != 0);
          if (!__pyx_t_28) {
          } else {
            __pyx_t_27 = __pyx_t_28;
            goto __pyx_L19_bool_binop_done;
          }
          __pyx_t_28 = ((__pyx_v_k != 0) != 0);
          __pyx_t_27 = __pyx_t_28;
          __pyx_L19_bool_binop_done:;
          if (__pyx_t_27) {

            /* "fatslimlib/core_base.pyx":479
 *                     # A shift is only useful when it is trilinic
 *                     if j != 0 or k != 0:
 *                         d2old = 0             # <<<<<<<<<<<<<<
 *                         d2new = 0
 * 
 */
            __pyx_v_d2old = 0.0;

            /* "fatslimlib/core_base.pyx":480
 *                     if j != 0 or k != 0:
 *                         d2old = 0
 *                         d2new = 0             # <<<<<<<<<<<<<<
 * 
 *                         for d in range(DIM):
 */
            __pyx_v_d2new = 0.0;

            /* "fatslimlib/core_base.pyx":482
 *                         d2new = 0
 * 
 *                         for d in range(DIM):             # <<<<<<<<<<<<<<
 *                             trial[d] = i*box[XX, d] + j*box[YY, d] + k*box[ZZ, d]
 * 
 */
            for (__pyx_t_29 = 0; __pyx_t_29 < 3; __pyx_t_29+=1) {
              __pyx_v_d = __pyx_t_29;

              /* "fatslimlib/core_base.pyx":483
 * 
 *                         for d in range(DIM):
 *                             trial[d] = i*box[XX, d] + j*box[YY, d] + k*box[ZZ, d]             # <<<<<<<<<<<<<<
 * 
 *                             # Choose the vector within the brick around 0,0,0 that
 */
              __pyx_t_30 = 0;
              __pyx_t_31 = __pyx_v_d;
              __pyx_t_32 = 1;
              __pyx_t_33 = __pyx_v_d;
              __pyx_t_34 = 2;
              __pyx_t_35 = __pyx_v_d;
              (__pyx_v_trial[__pyx_v_d]) = (((__pyx_v_i * (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_30 * __pyx_v_box.strides[0]) )) + __pyx_t_31)) )))) + (__pyx_v_j * (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_32 * __pyx_v_box.strides[0]) )) + __pyx_t_33)) ))))) + (__pyx_v_k * (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_34 * __pyx_v_box.strides[0]) )) + __pyx_t_35)) )))));

              /* "fatslimlib/core_base.pyx":488
 *                             # will become the shortest due to shift try.
 * 
 *                             if d == DIM:             # <<<<<<<<<<<<<<
 *                                 trial[d] = 0
 *                                 pos[d] = 0
 */
              __pyx_t_27 = ((__pyx_v_d == 3) != 0);
              if (__pyx_t_27) {

                /* "fatslimlib/core_base.pyx":489
 * 
 *                             if d == DIM:
 *                                 trial[d] = 0             # <<<<<<<<<<<<<<
 *                                 pos[d] = 0
 *                             else:
 */
                (__pyx_v_trial[__pyx_v_d]) = 0.0;

                /* "fatslimlib/core_base.pyx":490
 *                             if d == DIM:
 *                                 trial[d] = 0
 *                                 pos[d] = 0             # <<<<<<<<<<<<<<
 *                             else:
 *                                 if trial[d] < 0:
 */
                (__pyx_v_pos[__pyx_v_d]) = 0.0;

                /* "fatslimlib/core_base.pyx":488
 *                             # will become the shortest due to shift try.
 * 
 *                             if d == DIM:             # <<<<<<<<<<<<<<
 *                                 trial[d] = 0
 *                                 pos[d] = 0
 */
                goto __pyx_L23;
              }

              /* "fatslimlib/core_base.pyx":492
 *                                 pos[d] = 0
 *                             else:
 *                                 if trial[d] < 0:             # <<<<<<<<<<<<<<
 *                                     pos[d] = real_min(self.c_pbcbox.hbox_diag[d], -trial[d])
 *                                 else:
 */
              /*else*/ {
                __pyx_t_27 = (((__pyx_v_trial[__pyx_v_d]) < 0.0) != 0);
                if (__pyx_t_27) {

                  /* "fatslimlib/core_base.pyx":493
 *                             else:
 *                                 if trial[d] < 0:
 *                                     pos[d] = real_min(self.c_pbcbox.hbox_diag[d], -trial[d])             # <<<<<<<<<<<<<<
 *                                 else:
 *                                     pos[d] = max(-self.c_pbcbox.hbox_diag[d], -trial[d])
 */
                  (__pyx_v_pos[__pyx_v_d]) = real_min((__pyx_v_self->c_pbcbox.hbox_diag[__pyx_v_d]), (-(__pyx_v_trial[__pyx_v_d])));

                  /* "fatslimlib/core_base.pyx":492
 *                                 pos[d] = 0
 *                             else:
 *                                 if trial[d] < 0:             # <<<<<<<<<<<<<<
 *                                     pos[d] = real_min(self.c_pbcbox.hbox_diag[d], -trial[d])
 *                                 else:
 */
                  goto __pyx_L24;
                }

                /* "fatslimlib/core_base.pyx":495
 *                                     pos[d] = real_min(self.c_pbcbox.hbox_diag[d], -trial[d])
 *                                 else:
 *                                     pos[d] = max(-self.c_pbcbox.hbox_diag[d], -trial[d])             # <<<<<<<<<<<<<<
 * 
 *                             d2old += sqrt(pos[d])
 */
                /*else*/ {
                  __pyx_t_36 = (-(__pyx_v_trial[__pyx_v_d]));
                  __pyx_t_37 = (-(__pyx_v_self->c_pbcbox.hbox_diag[__pyx_v_d]));
                  if (((__pyx_t_36 > __pyx_t_37) != 0)) {
                    __pyx_t_38 = __pyx_t_36;
                  } else {
                    __pyx_t_38 = __pyx_t_37;
                  }
                  (__pyx_v_pos[__pyx_v_d]) = __pyx_t_38;
                }
                __pyx_L24:;
              }
              __pyx_L23:;

              /* "fatslimlib/core_base.pyx":497
 *                                     pos[d] = max(-self.c_pbcbox.hbox_diag[d], -trial[d])
 * 
 *                             d2old += sqrt(pos[d])             # <<<<<<<<<<<<<<
 *                             d2new += sqrt(pos[d] + trial[d])
 * 
 */
              __pyx_v_d2old = (__pyx_v_d2old + sqrt((__pyx_v_pos[__pyx_v_d])));

              /* "fatslimlib/core_base.pyx":498
 * 
 *                             d2old += sqrt(pos[d])
 *                             d2new += sqrt(pos[d] + trial[d])             # <<<<<<<<<<<<<<
 * 
 *                         if BOX_MARGIN*d2new < d2old:
 */
              __pyx_v_d2new = (__pyx_v_d2new + sqrt(((__pyx_v_pos[__pyx_v_d]) + (__pyx_v_trial[__pyx_v_d]))));
            }

            /* "fatslimlib/core_base.pyx":500
 *                             d2new += sqrt(pos[d] + trial[d])
 * 
 *                         if BOX_MARGIN*d2new < d2old:             # <<<<<<<<<<<<<<
 *                             if  not (j < -1 or j > 1 or k < -1 or k > 1):
 *                                 use = True
 */
            __pyx_t_27 = (((1.001 * __pyx_v_d2new) < __pyx_v_d2old) != 0);
            if (__pyx_t_27) {

              /* "fatslimlib/core_base.pyx":501
 * 
 *                         if BOX_MARGIN*d2new < d2old:
 *                             if  not (j < -1 or j > 1 or k < -1 or k > 1):             # <<<<<<<<<<<<<<
 *                                 use = True
 * 
 */
              __pyx_t_28 = ((__pyx_v_j < -1L) != 0);
              if (!__pyx_t_28) {
              } else {
                __pyx_t_27 = __pyx_t_28;
                goto __pyx_L27_bool_binop_done;
              }
              __pyx_t_28 = ((__pyx_v_j > 1) != 0);
              if (!__pyx_t_28) {
              } else {
                __pyx_t_27 = __pyx_t_28;
                goto __pyx_L27_bool_binop_done;
              }
              __pyx_t_28 = ((__pyx_v_k < -1L) != 0);
              if (!__pyx_t_28) {
              } else {
                __pyx_t_27 = __pyx_t_28;
                goto __pyx_L27_bool_binop_done;
              }
              __pyx_t_28 = ((__pyx_v_k > 1) != 0);
              __pyx_t_27 = __pyx_t_28;
              __pyx_L27_bool_binop_done:;
              __pyx_t_28 = ((!__pyx_t_27) != 0);
              if (__pyx_t_28) {

                /* "fatslimlib/core_base.pyx":502
 *                         if BOX_MARGIN*d2new < d2old:
 *                             if  not (j < -1 or j > 1 or k < -1 or k > 1):
 *                                 use = True             # <<<<<<<<<<<<<<
 * 
 *                                 for dd in range(DIM):
 */
                __pyx_v_use = 1;

                /* "fatslimlib/core_base.pyx":504
 *                                 use = True
 * 
 *                                 for dd in range(DIM):             # <<<<<<<<<<<<<<
 *                                     if dd == 0:
 *                                         shift = i
 */
                for (__pyx_t_39 = 0; __pyx_t_39 < 3; __pyx_t_39+=1) {
                  __pyx_v_dd = __pyx_t_39;

                  /* "fatslimlib/core_base.pyx":505
 * 
 *                                 for dd in range(DIM):
 *                                     if dd == 0:             # <<<<<<<<<<<<<<
 *                                         shift = i
 *                                     elif dd == 1:
 */
                  switch (__pyx_v_dd) {
                    case 0:

                    /* "fatslimlib/core_base.pyx":506
 *                                 for dd in range(DIM):
 *                                     if dd == 0:
 *                                         shift = i             # <<<<<<<<<<<<<<
 *                                     elif dd == 1:
 *                                         shift = j
 */
                    __pyx_v_shift = __pyx_v_i;

                    /* "fatslimlib/core_base.pyx":505
 * 
 *                                 for dd in range(DIM):
 *                                     if dd == 0:             # <<<<<<<<<<<<<<
 *                                         shift = i
 *                                     elif dd == 1:
 */
                    break;
                    case 1:

                    /* "fatslimlib/core_base.pyx":508
 *                                         shift = i
 *                                     elif dd == 1:
 *                                         shift = j             # <<<<<<<<<<<<<<
 *                                     else:
 *                                         shift = k
 */
                    __pyx_v_shift = __pyx_v_j;

                    /* "fatslimlib/core_base.pyx":507
 *                                     if dd == 0:
 *                                         shift = i
 *                                     elif dd == 1:             # <<<<<<<<<<<<<<
 *                                         shift = j
 *                                     else:
 */
                    break;
                    default:

                    /* "fatslimlib/core_base.pyx":510
 *                                         shift = j
 *                                     else:
 *                                         shift = k             # <<<<<<<<<<<<<<
 * 
 *                                     if shift:
 */
                    __pyx_v_shift = __pyx_v_k;
                    break;
                  }

                  /* "fatslimlib/core_base.pyx":512
 *                                         shift = k
 * 
 *                                     if shift:             # <<<<<<<<<<<<<<
 *                                         d2new_c = 0
 * 
 */
                  __pyx_t_28 = (__pyx_v_shift != 0);
                  if (__pyx_t_28) {

                    /* "fatslimlib/core_base.pyx":513
 * 
 *                                     if shift:
 *                                         d2new_c = 0             # <<<<<<<<<<<<<<
 * 
 *                                         for d in range(DIM):
 */
                    __pyx_v_d2new_c = 0.0;

                    /* "fatslimlib/core_base.pyx":515
 *                                         d2new_c = 0
 * 
 *                                         for d in range(DIM):             # <<<<<<<<<<<<<<
 *                                             d2new_c += sqrt(pos[d] + trial[d] - shift*box[dd, d])
 * 
 */
                    for (__pyx_t_29 = 0; __pyx_t_29 < 3; __pyx_t_29+=1) {
                      __pyx_v_d = __pyx_t_29;

                      /* "fatslimlib/core_base.pyx":516
 * 
 *                                         for d in range(DIM):
 *                                             d2new_c += sqrt(pos[d] + trial[d] - shift*box[dd, d])             # <<<<<<<<<<<<<<
 * 
 *                                         if d2new_c <= BOX_MARGIN*d2new:
 */
                      __pyx_t_40 = __pyx_v_dd;
                      __pyx_t_41 = __pyx_v_d;
                      __pyx_v_d2new_c = (__pyx_v_d2new_c + sqrt((((__pyx_v_pos[__pyx_v_d]) + (__pyx_v_trial[__pyx_v_d])) - (__pyx_v_shift * (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_40 * __pyx_v_box.strides[0]) )) + __pyx_t_41)) )))))));
                    }

                    /* "fatslimlib/core_base.pyx":518
 *                                             d2new_c += sqrt(pos[d] + trial[d] - shift*box[dd, d])
 * 
 *                                         if d2new_c <= BOX_MARGIN*d2new:             # <<<<<<<<<<<<<<
 *                                             use = False
 * 
 */
                    __pyx_t_28 = ((__pyx_v_d2new_c <= (1.001 * __pyx_v_d2new)) != 0);
                    if (__pyx_t_28) {

                      /* "fatslimlib/core_base.pyx":519
 * 
 *                                         if d2new_c <= BOX_MARGIN*d2new:
 *                                             use = False             # <<<<<<<<<<<<<<
 * 
 *                                 if use: # Accept this shift vector.
 */
                      __pyx_v_use = 0;

                      /* "fatslimlib/core_base.pyx":518
 *                                             d2new_c += sqrt(pos[d] + trial[d] - shift*box[dd, d])
 * 
 *                                         if d2new_c <= BOX_MARGIN*d2new:             # <<<<<<<<<<<<<<
 *                                             use = False
 * 
 */
                    }

                    /* "fatslimlib/core_base.pyx":512
 *                                         shift = k
 * 
 *                                     if shift:             # <<<<<<<<<<<<<<
 *                                         d2new_c = 0
 * 
 */
                  }
                }

                /* "fatslimlib/core_base.pyx":521
 *                                             use = False
 * 
 *                                 if use: # Accept this shift vector.             # <<<<<<<<<<<<<<
 *                                     if self.c_pbcbox.ntric_vec >= MAX_NTRICVEC:
 *                                         with gil:
 */
                __pyx_t_28 = (__pyx_v_use != 0);
                if (__pyx_t_28) {

                  /* "fatslimlib/core_base.pyx":522
 * 
 *                                 if use: # Accept this shift vector.
 *                                     if self.c_pbcbox.ntric_vec >= MAX_NTRICVEC:             # <<<<<<<<<<<<<<
 *                                         with gil:
 *                                             print("\nWARNING: Found more than %d triclinic "
 */
                  __pyx_t_28 = ((__pyx_v_self->c_pbcbox.ntric_vec >= 12) != 0);
                  if (__pyx_t_28) {

                    /* "fatslimlib/core_base.pyx":523
 *                                 if use: # Accept this shift vector.
 *                                     if self.c_pbcbox.ntric_vec >= MAX_NTRICVEC:
 *                                         with gil:             # <<<<<<<<<<<<<<
 *                                             print("\nWARNING: Found more than %d triclinic "
 *                                                   "correction vectors, ignoring some."
 */
                    {
                        #ifdef WITH_THREAD
                        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
                        #endif
                        /*try:*/ {

                          /* "fatslimlib/core_base.pyx":526
 *                                             print("\nWARNING: Found more than %d triclinic "
 *                                                   "correction vectors, ignoring some."
 *                                                   % MAX_NTRICVEC)             # <<<<<<<<<<<<<<
 *                                             print("  There is probably something wrong with "
 *                                                   "your box.")
 */
                          __pyx_t_42 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_WARNING_Found_more_than_d_tricl, __pyx_int_12); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 526, __pyx_L42_error)
                          __Pyx_GOTREF(__pyx_t_42);

                          /* "fatslimlib/core_base.pyx":524
 *                                     if self.c_pbcbox.ntric_vec >= MAX_NTRICVEC:
 *                                         with gil:
 *                                             print("\nWARNING: Found more than %d triclinic "             # <<<<<<<<<<<<<<
 *                                                   "correction vectors, ignoring some."
 *                                                   % MAX_NTRICVEC)
 */
                          __pyx_t_43 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_42); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 524, __pyx_L42_error)
                          __Pyx_GOTREF(__pyx_t_43);
                          __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;
                          __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;

                          /* "fatslimlib/core_base.pyx":527
 *                                                   "correction vectors, ignoring some."
 *                                                   % MAX_NTRICVEC)
 *                                             print("  There is probably something wrong with "             # <<<<<<<<<<<<<<
 *                                                   "your box.")
 *                                             print(box)
 */
                          __pyx_t_43 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 527, __pyx_L42_error)
                          __Pyx_GOTREF(__pyx_t_43);
                          __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;

                          /* "fatslimlib/core_base.pyx":529
 *                                             print("  There is probably something wrong with "
 *                                                   "your box.")
 *                                             print(box)             # <<<<<<<<<<<<<<
 *                                     else:
 *                                         for d in range(DIM):
 */
                          __pyx_t_43 = __pyx_memoryview_fromslice(__pyx_v_box, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_43)) __PYX_ERR(0, 529, __pyx_L42_error)
                          __Pyx_GOTREF(__pyx_t_43);
                          __pyx_t_42 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_43); if (unlikely(!__pyx_t_42)) __PYX_ERR(0, 529, __pyx_L42_error)
                          __Pyx_GOTREF(__pyx_t_42);
                          __Pyx_DECREF(__pyx_t_43); __pyx_t_43 = 0;
                          __Pyx_DECREF(__pyx_t_42); __pyx_t_42 = 0;
                        }

                        /* "fatslimlib/core_base.pyx":523
 *                                 if use: # Accept this shift vector.
 *                                     if self.c_pbcbox.ntric_vec >= MAX_NTRICVEC:
 *                                         with gil:             # <<<<<<<<<<<<<<
 *                                             print("\nWARNING: Found more than %d triclinic "
 *                                                   "correction vectors, ignoring some."
 */
                        /*finally:*/ {
                          /*normal exit:*/{
                            #ifdef WITH_THREAD
                            __Pyx_PyGILState_Release(__pyx_gilstate_save);
                            #endif
                            goto __pyx_L43;
                          }
                          __pyx_L42_error: {
                            #ifdef WITH_THREAD
                            __Pyx_PyGILState_Release(__pyx_gilstate_save);
                            #endif
                            goto __pyx_L4_error;
                          }
                          __pyx_L43:;
                        }
                    }

                    /* "fatslimlib/core_base.pyx":522
 * 
 *                                 if use: # Accept this shift vector.
 *                                     if self.c_pbcbox.ntric_vec >= MAX_NTRICVEC:             # <<<<<<<<<<<<<<
 *                                         with gil:
 *                                             print("\nWARNING: Found more than %d triclinic "
 */
                    goto __pyx_L38;
                  }

                  /* "fatslimlib/core_base.pyx":531
 *                                             print(box)
 *                                     else:
 *                                         for d in range(DIM):             # <<<<<<<<<<<<<<
 *                                             self.c_pbcbox.tric_vec[self.c_pbcbox.ntric_vec][d] = \
 *                                                 trial[d]
 */
                  /*else*/ {
                    for (__pyx_t_29 = 0; __pyx_t_29 < 3; __pyx_t_29+=1) {
                      __pyx_v_d = __pyx_t_29;

                      /* "fatslimlib/core_base.pyx":532
 *                                     else:
 *                                         for d in range(DIM):
 *                                             self.c_pbcbox.tric_vec[self.c_pbcbox.ntric_vec][d] = \             # <<<<<<<<<<<<<<
 *                                                 trial[d]
 *                                         self.c_pbcbox.tric_shift[self.c_pbcbox.ntric_vec][XX] = i
 */
                      ((__pyx_v_self->c_pbcbox.tric_vec[__pyx_v_self->c_pbcbox.ntric_vec])[__pyx_v_d]) = (__pyx_v_trial[__pyx_v_d]);
                    }

                    /* "fatslimlib/core_base.pyx":534
 *                                             self.c_pbcbox.tric_vec[self.c_pbcbox.ntric_vec][d] = \
 *                                                 trial[d]
 *                                         self.c_pbcbox.tric_shift[self.c_pbcbox.ntric_vec][XX] = i             # <<<<<<<<<<<<<<
 *                                         self.c_pbcbox.tric_shift[self.c_pbcbox.ntric_vec][YY] = j
 *                                         self.c_pbcbox.tric_shift[self.c_pbcbox.ntric_vec][ZZ] = k
 */
                    ((__pyx_v_self->c_pbcbox.tric_shift[__pyx_v_self->c_pbcbox.ntric_vec])[0]) = __pyx_v_i;

                    /* "fatslimlib/core_base.pyx":535
 *                                                 trial[d]
 *                                         self.c_pbcbox.tric_shift[self.c_pbcbox.ntric_vec][XX] = i
 *                                         self.c_pbcbox.tric_shift[self.c_pbcbox.ntric_vec][YY] = j             # <<<<<<<<<<<<<<
 *                                         self.c_pbcbox.tric_shift[self.c_pbcbox.ntric_vec][ZZ] = k
 *                                         self.c_pbcbox.ntric_vec += 1
 */
                    ((__pyx_v_self->c_pbcbox.tric_shift[__pyx_v_self->c_pbcbox.ntric_vec])[1]) = __pyx_v_j;

                    /* "fatslimlib/core_base.pyx":536
 *                                         self.c_pbcbox.tric_shift[self.c_pbcbox.ntric_vec][XX] = i
 *                                         self.c_pbcbox.tric_shift[self.c_pbcbox.ntric_vec][YY] = j
 *                                         self.c_pbcbox.tric_shift[self.c_pbcbox.ntric_vec][ZZ] = k             # <<<<<<<<<<<<<<
 *                                         self.c_pbcbox.ntric_vec += 1
 * 
 */
                    ((__pyx_v_self->c_pbcbox.tric_shift[__pyx_v_self->c_pbcbox.ntric_vec])[2]) = __pyx_v_k;

                    /* "fatslimlib/core_base.pyx":537
 *                                         self.c_pbcbox.tric_shift[self.c_pbcbox.ntric_vec][YY] = j
 *                                         self.c_pbcbox.tric_shift[self.c_pbcbox.ntric_vec][ZZ] = k
 *                                         self.c_pbcbox.ntric_vec += 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
                    __pyx_v_self->c_pbcbox.ntric_vec = (__pyx_v_self->c_pbcbox.ntric_vec + 1);
                  }
                  __pyx_L38:;

                  /* "fatslimlib/core_base.pyx":521
 *                                             use = False
 * 
 *                                 if use: # Accept this shift vector.             # <<<<<<<<<<<<<<
 *                                     if self.c_pbcbox.ntric_vec >= MAX_NTRICVEC:
 *                                         with gil:
 */
                }

                /* "fatslimlib/core_base.pyx":501
 * 
 *                         if BOX_MARGIN*d2new < d2old:
 *                             if  not (j < -1 or j > 1 or k < -1 or k > 1):             # <<<<<<<<<<<<<<
 *                                 use = True
 * 
 */
              }

              /* "fatslimlib/core_base.pyx":500
 *                             d2new += sqrt(pos[d] + trial[d])
 * 
 *                         if BOX_MARGIN*d2new < d2old:             # <<<<<<<<<<<<<<
 *                             if  not (j < -1 or j > 1 or k < -1 or k > 1):
 *                                 use = True
 */
            }

            /* "fatslimlib/core_base.pyx":478
 * 
 *                     # A shift is only useful when it is trilinic
 *                     if j != 0 or k != 0:             # <<<<<<<<<<<<<<
 *                         d2old = 0
 *                         d2new = 0
 */
          }
        }
      }
    }
  }

  /* "fatslimlib/core_base.pyx":426
 * 
 *     cdef void fast_update(self, real[:,::1] box) nogil:
 *         cdef fsl_int i, j, k, d, jc, kc, shift             # <<<<<<<<<<<<<<
 *         cdef real d2old, d2new, d2new_c
 *         cdef rvec trial, pos
 */
  /*finally:*/ {
    /*normal exit:*/{
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L5;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "fatslimlib/core_base.pyx":425
 *         self.update(box)
 * 
 *     cdef void fast_update(self, real[:,::1] box) nogil:             # <<<<<<<<<<<<<<
 *         cdef fsl_int i, j, k, d, jc, kc, shift
 *         cdef real d2old, d2new, d2new_c
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_42);
  __Pyx_XDECREF(__pyx_t_43);
  __Pyx_WriteUnraisable("fatslimlib.core_base.PBCBox.fast_update", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "fatslimlib/core_base.pyx":540
 * 
 * 
 *     def update(self, real[:,::1] box):             # <<<<<<<<<<<<<<
 *         if box.shape[0] != DIM or box.shape[1] != DIM:
 *             raise ValueError("Box must be a %i x %i matrix. (shape: %i x %i)" %
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_3update(PyObject *__pyx_v_self, PyObject *__pyx_arg_box); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_3update(PyObject *__pyx_v_self, PyObject *__pyx_arg_box) {
  __Pyx_memviewslice __pyx_v_box = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update (wrapper)", 0);
  assert(__pyx_arg_box); {
    __pyx_v_box = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(__pyx_arg_box, PyBUF_WRITABLE); if (unlikely(!__pyx_v_box.memview)) __PYX_ERR(0, 540, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6PBCBox_2update(((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_v_self), __pyx_v_box);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_2update(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_box) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  __Pyx_RefNannySetupContext("update", 0);

  /* "fatslimlib/core_base.pyx":541
 * 
 *     def update(self, real[:,::1] box):
 *         if box.shape[0] != DIM or box.shape[1] != DIM:             # <<<<<<<<<<<<<<
 *             raise ValueError("Box must be a %i x %i matrix. (shape: %i x %i)" %
 *                              (DIM, DIM, box.shape[0], box.shape[1]))
 */
  __pyx_t_2 = (((__pyx_v_box.shape[0]) != 3) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (((__pyx_v_box.shape[1]) != 3) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "fatslimlib/core_base.pyx":543
 *         if box.shape[0] != DIM or box.shape[1] != DIM:
 *             raise ValueError("Box must be a %i x %i matrix. (shape: %i x %i)" %
 *                              (DIM, DIM, box.shape[0], box.shape[1]))             # <<<<<<<<<<<<<<
 *         if (box[XX, XX] == 0) or (box[YY, YY] == 0) or (box[ZZ, ZZ] == 0):
 *             raise ValueError("Box does not correspond to PBC=xyz")
 */
    __pyx_t_3 = PyInt_FromSsize_t((__pyx_v_box.shape[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 543, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyInt_FromSsize_t((__pyx_v_box.shape[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 543, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 543, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_int_3);
    __Pyx_INCREF(__pyx_int_3);
    __Pyx_GIVEREF(__pyx_int_3);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;

    /* "fatslimlib/core_base.pyx":542
 *     def update(self, real[:,::1] box):
 *         if box.shape[0] != DIM or box.shape[1] != DIM:
 *             raise ValueError("Box must be a %i x %i matrix. (shape: %i x %i)" %             # <<<<<<<<<<<<<<
 *                              (DIM, DIM, box.shape[0], box.shape[1]))
 *         if (box[XX, XX] == 0) or (box[YY, YY] == 0) or (box[ZZ, ZZ] == 0):
 */
    __pyx_t_4 = PyUnicode_Format(__pyx_kp_u_Box_must_be_a_i_x_i_matrix_shape, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 542, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":541
 * 
 *     def update(self, real[:,::1] box):
 *         if box.shape[0] != DIM or box.shape[1] != DIM:             # <<<<<<<<<<<<<<
 *             raise ValueError("Box must be a %i x %i matrix. (shape: %i x %i)" %
 *                              (DIM, DIM, box.shape[0], box.shape[1]))
 */
  }

  /* "fatslimlib/core_base.pyx":544
 *             raise ValueError("Box must be a %i x %i matrix. (shape: %i x %i)" %
 *                              (DIM, DIM, box.shape[0], box.shape[1]))
 *         if (box[XX, XX] == 0) or (box[YY, YY] == 0) or (box[ZZ, ZZ] == 0):             # <<<<<<<<<<<<<<
 *             raise ValueError("Box does not correspond to PBC=xyz")
 *         self.fast_update(box)
 */
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_2 = (((*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_6 * __pyx_v_box.strides[0]) )) + __pyx_t_7)) ))) == 0.0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_8 = 1;
  __pyx_t_9 = 1;
  __pyx_t_2 = (((*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_8 * __pyx_v_box.strides[0]) )) + __pyx_t_9)) ))) == 0.0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_10 = 2;
  __pyx_t_11 = 2;
  __pyx_t_2 = (((*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_box.data + __pyx_t_10 * __pyx_v_box.strides[0]) )) + __pyx_t_11)) ))) == 0.0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "fatslimlib/core_base.pyx":545
 *                              (DIM, DIM, box.shape[0], box.shape[1]))
 *         if (box[XX, XX] == 0) or (box[YY, YY] == 0) or (box[ZZ, ZZ] == 0):
 *             raise ValueError("Box does not correspond to PBC=xyz")             # <<<<<<<<<<<<<<
 *         self.fast_update(box)
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 545, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":544
 *             raise ValueError("Box must be a %i x %i matrix. (shape: %i x %i)" %
 *                              (DIM, DIM, box.shape[0], box.shape[1]))
 *         if (box[XX, XX] == 0) or (box[YY, YY] == 0) or (box[ZZ, ZZ] == 0):             # <<<<<<<<<<<<<<
 *             raise ValueError("Box does not correspond to PBC=xyz")
 *         self.fast_update(box)
 */
  }

  /* "fatslimlib/core_base.pyx":546
 *         if (box[XX, XX] == 0) or (box[YY, YY] == 0) or (box[ZZ, ZZ] == 0):
 *             raise ValueError("Box does not correspond to PBC=xyz")
 *         self.fast_update(box)             # <<<<<<<<<<<<<<
 * 
 *     property box:
 */
  ((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_self->__pyx_vtab)->fast_update(__pyx_v_self, __pyx_v_box);

  /* "fatslimlib/core_base.pyx":540
 * 
 * 
 *     def update(self, real[:,::1] box):             # <<<<<<<<<<<<<<
 *         if box.shape[0] != DIM or box.shape[1] != DIM:
 *             raise ValueError("Box must be a %i x %i matrix. (shape: %i x %i)" %
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_box, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":549
 * 
 *     property box:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             box_py = np.empty((DIM, DIM))
 *             cdef real[:,::1] box_memview = box_py
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_3box_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_3box_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6PBCBox_3box___get__(((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_3box___get__(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self) {
  PyObject *__pyx_v_box_py = NULL;
  __Pyx_memviewslice __pyx_v_box_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_memviewslice __pyx_t_4 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":550
 *     property box:
 *         def __get__(self):
 *             box_py = np.empty((DIM, DIM))             # <<<<<<<<<<<<<<
 *             cdef real[:,::1] box_memview = box_py
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_tuple__14) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_tuple__14);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_box_py = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":551
 *         def __get__(self):
 *             box_py = np.empty((DIM, DIM))
 *             cdef real[:,::1] box_memview = box_py             # <<<<<<<<<<<<<<
 * 
 *             mat_to_memview(self.c_pbcbox.box, box_memview)
 */
  __pyx_t_4 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(__pyx_v_box_py, PyBUF_WRITABLE); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(0, 551, __pyx_L1_error)
  __pyx_v_box_memview = __pyx_t_4;
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;

  /* "fatslimlib/core_base.pyx":553
 *             cdef real[:,::1] box_memview = box_py
 * 
 *             mat_to_memview(self.c_pbcbox.box, box_memview)             # <<<<<<<<<<<<<<
 * 
 *             return box_py
 */
  __pyx_f_10fatslimlib_9core_base_mat_to_memview(__pyx_v_self->c_pbcbox.box, __pyx_v_box_memview);

  /* "fatslimlib/core_base.pyx":555
 *             mat_to_memview(self.c_pbcbox.box, box_memview)
 * 
 *             return box_py             # <<<<<<<<<<<<<<
 * 
 *     property center:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_box_py);
  __pyx_r = __pyx_v_box_py;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":549
 * 
 *     property box:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             box_py = np.empty((DIM, DIM))
 *             cdef real[:,::1] box_memview = box_py
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __PYX_XDEC_MEMVIEW(&__pyx_t_4, 1);
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.box.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_box_py);
  __PYX_XDEC_MEMVIEW(&__pyx_v_box_memview, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":558
 * 
 *     property center:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return np.array([self.center[XX], self.center[YY], self.center[ZZ]], dtype=np.float64)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_6center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_6center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6PBCBox_6center___get__(((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_6center___get__(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":559
 *     property center:
 *         def __get__(self):
 *             return np.array([self.center[XX], self.center[YY], self.center[ZZ]], dtype=np.float64)             # <<<<<<<<<<<<<<
 * 
 *     property bbox_center:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->center[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyFloat_FromDouble((__pyx_v_self->center[1])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble((__pyx_v_self->center[2])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyList_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyList_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float64); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":558
 * 
 *     property center:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return np.array([self.center[XX], self.center[YY], self.center[ZZ]], dtype=np.float64)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":562
 * 
 *     property bbox_center:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return np.array([self.bbox_center[XX], self.bbox_center[YY], self.bbox_center[ZZ]],
 *                             dtype=np.float64)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_11bbox_center_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_11bbox_center_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6PBCBox_11bbox_center___get__(((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_11bbox_center___get__(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":563
 *     property bbox_center:
 *         def __get__(self):
 *             return np.array([self.bbox_center[XX], self.bbox_center[YY], self.bbox_center[ZZ]],             # <<<<<<<<<<<<<<
 *                             dtype=np.float64)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->bbox_center[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyFloat_FromDouble((__pyx_v_self->bbox_center[1])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble((__pyx_v_self->bbox_center[2])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyList_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyList_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
  __pyx_t_5 = 0;

  /* "fatslimlib/core_base.pyx":564
 *         def __get__(self):
 *             return np.array([self.bbox_center[XX], self.bbox_center[YY], self.bbox_center[ZZ]],
 *                             dtype=np.float64)             # <<<<<<<<<<<<<<
 * 
 *     def to_gro(self):
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_float64); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":563
 *     property bbox_center:
 *         def __get__(self):
 *             return np.array([self.bbox_center[XX], self.bbox_center[YY], self.bbox_center[ZZ]],             # <<<<<<<<<<<<<<
 *                             dtype=np.float64)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":562
 * 
 *     property bbox_center:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return np.array([self.bbox_center[XX], self.bbox_center[YY], self.bbox_center[ZZ]],
 *                             dtype=np.float64)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.bbox_center.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":566
 *                             dtype=np.float64)
 * 
 *     def to_gro(self):             # <<<<<<<<<<<<<<
 *         ret_str = "%.5f %.5f %.5f %.5f %.5f %.5f %.5f %.5f %.5f" % (
 *             self.c_pbcbox.box[XX][XX], self.c_pbcbox.box[YY][YY], self.c_pbcbox.box[ZZ][ZZ],
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_5to_gro(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_5to_gro(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_gro (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6PBCBox_4to_gro(((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_4to_gro(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self) {
  PyObject *__pyx_v_ret_str = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("to_gro", 0);

  /* "fatslimlib/core_base.pyx":567
 * 
 *     def to_gro(self):
 *         ret_str = "%.5f %.5f %.5f %.5f %.5f %.5f %.5f %.5f %.5f" % (             # <<<<<<<<<<<<<<
 *             self.c_pbcbox.box[XX][XX], self.c_pbcbox.box[YY][YY], self.c_pbcbox.box[ZZ][ZZ],
 *             self.c_pbcbox.box[XX][YY], self.c_pbcbox.box[XX][ZZ],
 */
  __pyx_t_1 = PyTuple_New(17); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;

  /* "fatslimlib/core_base.pyx":568
 *     def to_gro(self):
 *         ret_str = "%.5f %.5f %.5f %.5f %.5f %.5f %.5f %.5f %.5f" % (
 *             self.c_pbcbox.box[XX][XX], self.c_pbcbox.box[YY][YY], self.c_pbcbox.box[ZZ][ZZ],             # <<<<<<<<<<<<<<
 *             self.c_pbcbox.box[XX][YY], self.c_pbcbox.box[XX][ZZ],
 *             self.c_pbcbox.box[YY][XX], self.c_pbcbox.box[YY][ZZ],
 */
  __pyx_t_4 = PyFloat_FromDouble(((__pyx_v_self->c_pbcbox.box[0])[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_Format(__pyx_t_4, __pyx_kp_u_5f); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u__4);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__4);
  __pyx_t_5 = PyFloat_FromDouble(((__pyx_v_self->c_pbcbox.box[1])[1])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_Format(__pyx_t_5, __pyx_kp_u_5f); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__4);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u__4);
  __pyx_t_4 = PyFloat_FromDouble(((__pyx_v_self->c_pbcbox.box[2])[2])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_Format(__pyx_t_4, __pyx_kp_u_5f); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u__4);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__4);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_kp_u__4);

  /* "fatslimlib/core_base.pyx":569
 *         ret_str = "%.5f %.5f %.5f %.5f %.5f %.5f %.5f %.5f %.5f" % (
 *             self.c_pbcbox.box[XX][XX], self.c_pbcbox.box[YY][YY], self.c_pbcbox.box[ZZ][ZZ],
 *             self.c_pbcbox.box[XX][YY], self.c_pbcbox.box[XX][ZZ],             # <<<<<<<<<<<<<<
 *             self.c_pbcbox.box[YY][XX], self.c_pbcbox.box[YY][ZZ],
 *             self.c_pbcbox.box[ZZ][XX], self.c_pbcbox.box[ZZ][YY],
 */
  __pyx_t_5 = PyFloat_FromDouble(((__pyx_v_self->c_pbcbox.box[0])[1])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_Format(__pyx_t_5, __pyx_kp_u_5f); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__4);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__4);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_kp_u__4);
  __pyx_t_4 = PyFloat_FromDouble(((__pyx_v_self->c_pbcbox.box[0])[2])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_Format(__pyx_t_4, __pyx_kp_u_5f); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u__4);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__4);
  PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_kp_u__4);

  /* "fatslimlib/core_base.pyx":570
 *             self.c_pbcbox.box[XX][XX], self.c_pbcbox.box[YY][YY], self.c_pbcbox.box[ZZ][ZZ],
 *             self.c_pbcbox.box[XX][YY], self.c_pbcbox.box[XX][ZZ],
 *             self.c_pbcbox.box[YY][XX], self.c_pbcbox.box[YY][ZZ],             # <<<<<<<<<<<<<<
 *             self.c_pbcbox.box[ZZ][XX], self.c_pbcbox.box[ZZ][YY],
 *         )
 */
  __pyx_t_5 = PyFloat_FromDouble(((__pyx_v_self->c_pbcbox.box[1])[0])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_Format(__pyx_t_5, __pyx_kp_u_5f); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 10, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__4);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__4);
  PyTuple_SET_ITEM(__pyx_t_1, 11, __pyx_kp_u__4);
  __pyx_t_4 = PyFloat_FromDouble(((__pyx_v_self->c_pbcbox.box[1])[2])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_Format(__pyx_t_4, __pyx_kp_u_5f); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 12, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u__4);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__4);
  PyTuple_SET_ITEM(__pyx_t_1, 13, __pyx_kp_u__4);

  /* "fatslimlib/core_base.pyx":571
 *             self.c_pbcbox.box[XX][YY], self.c_pbcbox.box[XX][ZZ],
 *             self.c_pbcbox.box[YY][XX], self.c_pbcbox.box[YY][ZZ],
 *             self.c_pbcbox.box[ZZ][XX], self.c_pbcbox.box[ZZ][YY],             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_5 = PyFloat_FromDouble(((__pyx_v_self->c_pbcbox.box[2])[0])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_Format(__pyx_t_5, __pyx_kp_u_5f); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 14, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__4);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__4);
  PyTuple_SET_ITEM(__pyx_t_1, 15, __pyx_kp_u__4);
  __pyx_t_4 = PyFloat_FromDouble(((__pyx_v_self->c_pbcbox.box[2])[1])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_Format(__pyx_t_4, __pyx_kp_u_5f); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 16, __pyx_t_5);
  __pyx_t_5 = 0;

  /* "fatslimlib/core_base.pyx":567
 * 
 *     def to_gro(self):
 *         ret_str = "%.5f %.5f %.5f %.5f %.5f %.5f %.5f %.5f %.5f" % (             # <<<<<<<<<<<<<<
 *             self.c_pbcbox.box[XX][XX], self.c_pbcbox.box[YY][YY], self.c_pbcbox.box[ZZ][ZZ],
 *             self.c_pbcbox.box[XX][YY], self.c_pbcbox.box[XX][ZZ],
 */
  __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_1, 17, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_ret_str = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "fatslimlib/core_base.pyx":574
 *         )
 * 
 *         return ret_str             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ret_str);
  __pyx_r = __pyx_v_ret_str;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":566
 *                             dtype=np.float64)
 * 
 *     def to_gro(self):             # <<<<<<<<<<<<<<
 *         ret_str = "%.5f %.5f %.5f %.5f %.5f %.5f %.5f %.5f %.5f" % (
 *             self.c_pbcbox.box[XX][XX], self.c_pbcbox.box[YY][YY], self.c_pbcbox.box[ZZ][ZZ],
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.to_gro", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret_str);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":576
 *         return ret_str
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(getattr(self, "box"))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_7__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_7__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6PBCBox_6__str__(((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_6__str__(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "fatslimlib/core_base.pyx":577
 * 
 *     def __str__(self):
 *         return str(getattr(self, "box"))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_GetAttr(((PyObject *)__pyx_v_self), __pyx_n_u_box); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":576
 *         return ret_str
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(getattr(self, "box"))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":580
 * 
 * 
 *     cdef void fast_pbc_dx(self, rvec ref, rvec other, rvec dx) nogil:             # <<<<<<<<<<<<<<
 *         fast_pbc_dx_c(self.c_pbcbox, ref, other, dx)
 * 
 */

static void __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_pbc_dx(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, real *__pyx_v_ref, real *__pyx_v_other, real *__pyx_v_dx) {

  /* "fatslimlib/core_base.pyx":581
 * 
 *     cdef void fast_pbc_dx(self, rvec ref, rvec other, rvec dx) nogil:
 *         fast_pbc_dx_c(self.c_pbcbox, ref, other, dx)             # <<<<<<<<<<<<<<
 * 
 *     cdef void fast_pbc_dx_leaflet(self, rvec ref, rvec other, rvec dx, rvec ref_normal) nogil:
 */
  __pyx_f_10fatslimlib_9core_base_fast_pbc_dx_c(__pyx_v_self->c_pbcbox, __pyx_v_ref, __pyx_v_other, __pyx_v_dx);

  /* "fatslimlib/core_base.pyx":580
 * 
 * 
 *     cdef void fast_pbc_dx(self, rvec ref, rvec other, rvec dx) nogil:             # <<<<<<<<<<<<<<
 *         fast_pbc_dx_c(self.c_pbcbox, ref, other, dx)
 * 
 */

  /* function exit code */
}

/* "fatslimlib/core_base.pyx":583
 *         fast_pbc_dx_c(self.c_pbcbox, ref, other, dx)
 * 
 *     cdef void fast_pbc_dx_leaflet(self, rvec ref, rvec other, rvec dx, rvec ref_normal) nogil:             # <<<<<<<<<<<<<<
 *         cdef fsl_int i, j, xoffset, yoffset, zoffset, notset=1
 *         cdef rvec dx_raw, dx_try, offset, offset_x, offset_y, offset_z
 */

static void __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_pbc_dx_leaflet(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, real *__pyx_v_ref, real *__pyx_v_other, real *__pyx_v_dx, real *__pyx_v_ref_normal) {
  fsl_int __pyx_v_i;
  CYTHON_UNUSED fsl_int __pyx_v_notset;
  real __pyx_v_coord_val;
  real __pyx_v_tmp;
  fsl_int __pyx_v_shift_axis;
  fsl_int __pyx_v_sign;
  int __pyx_t_1;
  fsl_int __pyx_t_2;
  fsl_int __pyx_t_3;

  /* "fatslimlib/core_base.pyx":584
 * 
 *     cdef void fast_pbc_dx_leaflet(self, rvec ref, rvec other, rvec dx, rvec ref_normal) nogil:
 *         cdef fsl_int i, j, xoffset, yoffset, zoffset, notset=1             # <<<<<<<<<<<<<<
 *         cdef rvec dx_raw, dx_try, offset, offset_x, offset_y, offset_z
 *         cdef real d2min, d2try
 */
  __pyx_v_notset = 1;

  /* "fatslimlib/core_base.pyx":591
 * 
 *         # First: get the actual pbc dx
 *         self.fast_pbc_dx(ref, other, dx)             # <<<<<<<<<<<<<<
 * 
 *         # If dprod(dx, normal) < 0, dx goes through the bilayer
 */
  ((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_self->__pyx_vtab)->fast_pbc_dx(__pyx_v_self, __pyx_v_ref, __pyx_v_other, __pyx_v_dx);

  /* "fatslimlib/core_base.pyx":595
 *         # If dprod(dx, normal) < 0, dx goes through the bilayer
 *         # because dx = vector from ref to other
 *         if rvec_dprod(dx, ref_normal) < 0:             # <<<<<<<<<<<<<<
 *             return # Nothing to do!
 * 
 */
  __pyx_t_1 = ((rvec_dprod(__pyx_v_dx, __pyx_v_ref_normal) < 0.0) != 0);
  if (__pyx_t_1) {

    /* "fatslimlib/core_base.pyx":596
 *         # because dx = vector from ref to other
 *         if rvec_dprod(dx, ref_normal) < 0:
 *             return # Nothing to do!             # <<<<<<<<<<<<<<
 * 
 *         # Second: If the pbc dx is not right, we need to find the correct direction to go
 */
    goto __pyx_L0;

    /* "fatslimlib/core_base.pyx":595
 *         # If dprod(dx, normal) < 0, dx goes through the bilayer
 *         # because dx = vector from ref to other
 *         if rvec_dprod(dx, ref_normal) < 0:             # <<<<<<<<<<<<<<
 *             return # Nothing to do!
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":599
 * 
 *         # Second: If the pbc dx is not right, we need to find the correct direction to go
 *         shift_axis = ZZ             # <<<<<<<<<<<<<<
 *         coord_val = fabs(ref_normal[ZZ])
 *         for i in range(DIM-1):
 */
  __pyx_v_shift_axis = 2;

  /* "fatslimlib/core_base.pyx":600
 *         # Second: If the pbc dx is not right, we need to find the correct direction to go
 *         shift_axis = ZZ
 *         coord_val = fabs(ref_normal[ZZ])             # <<<<<<<<<<<<<<
 *         for i in range(DIM-1):
 *             tmp = fabs(ref_normal[i])
 */
  __pyx_v_coord_val = fabs((__pyx_v_ref_normal[2]));

  /* "fatslimlib/core_base.pyx":601
 *         shift_axis = ZZ
 *         coord_val = fabs(ref_normal[ZZ])
 *         for i in range(DIM-1):             # <<<<<<<<<<<<<<
 *             tmp = fabs(ref_normal[i])
 *             if tmp > coord_val:
 */
  for (__pyx_t_2 = 0; __pyx_t_2 < 2; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "fatslimlib/core_base.pyx":602
 *         coord_val = fabs(ref_normal[ZZ])
 *         for i in range(DIM-1):
 *             tmp = fabs(ref_normal[i])             # <<<<<<<<<<<<<<
 *             if tmp > coord_val:
 *                 shift_axis = i
 */
    __pyx_v_tmp = fabs((__pyx_v_ref_normal[__pyx_v_i]));

    /* "fatslimlib/core_base.pyx":603
 *         for i in range(DIM-1):
 *             tmp = fabs(ref_normal[i])
 *             if tmp > coord_val:             # <<<<<<<<<<<<<<
 *                 shift_axis = i
 * 
 */
    __pyx_t_1 = ((__pyx_v_tmp > __pyx_v_coord_val) != 0);
    if (__pyx_t_1) {

      /* "fatslimlib/core_base.pyx":604
 *             tmp = fabs(ref_normal[i])
 *             if tmp > coord_val:
 *                 shift_axis = i             # <<<<<<<<<<<<<<
 * 
 *         # Third: once we found the correct direction to shift,
 */
      __pyx_v_shift_axis = __pyx_v_i;

      /* "fatslimlib/core_base.pyx":603
 *         for i in range(DIM-1):
 *             tmp = fabs(ref_normal[i])
 *             if tmp > coord_val:             # <<<<<<<<<<<<<<
 *                 shift_axis = i
 * 
 */
    }
  }

  /* "fatslimlib/core_base.pyx":608
 *         # Third: once we found the correct direction to shift,
 *         # make one shift in that direction to get the right distance
 *         sign = 1             # <<<<<<<<<<<<<<
 *         if ref_normal[shift_axis] > 0:
 *             sign = -1
 */
  __pyx_v_sign = 1;

  /* "fatslimlib/core_base.pyx":609
 *         # make one shift in that direction to get the right distance
 *         sign = 1
 *         if ref_normal[shift_axis] > 0:             # <<<<<<<<<<<<<<
 *             sign = -1
 *         for i in range (shift_axis, -1, -1):
 */
  __pyx_t_1 = (((__pyx_v_ref_normal[__pyx_v_shift_axis]) > 0.0) != 0);
  if (__pyx_t_1) {

    /* "fatslimlib/core_base.pyx":610
 *         sign = 1
 *         if ref_normal[shift_axis] > 0:
 *             sign = -1             # <<<<<<<<<<<<<<
 *         for i in range (shift_axis, -1, -1):
 *             dx[i] += sign * self.c_pbcbox.box[shift_axis][i]
 */
    __pyx_v_sign = -1;

    /* "fatslimlib/core_base.pyx":609
 *         # make one shift in that direction to get the right distance
 *         sign = 1
 *         if ref_normal[shift_axis] > 0:             # <<<<<<<<<<<<<<
 *             sign = -1
 *         for i in range (shift_axis, -1, -1):
 */
  }

  /* "fatslimlib/core_base.pyx":611
 *         if ref_normal[shift_axis] > 0:
 *             sign = -1
 *         for i in range (shift_axis, -1, -1):             # <<<<<<<<<<<<<<
 *             dx[i] += sign * self.c_pbcbox.box[shift_axis][i]
 * 
 */
  for (__pyx_t_2 = __pyx_v_shift_axis; __pyx_t_2 > -1; __pyx_t_2-=1) {
    __pyx_v_i = __pyx_t_2;

    /* "fatslimlib/core_base.pyx":612
 *             sign = -1
 *         for i in range (shift_axis, -1, -1):
 *             dx[i] += sign * self.c_pbcbox.box[shift_axis][i]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_v_i;
    (__pyx_v_dx[__pyx_t_3]) = ((__pyx_v_dx[__pyx_t_3]) + (__pyx_v_sign * ((__pyx_v_self->c_pbcbox.box[__pyx_v_shift_axis])[__pyx_v_i])));
  }

  /* "fatslimlib/core_base.pyx":583
 *         fast_pbc_dx_c(self.c_pbcbox, ref, other, dx)
 * 
 *     cdef void fast_pbc_dx_leaflet(self, rvec ref, rvec other, rvec dx, rvec ref_normal) nogil:             # <<<<<<<<<<<<<<
 *         cdef fsl_int i, j, xoffset, yoffset, zoffset, notset=1
 *         cdef rvec dx_raw, dx_try, offset, offset_x, offset_y, offset_z
 */

  /* function exit code */
  __pyx_L0:;
}

/* "fatslimlib/core_base.pyx":615
 * 
 * 
 *     def dx(self, real[:] a_memview, real[:] b_memview):             # <<<<<<<<<<<<<<
 *         cdef rvec dx
 *         cdef real[:] dx_memview = np.empty(DIM)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_9dx(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_9dx(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_a_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_b_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dx (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_a_memview,&__pyx_n_s_b_memview,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_a_memview)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b_memview)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dx", 1, 2, 2, 1); __PYX_ERR(0, 615, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dx") < 0)) __PYX_ERR(0, 615, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_a_memview = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_a_memview.memview)) __PYX_ERR(0, 615, __pyx_L3_error)
    __pyx_v_b_memview = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_b_memview.memview)) __PYX_ERR(0, 615, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dx", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 615, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.dx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6PBCBox_8dx(((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_v_self), __pyx_v_a_memview, __pyx_v_b_memview);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_8dx(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_a_memview, __Pyx_memviewslice __pyx_v_b_memview) {
  rvec __pyx_v_dx;
  __Pyx_memviewslice __pyx_v_dx_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_memviewslice __pyx_t_4 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("dx", 0);

  /* "fatslimlib/core_base.pyx":617
 *     def dx(self, real[:] a_memview, real[:] b_memview):
 *         cdef rvec dx
 *         cdef real[:] dx_memview = np.empty(DIM)             # <<<<<<<<<<<<<<
 * 
 *         self.fast_pbc_dx(<rvec> &a_memview[0], <rvec> &b_memview[0], dx)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_int_3) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_int_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(0, 617, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_dx_memview = __pyx_t_4;
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;

  /* "fatslimlib/core_base.pyx":619
 *         cdef real[:] dx_memview = np.empty(DIM)
 * 
 *         self.fast_pbc_dx(<rvec> &a_memview[0], <rvec> &b_memview[0], dx)             # <<<<<<<<<<<<<<
 *         rvec_to_memview(dx, dx_memview)
 * 
 */
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  ((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_self->__pyx_vtab)->fast_pbc_dx(__pyx_v_self, ((real *)(&(*((real *) ( /* dim=0 */ (__pyx_v_a_memview.data + __pyx_t_5 * __pyx_v_a_memview.strides[0]) ))))), ((real *)(&(*((real *) ( /* dim=0 */ (__pyx_v_b_memview.data + __pyx_t_6 * __pyx_v_b_memview.strides[0]) ))))), __pyx_v_dx);

  /* "fatslimlib/core_base.pyx":620
 * 
 *         self.fast_pbc_dx(<rvec> &a_memview[0], <rvec> &b_memview[0], dx)
 *         rvec_to_memview(dx, dx_memview)             # <<<<<<<<<<<<<<
 * 
 *         return np.asarray(dx_memview)
 */
  __pyx_f_10fatslimlib_9core_base_rvec_to_memview(__pyx_v_dx, __pyx_v_dx_memview);

  /* "fatslimlib/core_base.pyx":622
 *         rvec_to_memview(dx, dx_memview)
 * 
 *         return np.asarray(dx_memview)             # <<<<<<<<<<<<<<
 * 
 *     def dx_leaflet(self, real[:] a_memview, real[:] b_memview, real[:] normal_memview):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_asarray); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dx_memview, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_7, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":615
 * 
 * 
 *     def dx(self, real[:] a_memview, real[:] b_memview):             # <<<<<<<<<<<<<<
 *         cdef rvec dx
 *         cdef real[:] dx_memview = np.empty(DIM)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __PYX_XDEC_MEMVIEW(&__pyx_t_4, 1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.dx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_dx_memview, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_a_memview, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_b_memview, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":624
 *         return np.asarray(dx_memview)
 * 
 *     def dx_leaflet(self, real[:] a_memview, real[:] b_memview, real[:] normal_memview):             # <<<<<<<<<<<<<<
 *         cdef rvec dx
 *         cdef real[:] dx_memview = np.empty(DIM)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_11dx_leaflet(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_11dx_leaflet(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_a_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_b_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_normal_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dx_leaflet (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_a_memview,&__pyx_n_s_b_memview,&__pyx_n_s_normal_memview,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_a_memview)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b_memview)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dx_leaflet", 1, 3, 3, 1); __PYX_ERR(0, 624, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_normal_memview)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dx_leaflet", 1, 3, 3, 2); __PYX_ERR(0, 624, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dx_leaflet") < 0)) __PYX_ERR(0, 624, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_a_memview = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_a_memview.memview)) __PYX_ERR(0, 624, __pyx_L3_error)
    __pyx_v_b_memview = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_b_memview.memview)) __PYX_ERR(0, 624, __pyx_L3_error)
    __pyx_v_normal_memview = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_normal_memview.memview)) __PYX_ERR(0, 624, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dx_leaflet", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 624, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.dx_leaflet", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6PBCBox_10dx_leaflet(((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_v_self), __pyx_v_a_memview, __pyx_v_b_memview, __pyx_v_normal_memview);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_10dx_leaflet(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_a_memview, __Pyx_memviewslice __pyx_v_b_memview, __Pyx_memviewslice __pyx_v_normal_memview) {
  rvec __pyx_v_dx;
  __Pyx_memviewslice __pyx_v_dx_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_memviewslice __pyx_t_4 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("dx_leaflet", 0);

  /* "fatslimlib/core_base.pyx":626
 *     def dx_leaflet(self, real[:] a_memview, real[:] b_memview, real[:] normal_memview):
 *         cdef rvec dx
 *         cdef real[:] dx_memview = np.empty(DIM)             # <<<<<<<<<<<<<<
 * 
 *         self.fast_pbc_dx_leaflet(<rvec> &a_memview[0], <rvec> &b_memview[0],
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_int_3) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_int_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(0, 626, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_dx_memview = __pyx_t_4;
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;

  /* "fatslimlib/core_base.pyx":628
 *         cdef real[:] dx_memview = np.empty(DIM)
 * 
 *         self.fast_pbc_dx_leaflet(<rvec> &a_memview[0], <rvec> &b_memview[0],             # <<<<<<<<<<<<<<
 *                                  dx, <rvec> &normal_memview[0])
 *         rvec_to_memview(dx, dx_memview)
 */
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;

  /* "fatslimlib/core_base.pyx":629
 * 
 *         self.fast_pbc_dx_leaflet(<rvec> &a_memview[0], <rvec> &b_memview[0],
 *                                  dx, <rvec> &normal_memview[0])             # <<<<<<<<<<<<<<
 *         rvec_to_memview(dx, dx_memview)
 * 
 */
  __pyx_t_7 = 0;

  /* "fatslimlib/core_base.pyx":628
 *         cdef real[:] dx_memview = np.empty(DIM)
 * 
 *         self.fast_pbc_dx_leaflet(<rvec> &a_memview[0], <rvec> &b_memview[0],             # <<<<<<<<<<<<<<
 *                                  dx, <rvec> &normal_memview[0])
 *         rvec_to_memview(dx, dx_memview)
 */
  ((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_self->__pyx_vtab)->fast_pbc_dx_leaflet(__pyx_v_self, ((real *)(&(*((real *) ( /* dim=0 */ (__pyx_v_a_memview.data + __pyx_t_5 * __pyx_v_a_memview.strides[0]) ))))), ((real *)(&(*((real *) ( /* dim=0 */ (__pyx_v_b_memview.data + __pyx_t_6 * __pyx_v_b_memview.strides[0]) ))))), __pyx_v_dx, ((real *)(&(*((real *) ( /* dim=0 */ (__pyx_v_normal_memview.data + __pyx_t_7 * __pyx_v_normal_memview.strides[0]) ))))));

  /* "fatslimlib/core_base.pyx":630
 *         self.fast_pbc_dx_leaflet(<rvec> &a_memview[0], <rvec> &b_memview[0],
 *                                  dx, <rvec> &normal_memview[0])
 *         rvec_to_memview(dx, dx_memview)             # <<<<<<<<<<<<<<
 * 
 *         return np.asarray(dx_memview)
 */
  __pyx_f_10fatslimlib_9core_base_rvec_to_memview(__pyx_v_dx, __pyx_v_dx_memview);

  /* "fatslimlib/core_base.pyx":632
 *         rvec_to_memview(dx, dx_memview)
 * 
 *         return np.asarray(dx_memview)             # <<<<<<<<<<<<<<
 * 
 *     cdef real fast_distance2(self, rvec a, rvec b) nogil:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_asarray); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dx_memview, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_8, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":624
 *         return np.asarray(dx_memview)
 * 
 *     def dx_leaflet(self, real[:] a_memview, real[:] b_memview, real[:] normal_memview):             # <<<<<<<<<<<<<<
 *         cdef rvec dx
 *         cdef real[:] dx_memview = np.empty(DIM)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __PYX_XDEC_MEMVIEW(&__pyx_t_4, 1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.dx_leaflet", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_dx_memview, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_a_memview, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_b_memview, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_normal_memview, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":634
 *         return np.asarray(dx_memview)
 * 
 *     cdef real fast_distance2(self, rvec a, rvec b) nogil:             # <<<<<<<<<<<<<<
 *         cdef rvec dx
 *         self.fast_pbc_dx(a, b, dx)
 */

static real __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_distance2(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, real *__pyx_v_a, real *__pyx_v_b) {
  rvec __pyx_v_dx;
  real __pyx_r;

  /* "fatslimlib/core_base.pyx":636
 *     cdef real fast_distance2(self, rvec a, rvec b) nogil:
 *         cdef rvec dx
 *         self.fast_pbc_dx(a, b, dx)             # <<<<<<<<<<<<<<
 *         return rvec_norm2(dx)
 * 
 */
  ((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_self->__pyx_vtab)->fast_pbc_dx(__pyx_v_self, __pyx_v_a, __pyx_v_b, __pyx_v_dx);

  /* "fatslimlib/core_base.pyx":637
 *         cdef rvec dx
 *         self.fast_pbc_dx(a, b, dx)
 *         return rvec_norm2(dx)             # <<<<<<<<<<<<<<
 * 
 *     cdef real fast_distance(self, rvec a, rvec b) nogil:
 */
  __pyx_r = rvec_norm2(__pyx_v_dx);
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":634
 *         return np.asarray(dx_memview)
 * 
 *     cdef real fast_distance2(self, rvec a, rvec b) nogil:             # <<<<<<<<<<<<<<
 *         cdef rvec dx
 *         self.fast_pbc_dx(a, b, dx)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":639
 *         return rvec_norm2(dx)
 * 
 *     cdef real fast_distance(self, rvec a, rvec b) nogil:             # <<<<<<<<<<<<<<
 *         return sqrt(self.fast_distance2(a,b))
 * 
 */

static real __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_distance(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, real *__pyx_v_a, real *__pyx_v_b) {
  real __pyx_r;

  /* "fatslimlib/core_base.pyx":640
 * 
 *     cdef real fast_distance(self, rvec a, rvec b) nogil:
 *         return sqrt(self.fast_distance2(a,b))             # <<<<<<<<<<<<<<
 * 
 *     cdef real fast_leaflet_distance2(self, rvec a, rvec b, rvec normal) nogil:
 */
  __pyx_r = sqrt(((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_self->__pyx_vtab)->fast_distance2(__pyx_v_self, __pyx_v_a, __pyx_v_b));
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":639
 *         return rvec_norm2(dx)
 * 
 *     cdef real fast_distance(self, rvec a, rvec b) nogil:             # <<<<<<<<<<<<<<
 *         return sqrt(self.fast_distance2(a,b))
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":642
 *         return sqrt(self.fast_distance2(a,b))
 * 
 *     cdef real fast_leaflet_distance2(self, rvec a, rvec b, rvec normal) nogil:             # <<<<<<<<<<<<<<
 *         cdef rvec dx
 *         self.fast_pbc_dx_leaflet(a, b, dx, normal)
 */

static real __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_leaflet_distance2(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, real *__pyx_v_a, real *__pyx_v_b, real *__pyx_v_normal) {
  rvec __pyx_v_dx;
  real __pyx_r;

  /* "fatslimlib/core_base.pyx":644
 *     cdef real fast_leaflet_distance2(self, rvec a, rvec b, rvec normal) nogil:
 *         cdef rvec dx
 *         self.fast_pbc_dx_leaflet(a, b, dx, normal)             # <<<<<<<<<<<<<<
 *         return rvec_norm2(dx)
 * 
 */
  ((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_self->__pyx_vtab)->fast_pbc_dx_leaflet(__pyx_v_self, __pyx_v_a, __pyx_v_b, __pyx_v_dx, __pyx_v_normal);

  /* "fatslimlib/core_base.pyx":645
 *         cdef rvec dx
 *         self.fast_pbc_dx_leaflet(a, b, dx, normal)
 *         return rvec_norm2(dx)             # <<<<<<<<<<<<<<
 * 
 *     cdef real fast_leaflet_distance(self, rvec a, rvec b, rvec normal) nogil:
 */
  __pyx_r = rvec_norm2(__pyx_v_dx);
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":642
 *         return sqrt(self.fast_distance2(a,b))
 * 
 *     cdef real fast_leaflet_distance2(self, rvec a, rvec b, rvec normal) nogil:             # <<<<<<<<<<<<<<
 *         cdef rvec dx
 *         self.fast_pbc_dx_leaflet(a, b, dx, normal)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":647
 *         return rvec_norm2(dx)
 * 
 *     cdef real fast_leaflet_distance(self, rvec a, rvec b, rvec normal) nogil:             # <<<<<<<<<<<<<<
 *         return sqrt(self.fast_leaflet_distance2(a,b,normal))
 * 
 */

static real __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_leaflet_distance(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, real *__pyx_v_a, real *__pyx_v_b, real *__pyx_v_normal) {
  real __pyx_r;

  /* "fatslimlib/core_base.pyx":648
 * 
 *     cdef real fast_leaflet_distance(self, rvec a, rvec b, rvec normal) nogil:
 *         return sqrt(self.fast_leaflet_distance2(a,b,normal))             # <<<<<<<<<<<<<<
 * 
 *     def distance2(self, real[:] a_memview, real[:] b_memview):
 */
  __pyx_r = sqrt(((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_self->__pyx_vtab)->fast_leaflet_distance2(__pyx_v_self, __pyx_v_a, __pyx_v_b, __pyx_v_normal));
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":647
 *         return rvec_norm2(dx)
 * 
 *     cdef real fast_leaflet_distance(self, rvec a, rvec b, rvec normal) nogil:             # <<<<<<<<<<<<<<
 *         return sqrt(self.fast_leaflet_distance2(a,b,normal))
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":650
 *         return sqrt(self.fast_leaflet_distance2(a,b,normal))
 * 
 *     def distance2(self, real[:] a_memview, real[:] b_memview):             # <<<<<<<<<<<<<<
 *         return self.fast_distance2(<rvec> &a_memview[0], <rvec> &b_memview[0])
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_13distance2(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_13distance2(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_a_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_b_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("distance2 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_a_memview,&__pyx_n_s_b_memview,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_a_memview)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b_memview)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("distance2", 1, 2, 2, 1); __PYX_ERR(0, 650, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "distance2") < 0)) __PYX_ERR(0, 650, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_a_memview = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_a_memview.memview)) __PYX_ERR(0, 650, __pyx_L3_error)
    __pyx_v_b_memview = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_b_memview.memview)) __PYX_ERR(0, 650, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("distance2", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 650, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.distance2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6PBCBox_12distance2(((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_v_self), __pyx_v_a_memview, __pyx_v_b_memview);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_12distance2(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_a_memview, __Pyx_memviewslice __pyx_v_b_memview) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("distance2", 0);

  /* "fatslimlib/core_base.pyx":651
 * 
 *     def distance2(self, real[:] a_memview, real[:] b_memview):
 *         return self.fast_distance2(<rvec> &a_memview[0], <rvec> &b_memview[0])             # <<<<<<<<<<<<<<
 * 
 *     def distance(self, real[:] a_memview, real[:] b_memview):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = PyFloat_FromDouble(((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_self->__pyx_vtab)->fast_distance2(__pyx_v_self, ((real *)(&(*((real *) ( /* dim=0 */ (__pyx_v_a_memview.data + __pyx_t_1 * __pyx_v_a_memview.strides[0]) ))))), ((real *)(&(*((real *) ( /* dim=0 */ (__pyx_v_b_memview.data + __pyx_t_2 * __pyx_v_b_memview.strides[0]) ))))))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":650
 *         return sqrt(self.fast_leaflet_distance2(a,b,normal))
 * 
 *     def distance2(self, real[:] a_memview, real[:] b_memview):             # <<<<<<<<<<<<<<
 *         return self.fast_distance2(<rvec> &a_memview[0], <rvec> &b_memview[0])
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.distance2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_a_memview, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_b_memview, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":653
 *         return self.fast_distance2(<rvec> &a_memview[0], <rvec> &b_memview[0])
 * 
 *     def distance(self, real[:] a_memview, real[:] b_memview):             # <<<<<<<<<<<<<<
 *         return sqrt(self.distance2(a_memview, b_memview))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_15distance(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_15distance(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_a_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_b_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("distance (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_a_memview,&__pyx_n_s_b_memview,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_a_memview)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b_memview)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("distance", 1, 2, 2, 1); __PYX_ERR(0, 653, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "distance") < 0)) __PYX_ERR(0, 653, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_a_memview = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_a_memview.memview)) __PYX_ERR(0, 653, __pyx_L3_error)
    __pyx_v_b_memview = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_b_memview.memview)) __PYX_ERR(0, 653, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("distance", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 653, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.distance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6PBCBox_14distance(((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_v_self), __pyx_v_a_memview, __pyx_v_b_memview);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_14distance(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_a_memview, __Pyx_memviewslice __pyx_v_b_memview) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  double __pyx_t_8;
  __Pyx_RefNannySetupContext("distance", 0);

  /* "fatslimlib/core_base.pyx":654
 * 
 *     def distance(self, real[:] a_memview, real[:] b_memview):
 *         return sqrt(self.distance2(a_memview, b_memview))             # <<<<<<<<<<<<<<
 * 
 *     def leaflet_distance2(self, real[:] a_memview, real[:] b_memview, real[:] normal_memview):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_distance2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_a_memview, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_b_memview, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 654, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 654, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 654, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_4);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 654, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyFloat_FromDouble(sqrt(__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":653
 *         return self.fast_distance2(<rvec> &a_memview[0], <rvec> &b_memview[0])
 * 
 *     def distance(self, real[:] a_memview, real[:] b_memview):             # <<<<<<<<<<<<<<
 *         return sqrt(self.distance2(a_memview, b_memview))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.distance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_a_memview, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_b_memview, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":656
 *         return sqrt(self.distance2(a_memview, b_memview))
 * 
 *     def leaflet_distance2(self, real[:] a_memview, real[:] b_memview, real[:] normal_memview):             # <<<<<<<<<<<<<<
 *         return self.fast_leaflet_distance2(<rvec> &a_memview[0], <rvec> &b_memview[0],
 *                                            <rvec> &normal_memview[0])
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_17leaflet_distance2(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_17leaflet_distance2(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_a_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_b_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_normal_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("leaflet_distance2 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_a_memview,&__pyx_n_s_b_memview,&__pyx_n_s_normal_memview,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_a_memview)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b_memview)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("leaflet_distance2", 1, 3, 3, 1); __PYX_ERR(0, 656, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_normal_memview)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("leaflet_distance2", 1, 3, 3, 2); __PYX_ERR(0, 656, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "leaflet_distance2") < 0)) __PYX_ERR(0, 656, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_a_memview = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_a_memview.memview)) __PYX_ERR(0, 656, __pyx_L3_error)
    __pyx_v_b_memview = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_b_memview.memview)) __PYX_ERR(0, 656, __pyx_L3_error)
    __pyx_v_normal_memview = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_normal_memview.memview)) __PYX_ERR(0, 656, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("leaflet_distance2", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 656, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.leaflet_distance2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6PBCBox_16leaflet_distance2(((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_v_self), __pyx_v_a_memview, __pyx_v_b_memview, __pyx_v_normal_memview);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_16leaflet_distance2(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_a_memview, __Pyx_memviewslice __pyx_v_b_memview, __Pyx_memviewslice __pyx_v_normal_memview) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("leaflet_distance2", 0);

  /* "fatslimlib/core_base.pyx":657
 * 
 *     def leaflet_distance2(self, real[:] a_memview, real[:] b_memview, real[:] normal_memview):
 *         return self.fast_leaflet_distance2(<rvec> &a_memview[0], <rvec> &b_memview[0],             # <<<<<<<<<<<<<<
 *                                            <rvec> &normal_memview[0])
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":658
 *     def leaflet_distance2(self, real[:] a_memview, real[:] b_memview, real[:] normal_memview):
 *         return self.fast_leaflet_distance2(<rvec> &a_memview[0], <rvec> &b_memview[0],
 *                                            <rvec> &normal_memview[0])             # <<<<<<<<<<<<<<
 * 
 *     def leaflet_distance(self, real[:] a_memview, real[:] b_memview, real[:] normal_memview):
 */
  __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":657
 * 
 *     def leaflet_distance2(self, real[:] a_memview, real[:] b_memview, real[:] normal_memview):
 *         return self.fast_leaflet_distance2(<rvec> &a_memview[0], <rvec> &b_memview[0],             # <<<<<<<<<<<<<<
 *                                            <rvec> &normal_memview[0])
 * 
 */
  __pyx_t_4 = PyFloat_FromDouble(((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_self->__pyx_vtab)->fast_leaflet_distance2(__pyx_v_self, ((real *)(&(*((real *) ( /* dim=0 */ (__pyx_v_a_memview.data + __pyx_t_1 * __pyx_v_a_memview.strides[0]) ))))), ((real *)(&(*((real *) ( /* dim=0 */ (__pyx_v_b_memview.data + __pyx_t_2 * __pyx_v_b_memview.strides[0]) ))))), ((real *)(&(*((real *) ( /* dim=0 */ (__pyx_v_normal_memview.data + __pyx_t_3 * __pyx_v_normal_memview.strides[0]) ))))))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":656
 *         return sqrt(self.distance2(a_memview, b_memview))
 * 
 *     def leaflet_distance2(self, real[:] a_memview, real[:] b_memview, real[:] normal_memview):             # <<<<<<<<<<<<<<
 *         return self.fast_leaflet_distance2(<rvec> &a_memview[0], <rvec> &b_memview[0],
 *                                            <rvec> &normal_memview[0])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.leaflet_distance2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_a_memview, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_b_memview, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_normal_memview, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":660
 *                                            <rvec> &normal_memview[0])
 * 
 *     def leaflet_distance(self, real[:] a_memview, real[:] b_memview, real[:] normal_memview):             # <<<<<<<<<<<<<<
 *         return sqrt(self.leaflet_distance2(a_memview, b_memview, normal_memview))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_19leaflet_distance(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_19leaflet_distance(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  __Pyx_memviewslice __pyx_v_a_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_b_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_normal_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("leaflet_distance (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_a_memview,&__pyx_n_s_b_memview,&__pyx_n_s_normal_memview,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_a_memview)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b_memview)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("leaflet_distance", 1, 3, 3, 1); __PYX_ERR(0, 660, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_normal_memview)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("leaflet_distance", 1, 3, 3, 2); __PYX_ERR(0, 660, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "leaflet_distance") < 0)) __PYX_ERR(0, 660, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_a_memview = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(values[0], PyBUF_WRITABLE); if (unlikely(!__pyx_v_a_memview.memview)) __PYX_ERR(0, 660, __pyx_L3_error)
    __pyx_v_b_memview = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(values[1], PyBUF_WRITABLE); if (unlikely(!__pyx_v_b_memview.memview)) __PYX_ERR(0, 660, __pyx_L3_error)
    __pyx_v_normal_memview = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(values[2], PyBUF_WRITABLE); if (unlikely(!__pyx_v_normal_memview.memview)) __PYX_ERR(0, 660, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("leaflet_distance", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 660, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.leaflet_distance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6PBCBox_18leaflet_distance(((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_v_self), __pyx_v_a_memview, __pyx_v_b_memview, __pyx_v_normal_memview);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_18leaflet_distance(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_a_memview, __Pyx_memviewslice __pyx_v_b_memview, __Pyx_memviewslice __pyx_v_normal_memview) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  double __pyx_t_9;
  __Pyx_RefNannySetupContext("leaflet_distance", 0);

  /* "fatslimlib/core_base.pyx":661
 * 
 *     def leaflet_distance(self, real[:] a_memview, real[:] b_memview, real[:] normal_memview):
 *         return sqrt(self.leaflet_distance2(a_memview, b_memview, normal_memview))             # <<<<<<<<<<<<<<
 * 
 *     cdef void fast_pbc_xcm(self, real[:, ::1] coords, rvec xcm, fsl_int limit=NOTSET) nogil:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_leaflet_distance2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_a_memview, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_b_memview, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_normal_memview, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_t_3, __pyx_t_4, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 661, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_t_3, __pyx_t_4, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 661, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  {
    __pyx_t_8 = PyTuple_New(3+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 661, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_8, 2+__pyx_t_7, __pyx_t_5);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 661, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyFloat_FromDouble(sqrt(__pyx_t_9)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":660
 *                                            <rvec> &normal_memview[0])
 * 
 *     def leaflet_distance(self, real[:] a_memview, real[:] b_memview, real[:] normal_memview):             # <<<<<<<<<<<<<<
 *         return sqrt(self.leaflet_distance2(a_memview, b_memview, normal_memview))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.leaflet_distance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_a_memview, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_b_memview, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_normal_memview, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":663
 *         return sqrt(self.leaflet_distance2(a_memview, b_memview, normal_memview))
 * 
 *     cdef void fast_pbc_xcm(self, real[:, ::1] coords, rvec xcm, fsl_int limit=NOTSET) nogil:             # <<<<<<<<<<<<<<
 *         self.fast_pbc_xcm_from_ref(coords, &coords[0, XX], xcm, limit)
 * 
 */

static void __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_coords, real *__pyx_v_xcm, struct __pyx_opt_args_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm *__pyx_optional_args) {
  fsl_int __pyx_v_limit = ((fsl_int)-12345);
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  struct __pyx_opt_args_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm_from_ref __pyx_t_3;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_limit = __pyx_optional_args->limit;
    }
  }

  /* "fatslimlib/core_base.pyx":664
 * 
 *     cdef void fast_pbc_xcm(self, real[:, ::1] coords, rvec xcm, fsl_int limit=NOTSET) nogil:
 *         self.fast_pbc_xcm_from_ref(coords, &coords[0, XX], xcm, limit)             # <<<<<<<<<<<<<<
 * 
 *     cdef void fast_pbc_xcm_from_ref(self, real[:, ::1] coords, rvec ref,
 */
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3.__pyx_n = 1;
  __pyx_t_3.limit = __pyx_v_limit;
  ((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_self->__pyx_vtab)->fast_pbc_xcm_from_ref(__pyx_v_self, __pyx_v_coords, (&(*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_1 * __pyx_v_coords.strides[0]) )) + __pyx_t_2)) )))), __pyx_v_xcm, &__pyx_t_3); 

  /* "fatslimlib/core_base.pyx":663
 *         return sqrt(self.leaflet_distance2(a_memview, b_memview, normal_memview))
 * 
 *     cdef void fast_pbc_xcm(self, real[:, ::1] coords, rvec xcm, fsl_int limit=NOTSET) nogil:             # <<<<<<<<<<<<<<
 *         self.fast_pbc_xcm_from_ref(coords, &coords[0, XX], xcm, limit)
 * 
 */

  /* function exit code */
}

/* "fatslimlib/core_base.pyx":666
 *         self.fast_pbc_xcm_from_ref(coords, &coords[0, XX], xcm, limit)
 * 
 *     cdef void fast_pbc_xcm_from_ref(self, real[:, ::1] coords, rvec ref,             # <<<<<<<<<<<<<<
 *                                     rvec xcm, fsl_int limit=NOTSET) nogil:
 *         fast_pbc_xcm_from_ref_c(self.c_pbcbox, coords, ref, xcm, limit)
 */

static void __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm_from_ref(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_coords, real *__pyx_v_ref, real *__pyx_v_xcm, struct __pyx_opt_args_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm_from_ref *__pyx_optional_args) {
  fsl_int __pyx_v_limit = ((fsl_int)-12345);
  struct __pyx_opt_args_10fatslimlib_9core_base_fast_pbc_xcm_from_ref_c __pyx_t_1;
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_limit = __pyx_optional_args->limit;
    }
  }

  /* "fatslimlib/core_base.pyx":668
 *     cdef void fast_pbc_xcm_from_ref(self, real[:, ::1] coords, rvec ref,
 *                                     rvec xcm, fsl_int limit=NOTSET) nogil:
 *         fast_pbc_xcm_from_ref_c(self.c_pbcbox, coords, ref, xcm, limit)             # <<<<<<<<<<<<<<
 * 
 *     def get_xcm(self, real[:, ::1] coords):
 */
  __pyx_t_1.__pyx_n = 1;
  __pyx_t_1.first = __pyx_v_limit;
  __pyx_f_10fatslimlib_9core_base_fast_pbc_xcm_from_ref_c(__pyx_v_self->c_pbcbox, __pyx_v_coords, __pyx_v_ref, __pyx_v_xcm, &__pyx_t_1); 

  /* "fatslimlib/core_base.pyx":666
 *         self.fast_pbc_xcm_from_ref(coords, &coords[0, XX], xcm, limit)
 * 
 *     cdef void fast_pbc_xcm_from_ref(self, real[:, ::1] coords, rvec ref,             # <<<<<<<<<<<<<<
 *                                     rvec xcm, fsl_int limit=NOTSET) nogil:
 *         fast_pbc_xcm_from_ref_c(self.c_pbcbox, coords, ref, xcm, limit)
 */

  /* function exit code */
}

/* "fatslimlib/core_base.pyx":670
 *         fast_pbc_xcm_from_ref_c(self.c_pbcbox, coords, ref, xcm, limit)
 * 
 *     def get_xcm(self, real[:, ::1] coords):             # <<<<<<<<<<<<<<
 *         cdef rvec xcm
 *         self.fast_pbc_xcm(coords, xcm)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_21get_xcm(PyObject *__pyx_v_self, PyObject *__pyx_arg_coords); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_21get_xcm(PyObject *__pyx_v_self, PyObject *__pyx_arg_coords) {
  __Pyx_memviewslice __pyx_v_coords = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_xcm (wrapper)", 0);
  assert(__pyx_arg_coords); {
    __pyx_v_coords = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(__pyx_arg_coords, PyBUF_WRITABLE); if (unlikely(!__pyx_v_coords.memview)) __PYX_ERR(0, 670, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.get_xcm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6PBCBox_20get_xcm(((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_v_self), __pyx_v_coords);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_20get_xcm(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_coords) {
  rvec __pyx_v_xcm;
  PyObject *__pyx_v_xcm_py = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_memviewslice __pyx_t_4 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannySetupContext("get_xcm", 0);

  /* "fatslimlib/core_base.pyx":672
 *     def get_xcm(self, real[:, ::1] coords):
 *         cdef rvec xcm
 *         self.fast_pbc_xcm(coords, xcm)             # <<<<<<<<<<<<<<
 * 
 *         xcm_py = np.empty(DIM)
 */
  ((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_self->__pyx_vtab)->fast_pbc_xcm(__pyx_v_self, __pyx_v_coords, __pyx_v_xcm, NULL);

  /* "fatslimlib/core_base.pyx":674
 *         self.fast_pbc_xcm(coords, xcm)
 * 
 *         xcm_py = np.empty(DIM)             # <<<<<<<<<<<<<<
 *         rvec_to_memview(xcm, xcm_py)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_int_3) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_int_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_xcm_py = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":675
 * 
 *         xcm_py = np.empty(DIM)
 *         rvec_to_memview(xcm, xcm_py)             # <<<<<<<<<<<<<<
 * 
 *         return xcm_py
 */
  __pyx_t_4 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(__pyx_v_xcm_py, PyBUF_WRITABLE); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(0, 675, __pyx_L1_error)
  __pyx_f_10fatslimlib_9core_base_rvec_to_memview(__pyx_v_xcm, __pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_4, 1);
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;

  /* "fatslimlib/core_base.pyx":677
 *         rvec_to_memview(xcm, xcm_py)
 * 
 *         return xcm_py             # <<<<<<<<<<<<<<
 * 
 *     def asarray(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_xcm_py);
  __pyx_r = __pyx_v_xcm_py;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":670
 *         fast_pbc_xcm_from_ref_c(self.c_pbcbox, coords, ref, xcm, limit)
 * 
 *     def get_xcm(self, real[:, ::1] coords):             # <<<<<<<<<<<<<<
 *         cdef rvec xcm
 *         self.fast_pbc_xcm(coords, xcm)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __PYX_XDEC_MEMVIEW(&__pyx_t_4, 1);
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.get_xcm", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_coords, 1);
  __Pyx_XDECREF(__pyx_v_xcm_py);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":679
 *         return xcm_py
 * 
 *     def asarray(self):             # <<<<<<<<<<<<<<
 *         cdef fsl_int i, j
 *         box = np.empty((DIM, DIM))
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_23asarray(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_23asarray(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("asarray (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6PBCBox_22asarray(((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_22asarray(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self) {
  fsl_int __pyx_v_i;
  fsl_int __pyx_v_j;
  PyObject *__pyx_v_box = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  fsl_int __pyx_t_4;
  fsl_int __pyx_t_5;
  __Pyx_RefNannySetupContext("asarray", 0);

  /* "fatslimlib/core_base.pyx":681
 *     def asarray(self):
 *         cdef fsl_int i, j
 *         box = np.empty((DIM, DIM))             # <<<<<<<<<<<<<<
 * 
 *         for i in range(DIM):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_tuple__14) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_tuple__14);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_box = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":683
 *         box = np.empty((DIM, DIM))
 * 
 *         for i in range(DIM):             # <<<<<<<<<<<<<<
 *             for j in range(DIM):
 *                 box[i][j] = self.c_pbcbox.box[i][j]
 */
  for (__pyx_t_4 = 0; __pyx_t_4 < 3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "fatslimlib/core_base.pyx":684
 * 
 *         for i in range(DIM):
 *             for j in range(DIM):             # <<<<<<<<<<<<<<
 *                 box[i][j] = self.c_pbcbox.box[i][j]
 * 
 */
    for (__pyx_t_5 = 0; __pyx_t_5 < 3; __pyx_t_5+=1) {
      __pyx_v_j = __pyx_t_5;

      /* "fatslimlib/core_base.pyx":685
 *         for i in range(DIM):
 *             for j in range(DIM):
 *                 box[i][j] = self.c_pbcbox.box[i][j]             # <<<<<<<<<<<<<<
 * 
 *         return box
 */
      __pyx_t_1 = PyFloat_FromDouble(((__pyx_v_self->c_pbcbox.box[__pyx_v_i])[__pyx_v_j])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 685, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_box, __pyx_v_i, fsl_int, 1, __Pyx_PyInt_From_fsl_int, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 685, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (unlikely(__Pyx_SetItemInt(__pyx_t_3, __pyx_v_j, __pyx_t_1, fsl_int, 1, __Pyx_PyInt_From_fsl_int, 0, 0, 0) < 0)) __PYX_ERR(0, 685, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
  }

  /* "fatslimlib/core_base.pyx":687
 *                 box[i][j] = self.c_pbcbox.box[i][j]
 * 
 *         return box             # <<<<<<<<<<<<<<
 * 
 *     cdef real[:, ::1]fast_put_atoms_in_bbox(self, real[:,::1] coords) nogil:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_box);
  __pyx_r = __pyx_v_box;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":679
 *         return xcm_py
 * 
 *     def asarray(self):             # <<<<<<<<<<<<<<
 *         cdef fsl_int i, j
 *         box = np.empty((DIM, DIM))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.asarray", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_box);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":689
 *         return box
 * 
 *     cdef real[:, ::1]fast_put_atoms_in_bbox(self, real[:,::1] coords) nogil:             # <<<<<<<<<<<<<<
 *         cdef fsl_int i, m, d, natoms, wd = 0
 *         cdef real[:,::1] bbox_coords
 */

static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_6PBCBox_fast_put_atoms_in_bbox(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_coords) {
  fsl_int __pyx_v_i;
  fsl_int __pyx_v_m;
  fsl_int __pyx_v_d;
  fsl_int __pyx_v_natoms;
  CYTHON_UNUSED fsl_int __pyx_v_wd;
  __Pyx_memviewslice __pyx_v_bbox_coords = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_memviewslice __pyx_t_5 = { 0, 0, { 0 }, { 0 }, { 0 } };
  fsl_int __pyx_t_6;
  fsl_int __pyx_t_7;
  fsl_int __pyx_t_8;
  fsl_int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  long __pyx_t_12;
  long __pyx_t_13;
  fsl_int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("fast_put_atoms_in_bbox", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "fatslimlib/core_base.pyx":690
 * 
 *     cdef real[:, ::1]fast_put_atoms_in_bbox(self, real[:,::1] coords) nogil:
 *         cdef fsl_int i, m, d, natoms, wd = 0             # <<<<<<<<<<<<<<
 *         cdef real[:,::1] bbox_coords
 * 
 */
  /*try:*/ {
    __pyx_v_wd = 0;

    /* "fatslimlib/core_base.pyx":693
 *         cdef real[:,::1] bbox_coords
 * 
 *         natoms = coords.shape[0]             # <<<<<<<<<<<<<<
 *         with gil:
 *             if natoms == 0:
 */
    __pyx_v_natoms = (__pyx_v_coords.shape[0]);

    /* "fatslimlib/core_base.pyx":694
 * 
 *         natoms = coords.shape[0]
 *         with gil:             # <<<<<<<<<<<<<<
 *             if natoms == 0:
 *                 bbox_coords = np.empty((0, DIM))
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "fatslimlib/core_base.pyx":695
 *         natoms = coords.shape[0]
 *         with gil:
 *             if natoms == 0:             # <<<<<<<<<<<<<<
 *                 bbox_coords = np.empty((0, DIM))
 *             else:
 */
          __pyx_t_1 = ((__pyx_v_natoms == 0) != 0);
          if (__pyx_t_1) {

            /* "fatslimlib/core_base.pyx":696
 *         with gil:
 *             if natoms == 0:
 *                 bbox_coords = np.empty((0, DIM))             # <<<<<<<<<<<<<<
 *             else:
 *                 bbox_coords = coords.copy()
 */
            __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 696, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_empty); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 696, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_tuple__15) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_tuple__15);
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 696, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_5 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 696, __pyx_L7_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_v_bbox_coords = __pyx_t_5;
            __pyx_t_5.memview = NULL;
            __pyx_t_5.data = NULL;

            /* "fatslimlib/core_base.pyx":695
 *         natoms = coords.shape[0]
 *         with gil:
 *             if natoms == 0:             # <<<<<<<<<<<<<<
 *                 bbox_coords = np.empty((0, DIM))
 *             else:
 */
            goto __pyx_L9;
          }

          /* "fatslimlib/core_base.pyx":698
 *                 bbox_coords = np.empty((0, DIM))
 *             else:
 *                 bbox_coords = coords.copy()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(natoms):
 */
          /*else*/ {
            __pyx_t_5 = __pyx_memoryview_copy_slice_d_dc_nn_real_c(__pyx_v_coords); if (unlikely(!__pyx_t_5.memview)) __PYX_ERR(0, 698, __pyx_L7_error)
            __pyx_v_bbox_coords = __pyx_t_5;
            __pyx_t_5.memview = NULL;
            __pyx_t_5.data = NULL;
          }
          __pyx_L9:;
        }

        /* "fatslimlib/core_base.pyx":694
 * 
 *         natoms = coords.shape[0]
 *         with gil:             # <<<<<<<<<<<<<<
 *             if natoms == 0:
 *                 bbox_coords = np.empty((0, DIM))
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }

    /* "fatslimlib/core_base.pyx":700
 *                 bbox_coords = coords.copy()
 * 
 *         for i in range(natoms):             # <<<<<<<<<<<<<<
 *             for m in range(DIM - 1, -1, -1):
 *                 while bbox_coords[i, m] < 0:
 */
    __pyx_t_6 = __pyx_v_natoms;
    __pyx_t_7 = __pyx_t_6;
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
      __pyx_v_i = __pyx_t_8;

      /* "fatslimlib/core_base.pyx":701
 * 
 *         for i in range(natoms):
 *             for m in range(DIM - 1, -1, -1):             # <<<<<<<<<<<<<<
 *                 while bbox_coords[i, m] < 0:
 *                     for d in range(m+1):
 */
      for (__pyx_t_9 = 2; __pyx_t_9 > -1; __pyx_t_9-=1) {
        __pyx_v_m = __pyx_t_9;

        /* "fatslimlib/core_base.pyx":702
 *         for i in range(natoms):
 *             for m in range(DIM - 1, -1, -1):
 *                 while bbox_coords[i, m] < 0:             # <<<<<<<<<<<<<<
 *                     for d in range(m+1):
 *                         bbox_coords[i, d] += self.c_pbcbox.box[m][d]
 */
        while (1) {
          __pyx_t_10 = __pyx_v_i;
          __pyx_t_11 = __pyx_v_m;
          __pyx_t_1 = (((*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_bbox_coords.data + __pyx_t_10 * __pyx_v_bbox_coords.strides[0]) )) + __pyx_t_11)) ))) < 0.0) != 0);
          if (!__pyx_t_1) break;

          /* "fatslimlib/core_base.pyx":703
 *             for m in range(DIM - 1, -1, -1):
 *                 while bbox_coords[i, m] < 0:
 *                     for d in range(m+1):             # <<<<<<<<<<<<<<
 *                         bbox_coords[i, d] += self.c_pbcbox.box[m][d]
 *                 while bbox_coords[i, m] >= self.c_pbcbox.box[m][m]:
 */
          __pyx_t_12 = (__pyx_v_m + 1);
          __pyx_t_13 = __pyx_t_12;
          for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
            __pyx_v_d = __pyx_t_14;

            /* "fatslimlib/core_base.pyx":704
 *                 while bbox_coords[i, m] < 0:
 *                     for d in range(m+1):
 *                         bbox_coords[i, d] += self.c_pbcbox.box[m][d]             # <<<<<<<<<<<<<<
 *                 while bbox_coords[i, m] >= self.c_pbcbox.box[m][m]:
 *                     for d in range(m+1):
 */
            __pyx_t_15 = __pyx_v_i;
            __pyx_t_16 = __pyx_v_d;
            *((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_bbox_coords.data + __pyx_t_15 * __pyx_v_bbox_coords.strides[0]) )) + __pyx_t_16)) )) += ((__pyx_v_self->c_pbcbox.box[__pyx_v_m])[__pyx_v_d]);
          }
        }

        /* "fatslimlib/core_base.pyx":705
 *                     for d in range(m+1):
 *                         bbox_coords[i, d] += self.c_pbcbox.box[m][d]
 *                 while bbox_coords[i, m] >= self.c_pbcbox.box[m][m]:             # <<<<<<<<<<<<<<
 *                     for d in range(m+1):
 *                         bbox_coords[i, d] -= self.c_pbcbox.box[m][d]
 */
        while (1) {
          __pyx_t_17 = __pyx_v_i;
          __pyx_t_18 = __pyx_v_m;
          __pyx_t_1 = (((*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_bbox_coords.data + __pyx_t_17 * __pyx_v_bbox_coords.strides[0]) )) + __pyx_t_18)) ))) >= ((__pyx_v_self->c_pbcbox.box[__pyx_v_m])[__pyx_v_m])) != 0);
          if (!__pyx_t_1) break;

          /* "fatslimlib/core_base.pyx":706
 *                         bbox_coords[i, d] += self.c_pbcbox.box[m][d]
 *                 while bbox_coords[i, m] >= self.c_pbcbox.box[m][m]:
 *                     for d in range(m+1):             # <<<<<<<<<<<<<<
 *                         bbox_coords[i, d] -= self.c_pbcbox.box[m][d]
 *         return bbox_coords
 */
          __pyx_t_12 = (__pyx_v_m + 1);
          __pyx_t_13 = __pyx_t_12;
          for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
            __pyx_v_d = __pyx_t_14;

            /* "fatslimlib/core_base.pyx":707
 *                 while bbox_coords[i, m] >= self.c_pbcbox.box[m][m]:
 *                     for d in range(m+1):
 *                         bbox_coords[i, d] -= self.c_pbcbox.box[m][d]             # <<<<<<<<<<<<<<
 *         return bbox_coords
 * 
 */
            __pyx_t_19 = __pyx_v_i;
            __pyx_t_20 = __pyx_v_d;
            *((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_bbox_coords.data + __pyx_t_19 * __pyx_v_bbox_coords.strides[0]) )) + __pyx_t_20)) )) -= ((__pyx_v_self->c_pbcbox.box[__pyx_v_m])[__pyx_v_d]);
          }
        }
      }
    }

    /* "fatslimlib/core_base.pyx":708
 *                     for d in range(m+1):
 *                         bbox_coords[i, d] -= self.c_pbcbox.box[m][d]
 *         return bbox_coords             # <<<<<<<<<<<<<<
 * 
 *     def put_atoms_in_bbox(self, real[:,::1] coords):
 */
    __PYX_INC_MEMVIEW(&__pyx_v_bbox_coords, 1);
    __pyx_r = __pyx_v_bbox_coords;
    goto __pyx_L3_return;
  }

  /* "fatslimlib/core_base.pyx":690
 * 
 *     cdef real[:, ::1]fast_put_atoms_in_bbox(self, real[:,::1] coords) nogil:
 *         cdef fsl_int i, m, d, natoms, wd = 0             # <<<<<<<<<<<<<<
 *         cdef real[:,::1] bbox_coords
 * 
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "fatslimlib/core_base.pyx":689
 *         return box
 * 
 *     cdef real[:, ::1]fast_put_atoms_in_bbox(self, real[:,::1] coords) nogil:             # <<<<<<<<<<<<<<
 *         cdef fsl_int i, m, d, natoms, wd = 0
 *         cdef real[:,::1] bbox_coords
 */

  /* function exit code */
  __Pyx_pretend_to_initialize(&__pyx_r);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __PYX_XDEC_MEMVIEW(&__pyx_t_5, 0);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.fast_put_atoms_in_bbox", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_bbox_coords, 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":710
 *         return bbox_coords
 * 
 *     def put_atoms_in_bbox(self, real[:,::1] coords):             # <<<<<<<<<<<<<<
 *         return np.asarray(self.fast_put_atoms_in_bbox(coords))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_25put_atoms_in_bbox(PyObject *__pyx_v_self, PyObject *__pyx_arg_coords); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_25put_atoms_in_bbox(PyObject *__pyx_v_self, PyObject *__pyx_arg_coords) {
  __Pyx_memviewslice __pyx_v_coords = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("put_atoms_in_bbox (wrapper)", 0);
  assert(__pyx_arg_coords); {
    __pyx_v_coords = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(__pyx_arg_coords, PyBUF_WRITABLE); if (unlikely(!__pyx_v_coords.memview)) __PYX_ERR(0, 710, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.put_atoms_in_bbox", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6PBCBox_24put_atoms_in_bbox(((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_v_self), __pyx_v_coords);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_24put_atoms_in_bbox(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, __Pyx_memviewslice __pyx_v_coords) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_memviewslice __pyx_t_4 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("put_atoms_in_bbox", 0);

  /* "fatslimlib/core_base.pyx":711
 * 
 *     def put_atoms_in_bbox(self, real[:,::1] coords):
 *         return np.asarray(self.fast_put_atoms_in_bbox(coords))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_asarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_self->__pyx_vtab)->fast_put_atoms_in_bbox(__pyx_v_self, __pyx_v_coords); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(0, 711, __pyx_L1_error)
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_t_4, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __PYX_XDEC_MEMVIEW(&__pyx_t_4, 1);
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":710
 *         return bbox_coords
 * 
 *     def put_atoms_in_bbox(self, real[:,::1] coords):             # <<<<<<<<<<<<<<
 *         return np.asarray(self.fast_put_atoms_in_bbox(coords))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __PYX_XDEC_MEMVIEW(&__pyx_t_4, 1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.put_atoms_in_bbox", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_coords, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("Pickling of struct members such as self.c_pbcbox must be explicitly requested with @auto_pickle(True)")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_27__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_27__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6PBCBox_26__reduce_cython__(((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_26__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("Pickling of struct members such as self.c_pbcbox must be explicitly requested with @auto_pickle(True)")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("Pickling of struct members such as self.c_pbcbox must be explicitly requested with @auto_pickle(True)")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("Pickling of struct members such as self.c_pbcbox must be explicitly requested with @auto_pickle(True)")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("Pickling of struct members such as self.c_pbcbox must be explicitly requested with @auto_pickle(True)")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("Pickling of struct members such as self.c_pbcbox must be explicitly requested with @auto_pickle(True)")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_29__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_6PBCBox_29__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_6PBCBox_28__setstate_cython__(((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_6PBCBox_28__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("Pickling of struct members such as self.c_pbcbox must be explicitly requested with @auto_pickle(True)")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("Pickling of struct members such as self.c_pbcbox must be explicitly requested with @auto_pickle(True)")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("Pickling of struct members such as self.c_pbcbox must be explicitly requested with @auto_pickle(True)")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("Pickling of struct members such as self.c_pbcbox must be explicitly requested with @auto_pickle(True)")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.PBCBox.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":716
 * cdef class TopologyGroup(object):
 * 
 *     def __init__(self, Topology parent, str resname="group", fsl_int resid=1):             # <<<<<<<<<<<<<<
 *         self.parent = parent
 *         self.resname = resname.encode()
 */

/* Python wrapper */
static int __pyx_pw_10fatslimlib_9core_base_13TopologyGroup_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10fatslimlib_9core_base_13TopologyGroup_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_parent = 0;
  PyObject *__pyx_v_resname = 0;
  fsl_int __pyx_v_resid;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_parent,&__pyx_n_s_resname,&__pyx_n_s_resid,0};
    PyObject* values[3] = {0,0,0};
    values[1] = ((PyObject*)__pyx_n_u_group);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parent)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_resname);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_resid);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 716, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_parent = ((struct __pyx_obj_10fatslimlib_9core_base_Topology *)values[0]);
    __pyx_v_resname = ((PyObject*)values[1]);
    if (values[2]) {
      __pyx_v_resid = __Pyx_PyInt_As_fsl_int(values[2]); if (unlikely((__pyx_v_resid == ((fsl_int)-1)) && PyErr_Occurred())) __PYX_ERR(0, 716, __pyx_L3_error)
    } else {
      __pyx_v_resid = ((fsl_int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 716, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.TopologyGroup.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_parent), __pyx_ptype_10fatslimlib_9core_base_Topology, 1, "parent", 0))) __PYX_ERR(0, 716, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_resname), (&PyUnicode_Type), 1, "resname", 1))) __PYX_ERR(0, 716, __pyx_L1_error)
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_13TopologyGroup___init__(((struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *)__pyx_v_self), __pyx_v_parent, __pyx_v_resname, __pyx_v_resid);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10fatslimlib_9core_base_13TopologyGroup___init__(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self, struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_parent, PyObject *__pyx_v_resname, fsl_int __pyx_v_resid) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "fatslimlib/core_base.pyx":717
 * 
 *     def __init__(self, Topology parent, str resname="group", fsl_int resid=1):
 *         self.parent = parent             # <<<<<<<<<<<<<<
 *         self.resname = resname.encode()
 *         self.resid = resid
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_parent));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_parent));
  __Pyx_GOTREF(__pyx_v_self->parent);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->parent));
  __pyx_v_self->parent = __pyx_v_parent;

  /* "fatslimlib/core_base.pyx":718
 *     def __init__(self, Topology parent, str resname="group", fsl_int resid=1):
 *         self.parent = parent
 *         self.resname = resname.encode()             # <<<<<<<<<<<<<<
 *         self.resid = resid
 *         self.name = self.resname + b"%i" % resid
 */
  if (unlikely(__pyx_v_resname == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 718, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_resname, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->resname);
  __Pyx_DECREF(__pyx_v_self->resname);
  __pyx_v_self->resname = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":719
 *         self.parent = parent
 *         self.resname = resname.encode()
 *         self.resid = resid             # <<<<<<<<<<<<<<
 *         self.name = self.resname + b"%i" % resid
 *         self.atoms = NULL
 */
  __pyx_v_self->resid = __pyx_v_resid;

  /* "fatslimlib/core_base.pyx":720
 *         self.resname = resname.encode()
 *         self.resid = resid
 *         self.name = self.resname + b"%i" % resid             # <<<<<<<<<<<<<<
 *         self.atoms = NULL
 *         self.size = 0
 */
  __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_v_resid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Remainder(__pyx_kp_b_i, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_v_self->resname, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":721
 *         self.resid = resid
 *         self.name = self.resname + b"%i" % resid
 *         self.atoms = NULL             # <<<<<<<<<<<<<<
 *         self.size = 0
 *         self.allocated_size = 0
 */
  __pyx_v_self->atoms = NULL;

  /* "fatslimlib/core_base.pyx":722
 *         self.name = self.resname + b"%i" % resid
 *         self.atoms = NULL
 *         self.size = 0             # <<<<<<<<<<<<<<
 *         self.allocated_size = 0
 * 
 */
  __pyx_v_self->size = 0;

  /* "fatslimlib/core_base.pyx":723
 *         self.atoms = NULL
 *         self.size = 0
 *         self.allocated_size = 0             # <<<<<<<<<<<<<<
 * 
 *     cdef bint fast_contains(self, topol_atom_t atom) nogil except *:
 */
  __pyx_v_self->allocated_size = 0;

  /* "fatslimlib/core_base.pyx":716
 * cdef class TopologyGroup(object):
 * 
 *     def __init__(self, Topology parent, str resname="group", fsl_int resid=1):             # <<<<<<<<<<<<<<
 *         self.parent = parent
 *         self.resname = resname.encode()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("fatslimlib.core_base.TopologyGroup.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":725
 *         self.allocated_size = 0
 * 
 *     cdef bint fast_contains(self, topol_atom_t atom) nogil except *:             # <<<<<<<<<<<<<<
 *         cdef fsl_int i
 *         for i in range(self.size):
 */

static int __pyx_f_10fatslimlib_9core_base_13TopologyGroup_fast_contains(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self, struct __pyx_t_10fatslimlib_9core_base_topol_atom_t __pyx_v_atom) {
  fsl_int __pyx_v_i;
  int __pyx_r;
  fsl_int __pyx_t_1;
  fsl_int __pyx_t_2;
  fsl_int __pyx_t_3;
  int __pyx_t_4;

  /* "fatslimlib/core_base.pyx":727
 *     cdef bint fast_contains(self, topol_atom_t atom) nogil except *:
 *         cdef fsl_int i
 *         for i in range(self.size):             # <<<<<<<<<<<<<<
 *             if self.atoms[i].atomid == atom.atomid:
 *                 return True
 */
  __pyx_t_1 = __pyx_v_self->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "fatslimlib/core_base.pyx":728
 *         cdef fsl_int i
 *         for i in range(self.size):
 *             if self.atoms[i].atomid == atom.atomid:             # <<<<<<<<<<<<<<
 *                 return True
 *         return False
 */
    __pyx_t_4 = (((__pyx_v_self->atoms[__pyx_v_i]).atomid == __pyx_v_atom.atomid) != 0);
    if (__pyx_t_4) {

      /* "fatslimlib/core_base.pyx":729
 *         for i in range(self.size):
 *             if self.atoms[i].atomid == atom.atomid:
 *                 return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "fatslimlib/core_base.pyx":728
 *         cdef fsl_int i
 *         for i in range(self.size):
 *             if self.atoms[i].atomid == atom.atomid:             # <<<<<<<<<<<<<<
 *                 return True
 *         return False
 */
    }
  }

  /* "fatslimlib/core_base.pyx":730
 *             if self.atoms[i].atomid == atom.atomid:
 *                 return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     cdef void add_atom(self, topol_atom_t atom) nogil except *:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":725
 *         self.allocated_size = 0
 * 
 *     cdef bint fast_contains(self, topol_atom_t atom) nogil except *:             # <<<<<<<<<<<<<<
 *         cdef fsl_int i
 *         for i in range(self.size):
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":732
 *         return False
 * 
 *     cdef void add_atom(self, topol_atom_t atom) nogil except *:             # <<<<<<<<<<<<<<
 *         if self.fast_contains(atom):
 *             return
 */

static void __pyx_f_10fatslimlib_9core_base_13TopologyGroup_add_atom(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self, struct __pyx_t_10fatslimlib_9core_base_topol_atom_t __pyx_v_atom) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("add_atom", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "fatslimlib/core_base.pyx":733
 * 
 *     cdef void add_atom(self, topol_atom_t atom) nogil except *:
 *         if self.fast_contains(atom):             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  /*try:*/ {
    __pyx_t_1 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_TopologyGroup *)__pyx_v_self->__pyx_vtab)->fast_contains(__pyx_v_self, __pyx_v_atom); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 733, __pyx_L4_error)
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (__pyx_t_2) {

      /* "fatslimlib/core_base.pyx":734
 *     cdef void add_atom(self, topol_atom_t atom) nogil except *:
 *         if self.fast_contains(atom):
 *             return             # <<<<<<<<<<<<<<
 * 
 *         # Allocate memory
 */
      goto __pyx_L3_return;

      /* "fatslimlib/core_base.pyx":733
 * 
 *     cdef void add_atom(self, topol_atom_t atom) nogil except *:
 *         if self.fast_contains(atom):             # <<<<<<<<<<<<<<
 *             return
 * 
 */
    }

    /* "fatslimlib/core_base.pyx":737
 * 
 *         # Allocate memory
 *         with gil:             # <<<<<<<<<<<<<<
 *             while self.size + 1 > self.allocated_size:
 *                 self.allocated_size += RESIDUE_ALLOCATION_INCREMENT
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "fatslimlib/core_base.pyx":738
 *         # Allocate memory
 *         with gil:
 *             while self.size + 1 > self.allocated_size:             # <<<<<<<<<<<<<<
 *                 self.allocated_size += RESIDUE_ALLOCATION_INCREMENT
 * 
 */
          while (1) {
            __pyx_t_2 = (((__pyx_v_self->size + 1) > __pyx_v_self->allocated_size) != 0);
            if (!__pyx_t_2) break;

            /* "fatslimlib/core_base.pyx":739
 *         with gil:
 *             while self.size + 1 > self.allocated_size:
 *                 self.allocated_size += RESIDUE_ALLOCATION_INCREMENT             # <<<<<<<<<<<<<<
 * 
 *                 self.atoms = \
 */
            __pyx_v_self->allocated_size = (__pyx_v_self->allocated_size + 20);

            /* "fatslimlib/core_base.pyx":741
 *                 self.allocated_size += RESIDUE_ALLOCATION_INCREMENT
 * 
 *                 self.atoms = \             # <<<<<<<<<<<<<<
 *                     <topol_atom_t *> PyMem_Realloc(self.atoms,
 *                                                    sizeof(topol_atom_t ) * self.allocated_size)
 */
            __pyx_v_self->atoms = ((struct __pyx_t_10fatslimlib_9core_base_topol_atom_t *)PyMem_Realloc(__pyx_v_self->atoms, ((sizeof(struct __pyx_t_10fatslimlib_9core_base_topol_atom_t)) * __pyx_v_self->allocated_size)));

            /* "fatslimlib/core_base.pyx":744
 *                     <topol_atom_t *> PyMem_Realloc(self.atoms,
 *                                                    sizeof(topol_atom_t ) * self.allocated_size)
 *                 if self.atoms == NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 * 
 */
            __pyx_t_2 = ((__pyx_v_self->atoms == NULL) != 0);
            if (unlikely(__pyx_t_2)) {

              /* "fatslimlib/core_base.pyx":745
 *                                                    sizeof(topol_atom_t ) * self.allocated_size)
 *                 if self.atoms == NULL:
 *                     raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         self.atoms[self.size] = atom
 */
              PyErr_NoMemory(); __PYX_ERR(0, 745, __pyx_L8_error)

              /* "fatslimlib/core_base.pyx":744
 *                     <topol_atom_t *> PyMem_Realloc(self.atoms,
 *                                                    sizeof(topol_atom_t ) * self.allocated_size)
 *                 if self.atoms == NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 * 
 */
            }
          }
        }

        /* "fatslimlib/core_base.pyx":737
 * 
 *         # Allocate memory
 *         with gil:             # <<<<<<<<<<<<<<
 *             while self.size + 1 > self.allocated_size:
 *                 self.allocated_size += RESIDUE_ALLOCATION_INCREMENT
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L9;
          }
          __pyx_L8_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L9:;
        }
    }

    /* "fatslimlib/core_base.pyx":747
 *                     raise MemoryError
 * 
 *         self.atoms[self.size] = atom             # <<<<<<<<<<<<<<
 *         self.size += 1
 * 
 */
    (__pyx_v_self->atoms[__pyx_v_self->size]) = __pyx_v_atom;

    /* "fatslimlib/core_base.pyx":748
 * 
 *         self.atoms[self.size] = atom
 *         self.size += 1             # <<<<<<<<<<<<<<
 * 
 *     cdef void merge_other_group(self, TopologyGroup other) nogil:
 */
    __pyx_v_self->size = (__pyx_v_self->size + 1);
  }

  /* "fatslimlib/core_base.pyx":733
 * 
 *     cdef void add_atom(self, topol_atom_t atom) nogil except *:
 *         if self.fast_contains(atom):             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L5;
    }
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "fatslimlib/core_base.pyx":732
 *         return False
 * 
 *     cdef void add_atom(self, topol_atom_t atom) nogil except *:             # <<<<<<<<<<<<<<
 *         if self.fast_contains(atom):
 *             return
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.TopologyGroup.add_atom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "fatslimlib/core_base.pyx":750
 *         self.size += 1
 * 
 *     cdef void merge_other_group(self, TopologyGroup other) nogil:             # <<<<<<<<<<<<<<
 *         cdef fsl_int i
 *         for i in range(other.size):
 */

static void __pyx_f_10fatslimlib_9core_base_13TopologyGroup_merge_other_group(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self, struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_other) {
  fsl_int __pyx_v_i;
  fsl_int __pyx_t_1;
  fsl_int __pyx_t_2;
  fsl_int __pyx_t_3;

  /* "fatslimlib/core_base.pyx":752
 *     cdef void merge_other_group(self, TopologyGroup other) nogil:
 *         cdef fsl_int i
 *         for i in range(other.size):             # <<<<<<<<<<<<<<
 *             self.add_atom(other.atoms[i])
 * 
 */
  __pyx_t_1 = __pyx_v_other->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "fatslimlib/core_base.pyx":753
 *         cdef fsl_int i
 *         for i in range(other.size):
 *             self.add_atom(other.atoms[i])             # <<<<<<<<<<<<<<
 * 
 *     cdef fsl_int[:] fast_get_atomids(self) nogil:
 */
    ((struct __pyx_vtabstruct_10fatslimlib_9core_base_TopologyGroup *)__pyx_v_self->__pyx_vtab)->add_atom(__pyx_v_self, (__pyx_v_other->atoms[__pyx_v_i])); if (unlikely(__Pyx_ErrOccurredWithGIL())) __PYX_ERR(0, 753, __pyx_L1_error)
  }

  /* "fatslimlib/core_base.pyx":750
 *         self.size += 1
 * 
 *     cdef void merge_other_group(self, TopologyGroup other) nogil:             # <<<<<<<<<<<<<<
 *         cdef fsl_int i
 *         for i in range(other.size):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("fatslimlib.core_base.TopologyGroup.merge_other_group", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "fatslimlib/core_base.pyx":755
 *             self.add_atom(other.atoms[i])
 * 
 *     cdef fsl_int[:] fast_get_atomids(self) nogil:             # <<<<<<<<<<<<<<
 *         cdef fsl_int[:] atomids
 *         cdef fsl_int i
 */

static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_13TopologyGroup_fast_get_atomids(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_atomids = { 0, 0, { 0 }, { 0 }, { 0 } };
  fsl_int __pyx_v_i;
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_memviewslice __pyx_t_6 = { 0, 0, { 0 }, { 0 }, { 0 } };
  fsl_int __pyx_t_7;
  fsl_int __pyx_t_8;
  fsl_int __pyx_t_9;
  fsl_int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("fast_get_atomids", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "fatslimlib/core_base.pyx":756
 * 
 *     cdef fsl_int[:] fast_get_atomids(self) nogil:
 *         cdef fsl_int[:] atomids             # <<<<<<<<<<<<<<
 *         cdef fsl_int i
 *         cdef topol_atom_t atom
 */
  /*try:*/ {

    /* "fatslimlib/core_base.pyx":761
 * 
 *         # Create numpy array
 *         with gil:             # <<<<<<<<<<<<<<
 *             atomids = np.empty(self.size, dtype=np.int64)
 * 
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "fatslimlib/core_base.pyx":762
 *         # Create numpy array
 *         with gil:
 *             atomids = np.empty(self.size, dtype=np.int64)             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.size):
 */
          __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 762, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 762, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_v_self->size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 762, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 762, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_1);
          PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
          __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 762, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 762, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_int64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 762, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 762, __pyx_L7_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 762, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_6 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_6.memview)) __PYX_ERR(0, 762, __pyx_L7_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_v_atomids = __pyx_t_6;
          __pyx_t_6.memview = NULL;
          __pyx_t_6.data = NULL;
        }

        /* "fatslimlib/core_base.pyx":761
 * 
 *         # Create numpy array
 *         with gil:             # <<<<<<<<<<<<<<
 *             atomids = np.empty(self.size, dtype=np.int64)
 * 
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L8;
          }
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L8:;
        }
    }

    /* "fatslimlib/core_base.pyx":764
 *             atomids = np.empty(self.size, dtype=np.int64)
 * 
 *         for i in range(self.size):             # <<<<<<<<<<<<<<
 *             atomids[i] = self.atoms[i].atomid
 * 
 */
    __pyx_t_7 = __pyx_v_self->size;
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
      __pyx_v_i = __pyx_t_9;

      /* "fatslimlib/core_base.pyx":765
 * 
 *         for i in range(self.size):
 *             atomids[i] = self.atoms[i].atomid             # <<<<<<<<<<<<<<
 * 
 *         return atomids
 */
      __pyx_t_10 = (__pyx_v_self->atoms[__pyx_v_i]).atomid;
      __pyx_t_11 = __pyx_v_i;
      *((fsl_int *) ( /* dim=0 */ (__pyx_v_atomids.data + __pyx_t_11 * __pyx_v_atomids.strides[0]) )) = __pyx_t_10;
    }

    /* "fatslimlib/core_base.pyx":767
 *             atomids[i] = self.atoms[i].atomid
 * 
 *         return atomids             # <<<<<<<<<<<<<<
 * 
 *     cdef object group_by_resid(self):
 */
    __PYX_INC_MEMVIEW(&__pyx_v_atomids, 1);
    __pyx_r = __pyx_v_atomids;
    goto __pyx_L3_return;
  }

  /* "fatslimlib/core_base.pyx":756
 * 
 *     cdef fsl_int[:] fast_get_atomids(self) nogil:
 *         cdef fsl_int[:] atomids             # <<<<<<<<<<<<<<
 *         cdef fsl_int i
 *         cdef topol_atom_t atom
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "fatslimlib/core_base.pyx":755
 *             self.add_atom(other.atoms[i])
 * 
 *     cdef fsl_int[:] fast_get_atomids(self) nogil:             # <<<<<<<<<<<<<<
 *         cdef fsl_int[:] atomids
 *         cdef fsl_int i
 */

  /* function exit code */
  __Pyx_pretend_to_initialize(&__pyx_r);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __PYX_XDEC_MEMVIEW(&__pyx_t_6, 0);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("fatslimlib.core_base.TopologyGroup.fast_get_atomids", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_atomids, 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":769
 *         return atomids
 * 
 *     cdef object group_by_resid(self):             # <<<<<<<<<<<<<<
 *         cdef fsl_int i, internal_id
 *         cdef fsl_int atomid
 */

static PyObject *__pyx_f_10fatslimlib_9core_base_13TopologyGroup_group_by_resid(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self) {
  fsl_int __pyx_v_i;
  fsl_int __pyx_v_last_resid;
  struct __pyx_t_10fatslimlib_9core_base_topol_atom_t __pyx_v_atom;
  struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_new_group = 0;
  PyObject *__pyx_v_groups = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  fsl_int __pyx_t_2;
  fsl_int __pyx_t_3;
  fsl_int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  fsl_int __pyx_t_7;
  __Pyx_RefNannySetupContext("group_by_resid", 0);

  /* "fatslimlib/core_base.pyx":772
 *         cdef fsl_int i, internal_id
 *         cdef fsl_int atomid
 *         cdef fsl_int last_resid = -1             # <<<<<<<<<<<<<<
 *         cdef topol_atom_t atom
 *         cdef TopologyGroup new_group
 */
  __pyx_v_last_resid = -1;

  /* "fatslimlib/core_base.pyx":776
 *         cdef TopologyGroup new_group
 * 
 *         groups = []             # <<<<<<<<<<<<<<
 *         for i in range(self.size):
 *             atom = self.atoms[i]
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_groups = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":777
 * 
 *         groups = []
 *         for i in range(self.size):             # <<<<<<<<<<<<<<
 *             atom = self.atoms[i]
 * 
 */
  __pyx_t_2 = __pyx_v_self->size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "fatslimlib/core_base.pyx":778
 *         groups = []
 *         for i in range(self.size):
 *             atom = self.atoms[i]             # <<<<<<<<<<<<<<
 * 
 *             if atom.resid != last_resid:
 */
    __pyx_v_atom = (__pyx_v_self->atoms[__pyx_v_i]);

    /* "fatslimlib/core_base.pyx":780
 *             atom = self.atoms[i]
 * 
 *             if atom.resid != last_resid:             # <<<<<<<<<<<<<<
 *                 new_group = TopologyGroup(self.parent)
 *                 groups.append(new_group)
 */
    __pyx_t_5 = ((__pyx_v_atom.resid != __pyx_v_last_resid) != 0);
    if (__pyx_t_5) {

      /* "fatslimlib/core_base.pyx":781
 * 
 *             if atom.resid != last_resid:
 *                 new_group = TopologyGroup(self.parent)             # <<<<<<<<<<<<<<
 *                 groups.append(new_group)
 *                 last_resid = atom.resid
 */
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10fatslimlib_9core_base_TopologyGroup), ((PyObject *)__pyx_v_self->parent)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 781, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_new_group, ((struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "fatslimlib/core_base.pyx":782
 *             if atom.resid != last_resid:
 *                 new_group = TopologyGroup(self.parent)
 *                 groups.append(new_group)             # <<<<<<<<<<<<<<
 *                 last_resid = atom.resid
 *             new_group.add_atom(atom)
 */
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_groups, ((PyObject *)__pyx_v_new_group)); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 782, __pyx_L1_error)

      /* "fatslimlib/core_base.pyx":783
 *                 new_group = TopologyGroup(self.parent)
 *                 groups.append(new_group)
 *                 last_resid = atom.resid             # <<<<<<<<<<<<<<
 *             new_group.add_atom(atom)
 * 
 */
      __pyx_t_7 = __pyx_v_atom.resid;
      __pyx_v_last_resid = __pyx_t_7;

      /* "fatslimlib/core_base.pyx":780
 *             atom = self.atoms[i]
 * 
 *             if atom.resid != last_resid:             # <<<<<<<<<<<<<<
 *                 new_group = TopologyGroup(self.parent)
 *                 groups.append(new_group)
 */
    }

    /* "fatslimlib/core_base.pyx":784
 *                 groups.append(new_group)
 *                 last_resid = atom.resid
 *             new_group.add_atom(atom)             # <<<<<<<<<<<<<<
 * 
 *         return groups
 */
    if (unlikely(!__pyx_v_new_group)) { __Pyx_RaiseUnboundLocalError("new_group"); __PYX_ERR(0, 784, __pyx_L1_error) }
    ((struct __pyx_vtabstruct_10fatslimlib_9core_base_TopologyGroup *)__pyx_v_new_group->__pyx_vtab)->add_atom(__pyx_v_new_group, __pyx_v_atom); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 784, __pyx_L1_error)
  }

  /* "fatslimlib/core_base.pyx":786
 *             new_group.add_atom(atom)
 * 
 *         return groups             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_groups);
  __pyx_r = __pyx_v_groups;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":769
 *         return atomids
 * 
 *     cdef object group_by_resid(self):             # <<<<<<<<<<<<<<
 *         cdef fsl_int i, internal_id
 *         cdef fsl_int atomid
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.TopologyGroup.group_by_resid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_new_group);
  __Pyx_XDECREF(__pyx_v_groups);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":788
 *         return groups
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.size
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_10fatslimlib_9core_base_13TopologyGroup_3__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_10fatslimlib_9core_base_13TopologyGroup_3__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_13TopologyGroup_2__len__(((struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_10fatslimlib_9core_base_13TopologyGroup_2__len__(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "fatslimlib/core_base.pyx":789
 * 
 *     def __len__(self):
 *         return self.size             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_r = __pyx_v_self->size;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":788
 *         return groups
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.size
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":791
 *         return self.size
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self.atoms)
 * 
 */

/* Python wrapper */
static void __pyx_pw_10fatslimlib_9core_base_13TopologyGroup_5__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10fatslimlib_9core_base_13TopologyGroup_5__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_10fatslimlib_9core_base_13TopologyGroup_4__dealloc__(((struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_10fatslimlib_9core_base_13TopologyGroup_4__dealloc__(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "fatslimlib/core_base.pyx":792
 * 
 *     def __dealloc__(self):
 *         PyMem_Free(self.atoms)             # <<<<<<<<<<<<<<
 * 
 *     property atomids:
 */
  PyMem_Free(__pyx_v_self->atoms);

  /* "fatslimlib/core_base.pyx":791
 *         return self.size
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self.atoms)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "fatslimlib/core_base.pyx":795
 * 
 *     property atomids:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return np.asarray(self.fast_get_atomids())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_13TopologyGroup_7atomids_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_13TopologyGroup_7atomids_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_13TopologyGroup_7atomids___get__(((struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_13TopologyGroup_7atomids___get__(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_memviewslice __pyx_t_4 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":796
 *     property atomids:
 *         def __get__(self):
 *             return np.asarray(self.fast_get_atomids())             # <<<<<<<<<<<<<<
 * 
 *     property residue_atomids:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_asarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_TopologyGroup *)__pyx_v_self->__pyx_vtab)->fast_get_atomids(__pyx_v_self); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(0, 796, __pyx_L1_error)
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_t_4, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_fsl_int, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_fsl_int, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __PYX_XDEC_MEMVIEW(&__pyx_t_4, 1);
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":795
 * 
 *     property atomids:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return np.asarray(self.fast_get_atomids())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __PYX_XDEC_MEMVIEW(&__pyx_t_4, 1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.TopologyGroup.atomids.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":799
 * 
 *     property residue_atomids:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef fsl_int i, internal_id
 *             cdef fsl_int atomid
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_13TopologyGroup_15residue_atomids_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_13TopologyGroup_15residue_atomids_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_13TopologyGroup_15residue_atomids___get__(((struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_13TopologyGroup_15residue_atomids___get__(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self) {
  fsl_int __pyx_v_i;
  fsl_int __pyx_v_last_resid;
  struct __pyx_t_10fatslimlib_9core_base_topol_atom_t __pyx_v_atom;
  PyObject *__pyx_v_atomids = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  fsl_int __pyx_t_2;
  fsl_int __pyx_t_3;
  fsl_int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  fsl_int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":802
 *             cdef fsl_int i, internal_id
 *             cdef fsl_int atomid
 *             cdef fsl_int last_resid = -1             # <<<<<<<<<<<<<<
 *             cdef topol_atom_t atom
 * 
 */
  __pyx_v_last_resid = -1;

  /* "fatslimlib/core_base.pyx":805
 *             cdef topol_atom_t atom
 * 
 *             atomids = []             # <<<<<<<<<<<<<<
 *             for i in range(self.size):
 *                 atom = self.atoms[i]
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_atomids = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":806
 * 
 *             atomids = []
 *             for i in range(self.size):             # <<<<<<<<<<<<<<
 *                 atom = self.atoms[i]
 * 
 */
  __pyx_t_2 = __pyx_v_self->size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "fatslimlib/core_base.pyx":807
 *             atomids = []
 *             for i in range(self.size):
 *                 atom = self.atoms[i]             # <<<<<<<<<<<<<<
 * 
 *                 if atom.resid != last_resid:
 */
    __pyx_v_atom = (__pyx_v_self->atoms[__pyx_v_i]);

    /* "fatslimlib/core_base.pyx":809
 *                 atom = self.atoms[i]
 * 
 *                 if atom.resid != last_resid:             # <<<<<<<<<<<<<<
 *                     atomids.extend(list(self.parent._residues[atom.resid].atomids))
 *                     last_resid = atom.resid
 */
    __pyx_t_5 = ((__pyx_v_atom.resid != __pyx_v_last_resid) != 0);
    if (__pyx_t_5) {

      /* "fatslimlib/core_base.pyx":810
 * 
 *                 if atom.resid != last_resid:
 *                     atomids.extend(list(self.parent._residues[atom.resid].atomids))             # <<<<<<<<<<<<<<
 *                     last_resid = atom.resid
 *             return np.array(atomids, dtype=np.int64)
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->parent), __pyx_n_s_residues); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 810, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_1, __pyx_v_atom.resid, fsl_int, 1, __Pyx_PyInt_From_fsl_int, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 810, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_atomids); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 810, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 810, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_7 = __Pyx_PyList_Extend(__pyx_v_atomids, __pyx_t_6); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 810, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "fatslimlib/core_base.pyx":811
 *                 if atom.resid != last_resid:
 *                     atomids.extend(list(self.parent._residues[atom.resid].atomids))
 *                     last_resid = atom.resid             # <<<<<<<<<<<<<<
 *             return np.array(atomids, dtype=np.int64)
 * 
 */
      __pyx_t_8 = __pyx_v_atom.resid;
      __pyx_v_last_resid = __pyx_t_8;

      /* "fatslimlib/core_base.pyx":809
 *                 atom = self.atoms[i]
 * 
 *                 if atom.resid != last_resid:             # <<<<<<<<<<<<<<
 *                     atomids.extend(list(self.parent._residues[atom.resid].atomids))
 *                     last_resid = atom.resid
 */
    }
  }

  /* "fatslimlib/core_base.pyx":812
 *                     atomids.extend(list(self.parent._residues[atom.resid].atomids))
 *                     last_resid = atom.resid
 *             return np.array(atomids, dtype=np.int64)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_atomids);
  __Pyx_GIVEREF(__pyx_v_atomids);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_atomids);
  __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_np); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_int64); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_dtype, __pyx_t_11) < 0) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, __pyx_t_9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_r = __pyx_t_11;
  __pyx_t_11 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":799
 * 
 *     property residue_atomids:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef fsl_int i, internal_id
 *             cdef fsl_int atomid
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("fatslimlib.core_base.TopologyGroup.residue_atomids.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_atomids);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":814
 *             return np.array(atomids, dtype=np.int64)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "TopologyGroup '%s': %i atoms" % (self.name, self.size)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_13TopologyGroup_7__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_13TopologyGroup_7__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_13TopologyGroup_6__str__(((struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_13TopologyGroup_6__str__(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "fatslimlib/core_base.pyx":815
 * 
 *     def __str__(self):
 *         return "TopologyGroup '%s': %i atoms" % (self.name, self.size)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_v_self->size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyUnicode_Format(__pyx_kp_u_TopologyGroup_s_i_atoms, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":814
 *             return np.array(atomids, dtype=np.int64)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "TopologyGroup '%s': %i atoms" % (self.name, self.size)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("fatslimlib.core_base.TopologyGroup.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pxd":126
 * cdef class TopologyGroup(object):
 *     # Attrbutes
 *     cdef readonly Topology parent             # <<<<<<<<<<<<<<
 *     cdef bytes name
 *     cdef bytes resname
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_13TopologyGroup_6parent_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_13TopologyGroup_6parent_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_13TopologyGroup_6parent___get__(((struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_13TopologyGroup_6parent___get__(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->parent));
  __pyx_r = ((PyObject *)__pyx_v_self->parent);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pxd":129
 *     cdef bytes name
 *     cdef bytes resname
 *     cdef readonly fsl_int resid             # <<<<<<<<<<<<<<
 *     cdef topol_atom_t *atoms
 *     cdef readonly fsl_int size
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_13TopologyGroup_5resid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_13TopologyGroup_5resid_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_13TopologyGroup_5resid___get__(((struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_13TopologyGroup_5resid___get__(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_v_self->resid); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.TopologyGroup.resid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pxd":131
 *     cdef readonly fsl_int resid
 *     cdef topol_atom_t *atoms
 *     cdef readonly fsl_int size             # <<<<<<<<<<<<<<
 *     cdef fsl_int allocated_size
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_13TopologyGroup_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_13TopologyGroup_4size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_13TopologyGroup_4size___get__(((struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_13TopologyGroup_4size___get__(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_v_self->size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.TopologyGroup.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.atoms cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_13TopologyGroup_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_13TopologyGroup_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_13TopologyGroup_8__reduce_cython__(((struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_13TopologyGroup_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.atoms cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.atoms cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.atoms cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.TopologyGroup.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.atoms cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.atoms cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_13TopologyGroup_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_13TopologyGroup_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_13TopologyGroup_10__setstate_cython__(((struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_13TopologyGroup_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.atoms cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.atoms cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.atoms cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.atoms cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.TopologyGroup.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":818
 * 
 * 
 * cdef list topol_residue_atomids_as_list(topol_residue_t *residue):             # <<<<<<<<<<<<<<
 *     cdef fsl_int i
 * 
 */

static PyObject *__pyx_f_10fatslimlib_9core_base_topol_residue_atomids_as_list(struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *__pyx_v_residue) {
  fsl_int __pyx_v_i;
  PyObject *__pyx_v_obj = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  fsl_int __pyx_t_2;
  fsl_int __pyx_t_3;
  fsl_int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("topol_residue_atomids_as_list", 0);

  /* "fatslimlib/core_base.pyx":821
 *     cdef fsl_int i
 * 
 *     obj = []             # <<<<<<<<<<<<<<
 *     for i in range(residue.size):
 *         obj.append(residue.atomids[i])
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_obj = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":822
 * 
 *     obj = []
 *     for i in range(residue.size):             # <<<<<<<<<<<<<<
 *         obj.append(residue.atomids[i])
 * 
 */
  __pyx_t_2 = __pyx_v_residue->size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "fatslimlib/core_base.pyx":823
 *     obj = []
 *     for i in range(residue.size):
 *         obj.append(residue.atomids[i])             # <<<<<<<<<<<<<<
 * 
 *     return obj
 */
    __pyx_t_1 = __Pyx_PyInt_From_fsl_int((__pyx_v_residue->atomids[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 823, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_obj, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 823, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "fatslimlib/core_base.pyx":825
 *         obj.append(residue.atomids[i])
 * 
 *     return obj             # <<<<<<<<<<<<<<
 * 
 * cdef fsl_int topol_residue_atomid_index(topol_residue_t *residue, fsl_int atomid):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_obj);
  __pyx_r = __pyx_v_obj;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":818
 * 
 * 
 * cdef list topol_residue_atomids_as_list(topol_residue_t *residue):             # <<<<<<<<<<<<<<
 *     cdef fsl_int i
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.topol_residue_atomids_as_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":827
 *     return obj
 * 
 * cdef fsl_int topol_residue_atomid_index(topol_residue_t *residue, fsl_int atomid):             # <<<<<<<<<<<<<<
 *     cdef fsl_int i
 *     for i in range(residue.size):
 */

static fsl_int __pyx_f_10fatslimlib_9core_base_topol_residue_atomid_index(struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *__pyx_v_residue, fsl_int __pyx_v_atomid) {
  fsl_int __pyx_v_i;
  fsl_int __pyx_r;
  __Pyx_RefNannyDeclarations
  fsl_int __pyx_t_1;
  fsl_int __pyx_t_2;
  fsl_int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("topol_residue_atomid_index", 0);

  /* "fatslimlib/core_base.pyx":829
 * cdef fsl_int topol_residue_atomid_index(topol_residue_t *residue, fsl_int atomid):
 *     cdef fsl_int i
 *     for i in range(residue.size):             # <<<<<<<<<<<<<<
 *         if residue.atomids[i] == atomid:
 *             return i
 */
  __pyx_t_1 = __pyx_v_residue->size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "fatslimlib/core_base.pyx":830
 *     cdef fsl_int i
 *     for i in range(residue.size):
 *         if residue.atomids[i] == atomid:             # <<<<<<<<<<<<<<
 *             return i
 *     return -1
 */
    __pyx_t_4 = (((__pyx_v_residue->atomids[__pyx_v_i]) == __pyx_v_atomid) != 0);
    if (__pyx_t_4) {

      /* "fatslimlib/core_base.pyx":831
 *     for i in range(residue.size):
 *         if residue.atomids[i] == atomid:
 *             return i             # <<<<<<<<<<<<<<
 *     return -1
 * 
 */
      __pyx_r = __pyx_v_i;
      goto __pyx_L0;

      /* "fatslimlib/core_base.pyx":830
 *     cdef fsl_int i
 *     for i in range(residue.size):
 *         if residue.atomids[i] == atomid:             # <<<<<<<<<<<<<<
 *             return i
 *     return -1
 */
    }
  }

  /* "fatslimlib/core_base.pyx":832
 *         if residue.atomids[i] == atomid:
 *             return i
 *     return -1             # <<<<<<<<<<<<<<
 * 
 * cdef dict topol_residue_to_dict(topol_residue_t *residue, bytes resname):
 */
  __pyx_r = -1;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":827
 *     return obj
 * 
 * cdef fsl_int topol_residue_atomid_index(topol_residue_t *residue, fsl_int atomid):             # <<<<<<<<<<<<<<
 *     cdef fsl_int i
 *     for i in range(residue.size):
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":834
 *     return -1
 * 
 * cdef dict topol_residue_to_dict(topol_residue_t *residue, bytes resname):             # <<<<<<<<<<<<<<
 *     atomids = []
 *     for i in range(residue.size):
 */

static PyObject *__pyx_f_10fatslimlib_9core_base_topol_residue_to_dict(struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *__pyx_v_residue, PyObject *__pyx_v_resname) {
  PyObject *__pyx_v_atomids = NULL;
  fsl_int __pyx_v_i;
  PyObject *__pyx_v_resdict = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  fsl_int __pyx_t_2;
  fsl_int __pyx_t_3;
  fsl_int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("topol_residue_to_dict", 0);

  /* "fatslimlib/core_base.pyx":835
 * 
 * cdef dict topol_residue_to_dict(topol_residue_t *residue, bytes resname):
 *     atomids = []             # <<<<<<<<<<<<<<
 *     for i in range(residue.size):
 *         atomids.append(residue.atomids[i])
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_atomids = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":836
 * cdef dict topol_residue_to_dict(topol_residue_t *residue, bytes resname):
 *     atomids = []
 *     for i in range(residue.size):             # <<<<<<<<<<<<<<
 *         atomids.append(residue.atomids[i])
 *     atomids.sort()
 */
  __pyx_t_2 = __pyx_v_residue->size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "fatslimlib/core_base.pyx":837
 *     atomids = []
 *     for i in range(residue.size):
 *         atomids.append(residue.atomids[i])             # <<<<<<<<<<<<<<
 *     atomids.sort()
 * 
 */
    __pyx_t_1 = __Pyx_PyInt_From_fsl_int((__pyx_v_residue->atomids[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 837, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_atomids, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 837, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "fatslimlib/core_base.pyx":838
 *     for i in range(residue.size):
 *         atomids.append(residue.atomids[i])
 *     atomids.sort()             # <<<<<<<<<<<<<<
 * 
 *     resdict = {
 */
  __pyx_t_5 = PyList_Sort(__pyx_v_atomids); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 838, __pyx_L1_error)

  /* "fatslimlib/core_base.pyx":841
 * 
 *     resdict = {
 *         "resname": str(resname.decode()),             # <<<<<<<<<<<<<<
 *         "resid": residue.resid,
 *         "name": str(resname.decode()) + "%i" % residue.resid,
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(__pyx_v_resname == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "decode");
    __PYX_ERR(0, 841, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_decode_bytes(__pyx_v_resname, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_resname, __pyx_t_7) < 0) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "fatslimlib/core_base.pyx":842
 *     resdict = {
 *         "resname": str(resname.decode()),
 *         "resid": residue.resid,             # <<<<<<<<<<<<<<
 *         "name": str(resname.decode()) + "%i" % residue.resid,
 *         "atomids": np.array(atomids, dtype=np.int64),
 */
  __pyx_t_7 = __Pyx_PyInt_From_fsl_int(__pyx_v_residue->resid); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_resid, __pyx_t_7) < 0) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "fatslimlib/core_base.pyx":843
 *         "resname": str(resname.decode()),
 *         "resid": residue.resid,
 *         "name": str(resname.decode()) + "%i" % residue.resid,             # <<<<<<<<<<<<<<
 *         "atomids": np.array(atomids, dtype=np.int64),
 *         "size": residue.size}
 */
  if (unlikely(__pyx_v_resname == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "decode");
    __PYX_ERR(0, 843, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_decode_bytes(__pyx_v_resname, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyInt_From_fsl_int(__pyx_v_residue->resid); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_i, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyUnicode_Concat(__pyx_t_6, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_name, __pyx_t_7) < 0) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "fatslimlib/core_base.pyx":844
 *         "resid": residue.resid,
 *         "name": str(resname.decode()) + "%i" % residue.resid,
 *         "atomids": np.array(atomids, dtype=np.int64),             # <<<<<<<<<<<<<<
 *         "size": residue.size}
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_array); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_atomids);
  __Pyx_GIVEREF(__pyx_v_atomids);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_atomids);
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_np); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_int64); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_dtype, __pyx_t_10) < 0) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_7, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_atomids, __pyx_t_10) < 0) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "fatslimlib/core_base.pyx":845
 *         "name": str(resname.decode()) + "%i" % residue.resid,
 *         "atomids": np.array(atomids, dtype=np.int64),
 *         "size": residue.size}             # <<<<<<<<<<<<<<
 * 
 *     return resdict
 */
  __pyx_t_10 = __Pyx_PyInt_From_fsl_int(__pyx_v_residue->size); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_size, __pyx_t_10) < 0) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_resdict = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":847
 *         "size": residue.size}
 * 
 *     return resdict             # <<<<<<<<<<<<<<
 * 
 * cdef void topol_residue_append_atom(topol_residue_t *residue, topol_atom_t *atom):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_resdict);
  __pyx_r = __pyx_v_resdict;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":834
 *     return -1
 * 
 * cdef dict topol_residue_to_dict(topol_residue_t *residue, bytes resname):             # <<<<<<<<<<<<<<
 *     atomids = []
 *     for i in range(residue.size):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("fatslimlib.core_base.topol_residue_to_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_atomids);
  __Pyx_XDECREF(__pyx_v_resdict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":849
 *     return resdict
 * 
 * cdef void topol_residue_append_atom(topol_residue_t *residue, topol_atom_t *atom):             # <<<<<<<<<<<<<<
 *     if residue == NULL:
 *         raise RuntimeError
 */

static void __pyx_f_10fatslimlib_9core_base_topol_residue_append_atom(struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *__pyx_v_residue, struct __pyx_t_10fatslimlib_9core_base_topol_atom_t *__pyx_v_atom) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  fsl_int __pyx_t_2;
  __Pyx_RefNannySetupContext("topol_residue_append_atom", 0);

  /* "fatslimlib/core_base.pyx":850
 * 
 * cdef void topol_residue_append_atom(topol_residue_t *residue, topol_atom_t *atom):
 *     if residue == NULL:             # <<<<<<<<<<<<<<
 *         raise RuntimeError
 * 
 */
  __pyx_t_1 = ((__pyx_v_residue == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "fatslimlib/core_base.pyx":851
 * cdef void topol_residue_append_atom(topol_residue_t *residue, topol_atom_t *atom):
 *     if residue == NULL:
 *         raise RuntimeError             # <<<<<<<<<<<<<<
 * 
 *     if residue.size == residue.allocated_size:
 */
    __Pyx_Raise(__pyx_builtin_RuntimeError, 0, 0, 0);
    __PYX_ERR(0, 851, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":850
 * 
 * cdef void topol_residue_append_atom(topol_residue_t *residue, topol_atom_t *atom):
 *     if residue == NULL:             # <<<<<<<<<<<<<<
 *         raise RuntimeError
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":853
 *         raise RuntimeError
 * 
 *     if residue.size == residue.allocated_size:             # <<<<<<<<<<<<<<
 *         residue.allocated_size += RESIDUE_ALLOCATION_INCREMENT
 *         residue.atom_internal_ids = \
 */
  __pyx_t_1 = ((__pyx_v_residue->size == __pyx_v_residue->allocated_size) != 0);
  if (__pyx_t_1) {

    /* "fatslimlib/core_base.pyx":854
 * 
 *     if residue.size == residue.allocated_size:
 *         residue.allocated_size += RESIDUE_ALLOCATION_INCREMENT             # <<<<<<<<<<<<<<
 *         residue.atom_internal_ids = \
 *             <fsl_int *> PyMem_Realloc(<void *> residue.atom_internal_ids,
 */
    __pyx_v_residue->allocated_size = (__pyx_v_residue->allocated_size + 20);

    /* "fatslimlib/core_base.pyx":855
 *     if residue.size == residue.allocated_size:
 *         residue.allocated_size += RESIDUE_ALLOCATION_INCREMENT
 *         residue.atom_internal_ids = \             # <<<<<<<<<<<<<<
 *             <fsl_int *> PyMem_Realloc(<void *> residue.atom_internal_ids,
 *                                   sizeof(fsl_int) * residue.allocated_size)
 */
    __pyx_v_residue->atom_internal_ids = ((fsl_int *)PyMem_Realloc(((void *)__pyx_v_residue->atom_internal_ids), ((sizeof(fsl_int)) * __pyx_v_residue->allocated_size)));

    /* "fatslimlib/core_base.pyx":858
 *             <fsl_int *> PyMem_Realloc(<void *> residue.atom_internal_ids,
 *                                   sizeof(fsl_int) * residue.allocated_size)
 *         if residue.atom_internal_ids == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 *         residue.atomids = \
 */
    __pyx_t_1 = ((__pyx_v_residue->atom_internal_ids == NULL) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "fatslimlib/core_base.pyx":859
 *                                   sizeof(fsl_int) * residue.allocated_size)
 *         if residue.atom_internal_ids == NULL:
 *             raise MemoryError             # <<<<<<<<<<<<<<
 *         residue.atomids = \
 *             <fsl_int *> PyMem_Realloc(<void *> residue.atomids, sizeof(fsl_int) * residue.allocated_size)
 */
      PyErr_NoMemory(); __PYX_ERR(0, 859, __pyx_L1_error)

      /* "fatslimlib/core_base.pyx":858
 *             <fsl_int *> PyMem_Realloc(<void *> residue.atom_internal_ids,
 *                                   sizeof(fsl_int) * residue.allocated_size)
 *         if residue.atom_internal_ids == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 *         residue.atomids = \
 */
    }

    /* "fatslimlib/core_base.pyx":860
 *         if residue.atom_internal_ids == NULL:
 *             raise MemoryError
 *         residue.atomids = \             # <<<<<<<<<<<<<<
 *             <fsl_int *> PyMem_Realloc(<void *> residue.atomids, sizeof(fsl_int) * residue.allocated_size)
 *         if residue.atomids == NULL:
 */
    __pyx_v_residue->atomids = ((fsl_int *)PyMem_Realloc(((void *)__pyx_v_residue->atomids), ((sizeof(fsl_int)) * __pyx_v_residue->allocated_size)));

    /* "fatslimlib/core_base.pyx":862
 *         residue.atomids = \
 *             <fsl_int *> PyMem_Realloc(<void *> residue.atomids, sizeof(fsl_int) * residue.allocated_size)
 *         if residue.atomids == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 * 
 */
    __pyx_t_1 = ((__pyx_v_residue->atomids == NULL) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "fatslimlib/core_base.pyx":863
 *             <fsl_int *> PyMem_Realloc(<void *> residue.atomids, sizeof(fsl_int) * residue.allocated_size)
 *         if residue.atomids == NULL:
 *             raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *     atom.resid = residue.resid
 */
      PyErr_NoMemory(); __PYX_ERR(0, 863, __pyx_L1_error)

      /* "fatslimlib/core_base.pyx":862
 *         residue.atomids = \
 *             <fsl_int *> PyMem_Realloc(<void *> residue.atomids, sizeof(fsl_int) * residue.allocated_size)
 *         if residue.atomids == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 * 
 */
    }

    /* "fatslimlib/core_base.pyx":853
 *         raise RuntimeError
 * 
 *     if residue.size == residue.allocated_size:             # <<<<<<<<<<<<<<
 *         residue.allocated_size += RESIDUE_ALLOCATION_INCREMENT
 *         residue.atom_internal_ids = \
 */
  }

  /* "fatslimlib/core_base.pyx":865
 *             raise MemoryError
 * 
 *     atom.resid = residue.resid             # <<<<<<<<<<<<<<
 *     atom.residue_internal_id = residue.internal_id
 *     residue.atom_internal_ids[residue.size] = atom.internal_id
 */
  __pyx_t_2 = __pyx_v_residue->resid;
  __pyx_v_atom->resid = __pyx_t_2;

  /* "fatslimlib/core_base.pyx":866
 * 
 *     atom.resid = residue.resid
 *     atom.residue_internal_id = residue.internal_id             # <<<<<<<<<<<<<<
 *     residue.atom_internal_ids[residue.size] = atom.internal_id
 *     residue.atomids[residue.size] = atom.atomid
 */
  __pyx_t_2 = __pyx_v_residue->internal_id;
  __pyx_v_atom->residue_internal_id = __pyx_t_2;

  /* "fatslimlib/core_base.pyx":867
 *     atom.resid = residue.resid
 *     atom.residue_internal_id = residue.internal_id
 *     residue.atom_internal_ids[residue.size] = atom.internal_id             # <<<<<<<<<<<<<<
 *     residue.atomids[residue.size] = atom.atomid
 *     residue.size += 1
 */
  __pyx_t_2 = __pyx_v_atom->internal_id;
  (__pyx_v_residue->atom_internal_ids[__pyx_v_residue->size]) = __pyx_t_2;

  /* "fatslimlib/core_base.pyx":868
 *     atom.residue_internal_id = residue.internal_id
 *     residue.atom_internal_ids[residue.size] = atom.internal_id
 *     residue.atomids[residue.size] = atom.atomid             # <<<<<<<<<<<<<<
 *     residue.size += 1
 * 
 */
  __pyx_t_2 = __pyx_v_atom->atomid;
  (__pyx_v_residue->atomids[__pyx_v_residue->size]) = __pyx_t_2;

  /* "fatslimlib/core_base.pyx":869
 *     residue.atom_internal_ids[residue.size] = atom.internal_id
 *     residue.atomids[residue.size] = atom.atomid
 *     residue.size += 1             # <<<<<<<<<<<<<<
 * 
 * cdef class Topology(object):
 */
  __pyx_v_residue->size = (__pyx_v_residue->size + 1);

  /* "fatslimlib/core_base.pyx":849
 *     return resdict
 * 
 * cdef void topol_residue_append_atom(topol_residue_t *residue, topol_atom_t *atom):             # <<<<<<<<<<<<<<
 *     if residue == NULL:
 *         raise RuntimeError
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("fatslimlib.core_base.topol_residue_append_atom", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "fatslimlib/core_base.pyx":873
 * cdef class Topology(object):
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.atomnames = \
 *             <strname_t *> PyMem_Realloc(NULL, sizeof(strname_t) * NAME_ALLOCATION_INCREMENT)
 */

/* Python wrapper */
static int __pyx_pw_10fatslimlib_9core_base_8Topology_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10fatslimlib_9core_base_8Topology_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 0))) return -1;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_8Topology___init__(((struct __pyx_obj_10fatslimlib_9core_base_Topology *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10fatslimlib_9core_base_8Topology___init__(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self) {
  long __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  long __pyx_t_2;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "fatslimlib/core_base.pyx":874
 * 
 *     def __init__(self):
 *         self.atomnames = \             # <<<<<<<<<<<<<<
 *             <strname_t *> PyMem_Realloc(NULL, sizeof(strname_t) * NAME_ALLOCATION_INCREMENT)
 *         if self.atomnames == NULL:
 */
  __pyx_v_self->atomnames = ((__pyx_t_10fatslimlib_9core_base_strname_t *)PyMem_Realloc(NULL, ((sizeof(__pyx_t_10fatslimlib_9core_base_strname_t)) * 10)));

  /* "fatslimlib/core_base.pyx":876
 *         self.atomnames = \
 *             <strname_t *> PyMem_Realloc(NULL, sizeof(strname_t) * NAME_ALLOCATION_INCREMENT)
 *         if self.atomnames == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 *         self.atomnames_size = 0
 */
  __pyx_t_1 = ((__pyx_v_self->atomnames == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "fatslimlib/core_base.pyx":877
 *             <strname_t *> PyMem_Realloc(NULL, sizeof(strname_t) * NAME_ALLOCATION_INCREMENT)
 *         if self.atomnames == NULL:
 *             raise MemoryError             # <<<<<<<<<<<<<<
 *         self.atomnames_size = 0
 *         self.atomnames_allocated_size = NAME_ALLOCATION_INCREMENT
 */
    PyErr_NoMemory(); __PYX_ERR(0, 877, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":876
 *         self.atomnames = \
 *             <strname_t *> PyMem_Realloc(NULL, sizeof(strname_t) * NAME_ALLOCATION_INCREMENT)
 *         if self.atomnames == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 *         self.atomnames_size = 0
 */
  }

  /* "fatslimlib/core_base.pyx":878
 *         if self.atomnames == NULL:
 *             raise MemoryError
 *         self.atomnames_size = 0             # <<<<<<<<<<<<<<
 *         self.atomnames_allocated_size = NAME_ALLOCATION_INCREMENT
 * 
 */
  __pyx_v_self->atomnames_size = 0;

  /* "fatslimlib/core_base.pyx":879
 *             raise MemoryError
 *         self.atomnames_size = 0
 *         self.atomnames_allocated_size = NAME_ALLOCATION_INCREMENT             # <<<<<<<<<<<<<<
 * 
 *         self.resnames = \
 */
  __pyx_v_self->atomnames_allocated_size = 10;

  /* "fatslimlib/core_base.pyx":881
 *         self.atomnames_allocated_size = NAME_ALLOCATION_INCREMENT
 * 
 *         self.resnames = \             # <<<<<<<<<<<<<<
 *             <strname_t *> PyMem_Realloc(NULL, sizeof(strname_t) * NAME_ALLOCATION_INCREMENT)
 *         if self.resnames == NULL:
 */
  __pyx_v_self->resnames = ((__pyx_t_10fatslimlib_9core_base_strname_t *)PyMem_Realloc(NULL, ((sizeof(__pyx_t_10fatslimlib_9core_base_strname_t)) * 10)));

  /* "fatslimlib/core_base.pyx":883
 *         self.resnames = \
 *             <strname_t *> PyMem_Realloc(NULL, sizeof(strname_t) * NAME_ALLOCATION_INCREMENT)
 *         if self.resnames == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 *         self.resnames_size = 0
 */
  __pyx_t_1 = ((__pyx_v_self->resnames == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "fatslimlib/core_base.pyx":884
 *             <strname_t *> PyMem_Realloc(NULL, sizeof(strname_t) * NAME_ALLOCATION_INCREMENT)
 *         if self.resnames == NULL:
 *             raise MemoryError             # <<<<<<<<<<<<<<
 *         self.resnames_size = 0
 *         self.resnames_allocated_size = NAME_ALLOCATION_INCREMENT
 */
    PyErr_NoMemory(); __PYX_ERR(0, 884, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":883
 *         self.resnames = \
 *             <strname_t *> PyMem_Realloc(NULL, sizeof(strname_t) * NAME_ALLOCATION_INCREMENT)
 *         if self.resnames == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 *         self.resnames_size = 0
 */
  }

  /* "fatslimlib/core_base.pyx":885
 *         if self.resnames == NULL:
 *             raise MemoryError
 *         self.resnames_size = 0             # <<<<<<<<<<<<<<
 *         self.resnames_allocated_size = NAME_ALLOCATION_INCREMENT
 * 
 */
  __pyx_v_self->resnames_size = 0;

  /* "fatslimlib/core_base.pyx":886
 *             raise MemoryError
 *         self.resnames_size = 0
 *         self.resnames_allocated_size = NAME_ALLOCATION_INCREMENT             # <<<<<<<<<<<<<<
 * 
 *         self.atoms = NULL
 */
  __pyx_v_self->resnames_allocated_size = 10;

  /* "fatslimlib/core_base.pyx":888
 *         self.resnames_allocated_size = NAME_ALLOCATION_INCREMENT
 * 
 *         self.atoms = NULL             # <<<<<<<<<<<<<<
 *         self.atoms_size = 0
 *         self.atoms_allocated_size = 0
 */
  __pyx_v_self->atoms = NULL;

  /* "fatslimlib/core_base.pyx":889
 * 
 *         self.atoms = NULL
 *         self.atoms_size = 0             # <<<<<<<<<<<<<<
 *         self.atoms_allocated_size = 0
 *         self.set_atoms_allocation(ATOMS_ALLOCATION_INCREMENT)
 */
  __pyx_v_self->atoms_size = 0;

  /* "fatslimlib/core_base.pyx":890
 *         self.atoms = NULL
 *         self.atoms_size = 0
 *         self.atoms_allocated_size = 0             # <<<<<<<<<<<<<<
 *         self.set_atoms_allocation(ATOMS_ALLOCATION_INCREMENT)
 * 
 */
  __pyx_v_self->atoms_allocated_size = 0;

  /* "fatslimlib/core_base.pyx":891
 *         self.atoms_size = 0
 *         self.atoms_allocated_size = 0
 *         self.set_atoms_allocation(ATOMS_ALLOCATION_INCREMENT)             # <<<<<<<<<<<<<<
 * 
 *         self.atomids_to_internalids = \
 */
  ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Topology *)__pyx_v_self->__pyx_vtab)->set_atoms_allocation(__pyx_v_self, 0x3E8);

  /* "fatslimlib/core_base.pyx":893
 *         self.set_atoms_allocation(ATOMS_ALLOCATION_INCREMENT)
 * 
 *         self.atomids_to_internalids = \             # <<<<<<<<<<<<<<
 *             <fsl_int *> PyMem_Realloc(NULL, sizeof(fsl_int) * ATOMS_ALLOCATION_INCREMENT)
 *         if self.atomids_to_internalids == NULL:
 */
  __pyx_v_self->atomids_to_internalids = ((fsl_int *)PyMem_Realloc(NULL, ((sizeof(fsl_int)) * 0x3E8)));

  /* "fatslimlib/core_base.pyx":895
 *         self.atomids_to_internalids = \
 *             <fsl_int *> PyMem_Realloc(NULL, sizeof(fsl_int) * ATOMS_ALLOCATION_INCREMENT)
 *         if self.atomids_to_internalids == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 *         for i in range(ATOMS_ALLOCATION_INCREMENT):
 */
  __pyx_t_1 = ((__pyx_v_self->atomids_to_internalids == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "fatslimlib/core_base.pyx":896
 *             <fsl_int *> PyMem_Realloc(NULL, sizeof(fsl_int) * ATOMS_ALLOCATION_INCREMENT)
 *         if self.atomids_to_internalids == NULL:
 *             raise MemoryError             # <<<<<<<<<<<<<<
 *         for i in range(ATOMS_ALLOCATION_INCREMENT):
 *             self.atomids_to_internalids[i] = NOTSET
 */
    PyErr_NoMemory(); __PYX_ERR(0, 896, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":895
 *         self.atomids_to_internalids = \
 *             <fsl_int *> PyMem_Realloc(NULL, sizeof(fsl_int) * ATOMS_ALLOCATION_INCREMENT)
 *         if self.atomids_to_internalids == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 *         for i in range(ATOMS_ALLOCATION_INCREMENT):
 */
  }

  /* "fatslimlib/core_base.pyx":897
 *         if self.atomids_to_internalids == NULL:
 *             raise MemoryError
 *         for i in range(ATOMS_ALLOCATION_INCREMENT):             # <<<<<<<<<<<<<<
 *             self.atomids_to_internalids[i] = NOTSET
 *         self.atomids_to_internalids_size = 0
 */
  for (__pyx_t_2 = 0; __pyx_t_2 < 0x3E8; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "fatslimlib/core_base.pyx":898
 *             raise MemoryError
 *         for i in range(ATOMS_ALLOCATION_INCREMENT):
 *             self.atomids_to_internalids[i] = NOTSET             # <<<<<<<<<<<<<<
 *         self.atomids_to_internalids_size = 0
 *         self.atomids_to_internalids_allocated_size = ATOMS_ALLOCATION_INCREMENT
 */
    (__pyx_v_self->atomids_to_internalids[__pyx_v_i]) = -12345;
  }

  /* "fatslimlib/core_base.pyx":899
 *         for i in range(ATOMS_ALLOCATION_INCREMENT):
 *             self.atomids_to_internalids[i] = NOTSET
 *         self.atomids_to_internalids_size = 0             # <<<<<<<<<<<<<<
 *         self.atomids_to_internalids_allocated_size = ATOMS_ALLOCATION_INCREMENT
 * 
 */
  __pyx_v_self->atomids_to_internalids_size = 0;

  /* "fatslimlib/core_base.pyx":900
 *             self.atomids_to_internalids[i] = NOTSET
 *         self.atomids_to_internalids_size = 0
 *         self.atomids_to_internalids_allocated_size = ATOMS_ALLOCATION_INCREMENT             # <<<<<<<<<<<<<<
 * 
 *         self.residues = NULL
 */
  __pyx_v_self->atomids_to_internalids_allocated_size = 0x3E8;

  /* "fatslimlib/core_base.pyx":902
 *         self.atomids_to_internalids_allocated_size = ATOMS_ALLOCATION_INCREMENT
 * 
 *         self.residues = NULL             # <<<<<<<<<<<<<<
 *         self.residues_size = 0
 *         self.residues_allocated_size = 0
 */
  __pyx_v_self->residues = NULL;

  /* "fatslimlib/core_base.pyx":903
 * 
 *         self.residues = NULL
 *         self.residues_size = 0             # <<<<<<<<<<<<<<
 *         self.residues_allocated_size = 0
 *         self.set_residues_allocation(RESIDUES_ALLOCATION_INCREMENT)
 */
  __pyx_v_self->residues_size = 0;

  /* "fatslimlib/core_base.pyx":904
 *         self.residues = NULL
 *         self.residues_size = 0
 *         self.residues_allocated_size = 0             # <<<<<<<<<<<<<<
 *         self.set_residues_allocation(RESIDUES_ALLOCATION_INCREMENT)
 * 
 */
  __pyx_v_self->residues_allocated_size = 0;

  /* "fatslimlib/core_base.pyx":905
 *         self.residues_size = 0
 *         self.residues_allocated_size = 0
 *         self.set_residues_allocation(RESIDUES_ALLOCATION_INCREMENT)             # <<<<<<<<<<<<<<
 * 
 *         self.resids_to_internalids = \
 */
  ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Topology *)__pyx_v_self->__pyx_vtab)->set_residues_allocation(__pyx_v_self, 50);

  /* "fatslimlib/core_base.pyx":907
 *         self.set_residues_allocation(RESIDUES_ALLOCATION_INCREMENT)
 * 
 *         self.resids_to_internalids = \             # <<<<<<<<<<<<<<
 *             <fsl_int *> PyMem_Realloc(NULL, sizeof(fsl_int) * RESIDUES_ALLOCATION_INCREMENT)
 *         if self.resids_to_internalids == NULL:
 */
  __pyx_v_self->resids_to_internalids = ((fsl_int *)PyMem_Realloc(NULL, ((sizeof(fsl_int)) * 50)));

  /* "fatslimlib/core_base.pyx":909
 *         self.resids_to_internalids = \
 *             <fsl_int *> PyMem_Realloc(NULL, sizeof(fsl_int) * RESIDUES_ALLOCATION_INCREMENT)
 *         if self.resids_to_internalids == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 *         for i in range(RESIDUES_ALLOCATION_INCREMENT):
 */
  __pyx_t_1 = ((__pyx_v_self->resids_to_internalids == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "fatslimlib/core_base.pyx":910
 *             <fsl_int *> PyMem_Realloc(NULL, sizeof(fsl_int) * RESIDUES_ALLOCATION_INCREMENT)
 *         if self.resids_to_internalids == NULL:
 *             raise MemoryError             # <<<<<<<<<<<<<<
 *         for i in range(RESIDUES_ALLOCATION_INCREMENT):
 *             self.resids_to_internalids[i] = NOTSET
 */
    PyErr_NoMemory(); __PYX_ERR(0, 910, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":909
 *         self.resids_to_internalids = \
 *             <fsl_int *> PyMem_Realloc(NULL, sizeof(fsl_int) * RESIDUES_ALLOCATION_INCREMENT)
 *         if self.resids_to_internalids == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 *         for i in range(RESIDUES_ALLOCATION_INCREMENT):
 */
  }

  /* "fatslimlib/core_base.pyx":911
 *         if self.resids_to_internalids == NULL:
 *             raise MemoryError
 *         for i in range(RESIDUES_ALLOCATION_INCREMENT):             # <<<<<<<<<<<<<<
 *             self.resids_to_internalids[i] = NOTSET
 * 
 */
  for (__pyx_t_2 = 0; __pyx_t_2 < 50; __pyx_t_2+=1) {
    __pyx_v_i = __pyx_t_2;

    /* "fatslimlib/core_base.pyx":912
 *             raise MemoryError
 *         for i in range(RESIDUES_ALLOCATION_INCREMENT):
 *             self.resids_to_internalids[i] = NOTSET             # <<<<<<<<<<<<<<
 * 
 *         self.resids_to_internalids_size = 0
 */
    (__pyx_v_self->resids_to_internalids[__pyx_v_i]) = -12345;
  }

  /* "fatslimlib/core_base.pyx":914
 *             self.resids_to_internalids[i] = NOTSET
 * 
 *         self.resids_to_internalids_size = 0             # <<<<<<<<<<<<<<
 *         self.resids_to_internalids_allocated_size = RESIDUES_ALLOCATION_INCREMENT
 * 
 */
  __pyx_v_self->resids_to_internalids_size = 0;

  /* "fatslimlib/core_base.pyx":915
 * 
 *         self.resids_to_internalids_size = 0
 *         self.resids_to_internalids_allocated_size = RESIDUES_ALLOCATION_INCREMENT             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->resids_to_internalids_allocated_size = 50;

  /* "fatslimlib/core_base.pyx":873
 * cdef class Topology(object):
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.atomnames = \
 *             <strname_t *> PyMem_Realloc(NULL, sizeof(strname_t) * NAME_ALLOCATION_INCREMENT)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.Topology.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":917
 *         self.resids_to_internalids_allocated_size = RESIDUES_ALLOCATION_INCREMENT
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         cdef fsl_int i
 *         PyMem_Free(self.atomnames)
 */

/* Python wrapper */
static void __pyx_pw_10fatslimlib_9core_base_8Topology_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10fatslimlib_9core_base_8Topology_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_10fatslimlib_9core_base_8Topology_2__dealloc__(((struct __pyx_obj_10fatslimlib_9core_base_Topology *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_10fatslimlib_9core_base_8Topology_2__dealloc__(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self) {
  fsl_int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  fsl_int __pyx_t_1;
  fsl_int __pyx_t_2;
  fsl_int __pyx_t_3;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "fatslimlib/core_base.pyx":919
 *     def __dealloc__(self):
 *         cdef fsl_int i
 *         PyMem_Free(self.atomnames)             # <<<<<<<<<<<<<<
 *         PyMem_Free(self.resnames)
 * 
 */
  PyMem_Free(__pyx_v_self->atomnames);

  /* "fatslimlib/core_base.pyx":920
 *         cdef fsl_int i
 *         PyMem_Free(self.atomnames)
 *         PyMem_Free(self.resnames)             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.residues_size):
 */
  PyMem_Free(__pyx_v_self->resnames);

  /* "fatslimlib/core_base.pyx":922
 *         PyMem_Free(self.resnames)
 * 
 *         for i in range(self.residues_size):             # <<<<<<<<<<<<<<
 *             PyMem_Free(self.residues[i].atom_internal_ids)
 *             PyMem_Free(self.residues[i].atomids)
 */
  __pyx_t_1 = __pyx_v_self->residues_size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "fatslimlib/core_base.pyx":923
 * 
 *         for i in range(self.residues_size):
 *             PyMem_Free(self.residues[i].atom_internal_ids)             # <<<<<<<<<<<<<<
 *             PyMem_Free(self.residues[i].atomids)
 *         PyMem_Free(self.residues)
 */
    PyMem_Free((__pyx_v_self->residues[__pyx_v_i]).atom_internal_ids);

    /* "fatslimlib/core_base.pyx":924
 *         for i in range(self.residues_size):
 *             PyMem_Free(self.residues[i].atom_internal_ids)
 *             PyMem_Free(self.residues[i].atomids)             # <<<<<<<<<<<<<<
 *         PyMem_Free(self.residues)
 *         PyMem_Free(self.resids_to_internalids)
 */
    PyMem_Free((__pyx_v_self->residues[__pyx_v_i]).atomids);
  }

  /* "fatslimlib/core_base.pyx":925
 *             PyMem_Free(self.residues[i].atom_internal_ids)
 *             PyMem_Free(self.residues[i].atomids)
 *         PyMem_Free(self.residues)             # <<<<<<<<<<<<<<
 *         PyMem_Free(self.resids_to_internalids)
 * 
 */
  PyMem_Free(__pyx_v_self->residues);

  /* "fatslimlib/core_base.pyx":926
 *             PyMem_Free(self.residues[i].atomids)
 *         PyMem_Free(self.residues)
 *         PyMem_Free(self.resids_to_internalids)             # <<<<<<<<<<<<<<
 * 
 *         PyMem_Free(self.atoms)
 */
  PyMem_Free(__pyx_v_self->resids_to_internalids);

  /* "fatslimlib/core_base.pyx":928
 *         PyMem_Free(self.resids_to_internalids)
 * 
 *         PyMem_Free(self.atoms)             # <<<<<<<<<<<<<<
 *         PyMem_Free(self.atomids_to_internalids)
 * 
 */
  PyMem_Free(__pyx_v_self->atoms);

  /* "fatslimlib/core_base.pyx":929
 * 
 *         PyMem_Free(self.atoms)
 *         PyMem_Free(self.atomids_to_internalids)             # <<<<<<<<<<<<<<
 * 
 *     cdef void set_atoms_allocation(self, fsl_int size):
 */
  PyMem_Free(__pyx_v_self->atomids_to_internalids);

  /* "fatslimlib/core_base.pyx":917
 *         self.resids_to_internalids_allocated_size = RESIDUES_ALLOCATION_INCREMENT
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         cdef fsl_int i
 *         PyMem_Free(self.atomnames)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "fatslimlib/core_base.pyx":931
 *         PyMem_Free(self.atomids_to_internalids)
 * 
 *     cdef void set_atoms_allocation(self, fsl_int size):             # <<<<<<<<<<<<<<
 *         cdef fsl_int i
 *         self.atoms = <topol_atom_t *> PyMem_Realloc(<void *> self.atoms,
 */

static void __pyx_f_10fatslimlib_9core_base_8Topology_set_atoms_allocation(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_size) {
  fsl_int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  fsl_int __pyx_t_2;
  fsl_int __pyx_t_3;
  fsl_int __pyx_t_4;
  __Pyx_RefNannySetupContext("set_atoms_allocation", 0);

  /* "fatslimlib/core_base.pyx":933
 *     cdef void set_atoms_allocation(self, fsl_int size):
 *         cdef fsl_int i
 *         self.atoms = <topol_atom_t *> PyMem_Realloc(<void *> self.atoms,             # <<<<<<<<<<<<<<
 *                                                     sizeof(topol_atom_t) * size)
 *         if self.atoms == NULL:
 */
  __pyx_v_self->atoms = ((struct __pyx_t_10fatslimlib_9core_base_topol_atom_t *)PyMem_Realloc(((void *)__pyx_v_self->atoms), ((sizeof(struct __pyx_t_10fatslimlib_9core_base_topol_atom_t)) * __pyx_v_size)));

  /* "fatslimlib/core_base.pyx":935
 *         self.atoms = <topol_atom_t *> PyMem_Realloc(<void *> self.atoms,
 *                                                     sizeof(topol_atom_t) * size)
 *         if self.atoms == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->atoms == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "fatslimlib/core_base.pyx":936
 *                                                     sizeof(topol_atom_t) * size)
 *         if self.atoms == NULL:
 *             raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self.atoms_allocated_size, size):
 */
    PyErr_NoMemory(); __PYX_ERR(0, 936, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":935
 *         self.atoms = <topol_atom_t *> PyMem_Realloc(<void *> self.atoms,
 *                                                     sizeof(topol_atom_t) * size)
 *         if self.atoms == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":938
 *             raise MemoryError
 * 
 *         for i in range(self.atoms_allocated_size, size):             # <<<<<<<<<<<<<<
 *             self.atoms[i].atomid = NOTSET
 *             self.atoms[i].internal_id = NOTSET
 */
  __pyx_t_2 = __pyx_v_size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = __pyx_v_self->atoms_allocated_size; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "fatslimlib/core_base.pyx":939
 * 
 *         for i in range(self.atoms_allocated_size, size):
 *             self.atoms[i].atomid = NOTSET             # <<<<<<<<<<<<<<
 *             self.atoms[i].internal_id = NOTSET
 *             self.atoms[i].resid = NOTSET
 */
    (__pyx_v_self->atoms[__pyx_v_i]).atomid = -12345;

    /* "fatslimlib/core_base.pyx":940
 *         for i in range(self.atoms_allocated_size, size):
 *             self.atoms[i].atomid = NOTSET
 *             self.atoms[i].internal_id = NOTSET             # <<<<<<<<<<<<<<
 *             self.atoms[i].resid = NOTSET
 *             self.atoms[i].residue_internal_id = NOTSET
 */
    (__pyx_v_self->atoms[__pyx_v_i]).internal_id = -12345;

    /* "fatslimlib/core_base.pyx":941
 *             self.atoms[i].atomid = NOTSET
 *             self.atoms[i].internal_id = NOTSET
 *             self.atoms[i].resid = NOTSET             # <<<<<<<<<<<<<<
 *             self.atoms[i].residue_internal_id = NOTSET
 *             self.atoms[i].name_id = NOTSET
 */
    (__pyx_v_self->atoms[__pyx_v_i]).resid = -12345;

    /* "fatslimlib/core_base.pyx":942
 *             self.atoms[i].internal_id = NOTSET
 *             self.atoms[i].resid = NOTSET
 *             self.atoms[i].residue_internal_id = NOTSET             # <<<<<<<<<<<<<<
 *             self.atoms[i].name_id = NOTSET
 * 
 */
    (__pyx_v_self->atoms[__pyx_v_i]).residue_internal_id = -12345;

    /* "fatslimlib/core_base.pyx":943
 *             self.atoms[i].resid = NOTSET
 *             self.atoms[i].residue_internal_id = NOTSET
 *             self.atoms[i].name_id = NOTSET             # <<<<<<<<<<<<<<
 * 
 *         self.atoms_allocated_size = size
 */
    (__pyx_v_self->atoms[__pyx_v_i]).name_id = -12345;
  }

  /* "fatslimlib/core_base.pyx":945
 *             self.atoms[i].name_id = NOTSET
 * 
 *         self.atoms_allocated_size = size             # <<<<<<<<<<<<<<
 * 
 *     cdef void set_residues_allocation(self, fsl_int size):
 */
  __pyx_v_self->atoms_allocated_size = __pyx_v_size;

  /* "fatslimlib/core_base.pyx":931
 *         PyMem_Free(self.atomids_to_internalids)
 * 
 *     cdef void set_atoms_allocation(self, fsl_int size):             # <<<<<<<<<<<<<<
 *         cdef fsl_int i
 *         self.atoms = <topol_atom_t *> PyMem_Realloc(<void *> self.atoms,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("fatslimlib.core_base.Topology.set_atoms_allocation", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "fatslimlib/core_base.pyx":947
 *         self.atoms_allocated_size = size
 * 
 *     cdef void set_residues_allocation(self, fsl_int size):             # <<<<<<<<<<<<<<
 *         cdef fsl_int i
 *         self.residues = <topol_residue_t *> PyMem_Realloc(<void *>self.residues,
 */

static void __pyx_f_10fatslimlib_9core_base_8Topology_set_residues_allocation(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_size) {
  fsl_int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  fsl_int __pyx_t_2;
  fsl_int __pyx_t_3;
  fsl_int __pyx_t_4;
  __Pyx_RefNannySetupContext("set_residues_allocation", 0);

  /* "fatslimlib/core_base.pyx":949
 *     cdef void set_residues_allocation(self, fsl_int size):
 *         cdef fsl_int i
 *         self.residues = <topol_residue_t *> PyMem_Realloc(<void *>self.residues,             # <<<<<<<<<<<<<<
 *                                                           sizeof(topol_residue_t) * size)
 *         if self.residues == NULL:
 */
  __pyx_v_self->residues = ((struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *)PyMem_Realloc(((void *)__pyx_v_self->residues), ((sizeof(struct __pyx_t_10fatslimlib_9core_base_topol_residue_t)) * __pyx_v_size)));

  /* "fatslimlib/core_base.pyx":951
 *         self.residues = <topol_residue_t *> PyMem_Realloc(<void *>self.residues,
 *                                                           sizeof(topol_residue_t) * size)
 *         if self.residues == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->residues == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "fatslimlib/core_base.pyx":952
 *                                                           sizeof(topol_residue_t) * size)
 *         if self.residues == NULL:
 *             raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         # Initialize residues
 */
    PyErr_NoMemory(); __PYX_ERR(0, 952, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":951
 *         self.residues = <topol_residue_t *> PyMem_Realloc(<void *>self.residues,
 *                                                           sizeof(topol_residue_t) * size)
 *         if self.residues == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":955
 * 
 *         # Initialize residues
 *         for i in range(self.residues_allocated_size, size):             # <<<<<<<<<<<<<<
 * 
 *             self.residues[i].name_id = NOTSET
 */
  __pyx_t_2 = __pyx_v_size;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = __pyx_v_self->residues_allocated_size; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "fatslimlib/core_base.pyx":957
 *         for i in range(self.residues_allocated_size, size):
 * 
 *             self.residues[i].name_id = NOTSET             # <<<<<<<<<<<<<<
 *             self.residues[i].internal_id = NOTSET
 *             self.residues[i].resid = NOTSET
 */
    (__pyx_v_self->residues[__pyx_v_i]).name_id = -12345;

    /* "fatslimlib/core_base.pyx":958
 * 
 *             self.residues[i].name_id = NOTSET
 *             self.residues[i].internal_id = NOTSET             # <<<<<<<<<<<<<<
 *             self.residues[i].resid = NOTSET
 *             self.residues[i].atom_internal_ids = \
 */
    (__pyx_v_self->residues[__pyx_v_i]).internal_id = -12345;

    /* "fatslimlib/core_base.pyx":959
 *             self.residues[i].name_id = NOTSET
 *             self.residues[i].internal_id = NOTSET
 *             self.residues[i].resid = NOTSET             # <<<<<<<<<<<<<<
 *             self.residues[i].atom_internal_ids = \
 *                 <fsl_int *> PyMem_Realloc(NULL, sizeof(fsl_int) * RESIDUE_ALLOCATION_INCREMENT)
 */
    (__pyx_v_self->residues[__pyx_v_i]).resid = -12345;

    /* "fatslimlib/core_base.pyx":960
 *             self.residues[i].internal_id = NOTSET
 *             self.residues[i].resid = NOTSET
 *             self.residues[i].atom_internal_ids = \             # <<<<<<<<<<<<<<
 *                 <fsl_int *> PyMem_Realloc(NULL, sizeof(fsl_int) * RESIDUE_ALLOCATION_INCREMENT)
 *             self.residues[i].atomids = \
 */
    (__pyx_v_self->residues[__pyx_v_i]).atom_internal_ids = ((fsl_int *)PyMem_Realloc(NULL, ((sizeof(fsl_int)) * 20)));

    /* "fatslimlib/core_base.pyx":962
 *             self.residues[i].atom_internal_ids = \
 *                 <fsl_int *> PyMem_Realloc(NULL, sizeof(fsl_int) * RESIDUE_ALLOCATION_INCREMENT)
 *             self.residues[i].atomids = \             # <<<<<<<<<<<<<<
 *                 <fsl_int *> PyMem_Realloc(NULL, sizeof(fsl_int) * RESIDUE_ALLOCATION_INCREMENT)
 *             self.residues[i].allocated_size = RESIDUE_ALLOCATION_INCREMENT
 */
    (__pyx_v_self->residues[__pyx_v_i]).atomids = ((fsl_int *)PyMem_Realloc(NULL, ((sizeof(fsl_int)) * 20)));

    /* "fatslimlib/core_base.pyx":964
 *             self.residues[i].atomids = \
 *                 <fsl_int *> PyMem_Realloc(NULL, sizeof(fsl_int) * RESIDUE_ALLOCATION_INCREMENT)
 *             self.residues[i].allocated_size = RESIDUE_ALLOCATION_INCREMENT             # <<<<<<<<<<<<<<
 *             self.residues[i].size = 0
 * 
 */
    (__pyx_v_self->residues[__pyx_v_i]).allocated_size = 20;

    /* "fatslimlib/core_base.pyx":965
 *                 <fsl_int *> PyMem_Realloc(NULL, sizeof(fsl_int) * RESIDUE_ALLOCATION_INCREMENT)
 *             self.residues[i].allocated_size = RESIDUE_ALLOCATION_INCREMENT
 *             self.residues[i].size = 0             # <<<<<<<<<<<<<<
 * 
 *         self.residues_allocated_size = size
 */
    (__pyx_v_self->residues[__pyx_v_i]).size = 0;
  }

  /* "fatslimlib/core_base.pyx":967
 *             self.residues[i].size = 0
 * 
 *         self.residues_allocated_size = size             # <<<<<<<<<<<<<<
 * 
 *     def append(self, fsl_int resid, str resname_str, str atomname_str, fsl_int atomid, bint strip_name=True):
 */
  __pyx_v_self->residues_allocated_size = __pyx_v_size;

  /* "fatslimlib/core_base.pyx":947
 *         self.atoms_allocated_size = size
 * 
 *     cdef void set_residues_allocation(self, fsl_int size):             # <<<<<<<<<<<<<<
 *         cdef fsl_int i
 *         self.residues = <topol_residue_t *> PyMem_Realloc(<void *>self.residues,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("fatslimlib.core_base.Topology.set_residues_allocation", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "fatslimlib/core_base.pyx":969
 *         self.residues_allocated_size = size
 * 
 *     def append(self, fsl_int resid, str resname_str, str atomname_str, fsl_int atomid, bint strip_name=True):             # <<<<<<<<<<<<<<
 *         cdef bytes resname, atomname
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_8Topology_5append(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_8Topology_5append(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  fsl_int __pyx_v_resid;
  PyObject *__pyx_v_resname_str = 0;
  PyObject *__pyx_v_atomname_str = 0;
  fsl_int __pyx_v_atomid;
  int __pyx_v_strip_name;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_resid,&__pyx_n_s_resname_str,&__pyx_n_s_atomname_str,&__pyx_n_s_atomid,&__pyx_n_s_strip_name,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_resid)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_resname_str)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("append", 0, 4, 5, 1); __PYX_ERR(0, 969, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_atomname_str)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("append", 0, 4, 5, 2); __PYX_ERR(0, 969, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_atomid)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("append", 0, 4, 5, 3); __PYX_ERR(0, 969, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strip_name);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "append") < 0)) __PYX_ERR(0, 969, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_resid = __Pyx_PyInt_As_fsl_int(values[0]); if (unlikely((__pyx_v_resid == ((fsl_int)-1)) && PyErr_Occurred())) __PYX_ERR(0, 969, __pyx_L3_error)
    __pyx_v_resname_str = ((PyObject*)values[1]);
    __pyx_v_atomname_str = ((PyObject*)values[2]);
    __pyx_v_atomid = __Pyx_PyInt_As_fsl_int(values[3]); if (unlikely((__pyx_v_atomid == ((fsl_int)-1)) && PyErr_Occurred())) __PYX_ERR(0, 969, __pyx_L3_error)
    if (values[4]) {
      __pyx_v_strip_name = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_strip_name == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 969, __pyx_L3_error)
    } else {
      __pyx_v_strip_name = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("append", 0, 4, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 969, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.Topology.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_resname_str), (&PyUnicode_Type), 1, "resname_str", 1))) __PYX_ERR(0, 969, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_atomname_str), (&PyUnicode_Type), 1, "atomname_str", 1))) __PYX_ERR(0, 969, __pyx_L1_error)
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_8Topology_4append(((struct __pyx_obj_10fatslimlib_9core_base_Topology *)__pyx_v_self), __pyx_v_resid, __pyx_v_resname_str, __pyx_v_atomname_str, __pyx_v_atomid, __pyx_v_strip_name);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_8Topology_4append(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_resid, PyObject *__pyx_v_resname_str, PyObject *__pyx_v_atomname_str, fsl_int __pyx_v_atomid, int __pyx_v_strip_name) {
  PyObject *__pyx_v_resname = 0;
  PyObject *__pyx_v_atomname = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("append", 0);

  /* "fatslimlib/core_base.pyx":972
 *         cdef bytes resname, atomname
 * 
 *         if strip_name:             # <<<<<<<<<<<<<<
 *             resname = resname_str.strip().encode()
 *             atomname = atomname_str.strip().encode()
 */
  __pyx_t_1 = (__pyx_v_strip_name != 0);
  if (__pyx_t_1) {

    /* "fatslimlib/core_base.pyx":973
 * 
 *         if strip_name:
 *             resname = resname_str.strip().encode()             # <<<<<<<<<<<<<<
 *             atomname = atomname_str.strip().encode()
 *         else:
 */
    __pyx_t_3 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyUnicode_Type_strip, __pyx_v_resname_str); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 973, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 973, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 973, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 973, __pyx_L1_error)
    __pyx_v_resname = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "fatslimlib/core_base.pyx":974
 *         if strip_name:
 *             resname = resname_str.strip().encode()
 *             atomname = atomname_str.strip().encode()             # <<<<<<<<<<<<<<
 *         else:
 *             resname = resname_str.encode()
 */
    __pyx_t_4 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyUnicode_Type_strip, __pyx_v_atomname_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 974, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 974, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 974, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 974, __pyx_L1_error)
    __pyx_v_atomname = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "fatslimlib/core_base.pyx":972
 *         cdef bytes resname, atomname
 * 
 *         if strip_name:             # <<<<<<<<<<<<<<
 *             resname = resname_str.strip().encode()
 *             atomname = atomname_str.strip().encode()
 */
    goto __pyx_L3;
  }

  /* "fatslimlib/core_base.pyx":976
 *             atomname = atomname_str.strip().encode()
 *         else:
 *             resname = resname_str.encode()             # <<<<<<<<<<<<<<
 *             atomname = atomname_str.encode()
 * 
 */
  /*else*/ {
    if (unlikely(__pyx_v_resname_str == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
      __PYX_ERR(0, 976, __pyx_L1_error)
    }
    __pyx_t_2 = PyUnicode_AsEncodedString(__pyx_v_resname_str, NULL, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 976, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_resname = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "fatslimlib/core_base.pyx":977
 *         else:
 *             resname = resname_str.encode()
 *             atomname = atomname_str.encode()             # <<<<<<<<<<<<<<
 * 
 *         self.internal_append(resid, resname, atomname, atomid)
 */
    if (unlikely(__pyx_v_atomname_str == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
      __PYX_ERR(0, 977, __pyx_L1_error)
    }
    __pyx_t_2 = PyUnicode_AsEncodedString(__pyx_v_atomname_str, NULL, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_atomname = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "fatslimlib/core_base.pyx":979
 *             atomname = atomname_str.encode()
 * 
 *         self.internal_append(resid, resname, atomname, atomid)             # <<<<<<<<<<<<<<
 * 
 *     cdef void internal_append(self, fsl_int resid, bytes resname, bytes atomname, fsl_int atomid):
 */
  ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Topology *)__pyx_v_self->__pyx_vtab)->internal_append(__pyx_v_self, __pyx_v_resid, __pyx_v_resname, __pyx_v_atomname, __pyx_v_atomid);

  /* "fatslimlib/core_base.pyx":969
 *         self.residues_allocated_size = size
 * 
 *     def append(self, fsl_int resid, str resname_str, str atomname_str, fsl_int atomid, bint strip_name=True):             # <<<<<<<<<<<<<<
 *         cdef bytes resname, atomname
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("fatslimlib.core_base.Topology.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_resname);
  __Pyx_XDECREF(__pyx_v_atomname);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":981
 *         self.internal_append(resid, resname, atomname, atomid)
 * 
 *     cdef void internal_append(self, fsl_int resid, bytes resname, bytes atomname, fsl_int atomid):             # <<<<<<<<<<<<<<
 *         cdef fsl_int i
 *         cdef fsl_int atomname_id, resname_id
 */

static void __pyx_f_10fatslimlib_9core_base_8Topology_internal_append(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_resid, PyObject *__pyx_v_resname, PyObject *__pyx_v_atomname, fsl_int __pyx_v_atomid) {
  fsl_int __pyx_v_i;
  fsl_int __pyx_v_atomname_id;
  fsl_int __pyx_v_resname_id;
  fsl_int __pyx_v_residue_internal_id;
  fsl_int __pyx_v_atom_internal_id;
  int __pyx_v_residue_created;
  struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *__pyx_v_residue;
  struct __pyx_t_10fatslimlib_9core_base_topol_atom_t *__pyx_v_atom;
  __Pyx_RefNannyDeclarations
  fsl_int __pyx_t_1;
  fsl_int __pyx_t_2;
  fsl_int __pyx_t_3;
  char const *__pyx_t_4;
  int __pyx_t_5;
  char const *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("internal_append", 0);

  /* "fatslimlib/core_base.pyx":986
 *         cdef fsl_int residue_internal_id, atom_internal_id
 *         cdef fsl_int old_size
 *         cdef bint residue_created = False             # <<<<<<<<<<<<<<
 *         cdef topol_residue_t *residue
 *         cdef topol_atom_t *atom
 */
  __pyx_v_residue_created = 0;

  /* "fatslimlib/core_base.pyx":991
 * 
 *         # Handles atomname
 *         atomname_id = -1             # <<<<<<<<<<<<<<
 *         for i in range(self.atomnames_size):
 *             if strncmp(self.atomnames[i], atomname, NAME_SIZE-1) == 0:
 */
  __pyx_v_atomname_id = -1;

  /* "fatslimlib/core_base.pyx":992
 *         # Handles atomname
 *         atomname_id = -1
 *         for i in range(self.atomnames_size):             # <<<<<<<<<<<<<<
 *             if strncmp(self.atomnames[i], atomname, NAME_SIZE-1) == 0:
 *                 atomname_id = i
 */
  __pyx_t_1 = __pyx_v_self->atomnames_size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "fatslimlib/core_base.pyx":993
 *         atomname_id = -1
 *         for i in range(self.atomnames_size):
 *             if strncmp(self.atomnames[i], atomname, NAME_SIZE-1) == 0:             # <<<<<<<<<<<<<<
 *                 atomname_id = i
 *                 break
 */
    if (unlikely(__pyx_v_atomname == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 993, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PyBytes_AsString(__pyx_v_atomname); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 993, __pyx_L1_error)
    __pyx_t_5 = ((strncmp((__pyx_v_self->atomnames[__pyx_v_i]), __pyx_t_4, 5) == 0) != 0);
    if (__pyx_t_5) {

      /* "fatslimlib/core_base.pyx":994
 *         for i in range(self.atomnames_size):
 *             if strncmp(self.atomnames[i], atomname, NAME_SIZE-1) == 0:
 *                 atomname_id = i             # <<<<<<<<<<<<<<
 *                 break
 *         if atomname_id == -1: # We need to add the new atomname
 */
      __pyx_v_atomname_id = __pyx_v_i;

      /* "fatslimlib/core_base.pyx":995
 *             if strncmp(self.atomnames[i], atomname, NAME_SIZE-1) == 0:
 *                 atomname_id = i
 *                 break             # <<<<<<<<<<<<<<
 *         if atomname_id == -1: # We need to add the new atomname
 *             # Allocate memory if necessary
 */
      goto __pyx_L4_break;

      /* "fatslimlib/core_base.pyx":993
 *         atomname_id = -1
 *         for i in range(self.atomnames_size):
 *             if strncmp(self.atomnames[i], atomname, NAME_SIZE-1) == 0:             # <<<<<<<<<<<<<<
 *                 atomname_id = i
 *                 break
 */
    }
  }
  __pyx_L4_break:;

  /* "fatslimlib/core_base.pyx":996
 *                 atomname_id = i
 *                 break
 *         if atomname_id == -1: # We need to add the new atomname             # <<<<<<<<<<<<<<
 *             # Allocate memory if necessary
 *             if self.atomnames_size == self.atomnames_allocated_size:
 */
  __pyx_t_5 = ((__pyx_v_atomname_id == -1L) != 0);
  if (__pyx_t_5) {

    /* "fatslimlib/core_base.pyx":998
 *         if atomname_id == -1: # We need to add the new atomname
 *             # Allocate memory if necessary
 *             if self.atomnames_size == self.atomnames_allocated_size:             # <<<<<<<<<<<<<<
 *                 self.atomnames_allocated_size += NAME_ALLOCATION_INCREMENT
 *                 self.atomnames = <strname_t *> PyMem_Realloc(<void *> self.atomnames,
 */
    __pyx_t_5 = ((__pyx_v_self->atomnames_size == __pyx_v_self->atomnames_allocated_size) != 0);
    if (__pyx_t_5) {

      /* "fatslimlib/core_base.pyx":999
 *             # Allocate memory if necessary
 *             if self.atomnames_size == self.atomnames_allocated_size:
 *                 self.atomnames_allocated_size += NAME_ALLOCATION_INCREMENT             # <<<<<<<<<<<<<<
 *                 self.atomnames = <strname_t *> PyMem_Realloc(<void *> self.atomnames,
 *                                                              sizeof(strname_t) *
 */
      __pyx_v_self->atomnames_allocated_size = (__pyx_v_self->atomnames_allocated_size + 10);

      /* "fatslimlib/core_base.pyx":1000
 *             if self.atomnames_size == self.atomnames_allocated_size:
 *                 self.atomnames_allocated_size += NAME_ALLOCATION_INCREMENT
 *                 self.atomnames = <strname_t *> PyMem_Realloc(<void *> self.atomnames,             # <<<<<<<<<<<<<<
 *                                                              sizeof(strname_t) *
 *                                                              self.atomnames_allocated_size)
 */
      __pyx_v_self->atomnames = ((__pyx_t_10fatslimlib_9core_base_strname_t *)PyMem_Realloc(((void *)__pyx_v_self->atomnames), ((sizeof(__pyx_t_10fatslimlib_9core_base_strname_t)) * __pyx_v_self->atomnames_allocated_size)));

      /* "fatslimlib/core_base.pyx":1003
 *                                                              sizeof(strname_t) *
 *                                                              self.atomnames_allocated_size)
 *                 if self.atomnames == NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 * 
 */
      __pyx_t_5 = ((__pyx_v_self->atomnames == NULL) != 0);
      if (unlikely(__pyx_t_5)) {

        /* "fatslimlib/core_base.pyx":1004
 *                                                              self.atomnames_allocated_size)
 *                 if self.atomnames == NULL:
 *                     raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *             atomname_id = self.atomnames_size
 */
        PyErr_NoMemory(); __PYX_ERR(0, 1004, __pyx_L1_error)

        /* "fatslimlib/core_base.pyx":1003
 *                                                              sizeof(strname_t) *
 *                                                              self.atomnames_allocated_size)
 *                 if self.atomnames == NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 * 
 */
      }

      /* "fatslimlib/core_base.pyx":998
 *         if atomname_id == -1: # We need to add the new atomname
 *             # Allocate memory if necessary
 *             if self.atomnames_size == self.atomnames_allocated_size:             # <<<<<<<<<<<<<<
 *                 self.atomnames_allocated_size += NAME_ALLOCATION_INCREMENT
 *                 self.atomnames = <strname_t *> PyMem_Realloc(<void *> self.atomnames,
 */
    }

    /* "fatslimlib/core_base.pyx":1006
 *                     raise MemoryError
 * 
 *             atomname_id = self.atomnames_size             # <<<<<<<<<<<<<<
 *             self.atomnames_size += 1
 *             strncpy(self.atomnames[atomname_id], atomname, NAME_SIZE)
 */
    __pyx_t_1 = __pyx_v_self->atomnames_size;
    __pyx_v_atomname_id = __pyx_t_1;

    /* "fatslimlib/core_base.pyx":1007
 * 
 *             atomname_id = self.atomnames_size
 *             self.atomnames_size += 1             # <<<<<<<<<<<<<<
 *             strncpy(self.atomnames[atomname_id], atomname, NAME_SIZE)
 *             self.atomnames[atomname_id][NAME_SIZE-1] = b'\0'
 */
    __pyx_v_self->atomnames_size = (__pyx_v_self->atomnames_size + 1);

    /* "fatslimlib/core_base.pyx":1008
 *             atomname_id = self.atomnames_size
 *             self.atomnames_size += 1
 *             strncpy(self.atomnames[atomname_id], atomname, NAME_SIZE)             # <<<<<<<<<<<<<<
 *             self.atomnames[atomname_id][NAME_SIZE-1] = b'\0'
 * 
 */
    if (unlikely(__pyx_v_atomname == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 1008, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyBytes_AsString(__pyx_v_atomname); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 1008, __pyx_L1_error)
    (void)(strncpy((__pyx_v_self->atomnames[__pyx_v_atomname_id]), __pyx_t_6, 6));

    /* "fatslimlib/core_base.pyx":1009
 *             self.atomnames_size += 1
 *             strncpy(self.atomnames[atomname_id], atomname, NAME_SIZE)
 *             self.atomnames[atomname_id][NAME_SIZE-1] = b'\0'             # <<<<<<<<<<<<<<
 * 
 *         # Handles resname
 */
    ((__pyx_v_self->atomnames[__pyx_v_atomname_id])[5]) = '\x00';

    /* "fatslimlib/core_base.pyx":996
 *                 atomname_id = i
 *                 break
 *         if atomname_id == -1: # We need to add the new atomname             # <<<<<<<<<<<<<<
 *             # Allocate memory if necessary
 *             if self.atomnames_size == self.atomnames_allocated_size:
 */
  }

  /* "fatslimlib/core_base.pyx":1012
 * 
 *         # Handles resname
 *         resname_id = -1             # <<<<<<<<<<<<<<
 *         for i in range(self.resnames_size):
 *             if strncmp(self.resnames[i], resname, NAME_SIZE-1) == 0:
 */
  __pyx_v_resname_id = -1;

  /* "fatslimlib/core_base.pyx":1013
 *         # Handles resname
 *         resname_id = -1
 *         for i in range(self.resnames_size):             # <<<<<<<<<<<<<<
 *             if strncmp(self.resnames[i], resname, NAME_SIZE-1) == 0:
 *                 resname_id = i
 */
  __pyx_t_1 = __pyx_v_self->resnames_size;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "fatslimlib/core_base.pyx":1014
 *         resname_id = -1
 *         for i in range(self.resnames_size):
 *             if strncmp(self.resnames[i], resname, NAME_SIZE-1) == 0:             # <<<<<<<<<<<<<<
 *                 resname_id = i
 *                 break
 */
    if (unlikely(__pyx_v_resname == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 1014, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PyBytes_AsString(__pyx_v_resname); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 1014, __pyx_L1_error)
    __pyx_t_5 = ((strncmp((__pyx_v_self->resnames[__pyx_v_i]), __pyx_t_4, 5) == 0) != 0);
    if (__pyx_t_5) {

      /* "fatslimlib/core_base.pyx":1015
 *         for i in range(self.resnames_size):
 *             if strncmp(self.resnames[i], resname, NAME_SIZE-1) == 0:
 *                 resname_id = i             # <<<<<<<<<<<<<<
 *                 break
 *         if resname_id == -1: # We need to add the new resname
 */
      __pyx_v_resname_id = __pyx_v_i;

      /* "fatslimlib/core_base.pyx":1016
 *             if strncmp(self.resnames[i], resname, NAME_SIZE-1) == 0:
 *                 resname_id = i
 *                 break             # <<<<<<<<<<<<<<
 *         if resname_id == -1: # We need to add the new resname
 *             # Allocate memory if necessary
 */
      goto __pyx_L10_break;

      /* "fatslimlib/core_base.pyx":1014
 *         resname_id = -1
 *         for i in range(self.resnames_size):
 *             if strncmp(self.resnames[i], resname, NAME_SIZE-1) == 0:             # <<<<<<<<<<<<<<
 *                 resname_id = i
 *                 break
 */
    }
  }
  __pyx_L10_break:;

  /* "fatslimlib/core_base.pyx":1017
 *                 resname_id = i
 *                 break
 *         if resname_id == -1: # We need to add the new resname             # <<<<<<<<<<<<<<
 *             # Allocate memory if necessary
 *             if self.resnames_size == self.resnames_allocated_size:
 */
  __pyx_t_5 = ((__pyx_v_resname_id == -1L) != 0);
  if (__pyx_t_5) {

    /* "fatslimlib/core_base.pyx":1019
 *         if resname_id == -1: # We need to add the new resname
 *             # Allocate memory if necessary
 *             if self.resnames_size == self.resnames_allocated_size:             # <<<<<<<<<<<<<<
 *                 self.resnames_allocated_size += NAME_ALLOCATION_INCREMENT
 *                 self.resnames = <strname_t *> PyMem_Realloc(<void *> self.resnames,
 */
    __pyx_t_5 = ((__pyx_v_self->resnames_size == __pyx_v_self->resnames_allocated_size) != 0);
    if (__pyx_t_5) {

      /* "fatslimlib/core_base.pyx":1020
 *             # Allocate memory if necessary
 *             if self.resnames_size == self.resnames_allocated_size:
 *                 self.resnames_allocated_size += NAME_ALLOCATION_INCREMENT             # <<<<<<<<<<<<<<
 *                 self.resnames = <strname_t *> PyMem_Realloc(<void *> self.resnames,
 *                                                             sizeof(strname_t) *
 */
      __pyx_v_self->resnames_allocated_size = (__pyx_v_self->resnames_allocated_size + 10);

      /* "fatslimlib/core_base.pyx":1021
 *             if self.resnames_size == self.resnames_allocated_size:
 *                 self.resnames_allocated_size += NAME_ALLOCATION_INCREMENT
 *                 self.resnames = <strname_t *> PyMem_Realloc(<void *> self.resnames,             # <<<<<<<<<<<<<<
 *                                                             sizeof(strname_t) *
 *                                                             self.resnames_allocated_size)
 */
      __pyx_v_self->resnames = ((__pyx_t_10fatslimlib_9core_base_strname_t *)PyMem_Realloc(((void *)__pyx_v_self->resnames), ((sizeof(__pyx_t_10fatslimlib_9core_base_strname_t)) * __pyx_v_self->resnames_allocated_size)));

      /* "fatslimlib/core_base.pyx":1024
 *                                                             sizeof(strname_t) *
 *                                                             self.resnames_allocated_size)
 *                 if self.resnames == NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 *             resname_id = self.resnames_size
 */
      __pyx_t_5 = ((__pyx_v_self->resnames == NULL) != 0);
      if (unlikely(__pyx_t_5)) {

        /* "fatslimlib/core_base.pyx":1025
 *                                                             self.resnames_allocated_size)
 *                 if self.resnames == NULL:
 *                     raise MemoryError             # <<<<<<<<<<<<<<
 *             resname_id = self.resnames_size
 *             self.resnames_size += 1
 */
        PyErr_NoMemory(); __PYX_ERR(0, 1025, __pyx_L1_error)

        /* "fatslimlib/core_base.pyx":1024
 *                                                             sizeof(strname_t) *
 *                                                             self.resnames_allocated_size)
 *                 if self.resnames == NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError
 *             resname_id = self.resnames_size
 */
      }

      /* "fatslimlib/core_base.pyx":1019
 *         if resname_id == -1: # We need to add the new resname
 *             # Allocate memory if necessary
 *             if self.resnames_size == self.resnames_allocated_size:             # <<<<<<<<<<<<<<
 *                 self.resnames_allocated_size += NAME_ALLOCATION_INCREMENT
 *                 self.resnames = <strname_t *> PyMem_Realloc(<void *> self.resnames,
 */
    }

    /* "fatslimlib/core_base.pyx":1026
 *                 if self.resnames == NULL:
 *                     raise MemoryError
 *             resname_id = self.resnames_size             # <<<<<<<<<<<<<<
 *             self.resnames_size += 1
 *             strncpy(self.resnames[resname_id], resname, NAME_SIZE)
 */
    __pyx_t_1 = __pyx_v_self->resnames_size;
    __pyx_v_resname_id = __pyx_t_1;

    /* "fatslimlib/core_base.pyx":1027
 *                     raise MemoryError
 *             resname_id = self.resnames_size
 *             self.resnames_size += 1             # <<<<<<<<<<<<<<
 *             strncpy(self.resnames[resname_id], resname, NAME_SIZE)
 *             self.resnames[resname_id][NAME_SIZE-1] = b'\0'
 */
    __pyx_v_self->resnames_size = (__pyx_v_self->resnames_size + 1);

    /* "fatslimlib/core_base.pyx":1028
 *             resname_id = self.resnames_size
 *             self.resnames_size += 1
 *             strncpy(self.resnames[resname_id], resname, NAME_SIZE)             # <<<<<<<<<<<<<<
 *             self.resnames[resname_id][NAME_SIZE-1] = b'\0'
 * 
 */
    if (unlikely(__pyx_v_resname == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 1028, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyBytes_AsString(__pyx_v_resname); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 1028, __pyx_L1_error)
    (void)(strncpy((__pyx_v_self->resnames[__pyx_v_resname_id]), __pyx_t_6, 6));

    /* "fatslimlib/core_base.pyx":1029
 *             self.resnames_size += 1
 *             strncpy(self.resnames[resname_id], resname, NAME_SIZE)
 *             self.resnames[resname_id][NAME_SIZE-1] = b'\0'             # <<<<<<<<<<<<<<
 * 
 *         # Handles resid
 */
    ((__pyx_v_self->resnames[__pyx_v_resname_id])[5]) = '\x00';

    /* "fatslimlib/core_base.pyx":1017
 *                 resname_id = i
 *                 break
 *         if resname_id == -1: # We need to add the new resname             # <<<<<<<<<<<<<<
 *             # Allocate memory if necessary
 *             if self.resnames_size == self.resnames_allocated_size:
 */
  }

  /* "fatslimlib/core_base.pyx":1034
 * 
 *         # First we need to handle the translation resid->internalid
 *         if resid >= self.resids_to_internalids_allocated_size:             # <<<<<<<<<<<<<<
 *             # Allocate enough memory to store the new resid plus extra for future resids
 *             self.resids_to_internalids_allocated_size += \
 */
  __pyx_t_5 = ((__pyx_v_resid >= __pyx_v_self->resids_to_internalids_allocated_size) != 0);
  if (__pyx_t_5) {

    /* "fatslimlib/core_base.pyx":1036
 *         if resid >= self.resids_to_internalids_allocated_size:
 *             # Allocate enough memory to store the new resid plus extra for future resids
 *             self.resids_to_internalids_allocated_size += \             # <<<<<<<<<<<<<<
 *                 (resid - self.resids_to_internalids_allocated_size) + RESIDUES_ALLOCATION_INCREMENT
 *             self.resids_to_internalids = \
 */
    __pyx_v_self->resids_to_internalids_allocated_size = (__pyx_v_self->resids_to_internalids_allocated_size + ((__pyx_v_resid - __pyx_v_self->resids_to_internalids_allocated_size) + 50));

    /* "fatslimlib/core_base.pyx":1038
 *             self.resids_to_internalids_allocated_size += \
 *                 (resid - self.resids_to_internalids_allocated_size) + RESIDUES_ALLOCATION_INCREMENT
 *             self.resids_to_internalids = \             # <<<<<<<<<<<<<<
 *                 <fsl_int *> PyMem_Realloc(<void *>self.resids_to_internalids,
 *                                       sizeof(fsl_int) * self.resids_to_internalids_allocated_size)
 */
    __pyx_v_self->resids_to_internalids = ((fsl_int *)PyMem_Realloc(((void *)__pyx_v_self->resids_to_internalids), ((sizeof(fsl_int)) * __pyx_v_self->resids_to_internalids_allocated_size)));

    /* "fatslimlib/core_base.pyx":1041
 *                 <fsl_int *> PyMem_Realloc(<void *>self.resids_to_internalids,
 *                                       sizeof(fsl_int) * self.resids_to_internalids_allocated_size)
 *             if self.resids_to_internalids == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    __pyx_t_5 = ((__pyx_v_self->resids_to_internalids == NULL) != 0);
    if (unlikely(__pyx_t_5)) {

      /* "fatslimlib/core_base.pyx":1042
 *                                       sizeof(fsl_int) * self.resids_to_internalids_allocated_size)
 *             if self.resids_to_internalids == NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *             # We also need to initialize the newly allocated memory:
 */
      PyErr_NoMemory(); __PYX_ERR(0, 1042, __pyx_L1_error)

      /* "fatslimlib/core_base.pyx":1041
 *                 <fsl_int *> PyMem_Realloc(<void *>self.resids_to_internalids,
 *                                       sizeof(fsl_int) * self.resids_to_internalids_allocated_size)
 *             if self.resids_to_internalids == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    }

    /* "fatslimlib/core_base.pyx":1046
 *             # We also need to initialize the newly allocated memory:
 *             for i in range(self.resids_to_internalids_size,
 *                            self.resids_to_internalids_allocated_size):             # <<<<<<<<<<<<<<
 *                 self.resids_to_internalids[i] = NOTSET
 *             self.resids_to_internalids_size = resid
 */
    __pyx_t_1 = __pyx_v_self->resids_to_internalids_allocated_size;

    /* "fatslimlib/core_base.pyx":1045
 * 
 *             # We also need to initialize the newly allocated memory:
 *             for i in range(self.resids_to_internalids_size,             # <<<<<<<<<<<<<<
 *                            self.resids_to_internalids_allocated_size):
 *                 self.resids_to_internalids[i] = NOTSET
 */
    __pyx_t_2 = __pyx_t_1;
    for (__pyx_t_3 = __pyx_v_self->resids_to_internalids_size; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
      __pyx_v_i = __pyx_t_3;

      /* "fatslimlib/core_base.pyx":1047
 *             for i in range(self.resids_to_internalids_size,
 *                            self.resids_to_internalids_allocated_size):
 *                 self.resids_to_internalids[i] = NOTSET             # <<<<<<<<<<<<<<
 *             self.resids_to_internalids_size = resid
 * 
 */
      (__pyx_v_self->resids_to_internalids[__pyx_v_i]) = -12345;
    }

    /* "fatslimlib/core_base.pyx":1048
 *                            self.resids_to_internalids_allocated_size):
 *                 self.resids_to_internalids[i] = NOTSET
 *             self.resids_to_internalids_size = resid             # <<<<<<<<<<<<<<
 * 
 *         # Watch out! resids start from 1 but are stored from 0
 */
    __pyx_v_self->resids_to_internalids_size = __pyx_v_resid;

    /* "fatslimlib/core_base.pyx":1034
 * 
 *         # First we need to handle the translation resid->internalid
 *         if resid >= self.resids_to_internalids_allocated_size:             # <<<<<<<<<<<<<<
 *             # Allocate enough memory to store the new resid plus extra for future resids
 *             self.resids_to_internalids_allocated_size += \
 */
  }

  /* "fatslimlib/core_base.pyx":1051
 * 
 *         # Watch out! resids start from 1 but are stored from 0
 *         residue_internal_id = self.resids_to_internalids[resid-1]             # <<<<<<<<<<<<<<
 * 
 *         # Second we need to handle the internal id
 */
  __pyx_v_residue_internal_id = (__pyx_v_self->resids_to_internalids[(__pyx_v_resid - 1)]);

  /* "fatslimlib/core_base.pyx":1054
 * 
 *         # Second we need to handle the internal id
 *         if residue_internal_id == NOTSET: # The residue is not registered yet             # <<<<<<<<<<<<<<
 *             # Allocate memory if necessary
 *             if self.residues_size == self.residues_allocated_size:
 */
  __pyx_t_5 = ((__pyx_v_residue_internal_id == -12345L) != 0);
  if (__pyx_t_5) {

    /* "fatslimlib/core_base.pyx":1056
 *         if residue_internal_id == NOTSET: # The residue is not registered yet
 *             # Allocate memory if necessary
 *             if self.residues_size == self.residues_allocated_size:             # <<<<<<<<<<<<<<
 *                 self.set_residues_allocation(self.residues_allocated_size
 *                                              + RESIDUES_ALLOCATION_INCREMENT)
 */
    __pyx_t_5 = ((__pyx_v_self->residues_size == __pyx_v_self->residues_allocated_size) != 0);
    if (__pyx_t_5) {

      /* "fatslimlib/core_base.pyx":1057
 *             # Allocate memory if necessary
 *             if self.residues_size == self.residues_allocated_size:
 *                 self.set_residues_allocation(self.residues_allocated_size             # <<<<<<<<<<<<<<
 *                                              + RESIDUES_ALLOCATION_INCREMENT)
 * 
 */
      ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Topology *)__pyx_v_self->__pyx_vtab)->set_residues_allocation(__pyx_v_self, (__pyx_v_self->residues_allocated_size + 50));

      /* "fatslimlib/core_base.pyx":1056
 *         if residue_internal_id == NOTSET: # The residue is not registered yet
 *             # Allocate memory if necessary
 *             if self.residues_size == self.residues_allocated_size:             # <<<<<<<<<<<<<<
 *                 self.set_residues_allocation(self.residues_allocated_size
 *                                              + RESIDUES_ALLOCATION_INCREMENT)
 */
    }

    /* "fatslimlib/core_base.pyx":1060
 *                                              + RESIDUES_ALLOCATION_INCREMENT)
 * 
 *             residue_internal_id = self.residues_size             # <<<<<<<<<<<<<<
 *             self.residues_size += 1
 *             residue_created = True
 */
    __pyx_t_1 = __pyx_v_self->residues_size;
    __pyx_v_residue_internal_id = __pyx_t_1;

    /* "fatslimlib/core_base.pyx":1061
 * 
 *             residue_internal_id = self.residues_size
 *             self.residues_size += 1             # <<<<<<<<<<<<<<
 *             residue_created = True
 *             self.resids_to_internalids[resid - 1] = residue_internal_id
 */
    __pyx_v_self->residues_size = (__pyx_v_self->residues_size + 1);

    /* "fatslimlib/core_base.pyx":1062
 *             residue_internal_id = self.residues_size
 *             self.residues_size += 1
 *             residue_created = True             # <<<<<<<<<<<<<<
 *             self.resids_to_internalids[resid - 1] = residue_internal_id
 *             self.resids_to_internalids_size = resid
 */
    __pyx_v_residue_created = 1;

    /* "fatslimlib/core_base.pyx":1063
 *             self.residues_size += 1
 *             residue_created = True
 *             self.resids_to_internalids[resid - 1] = residue_internal_id             # <<<<<<<<<<<<<<
 *             self.resids_to_internalids_size = resid
 * 
 */
    (__pyx_v_self->resids_to_internalids[(__pyx_v_resid - 1)]) = __pyx_v_residue_internal_id;

    /* "fatslimlib/core_base.pyx":1064
 *             residue_created = True
 *             self.resids_to_internalids[resid - 1] = residue_internal_id
 *             self.resids_to_internalids_size = resid             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_self->resids_to_internalids_size = __pyx_v_resid;

    /* "fatslimlib/core_base.pyx":1054
 * 
 *         # Second we need to handle the internal id
 *         if residue_internal_id == NOTSET: # The residue is not registered yet             # <<<<<<<<<<<<<<
 *             # Allocate memory if necessary
 *             if self.residues_size == self.residues_allocated_size:
 */
  }

  /* "fatslimlib/core_base.pyx":1070
 * 
 *         # First we need to handle the translation atomid->internalid
 *         if atomid >= self.atomids_to_internalids_allocated_size:             # <<<<<<<<<<<<<<
 *             # Allocate enough memory to store the new resid plus extra for future resids
 *             self.atomids_to_internalids_allocated_size += \
 */
  __pyx_t_5 = ((__pyx_v_atomid >= __pyx_v_self->atomids_to_internalids_allocated_size) != 0);
  if (__pyx_t_5) {

    /* "fatslimlib/core_base.pyx":1072
 *         if atomid >= self.atomids_to_internalids_allocated_size:
 *             # Allocate enough memory to store the new resid plus extra for future resids
 *             self.atomids_to_internalids_allocated_size += \             # <<<<<<<<<<<<<<
 *                 (atomid - self.atomids_to_internalids_allocated_size) + ATOMS_ALLOCATION_INCREMENT
 *             self.atomids_to_internalids = \
 */
    __pyx_v_self->atomids_to_internalids_allocated_size = (__pyx_v_self->atomids_to_internalids_allocated_size + ((__pyx_v_atomid - __pyx_v_self->atomids_to_internalids_allocated_size) + 0x3E8));

    /* "fatslimlib/core_base.pyx":1074
 *             self.atomids_to_internalids_allocated_size += \
 *                 (atomid - self.atomids_to_internalids_allocated_size) + ATOMS_ALLOCATION_INCREMENT
 *             self.atomids_to_internalids = \             # <<<<<<<<<<<<<<
 *                 <fsl_int *> PyMem_Realloc(<void *>self.atomids_to_internalids,
 *                                       sizeof(fsl_int) * self.atomids_to_internalids_allocated_size)
 */
    __pyx_v_self->atomids_to_internalids = ((fsl_int *)PyMem_Realloc(((void *)__pyx_v_self->atomids_to_internalids), ((sizeof(fsl_int)) * __pyx_v_self->atomids_to_internalids_allocated_size)));

    /* "fatslimlib/core_base.pyx":1077
 *                 <fsl_int *> PyMem_Realloc(<void *>self.atomids_to_internalids,
 *                                       sizeof(fsl_int) * self.atomids_to_internalids_allocated_size)
 *             if self.atomids_to_internalids == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    __pyx_t_5 = ((__pyx_v_self->atomids_to_internalids == NULL) != 0);
    if (unlikely(__pyx_t_5)) {

      /* "fatslimlib/core_base.pyx":1078
 *                                       sizeof(fsl_int) * self.atomids_to_internalids_allocated_size)
 *             if self.atomids_to_internalids == NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *             # We also need to initialize the newly allocated memory:
 */
      PyErr_NoMemory(); __PYX_ERR(0, 1078, __pyx_L1_error)

      /* "fatslimlib/core_base.pyx":1077
 *                 <fsl_int *> PyMem_Realloc(<void *>self.atomids_to_internalids,
 *                                       sizeof(fsl_int) * self.atomids_to_internalids_allocated_size)
 *             if self.atomids_to_internalids == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    }

    /* "fatslimlib/core_base.pyx":1082
 *             # We also need to initialize the newly allocated memory:
 *             for i in range(self.atomids_to_internalids_size,
 *                            self.atomids_to_internalids_allocated_size):             # <<<<<<<<<<<<<<
 *                 self.atomids_to_internalids[i] = NOTSET
 * 
 */
    __pyx_t_1 = __pyx_v_self->atomids_to_internalids_allocated_size;

    /* "fatslimlib/core_base.pyx":1081
 * 
 *             # We also need to initialize the newly allocated memory:
 *             for i in range(self.atomids_to_internalids_size,             # <<<<<<<<<<<<<<
 *                            self.atomids_to_internalids_allocated_size):
 *                 self.atomids_to_internalids[i] = NOTSET
 */
    __pyx_t_2 = __pyx_t_1;
    for (__pyx_t_3 = __pyx_v_self->atomids_to_internalids_size; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
      __pyx_v_i = __pyx_t_3;

      /* "fatslimlib/core_base.pyx":1083
 *             for i in range(self.atomids_to_internalids_size,
 *                            self.atomids_to_internalids_allocated_size):
 *                 self.atomids_to_internalids[i] = NOTSET             # <<<<<<<<<<<<<<
 * 
 *         # Watch out! atomids start from 1 but are stored from 0
 */
      (__pyx_v_self->atomids_to_internalids[__pyx_v_i]) = -12345;
    }

    /* "fatslimlib/core_base.pyx":1070
 * 
 *         # First we need to handle the translation atomid->internalid
 *         if atomid >= self.atomids_to_internalids_allocated_size:             # <<<<<<<<<<<<<<
 *             # Allocate enough memory to store the new resid plus extra for future resids
 *             self.atomids_to_internalids_allocated_size += \
 */
  }

  /* "fatslimlib/core_base.pyx":1086
 * 
 *         # Watch out! atomids start from 1 but are stored from 0
 *         atom_internal_id = self.atomids_to_internalids[atomid-1]             # <<<<<<<<<<<<<<
 * 
 *         # Second we create the atom
 */
  __pyx_v_atom_internal_id = (__pyx_v_self->atomids_to_internalids[(__pyx_v_atomid - 1)]);

  /* "fatslimlib/core_base.pyx":1089
 * 
 *         # Second we create the atom
 *         if atom_internal_id != NOTSET: # The atom is not registered yet             # <<<<<<<<<<<<<<
 *             raise IndexError("Atom with atomid #%i is already registered" % atomid)
 * 
 */
  __pyx_t_5 = ((__pyx_v_atom_internal_id != -12345L) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "fatslimlib/core_base.pyx":1090
 *         # Second we create the atom
 *         if atom_internal_id != NOTSET: # The atom is not registered yet
 *             raise IndexError("Atom with atomid #%i is already registered" % atomid)             # <<<<<<<<<<<<<<
 * 
 *         # Allocate memory if necessary
 */
    __pyx_t_7 = __Pyx_PyInt_From_fsl_int(__pyx_v_atomid); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1090, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_Atom_with_atomid_i_is_already_re, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1090, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1090, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 1090, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":1089
 * 
 *         # Second we create the atom
 *         if atom_internal_id != NOTSET: # The atom is not registered yet             # <<<<<<<<<<<<<<
 *             raise IndexError("Atom with atomid #%i is already registered" % atomid)
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":1093
 * 
 *         # Allocate memory if necessary
 *         if self.atoms_size == self.atoms_allocated_size:             # <<<<<<<<<<<<<<
 *             self.set_atoms_allocation(self.atoms_allocated_size + ATOMS_ALLOCATION_INCREMENT)
 * 
 */
  __pyx_t_5 = ((__pyx_v_self->atoms_size == __pyx_v_self->atoms_allocated_size) != 0);
  if (__pyx_t_5) {

    /* "fatslimlib/core_base.pyx":1094
 *         # Allocate memory if necessary
 *         if self.atoms_size == self.atoms_allocated_size:
 *             self.set_atoms_allocation(self.atoms_allocated_size + ATOMS_ALLOCATION_INCREMENT)             # <<<<<<<<<<<<<<
 * 
 *         atom_internal_id = self.atoms_size
 */
    ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Topology *)__pyx_v_self->__pyx_vtab)->set_atoms_allocation(__pyx_v_self, (__pyx_v_self->atoms_allocated_size + 0x3E8));

    /* "fatslimlib/core_base.pyx":1093
 * 
 *         # Allocate memory if necessary
 *         if self.atoms_size == self.atoms_allocated_size:             # <<<<<<<<<<<<<<
 *             self.set_atoms_allocation(self.atoms_allocated_size + ATOMS_ALLOCATION_INCREMENT)
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":1096
 *             self.set_atoms_allocation(self.atoms_allocated_size + ATOMS_ALLOCATION_INCREMENT)
 * 
 *         atom_internal_id = self.atoms_size             # <<<<<<<<<<<<<<
 *         self.atoms_size += 1
 * 
 */
  __pyx_t_1 = __pyx_v_self->atoms_size;
  __pyx_v_atom_internal_id = __pyx_t_1;

  /* "fatslimlib/core_base.pyx":1097
 * 
 *         atom_internal_id = self.atoms_size
 *         self.atoms_size += 1             # <<<<<<<<<<<<<<
 * 
 *         # Create the atom
 */
  __pyx_v_self->atoms_size = (__pyx_v_self->atoms_size + 1);

  /* "fatslimlib/core_base.pyx":1100
 * 
 *         # Create the atom
 *         atom = &self.atoms[atom_internal_id]             # <<<<<<<<<<<<<<
 *         atom.name_id = atomname_id
 *         atom.internal_id = atom_internal_id
 */
  __pyx_v_atom = (&(__pyx_v_self->atoms[__pyx_v_atom_internal_id]));

  /* "fatslimlib/core_base.pyx":1101
 *         # Create the atom
 *         atom = &self.atoms[atom_internal_id]
 *         atom.name_id = atomname_id             # <<<<<<<<<<<<<<
 *         atom.internal_id = atom_internal_id
 *         atom.atomid = atomid
 */
  __pyx_v_atom->name_id = __pyx_v_atomname_id;

  /* "fatslimlib/core_base.pyx":1102
 *         atom = &self.atoms[atom_internal_id]
 *         atom.name_id = atomname_id
 *         atom.internal_id = atom_internal_id             # <<<<<<<<<<<<<<
 *         atom.atomid = atomid
 *         self.atomids_to_internalids[atomid-1] = atom_internal_id
 */
  __pyx_v_atom->internal_id = __pyx_v_atom_internal_id;

  /* "fatslimlib/core_base.pyx":1103
 *         atom.name_id = atomname_id
 *         atom.internal_id = atom_internal_id
 *         atom.atomid = atomid             # <<<<<<<<<<<<<<
 *         self.atomids_to_internalids[atomid-1] = atom_internal_id
 *         self.atomids_to_internalids_size = atomid
 */
  __pyx_v_atom->atomid = __pyx_v_atomid;

  /* "fatslimlib/core_base.pyx":1104
 *         atom.internal_id = atom_internal_id
 *         atom.atomid = atomid
 *         self.atomids_to_internalids[atomid-1] = atom_internal_id             # <<<<<<<<<<<<<<
 *         self.atomids_to_internalids_size = atomid
 * 
 */
  (__pyx_v_self->atomids_to_internalids[(__pyx_v_atomid - 1)]) = __pyx_v_atom_internal_id;

  /* "fatslimlib/core_base.pyx":1105
 *         atom.atomid = atomid
 *         self.atomids_to_internalids[atomid-1] = atom_internal_id
 *         self.atomids_to_internalids_size = atomid             # <<<<<<<<<<<<<<
 * 
 *         # Populate the residue
 */
  __pyx_v_self->atomids_to_internalids_size = __pyx_v_atomid;

  /* "fatslimlib/core_base.pyx":1108
 * 
 *         # Populate the residue
 *         residue = &self.residues[residue_internal_id]             # <<<<<<<<<<<<<<
 * 
 *         if residue_created:
 */
  __pyx_v_residue = (&(__pyx_v_self->residues[__pyx_v_residue_internal_id]));

  /* "fatslimlib/core_base.pyx":1110
 *         residue = &self.residues[residue_internal_id]
 * 
 *         if residue_created:             # <<<<<<<<<<<<<<
 *             residue.name_id = resname_id
 *             residue.internal_id = residue_internal_id
 */
  __pyx_t_5 = (__pyx_v_residue_created != 0);
  if (__pyx_t_5) {

    /* "fatslimlib/core_base.pyx":1111
 * 
 *         if residue_created:
 *             residue.name_id = resname_id             # <<<<<<<<<<<<<<
 *             residue.internal_id = residue_internal_id
 *             residue.resid = resid
 */
    __pyx_v_residue->name_id = __pyx_v_resname_id;

    /* "fatslimlib/core_base.pyx":1112
 *         if residue_created:
 *             residue.name_id = resname_id
 *             residue.internal_id = residue_internal_id             # <<<<<<<<<<<<<<
 *             residue.resid = resid
 * 
 */
    __pyx_v_residue->internal_id = __pyx_v_residue_internal_id;

    /* "fatslimlib/core_base.pyx":1113
 *             residue.name_id = resname_id
 *             residue.internal_id = residue_internal_id
 *             residue.resid = resid             # <<<<<<<<<<<<<<
 * 
 *         # Add the atom to the residue
 */
    __pyx_v_residue->resid = __pyx_v_resid;

    /* "fatslimlib/core_base.pyx":1110
 *         residue = &self.residues[residue_internal_id]
 * 
 *         if residue_created:             # <<<<<<<<<<<<<<
 *             residue.name_id = resname_id
 *             residue.internal_id = residue_internal_id
 */
  }

  /* "fatslimlib/core_base.pyx":1116
 * 
 *         # Add the atom to the residue
 *         topol_residue_append_atom(residue, atom)             # <<<<<<<<<<<<<<
 * 
 *     cdef fsl_int fast_get_atom_internal_id(self, fsl_int atomid) nogil:
 */
  __pyx_f_10fatslimlib_9core_base_topol_residue_append_atom(__pyx_v_residue, __pyx_v_atom);

  /* "fatslimlib/core_base.pyx":981
 *         self.internal_append(resid, resname, atomname, atomid)
 * 
 *     cdef void internal_append(self, fsl_int resid, bytes resname, bytes atomname, fsl_int atomid):             # <<<<<<<<<<<<<<
 *         cdef fsl_int i
 *         cdef fsl_int atomname_id, resname_id
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_WriteUnraisable("fatslimlib.core_base.Topology.internal_append", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "fatslimlib/core_base.pyx":1118
 *         topol_residue_append_atom(residue, atom)
 * 
 *     cdef fsl_int fast_get_atom_internal_id(self, fsl_int atomid) nogil:             # <<<<<<<<<<<<<<
 *         if not 0 < atomid <= self.atomids_to_internalids_size:
 *             return -1
 */

static fsl_int __pyx_f_10fatslimlib_9core_base_8Topology_fast_get_atom_internal_id(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_atomid) {
  fsl_int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "fatslimlib/core_base.pyx":1119
 * 
 *     cdef fsl_int fast_get_atom_internal_id(self, fsl_int atomid) nogil:
 *         if not 0 < atomid <= self.atomids_to_internalids_size:             # <<<<<<<<<<<<<<
 *             return -1
 *         return self.atomids_to_internalids[atomid-1]
 */
  __pyx_t_1 = (0 < __pyx_v_atomid);
  if (__pyx_t_1) {
    __pyx_t_1 = (__pyx_v_atomid <= __pyx_v_self->atomids_to_internalids_size);
  }
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "fatslimlib/core_base.pyx":1120
 *     cdef fsl_int fast_get_atom_internal_id(self, fsl_int atomid) nogil:
 *         if not 0 < atomid <= self.atomids_to_internalids_size:
 *             return -1             # <<<<<<<<<<<<<<
 *         return self.atomids_to_internalids[atomid-1]
 * 
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "fatslimlib/core_base.pyx":1119
 * 
 *     cdef fsl_int fast_get_atom_internal_id(self, fsl_int atomid) nogil:
 *         if not 0 < atomid <= self.atomids_to_internalids_size:             # <<<<<<<<<<<<<<
 *             return -1
 *         return self.atomids_to_internalids[atomid-1]
 */
  }

  /* "fatslimlib/core_base.pyx":1121
 *         if not 0 < atomid <= self.atomids_to_internalids_size:
 *             return -1
 *         return self.atomids_to_internalids[atomid-1]             # <<<<<<<<<<<<<<
 * 
 *     cdef topol_atom_t *fast_get_atom(self, fsl_int atomid) nogil:
 */
  __pyx_r = (__pyx_v_self->atomids_to_internalids[(__pyx_v_atomid - 1)]);
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1118
 *         topol_residue_append_atom(residue, atom)
 * 
 *     cdef fsl_int fast_get_atom_internal_id(self, fsl_int atomid) nogil:             # <<<<<<<<<<<<<<
 *         if not 0 < atomid <= self.atomids_to_internalids_size:
 *             return -1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1123
 *         return self.atomids_to_internalids[atomid-1]
 * 
 *     cdef topol_atom_t *fast_get_atom(self, fsl_int atomid) nogil:             # <<<<<<<<<<<<<<
 *         cdef fsl_int internal_id = self.fast_get_atom_internal_id(atomid)
 * 
 */

static struct __pyx_t_10fatslimlib_9core_base_topol_atom_t *__pyx_f_10fatslimlib_9core_base_8Topology_fast_get_atom(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_atomid) {
  fsl_int __pyx_v_internal_id;
  struct __pyx_t_10fatslimlib_9core_base_topol_atom_t *__pyx_r;
  int __pyx_t_1;

  /* "fatslimlib/core_base.pyx":1124
 * 
 *     cdef topol_atom_t *fast_get_atom(self, fsl_int atomid) nogil:
 *         cdef fsl_int internal_id = self.fast_get_atom_internal_id(atomid)             # <<<<<<<<<<<<<<
 * 
 *         if internal_id < 0:
 */
  __pyx_v_internal_id = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Topology *)__pyx_v_self->__pyx_vtab)->fast_get_atom_internal_id(__pyx_v_self, __pyx_v_atomid);

  /* "fatslimlib/core_base.pyx":1126
 *         cdef fsl_int internal_id = self.fast_get_atom_internal_id(atomid)
 * 
 *         if internal_id < 0:             # <<<<<<<<<<<<<<
 *             return NULL
 * 
 */
  __pyx_t_1 = ((__pyx_v_internal_id < 0) != 0);
  if (__pyx_t_1) {

    /* "fatslimlib/core_base.pyx":1127
 * 
 *         if internal_id < 0:
 *             return NULL             # <<<<<<<<<<<<<<
 * 
 *         return &self.atoms[internal_id]
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "fatslimlib/core_base.pyx":1126
 *         cdef fsl_int internal_id = self.fast_get_atom_internal_id(atomid)
 * 
 *         if internal_id < 0:             # <<<<<<<<<<<<<<
 *             return NULL
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":1129
 *             return NULL
 * 
 *         return &self.atoms[internal_id]             # <<<<<<<<<<<<<<
 * 
 *     def get_atom(self, fsl_int atomid):
 */
  __pyx_r = (&(__pyx_v_self->atoms[__pyx_v_internal_id]));
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1123
 *         return self.atomids_to_internalids[atomid-1]
 * 
 *     cdef topol_atom_t *fast_get_atom(self, fsl_int atomid) nogil:             # <<<<<<<<<<<<<<
 *         cdef fsl_int internal_id = self.fast_get_atom_internal_id(atomid)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1131
 *         return &self.atoms[internal_id]
 * 
 *     def get_atom(self, fsl_int atomid):             # <<<<<<<<<<<<<<
 *         cdef topol_atom_t atom
 *         cdef topol_residue_t residue
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_8Topology_7get_atom(PyObject *__pyx_v_self, PyObject *__pyx_arg_atomid); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_8Topology_7get_atom(PyObject *__pyx_v_self, PyObject *__pyx_arg_atomid) {
  fsl_int __pyx_v_atomid;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_atom (wrapper)", 0);
  assert(__pyx_arg_atomid); {
    __pyx_v_atomid = __Pyx_PyInt_As_fsl_int(__pyx_arg_atomid); if (unlikely((__pyx_v_atomid == ((fsl_int)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1131, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.Topology.get_atom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_8Topology_6get_atom(((struct __pyx_obj_10fatslimlib_9core_base_Topology *)__pyx_v_self), ((fsl_int)__pyx_v_atomid));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_8Topology_6get_atom(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_atomid) {
  struct __pyx_t_10fatslimlib_9core_base_topol_atom_t __pyx_v_atom;
  struct __pyx_t_10fatslimlib_9core_base_topol_residue_t __pyx_v_residue;
  fsl_int __pyx_v_internal_id;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  char *__pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("get_atom", 0);

  /* "fatslimlib/core_base.pyx":1134
 *         cdef topol_atom_t atom
 *         cdef topol_residue_t residue
 *         cdef fsl_int internal_id = self.fast_get_atom_internal_id(atomid)             # <<<<<<<<<<<<<<
 *         cdef bytes resname, atomname
 * 
 */
  __pyx_v_internal_id = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Topology *)__pyx_v_self->__pyx_vtab)->fast_get_atom_internal_id(__pyx_v_self, __pyx_v_atomid);

  /* "fatslimlib/core_base.pyx":1137
 *         cdef bytes resname, atomname
 * 
 *         if internal_id < 0:             # <<<<<<<<<<<<<<
 *             for i in range(self.atomids_to_internalids_size):
 *                 print("Atomid #%i -> internal id #%i" % (i+1, self.atomids_to_internalids[i]))
 */
  __pyx_t_1 = ((__pyx_v_internal_id < 0) != 0);
  if (__pyx_t_1) {

    /* "fatslimlib/core_base.pyx":1138
 * 
 *         if internal_id < 0:
 *             for i in range(self.atomids_to_internalids_size):             # <<<<<<<<<<<<<<
 *                 print("Atomid #%i -> internal id #%i" % (i+1, self.atomids_to_internalids[i]))
 *             raise KeyError("atomid not valid: %i (max atomid:%i)" %
 */
    __pyx_t_2 = __Pyx_PyInt_From_fsl_int(__pyx_v_self->atomids_to_internalids_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1138, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1138, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1138, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1138, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1138, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_3); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 1138, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1138, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_5(__pyx_t_2);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1138, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "fatslimlib/core_base.pyx":1139
 *         if internal_id < 0:
 *             for i in range(self.atomids_to_internalids_size):
 *                 print("Atomid #%i -> internal id #%i" % (i+1, self.atomids_to_internalids[i]))             # <<<<<<<<<<<<<<
 *             raise KeyError("atomid not valid: %i (max atomid:%i)" %
 *                            (atomid, self.atomids_to_internalids_size))
 */
      __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_v_i, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1139, __pyx_L1_error)
      __pyx_t_7 = __Pyx_PyInt_From_fsl_int((__pyx_v_self->atomids_to_internalids[__pyx_t_6])); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_7);
      __pyx_t_3 = 0;
      __pyx_t_7 = 0;
      __pyx_t_7 = PyUnicode_Format(__pyx_kp_u_Atomid_i_internal_id_i, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1139, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "fatslimlib/core_base.pyx":1138
 * 
 *         if internal_id < 0:
 *             for i in range(self.atomids_to_internalids_size):             # <<<<<<<<<<<<<<
 *                 print("Atomid #%i -> internal id #%i" % (i+1, self.atomids_to_internalids[i]))
 *             raise KeyError("atomid not valid: %i (max atomid:%i)" %
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fatslimlib/core_base.pyx":1141
 *                 print("Atomid #%i -> internal id #%i" % (i+1, self.atomids_to_internalids[i]))
 *             raise KeyError("atomid not valid: %i (max atomid:%i)" %
 *                            (atomid, self.atomids_to_internalids_size))             # <<<<<<<<<<<<<<
 * 
 *         atom = self.atoms[internal_id]
 */
    __pyx_t_2 = __Pyx_PyInt_From_fsl_int(__pyx_v_atomid); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_8 = __Pyx_PyInt_From_fsl_int(__pyx_v_self->atomids_to_internalids_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_8);
    __pyx_t_2 = 0;
    __pyx_t_8 = 0;

    /* "fatslimlib/core_base.pyx":1140
 *             for i in range(self.atomids_to_internalids_size):
 *                 print("Atomid #%i -> internal id #%i" % (i+1, self.atomids_to_internalids[i]))
 *             raise KeyError("atomid not valid: %i (max atomid:%i)" %             # <<<<<<<<<<<<<<
 *                            (atomid, self.atomids_to_internalids_size))
 * 
 */
    __pyx_t_8 = PyUnicode_Format(__pyx_kp_u_atomid_not_valid_i_max_atomid_i, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 1140, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":1137
 *         cdef bytes resname, atomname
 * 
 *         if internal_id < 0:             # <<<<<<<<<<<<<<
 *             for i in range(self.atomids_to_internalids_size):
 *                 print("Atomid #%i -> internal id #%i" % (i+1, self.atomids_to_internalids[i]))
 */
  }

  /* "fatslimlib/core_base.pyx":1143
 *                            (atomid, self.atomids_to_internalids_size))
 * 
 *         atom = self.atoms[internal_id]             # <<<<<<<<<<<<<<
 *         residue = self.residues[atom.residue_internal_id]
 * 
 */
  __pyx_v_atom = (__pyx_v_self->atoms[__pyx_v_internal_id]);

  /* "fatslimlib/core_base.pyx":1144
 * 
 *         atom = self.atoms[internal_id]
 *         residue = self.residues[atom.residue_internal_id]             # <<<<<<<<<<<<<<
 * 
 *         return atom.resid, \
 */
  __pyx_v_residue = (__pyx_v_self->residues[__pyx_v_atom.residue_internal_id]);

  /* "fatslimlib/core_base.pyx":1146
 *         residue = self.residues[atom.residue_internal_id]
 * 
 *         return atom.resid, \             # <<<<<<<<<<<<<<
 *                str(self.resnames[residue.name_id].decode()), \
 *                str(self.atomnames[atom.name_id].decode()), \
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_7 = __Pyx_PyInt_From_fsl_int(__pyx_v_atom.resid); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "fatslimlib/core_base.pyx":1147
 * 
 *         return atom.resid, \
 *                str(self.resnames[residue.name_id].decode()), \             # <<<<<<<<<<<<<<
 *                str(self.atomnames[atom.name_id].decode()), \
 *                atom.atomid
 */
  __pyx_t_9 = (__pyx_v_self->resnames[__pyx_v_residue.name_id]);
  __pyx_t_8 = __Pyx_decode_c_string(__pyx_t_9, 0, strlen(__pyx_t_9), NULL, NULL, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "fatslimlib/core_base.pyx":1148
 *         return atom.resid, \
 *                str(self.resnames[residue.name_id].decode()), \
 *                str(self.atomnames[atom.name_id].decode()), \             # <<<<<<<<<<<<<<
 *                atom.atomid
 * 
 */
  __pyx_t_9 = (__pyx_v_self->atomnames[__pyx_v_atom.name_id]);
  __pyx_t_8 = __Pyx_decode_c_string(__pyx_t_9, 0, strlen(__pyx_t_9), NULL, NULL, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "fatslimlib/core_base.pyx":1149
 *                str(self.resnames[residue.name_id].decode()), \
 *                str(self.atomnames[atom.name_id].decode()), \
 *                atom.atomid             # <<<<<<<<<<<<<<
 * 
 *     cdef fsl_int fast_get_resid_from_atomid(self, fsl_int atomid) nogil:
 */
  __pyx_t_8 = __Pyx_PyInt_From_fsl_int(__pyx_v_atom.atomid); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "fatslimlib/core_base.pyx":1146
 *         residue = self.residues[atom.residue_internal_id]
 * 
 *         return atom.resid, \             # <<<<<<<<<<<<<<
 *                str(self.resnames[residue.name_id].decode()), \
 *                str(self.atomnames[atom.name_id].decode()), \
 */
  __pyx_t_10 = PyTuple_New(4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_t_8);
  __pyx_t_7 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_8 = 0;
  __pyx_r = __pyx_t_10;
  __pyx_t_10 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1131
 *         return &self.atoms[internal_id]
 * 
 *     def get_atom(self, fsl_int atomid):             # <<<<<<<<<<<<<<
 *         cdef topol_atom_t atom
 *         cdef topol_residue_t residue
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("fatslimlib.core_base.Topology.get_atom", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1151
 *                atom.atomid
 * 
 *     cdef fsl_int fast_get_resid_from_atomid(self, fsl_int atomid) nogil:             # <<<<<<<<<<<<<<
 *         cdef topol_atom_t atom
 *         cdef fsl_int internal_id = self.fast_get_atom_internal_id(atomid)
 */

static fsl_int __pyx_f_10fatslimlib_9core_base_8Topology_fast_get_resid_from_atomid(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_atomid) {
  struct __pyx_t_10fatslimlib_9core_base_topol_atom_t __pyx_v_atom;
  fsl_int __pyx_v_internal_id;
  fsl_int __pyx_r;
  int __pyx_t_1;

  /* "fatslimlib/core_base.pyx":1153
 *     cdef fsl_int fast_get_resid_from_atomid(self, fsl_int atomid) nogil:
 *         cdef topol_atom_t atom
 *         cdef fsl_int internal_id = self.fast_get_atom_internal_id(atomid)             # <<<<<<<<<<<<<<
 * 
 *         if internal_id < 0:
 */
  __pyx_v_internal_id = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Topology *)__pyx_v_self->__pyx_vtab)->fast_get_atom_internal_id(__pyx_v_self, __pyx_v_atomid);

  /* "fatslimlib/core_base.pyx":1155
 *         cdef fsl_int internal_id = self.fast_get_atom_internal_id(atomid)
 * 
 *         if internal_id < 0:             # <<<<<<<<<<<<<<
 *             return -1
 * 
 */
  __pyx_t_1 = ((__pyx_v_internal_id < 0) != 0);
  if (__pyx_t_1) {

    /* "fatslimlib/core_base.pyx":1156
 * 
 *         if internal_id < 0:
 *             return -1             # <<<<<<<<<<<<<<
 * 
 *         atom = self.atoms[internal_id]
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "fatslimlib/core_base.pyx":1155
 *         cdef fsl_int internal_id = self.fast_get_atom_internal_id(atomid)
 * 
 *         if internal_id < 0:             # <<<<<<<<<<<<<<
 *             return -1
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":1158
 *             return -1
 * 
 *         atom = self.atoms[internal_id]             # <<<<<<<<<<<<<<
 *         return atom.resid
 * 
 */
  __pyx_v_atom = (__pyx_v_self->atoms[__pyx_v_internal_id]);

  /* "fatslimlib/core_base.pyx":1159
 * 
 *         atom = self.atoms[internal_id]
 *         return atom.resid             # <<<<<<<<<<<<<<
 * 
 *     cdef topol_residue_t *fast_get_residue_from_atomid(self, fsl_int atomid) nogil:
 */
  __pyx_r = __pyx_v_atom.resid;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1151
 *                atom.atomid
 * 
 *     cdef fsl_int fast_get_resid_from_atomid(self, fsl_int atomid) nogil:             # <<<<<<<<<<<<<<
 *         cdef topol_atom_t atom
 *         cdef fsl_int internal_id = self.fast_get_atom_internal_id(atomid)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1161
 *         return atom.resid
 * 
 *     cdef topol_residue_t *fast_get_residue_from_atomid(self, fsl_int atomid) nogil:             # <<<<<<<<<<<<<<
 *         cdef topol_atom_t atom
 *         cdef fsl_int internal_id = self.fast_get_atom_internal_id(atomid)
 */

static struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *__pyx_f_10fatslimlib_9core_base_8Topology_fast_get_residue_from_atomid(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_atomid) {
  struct __pyx_t_10fatslimlib_9core_base_topol_atom_t __pyx_v_atom;
  fsl_int __pyx_v_internal_id;
  struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *__pyx_r;
  int __pyx_t_1;

  /* "fatslimlib/core_base.pyx":1163
 *     cdef topol_residue_t *fast_get_residue_from_atomid(self, fsl_int atomid) nogil:
 *         cdef topol_atom_t atom
 *         cdef fsl_int internal_id = self.fast_get_atom_internal_id(atomid)             # <<<<<<<<<<<<<<
 * 
 *         if internal_id < 0:
 */
  __pyx_v_internal_id = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Topology *)__pyx_v_self->__pyx_vtab)->fast_get_atom_internal_id(__pyx_v_self, __pyx_v_atomid);

  /* "fatslimlib/core_base.pyx":1165
 *         cdef fsl_int internal_id = self.fast_get_atom_internal_id(atomid)
 * 
 *         if internal_id < 0:             # <<<<<<<<<<<<<<
 *             return NULL
 *         atom = self.atoms[internal_id]
 */
  __pyx_t_1 = ((__pyx_v_internal_id < 0) != 0);
  if (__pyx_t_1) {

    /* "fatslimlib/core_base.pyx":1166
 * 
 *         if internal_id < 0:
 *             return NULL             # <<<<<<<<<<<<<<
 *         atom = self.atoms[internal_id]
 * 
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "fatslimlib/core_base.pyx":1165
 *         cdef fsl_int internal_id = self.fast_get_atom_internal_id(atomid)
 * 
 *         if internal_id < 0:             # <<<<<<<<<<<<<<
 *             return NULL
 *         atom = self.atoms[internal_id]
 */
  }

  /* "fatslimlib/core_base.pyx":1167
 *         if internal_id < 0:
 *             return NULL
 *         atom = self.atoms[internal_id]             # <<<<<<<<<<<<<<
 * 
 *         return &self.residues[atom.residue_internal_id]
 */
  __pyx_v_atom = (__pyx_v_self->atoms[__pyx_v_internal_id]);

  /* "fatslimlib/core_base.pyx":1169
 *         atom = self.atoms[internal_id]
 * 
 *         return &self.residues[atom.residue_internal_id]             # <<<<<<<<<<<<<<
 * 
 *     def get_residue_from_atomid(self, fsl_int atomid):
 */
  __pyx_r = (&(__pyx_v_self->residues[__pyx_v_atom.residue_internal_id]));
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1161
 *         return atom.resid
 * 
 *     cdef topol_residue_t *fast_get_residue_from_atomid(self, fsl_int atomid) nogil:             # <<<<<<<<<<<<<<
 *         cdef topol_atom_t atom
 *         cdef fsl_int internal_id = self.fast_get_atom_internal_id(atomid)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1171
 *         return &self.residues[atom.residue_internal_id]
 * 
 *     def get_residue_from_atomid(self, fsl_int atomid):             # <<<<<<<<<<<<<<
 *         cdef topol_residue_t *residue = self.fast_get_residue_from_atomid(atomid)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_8Topology_9get_residue_from_atomid(PyObject *__pyx_v_self, PyObject *__pyx_arg_atomid); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_8Topology_9get_residue_from_atomid(PyObject *__pyx_v_self, PyObject *__pyx_arg_atomid) {
  fsl_int __pyx_v_atomid;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_residue_from_atomid (wrapper)", 0);
  assert(__pyx_arg_atomid); {
    __pyx_v_atomid = __Pyx_PyInt_As_fsl_int(__pyx_arg_atomid); if (unlikely((__pyx_v_atomid == ((fsl_int)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1171, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.Topology.get_residue_from_atomid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_8Topology_8get_residue_from_atomid(((struct __pyx_obj_10fatslimlib_9core_base_Topology *)__pyx_v_self), ((fsl_int)__pyx_v_atomid));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_8Topology_8get_residue_from_atomid(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_atomid) {
  struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *__pyx_v_residue;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("get_residue_from_atomid", 0);

  /* "fatslimlib/core_base.pyx":1172
 * 
 *     def get_residue_from_atomid(self, fsl_int atomid):
 *         cdef topol_residue_t *residue = self.fast_get_residue_from_atomid(atomid)             # <<<<<<<<<<<<<<
 * 
 *         if residue == NULL:
 */
  __pyx_v_residue = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Topology *)__pyx_v_self->__pyx_vtab)->fast_get_residue_from_atomid(__pyx_v_self, __pyx_v_atomid);

  /* "fatslimlib/core_base.pyx":1174
 *         cdef topol_residue_t *residue = self.fast_get_residue_from_atomid(atomid)
 * 
 *         if residue == NULL:             # <<<<<<<<<<<<<<
 *             raise KeyError("atomid not valid: %i" % atomid)
 * 
 */
  __pyx_t_1 = ((__pyx_v_residue == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "fatslimlib/core_base.pyx":1175
 * 
 *         if residue == NULL:
 *             raise KeyError("atomid not valid: %i" % atomid)             # <<<<<<<<<<<<<<
 * 
 *         return topol_residue_to_dict(residue, self.resnames[residue.name_id])
 */
    __pyx_t_2 = __Pyx_PyInt_From_fsl_int(__pyx_v_atomid); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_atomid_not_valid_i, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1175, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":1174
 *         cdef topol_residue_t *residue = self.fast_get_residue_from_atomid(atomid)
 * 
 *         if residue == NULL:             # <<<<<<<<<<<<<<
 *             raise KeyError("atomid not valid: %i" % atomid)
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":1177
 *             raise KeyError("atomid not valid: %i" % atomid)
 * 
 *         return topol_residue_to_dict(residue, self.resnames[residue.name_id])             # <<<<<<<<<<<<<<
 * 
 *     cdef fsl_int fast_get_residue_internal_id(self, fsl_int resid) nogil:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_FromString((__pyx_v_self->resnames[__pyx_v_residue->name_id])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_10fatslimlib_9core_base_topol_residue_to_dict(__pyx_v_residue, ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1171
 *         return &self.residues[atom.residue_internal_id]
 * 
 *     def get_residue_from_atomid(self, fsl_int atomid):             # <<<<<<<<<<<<<<
 *         cdef topol_residue_t *residue = self.fast_get_residue_from_atomid(atomid)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("fatslimlib.core_base.Topology.get_residue_from_atomid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1179
 *         return topol_residue_to_dict(residue, self.resnames[residue.name_id])
 * 
 *     cdef fsl_int fast_get_residue_internal_id(self, fsl_int resid) nogil:             # <<<<<<<<<<<<<<
 *         if not 0 < resid <= self.resids_to_internalids_size:
 *             return -1
 */

static fsl_int __pyx_f_10fatslimlib_9core_base_8Topology_fast_get_residue_internal_id(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_resid) {
  fsl_int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;

  /* "fatslimlib/core_base.pyx":1180
 * 
 *     cdef fsl_int fast_get_residue_internal_id(self, fsl_int resid) nogil:
 *         if not 0 < resid <= self.resids_to_internalids_size:             # <<<<<<<<<<<<<<
 *             return -1
 *         return self.resids_to_internalids[resid-1]
 */
  __pyx_t_1 = (0 < __pyx_v_resid);
  if (__pyx_t_1) {
    __pyx_t_1 = (__pyx_v_resid <= __pyx_v_self->resids_to_internalids_size);
  }
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "fatslimlib/core_base.pyx":1181
 *     cdef fsl_int fast_get_residue_internal_id(self, fsl_int resid) nogil:
 *         if not 0 < resid <= self.resids_to_internalids_size:
 *             return -1             # <<<<<<<<<<<<<<
 *         return self.resids_to_internalids[resid-1]
 * 
 */
    __pyx_r = -1;
    goto __pyx_L0;

    /* "fatslimlib/core_base.pyx":1180
 * 
 *     cdef fsl_int fast_get_residue_internal_id(self, fsl_int resid) nogil:
 *         if not 0 < resid <= self.resids_to_internalids_size:             # <<<<<<<<<<<<<<
 *             return -1
 *         return self.resids_to_internalids[resid-1]
 */
  }

  /* "fatslimlib/core_base.pyx":1182
 *         if not 0 < resid <= self.resids_to_internalids_size:
 *             return -1
 *         return self.resids_to_internalids[resid-1]             # <<<<<<<<<<<<<<
 * 
 *     cdef topol_residue_t *fast_get_residue(self, fsl_int resid) nogil:
 */
  __pyx_r = (__pyx_v_self->resids_to_internalids[(__pyx_v_resid - 1)]);
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1179
 *         return topol_residue_to_dict(residue, self.resnames[residue.name_id])
 * 
 *     cdef fsl_int fast_get_residue_internal_id(self, fsl_int resid) nogil:             # <<<<<<<<<<<<<<
 *         if not 0 < resid <= self.resids_to_internalids_size:
 *             return -1
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1184
 *         return self.resids_to_internalids[resid-1]
 * 
 *     cdef topol_residue_t *fast_get_residue(self, fsl_int resid) nogil:             # <<<<<<<<<<<<<<
 *         cdef fsl_int internal_id = self.fast_get_residue_internal_id(resid)
 * 
 */

static struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *__pyx_f_10fatslimlib_9core_base_8Topology_fast_get_residue(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, fsl_int __pyx_v_resid) {
  fsl_int __pyx_v_internal_id;
  struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *__pyx_r;
  int __pyx_t_1;

  /* "fatslimlib/core_base.pyx":1185
 * 
 *     cdef topol_residue_t *fast_get_residue(self, fsl_int resid) nogil:
 *         cdef fsl_int internal_id = self.fast_get_residue_internal_id(resid)             # <<<<<<<<<<<<<<
 * 
 *         if internal_id < 0:
 */
  __pyx_v_internal_id = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Topology *)__pyx_v_self->__pyx_vtab)->fast_get_residue_internal_id(__pyx_v_self, __pyx_v_resid);

  /* "fatslimlib/core_base.pyx":1187
 *         cdef fsl_int internal_id = self.fast_get_residue_internal_id(resid)
 * 
 *         if internal_id < 0:             # <<<<<<<<<<<<<<
 *             return NULL
 * 
 */
  __pyx_t_1 = ((__pyx_v_internal_id < 0) != 0);
  if (__pyx_t_1) {

    /* "fatslimlib/core_base.pyx":1188
 * 
 *         if internal_id < 0:
 *             return NULL             # <<<<<<<<<<<<<<
 * 
 *         return &self.residues[internal_id]
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "fatslimlib/core_base.pyx":1187
 *         cdef fsl_int internal_id = self.fast_get_residue_internal_id(resid)
 * 
 *         if internal_id < 0:             # <<<<<<<<<<<<<<
 *             return NULL
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":1190
 *             return NULL
 * 
 *         return &self.residues[internal_id]             # <<<<<<<<<<<<<<
 * 
 *     def get_residue(self, resid):
 */
  __pyx_r = (&(__pyx_v_self->residues[__pyx_v_internal_id]));
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1184
 *         return self.resids_to_internalids[resid-1]
 * 
 *     cdef topol_residue_t *fast_get_residue(self, fsl_int resid) nogil:             # <<<<<<<<<<<<<<
 *         cdef fsl_int internal_id = self.fast_get_residue_internal_id(resid)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1192
 *         return &self.residues[internal_id]
 * 
 *     def get_residue(self, resid):             # <<<<<<<<<<<<<<
 *         cdef topol_residue_t *residue = self.fast_get_residue(resid)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_8Topology_11get_residue(PyObject *__pyx_v_self, PyObject *__pyx_v_resid); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_8Topology_11get_residue(PyObject *__pyx_v_self, PyObject *__pyx_v_resid) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_residue (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_8Topology_10get_residue(((struct __pyx_obj_10fatslimlib_9core_base_Topology *)__pyx_v_self), ((PyObject *)__pyx_v_resid));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_8Topology_10get_residue(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, PyObject *__pyx_v_resid) {
  struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *__pyx_v_residue;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  fsl_int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("get_residue", 0);

  /* "fatslimlib/core_base.pyx":1193
 * 
 *     def get_residue(self, resid):
 *         cdef topol_residue_t *residue = self.fast_get_residue(resid)             # <<<<<<<<<<<<<<
 * 
 *         if residue == NULL:
 */
  __pyx_t_1 = __Pyx_PyInt_As_fsl_int(__pyx_v_resid); if (unlikely((__pyx_t_1 == ((fsl_int)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1193, __pyx_L1_error)
  __pyx_v_residue = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Topology *)__pyx_v_self->__pyx_vtab)->fast_get_residue(__pyx_v_self, __pyx_t_1);

  /* "fatslimlib/core_base.pyx":1195
 *         cdef topol_residue_t *residue = self.fast_get_residue(resid)
 * 
 *         if residue == NULL:             # <<<<<<<<<<<<<<
 *             raise KeyError("resid not valid: %i" % resid)
 * 
 */
  __pyx_t_2 = ((__pyx_v_residue == NULL) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "fatslimlib/core_base.pyx":1196
 * 
 *         if residue == NULL:
 *             raise KeyError("resid not valid: %i" % resid)             # <<<<<<<<<<<<<<
 * 
 *         return topol_residue_to_dict(residue, self.resnames[residue.name_id])
 */
    __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_resid_not_valid_i, __pyx_v_resid); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 1196, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":1195
 *         cdef topol_residue_t *residue = self.fast_get_residue(resid)
 * 
 *         if residue == NULL:             # <<<<<<<<<<<<<<
 *             raise KeyError("resid not valid: %i" % resid)
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":1198
 *             raise KeyError("resid not valid: %i" % resid)
 * 
 *         return topol_residue_to_dict(residue, self.resnames[residue.name_id])             # <<<<<<<<<<<<<<
 * 
 *     property nresidues:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_FromString((__pyx_v_self->resnames[__pyx_v_residue->name_id])); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __pyx_f_10fatslimlib_9core_base_topol_residue_to_dict(__pyx_v_residue, ((PyObject*)__pyx_t_4)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1192
 *         return &self.residues[internal_id]
 * 
 *     def get_residue(self, resid):             # <<<<<<<<<<<<<<
 *         cdef topol_residue_t *residue = self.fast_get_residue(resid)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("fatslimlib.core_base.Topology.get_residue", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1201
 * 
 *     property nresidues:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.residues_size
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_8Topology_9nresidues_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_8Topology_9nresidues_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_8Topology_9nresidues___get__(((struct __pyx_obj_10fatslimlib_9core_base_Topology *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_8Topology_9nresidues___get__(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":1202
 *     property nresidues:
 *         def __get__(self):
 *             return self.residues_size             # <<<<<<<<<<<<<<
 * 
 *     property natoms:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_v_self->residues_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1201
 * 
 *     property nresidues:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.residues_size
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Topology.nresidues.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1205
 * 
 *     property natoms:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.atoms_size
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_8Topology_6natoms_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_8Topology_6natoms_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_8Topology_6natoms___get__(((struct __pyx_obj_10fatslimlib_9core_base_Topology *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_8Topology_6natoms___get__(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":1206
 *     property natoms:
 *         def __get__(self):
 *             return self.atoms_size             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_v_self->atoms_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1205
 * 
 *     property natoms:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.atoms_size
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Topology.natoms.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1208
 *             return self.atoms_size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.atoms_size
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_10fatslimlib_9core_base_8Topology_13__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_10fatslimlib_9core_base_8Topology_13__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_8Topology_12__len__(((struct __pyx_obj_10fatslimlib_9core_base_Topology *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_10fatslimlib_9core_base_8Topology_12__len__(struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "fatslimlib/core_base.pyx":1209
 * 
 *     def __len__(self):
 *         return self.atoms_size             # <<<<<<<<<<<<<<
 * 
 * cdef class TopologyReader(object):
 */
  __pyx_r = __pyx_v_self->atoms_size;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1208
 *             return self.atoms_size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.atoms_size
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.atomids_to_internalids,self.atomnames,self.atoms,self.resids_to_internalids,self.residues,self.resnames cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_8Topology_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_8Topology_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_8Topology_14__reduce_cython__(((struct __pyx_obj_10fatslimlib_9core_base_Topology *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_8Topology_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.atomids_to_internalids,self.atomnames,self.atoms,self.resids_to_internalids,self.residues,self.resnames cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.atomids_to_internalids,self.atomnames,self.atoms,self.resids_to_internalids,self.residues,self.resnames cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.atomids_to_internalids,self.atomnames,self.atoms,self.resids_to_internalids,self.residues,self.resnames cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Topology.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.atomids_to_internalids,self.atomnames,self.atoms,self.resids_to_internalids,self.residues,self.resnames cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.atomids_to_internalids,self.atomnames,self.atoms,self.resids_to_internalids,self.residues,self.resnames cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_8Topology_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_8Topology_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_8Topology_16__setstate_cython__(((struct __pyx_obj_10fatslimlib_9core_base_Topology *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_8Topology_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.atomids_to_internalids,self.atomnames,self.atoms,self.resids_to_internalids,self.residues,self.resnames cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.atomids_to_internalids,self.atomnames,self.atoms,self.resids_to_internalids,self.residues,self.resnames cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.atomids_to_internalids,self.atomnames,self.atoms,self.resids_to_internalids,self.residues,self.resnames cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.atomids_to_internalids,self.atomnames,self.atoms,self.resids_to_internalids,self.residues,self.resnames cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Topology.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1213
 * cdef class TopologyReader(object):
 * 
 *     def __init__(self, str filename, bint verbose=True):             # <<<<<<<<<<<<<<
 *         self.filename = filename.encode()
 *         self.topology = Topology()
 */

/* Python wrapper */
static int __pyx_pw_10fatslimlib_9core_base_14TopologyReader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10fatslimlib_9core_base_14TopologyReader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_filename = 0;
  int __pyx_v_verbose;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_filename,&__pyx_n_s_verbose,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_verbose);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1213, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_filename = ((PyObject*)values[0]);
    if (values[1]) {
      __pyx_v_verbose = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_verbose == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1213, __pyx_L3_error)
    } else {
      __pyx_v_verbose = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1213, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.TopologyReader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_filename), (&PyUnicode_Type), 1, "filename", 1))) __PYX_ERR(0, 1213, __pyx_L1_error)
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_14TopologyReader___init__(((struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *)__pyx_v_self), __pyx_v_filename, __pyx_v_verbose);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10fatslimlib_9core_base_14TopologyReader___init__(struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *__pyx_v_self, PyObject *__pyx_v_filename, int __pyx_v_verbose) {
  PyObject *__pyx_v_begin = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "fatslimlib/core_base.pyx":1214
 * 
 *     def __init__(self, str filename, bint verbose=True):
 *         self.filename = filename.encode()             # <<<<<<<<<<<<<<
 *         self.topology = Topology()
 * 
 */
  if (unlikely(__pyx_v_filename == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 1214, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_filename, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->filename);
  __Pyx_DECREF(__pyx_v_self->filename);
  __pyx_v_self->filename = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":1215
 *     def __init__(self, str filename, bint verbose=True):
 *         self.filename = filename.encode()
 *         self.topology = Topology()             # <<<<<<<<<<<<<<
 * 
 *         # load topology file
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_10fatslimlib_9core_base_Topology)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->topology);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->topology));
  __pyx_v_self->topology = ((struct __pyx_obj_10fatslimlib_9core_base_Topology *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":1218
 * 
 *         # load topology file
 *         verbose_print("Loading topology from '%s'... " %             # <<<<<<<<<<<<<<
 *                       os.path.basename(self.filename.decode()),
 *                       verbose,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_verbose_print); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "fatslimlib/core_base.pyx":1219
 *         # load topology file
 *         verbose_print("Loading topology from '%s'... " %
 *                       os.path.basename(self.filename.decode()),             # <<<<<<<<<<<<<<
 *                       verbose,
 *                       end="")
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_path); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_basename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(__pyx_v_self->filename == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "decode");
    __PYX_ERR(0, 1219, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_decode_bytes(__pyx_v_self->filename, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":1218
 * 
 *         # load topology file
 *         verbose_print("Loading topology from '%s'... " %             # <<<<<<<<<<<<<<
 *                       os.path.basename(self.filename.decode()),
 *                       verbose,
 */
  __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Loading_topology_from_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":1220
 *         verbose_print("Loading topology from '%s'... " %
 *                       os.path.basename(self.filename.decode()),
 *                       verbose,             # <<<<<<<<<<<<<<
 *                       end="")
 *         begin = time()
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_verbose); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "fatslimlib/core_base.pyx":1218
 * 
 *         # load topology file
 *         verbose_print("Loading topology from '%s'... " %             # <<<<<<<<<<<<<<
 *                       os.path.basename(self.filename.decode()),
 *                       verbose,
 */
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":1221
 *                       os.path.basename(self.filename.decode()),
 *                       verbose,
 *                       end="")             # <<<<<<<<<<<<<<
 *         begin = time()
 *         self.load()
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_end, __pyx_kp_u__22) < 0) __PYX_ERR(0, 1221, __pyx_L1_error)

  /* "fatslimlib/core_base.pyx":1218
 * 
 *         # load topology file
 *         verbose_print("Loading topology from '%s'... " %             # <<<<<<<<<<<<<<
 *                       os.path.basename(self.filename.decode()),
 *                       verbose,
 */
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":1222
 *                       verbose,
 *                       end="")
 *         begin = time()             # <<<<<<<<<<<<<<
 *         self.load()
 *         verbose_print("%i topology atoms (from %i residues) loaded in %s" %
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_begin = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":1223
 *                       end="")
 *         begin = time()
 *         self.load()             # <<<<<<<<<<<<<<
 *         verbose_print("%i topology atoms (from %i residues) loaded in %s" %
 *                       (self.topology.natoms,
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_TopologyReader *)__pyx_v_self->__pyx_vtab)->load(__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":1224
 *         begin = time()
 *         self.load()
 *         verbose_print("%i topology atoms (from %i residues) loaded in %s" %             # <<<<<<<<<<<<<<
 *                       (self.topology.natoms,
 *                        self.topology.nresidues,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_verbose_print); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "fatslimlib/core_base.pyx":1225
 *         self.load()
 *         verbose_print("%i topology atoms (from %i residues) loaded in %s" %
 *                       (self.topology.natoms,             # <<<<<<<<<<<<<<
 *                        self.topology.nresidues,
 *                        pretty_delta(begin, time())),
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->topology), __pyx_n_s_natoms); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "fatslimlib/core_base.pyx":1226
 *         verbose_print("%i topology atoms (from %i residues) loaded in %s" %
 *                       (self.topology.natoms,
 *                        self.topology.nresidues,             # <<<<<<<<<<<<<<
 *                        pretty_delta(begin, time())),
 *                       verbose)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->topology), __pyx_n_s_nresidues); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "fatslimlib/core_base.pyx":1227
 *                       (self.topology.natoms,
 *                        self.topology.nresidues,
 *                        pretty_delta(begin, time())),             # <<<<<<<<<<<<<<
 *                       verbose)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_time); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __pyx_f_10fatslimlib_9core_base_pretty_delta(__pyx_v_begin, __pyx_t_5, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fatslimlib/core_base.pyx":1225
 *         self.load()
 *         verbose_print("%i topology atoms (from %i residues) loaded in %s" %
 *                       (self.topology.natoms,             # <<<<<<<<<<<<<<
 *                        self.topology.nresidues,
 *                        pretty_delta(begin, time())),
 */
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_6);
  __pyx_t_4 = 0;
  __pyx_t_1 = 0;
  __pyx_t_6 = 0;

  /* "fatslimlib/core_base.pyx":1224
 *         begin = time()
 *         self.load()
 *         verbose_print("%i topology atoms (from %i residues) loaded in %s" %             # <<<<<<<<<<<<<<
 *                       (self.topology.natoms,
 *                        self.topology.nresidues,
 */
  __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_i_topology_atoms_from_i_residue, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fatslimlib/core_base.pyx":1228
 *                        self.topology.nresidues,
 *                        pretty_delta(begin, time())),
 *                       verbose)             # <<<<<<<<<<<<<<
 * 
 *     cdef load(self):
 */
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_verbose); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_1 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_t_6, __pyx_t_5};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1224, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_t_6, __pyx_t_5};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1224, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  } else
  #endif
  {
    __pyx_t_4 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__pyx_t_1) {
      __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_8, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_8, __pyx_t_5);
    __pyx_t_6 = 0;
    __pyx_t_5 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1224, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":1213
 * cdef class TopologyReader(object):
 * 
 *     def __init__(self, str filename, bint verbose=True):             # <<<<<<<<<<<<<<
 *         self.filename = filename.encode()
 *         self.topology = Topology()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("fatslimlib.core_base.TopologyReader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_begin);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1230
 *                       verbose)
 * 
 *     cdef load(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

static PyObject *__pyx_f_10fatslimlib_9core_base_14TopologyReader_load(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("load", 0);

  /* "fatslimlib/core_base.pyx":1231
 * 
 *     cdef load(self):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(0, 1231, __pyx_L1_error)

  /* "fatslimlib/core_base.pyx":1230
 *                       verbose)
 * 
 *     cdef load(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.TopologyReader.load", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pxd":183
 *     # Attributes
 *     cdef bytes filename
 *     cdef readonly Topology topology             # <<<<<<<<<<<<<<
 * 
 *     # Methods
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_14TopologyReader_8topology_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_14TopologyReader_8topology_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_14TopologyReader_8topology___get__(((struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_14TopologyReader_8topology___get__(struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->topology));
  __pyx_r = ((PyObject *)__pyx_v_self->topology);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_14TopologyReader_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_14TopologyReader_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_14TopologyReader_2__reduce_cython__(((struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_14TopologyReader_2__reduce_cython__(struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.filename, self.topology)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->filename);
  __Pyx_GIVEREF(__pyx_v_self->filename);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->filename);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->topology));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->topology));
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->topology));
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.filename, self.topology)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.filename, self.topology)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.filename is not None or self.topology is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.filename, self.topology)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.filename is not None or self.topology is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_TopologyReader, (type(self), 0x8dafc6b, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->filename != ((PyObject*)Py_None));
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->topology) != Py_None);
    __pyx_t_2 = (__pyx_t_5 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.filename is not None or self.topology is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TopologyReader, (type(self), 0x8dafc6b, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.filename is not None or self.topology is not None
 *     if use_setstate:
 *         return __pyx_unpickle_TopologyReader, (type(self), 0x8dafc6b, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_TopologyReader, (type(self), 0x8dafc6b, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_TopologyReader); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_148569195);
    __Pyx_GIVEREF(__pyx_int_148569195);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_148569195);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.filename is not None or self.topology is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_TopologyReader, (type(self), 0x8dafc6b, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_TopologyReader, (type(self), 0x8dafc6b, None), state
 *     else:
 *         return __pyx_unpickle_TopologyReader, (type(self), 0x8dafc6b, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TopologyReader__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_TopologyReader); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_148569195);
    __Pyx_GIVEREF(__pyx_int_148569195);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_148569195);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("fatslimlib.core_base.TopologyReader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TopologyReader, (type(self), 0x8dafc6b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TopologyReader__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_14TopologyReader_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_14TopologyReader_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_14TopologyReader_4__setstate_cython__(((struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_14TopologyReader_4__setstate_cython__(struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_TopologyReader, (type(self), 0x8dafc6b, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_TopologyReader__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10fatslimlib_9core_base___pyx_unpickle_TopologyReader__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_TopologyReader, (type(self), 0x8dafc6b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_TopologyReader__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.TopologyReader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1237
 * cdef class CoordinateReader(object):
 * 
 *     def __init__(self, str filename, bint verbose=True):             # <<<<<<<<<<<<<<
 *         self.filename = filename.encode()
 * 
 */

/* Python wrapper */
static int __pyx_pw_10fatslimlib_9core_base_16CoordinateReader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10fatslimlib_9core_base_16CoordinateReader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_filename = 0;
  int __pyx_v_verbose;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_filename,&__pyx_n_s_verbose,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_verbose);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1237, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_filename = ((PyObject*)values[0]);
    if (values[1]) {
      __pyx_v_verbose = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_verbose == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1237, __pyx_L3_error)
    } else {
      __pyx_v_verbose = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1237, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.CoordinateReader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_filename), (&PyUnicode_Type), 1, "filename", 1))) __PYX_ERR(0, 1237, __pyx_L1_error)
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_16CoordinateReader___init__(((struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *)__pyx_v_self), __pyx_v_filename, __pyx_v_verbose);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10fatslimlib_9core_base_16CoordinateReader___init__(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_self, PyObject *__pyx_v_filename, int __pyx_v_verbose) {
  PyObject *__pyx_v_begin = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "fatslimlib/core_base.pyx":1238
 * 
 *     def __init__(self, str filename, bint verbose=True):
 *         self.filename = filename.encode()             # <<<<<<<<<<<<<<
 * 
 *         # Preload file
 */
  if (unlikely(__pyx_v_filename == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 1238, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_filename, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->filename);
  __Pyx_DECREF(__pyx_v_self->filename);
  __pyx_v_self->filename = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":1241
 * 
 *         # Preload file
 *         verbose_print("Preloading coordinates from '%s'... " %             # <<<<<<<<<<<<<<
 *                       os.path.basename(self.filename.decode()),
 *                       verbose,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_verbose_print); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "fatslimlib/core_base.pyx":1242
 *         # Preload file
 *         verbose_print("Preloading coordinates from '%s'... " %
 *                       os.path.basename(self.filename.decode()),             # <<<<<<<<<<<<<<
 *                       verbose,
 *                       end="")
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_path); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_basename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(__pyx_v_self->filename == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "decode");
    __PYX_ERR(0, 1242, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_decode_bytes(__pyx_v_self->filename, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":1241
 * 
 *         # Preload file
 *         verbose_print("Preloading coordinates from '%s'... " %             # <<<<<<<<<<<<<<
 *                       os.path.basename(self.filename.decode()),
 *                       verbose,
 */
  __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Preloading_coordinates_from_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":1243
 *         verbose_print("Preloading coordinates from '%s'... " %
 *                       os.path.basename(self.filename.decode()),
 *                       verbose,             # <<<<<<<<<<<<<<
 *                       end="")
 *         begin = time()
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_verbose); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "fatslimlib/core_base.pyx":1241
 * 
 *         # Preload file
 *         verbose_print("Preloading coordinates from '%s'... " %             # <<<<<<<<<<<<<<
 *                       os.path.basename(self.filename.decode()),
 *                       verbose,
 */
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":1244
 *                       os.path.basename(self.filename.decode()),
 *                       verbose,
 *                       end="")             # <<<<<<<<<<<<<<
 *         begin = time()
 *         self.preload()
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_end, __pyx_kp_u__22) < 0) __PYX_ERR(0, 1244, __pyx_L1_error)

  /* "fatslimlib/core_base.pyx":1241
 * 
 *         # Preload file
 *         verbose_print("Preloading coordinates from '%s'... " %             # <<<<<<<<<<<<<<
 *                       os.path.basename(self.filename.decode()),
 *                       verbose,
 */
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":1245
 *                       verbose,
 *                       end="")
 *         begin = time()             # <<<<<<<<<<<<<<
 *         self.preload()
 *         verbose_print("%i frames preloaded in %s" % (self.nframes, pretty_delta(begin, time())),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_begin = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":1246
 *                       end="")
 *         begin = time()
 *         self.preload()             # <<<<<<<<<<<<<<
 *         verbose_print("%i frames preloaded in %s" % (self.nframes, pretty_delta(begin, time())),
 *                       verbose)
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_CoordinateReader *)__pyx_v_self->__pyx_vtab)->preload(__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":1247
 *         begin = time()
 *         self.preload()
 *         verbose_print("%i frames preloaded in %s" % (self.nframes, pretty_delta(begin, time())),             # <<<<<<<<<<<<<<
 *                       verbose)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_verbose_print); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyInt_From_fsl_int(__pyx_v_self->nframes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_time); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_f_10fatslimlib_9core_base_pretty_delta(__pyx_v_begin, __pyx_t_1, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_5 = PyUnicode_Format(__pyx_kp_u_i_frames_preloaded_in_s, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":1248
 *         self.preload()
 *         verbose_print("%i frames preloaded in %s" % (self.nframes, pretty_delta(begin, time())),
 *                       verbose)             # <<<<<<<<<<<<<<
 * 
 *     cdef preload(self):
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_verbose); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_5, __pyx_t_1};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1247, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_5, __pyx_t_1};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1247, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_7, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_7, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":1237
 * cdef class CoordinateReader(object):
 * 
 *     def __init__(self, str filename, bint verbose=True):             # <<<<<<<<<<<<<<
 *         self.filename = filename.encode()
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("fatslimlib.core_base.CoordinateReader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_begin);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1250
 *                       verbose)
 * 
 *     cdef preload(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

static PyObject *__pyx_f_10fatslimlib_9core_base_16CoordinateReader_preload(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("preload", 0);

  /* "fatslimlib/core_base.pyx":1251
 * 
 *     cdef preload(self):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     cdef PBCBox load_box(self, fsl_int frame_id):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(0, 1251, __pyx_L1_error)

  /* "fatslimlib/core_base.pyx":1250
 *                       verbose)
 * 
 *     cdef preload(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.CoordinateReader.preload", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1253
 *         raise NotImplementedError
 * 
 *     cdef PBCBox load_box(self, fsl_int frame_id):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

static struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_f_10fatslimlib_9core_base_16CoordinateReader_load_box(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_self, CYTHON_UNUSED fsl_int __pyx_v_frame_id) {
  struct __pyx_obj_10fatslimlib_9core_base_PBCBox *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("load_box", 0);

  /* "fatslimlib/core_base.pyx":1254
 * 
 *     cdef PBCBox load_box(self, fsl_int frame_id):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     cdef assert_frame_id(self, fsl_int frame_id):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(0, 1254, __pyx_L1_error)

  /* "fatslimlib/core_base.pyx":1253
 *         raise NotImplementedError
 * 
 *     cdef PBCBox load_box(self, fsl_int frame_id):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.CoordinateReader.load_box", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1256
 *         raise NotImplementedError
 * 
 *     cdef assert_frame_id(self, fsl_int frame_id):             # <<<<<<<<<<<<<<
 *         if not 0 <= frame_id < self.nframes:
 *             raise IndexError("Frame index error: %i (%i frames)" % (frame_id, self.nframes))
 */

static PyObject *__pyx_f_10fatslimlib_9core_base_16CoordinateReader_assert_frame_id(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_self, fsl_int __pyx_v_frame_id) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("assert_frame_id", 0);

  /* "fatslimlib/core_base.pyx":1257
 * 
 *     cdef assert_frame_id(self, fsl_int frame_id):
 *         if not 0 <= frame_id < self.nframes:             # <<<<<<<<<<<<<<
 *             raise IndexError("Frame index error: %i (%i frames)" % (frame_id, self.nframes))
 * 
 */
  __pyx_t_1 = (0 <= __pyx_v_frame_id);
  if (__pyx_t_1) {
    __pyx_t_1 = (__pyx_v_frame_id < __pyx_v_self->nframes);
  }
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "fatslimlib/core_base.pyx":1258
 *     cdef assert_frame_id(self, fsl_int frame_id):
 *         if not 0 <= frame_id < self.nframes:
 *             raise IndexError("Frame index error: %i (%i frames)" % (frame_id, self.nframes))             # <<<<<<<<<<<<<<
 * 
 *     cdef real[:,::1] load_coords(self, fsl_int frame_id, fsl_int[:] atomids) nogil except *:
 */
    __pyx_t_3 = __Pyx_PyInt_From_fsl_int(__pyx_v_frame_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyInt_From_fsl_int(__pyx_v_self->nframes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = PyUnicode_Format(__pyx_kp_u_Frame_index_error_i_i_frames, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 1258, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":1257
 * 
 *     cdef assert_frame_id(self, fsl_int frame_id):
 *         if not 0 <= frame_id < self.nframes:             # <<<<<<<<<<<<<<
 *             raise IndexError("Frame index error: %i (%i frames)" % (frame_id, self.nframes))
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":1256
 *         raise NotImplementedError
 * 
 *     cdef assert_frame_id(self, fsl_int frame_id):             # <<<<<<<<<<<<<<
 *         if not 0 <= frame_id < self.nframes:
 *             raise IndexError("Frame index error: %i (%i frames)" % (frame_id, self.nframes))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.CoordinateReader.assert_frame_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1260
 *             raise IndexError("Frame index error: %i (%i frames)" % (frame_id, self.nframes))
 * 
 *     cdef real[:,::1] load_coords(self, fsl_int frame_id, fsl_int[:] atomids) nogil except *:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise NotImplementedError
 */

static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_16CoordinateReader_load_coords(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_self, CYTHON_UNUSED fsl_int __pyx_v_frame_id, CYTHON_UNUSED __Pyx_memviewslice __pyx_v_atomids) {
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("load_coords", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "fatslimlib/core_base.pyx":1261
 * 
 *     cdef real[:,::1] load_coords(self, fsl_int frame_id, fsl_int[:] atomids) nogil except *:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError
 * 
 */
  /*try:*/ {
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "fatslimlib/core_base.pyx":1262
 *     cdef real[:,::1] load_coords(self, fsl_int frame_id, fsl_int[:] atomids) nogil except *:
 *         with gil:
 *             raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
          __PYX_ERR(0, 1262, __pyx_L7_error)
        }

        /* "fatslimlib/core_base.pyx":1261
 * 
 *     cdef real[:,::1] load_coords(self, fsl_int frame_id, fsl_int[:] atomids) nogil except *:
 *         with gil:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError
 * 
 */
        /*finally:*/ {
          __pyx_L7_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
        }
    }
  }
  /*finally:*/ {
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "fatslimlib/core_base.pyx":1260
 *             raise IndexError("Frame index error: %i (%i frames)" % (frame_id, self.nframes))
 * 
 *     cdef real[:,::1] load_coords(self, fsl_int frame_id, fsl_int[:] atomids) nogil except *:             # <<<<<<<<<<<<<<
 *         with gil:
 *             raise NotImplementedError
 */

  /* function exit code */
  __Pyx_pretend_to_initialize(&__pyx_r);
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("fatslimlib.core_base.CoordinateReader.load_coords", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "fatslimlib/core_base.pxd":192
 *     # Attributes
 *     cdef bytes filename
 *     cdef readonly fsl_int nframes             # <<<<<<<<<<<<<<
 *     cdef fsl_int[:] coordinate_offsets
 *     cdef fsl_int[:] box_offsets
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_16CoordinateReader_7nframes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_16CoordinateReader_7nframes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_16CoordinateReader_7nframes___get__(((struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_16CoordinateReader_7nframes___get__(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_v_self->nframes); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.CoordinateReader.nframes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pxd":195
 *     cdef fsl_int[:] coordinate_offsets
 *     cdef fsl_int[:] box_offsets
 *     cdef readonly real[:] timesteps             # <<<<<<<<<<<<<<
 * 
 *     # Methods
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_16CoordinateReader_9timesteps_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_16CoordinateReader_9timesteps_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_16CoordinateReader_9timesteps___get__(((struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_16CoordinateReader_9timesteps___get__(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_self->timesteps.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(2, 195, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->timesteps, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.CoordinateReader.timesteps.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_16CoordinateReader_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_16CoordinateReader_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_16CoordinateReader_2__reduce_cython__(((struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_16CoordinateReader_2__reduce_cython__(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.box_offsets, self.coordinate_offsets, self.filename, self.nframes, self.timesteps)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  if (unlikely(!__pyx_v_self->box_offsets.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 5, __pyx_L1_error)}
  __pyx_t_1 = __pyx_memoryview_fromslice(__pyx_v_self->box_offsets, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_fsl_int, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_fsl_int, 0);; if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(!__pyx_v_self->coordinate_offsets.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 5, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->coordinate_offsets, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_fsl_int, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_fsl_int, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_fsl_int(__pyx_v_self->nframes); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->timesteps.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->timesteps, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(5); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->filename);
  __Pyx_GIVEREF(__pyx_v_self->filename);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_self->filename);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.box_offsets, self.coordinate_offsets, self.filename, self.nframes, self.timesteps)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_5 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v__dict = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "(tree fragment)":7
 *     state = (self.box_offsets, self.coordinate_offsets, self.filename, self.nframes, self.timesteps)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_6 = (__pyx_v__dict != Py_None);
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (__pyx_t_7) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.filename is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.box_offsets, self.coordinate_offsets, self.filename, self.nframes, self.timesteps)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.filename is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_CoordinateReader, (type(self), 0x500400c, None), state
 */
  /*else*/ {
    __pyx_t_7 = (__pyx_v_self->filename != ((PyObject*)Py_None));
    __pyx_v_use_setstate = __pyx_t_7;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.filename is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CoordinateReader, (type(self), 0x500400c, None), state
 *     else:
 */
  __pyx_t_7 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_7) {

    /* "(tree fragment)":13
 *         use_setstate = self.filename is not None
 *     if use_setstate:
 *         return __pyx_unpickle_CoordinateReader, (type(self), 0x500400c, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_CoordinateReader, (type(self), 0x500400c, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_CoordinateReader); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_83902476);
    __Pyx_GIVEREF(__pyx_int_83902476);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_83902476);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_5, 2, Py_None);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.filename is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_CoordinateReader, (type(self), 0x500400c, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_CoordinateReader, (type(self), 0x500400c, None), state
 *     else:
 *         return __pyx_unpickle_CoordinateReader, (type(self), 0x500400c, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CoordinateReader__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_CoordinateReader); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_83902476);
    __Pyx_GIVEREF(__pyx_int_83902476);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_83902476);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.CoordinateReader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CoordinateReader, (type(self), 0x500400c, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CoordinateReader__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_16CoordinateReader_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_16CoordinateReader_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_16CoordinateReader_4__setstate_cython__(((struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_16CoordinateReader_4__setstate_cython__(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_CoordinateReader, (type(self), 0x500400c, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_CoordinateReader__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10fatslimlib_9core_base___pyx_unpickle_CoordinateReader__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_CoordinateReader, (type(self), 0x500400c, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_CoordinateReader__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.CoordinateReader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1267
 * cdef class IndexReader(object):
 * 
 *     def __init__(self, str filename, bint verbose=True):             # <<<<<<<<<<<<<<
 *         self.filename = filename.encode()
 *         self.loaded = False
 */

/* Python wrapper */
static int __pyx_pw_10fatslimlib_9core_base_11IndexReader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10fatslimlib_9core_base_11IndexReader_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_filename = 0;
  int __pyx_v_verbose;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_filename,&__pyx_n_s_verbose,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_verbose);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1267, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_filename = ((PyObject*)values[0]);
    if (values[1]) {
      __pyx_v_verbose = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_verbose == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1267, __pyx_L3_error)
    } else {
      __pyx_v_verbose = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1267, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.IndexReader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_filename), (&PyUnicode_Type), 1, "filename", 1))) __PYX_ERR(0, 1267, __pyx_L1_error)
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_11IndexReader___init__(((struct __pyx_obj_10fatslimlib_9core_base_IndexReader *)__pyx_v_self), __pyx_v_filename, __pyx_v_verbose);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10fatslimlib_9core_base_11IndexReader___init__(struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_self, PyObject *__pyx_v_filename, int __pyx_v_verbose) {
  PyObject *__pyx_v_begin = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "fatslimlib/core_base.pyx":1268
 * 
 *     def __init__(self, str filename, bint verbose=True):
 *         self.filename = filename.encode()             # <<<<<<<<<<<<<<
 *         self.loaded = False
 * 
 */
  if (unlikely(__pyx_v_filename == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 1268, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_filename, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->filename);
  __Pyx_DECREF(__pyx_v_self->filename);
  __pyx_v_self->filename = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":1269
 *     def __init__(self, str filename, bint verbose=True):
 *         self.filename = filename.encode()
 *         self.loaded = False             # <<<<<<<<<<<<<<
 * 
 *         self.groups = {}
 */
  __pyx_v_self->loaded = 0;

  /* "fatslimlib/core_base.pyx":1271
 *         self.loaded = False
 * 
 *         self.groups = {}             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->groups);
  __Pyx_DECREF(__pyx_v_self->groups);
  __pyx_v_self->groups = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":1274
 * 
 * 
 *         verbose_print("Loading groups from '%s'... " %             # <<<<<<<<<<<<<<
 *                       os.path.basename(self.filename),
 *                       verbose,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_verbose_print); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "fatslimlib/core_base.pyx":1275
 * 
 *         verbose_print("Loading groups from '%s'... " %
 *                       os.path.basename(self.filename),             # <<<<<<<<<<<<<<
 *                       verbose,
 *                       end="")
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_path); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_basename); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_self->filename) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_self->filename);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":1274
 * 
 * 
 *         verbose_print("Loading groups from '%s'... " %             # <<<<<<<<<<<<<<
 *                       os.path.basename(self.filename),
 *                       verbose,
 */
  __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Loading_groups_from_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":1276
 *         verbose_print("Loading groups from '%s'... " %
 *                       os.path.basename(self.filename),
 *                       verbose,             # <<<<<<<<<<<<<<
 *                       end="")
 *         begin = time()
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_verbose); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "fatslimlib/core_base.pyx":1274
 * 
 * 
 *         verbose_print("Loading groups from '%s'... " %             # <<<<<<<<<<<<<<
 *                       os.path.basename(self.filename),
 *                       verbose,
 */
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":1277
 *                       os.path.basename(self.filename),
 *                       verbose,
 *                       end="")             # <<<<<<<<<<<<<<
 *         begin = time()
 *         self.load()
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_end, __pyx_kp_u__22) < 0) __PYX_ERR(0, 1277, __pyx_L1_error)

  /* "fatslimlib/core_base.pyx":1274
 * 
 * 
 *         verbose_print("Loading groups from '%s'... " %             # <<<<<<<<<<<<<<
 *                       os.path.basename(self.filename),
 *                       verbose,
 */
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":1278
 *                       verbose,
 *                       end="")
 *         begin = time()             # <<<<<<<<<<<<<<
 *         self.load()
 *         self.loaded = True
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1278, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_begin = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":1279
 *                       end="")
 *         begin = time()
 *         self.load()             # <<<<<<<<<<<<<<
 *         self.loaded = True
 *         verbose_print("%i groups loaded in %s" % (len(self), pretty_delta(begin, time())),
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_load); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":1280
 *         begin = time()
 *         self.load()
 *         self.loaded = True             # <<<<<<<<<<<<<<
 *         verbose_print("%i groups loaded in %s" % (len(self), pretty_delta(begin, time())),
 *                       verbose)
 */
  __pyx_v_self->loaded = 1;

  /* "fatslimlib/core_base.pyx":1281
 *         self.load()
 *         self.loaded = True
 *         verbose_print("%i groups loaded in %s" % (len(self), pretty_delta(begin, time())),             # <<<<<<<<<<<<<<
 *                       verbose)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_verbose_print); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1281, __pyx_L1_error)
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_time); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_1 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __pyx_f_10fatslimlib_9core_base_pretty_delta(__pyx_v_begin, __pyx_t_1, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_6);
  __pyx_t_4 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_i_groups_loaded_in_s, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":1282
 *         self.loaded = True
 *         verbose_print("%i groups loaded in %s" % (len(self), pretty_delta(begin, time())),
 *                       verbose)             # <<<<<<<<<<<<<<
 * 
 *     def load(self):
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_verbose); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_6, __pyx_t_1};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1281, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_6, __pyx_t_1};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1281, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_8, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_8, __pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":1267
 * cdef class IndexReader(object):
 * 
 *     def __init__(self, str filename, bint verbose=True):             # <<<<<<<<<<<<<<
 *         self.filename = filename.encode()
 *         self.loaded = False
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("fatslimlib.core_base.IndexReader.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_begin);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1284
 *                       verbose)
 * 
 *     def load(self):             # <<<<<<<<<<<<<<
 *         if self.loaded:
 *             return
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_11IndexReader_3load(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_11IndexReader_3load(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("load (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_11IndexReader_2load(((struct __pyx_obj_10fatslimlib_9core_base_IndexReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_11IndexReader_2load(struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("load", 0);

  /* "fatslimlib/core_base.pyx":1285
 * 
 *     def load(self):
 *         if self.loaded:             # <<<<<<<<<<<<<<
 *             return
 *         self.fast_load()
 */
  __pyx_t_1 = (__pyx_v_self->loaded != 0);
  if (__pyx_t_1) {

    /* "fatslimlib/core_base.pyx":1286
 *     def load(self):
 *         if self.loaded:
 *             return             # <<<<<<<<<<<<<<
 *         self.fast_load()
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "fatslimlib/core_base.pyx":1285
 * 
 *     def load(self):
 *         if self.loaded:             # <<<<<<<<<<<<<<
 *             return
 *         self.fast_load()
 */
  }

  /* "fatslimlib/core_base.pyx":1287
 *         if self.loaded:
 *             return
 *         self.fast_load()             # <<<<<<<<<<<<<<
 * 
 *     cdef fast_load(self):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_IndexReader *)__pyx_v_self->__pyx_vtab)->fast_load(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":1284
 *                       verbose)
 * 
 *     def load(self):             # <<<<<<<<<<<<<<
 *         if self.loaded:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("fatslimlib.core_base.IndexReader.load", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1289
 *         self.fast_load()
 * 
 *     cdef fast_load(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

static PyObject *__pyx_f_10fatslimlib_9core_base_11IndexReader_fast_load(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fast_load", 0);

  /* "fatslimlib/core_base.pyx":1290
 * 
 *     cdef fast_load(self):
 *         raise NotImplementedError             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_Raise(__pyx_builtin_NotImplementedError, 0, 0, 0);
  __PYX_ERR(0, 1290, __pyx_L1_error)

  /* "fatslimlib/core_base.pyx":1289
 *         self.fast_load()
 * 
 *     cdef fast_load(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.IndexReader.fast_load", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1292
 *         raise NotImplementedError
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.groups)
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_10fatslimlib_9core_base_11IndexReader_5__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_10fatslimlib_9core_base_11IndexReader_5__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_11IndexReader_4__len__(((struct __pyx_obj_10fatslimlib_9core_base_IndexReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_10fatslimlib_9core_base_11IndexReader_4__len__(struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "fatslimlib/core_base.pyx":1293
 * 
 *     def __len__(self):
 *         return len(self.groups)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, str item):
 */
  __pyx_t_1 = __pyx_v_self->groups;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1293, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1292
 *         raise NotImplementedError
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.groups)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.IndexReader.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1295
 *         return len(self.groups)
 * 
 *     def __getitem__(self, str item):             # <<<<<<<<<<<<<<
 *         return self.groups[item.encode()]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_11IndexReader_7__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_11IndexReader_7__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_item), (&PyUnicode_Type), 1, "item", 1))) __PYX_ERR(0, 1295, __pyx_L1_error)
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_11IndexReader_6__getitem__(((struct __pyx_obj_10fatslimlib_9core_base_IndexReader *)__pyx_v_self), ((PyObject*)__pyx_v_item));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_11IndexReader_6__getitem__(struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "fatslimlib/core_base.pyx":1296
 * 
 *     def __getitem__(self, str item):
 *         return self.groups[item.encode()]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_item == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 1296, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_item, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_v_self->groups, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1295
 *         return len(self.groups)
 * 
 *     def __getitem__(self, str item):             # <<<<<<<<<<<<<<
 *         return self.groups[item.encode()]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("fatslimlib.core_base.IndexReader.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pxd":208
 *     cdef bint loaded
 *     cdef bytes filename
 *     cdef readonly object groups             # <<<<<<<<<<<<<<
 * 
 *     # Methods
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_11IndexReader_6groups_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_11IndexReader_6groups_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_11IndexReader_6groups___get__(((struct __pyx_obj_10fatslimlib_9core_base_IndexReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_11IndexReader_6groups___get__(struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->groups);
  __pyx_r = __pyx_v_self->groups;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_11IndexReader_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_11IndexReader_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_11IndexReader_8__reduce_cython__(((struct __pyx_obj_10fatslimlib_9core_base_IndexReader *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_11IndexReader_8__reduce_cython__(struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.filename, self.groups, self.loaded)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->loaded); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->filename);
  __Pyx_GIVEREF(__pyx_v_self->filename);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->filename);
  __Pyx_INCREF(__pyx_v_self->groups);
  __Pyx_GIVEREF(__pyx_v_self->groups);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self->groups);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.filename, self.groups, self.loaded)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.filename, self.groups, self.loaded)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.filename is not None or self.groups is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.filename, self.groups, self.loaded)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.filename is not None or self.groups is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_IndexReader, (type(self), 0x7ac6dcc, None), state
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->filename != ((PyObject*)Py_None));
    __pyx_t_5 = (__pyx_t_3 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->groups != Py_None);
    __pyx_t_3 = (__pyx_t_5 != 0);
    __pyx_t_4 = __pyx_t_3;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.filename is not None or self.groups is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IndexReader, (type(self), 0x7ac6dcc, None), state
 *     else:
 */
  __pyx_t_4 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_4) {

    /* "(tree fragment)":13
 *         use_setstate = self.filename is not None or self.groups is not None
 *     if use_setstate:
 *         return __pyx_unpickle_IndexReader, (type(self), 0x7ac6dcc, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_IndexReader, (type(self), 0x7ac6dcc, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_IndexReader); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_128740812);
    __Pyx_GIVEREF(__pyx_int_128740812);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_128740812);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.filename is not None or self.groups is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IndexReader, (type(self), 0x7ac6dcc, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_IndexReader, (type(self), 0x7ac6dcc, None), state
 *     else:
 *         return __pyx_unpickle_IndexReader, (type(self), 0x7ac6dcc, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IndexReader__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_IndexReader); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_128740812);
    __Pyx_GIVEREF(__pyx_int_128740812);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_128740812);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_6 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("fatslimlib.core_base.IndexReader.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IndexReader, (type(self), 0x7ac6dcc, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IndexReader__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_11IndexReader_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_11IndexReader_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_11IndexReader_10__setstate_cython__(((struct __pyx_obj_10fatslimlib_9core_base_IndexReader *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_11IndexReader_10__setstate_cython__(struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_IndexReader, (type(self), 0x7ac6dcc, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IndexReader__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10fatslimlib_9core_base___pyx_unpickle_IndexReader__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IndexReader, (type(self), 0x7ac6dcc, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IndexReader__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.IndexReader.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1300
 * 
 * cdef class Frame(object):
 *     def __init__(self, Trajectory trajectory, fsl_int index, real timestep):             # <<<<<<<<<<<<<<
 *         self.trajectory = trajectory
 *         self.topology = trajectory.topology
 */

/* Python wrapper */
static int __pyx_pw_10fatslimlib_9core_base_5Frame_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10fatslimlib_9core_base_5Frame_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_trajectory = 0;
  fsl_int __pyx_v_index;
  real __pyx_v_timestep;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_trajectory,&__pyx_n_s_index,&__pyx_n_s_timestep,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trajectory)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 1300, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_timestep)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 1300, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1300, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_trajectory = ((struct __pyx_obj_10fatslimlib_9core_base_Trajectory *)values[0]);
    __pyx_v_index = __Pyx_PyInt_As_fsl_int(values[1]); if (unlikely((__pyx_v_index == ((fsl_int)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1300, __pyx_L3_error)
    __pyx_v_timestep = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_timestep == ((real)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1300, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1300, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_trajectory), __pyx_ptype_10fatslimlib_9core_base_Trajectory, 1, "trajectory", 0))) __PYX_ERR(0, 1300, __pyx_L1_error)
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame___init__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self), __pyx_v_trajectory, __pyx_v_index, __pyx_v_timestep);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10fatslimlib_9core_base_5Frame___init__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self, struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_trajectory, fsl_int __pyx_v_index, real __pyx_v_timestep) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "fatslimlib/core_base.pyx":1301
 * cdef class Frame(object):
 *     def __init__(self, Trajectory trajectory, fsl_int index, real timestep):
 *         self.trajectory = trajectory             # <<<<<<<<<<<<<<
 *         self.topology = trajectory.topology
 *         self.coords_reader = trajectory.coords_reader
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_trajectory));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_trajectory));
  __Pyx_GOTREF(__pyx_v_self->trajectory);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->trajectory));
  __pyx_v_self->trajectory = __pyx_v_trajectory;

  /* "fatslimlib/core_base.pyx":1302
 *     def __init__(self, Trajectory trajectory, fsl_int index, real timestep):
 *         self.trajectory = trajectory
 *         self.topology = trajectory.topology             # <<<<<<<<<<<<<<
 *         self.coords_reader = trajectory.coords_reader
 *         self.index = index
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_trajectory->topology);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->topology);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->topology));
  __pyx_v_self->topology = ((struct __pyx_obj_10fatslimlib_9core_base_Topology *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":1303
 *         self.trajectory = trajectory
 *         self.topology = trajectory.topology
 *         self.coords_reader = trajectory.coords_reader             # <<<<<<<<<<<<<<
 *         self.index = index
 *         self.timestep = timestep
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_trajectory->coords_reader);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->coords_reader);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->coords_reader));
  __pyx_v_self->coords_reader = ((struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":1304
 *         self.topology = trajectory.topology
 *         self.coords_reader = trajectory.coords_reader
 *         self.index = index             # <<<<<<<<<<<<<<
 *         self.timestep = timestep
 *         self.box = self.coords_reader.load_box(index)
 */
  __pyx_v_self->index = __pyx_v_index;

  /* "fatslimlib/core_base.pyx":1305
 *         self.coords_reader = trajectory.coords_reader
 *         self.index = index
 *         self.timestep = timestep             # <<<<<<<<<<<<<<
 *         self.box = self.coords_reader.load_box(index)
 *         # -1 line below => there is one more offset compared to actual number of lipids!
 */
  __pyx_v_self->timestep = __pyx_v_timestep;

  /* "fatslimlib/core_base.pyx":1306
 *         self.index = index
 *         self.timestep = timestep
 *         self.box = self.coords_reader.load_box(index)             # <<<<<<<<<<<<<<
 *         # -1 line below => there is one more offset compared to actual number of lipids!
 *         self.size = self.trajectory.lipid_atomids_offsets.shape[0] - 1
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_10fatslimlib_9core_base_CoordinateReader *)__pyx_v_self->coords_reader->__pyx_vtab)->load_box(__pyx_v_self->coords_reader, __pyx_v_index)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->box);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->box));
  __pyx_v_self->box = ((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":1308
 *         self.box = self.coords_reader.load_box(index)
 *         # -1 line below => there is one more offset compared to actual number of lipids!
 *         self.size = self.trajectory.lipid_atomids_offsets.shape[0] - 1             # <<<<<<<<<<<<<<
 * 
 *         # Headgroups
 */
  if (unlikely(!__pyx_v_self->trajectory->lipid_atomids_offsets.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1308, __pyx_L1_error)}
  __pyx_v_self->size = ((__pyx_v_self->trajectory->lipid_atomids_offsets.shape[0]) - 1);

  /* "fatslimlib/core_base.pyx":1311
 * 
 *         # Headgroups
 *         self.hg_group_coords = None             # <<<<<<<<<<<<<<
 *         self.hg_coords_bbox = None
 *         # Lipids
 */
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1311, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->hg_group_coords, 0);
  __pyx_v_self->hg_group_coords = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "fatslimlib/core_base.pyx":1312
 *         # Headgroups
 *         self.hg_group_coords = None
 *         self.hg_coords_bbox = None             # <<<<<<<<<<<<<<
 *         # Lipids
 *         self.lipid_coords = None
 */
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1312, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->hg_coords_bbox, 0);
  __pyx_v_self->hg_coords_bbox = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "fatslimlib/core_base.pyx":1314
 *         self.hg_coords_bbox = None
 *         # Lipids
 *         self.lipid_coords = None             # <<<<<<<<<<<<<<
 *         self.lipid_coords_aslist = None
 *         self.lipid_coords_bbox = None
 */
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->lipid_coords, 0);
  __pyx_v_self->lipid_coords = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "fatslimlib/core_base.pyx":1315
 *         # Lipids
 *         self.lipid_coords = None
 *         self.lipid_coords_aslist = None             # <<<<<<<<<<<<<<
 *         self.lipid_coords_bbox = None
 *         self.lipids_coords_bbox_aslist = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->lipid_coords_aslist);
  __Pyx_DECREF(__pyx_v_self->lipid_coords_aslist);
  __pyx_v_self->lipid_coords_aslist = ((PyObject*)Py_None);

  /* "fatslimlib/core_base.pyx":1316
 *         self.lipid_coords = None
 *         self.lipid_coords_aslist = None
 *         self.lipid_coords_bbox = None             # <<<<<<<<<<<<<<
 *         self.lipids_coords_bbox_aslist = None
 *         self.lipid_atomids_offsets = self.trajectory.lipid_atomids_offsets
 */
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1316, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->lipid_coords_bbox, 0);
  __pyx_v_self->lipid_coords_bbox = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "fatslimlib/core_base.pyx":1317
 *         self.lipid_coords_aslist = None
 *         self.lipid_coords_bbox = None
 *         self.lipids_coords_bbox_aslist = None             # <<<<<<<<<<<<<<
 *         self.lipid_atomids_offsets = self.trajectory.lipid_atomids_offsets
 *         # Simplified lipids
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->lipids_coords_bbox_aslist);
  __Pyx_DECREF(__pyx_v_self->lipids_coords_bbox_aslist);
  __pyx_v_self->lipids_coords_bbox_aslist = ((PyObject*)Py_None);

  /* "fatslimlib/core_base.pyx":1318
 *         self.lipid_coords_bbox = None
 *         self.lipids_coords_bbox_aslist = None
 *         self.lipid_atomids_offsets = self.trajectory.lipid_atomids_offsets             # <<<<<<<<<<<<<<
 *         # Simplified lipids
 *         self.bead_coords = None
 */
  if (unlikely(!__pyx_v_self->trajectory->lipid_atomids_offsets.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1318, __pyx_L1_error)}
  __pyx_t_3 = __pyx_v_self->trajectory->lipid_atomids_offsets;
  __PYX_INC_MEMVIEW(&__pyx_t_3, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->lipid_atomids_offsets, 0);
  __pyx_v_self->lipid_atomids_offsets = __pyx_t_3;
  __pyx_t_3.memview = NULL;
  __pyx_t_3.data = NULL;

  /* "fatslimlib/core_base.pyx":1320
 *         self.lipid_atomids_offsets = self.trajectory.lipid_atomids_offsets
 *         # Simplified lipids
 *         self.bead_coords = None             # <<<<<<<<<<<<<<
 *         self.bead_coords_bbox = None
 *         # Interacting groups
 */
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1320, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->bead_coords, 0);
  __pyx_v_self->bead_coords = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "fatslimlib/core_base.pyx":1321
 *         # Simplified lipids
 *         self.bead_coords = None
 *         self.bead_coords_bbox = None             # <<<<<<<<<<<<<<
 *         # Interacting groups
 *         self.interacting_group_coords = None
 */
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1321, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->bead_coords_bbox, 0);
  __pyx_v_self->bead_coords_bbox = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "fatslimlib/core_base.pyx":1323
 *         self.bead_coords_bbox = None
 *         # Interacting groups
 *         self.interacting_group_coords = None             # <<<<<<<<<<<<<<
 *         self.interacting_group_coords_bbox = None
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1323, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->interacting_group_coords, 0);
  __pyx_v_self->interacting_group_coords = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "fatslimlib/core_base.pyx":1324
 *         # Interacting groups
 *         self.interacting_group_coords = None
 *         self.interacting_group_coords_bbox = None             # <<<<<<<<<<<<<<
 * 
 *         # NS-related
 */
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1324, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->interacting_group_coords_bbox, 0);
  __pyx_v_self->interacting_group_coords_bbox = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "fatslimlib/core_base.pyx":1327
 * 
 *         # NS-related
 *         self.proximity_cutoff = NOTSET             # <<<<<<<<<<<<<<
 *         self.neighbors = NULL
 * 
 */
  __pyx_v_self->proximity_cutoff = -12345.0;

  /* "fatslimlib/core_base.pyx":1328
 *         # NS-related
 *         self.proximity_cutoff = NOTSET
 *         self.neighbors = NULL             # <<<<<<<<<<<<<<
 * 
 *         # Directions & Normals
 */
  __pyx_v_self->neighbors = NULL;

  /* "fatslimlib/core_base.pyx":1331
 * 
 *         # Directions & Normals
 *         self.directions = None             # <<<<<<<<<<<<<<
 *         self.normals = None
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1331, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->directions, 0);
  __pyx_v_self->directions = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "fatslimlib/core_base.pyx":1332
 *         # Directions & Normals
 *         self.directions = None
 *         self.normals = None             # <<<<<<<<<<<<<<
 * 
 *         # Aggregates
 */
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->normals, 0);
  __pyx_v_self->normals = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "fatslimlib/core_base.pyx":1335
 * 
 *         # Aggregates
 *         self.aggregates = []             # <<<<<<<<<<<<<<
 *         self.aggregates_retrieved = False
 *         self.aggregates_cutoff = NOTSET
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->aggregates);
  __Pyx_DECREF(__pyx_v_self->aggregates);
  __pyx_v_self->aggregates = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":1336
 *         # Aggregates
 *         self.aggregates = []
 *         self.aggregates_retrieved = False             # <<<<<<<<<<<<<<
 *         self.aggregates_cutoff = NOTSET
 * 
 */
  __pyx_v_self->aggregates_retrieved = 0;

  /* "fatslimlib/core_base.pyx":1337
 *         self.aggregates = []
 *         self.aggregates_retrieved = False
 *         self.aggregates_cutoff = NOTSET             # <<<<<<<<<<<<<<
 * 
 *         # Leaflets & Membranes
 */
  __pyx_v_self->aggregates_cutoff = -12345.0;

  /* "fatslimlib/core_base.pyx":1340
 * 
 *         # Leaflets & Membranes
 *         self.membranes = []             # <<<<<<<<<<<<<<
 *         self.membranes_retrieved = False
 *         self.membranes_cutoff = NOTSET
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->membranes);
  __Pyx_DECREF(__pyx_v_self->membranes);
  __pyx_v_self->membranes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":1341
 *         # Leaflets & Membranes
 *         self.membranes = []
 *         self.membranes_retrieved = False             # <<<<<<<<<<<<<<
 *         self.membranes_cutoff = NOTSET
 * 
 */
  __pyx_v_self->membranes_retrieved = 0;

  /* "fatslimlib/core_base.pyx":1342
 *         self.membranes = []
 *         self.membranes_retrieved = False
 *         self.membranes_cutoff = NOTSET             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->membranes_cutoff = -12345.0;

  /* "fatslimlib/core_base.pyx":1300
 * 
 * cdef class Frame(object):
 *     def __init__(self, Trajectory trajectory, fsl_int index, real timestep):             # <<<<<<<<<<<<<<
 *         self.trajectory = trajectory
 *         self.topology = trajectory.topology
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 1);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1344
 *         self.membranes_cutoff = NOTSET
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         free_neighborhood_holder(self.neighbors)
 *         self.neighbors = NULL
 */

/* Python wrapper */
static void __pyx_pw_10fatslimlib_9core_base_5Frame_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10fatslimlib_9core_base_5Frame_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_10fatslimlib_9core_base_5Frame_2__dealloc__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_10fatslimlib_9core_base_5Frame_2__dealloc__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "fatslimlib/core_base.pyx":1345
 * 
 *     def __dealloc__(self):
 *         free_neighborhood_holder(self.neighbors)             # <<<<<<<<<<<<<<
 *         self.neighbors = NULL
 * 
 */
  __pyx_f_10fatslimlib_7core_ns_free_neighborhood_holder(__pyx_v_self->neighbors);

  /* "fatslimlib/core_base.pyx":1346
 *     def __dealloc__(self):
 *         free_neighborhood_holder(self.neighbors)
 *         self.neighbors = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __pyx_v_self->neighbors = NULL;

  /* "fatslimlib/core_base.pyx":1344
 *         self.membranes_cutoff = NOTSET
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         free_neighborhood_holder(self.neighbors)
 *         self.neighbors = NULL
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "fatslimlib/core_base.pyx":1348
 *         self.neighbors = NULL
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.size
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_10fatslimlib_9core_base_5Frame_5__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_10fatslimlib_9core_base_5Frame_5__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_4__len__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_10fatslimlib_9core_base_5Frame_4__len__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "fatslimlib/core_base.pyx":1349
 * 
 *     def __len__(self):
 *         return self.size             # <<<<<<<<<<<<<<
 * 
 *     cpdef fsl_int get_forcefield_type(self):
 */
  __pyx_r = __pyx_v_self->size;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1348
 *         self.neighbors = NULL
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.size
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1351
 *         return self.size
 * 
 *     cpdef fsl_int get_forcefield_type(self):             # <<<<<<<<<<<<<<
 *         return self.trajectory.get_forcefield_type()
 * 
 */

static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_7get_forcefield_type(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static fsl_int __pyx_f_10fatslimlib_9core_base_5Frame_get_forcefield_type(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self, int __pyx_skip_dispatch) {
  fsl_int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  fsl_int __pyx_t_5;
  __Pyx_RefNannySetupContext("get_forcefield_type", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_forcefield_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1351, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10fatslimlib_9core_base_5Frame_7get_forcefield_type)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1351, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_fsl_int(__pyx_t_2); if (unlikely((__pyx_t_5 == ((fsl_int)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1351, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "fatslimlib/core_base.pyx":1352
 * 
 *     cpdef fsl_int get_forcefield_type(self):
 *         return self.trajectory.get_forcefield_type()             # <<<<<<<<<<<<<<
 * 
 *     cdef real[:,::1] fast_get_hg_group_coords(self) nogil:
 */
  __pyx_r = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Trajectory *)__pyx_v_self->trajectory->__pyx_vtab)->get_forcefield_type(__pyx_v_self->trajectory, 0);
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1351
 *         return self.size
 * 
 *     cpdef fsl_int get_forcefield_type(self):             # <<<<<<<<<<<<<<
 *         return self.trajectory.get_forcefield_type()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("fatslimlib.core_base.Frame.get_forcefield_type", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_7get_forcefield_type(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_7get_forcefield_type(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_forcefield_type (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_6get_forcefield_type(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_6get_forcefield_type(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_forcefield_type", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_f_10fatslimlib_9core_base_5Frame_get_forcefield_type(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.get_forcefield_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1354
 *         return self.trajectory.get_forcefield_type()
 * 
 *     cdef real[:,::1] fast_get_hg_group_coords(self) nogil:             # <<<<<<<<<<<<<<
 *         if self.hg_group_coords is None:
 *             with gil:
 */

static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_hg_group_coords(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("fast_get_hg_group_coords", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "fatslimlib/core_base.pyx":1355
 * 
 *     cdef real[:,::1] fast_get_hg_group_coords(self) nogil:
 *         if self.hg_group_coords is None:             # <<<<<<<<<<<<<<
 *             with gil:
 *                 self.hg_group_coords = self.coords_reader.\
 */
  /*try:*/ {
    if (unlikely(!__pyx_v_self->hg_group_coords.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1355, __pyx_L4_error)}
    __pyx_t_1 = ((((PyObject *) __pyx_v_self->hg_group_coords.memview) == Py_None) != 0);
    if (__pyx_t_1) {

      /* "fatslimlib/core_base.pyx":1356
 *     cdef real[:,::1] fast_get_hg_group_coords(self) nogil:
 *         if self.hg_group_coords is None:
 *             with gil:             # <<<<<<<<<<<<<<
 *                 self.hg_group_coords = self.coords_reader.\
 *                     load_coords(self.index,self.trajectory.hg_group_atomids)
 */
      {
          #ifdef WITH_THREAD
          PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
          #endif
          /*try:*/ {

            /* "fatslimlib/core_base.pyx":1358
 *             with gil:
 *                 self.hg_group_coords = self.coords_reader.\
 *                     load_coords(self.index,self.trajectory.hg_group_atomids)             # <<<<<<<<<<<<<<
 *         return self.hg_group_coords
 * 
 */
            if (unlikely(!__pyx_v_self->trajectory->hg_group_atomids.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1358, __pyx_L8_error)}
            __pyx_t_2 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_CoordinateReader *)__pyx_v_self->coords_reader->__pyx_vtab)->load_coords(__pyx_v_self->coords_reader, __pyx_v_self->index, __pyx_v_self->trajectory->hg_group_atomids); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1358, __pyx_L8_error)

            /* "fatslimlib/core_base.pyx":1357
 *         if self.hg_group_coords is None:
 *             with gil:
 *                 self.hg_group_coords = self.coords_reader.\             # <<<<<<<<<<<<<<
 *                     load_coords(self.index,self.trajectory.hg_group_atomids)
 *         return self.hg_group_coords
 */
            __PYX_XDEC_MEMVIEW(&__pyx_v_self->hg_group_coords, 0);
            __pyx_v_self->hg_group_coords = __pyx_t_2;
            __pyx_t_2.memview = NULL;
            __pyx_t_2.data = NULL;
          }

          /* "fatslimlib/core_base.pyx":1356
 *     cdef real[:,::1] fast_get_hg_group_coords(self) nogil:
 *         if self.hg_group_coords is None:
 *             with gil:             # <<<<<<<<<<<<<<
 *                 self.hg_group_coords = self.coords_reader.\
 *                     load_coords(self.index,self.trajectory.hg_group_atomids)
 */
          /*finally:*/ {
            /*normal exit:*/{
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              goto __pyx_L9;
            }
            __pyx_L8_error: {
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              goto __pyx_L4_error;
            }
            __pyx_L9:;
          }
      }

      /* "fatslimlib/core_base.pyx":1355
 * 
 *     cdef real[:,::1] fast_get_hg_group_coords(self) nogil:
 *         if self.hg_group_coords is None:             # <<<<<<<<<<<<<<
 *             with gil:
 *                 self.hg_group_coords = self.coords_reader.\
 */
    }

    /* "fatslimlib/core_base.pyx":1359
 *                 self.hg_group_coords = self.coords_reader.\
 *                     load_coords(self.index,self.trajectory.hg_group_atomids)
 *         return self.hg_group_coords             # <<<<<<<<<<<<<<
 * 
 *     # Associated python property
 */
    if (unlikely(!__pyx_v_self->hg_group_coords.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1359, __pyx_L4_error)}
    __PYX_INC_MEMVIEW(&__pyx_v_self->hg_group_coords, 1);
    __pyx_r = __pyx_v_self->hg_group_coords;
    goto __pyx_L3_return;
  }

  /* "fatslimlib/core_base.pyx":1355
 * 
 *     cdef real[:,::1] fast_get_hg_group_coords(self) nogil:
 *         if self.hg_group_coords is None:             # <<<<<<<<<<<<<<
 *             with gil:
 *                 self.hg_group_coords = self.coords_reader.\
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "fatslimlib/core_base.pyx":1354
 *         return self.trajectory.get_forcefield_type()
 * 
 *     cdef real[:,::1] fast_get_hg_group_coords(self) nogil:             # <<<<<<<<<<<<<<
 *         if self.hg_group_coords is None:
 *             with gil:
 */

  /* function exit code */
  __Pyx_pretend_to_initialize(&__pyx_r);
  goto __pyx_L0;
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 0);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.fast_get_hg_group_coords", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1363
 *     # Associated python property
 *     property hg_group_coords:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef real[:, ::1] memview
 *             with nogil:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_15hg_group_coords_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_15hg_group_coords_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_15hg_group_coords___get__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_15hg_group_coords___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":1365
 *         def __get__(self):
 *             cdef real[:, ::1] memview
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 memview = self.fast_get_hg_group_coords()
 *             return np.asarray(memview)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "fatslimlib/core_base.pyx":1366
 *             cdef real[:, ::1] memview
 *             with nogil:
 *                 memview = self.fast_get_hg_group_coords()             # <<<<<<<<<<<<<<
 *             return np.asarray(memview)
 * 
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_hg_group_coords(__pyx_v_self); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(0, 1366, __pyx_L4_error)
        __pyx_v_memview = __pyx_t_1;
        __pyx_t_1.memview = NULL;
        __pyx_t_1.data = NULL;
      }

      /* "fatslimlib/core_base.pyx":1365
 *         def __get__(self):
 *             cdef real[:, ::1] memview
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 memview = self.fast_get_hg_group_coords()
 *             return np.asarray(memview)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "fatslimlib/core_base.pyx":1367
 *             with nogil:
 *                 memview = self.fast_get_hg_group_coords()
 *             return np.asarray(memview)             # <<<<<<<<<<<<<<
 * 
 *     cdef real[:,::1] fast_get_hg_group_coords_bbox(self) nogil:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_asarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_memview, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1363
 *     # Associated python property
 *     property hg_group_coords:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef real[:, ::1] memview
 *             with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.hg_group_coords.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_memview, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1369
 *             return np.asarray(memview)
 * 
 *     cdef real[:,::1] fast_get_hg_group_coords_bbox(self) nogil:             # <<<<<<<<<<<<<<
 *         if self.hg_coords_bbox is None:
 *             self.hg_coords_bbox = self.box.fast_put_atoms_in_bbox(self.fast_get_hg_group_coords())
 */

static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_hg_group_coords_bbox(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_1;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };

  /* "fatslimlib/core_base.pyx":1370
 * 
 *     cdef real[:,::1] fast_get_hg_group_coords_bbox(self) nogil:
 *         if self.hg_coords_bbox is None:             # <<<<<<<<<<<<<<
 *             self.hg_coords_bbox = self.box.fast_put_atoms_in_bbox(self.fast_get_hg_group_coords())
 *         return self.hg_coords_bbox
 */
  if (unlikely(!__pyx_v_self->hg_coords_bbox.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1370, __pyx_L1_error)}
  __pyx_t_1 = ((((PyObject *) __pyx_v_self->hg_coords_bbox.memview) == Py_None) != 0);
  if (__pyx_t_1) {

    /* "fatslimlib/core_base.pyx":1371
 *     cdef real[:,::1] fast_get_hg_group_coords_bbox(self) nogil:
 *         if self.hg_coords_bbox is None:
 *             self.hg_coords_bbox = self.box.fast_put_atoms_in_bbox(self.fast_get_hg_group_coords())             # <<<<<<<<<<<<<<
 *         return self.hg_coords_bbox
 * 
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_hg_group_coords(__pyx_v_self); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1371, __pyx_L1_error)
    __pyx_t_3 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_self->box->__pyx_vtab)->fast_put_atoms_in_bbox(__pyx_v_self->box, __pyx_t_2); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(0, 1371, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_t_2, 0);
    __pyx_t_2.memview = NULL;
    __pyx_t_2.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->hg_coords_bbox, 0);
    __pyx_v_self->hg_coords_bbox = __pyx_t_3;
    __pyx_t_3.memview = NULL;
    __pyx_t_3.data = NULL;

    /* "fatslimlib/core_base.pyx":1370
 * 
 *     cdef real[:,::1] fast_get_hg_group_coords_bbox(self) nogil:
 *         if self.hg_coords_bbox is None:             # <<<<<<<<<<<<<<
 *             self.hg_coords_bbox = self.box.fast_put_atoms_in_bbox(self.fast_get_hg_group_coords())
 *         return self.hg_coords_bbox
 */
  }

  /* "fatslimlib/core_base.pyx":1372
 *         if self.hg_coords_bbox is None:
 *             self.hg_coords_bbox = self.box.fast_put_atoms_in_bbox(self.fast_get_hg_group_coords())
 *         return self.hg_coords_bbox             # <<<<<<<<<<<<<<
 * 
 *     # Associated python property
 */
  if (unlikely(!__pyx_v_self->hg_coords_bbox.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1372, __pyx_L1_error)}
  __PYX_INC_MEMVIEW(&__pyx_v_self->hg_coords_bbox, 1);
  __pyx_r = __pyx_v_self->hg_coords_bbox;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1369
 *             return np.asarray(memview)
 * 
 *     cdef real[:,::1] fast_get_hg_group_coords_bbox(self) nogil:             # <<<<<<<<<<<<<<
 *         if self.hg_coords_bbox is None:
 *             self.hg_coords_bbox = self.box.fast_put_atoms_in_bbox(self.fast_get_hg_group_coords())
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 0);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 0);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("fatslimlib.core_base.Frame.fast_get_hg_group_coords_bbox", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1376
 *     # Associated python property
 *     property hg_group_coords_bbox:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef real[:, ::1] memview
 *             with nogil:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_20hg_group_coords_bbox_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_20hg_group_coords_bbox_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_20hg_group_coords_bbox___get__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_20hg_group_coords_bbox___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":1378
 *         def __get__(self):
 *             cdef real[:, ::1] memview
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 memview = self.fast_get_hg_group_coords_bbox()
 *             return np.asarray(memview)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "fatslimlib/core_base.pyx":1379
 *             cdef real[:, ::1] memview
 *             with nogil:
 *                 memview = self.fast_get_hg_group_coords_bbox()             # <<<<<<<<<<<<<<
 *             return np.asarray(memview)
 * 
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_hg_group_coords_bbox(__pyx_v_self); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(0, 1379, __pyx_L4_error)
        __pyx_v_memview = __pyx_t_1;
        __pyx_t_1.memview = NULL;
        __pyx_t_1.data = NULL;
      }

      /* "fatslimlib/core_base.pyx":1378
 *         def __get__(self):
 *             cdef real[:, ::1] memview
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 memview = self.fast_get_hg_group_coords_bbox()
 *             return np.asarray(memview)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "fatslimlib/core_base.pyx":1380
 *             with nogil:
 *                 memview = self.fast_get_hg_group_coords_bbox()
 *             return np.asarray(memview)             # <<<<<<<<<<<<<<
 * 
 *     cdef real[:, ::1] fast_get_lipid_coords(self) nogil:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_asarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_memview, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1376
 *     # Associated python property
 *     property hg_group_coords_bbox:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef real[:, ::1] memview
 *             with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.hg_group_coords_bbox.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_memview, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1382
 *             return np.asarray(memview)
 * 
 *     cdef real[:, ::1] fast_get_lipid_coords(self) nogil:             # <<<<<<<<<<<<<<
 *         if self.lipid_coords is None:
 *             self.lipid_coords = self.coords_reader.\
 */

static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_lipid_coords(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_1;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };

  /* "fatslimlib/core_base.pyx":1383
 * 
 *     cdef real[:, ::1] fast_get_lipid_coords(self) nogil:
 *         if self.lipid_coords is None:             # <<<<<<<<<<<<<<
 *             self.lipid_coords = self.coords_reader.\
 *                 load_coords(self.index, self.trajectory.lipid_atomids)
 */
  if (unlikely(!__pyx_v_self->lipid_coords.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1383, __pyx_L1_error)}
  __pyx_t_1 = ((((PyObject *) __pyx_v_self->lipid_coords.memview) == Py_None) != 0);
  if (__pyx_t_1) {

    /* "fatslimlib/core_base.pyx":1385
 *         if self.lipid_coords is None:
 *             self.lipid_coords = self.coords_reader.\
 *                 load_coords(self.index, self.trajectory.lipid_atomids)             # <<<<<<<<<<<<<<
 *         return self.lipid_coords
 * 
 */
    if (unlikely(!__pyx_v_self->trajectory->lipid_atomids.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1385, __pyx_L1_error)}
    __pyx_t_2 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_CoordinateReader *)__pyx_v_self->coords_reader->__pyx_vtab)->load_coords(__pyx_v_self->coords_reader, __pyx_v_self->index, __pyx_v_self->trajectory->lipid_atomids); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1385, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":1384
 *     cdef real[:, ::1] fast_get_lipid_coords(self) nogil:
 *         if self.lipid_coords is None:
 *             self.lipid_coords = self.coords_reader.\             # <<<<<<<<<<<<<<
 *                 load_coords(self.index, self.trajectory.lipid_atomids)
 *         return self.lipid_coords
 */
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->lipid_coords, 0);
    __pyx_v_self->lipid_coords = __pyx_t_2;
    __pyx_t_2.memview = NULL;
    __pyx_t_2.data = NULL;

    /* "fatslimlib/core_base.pyx":1383
 * 
 *     cdef real[:, ::1] fast_get_lipid_coords(self) nogil:
 *         if self.lipid_coords is None:             # <<<<<<<<<<<<<<
 *             self.lipid_coords = self.coords_reader.\
 *                 load_coords(self.index, self.trajectory.lipid_atomids)
 */
  }

  /* "fatslimlib/core_base.pyx":1386
 *             self.lipid_coords = self.coords_reader.\
 *                 load_coords(self.index, self.trajectory.lipid_atomids)
 *         return self.lipid_coords             # <<<<<<<<<<<<<<
 * 
 *     cdef list get_lipid_coords_aslist(self):
 */
  if (unlikely(!__pyx_v_self->lipid_coords.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1386, __pyx_L1_error)}
  __PYX_INC_MEMVIEW(&__pyx_v_self->lipid_coords, 1);
  __pyx_r = __pyx_v_self->lipid_coords;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1382
 *             return np.asarray(memview)
 * 
 *     cdef real[:, ::1] fast_get_lipid_coords(self) nogil:             # <<<<<<<<<<<<<<
 *         if self.lipid_coords is None:
 *             self.lipid_coords = self.coords_reader.\
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 0);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("fatslimlib.core_base.Frame.fast_get_lipid_coords", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1388
 *         return self.lipid_coords
 * 
 *     cdef list get_lipid_coords_aslist(self):             # <<<<<<<<<<<<<<
 *         cdef fsl_int[:] offsets = self.lipid_atomids_offsets
 *         cdef real[:, ::1] coords = self.fast_get_lipid_coords()
 */

static PyObject *__pyx_f_10fatslimlib_9core_base_5Frame_get_lipid_coords_aslist(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_offsets = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_coords = { 0, 0, { 0 }, { 0 }, { 0 } };
  fsl_int __pyx_v_i;
  PyObject *__pyx_v_coords_list = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  fsl_int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  __Pyx_RefNannySetupContext("get_lipid_coords_aslist", 0);

  /* "fatslimlib/core_base.pyx":1389
 * 
 *     cdef list get_lipid_coords_aslist(self):
 *         cdef fsl_int[:] offsets = self.lipid_atomids_offsets             # <<<<<<<<<<<<<<
 *         cdef real[:, ::1] coords = self.fast_get_lipid_coords()
 *         cdef fsl_int i
 */
  if (unlikely(!__pyx_v_self->lipid_atomids_offsets.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1389, __pyx_L1_error)}
  __pyx_t_1 = __pyx_v_self->lipid_atomids_offsets;
  __PYX_INC_MEMVIEW(&__pyx_t_1, 1);
  __pyx_v_offsets = __pyx_t_1;
  __pyx_t_1.memview = NULL;
  __pyx_t_1.data = NULL;

  /* "fatslimlib/core_base.pyx":1390
 *     cdef list get_lipid_coords_aslist(self):
 *         cdef fsl_int[:] offsets = self.lipid_atomids_offsets
 *         cdef real[:, ::1] coords = self.fast_get_lipid_coords()             # <<<<<<<<<<<<<<
 *         cdef fsl_int i
 *         if self.lipid_coords_aslist is None:
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_lipid_coords(__pyx_v_self); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1390, __pyx_L1_error)
  __pyx_v_coords = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "fatslimlib/core_base.pyx":1392
 *         cdef real[:, ::1] coords = self.fast_get_lipid_coords()
 *         cdef fsl_int i
 *         if self.lipid_coords_aslist is None:             # <<<<<<<<<<<<<<
 *             coords_list = []
 * 
 */
  __pyx_t_3 = (__pyx_v_self->lipid_coords_aslist == ((PyObject*)Py_None));
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "fatslimlib/core_base.pyx":1393
 *         cdef fsl_int i
 *         if self.lipid_coords_aslist is None:
 *             coords_list = []             # <<<<<<<<<<<<<<
 * 
 *             for i in range(offsets.shape[0] - 1):
 */
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_coords_list = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "fatslimlib/core_base.pyx":1395
 *             coords_list = []
 * 
 *             for i in range(offsets.shape[0] - 1):             # <<<<<<<<<<<<<<
 *                 coords_list.append(np.asarray(coords[offsets[i]:offsets[i+1]]))
 * 
 */
    __pyx_t_6 = ((__pyx_v_offsets.shape[0]) - 1);
    __pyx_t_7 = __pyx_t_6;
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
      __pyx_v_i = __pyx_t_8;

      /* "fatslimlib/core_base.pyx":1396
 * 
 *             for i in range(offsets.shape[0] - 1):
 *                 coords_list.append(np.asarray(coords[offsets[i]:offsets[i+1]]))             # <<<<<<<<<<<<<<
 * 
 *             self.lipid_coords_aslist = coords_list
 */
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_np); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_asarray); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_11 = __pyx_v_i;
      __pyx_t_12 = (__pyx_v_i + 1);
      __pyx_t_2.data = __pyx_v_coords.data;
      __pyx_t_2.memview = __pyx_v_coords.memview;
      __PYX_INC_MEMVIEW(&__pyx_t_2, 0);
      __pyx_t_13 = -1;
      if (unlikely(__pyx_memoryview_slice_memviewslice(
    &__pyx_t_2,
    __pyx_v_coords.shape[0], __pyx_v_coords.strides[0], __pyx_v_coords.suboffsets[0],
    0,
    0,
    &__pyx_t_13,
    (*((fsl_int *) ( /* dim=0 */ (__pyx_v_offsets.data + __pyx_t_11 * __pyx_v_offsets.strides[0]) ))),
    (*((fsl_int *) ( /* dim=0 */ (__pyx_v_offsets.data + __pyx_t_12 * __pyx_v_offsets.strides[0]) ))),
    0,
    1,
    1,
    0,
    1) < 0))
{
    __PYX_ERR(0, 1396, __pyx_L1_error)
}

__pyx_t_2.shape[1] = __pyx_v_coords.shape[1];
__pyx_t_2.strides[1] = __pyx_v_coords.strides[1];
    __pyx_t_2.suboffsets[1] = -1;

__pyx_t_9 = __pyx_memoryview_fromslice(__pyx_t_2, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
      __pyx_t_2.memview = NULL;
      __pyx_t_2.data = NULL;
      __pyx_t_14 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      __pyx_t_5 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_14, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_9);
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_15 = __Pyx_PyList_Append(__pyx_v_coords_list, __pyx_t_5); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 1396, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }

    /* "fatslimlib/core_base.pyx":1398
 *                 coords_list.append(np.asarray(coords[offsets[i]:offsets[i+1]]))
 * 
 *             self.lipid_coords_aslist = coords_list             # <<<<<<<<<<<<<<
 * 
 *         return self.lipid_coords_aslist
 */
    __Pyx_INCREF(__pyx_v_coords_list);
    __Pyx_GIVEREF(__pyx_v_coords_list);
    __Pyx_GOTREF(__pyx_v_self->lipid_coords_aslist);
    __Pyx_DECREF(__pyx_v_self->lipid_coords_aslist);
    __pyx_v_self->lipid_coords_aslist = __pyx_v_coords_list;

    /* "fatslimlib/core_base.pyx":1392
 *         cdef real[:, ::1] coords = self.fast_get_lipid_coords()
 *         cdef fsl_int i
 *         if self.lipid_coords_aslist is None:             # <<<<<<<<<<<<<<
 *             coords_list = []
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":1400
 *             self.lipid_coords_aslist = coords_list
 * 
 *         return self.lipid_coords_aslist             # <<<<<<<<<<<<<<
 * 
 *     # Associated python property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->lipid_coords_aslist);
  __pyx_r = __pyx_v_self->lipid_coords_aslist;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1388
 *         return self.lipid_coords
 * 
 *     cdef list get_lipid_coords_aslist(self):             # <<<<<<<<<<<<<<
 *         cdef fsl_int[:] offsets = self.lipid_atomids_offsets
 *         cdef real[:, ::1] coords = self.fast_get_lipid_coords()
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_1, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.get_lipid_coords_aslist", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_offsets, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_coords, 1);
  __Pyx_XDECREF(__pyx_v_coords_list);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1404
 *     # Associated python property
 *     property lipid_coords:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.get_lipid_coords_aslist()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_12lipid_coords_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_12lipid_coords_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_12lipid_coords___get__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_12lipid_coords___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":1405
 *     property lipid_coords:
 *         def __get__(self):
 *             return self.get_lipid_coords_aslist()             # <<<<<<<<<<<<<<
 * 
 *     cdef real[:, ::1] fast_get_lipid_coords_bbox(self) nogil:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->get_lipid_coords_aslist(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1404
 *     # Associated python property
 *     property lipid_coords:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.get_lipid_coords_aslist()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.lipid_coords.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1407
 *             return self.get_lipid_coords_aslist()
 * 
 *     cdef real[:, ::1] fast_get_lipid_coords_bbox(self) nogil:             # <<<<<<<<<<<<<<
 *         if self.lipid_coords_bbox is None:
 *             self.lipid_coords_bbox = self.box.fast_put_atoms_in_bbox(self.fast_get_lipid_coords())
 */

static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_lipid_coords_bbox(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_1;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };

  /* "fatslimlib/core_base.pyx":1408
 * 
 *     cdef real[:, ::1] fast_get_lipid_coords_bbox(self) nogil:
 *         if self.lipid_coords_bbox is None:             # <<<<<<<<<<<<<<
 *             self.lipid_coords_bbox = self.box.fast_put_atoms_in_bbox(self.fast_get_lipid_coords())
 *         return self.lipid_coords_bbox
 */
  if (unlikely(!__pyx_v_self->lipid_coords_bbox.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1408, __pyx_L1_error)}
  __pyx_t_1 = ((((PyObject *) __pyx_v_self->lipid_coords_bbox.memview) == Py_None) != 0);
  if (__pyx_t_1) {

    /* "fatslimlib/core_base.pyx":1409
 *     cdef real[:, ::1] fast_get_lipid_coords_bbox(self) nogil:
 *         if self.lipid_coords_bbox is None:
 *             self.lipid_coords_bbox = self.box.fast_put_atoms_in_bbox(self.fast_get_lipid_coords())             # <<<<<<<<<<<<<<
 *         return self.lipid_coords_bbox
 * 
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_lipid_coords(__pyx_v_self); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1409, __pyx_L1_error)
    __pyx_t_3 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_self->box->__pyx_vtab)->fast_put_atoms_in_bbox(__pyx_v_self->box, __pyx_t_2); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(0, 1409, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_t_2, 0);
    __pyx_t_2.memview = NULL;
    __pyx_t_2.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->lipid_coords_bbox, 0);
    __pyx_v_self->lipid_coords_bbox = __pyx_t_3;
    __pyx_t_3.memview = NULL;
    __pyx_t_3.data = NULL;

    /* "fatslimlib/core_base.pyx":1408
 * 
 *     cdef real[:, ::1] fast_get_lipid_coords_bbox(self) nogil:
 *         if self.lipid_coords_bbox is None:             # <<<<<<<<<<<<<<
 *             self.lipid_coords_bbox = self.box.fast_put_atoms_in_bbox(self.fast_get_lipid_coords())
 *         return self.lipid_coords_bbox
 */
  }

  /* "fatslimlib/core_base.pyx":1410
 *         if self.lipid_coords_bbox is None:
 *             self.lipid_coords_bbox = self.box.fast_put_atoms_in_bbox(self.fast_get_lipid_coords())
 *         return self.lipid_coords_bbox             # <<<<<<<<<<<<<<
 * 
 *     cdef list get_lipid_coords_bbox_aslist(self):
 */
  if (unlikely(!__pyx_v_self->lipid_coords_bbox.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1410, __pyx_L1_error)}
  __PYX_INC_MEMVIEW(&__pyx_v_self->lipid_coords_bbox, 1);
  __pyx_r = __pyx_v_self->lipid_coords_bbox;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1407
 *             return self.get_lipid_coords_aslist()
 * 
 *     cdef real[:, ::1] fast_get_lipid_coords_bbox(self) nogil:             # <<<<<<<<<<<<<<
 *         if self.lipid_coords_bbox is None:
 *             self.lipid_coords_bbox = self.box.fast_put_atoms_in_bbox(self.fast_get_lipid_coords())
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 0);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 0);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("fatslimlib.core_base.Frame.fast_get_lipid_coords_bbox", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1412
 *         return self.lipid_coords_bbox
 * 
 *     cdef list get_lipid_coords_bbox_aslist(self):             # <<<<<<<<<<<<<<
 *         cdef fsl_int[:] offsets = self.lipid_atomids_offsets
 *         cdef real[:, ::1] coords = self.fast_get_lipid_coords_bbox()
 */

static PyObject *__pyx_f_10fatslimlib_9core_base_5Frame_get_lipid_coords_bbox_aslist(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_offsets = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_coords = { 0, 0, { 0 }, { 0 }, { 0 } };
  fsl_int __pyx_v_i;
  PyObject *__pyx_v_coords_list = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  fsl_int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  __Pyx_RefNannySetupContext("get_lipid_coords_bbox_aslist", 0);

  /* "fatslimlib/core_base.pyx":1413
 * 
 *     cdef list get_lipid_coords_bbox_aslist(self):
 *         cdef fsl_int[:] offsets = self.lipid_atomids_offsets             # <<<<<<<<<<<<<<
 *         cdef real[:, ::1] coords = self.fast_get_lipid_coords_bbox()
 *         cdef fsl_int i
 */
  if (unlikely(!__pyx_v_self->lipid_atomids_offsets.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1413, __pyx_L1_error)}
  __pyx_t_1 = __pyx_v_self->lipid_atomids_offsets;
  __PYX_INC_MEMVIEW(&__pyx_t_1, 1);
  __pyx_v_offsets = __pyx_t_1;
  __pyx_t_1.memview = NULL;
  __pyx_t_1.data = NULL;

  /* "fatslimlib/core_base.pyx":1414
 *     cdef list get_lipid_coords_bbox_aslist(self):
 *         cdef fsl_int[:] offsets = self.lipid_atomids_offsets
 *         cdef real[:, ::1] coords = self.fast_get_lipid_coords_bbox()             # <<<<<<<<<<<<<<
 *         cdef fsl_int i
 *         if self.lipids_coords_bbox_aslist is None:
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_lipid_coords_bbox(__pyx_v_self); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1414, __pyx_L1_error)
  __pyx_v_coords = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "fatslimlib/core_base.pyx":1416
 *         cdef real[:, ::1] coords = self.fast_get_lipid_coords_bbox()
 *         cdef fsl_int i
 *         if self.lipids_coords_bbox_aslist is None:             # <<<<<<<<<<<<<<
 *             coords_list = []
 * 
 */
  __pyx_t_3 = (__pyx_v_self->lipids_coords_bbox_aslist == ((PyObject*)Py_None));
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "fatslimlib/core_base.pyx":1417
 *         cdef fsl_int i
 *         if self.lipids_coords_bbox_aslist is None:
 *             coords_list = []             # <<<<<<<<<<<<<<
 * 
 *             for i in range(offsets.shape[0] - 1):
 */
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_coords_list = ((PyObject*)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "fatslimlib/core_base.pyx":1419
 *             coords_list = []
 * 
 *             for i in range(offsets.shape[0] - 1):             # <<<<<<<<<<<<<<
 *                 coords_list.append(np.asarray(coords[offsets[i]:offsets[i+1]]))
 * 
 */
    __pyx_t_6 = ((__pyx_v_offsets.shape[0]) - 1);
    __pyx_t_7 = __pyx_t_6;
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
      __pyx_v_i = __pyx_t_8;

      /* "fatslimlib/core_base.pyx":1420
 * 
 *             for i in range(offsets.shape[0] - 1):
 *                 coords_list.append(np.asarray(coords[offsets[i]:offsets[i+1]]))             # <<<<<<<<<<<<<<
 * 
 *             self.lipids_coords_bbox_aslist = coords_list
 */
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_np); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1420, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_asarray); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1420, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_11 = __pyx_v_i;
      __pyx_t_12 = (__pyx_v_i + 1);
      __pyx_t_2.data = __pyx_v_coords.data;
      __pyx_t_2.memview = __pyx_v_coords.memview;
      __PYX_INC_MEMVIEW(&__pyx_t_2, 0);
      __pyx_t_13 = -1;
      if (unlikely(__pyx_memoryview_slice_memviewslice(
    &__pyx_t_2,
    __pyx_v_coords.shape[0], __pyx_v_coords.strides[0], __pyx_v_coords.suboffsets[0],
    0,
    0,
    &__pyx_t_13,
    (*((fsl_int *) ( /* dim=0 */ (__pyx_v_offsets.data + __pyx_t_11 * __pyx_v_offsets.strides[0]) ))),
    (*((fsl_int *) ( /* dim=0 */ (__pyx_v_offsets.data + __pyx_t_12 * __pyx_v_offsets.strides[0]) ))),
    0,
    1,
    1,
    0,
    1) < 0))
{
    __PYX_ERR(0, 1420, __pyx_L1_error)
}

__pyx_t_2.shape[1] = __pyx_v_coords.shape[1];
__pyx_t_2.strides[1] = __pyx_v_coords.strides[1];
    __pyx_t_2.suboffsets[1] = -1;

__pyx_t_9 = __pyx_memoryview_fromslice(__pyx_t_2, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1420, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
      __pyx_t_2.memview = NULL;
      __pyx_t_2.data = NULL;
      __pyx_t_14 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      __pyx_t_5 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_14, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_9);
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1420, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_15 = __Pyx_PyList_Append(__pyx_v_coords_list, __pyx_t_5); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 1420, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }

    /* "fatslimlib/core_base.pyx":1422
 *                 coords_list.append(np.asarray(coords[offsets[i]:offsets[i+1]]))
 * 
 *             self.lipids_coords_bbox_aslist = coords_list             # <<<<<<<<<<<<<<
 * 
 *         return self.lipids_coords_bbox_aslist
 */
    __Pyx_INCREF(__pyx_v_coords_list);
    __Pyx_GIVEREF(__pyx_v_coords_list);
    __Pyx_GOTREF(__pyx_v_self->lipids_coords_bbox_aslist);
    __Pyx_DECREF(__pyx_v_self->lipids_coords_bbox_aslist);
    __pyx_v_self->lipids_coords_bbox_aslist = __pyx_v_coords_list;

    /* "fatslimlib/core_base.pyx":1416
 *         cdef real[:, ::1] coords = self.fast_get_lipid_coords_bbox()
 *         cdef fsl_int i
 *         if self.lipids_coords_bbox_aslist is None:             # <<<<<<<<<<<<<<
 *             coords_list = []
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":1424
 *             self.lipids_coords_bbox_aslist = coords_list
 * 
 *         return self.lipids_coords_bbox_aslist             # <<<<<<<<<<<<<<
 * 
 *     # Associated python property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->lipids_coords_bbox_aslist);
  __pyx_r = __pyx_v_self->lipids_coords_bbox_aslist;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1412
 *         return self.lipid_coords_bbox
 * 
 *     cdef list get_lipid_coords_bbox_aslist(self):             # <<<<<<<<<<<<<<
 *         cdef fsl_int[:] offsets = self.lipid_atomids_offsets
 *         cdef real[:, ::1] coords = self.fast_get_lipid_coords_bbox()
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_1, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.get_lipid_coords_bbox_aslist", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_offsets, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_coords, 1);
  __Pyx_XDECREF(__pyx_v_coords_list);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1428
 *     # Associated python property
 *     property lipid_coords_bbox:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.get_lipid_coords_bbox_aslist()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_17lipid_coords_bbox_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_17lipid_coords_bbox_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_17lipid_coords_bbox___get__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_17lipid_coords_bbox___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":1429
 *     property lipid_coords_bbox:
 *         def __get__(self):
 *             return self.get_lipid_coords_bbox_aslist()             # <<<<<<<<<<<<<<
 * 
 *     cdef real[:,::1] fast_get_interacting_group_coords(self) nogil:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->get_lipid_coords_bbox_aslist(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1428
 *     # Associated python property
 *     property lipid_coords_bbox:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.get_lipid_coords_bbox_aslist()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.lipid_coords_bbox.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1431
 *             return self.get_lipid_coords_bbox_aslist()
 * 
 *     cdef real[:,::1] fast_get_interacting_group_coords(self) nogil:             # <<<<<<<<<<<<<<
 *         if self.interacting_group_coords is None:
 *             with gil:
 */

static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_interacting_group_coords(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("fast_get_interacting_group_coords", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "fatslimlib/core_base.pyx":1432
 * 
 *     cdef real[:,::1] fast_get_interacting_group_coords(self) nogil:
 *         if self.interacting_group_coords is None:             # <<<<<<<<<<<<<<
 *             with gil:
 *                 if self.trajectory.interacting_atomids.shape[0] > 0:
 */
  /*try:*/ {
    if (unlikely(!__pyx_v_self->interacting_group_coords.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1432, __pyx_L4_error)}
    __pyx_t_1 = ((((PyObject *) __pyx_v_self->interacting_group_coords.memview) == Py_None) != 0);
    if (__pyx_t_1) {

      /* "fatslimlib/core_base.pyx":1433
 *     cdef real[:,::1] fast_get_interacting_group_coords(self) nogil:
 *         if self.interacting_group_coords is None:
 *             with gil:             # <<<<<<<<<<<<<<
 *                 if self.trajectory.interacting_atomids.shape[0] > 0:
 *                     self.interacting_group_coords = self.coords_reader.\
 */
      {
          #ifdef WITH_THREAD
          PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
          #endif
          /*try:*/ {

            /* "fatslimlib/core_base.pyx":1434
 *         if self.interacting_group_coords is None:
 *             with gil:
 *                 if self.trajectory.interacting_atomids.shape[0] > 0:             # <<<<<<<<<<<<<<
 *                     self.interacting_group_coords = self.coords_reader.\
 *                         load_coords(self.index, self.trajectory.interacting_atomids)
 */
            if (unlikely(!__pyx_v_self->trajectory->interacting_atomids.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1434, __pyx_L8_error)}
            __pyx_t_1 = (((__pyx_v_self->trajectory->interacting_atomids.shape[0]) > 0) != 0);
            if (__pyx_t_1) {

              /* "fatslimlib/core_base.pyx":1436
 *                 if self.trajectory.interacting_atomids.shape[0] > 0:
 *                     self.interacting_group_coords = self.coords_reader.\
 *                         load_coords(self.index, self.trajectory.interacting_atomids)             # <<<<<<<<<<<<<<
 *                 else:
 *                     self.interacting_group_coords = np.empty((0, DIM))
 */
              if (unlikely(!__pyx_v_self->trajectory->interacting_atomids.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1436, __pyx_L8_error)}
              __pyx_t_2 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_CoordinateReader *)__pyx_v_self->coords_reader->__pyx_vtab)->load_coords(__pyx_v_self->coords_reader, __pyx_v_self->index, __pyx_v_self->trajectory->interacting_atomids); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1436, __pyx_L8_error)

              /* "fatslimlib/core_base.pyx":1435
 *             with gil:
 *                 if self.trajectory.interacting_atomids.shape[0] > 0:
 *                     self.interacting_group_coords = self.coords_reader.\             # <<<<<<<<<<<<<<
 *                         load_coords(self.index, self.trajectory.interacting_atomids)
 *                 else:
 */
              __PYX_XDEC_MEMVIEW(&__pyx_v_self->interacting_group_coords, 0);
              __pyx_v_self->interacting_group_coords = __pyx_t_2;
              __pyx_t_2.memview = NULL;
              __pyx_t_2.data = NULL;

              /* "fatslimlib/core_base.pyx":1434
 *         if self.interacting_group_coords is None:
 *             with gil:
 *                 if self.trajectory.interacting_atomids.shape[0] > 0:             # <<<<<<<<<<<<<<
 *                     self.interacting_group_coords = self.coords_reader.\
 *                         load_coords(self.index, self.trajectory.interacting_atomids)
 */
              goto __pyx_L10;
            }

            /* "fatslimlib/core_base.pyx":1438
 *                         load_coords(self.index, self.trajectory.interacting_atomids)
 *                 else:
 *                     self.interacting_group_coords = np.empty((0, DIM))             # <<<<<<<<<<<<<<
 *         return self.interacting_group_coords
 * 
 */
            /*else*/ {
              __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1438, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1438, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_4 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_4)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_4);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                }
              }
              __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_tuple__15) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_tuple__15);
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1438, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1438, __pyx_L8_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __PYX_XDEC_MEMVIEW(&__pyx_v_self->interacting_group_coords, 0);
              __pyx_v_self->interacting_group_coords = __pyx_t_2;
              __pyx_t_2.memview = NULL;
              __pyx_t_2.data = NULL;
            }
            __pyx_L10:;
          }

          /* "fatslimlib/core_base.pyx":1433
 *     cdef real[:,::1] fast_get_interacting_group_coords(self) nogil:
 *         if self.interacting_group_coords is None:
 *             with gil:             # <<<<<<<<<<<<<<
 *                 if self.trajectory.interacting_atomids.shape[0] > 0:
 *                     self.interacting_group_coords = self.coords_reader.\
 */
          /*finally:*/ {
            /*normal exit:*/{
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              goto __pyx_L9;
            }
            __pyx_L8_error: {
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              goto __pyx_L4_error;
            }
            __pyx_L9:;
          }
      }

      /* "fatslimlib/core_base.pyx":1432
 * 
 *     cdef real[:,::1] fast_get_interacting_group_coords(self) nogil:
 *         if self.interacting_group_coords is None:             # <<<<<<<<<<<<<<
 *             with gil:
 *                 if self.trajectory.interacting_atomids.shape[0] > 0:
 */
    }

    /* "fatslimlib/core_base.pyx":1439
 *                 else:
 *                     self.interacting_group_coords = np.empty((0, DIM))
 *         return self.interacting_group_coords             # <<<<<<<<<<<<<<
 * 
 *     # Associated python property
 */
    if (unlikely(!__pyx_v_self->interacting_group_coords.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1439, __pyx_L4_error)}
    __PYX_INC_MEMVIEW(&__pyx_v_self->interacting_group_coords, 1);
    __pyx_r = __pyx_v_self->interacting_group_coords;
    goto __pyx_L3_return;
  }

  /* "fatslimlib/core_base.pyx":1432
 * 
 *     cdef real[:,::1] fast_get_interacting_group_coords(self) nogil:
 *         if self.interacting_group_coords is None:             # <<<<<<<<<<<<<<
 *             with gil:
 *                 if self.trajectory.interacting_atomids.shape[0] > 0:
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "fatslimlib/core_base.pyx":1431
 *             return self.get_lipid_coords_bbox_aslist()
 * 
 *     cdef real[:,::1] fast_get_interacting_group_coords(self) nogil:             # <<<<<<<<<<<<<<
 *         if self.interacting_group_coords is None:
 *             with gil:
 */

  /* function exit code */
  __Pyx_pretend_to_initialize(&__pyx_r);
  goto __pyx_L0;
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 0);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.fast_get_interacting_group_coords", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1443
 *     # Associated python property
 *     property interacting_group_coords:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef real[:, ::1] memview
 *             with nogil:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_24interacting_group_coords_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_24interacting_group_coords_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_24interacting_group_coords___get__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_24interacting_group_coords___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":1445
 *         def __get__(self):
 *             cdef real[:, ::1] memview
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 memview = self.fast_get_interacting_group_coords()
 *             return np.asarray(memview)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "fatslimlib/core_base.pyx":1446
 *             cdef real[:, ::1] memview
 *             with nogil:
 *                 memview = self.fast_get_interacting_group_coords()             # <<<<<<<<<<<<<<
 *             return np.asarray(memview)
 * 
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_interacting_group_coords(__pyx_v_self); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(0, 1446, __pyx_L4_error)
        __pyx_v_memview = __pyx_t_1;
        __pyx_t_1.memview = NULL;
        __pyx_t_1.data = NULL;
      }

      /* "fatslimlib/core_base.pyx":1445
 *         def __get__(self):
 *             cdef real[:, ::1] memview
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 memview = self.fast_get_interacting_group_coords()
 *             return np.asarray(memview)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "fatslimlib/core_base.pyx":1447
 *             with nogil:
 *                 memview = self.fast_get_interacting_group_coords()
 *             return np.asarray(memview)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_asarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_memview, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1443
 *     # Associated python property
 *     property interacting_group_coords:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef real[:, ::1] memview
 *             with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.interacting_group_coords.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_memview, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1450
 * 
 * 
 *     cdef real[:,::1] fast_get_interacting_group_coords_bbox(self) nogil:             # <<<<<<<<<<<<<<
 *         cdef real[:, ::1] coords_bbox = self.fast_get_interacting_group_coords()
 *         if self.interacting_group_coords_bbox is None:
 */

static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_interacting_group_coords_bbox(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_coords_bbox = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_2;

  /* "fatslimlib/core_base.pyx":1451
 * 
 *     cdef real[:,::1] fast_get_interacting_group_coords_bbox(self) nogil:
 *         cdef real[:, ::1] coords_bbox = self.fast_get_interacting_group_coords()             # <<<<<<<<<<<<<<
 *         if self.interacting_group_coords_bbox is None:
 *             self.interacting_group_coords_bbox = self.box.fast_put_atoms_in_bbox(coords_bbox)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_interacting_group_coords(__pyx_v_self); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(0, 1451, __pyx_L1_error)
  __pyx_v_coords_bbox = __pyx_t_1;
  __pyx_t_1.memview = NULL;
  __pyx_t_1.data = NULL;

  /* "fatslimlib/core_base.pyx":1452
 *     cdef real[:,::1] fast_get_interacting_group_coords_bbox(self) nogil:
 *         cdef real[:, ::1] coords_bbox = self.fast_get_interacting_group_coords()
 *         if self.interacting_group_coords_bbox is None:             # <<<<<<<<<<<<<<
 *             self.interacting_group_coords_bbox = self.box.fast_put_atoms_in_bbox(coords_bbox)
 *         return self.interacting_group_coords_bbox
 */
  if (unlikely(!__pyx_v_self->interacting_group_coords_bbox.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1452, __pyx_L1_error)}
  __pyx_t_2 = ((((PyObject *) __pyx_v_self->interacting_group_coords_bbox.memview) == Py_None) != 0);
  if (__pyx_t_2) {

    /* "fatslimlib/core_base.pyx":1453
 *         cdef real[:, ::1] coords_bbox = self.fast_get_interacting_group_coords()
 *         if self.interacting_group_coords_bbox is None:
 *             self.interacting_group_coords_bbox = self.box.fast_put_atoms_in_bbox(coords_bbox)             # <<<<<<<<<<<<<<
 *         return self.interacting_group_coords_bbox
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_self->box->__pyx_vtab)->fast_put_atoms_in_bbox(__pyx_v_self->box, __pyx_v_coords_bbox); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(0, 1453, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->interacting_group_coords_bbox, 0);
    __pyx_v_self->interacting_group_coords_bbox = __pyx_t_1;
    __pyx_t_1.memview = NULL;
    __pyx_t_1.data = NULL;

    /* "fatslimlib/core_base.pyx":1452
 *     cdef real[:,::1] fast_get_interacting_group_coords_bbox(self) nogil:
 *         cdef real[:, ::1] coords_bbox = self.fast_get_interacting_group_coords()
 *         if self.interacting_group_coords_bbox is None:             # <<<<<<<<<<<<<<
 *             self.interacting_group_coords_bbox = self.box.fast_put_atoms_in_bbox(coords_bbox)
 *         return self.interacting_group_coords_bbox
 */
  }

  /* "fatslimlib/core_base.pyx":1454
 *         if self.interacting_group_coords_bbox is None:
 *             self.interacting_group_coords_bbox = self.box.fast_put_atoms_in_bbox(coords_bbox)
 *         return self.interacting_group_coords_bbox             # <<<<<<<<<<<<<<
 * 
 *     # Associated python property
 */
  if (unlikely(!__pyx_v_self->interacting_group_coords_bbox.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1454, __pyx_L1_error)}
  __PYX_INC_MEMVIEW(&__pyx_v_self->interacting_group_coords_bbox, 1);
  __pyx_r = __pyx_v_self->interacting_group_coords_bbox;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1450
 * 
 * 
 *     cdef real[:,::1] fast_get_interacting_group_coords_bbox(self) nogil:             # <<<<<<<<<<<<<<
 *         cdef real[:, ::1] coords_bbox = self.fast_get_interacting_group_coords()
 *         if self.interacting_group_coords_bbox is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_1, 0);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("fatslimlib.core_base.Frame.fast_get_interacting_group_coords_bbox", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_coords_bbox, 0);
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1458
 *     # Associated python property
 *     property interacting_group_coords_bbox:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef real[:, ::1] memview
 *             with nogil:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_29interacting_group_coords_bbox_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_29interacting_group_coords_bbox_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_29interacting_group_coords_bbox___get__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_29interacting_group_coords_bbox___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":1460
 *         def __get__(self):
 *             cdef real[:, ::1] memview
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 memview = self.fast_get_interacting_group_coords_bbox()
 *             return np.asarray(memview)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "fatslimlib/core_base.pyx":1461
 *             cdef real[:, ::1] memview
 *             with nogil:
 *                 memview = self.fast_get_interacting_group_coords_bbox()             # <<<<<<<<<<<<<<
 *             return np.asarray(memview)
 * 
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_interacting_group_coords_bbox(__pyx_v_self); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(0, 1461, __pyx_L4_error)
        __pyx_v_memview = __pyx_t_1;
        __pyx_t_1.memview = NULL;
        __pyx_t_1.data = NULL;
      }

      /* "fatslimlib/core_base.pyx":1460
 *         def __get__(self):
 *             cdef real[:, ::1] memview
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 memview = self.fast_get_interacting_group_coords_bbox()
 *             return np.asarray(memview)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "fatslimlib/core_base.pyx":1462
 *             with nogil:
 *                 memview = self.fast_get_interacting_group_coords_bbox()
 *             return np.asarray(memview)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_asarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_memview, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1458
 *     # Associated python property
 *     property interacting_group_coords_bbox:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef real[:, ::1] memview
 *             with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.interacting_group_coords_bbox.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_memview, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1465
 * 
 * 
 *     cdef real[:,::1] fast_get_bead_coords(self) nogil:             # <<<<<<<<<<<<<<
 *         cdef real[:, ::1] bead_coords, coords
 *         cdef real[:,::1] tmp_coords
 */

static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_bead_coords(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_bead_coords = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_coords = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_tmp_coords = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_hg_offsets = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_hg_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  fsl_int __pyx_v_i;
  fsl_int __pyx_v_j;
  fsl_int __pyx_v_residue_hg_size;
  rvec __pyx_v_xcm;
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  fsl_int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  fsl_int __pyx_t_13;
  fsl_int __pyx_t_14;
  fsl_int __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  Py_ssize_t __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  Py_ssize_t __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  Py_ssize_t __pyx_t_33;
  int __pyx_t_34;
  __Pyx_memviewslice __pyx_t_35 = { 0, 0, { 0 }, { 0 }, { 0 } };
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("fast_get_bead_coords", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "fatslimlib/core_base.pyx":1466
 * 
 *     cdef real[:,::1] fast_get_bead_coords(self) nogil:
 *         cdef real[:, ::1] bead_coords, coords             # <<<<<<<<<<<<<<
 *         cdef real[:,::1] tmp_coords
 *         cdef fsl_int[:] hg_offsets = self.trajectory.hg_bead_atomids_offsets
 */
  /*try:*/ {

    /* "fatslimlib/core_base.pyx":1468
 *         cdef real[:, ::1] bead_coords, coords
 *         cdef real[:,::1] tmp_coords
 *         cdef fsl_int[:] hg_offsets = self.trajectory.hg_bead_atomids_offsets             # <<<<<<<<<<<<<<
 *         cdef fsl_int[:] hg_indices = self.trajectory.lipid_hg_indices
 *         cdef fsl_int i, j, residue_hg_size
 */
    if (unlikely(!__pyx_v_self->trajectory->hg_bead_atomids_offsets.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1468, __pyx_L4_error)}
    __pyx_t_1 = __pyx_v_self->trajectory->hg_bead_atomids_offsets;
    __PYX_INC_MEMVIEW(&__pyx_t_1, 0);
    __pyx_v_hg_offsets = __pyx_t_1;
    __pyx_t_1.memview = NULL;
    __pyx_t_1.data = NULL;

    /* "fatslimlib/core_base.pyx":1469
 *         cdef real[:,::1] tmp_coords
 *         cdef fsl_int[:] hg_offsets = self.trajectory.hg_bead_atomids_offsets
 *         cdef fsl_int[:] hg_indices = self.trajectory.lipid_hg_indices             # <<<<<<<<<<<<<<
 *         cdef fsl_int i, j, residue_hg_size
 *         cdef rvec xcm
 */
    if (unlikely(!__pyx_v_self->trajectory->lipid_hg_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1469, __pyx_L4_error)}
    __pyx_t_1 = __pyx_v_self->trajectory->lipid_hg_indices;
    __PYX_INC_MEMVIEW(&__pyx_t_1, 0);
    __pyx_v_hg_indices = __pyx_t_1;
    __pyx_t_1.memview = NULL;
    __pyx_t_1.data = NULL;

    /* "fatslimlib/core_base.pyx":1473
 *         cdef rvec xcm
 * 
 *         if self.bead_coords is not None:             # <<<<<<<<<<<<<<
 *             return self.bead_coords
 * 
 */
    if (unlikely(!__pyx_v_self->bead_coords.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1473, __pyx_L4_error)}
    __pyx_t_2 = ((((PyObject *) __pyx_v_self->bead_coords.memview) != Py_None) != 0);
    if (__pyx_t_2) {

      /* "fatslimlib/core_base.pyx":1474
 * 
 *         if self.bead_coords is not None:
 *             return self.bead_coords             # <<<<<<<<<<<<<<
 * 
 *         # Load coords & allocate memory
 */
      if (unlikely(!__pyx_v_self->bead_coords.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1474, __pyx_L4_error)}
      __PYX_INC_MEMVIEW(&__pyx_v_self->bead_coords, 1);
      __pyx_r = __pyx_v_self->bead_coords;
      goto __pyx_L3_return;

      /* "fatslimlib/core_base.pyx":1473
 *         cdef rvec xcm
 * 
 *         if self.bead_coords is not None:             # <<<<<<<<<<<<<<
 *             return self.bead_coords
 * 
 */
    }

    /* "fatslimlib/core_base.pyx":1477
 * 
 *         # Load coords & allocate memory
 *         with gil:             # <<<<<<<<<<<<<<
 *             bead_coords = np.empty((self.trajectory.hg_bead_atomids_offsets.shape[0] - 1, DIM))
 *             tmp_coords = np.empty((self.trajectory.hg_bead_atomids_offsets.shape[0] - 1, DIM))
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "fatslimlib/core_base.pyx":1478
 *         # Load coords & allocate memory
 *         with gil:
 *             bead_coords = np.empty((self.trajectory.hg_bead_atomids_offsets.shape[0] - 1, DIM))             # <<<<<<<<<<<<<<
 *             tmp_coords = np.empty((self.trajectory.hg_bead_atomids_offsets.shape[0] - 1, DIM))
 *             coords = self.fast_get_lipid_coords()
 */
          __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1478, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1478, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_v_self->trajectory->hg_bead_atomids_offsets.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1478, __pyx_L8_error)}
          __pyx_t_4 = PyInt_FromSsize_t(((__pyx_v_self->trajectory->hg_bead_atomids_offsets.shape[0]) - 1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1478, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1478, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
          __Pyx_INCREF(__pyx_int_3);
          __Pyx_GIVEREF(__pyx_int_3);
          PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_3);
          __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1478, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1478, __pyx_L8_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_v_bead_coords = __pyx_t_7;
          __pyx_t_7.memview = NULL;
          __pyx_t_7.data = NULL;

          /* "fatslimlib/core_base.pyx":1479
 *         with gil:
 *             bead_coords = np.empty((self.trajectory.hg_bead_atomids_offsets.shape[0] - 1, DIM))
 *             tmp_coords = np.empty((self.trajectory.hg_bead_atomids_offsets.shape[0] - 1, DIM))             # <<<<<<<<<<<<<<
 *             coords = self.fast_get_lipid_coords()
 * 
 */
          __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1479, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1479, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_v_self->trajectory->hg_bead_atomids_offsets.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1479, __pyx_L8_error)}
          __pyx_t_5 = PyInt_FromSsize_t(((__pyx_v_self->trajectory->hg_bead_atomids_offsets.shape[0]) - 1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1479, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1479, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
          __Pyx_INCREF(__pyx_int_3);
          __Pyx_GIVEREF(__pyx_int_3);
          PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_3);
          __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
            }
          }
          __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1479, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(__pyx_t_3, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1479, __pyx_L8_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_v_tmp_coords = __pyx_t_7;
          __pyx_t_7.memview = NULL;
          __pyx_t_7.data = NULL;

          /* "fatslimlib/core_base.pyx":1480
 *             bead_coords = np.empty((self.trajectory.hg_bead_atomids_offsets.shape[0] - 1, DIM))
 *             tmp_coords = np.empty((self.trajectory.hg_bead_atomids_offsets.shape[0] - 1, DIM))
 *             coords = self.fast_get_lipid_coords()             # <<<<<<<<<<<<<<
 * 
 *         for i in range(hg_offsets.shape[0] - 1):
 */
          __pyx_t_7 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_lipid_coords(__pyx_v_self); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1480, __pyx_L8_error)
          __pyx_v_coords = __pyx_t_7;
          __pyx_t_7.memview = NULL;
          __pyx_t_7.data = NULL;
        }

        /* "fatslimlib/core_base.pyx":1477
 * 
 *         # Load coords & allocate memory
 *         with gil:             # <<<<<<<<<<<<<<
 *             bead_coords = np.empty((self.trajectory.hg_bead_atomids_offsets.shape[0] - 1, DIM))
 *             tmp_coords = np.empty((self.trajectory.hg_bead_atomids_offsets.shape[0] - 1, DIM))
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L9;
          }
          __pyx_L8_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L9:;
        }
    }

    /* "fatslimlib/core_base.pyx":1482
 *             coords = self.fast_get_lipid_coords()
 * 
 *         for i in range(hg_offsets.shape[0] - 1):             # <<<<<<<<<<<<<<
 *             residue_hg_size = hg_offsets[i+1] - hg_offsets[i]
 *             for j in range(residue_hg_size):
 */
    __pyx_t_8 = ((__pyx_v_hg_offsets.shape[0]) - 1);
    __pyx_t_9 = __pyx_t_8;
    for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
      __pyx_v_i = __pyx_t_10;

      /* "fatslimlib/core_base.pyx":1483
 * 
 *         for i in range(hg_offsets.shape[0] - 1):
 *             residue_hg_size = hg_offsets[i+1] - hg_offsets[i]             # <<<<<<<<<<<<<<
 *             for j in range(residue_hg_size):
 *                 tmp_coords[j, XX] = coords[hg_indices[hg_offsets[i] + j], XX]
 */
      __pyx_t_11 = (__pyx_v_i + 1);
      __pyx_t_12 = __pyx_v_i;
      __pyx_v_residue_hg_size = ((*((fsl_int *) ( /* dim=0 */ (__pyx_v_hg_offsets.data + __pyx_t_11 * __pyx_v_hg_offsets.strides[0]) ))) - (*((fsl_int *) ( /* dim=0 */ (__pyx_v_hg_offsets.data + __pyx_t_12 * __pyx_v_hg_offsets.strides[0]) ))));

      /* "fatslimlib/core_base.pyx":1484
 *         for i in range(hg_offsets.shape[0] - 1):
 *             residue_hg_size = hg_offsets[i+1] - hg_offsets[i]
 *             for j in range(residue_hg_size):             # <<<<<<<<<<<<<<
 *                 tmp_coords[j, XX] = coords[hg_indices[hg_offsets[i] + j], XX]
 *                 tmp_coords[j, YY] = coords[hg_indices[hg_offsets[i] + j], YY]
 */
      __pyx_t_13 = __pyx_v_residue_hg_size;
      __pyx_t_14 = __pyx_t_13;
      for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
        __pyx_v_j = __pyx_t_15;

        /* "fatslimlib/core_base.pyx":1485
 *             residue_hg_size = hg_offsets[i+1] - hg_offsets[i]
 *             for j in range(residue_hg_size):
 *                 tmp_coords[j, XX] = coords[hg_indices[hg_offsets[i] + j], XX]             # <<<<<<<<<<<<<<
 *                 tmp_coords[j, YY] = coords[hg_indices[hg_offsets[i] + j], YY]
 *                 tmp_coords[j, ZZ] = coords[hg_indices[hg_offsets[i] + j], ZZ]
 */
        __pyx_t_16 = __pyx_v_i;
        __pyx_t_17 = ((*((fsl_int *) ( /* dim=0 */ (__pyx_v_hg_offsets.data + __pyx_t_16 * __pyx_v_hg_offsets.strides[0]) ))) + __pyx_v_j);
        __pyx_t_18 = (*((fsl_int *) ( /* dim=0 */ (__pyx_v_hg_indices.data + __pyx_t_17 * __pyx_v_hg_indices.strides[0]) )));
        __pyx_t_19 = 0;
        __pyx_t_20 = __pyx_v_j;
        __pyx_t_21 = 0;
        *((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_tmp_coords.data + __pyx_t_20 * __pyx_v_tmp_coords.strides[0]) )) + __pyx_t_21)) )) = (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_18 * __pyx_v_coords.strides[0]) )) + __pyx_t_19)) )));

        /* "fatslimlib/core_base.pyx":1486
 *             for j in range(residue_hg_size):
 *                 tmp_coords[j, XX] = coords[hg_indices[hg_offsets[i] + j], XX]
 *                 tmp_coords[j, YY] = coords[hg_indices[hg_offsets[i] + j], YY]             # <<<<<<<<<<<<<<
 *                 tmp_coords[j, ZZ] = coords[hg_indices[hg_offsets[i] + j], ZZ]
 *             self.box.fast_pbc_xcm(tmp_coords[:residue_hg_size], xcm)
 */
        __pyx_t_22 = __pyx_v_i;
        __pyx_t_23 = ((*((fsl_int *) ( /* dim=0 */ (__pyx_v_hg_offsets.data + __pyx_t_22 * __pyx_v_hg_offsets.strides[0]) ))) + __pyx_v_j);
        __pyx_t_24 = (*((fsl_int *) ( /* dim=0 */ (__pyx_v_hg_indices.data + __pyx_t_23 * __pyx_v_hg_indices.strides[0]) )));
        __pyx_t_25 = 1;
        __pyx_t_26 = __pyx_v_j;
        __pyx_t_27 = 1;
        *((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_tmp_coords.data + __pyx_t_26 * __pyx_v_tmp_coords.strides[0]) )) + __pyx_t_27)) )) = (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_24 * __pyx_v_coords.strides[0]) )) + __pyx_t_25)) )));

        /* "fatslimlib/core_base.pyx":1487
 *                 tmp_coords[j, XX] = coords[hg_indices[hg_offsets[i] + j], XX]
 *                 tmp_coords[j, YY] = coords[hg_indices[hg_offsets[i] + j], YY]
 *                 tmp_coords[j, ZZ] = coords[hg_indices[hg_offsets[i] + j], ZZ]             # <<<<<<<<<<<<<<
 *             self.box.fast_pbc_xcm(tmp_coords[:residue_hg_size], xcm)
 *             rvec_to_memview(xcm, bead_coords[i])
 */
        __pyx_t_28 = __pyx_v_i;
        __pyx_t_29 = ((*((fsl_int *) ( /* dim=0 */ (__pyx_v_hg_offsets.data + __pyx_t_28 * __pyx_v_hg_offsets.strides[0]) ))) + __pyx_v_j);
        __pyx_t_30 = (*((fsl_int *) ( /* dim=0 */ (__pyx_v_hg_indices.data + __pyx_t_29 * __pyx_v_hg_indices.strides[0]) )));
        __pyx_t_31 = 2;
        __pyx_t_32 = __pyx_v_j;
        __pyx_t_33 = 2;
        *((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_tmp_coords.data + __pyx_t_32 * __pyx_v_tmp_coords.strides[0]) )) + __pyx_t_33)) )) = (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_30 * __pyx_v_coords.strides[0]) )) + __pyx_t_31)) )));
      }

      /* "fatslimlib/core_base.pyx":1488
 *                 tmp_coords[j, YY] = coords[hg_indices[hg_offsets[i] + j], YY]
 *                 tmp_coords[j, ZZ] = coords[hg_indices[hg_offsets[i] + j], ZZ]
 *             self.box.fast_pbc_xcm(tmp_coords[:residue_hg_size], xcm)             # <<<<<<<<<<<<<<
 *             rvec_to_memview(xcm, bead_coords[i])
 * 
 */
      __pyx_t_7.data = __pyx_v_tmp_coords.data;
      __pyx_t_7.memview = __pyx_v_tmp_coords.memview;
      __PYX_INC_MEMVIEW(&__pyx_t_7, 0);
      __pyx_t_34 = -1;
      if (unlikely(__pyx_memoryview_slice_memviewslice(
    &__pyx_t_7,
    __pyx_v_tmp_coords.shape[0], __pyx_v_tmp_coords.strides[0], __pyx_v_tmp_coords.suboffsets[0],
    0,
    0,
    &__pyx_t_34,
    0,
    __pyx_v_residue_hg_size,
    0,
    0,
    1,
    0,
    1) < 0))
{
    __PYX_ERR(0, 1488, __pyx_L4_error)
}

__pyx_t_7.shape[1] = __pyx_v_tmp_coords.shape[1];
__pyx_t_7.strides[1] = __pyx_v_tmp_coords.strides[1];
    __pyx_t_7.suboffsets[1] = -1;

((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_self->box->__pyx_vtab)->fast_pbc_xcm(__pyx_v_self->box, __pyx_t_7, __pyx_v_xcm, NULL);
      __PYX_XDEC_MEMVIEW(&__pyx_t_7, 0);
      __pyx_t_7.memview = NULL;
      __pyx_t_7.data = NULL;

      /* "fatslimlib/core_base.pyx":1489
 *                 tmp_coords[j, ZZ] = coords[hg_indices[hg_offsets[i] + j], ZZ]
 *             self.box.fast_pbc_xcm(tmp_coords[:residue_hg_size], xcm)
 *             rvec_to_memview(xcm, bead_coords[i])             # <<<<<<<<<<<<<<
 * 
 *         self.bead_coords = bead_coords
 */
      __pyx_t_35.data = __pyx_v_bead_coords.data;
      __pyx_t_35.memview = __pyx_v_bead_coords.memview;
      __PYX_INC_MEMVIEW(&__pyx_t_35, 0);
      {
    Py_ssize_t __pyx_tmp_idx = __pyx_v_i;
    Py_ssize_t __pyx_tmp_stride = __pyx_v_bead_coords.strides[0];
        if ((0)) __PYX_ERR(0, 1489, __pyx_L4_error)
        __pyx_t_35.data += __pyx_tmp_idx * __pyx_tmp_stride;
}

__pyx_t_35.shape[0] = __pyx_v_bead_coords.shape[1];
__pyx_t_35.strides[0] = __pyx_v_bead_coords.strides[1];
    __pyx_t_35.suboffsets[0] = -1;

__pyx_f_10fatslimlib_9core_base_rvec_to_memview(__pyx_v_xcm, __pyx_t_35);
      __PYX_XDEC_MEMVIEW(&__pyx_t_35, 0);
      __pyx_t_35.memview = NULL;
      __pyx_t_35.data = NULL;
    }

    /* "fatslimlib/core_base.pyx":1491
 *             rvec_to_memview(xcm, bead_coords[i])
 * 
 *         self.bead_coords = bead_coords             # <<<<<<<<<<<<<<
 *         return bead_coords
 * 
 */
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->bead_coords, 0);
    __PYX_INC_MEMVIEW(&__pyx_v_bead_coords, 1);
    __pyx_v_self->bead_coords = __pyx_v_bead_coords;

    /* "fatslimlib/core_base.pyx":1492
 * 
 *         self.bead_coords = bead_coords
 *         return bead_coords             # <<<<<<<<<<<<<<
 * 
 *     # Associated python property
 */
    __PYX_INC_MEMVIEW(&__pyx_v_bead_coords, 1);
    __pyx_r = __pyx_v_bead_coords;
    goto __pyx_L3_return;
  }

  /* "fatslimlib/core_base.pyx":1466
 * 
 *     cdef real[:,::1] fast_get_bead_coords(self) nogil:
 *         cdef real[:, ::1] bead_coords, coords             # <<<<<<<<<<<<<<
 *         cdef real[:,::1] tmp_coords
 *         cdef fsl_int[:] hg_offsets = self.trajectory.hg_bead_atomids_offsets
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "fatslimlib/core_base.pyx":1465
 * 
 * 
 *     cdef real[:,::1] fast_get_bead_coords(self) nogil:             # <<<<<<<<<<<<<<
 *         cdef real[:, ::1] bead_coords, coords
 *         cdef real[:,::1] tmp_coords
 */

  /* function exit code */
  __Pyx_pretend_to_initialize(&__pyx_r);
  goto __pyx_L0;
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_1, 0);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 0);
  __PYX_XDEC_MEMVIEW(&__pyx_t_35, 0);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.fast_get_bead_coords", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_bead_coords, 0);
  __PYX_XDEC_MEMVIEW(&__pyx_v_coords, 0);
  __PYX_XDEC_MEMVIEW(&__pyx_v_tmp_coords, 0);
  __PYX_XDEC_MEMVIEW(&__pyx_v_hg_offsets, 0);
  __PYX_XDEC_MEMVIEW(&__pyx_v_hg_indices, 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1496
 *     # Associated python property
 *     property bead_coords:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef real[:, ::1] memview
 *             with nogil:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_11bead_coords_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_11bead_coords_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_11bead_coords___get__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_11bead_coords___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":1498
 *         def __get__(self):
 *             cdef real[:, ::1] memview
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 memview = self.fast_get_bead_coords()
 *             return np.asarray(memview)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "fatslimlib/core_base.pyx":1499
 *             cdef real[:, ::1] memview
 *             with nogil:
 *                 memview = self.fast_get_bead_coords()             # <<<<<<<<<<<<<<
 *             return np.asarray(memview)
 * 
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_bead_coords(__pyx_v_self); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(0, 1499, __pyx_L4_error)
        __pyx_v_memview = __pyx_t_1;
        __pyx_t_1.memview = NULL;
        __pyx_t_1.data = NULL;
      }

      /* "fatslimlib/core_base.pyx":1498
 *         def __get__(self):
 *             cdef real[:, ::1] memview
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 memview = self.fast_get_bead_coords()
 *             return np.asarray(memview)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "fatslimlib/core_base.pyx":1500
 *             with nogil:
 *                 memview = self.fast_get_bead_coords()
 *             return np.asarray(memview)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_asarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_memview, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1496
 *     # Associated python property
 *     property bead_coords:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef real[:, ::1] memview
 *             with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.bead_coords.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_memview, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1503
 * 
 * 
 *     cdef real[:,::1] fast_get_bead_coords_bbox(self) nogil:             # <<<<<<<<<<<<<<
 *         if self.bead_coords_bbox is None:
 *             self.bead_coords_bbox = self.box.fast_put_atoms_in_bbox(self.fast_get_bead_coords())
 */

static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_bead_coords_bbox(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_1;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_3 = { 0, 0, { 0 }, { 0 }, { 0 } };

  /* "fatslimlib/core_base.pyx":1504
 * 
 *     cdef real[:,::1] fast_get_bead_coords_bbox(self) nogil:
 *         if self.bead_coords_bbox is None:             # <<<<<<<<<<<<<<
 *             self.bead_coords_bbox = self.box.fast_put_atoms_in_bbox(self.fast_get_bead_coords())
 *         return self.bead_coords_bbox
 */
  if (unlikely(!__pyx_v_self->bead_coords_bbox.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1504, __pyx_L1_error)}
  __pyx_t_1 = ((((PyObject *) __pyx_v_self->bead_coords_bbox.memview) == Py_None) != 0);
  if (__pyx_t_1) {

    /* "fatslimlib/core_base.pyx":1505
 *     cdef real[:,::1] fast_get_bead_coords_bbox(self) nogil:
 *         if self.bead_coords_bbox is None:
 *             self.bead_coords_bbox = self.box.fast_put_atoms_in_bbox(self.fast_get_bead_coords())             # <<<<<<<<<<<<<<
 *         return self.bead_coords_bbox
 * 
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_bead_coords(__pyx_v_self); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1505, __pyx_L1_error)
    __pyx_t_3 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox *)__pyx_v_self->box->__pyx_vtab)->fast_put_atoms_in_bbox(__pyx_v_self->box, __pyx_t_2); if (unlikely(!__pyx_t_3.memview)) __PYX_ERR(0, 1505, __pyx_L1_error)
    __PYX_XDEC_MEMVIEW(&__pyx_t_2, 0);
    __pyx_t_2.memview = NULL;
    __pyx_t_2.data = NULL;
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->bead_coords_bbox, 0);
    __pyx_v_self->bead_coords_bbox = __pyx_t_3;
    __pyx_t_3.memview = NULL;
    __pyx_t_3.data = NULL;

    /* "fatslimlib/core_base.pyx":1504
 * 
 *     cdef real[:,::1] fast_get_bead_coords_bbox(self) nogil:
 *         if self.bead_coords_bbox is None:             # <<<<<<<<<<<<<<
 *             self.bead_coords_bbox = self.box.fast_put_atoms_in_bbox(self.fast_get_bead_coords())
 *         return self.bead_coords_bbox
 */
  }

  /* "fatslimlib/core_base.pyx":1506
 *         if self.bead_coords_bbox is None:
 *             self.bead_coords_bbox = self.box.fast_put_atoms_in_bbox(self.fast_get_bead_coords())
 *         return self.bead_coords_bbox             # <<<<<<<<<<<<<<
 * 
 *     # Associated python property
 */
  if (unlikely(!__pyx_v_self->bead_coords_bbox.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1506, __pyx_L1_error)}
  __PYX_INC_MEMVIEW(&__pyx_v_self->bead_coords_bbox, 1);
  __pyx_r = __pyx_v_self->bead_coords_bbox;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1503
 * 
 * 
 *     cdef real[:,::1] fast_get_bead_coords_bbox(self) nogil:             # <<<<<<<<<<<<<<
 *         if self.bead_coords_bbox is None:
 *             self.bead_coords_bbox = self.box.fast_put_atoms_in_bbox(self.fast_get_bead_coords())
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 0);
  __PYX_XDEC_MEMVIEW(&__pyx_t_3, 0);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("fatslimlib.core_base.Frame.fast_get_bead_coords_bbox", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1510
 *     # Associated python property
 *     property bead_coords_bbox:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef real[:, ::1] memview
 *             with nogil:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_16bead_coords_bbox_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_16bead_coords_bbox_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_16bead_coords_bbox___get__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_16bead_coords_bbox___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":1512
 *         def __get__(self):
 *             cdef real[:, ::1] memview
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 memview = self.fast_get_bead_coords_bbox()
 *             return np.asarray(memview)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "fatslimlib/core_base.pyx":1513
 *             cdef real[:, ::1] memview
 *             with nogil:
 *                 memview = self.fast_get_bead_coords_bbox()             # <<<<<<<<<<<<<<
 *             return np.asarray(memview)
 * 
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_bead_coords_bbox(__pyx_v_self); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(0, 1513, __pyx_L4_error)
        __pyx_v_memview = __pyx_t_1;
        __pyx_t_1.memview = NULL;
        __pyx_t_1.data = NULL;
      }

      /* "fatslimlib/core_base.pyx":1512
 *         def __get__(self):
 *             cdef real[:, ::1] memview
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 memview = self.fast_get_bead_coords_bbox()
 *             return np.asarray(memview)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "fatslimlib/core_base.pyx":1514
 *             with nogil:
 *                 memview = self.fast_get_bead_coords_bbox()
 *             return np.asarray(memview)             # <<<<<<<<<<<<<<
 * 
 *     cdef real[:, ::1] fast_get_directions(self) nogil except *:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_asarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_memview, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1510
 *     # Associated python property
 *     property bead_coords_bbox:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef real[:, ::1] memview
 *             with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.bead_coords_bbox.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_memview, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1516
 *             return np.asarray(memview)
 * 
 *     cdef real[:, ::1] fast_get_directions(self) nogil except *:             # <<<<<<<<<<<<<<
 *         if self.directions is not None:
 *             return self.directions
 */

static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_directions(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  fsl_int __pyx_v_i;
  __Pyx_memviewslice __pyx_v_bead_coords = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_all_lipid_coords = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lipid_offsets = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lipid_atomids = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_hg_offsets = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_hg_atomids = { 0, 0, { 0 }, { 0 }, { 0 } };
  struct __pyx_t_10fatslimlib_9core_base_cPBCBox_t __pyx_v_box;
  __Pyx_memviewslice __pyx_v_directions = { 0, 0, { 0 }, { 0 }, { 0 } };
  fsl_int __pyx_v_start_offset;
  fsl_int __pyx_v_first_lipid_atomid;
  fsl_int __pyx_v_lipid_offset;
  fsl_int __pyx_v_hg_offset;
  fsl_int __pyx_v_last_hg_atomid;
  Py_ssize_t __pyx_v_nbeads;
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  struct __pyx_t_10fatslimlib_9core_base_cPBCBox_t __pyx_t_3;
  __Pyx_memviewslice __pyx_t_4 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("fast_get_directions", 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "fatslimlib/core_base.pyx":1517
 * 
 *     cdef real[:, ::1] fast_get_directions(self) nogil except *:
 *         if self.directions is not None:             # <<<<<<<<<<<<<<
 *             return self.directions
 * 
 */
  /*try:*/ {
    if (unlikely(!__pyx_v_self->directions.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1517, __pyx_L4_error)}
    __pyx_t_1 = ((((PyObject *) __pyx_v_self->directions.memview) != Py_None) != 0);
    if (__pyx_t_1) {

      /* "fatslimlib/core_base.pyx":1518
 *     cdef real[:, ::1] fast_get_directions(self) nogil except *:
 *         if self.directions is not None:
 *             return self.directions             # <<<<<<<<<<<<<<
 * 
 *         cdef fsl_int i
 */
      if (unlikely(!__pyx_v_self->directions.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1518, __pyx_L4_error)}
      __PYX_INC_MEMVIEW(&__pyx_v_self->directions, 1);
      __pyx_r = __pyx_v_self->directions;
      goto __pyx_L3_return;

      /* "fatslimlib/core_base.pyx":1517
 * 
 *     cdef real[:, ::1] fast_get_directions(self) nogil except *:
 *         if self.directions is not None:             # <<<<<<<<<<<<<<
 *             return self.directions
 * 
 */
    }

    /* "fatslimlib/core_base.pyx":1524
 *         cdef fsl_int beads
 *         cdef real[:,::1] lipid_coords, all_lipid_coords
 *         cdef fsl_int[:] lipid_offsets = self.lipid_atomids_offsets             # <<<<<<<<<<<<<<
 *         cdef fsl_int[:] lipid_atomids = self.trajectory.lipid_atomids
 *         cdef fsl_int[:] hg_offsets = self.trajectory.hg_bead_atomids_offsets
 */
    if (unlikely(!__pyx_v_self->lipid_atomids_offsets.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1524, __pyx_L4_error)}
    __pyx_t_2 = __pyx_v_self->lipid_atomids_offsets;
    __PYX_INC_MEMVIEW(&__pyx_t_2, 0);
    __pyx_v_lipid_offsets = __pyx_t_2;
    __pyx_t_2.memview = NULL;
    __pyx_t_2.data = NULL;

    /* "fatslimlib/core_base.pyx":1525
 *         cdef real[:,::1] lipid_coords, all_lipid_coords
 *         cdef fsl_int[:] lipid_offsets = self.lipid_atomids_offsets
 *         cdef fsl_int[:] lipid_atomids = self.trajectory.lipid_atomids             # <<<<<<<<<<<<<<
 *         cdef fsl_int[:] hg_offsets = self.trajectory.hg_bead_atomids_offsets
 *         cdef fsl_int[:] hg_atomids = self.trajectory.hg_bead_atomids
 */
    if (unlikely(!__pyx_v_self->trajectory->lipid_atomids.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1525, __pyx_L4_error)}
    __pyx_t_2 = __pyx_v_self->trajectory->lipid_atomids;
    __PYX_INC_MEMVIEW(&__pyx_t_2, 0);
    __pyx_v_lipid_atomids = __pyx_t_2;
    __pyx_t_2.memview = NULL;
    __pyx_t_2.data = NULL;

    /* "fatslimlib/core_base.pyx":1526
 *         cdef fsl_int[:] lipid_offsets = self.lipid_atomids_offsets
 *         cdef fsl_int[:] lipid_atomids = self.trajectory.lipid_atomids
 *         cdef fsl_int[:] hg_offsets = self.trajectory.hg_bead_atomids_offsets             # <<<<<<<<<<<<<<
 *         cdef fsl_int[:] hg_atomids = self.trajectory.hg_bead_atomids
 *         cdef cPBCBox_t box = self.box.c_pbcbox
 */
    if (unlikely(!__pyx_v_self->trajectory->hg_bead_atomids_offsets.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1526, __pyx_L4_error)}
    __pyx_t_2 = __pyx_v_self->trajectory->hg_bead_atomids_offsets;
    __PYX_INC_MEMVIEW(&__pyx_t_2, 0);
    __pyx_v_hg_offsets = __pyx_t_2;
    __pyx_t_2.memview = NULL;
    __pyx_t_2.data = NULL;

    /* "fatslimlib/core_base.pyx":1527
 *         cdef fsl_int[:] lipid_atomids = self.trajectory.lipid_atomids
 *         cdef fsl_int[:] hg_offsets = self.trajectory.hg_bead_atomids_offsets
 *         cdef fsl_int[:] hg_atomids = self.trajectory.hg_bead_atomids             # <<<<<<<<<<<<<<
 *         cdef cPBCBox_t box = self.box.c_pbcbox
 *         cdef real[:, ::1] directions
 */
    if (unlikely(!__pyx_v_self->trajectory->hg_bead_atomids.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1527, __pyx_L4_error)}
    __pyx_t_2 = __pyx_v_self->trajectory->hg_bead_atomids;
    __PYX_INC_MEMVIEW(&__pyx_t_2, 0);
    __pyx_v_hg_atomids = __pyx_t_2;
    __pyx_t_2.memview = NULL;
    __pyx_t_2.data = NULL;

    /* "fatslimlib/core_base.pyx":1528
 *         cdef fsl_int[:] hg_offsets = self.trajectory.hg_bead_atomids_offsets
 *         cdef fsl_int[:] hg_atomids = self.trajectory.hg_bead_atomids
 *         cdef cPBCBox_t box = self.box.c_pbcbox             # <<<<<<<<<<<<<<
 *         cdef real[:, ::1] directions
 *         cdef fsl_int start_offset, first_lipid_atomid, lipid_offset, hg_offset, last_hg_atomid
 */
    __pyx_t_3 = __pyx_v_self->box->c_pbcbox;
    __pyx_v_box = __pyx_t_3;

    /* "fatslimlib/core_base.pyx":1533
 * 
 *         # Loads coordinates
 *         all_lipid_coords = self.fast_get_lipid_coords_bbox()             # <<<<<<<<<<<<<<
 *         bead_coords = self.fast_get_bead_coords_bbox()
 *         nbeads = bead_coords.shape[0]
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_lipid_coords_bbox(__pyx_v_self); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(0, 1533, __pyx_L4_error)
    __pyx_v_all_lipid_coords = __pyx_t_4;
    __pyx_t_4.memview = NULL;
    __pyx_t_4.data = NULL;

    /* "fatslimlib/core_base.pyx":1534
 *         # Loads coordinates
 *         all_lipid_coords = self.fast_get_lipid_coords_bbox()
 *         bead_coords = self.fast_get_bead_coords_bbox()             # <<<<<<<<<<<<<<
 *         nbeads = bead_coords.shape[0]
 * 
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_bead_coords_bbox(__pyx_v_self); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(0, 1534, __pyx_L4_error)
    __pyx_v_bead_coords = __pyx_t_4;
    __pyx_t_4.memview = NULL;
    __pyx_t_4.data = NULL;

    /* "fatslimlib/core_base.pyx":1535
 *         all_lipid_coords = self.fast_get_lipid_coords_bbox()
 *         bead_coords = self.fast_get_bead_coords_bbox()
 *         nbeads = bead_coords.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         with gil:
 */
    __pyx_v_nbeads = (__pyx_v_bead_coords.shape[0]);

    /* "fatslimlib/core_base.pyx":1537
 *         nbeads = bead_coords.shape[0]
 * 
 *         with gil:             # <<<<<<<<<<<<<<
 *             directions = np.empty((nbeads, DIM))
 * 
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "fatslimlib/core_base.pyx":1538
 * 
 *         with gil:
 *             directions = np.empty((nbeads, DIM))             # <<<<<<<<<<<<<<
 * 
 *         #for i in range(nbeads):
 */
          __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1538, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_empty); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1538, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_nbeads); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1538, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1538, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_6);
          PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6);
          __Pyx_INCREF(__pyx_int_3);
          __Pyx_GIVEREF(__pyx_int_3);
          PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_int_3);
          __pyx_t_6 = 0;
          __pyx_t_6 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
            }
          }
          __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8);
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1538, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_4 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(0, 1538, __pyx_L8_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_v_directions = __pyx_t_4;
          __pyx_t_4.memview = NULL;
          __pyx_t_4.data = NULL;
        }

        /* "fatslimlib/core_base.pyx":1537
 *         nbeads = bead_coords.shape[0]
 * 
 *         with gil:             # <<<<<<<<<<<<<<
 *             directions = np.empty((nbeads, DIM))
 * 
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L9;
          }
          __pyx_L8_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L9:;
        }
    }

    /* "fatslimlib/core_base.pyx":1541
 * 
 *         #for i in range(nbeads):
 *         for i in prange(nbeads, schedule='dynamic', num_threads=OPENMP_NUM_THREADS):             # <<<<<<<<<<<<<<
 *             lipid_offset = lipid_offsets[i]
 *             first_lipid_atomid = lipid_atomids[lipid_offset]
 */
    __pyx_t_9 = __pyx_v_nbeads;
    if (1 == 0) abort();
    {
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
            #undef likely
            #undef unlikely
            #define likely(x)   (x)
            #define unlikely(x) (x)
        #endif
        __pyx_t_11 = (__pyx_t_9 - 0 + 1 - 1/abs(1)) / 1;
        if (__pyx_t_11 > 0)
        {
            #ifdef _OPENMP
            #pragma omp parallel num_threads(__pyx_v_10fatslimlib_9core_base_OPENMP_NUM_THREADS) private(__pyx_t_12, __pyx_t_13, __pyx_t_14, __pyx_t_15, __pyx_t_16, __pyx_t_17, __pyx_t_18, __pyx_t_19, __pyx_t_20)
            #endif /* _OPENMP */
            {
                #ifdef _OPENMP
                #pragma omp for lastprivate(__pyx_v_first_lipid_atomid) lastprivate(__pyx_v_hg_offset) firstprivate(__pyx_v_i) lastprivate(__pyx_v_i) lastprivate(__pyx_v_last_hg_atomid) lastprivate(__pyx_v_lipid_offset) lastprivate(__pyx_v_start_offset) schedule(dynamic)
                #endif /* _OPENMP */
                for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_11; __pyx_t_10++){
                    {
                        __pyx_v_i = (fsl_int)(0 + 1 * __pyx_t_10);
                        /* Initialize private variables to invalid values */
                        __pyx_v_first_lipid_atomid = ((fsl_int)0xbad0bad0);
                        __pyx_v_hg_offset = ((fsl_int)0xbad0bad0);
                        __pyx_v_last_hg_atomid = ((fsl_int)0xbad0bad0);
                        __pyx_v_lipid_offset = ((fsl_int)0xbad0bad0);
                        __pyx_v_start_offset = ((fsl_int)0xbad0bad0);

                        /* "fatslimlib/core_base.pyx":1542
 *         #for i in range(nbeads):
 *         for i in prange(nbeads, schedule='dynamic', num_threads=OPENMP_NUM_THREADS):
 *             lipid_offset = lipid_offsets[i]             # <<<<<<<<<<<<<<
 *             first_lipid_atomid = lipid_atomids[lipid_offset]
 *             hg_offset = hg_offsets[i+1]
 */
                        __pyx_t_12 = __pyx_v_i;
                        __pyx_v_lipid_offset = (*((fsl_int *) ( /* dim=0 */ (__pyx_v_lipid_offsets.data + __pyx_t_12 * __pyx_v_lipid_offsets.strides[0]) )));

                        /* "fatslimlib/core_base.pyx":1543
 *         for i in prange(nbeads, schedule='dynamic', num_threads=OPENMP_NUM_THREADS):
 *             lipid_offset = lipid_offsets[i]
 *             first_lipid_atomid = lipid_atomids[lipid_offset]             # <<<<<<<<<<<<<<
 *             hg_offset = hg_offsets[i+1]
 *             last_hg_atomid = hg_atomids[hg_offset -1]
 */
                        __pyx_t_13 = __pyx_v_lipid_offset;
                        __pyx_v_first_lipid_atomid = (*((fsl_int *) ( /* dim=0 */ (__pyx_v_lipid_atomids.data + __pyx_t_13 * __pyx_v_lipid_atomids.strides[0]) )));

                        /* "fatslimlib/core_base.pyx":1544
 *             lipid_offset = lipid_offsets[i]
 *             first_lipid_atomid = lipid_atomids[lipid_offset]
 *             hg_offset = hg_offsets[i+1]             # <<<<<<<<<<<<<<
 *             last_hg_atomid = hg_atomids[hg_offset -1]
 *             start_offset = lipid_offset + (last_hg_atomid - first_lipid_atomid)
 */
                        __pyx_t_14 = (__pyx_v_i + 1);
                        __pyx_v_hg_offset = (*((fsl_int *) ( /* dim=0 */ (__pyx_v_hg_offsets.data + __pyx_t_14 * __pyx_v_hg_offsets.strides[0]) )));

                        /* "fatslimlib/core_base.pyx":1545
 *             first_lipid_atomid = lipid_atomids[lipid_offset]
 *             hg_offset = hg_offsets[i+1]
 *             last_hg_atomid = hg_atomids[hg_offset -1]             # <<<<<<<<<<<<<<
 *             start_offset = lipid_offset + (last_hg_atomid - first_lipid_atomid)
 * 
 */
                        __pyx_t_15 = (__pyx_v_hg_offset - 1);
                        __pyx_v_last_hg_atomid = (*((fsl_int *) ( /* dim=0 */ (__pyx_v_hg_atomids.data + __pyx_t_15 * __pyx_v_hg_atomids.strides[0]) )));

                        /* "fatslimlib/core_base.pyx":1546
 *             hg_offset = hg_offsets[i+1]
 *             last_hg_atomid = hg_atomids[hg_offset -1]
 *             start_offset = lipid_offset + (last_hg_atomid - first_lipid_atomid)             # <<<<<<<<<<<<<<
 * 
 *             calculate_lipid_direction(&bead_coords[i, XX],
 */
                        __pyx_v_start_offset = (__pyx_v_lipid_offset + (__pyx_v_last_hg_atomid - __pyx_v_first_lipid_atomid));

                        /* "fatslimlib/core_base.pyx":1548
 *             start_offset = lipid_offset + (last_hg_atomid - first_lipid_atomid)
 * 
 *             calculate_lipid_direction(&bead_coords[i, XX],             # <<<<<<<<<<<<<<
 *                                       box,
 *                                       all_lipid_coords,
 */
                        __pyx_t_16 = __pyx_v_i;
                        __pyx_t_17 = 0;

                        /* "fatslimlib/core_base.pyx":1552
 *                                       all_lipid_coords,
 *                                       start_offset,
 *                                       lipid_offsets[i+1],             # <<<<<<<<<<<<<<
 *                                       &directions[i, XX])
 * 
 */
                        __pyx_t_18 = (__pyx_v_i + 1);

                        /* "fatslimlib/core_base.pyx":1553
 *                                       start_offset,
 *                                       lipid_offsets[i+1],
 *                                       &directions[i, XX])             # <<<<<<<<<<<<<<
 * 
 *         with gil:
 */
                        __pyx_t_19 = __pyx_v_i;
                        __pyx_t_20 = 0;

                        /* "fatslimlib/core_base.pyx":1548
 *             start_offset = lipid_offset + (last_hg_atomid - first_lipid_atomid)
 * 
 *             calculate_lipid_direction(&bead_coords[i, XX],             # <<<<<<<<<<<<<<
 *                                       box,
 *                                       all_lipid_coords,
 */
                        __pyx_f_10fatslimlib_9core_base_calculate_lipid_direction((&(*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_bead_coords.data + __pyx_t_16 * __pyx_v_bead_coords.strides[0]) )) + __pyx_t_17)) )))), __pyx_v_box, __pyx_v_all_lipid_coords, __pyx_v_start_offset, (*((fsl_int *) ( /* dim=0 */ (__pyx_v_lipid_offsets.data + __pyx_t_18 * __pyx_v_lipid_offsets.strides[0]) ))), (&(*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_directions.data + __pyx_t_19 * __pyx_v_directions.strides[0]) )) + __pyx_t_20)) )))));
                    }
                }
            }
        }
    }
    #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
        #undef likely
        #undef unlikely
        #define likely(x)   __builtin_expect(!!(x), 1)
        #define unlikely(x) __builtin_expect(!!(x), 0)
    #endif

    /* "fatslimlib/core_base.pyx":1555
 *                                       &directions[i, XX])
 * 
 *         with gil:             # <<<<<<<<<<<<<<
 *             self.directions = directions
 *         return directions
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "fatslimlib/core_base.pyx":1556
 * 
 *         with gil:
 *             self.directions = directions             # <<<<<<<<<<<<<<
 *         return directions
 * 
 */
          __PYX_XDEC_MEMVIEW(&__pyx_v_self->directions, 0);
          __PYX_INC_MEMVIEW(&__pyx_v_directions, 0);
          __pyx_v_self->directions = __pyx_v_directions;
        }

        /* "fatslimlib/core_base.pyx":1555
 *                                       &directions[i, XX])
 * 
 *         with gil:             # <<<<<<<<<<<<<<
 *             self.directions = directions
 *         return directions
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L18;
          }
          __pyx_L18:;
        }
    }

    /* "fatslimlib/core_base.pyx":1557
 *         with gil:
 *             self.directions = directions
 *         return directions             # <<<<<<<<<<<<<<
 * 
 *     # Associated python property
 */
    __PYX_INC_MEMVIEW(&__pyx_v_directions, 1);
    __pyx_r = __pyx_v_directions;
    goto __pyx_L3_return;
  }

  /* "fatslimlib/core_base.pyx":1517
 * 
 *     cdef real[:, ::1] fast_get_directions(self) nogil except *:
 *         if self.directions is not None:             # <<<<<<<<<<<<<<
 *             return self.directions
 * 
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "fatslimlib/core_base.pyx":1516
 *             return np.asarray(memview)
 * 
 *     cdef real[:, ::1] fast_get_directions(self) nogil except *:             # <<<<<<<<<<<<<<
 *         if self.directions is not None:
 *             return self.directions
 */

  /* function exit code */
  __Pyx_pretend_to_initialize(&__pyx_r);
  goto __pyx_L0;
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 0);
  __PYX_XDEC_MEMVIEW(&__pyx_t_4, 0);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.fast_get_directions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_bead_coords, 0);
  __PYX_XDEC_MEMVIEW(&__pyx_v_all_lipid_coords, 0);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lipid_offsets, 0);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lipid_atomids, 0);
  __PYX_XDEC_MEMVIEW(&__pyx_v_hg_offsets, 0);
  __PYX_XDEC_MEMVIEW(&__pyx_v_hg_atomids, 0);
  __PYX_XDEC_MEMVIEW(&__pyx_v_directions, 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1561
 *     # Associated python property
 *     property directions:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef real[:, ::1] memview
 *             with nogil:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_10directions_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_10directions_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_10directions___get__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_10directions___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":1563
 *         def __get__(self):
 *             cdef real[:, ::1] memview
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 memview = self.fast_get_directions()
 *             return np.asarray(memview)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "fatslimlib/core_base.pyx":1564
 *             cdef real[:, ::1] memview
 *             with nogil:
 *                 memview = self.fast_get_directions()             # <<<<<<<<<<<<<<
 *             return np.asarray(memview)
 * 
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_directions(__pyx_v_self); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(0, 1564, __pyx_L4_error)
        __pyx_v_memview = __pyx_t_1;
        __pyx_t_1.memview = NULL;
        __pyx_t_1.data = NULL;
      }

      /* "fatslimlib/core_base.pyx":1563
 *         def __get__(self):
 *             cdef real[:, ::1] memview
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 memview = self.fast_get_directions()
 *             return np.asarray(memview)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "fatslimlib/core_base.pyx":1565
 *             with nogil:
 *                 memview = self.fast_get_directions()
 *             return np.asarray(memview)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_asarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_memview, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1561
 *     # Associated python property
 *     property directions:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef real[:, ::1] memview
 *             with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.directions.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_memview, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1568
 * 
 * 
 *     cdef real[:, ::1] fast_get_normals(self, real proximity_cutoff=-1) nogil:             # <<<<<<<<<<<<<<
 *         cdef fsl_int ncoords = self.size
 *         cdef real[:, ::1] coords = self.fast_get_bead_coords_bbox()
 */

static __Pyx_memviewslice __pyx_f_10fatslimlib_9core_base_5Frame_fast_get_normals(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self, struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_fast_get_normals *__pyx_optional_args) {
  real __pyx_v_proximity_cutoff = ((real)-1.0);
  fsl_int __pyx_v_ncoords;
  __Pyx_memviewslice __pyx_v_coords = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_directions = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_normals = { 0, 0, { 0 }, { 0 }, { 0 } };
  fsl_int *__pyx_v_neighborhood_buffer;
  fsl_int __pyx_v_buffer_size;
  fsl_int __pyx_v_nid;
  fsl_int __pyx_v_i;
  __Pyx_memviewslice __pyx_r = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_RefNannyDeclarations
  fsl_int __pyx_t_1;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_3;
  int __pyx_t_4;
  real __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  fsl_int __pyx_t_10;
  fsl_int __pyx_t_11;
  fsl_int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("fast_get_normals", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_proximity_cutoff = __pyx_optional_args->proximity_cutoff;
    }
  }
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif

  /* "fatslimlib/core_base.pyx":1569
 * 
 *     cdef real[:, ::1] fast_get_normals(self, real proximity_cutoff=-1) nogil:
 *         cdef fsl_int ncoords = self.size             # <<<<<<<<<<<<<<
 *         cdef real[:, ::1] coords = self.fast_get_bead_coords_bbox()
 *         cdef real[:, ::1] directions = self.fast_get_directions()
 */
  /*try:*/ {
    __pyx_t_1 = __pyx_v_self->size;
    __pyx_v_ncoords = __pyx_t_1;

    /* "fatslimlib/core_base.pyx":1570
 *     cdef real[:, ::1] fast_get_normals(self, real proximity_cutoff=-1) nogil:
 *         cdef fsl_int ncoords = self.size
 *         cdef real[:, ::1] coords = self.fast_get_bead_coords_bbox()             # <<<<<<<<<<<<<<
 *         cdef real[:, ::1] directions = self.fast_get_directions()
 *         cdef real[:, ::1] normals
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_bead_coords_bbox(__pyx_v_self); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1570, __pyx_L4_error)
    __pyx_v_coords = __pyx_t_2;
    __pyx_t_2.memview = NULL;
    __pyx_t_2.data = NULL;

    /* "fatslimlib/core_base.pyx":1571
 *         cdef fsl_int ncoords = self.size
 *         cdef real[:, ::1] coords = self.fast_get_bead_coords_bbox()
 *         cdef real[:, ::1] directions = self.fast_get_directions()             # <<<<<<<<<<<<<<
 *         cdef real[:, ::1] normals
 *         cdef fsl_int *neighborhood_buffer = NULL
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_directions(__pyx_v_self); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1571, __pyx_L4_error)
    __pyx_v_directions = __pyx_t_2;
    __pyx_t_2.memview = NULL;
    __pyx_t_2.data = NULL;

    /* "fatslimlib/core_base.pyx":1573
 *         cdef real[:, ::1] directions = self.fast_get_directions()
 *         cdef real[:, ::1] normals
 *         cdef fsl_int *neighborhood_buffer = NULL             # <<<<<<<<<<<<<<
 *         cdef fsl_int buffer_size = 0
 *         cdef fsl_int nid, i
 */
    __pyx_v_neighborhood_buffer = NULL;

    /* "fatslimlib/core_base.pyx":1574
 *         cdef real[:, ::1] normals
 *         cdef fsl_int *neighborhood_buffer = NULL
 *         cdef fsl_int buffer_size = 0             # <<<<<<<<<<<<<<
 *         cdef fsl_int nid, i
 * 
 */
    __pyx_v_buffer_size = 0;

    /* "fatslimlib/core_base.pyx":1577
 *         cdef fsl_int nid, i
 * 
 *         if self.proximity_cutoff == proximity_cutoff and self.normals is not None:             # <<<<<<<<<<<<<<
 *             return self.normals
 * 
 */
    __pyx_t_4 = ((__pyx_v_self->proximity_cutoff == __pyx_v_proximity_cutoff) != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L7_bool_binop_done;
    }
    if (unlikely(!__pyx_v_self->normals.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1577, __pyx_L4_error)}
    __pyx_t_4 = ((((PyObject *) __pyx_v_self->normals.memview) != Py_None) != 0);
    __pyx_t_3 = __pyx_t_4;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_3) {

      /* "fatslimlib/core_base.pyx":1578
 * 
 *         if self.proximity_cutoff == proximity_cutoff and self.normals is not None:
 *             return self.normals             # <<<<<<<<<<<<<<
 * 
 *         if proximity_cutoff < 0:
 */
      if (unlikely(!__pyx_v_self->normals.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1578, __pyx_L4_error)}
      __PYX_INC_MEMVIEW(&__pyx_v_self->normals, 1);
      __pyx_r = __pyx_v_self->normals;
      goto __pyx_L3_return;

      /* "fatslimlib/core_base.pyx":1577
 *         cdef fsl_int nid, i
 * 
 *         if self.proximity_cutoff == proximity_cutoff and self.normals is not None:             # <<<<<<<<<<<<<<
 *             return self.normals
 * 
 */
    }

    /* "fatslimlib/core_base.pyx":1580
 *             return self.normals
 * 
 *         if proximity_cutoff < 0:             # <<<<<<<<<<<<<<
 *             if self.normals is not None:
 *                 return self.normals
 */
    __pyx_t_3 = ((__pyx_v_proximity_cutoff < 0.0) != 0);
    if (__pyx_t_3) {

      /* "fatslimlib/core_base.pyx":1581
 * 
 *         if proximity_cutoff < 0:
 *             if self.normals is not None:             # <<<<<<<<<<<<<<
 *                 return self.normals
 *             else:
 */
      if (unlikely(!__pyx_v_self->normals.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1581, __pyx_L4_error)}
      __pyx_t_3 = ((((PyObject *) __pyx_v_self->normals.memview) != Py_None) != 0);
      if (__pyx_t_3) {

        /* "fatslimlib/core_base.pyx":1582
 *         if proximity_cutoff < 0:
 *             if self.normals is not None:
 *                 return self.normals             # <<<<<<<<<<<<<<
 *             else:
 *                 self.proximity_cutoff = self.trajectory.normal_cutoff
 */
        if (unlikely(!__pyx_v_self->normals.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1582, __pyx_L4_error)}
        __PYX_INC_MEMVIEW(&__pyx_v_self->normals, 1);
        __pyx_r = __pyx_v_self->normals;
        goto __pyx_L3_return;

        /* "fatslimlib/core_base.pyx":1581
 * 
 *         if proximity_cutoff < 0:
 *             if self.normals is not None:             # <<<<<<<<<<<<<<
 *                 return self.normals
 *             else:
 */
      }

      /* "fatslimlib/core_base.pyx":1584
 *                 return self.normals
 *             else:
 *                 self.proximity_cutoff = self.trajectory.normal_cutoff             # <<<<<<<<<<<<<<
 *         else:
 *             self.proximity_cutoff = proximity_cutoff
 */
      /*else*/ {
        __pyx_t_5 = __pyx_v_self->trajectory->normal_cutoff;
        __pyx_v_self->proximity_cutoff = __pyx_t_5;
      }

      /* "fatslimlib/core_base.pyx":1580
 *             return self.normals
 * 
 *         if proximity_cutoff < 0:             # <<<<<<<<<<<<<<
 *             if self.normals is not None:
 *                 return self.normals
 */
      goto __pyx_L9;
    }

    /* "fatslimlib/core_base.pyx":1586
 *                 self.proximity_cutoff = self.trajectory.normal_cutoff
 *         else:
 *             self.proximity_cutoff = proximity_cutoff             # <<<<<<<<<<<<<<
 * 
 *         self.trajectory.normal_cutoff = self.proximity_cutoff
 */
    /*else*/ {
      __pyx_v_self->proximity_cutoff = __pyx_v_proximity_cutoff;
    }
    __pyx_L9:;

    /* "fatslimlib/core_base.pyx":1588
 *             self.proximity_cutoff = proximity_cutoff
 * 
 *         self.trajectory.normal_cutoff = self.proximity_cutoff             # <<<<<<<<<<<<<<
 * 
 *         # Delete old ref if needed
 */
    __pyx_t_5 = __pyx_v_self->proximity_cutoff;
    __pyx_v_self->trajectory->normal_cutoff = __pyx_t_5;

    /* "fatslimlib/core_base.pyx":1591
 * 
 *         # Delete old ref if needed
 *         if self.neighbors != NULL:             # <<<<<<<<<<<<<<
 *             free_neighborhood_holder(self.neighbors)
 *             self.neighbors = NULL
 */
    __pyx_t_3 = ((__pyx_v_self->neighbors != NULL) != 0);
    if (__pyx_t_3) {

      /* "fatslimlib/core_base.pyx":1592
 *         # Delete old ref if needed
 *         if self.neighbors != NULL:
 *             free_neighborhood_holder(self.neighbors)             # <<<<<<<<<<<<<<
 *             self.neighbors = NULL
 * 
 */
      __pyx_f_10fatslimlib_7core_ns_free_neighborhood_holder(__pyx_v_self->neighbors);

      /* "fatslimlib/core_base.pyx":1593
 *         if self.neighbors != NULL:
 *             free_neighborhood_holder(self.neighbors)
 *             self.neighbors = NULL             # <<<<<<<<<<<<<<
 * 
 *         # Retrieve neighborhoods
 */
      __pyx_v_self->neighbors = NULL;

      /* "fatslimlib/core_base.pyx":1591
 * 
 *         # Delete old ref if needed
 *         if self.neighbors != NULL:             # <<<<<<<<<<<<<<
 *             free_neighborhood_holder(self.neighbors)
 *             self.neighbors = NULL
 */
    }

    /* "fatslimlib/core_base.pyx":1596
 * 
 *         # Retrieve neighborhoods
 *         self.neighbors = fast_neighbor_search(coords, coords,             # <<<<<<<<<<<<<<
 *                                   self.box,
 *                                   self.proximity_cutoff)
 */
    __pyx_v_self->neighbors = __pyx_f_10fatslimlib_7core_ns_fast_neighbor_search(__pyx_v_coords, __pyx_v_coords, __pyx_v_self->box, __pyx_v_self->proximity_cutoff);

    /* "fatslimlib/core_base.pyx":1600
 *                                   self.proximity_cutoff)
 * 
 *         with gil:             # <<<<<<<<<<<<<<
 *             normals = np.empty((ncoords, DIM))
 * 
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {

          /* "fatslimlib/core_base.pyx":1601
 * 
 *         with gil:
 *             normals = np.empty((ncoords, DIM))             # <<<<<<<<<<<<<<
 * 
 *         for i in range(ncoords):
 */
          __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_np); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1601, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_empty); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1601, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = __Pyx_PyInt_From_fsl_int(__pyx_v_ncoords); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1601, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_9 = PyTuple_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1601, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7);
          __Pyx_INCREF(__pyx_int_3);
          __Pyx_GIVEREF(__pyx_int_3);
          PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_int_3);
          __pyx_t_7 = 0;
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_8, function);
            }
          }
          __pyx_t_6 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_7, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_9);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1601, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(__pyx_t_6, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1601, __pyx_L13_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_v_normals = __pyx_t_2;
          __pyx_t_2.memview = NULL;
          __pyx_t_2.data = NULL;
        }

        /* "fatslimlib/core_base.pyx":1600
 *                                   self.proximity_cutoff)
 * 
 *         with gil:             # <<<<<<<<<<<<<<
 *             normals = np.empty((ncoords, DIM))
 * 
 */
        /*finally:*/ {
          /*normal exit:*/{
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L14;
          }
          __pyx_L13_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
          __pyx_L14:;
        }
    }

    /* "fatslimlib/core_base.pyx":1603
 *             normals = np.empty((ncoords, DIM))
 * 
 *         for i in range(ncoords):             # <<<<<<<<<<<<<<
 *             if self.neighbors.neighborhoods[i].size > buffer_size:
 *                 buffer_size = self.neighbors.neighborhoods[i].size
 */
    __pyx_t_1 = __pyx_v_ncoords;
    __pyx_t_10 = __pyx_t_1;
    for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
      __pyx_v_i = __pyx_t_11;

      /* "fatslimlib/core_base.pyx":1604
 * 
 *         for i in range(ncoords):
 *             if self.neighbors.neighborhoods[i].size > buffer_size:             # <<<<<<<<<<<<<<
 *                 buffer_size = self.neighbors.neighborhoods[i].size
 * 
 */
      __pyx_t_3 = (((__pyx_v_self->neighbors->neighborhoods[__pyx_v_i])->size > __pyx_v_buffer_size) != 0);
      if (__pyx_t_3) {

        /* "fatslimlib/core_base.pyx":1605
 *         for i in range(ncoords):
 *             if self.neighbors.neighborhoods[i].size > buffer_size:
 *                 buffer_size = self.neighbors.neighborhoods[i].size             # <<<<<<<<<<<<<<
 * 
 *         neighborhood_buffer = <fsl_int *> malloc(buffer_size * sizeof(fsl_int) * OPENMP_NUM_THREADS)
 */
        __pyx_t_12 = (__pyx_v_self->neighbors->neighborhoods[__pyx_v_i])->size;
        __pyx_v_buffer_size = __pyx_t_12;

        /* "fatslimlib/core_base.pyx":1604
 * 
 *         for i in range(ncoords):
 *             if self.neighbors.neighborhoods[i].size > buffer_size:             # <<<<<<<<<<<<<<
 *                 buffer_size = self.neighbors.neighborhoods[i].size
 * 
 */
      }
    }

    /* "fatslimlib/core_base.pyx":1607
 *                 buffer_size = self.neighbors.neighborhoods[i].size
 * 
 *         neighborhood_buffer = <fsl_int *> malloc(buffer_size * sizeof(fsl_int) * OPENMP_NUM_THREADS)             # <<<<<<<<<<<<<<
 *         if neighborhood_buffer == NULL:
 *             abort()
 */
    __pyx_v_neighborhood_buffer = ((fsl_int *)malloc(((__pyx_v_buffer_size * (sizeof(fsl_int))) * __pyx_v_10fatslimlib_9core_base_OPENMP_NUM_THREADS)));

    /* "fatslimlib/core_base.pyx":1608
 * 
 *         neighborhood_buffer = <fsl_int *> malloc(buffer_size * sizeof(fsl_int) * OPENMP_NUM_THREADS)
 *         if neighborhood_buffer == NULL:             # <<<<<<<<<<<<<<
 *             abort()
 * 
 */
    __pyx_t_3 = ((__pyx_v_neighborhood_buffer == NULL) != 0);
    if (__pyx_t_3) {

      /* "fatslimlib/core_base.pyx":1609
 *         neighborhood_buffer = <fsl_int *> malloc(buffer_size * sizeof(fsl_int) * OPENMP_NUM_THREADS)
 *         if neighborhood_buffer == NULL:
 *             abort()             # <<<<<<<<<<<<<<
 * 
 *         for nid in prange(ncoords, schedule='dynamic', num_threads=OPENMP_NUM_THREADS):
 */
      abort();

      /* "fatslimlib/core_base.pyx":1608
 * 
 *         neighborhood_buffer = <fsl_int *> malloc(buffer_size * sizeof(fsl_int) * OPENMP_NUM_THREADS)
 *         if neighborhood_buffer == NULL:             # <<<<<<<<<<<<<<
 *             abort()
 * 
 */
    }

    /* "fatslimlib/core_base.pyx":1611
 *             abort()
 * 
 *         for nid in prange(ncoords, schedule='dynamic', num_threads=OPENMP_NUM_THREADS):             # <<<<<<<<<<<<<<
 *         #for nid in range(ncoords):
 *             calculate_neighborhood_normal(nid,
 */
    __pyx_t_1 = __pyx_v_ncoords;
    if (1 == 0) abort();
    {
        #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
            #undef likely
            #undef unlikely
            #define likely(x)   (x)
            #define unlikely(x) (x)
        #endif
        __pyx_t_11 = (__pyx_t_1 - 0 + 1 - 1/abs(1)) / 1;
        if (__pyx_t_11 > 0)
        {
            #ifdef _OPENMP
            #pragma omp parallel num_threads(__pyx_v_10fatslimlib_9core_base_OPENMP_NUM_THREADS) private(__pyx_t_13, __pyx_t_14)
            #endif /* _OPENMP */
            {
                #ifdef _OPENMP
                #pragma omp for firstprivate(__pyx_v_nid) lastprivate(__pyx_v_nid) schedule(dynamic)
                #endif /* _OPENMP */
                for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_11; __pyx_t_10++){
                    {
                        __pyx_v_nid = (fsl_int)(0 + 1 * __pyx_t_10);

                        /* "fatslimlib/core_base.pyx":1618
 *                                           self.neighbors.neighborhoods[nid],
 *                                           self.box,
 *                                           &normals[nid, XX],             # <<<<<<<<<<<<<<
 *                                           &neighborhood_buffer[omp_get_thread_num()
 *                                                                * buffer_size])
 */
                        __pyx_t_13 = __pyx_v_nid;
                        __pyx_t_14 = 0;

                        /* "fatslimlib/core_base.pyx":1613
 *         for nid in prange(ncoords, schedule='dynamic', num_threads=OPENMP_NUM_THREADS):
 *         #for nid in range(ncoords):
 *             calculate_neighborhood_normal(nid,             # <<<<<<<<<<<<<<
 *                                           coords,
 *                                           directions,
 */
                        __pyx_f_10fatslimlib_9core_base_calculate_neighborhood_normal(__pyx_v_nid, __pyx_v_coords, __pyx_v_directions, (__pyx_v_self->neighbors->neighborhoods[__pyx_v_nid]), __pyx_v_self->box, (&(*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_normals.data + __pyx_t_13 * __pyx_v_normals.strides[0]) )) + __pyx_t_14)) )))), (&(__pyx_v_neighborhood_buffer[(omp_get_thread_num() * __pyx_v_buffer_size)])));
                    }
                }
            }
        }
    }
    #if ((defined(__APPLE__) || defined(__OSX__)) && (defined(__GNUC__) && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))))
        #undef likely
        #undef unlikely
        #define likely(x)   __builtin_expect(!!(x), 1)
        #define unlikely(x) __builtin_expect(!!(x), 0)
    #endif

    /* "fatslimlib/core_base.pyx":1622
 *                                                                * buffer_size])
 * 
 *         free(neighborhood_buffer)             # <<<<<<<<<<<<<<
 * 
 *         self.normals = normals
 */
    free(__pyx_v_neighborhood_buffer);

    /* "fatslimlib/core_base.pyx":1624
 *         free(neighborhood_buffer)
 * 
 *         self.normals = normals             # <<<<<<<<<<<<<<
 *         return self.normals
 * 
 */
    __PYX_XDEC_MEMVIEW(&__pyx_v_self->normals, 0);
    __PYX_INC_MEMVIEW(&__pyx_v_normals, 1);
    __pyx_v_self->normals = __pyx_v_normals;

    /* "fatslimlib/core_base.pyx":1625
 * 
 *         self.normals = normals
 *         return self.normals             # <<<<<<<<<<<<<<
 * 
 *     # Associated python property
 */
    if (unlikely(!__pyx_v_self->normals.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1625, __pyx_L4_error)}
    __PYX_INC_MEMVIEW(&__pyx_v_self->normals, 1);
    __pyx_r = __pyx_v_self->normals;
    goto __pyx_L3_return;
  }

  /* "fatslimlib/core_base.pyx":1569
 * 
 *     cdef real[:, ::1] fast_get_normals(self, real proximity_cutoff=-1) nogil:
 *         cdef fsl_int ncoords = self.size             # <<<<<<<<<<<<<<
 *         cdef real[:, ::1] coords = self.fast_get_bead_coords_bbox()
 *         cdef real[:, ::1] directions = self.fast_get_directions()
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "fatslimlib/core_base.pyx":1568
 * 
 * 
 *     cdef real[:, ::1] fast_get_normals(self, real proximity_cutoff=-1) nogil:             # <<<<<<<<<<<<<<
 *         cdef fsl_int ncoords = self.size
 *         cdef real[:, ::1] coords = self.fast_get_bead_coords_bbox()
 */

  /* function exit code */
  __Pyx_pretend_to_initialize(&__pyx_r);
  goto __pyx_L0;
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 0);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __pyx_r.data = NULL;
  __pyx_r.memview = NULL;
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.fast_get_normals", __pyx_clineno, __pyx_lineno, __pyx_filename);
  goto __pyx_L2;
  __pyx_L0:;
  if (unlikely(!__pyx_r.memview)) {
    PyErr_SetString(PyExc_TypeError, "Memoryview return value is not initialized");
  }
  __pyx_L2:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_coords, 0);
  __PYX_XDEC_MEMVIEW(&__pyx_v_directions, 0);
  __PYX_XDEC_MEMVIEW(&__pyx_v_normals, 0);
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1629
 *     # Associated python property
 *     property normals:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef real[:, ::1] memview
 *             with nogil:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_7normals_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_7normals_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_7normals___get__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_7normals___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_memview = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":1631
 *         def __get__(self):
 *             cdef real[:, ::1] memview
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 memview = self.fast_get_normals()
 *             #memview = np.empty((self.size, DIM))
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "fatslimlib/core_base.pyx":1632
 *             cdef real[:, ::1] memview
 *             with nogil:
 *                 memview = self.fast_get_normals()             # <<<<<<<<<<<<<<
 *             #memview = np.empty((self.size, DIM))
 *             return np.asarray(memview)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame *)__pyx_v_self->__pyx_vtab)->fast_get_normals(__pyx_v_self, NULL); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(0, 1632, __pyx_L4_error)
        __pyx_v_memview = __pyx_t_1;
        __pyx_t_1.memview = NULL;
        __pyx_t_1.data = NULL;
      }

      /* "fatslimlib/core_base.pyx":1631
 *         def __get__(self):
 *             cdef real[:, ::1] memview
 *             with nogil:             # <<<<<<<<<<<<<<
 *                 memview = self.fast_get_normals()
 *             #memview = np.empty((self.size, DIM))
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L4_error: {
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L1_error;
        }
        __pyx_L5:;
      }
  }

  /* "fatslimlib/core_base.pyx":1634
 *                 memview = self.fast_get_normals()
 *             #memview = np.empty((self.size, DIM))
 *             return np.asarray(memview)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_asarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_memview, 2, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1629
 *     # Associated python property
 *     property normals:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef real[:, ::1] memview
 *             with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.normals.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_memview, 1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1637
 * 
 * 
 *     cpdef list get_aggregates(self, real cutoff=DEFAULT_PROXIMITY_CUTOFF, bint update=False):             # <<<<<<<<<<<<<<
 *         from .core_analysis import retrieve_aggregates
 * 
 */

static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_9get_aggregates(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_5Frame_get_aggregates(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_get_aggregates *__pyx_optional_args) {
  real __pyx_v_cutoff = ((real)2.0);
  int __pyx_v_update = ((int)0);
  PyObject *__pyx_v_retrieve_aggregates = NULL;
  PyObject *__pyx_v_traj_aggregates = NULL;
  PyObject *__pyx_v_aggregate = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_t_13;
  __Pyx_RefNannySetupContext("get_aggregates", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_cutoff = __pyx_optional_args->cutoff;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_update = __pyx_optional_args->update;
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_aggregates); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1637, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10fatslimlib_9core_base_5Frame_9get_aggregates)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_cutoff); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1637, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_update); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1637, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1637, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1637, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1637, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1637, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 1637, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "fatslimlib/core_base.pyx":1638
 * 
 *     cpdef list get_aggregates(self, real cutoff=DEFAULT_PROXIMITY_CUTOFF, bint update=False):
 *         from .core_analysis import retrieve_aggregates             # <<<<<<<<<<<<<<
 * 
 *         if self.aggregates_cutoff == cutoff and self.aggregates_retrieved and not update:
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_retrieve_aggregates);
  __Pyx_GIVEREF(__pyx_n_s_retrieve_aggregates);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_retrieve_aggregates);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_core_analysis, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_retrieve_aggregates); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_retrieve_aggregates = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":1640
 *         from .core_analysis import retrieve_aggregates
 * 
 *         if self.aggregates_cutoff == cutoff and self.aggregates_retrieved and not update:             # <<<<<<<<<<<<<<
 *             return self.aggregates
 * 
 */
  __pyx_t_10 = ((__pyx_v_self->aggregates_cutoff == __pyx_v_cutoff) != 0);
  if (__pyx_t_10) {
  } else {
    __pyx_t_9 = __pyx_t_10;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_10 = (__pyx_v_self->aggregates_retrieved != 0);
  if (__pyx_t_10) {
  } else {
    __pyx_t_9 = __pyx_t_10;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_10 = ((!(__pyx_v_update != 0)) != 0);
  __pyx_t_9 = __pyx_t_10;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_9) {

    /* "fatslimlib/core_base.pyx":1641
 * 
 *         if self.aggregates_cutoff == cutoff and self.aggregates_retrieved and not update:
 *             return self.aggregates             # <<<<<<<<<<<<<<
 * 
 *         self.aggregates_cutoff = cutoff
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_self->aggregates);
    __pyx_r = __pyx_v_self->aggregates;
    goto __pyx_L0;

    /* "fatslimlib/core_base.pyx":1640
 *         from .core_analysis import retrieve_aggregates
 * 
 *         if self.aggregates_cutoff == cutoff and self.aggregates_retrieved and not update:             # <<<<<<<<<<<<<<
 *             return self.aggregates
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":1643
 *             return self.aggregates
 * 
 *         self.aggregates_cutoff = cutoff             # <<<<<<<<<<<<<<
 *         if not update \
 *                 and self.trajectory.aggregates is not None \
 */
  __pyx_v_self->aggregates_cutoff = __pyx_v_cutoff;

  /* "fatslimlib/core_base.pyx":1644
 * 
 *         self.aggregates_cutoff = cutoff
 *         if not update \             # <<<<<<<<<<<<<<
 *                 and self.trajectory.aggregates is not None \
 *                 and self.trajectory.normal_cutoff == cutoff:
 */
  __pyx_t_10 = ((!(__pyx_v_update != 0)) != 0);
  if (__pyx_t_10) {
  } else {
    __pyx_t_9 = __pyx_t_10;
    goto __pyx_L8_bool_binop_done;
  }

  /* "fatslimlib/core_base.pyx":1645
 *         self.aggregates_cutoff = cutoff
 *         if not update \
 *                 and self.trajectory.aggregates is not None \             # <<<<<<<<<<<<<<
 *                 and self.trajectory.normal_cutoff == cutoff:
 *             traj_aggregates = self.trajectory.aggregates
 */
  __pyx_t_10 = (__pyx_v_self->trajectory->aggregates != ((PyObject*)Py_None));
  __pyx_t_11 = (__pyx_t_10 != 0);
  if (__pyx_t_11) {
  } else {
    __pyx_t_9 = __pyx_t_11;
    goto __pyx_L8_bool_binop_done;
  }

  /* "fatslimlib/core_base.pyx":1646
 *         if not update \
 *                 and self.trajectory.aggregates is not None \
 *                 and self.trajectory.normal_cutoff == cutoff:             # <<<<<<<<<<<<<<
 *             traj_aggregates = self.trajectory.aggregates
 *             self.aggregates = []
 */
  __pyx_t_11 = ((__pyx_v_self->trajectory->normal_cutoff == __pyx_v_cutoff) != 0);
  __pyx_t_9 = __pyx_t_11;
  __pyx_L8_bool_binop_done:;

  /* "fatslimlib/core_base.pyx":1644
 * 
 *         self.aggregates_cutoff = cutoff
 *         if not update \             # <<<<<<<<<<<<<<
 *                 and self.trajectory.aggregates is not None \
 *                 and self.trajectory.normal_cutoff == cutoff:
 */
  if (__pyx_t_9) {

    /* "fatslimlib/core_base.pyx":1647
 *                 and self.trajectory.aggregates is not None \
 *                 and self.trajectory.normal_cutoff == cutoff:
 *             traj_aggregates = self.trajectory.aggregates             # <<<<<<<<<<<<<<
 *             self.aggregates = []
 *             for aggregate in traj_aggregates:
 */
    __pyx_t_2 = __pyx_v_self->trajectory->aggregates;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_traj_aggregates = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "fatslimlib/core_base.pyx":1648
 *                 and self.trajectory.normal_cutoff == cutoff:
 *             traj_aggregates = self.trajectory.aggregates
 *             self.aggregates = []             # <<<<<<<<<<<<<<
 *             for aggregate in traj_aggregates:
 *                 self.aggregates.append(Aggregate(self, aggregate))
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->aggregates);
    __Pyx_DECREF(__pyx_v_self->aggregates);
    __pyx_v_self->aggregates = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "fatslimlib/core_base.pyx":1649
 *             traj_aggregates = self.trajectory.aggregates
 *             self.aggregates = []
 *             for aggregate in traj_aggregates:             # <<<<<<<<<<<<<<
 *                 self.aggregates.append(Aggregate(self, aggregate))
 *         else:
 */
    if (unlikely(__pyx_v_traj_aggregates == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 1649, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_v_traj_aggregates; __Pyx_INCREF(__pyx_t_2); __pyx_t_12 = 0;
    for (;;) {
      if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_2)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_12); __Pyx_INCREF(__pyx_t_1); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 1649, __pyx_L1_error)
      #else
      __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1649, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_aggregate, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "fatslimlib/core_base.pyx":1650
 *             self.aggregates = []
 *             for aggregate in traj_aggregates:
 *                 self.aggregates.append(Aggregate(self, aggregate))             # <<<<<<<<<<<<<<
 *         else:
 *             self.aggregates = retrieve_aggregates(self, cutoff)
 */
      if (unlikely(__pyx_v_self->aggregates == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
        __PYX_ERR(0, 1650, __pyx_L1_error)
      }
      __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1650, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self));
      __Pyx_INCREF(__pyx_v_aggregate);
      __Pyx_GIVEREF(__pyx_v_aggregate);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_aggregate);
      __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_10fatslimlib_13core_analysis_Aggregate), __pyx_t_1, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1650, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_self->aggregates, __pyx_t_5); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 1650, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "fatslimlib/core_base.pyx":1649
 *             traj_aggregates = self.trajectory.aggregates
 *             self.aggregates = []
 *             for aggregate in traj_aggregates:             # <<<<<<<<<<<<<<
 *                 self.aggregates.append(Aggregate(self, aggregate))
 *         else:
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fatslimlib/core_base.pyx":1644
 * 
 *         self.aggregates_cutoff = cutoff
 *         if not update \             # <<<<<<<<<<<<<<
 *                 and self.trajectory.aggregates is not None \
 *                 and self.trajectory.normal_cutoff == cutoff:
 */
    goto __pyx_L7;
  }

  /* "fatslimlib/core_base.pyx":1652
 *                 self.aggregates.append(Aggregate(self, aggregate))
 *         else:
 *             self.aggregates = retrieve_aggregates(self, cutoff)             # <<<<<<<<<<<<<<
 *             self.trajectory.aggregates = []
 *             for aggregate in self.aggregates:
 */
  /*else*/ {
    __pyx_t_5 = PyFloat_FromDouble(__pyx_v_cutoff); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1652, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_retrieve_aggregates);
    __pyx_t_1 = __pyx_v_retrieve_aggregates; __pyx_t_8 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, ((PyObject *)__pyx_v_self), __pyx_t_5};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1652, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, ((PyObject *)__pyx_v_self), __pyx_t_5};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1652, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1652, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_7, ((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_7, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1652, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 1652, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->aggregates);
    __Pyx_DECREF(__pyx_v_self->aggregates);
    __pyx_v_self->aggregates = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "fatslimlib/core_base.pyx":1653
 *         else:
 *             self.aggregates = retrieve_aggregates(self, cutoff)
 *             self.trajectory.aggregates = []             # <<<<<<<<<<<<<<
 *             for aggregate in self.aggregates:
 *                 self.trajectory.aggregates.append(np.asarray(aggregate.beadids))
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1653, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->trajectory->aggregates);
    __Pyx_DECREF(__pyx_v_self->trajectory->aggregates);
    __pyx_v_self->trajectory->aggregates = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "fatslimlib/core_base.pyx":1654
 *             self.aggregates = retrieve_aggregates(self, cutoff)
 *             self.trajectory.aggregates = []
 *             for aggregate in self.aggregates:             # <<<<<<<<<<<<<<
 *                 self.trajectory.aggregates.append(np.asarray(aggregate.beadids))
 *         self.aggregates_retrieved = True
 */
    if (unlikely(__pyx_v_self->aggregates == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 1654, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_v_self->aggregates; __Pyx_INCREF(__pyx_t_2); __pyx_t_12 = 0;
    for (;;) {
      if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_2)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_12); __Pyx_INCREF(__pyx_t_1); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 1654, __pyx_L1_error)
      #else
      __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1654, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_aggregate, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "fatslimlib/core_base.pyx":1655
 *             self.trajectory.aggregates = []
 *             for aggregate in self.aggregates:
 *                 self.trajectory.aggregates.append(np.asarray(aggregate.beadids))             # <<<<<<<<<<<<<<
 *         self.aggregates_retrieved = True
 * 
 */
      if (unlikely(__pyx_v_self->trajectory->aggregates == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
        __PYX_ERR(0, 1655, __pyx_L1_error)
      }
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1655, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_asarray); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1655, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_aggregate, __pyx_n_s_beadids); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1655, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_8, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1655, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_self->trajectory->aggregates, __pyx_t_1); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 1655, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "fatslimlib/core_base.pyx":1654
 *             self.aggregates = retrieve_aggregates(self, cutoff)
 *             self.trajectory.aggregates = []
 *             for aggregate in self.aggregates:             # <<<<<<<<<<<<<<
 *                 self.trajectory.aggregates.append(np.asarray(aggregate.beadids))
 *         self.aggregates_retrieved = True
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L7:;

  /* "fatslimlib/core_base.pyx":1656
 *             for aggregate in self.aggregates:
 *                 self.trajectory.aggregates.append(np.asarray(aggregate.beadids))
 *         self.aggregates_retrieved = True             # <<<<<<<<<<<<<<
 * 
 *         return self.aggregates
 */
  __pyx_v_self->aggregates_retrieved = 1;

  /* "fatslimlib/core_base.pyx":1658
 *         self.aggregates_retrieved = True
 * 
 *         return self.aggregates             # <<<<<<<<<<<<<<
 * 
 *     cpdef list get_membranes(self, real cutoff=DEFAULT_PROXIMITY_CUTOFF, bint update=False):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->aggregates);
  __pyx_r = __pyx_v_self->aggregates;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1637
 * 
 * 
 *     cpdef list get_aggregates(self, real cutoff=DEFAULT_PROXIMITY_CUTOFF, bint update=False):             # <<<<<<<<<<<<<<
 *         from .core_analysis import retrieve_aggregates
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.get_aggregates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_retrieve_aggregates);
  __Pyx_XDECREF(__pyx_v_traj_aggregates);
  __Pyx_XDECREF(__pyx_v_aggregate);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_9get_aggregates(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_9get_aggregates(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  real __pyx_v_cutoff;
  int __pyx_v_update;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_aggregates (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cutoff,&__pyx_n_s_update,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cutoff);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_update);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_aggregates") < 0)) __PYX_ERR(0, 1637, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_cutoff = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_cutoff == ((real)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1637, __pyx_L3_error)
    } else {
      __pyx_v_cutoff = ((real)2.0);
    }
    if (values[1]) {
      __pyx_v_update = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_update == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1637, __pyx_L3_error)
    } else {
      __pyx_v_update = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_aggregates", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1637, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.get_aggregates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_8get_aggregates(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self), __pyx_v_cutoff, __pyx_v_update);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_8get_aggregates(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self, real __pyx_v_cutoff, int __pyx_v_update) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_get_aggregates __pyx_t_2;
  __Pyx_RefNannySetupContext("get_aggregates", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.cutoff = __pyx_v_cutoff;
  __pyx_t_2.update = __pyx_v_update;
  __pyx_t_1 = __pyx_vtabptr_10fatslimlib_9core_base_Frame->get_aggregates(__pyx_v_self, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.get_aggregates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1660
 *         return self.aggregates
 * 
 *     cpdef list get_membranes(self, real cutoff=DEFAULT_PROXIMITY_CUTOFF, bint update=False):             # <<<<<<<<<<<<<<
 *         from .core_analysis import retrieve_membranes
 * 
 */

static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_11get_membranes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_10fatslimlib_9core_base_5Frame_get_membranes(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_get_membranes *__pyx_optional_args) {
  real __pyx_v_cutoff = ((real)2.0);
  int __pyx_v_update = ((int)0);
  PyObject *__pyx_v_retrieve_membranes = NULL;
  PyObject *__pyx_v_traj_membranes = NULL;
  PyObject *__pyx_v_membrane = NULL;
  struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *__pyx_v_l1 = NULL;
  struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *__pyx_v_l2 = NULL;
  PyObject *__pyx_v_memb_beadids = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_t_13;
  __Pyx_RefNannySetupContext("get_membranes", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_cutoff = __pyx_optional_args->cutoff;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_update = __pyx_optional_args->update;
      }
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_membranes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1660, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10fatslimlib_9core_base_5Frame_11get_membranes)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = PyFloat_FromDouble(__pyx_v_cutoff); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1660, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_update); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1660, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1660, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1660, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1660, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1660, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 1660, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "fatslimlib/core_base.pyx":1661
 * 
 *     cpdef list get_membranes(self, real cutoff=DEFAULT_PROXIMITY_CUTOFF, bint update=False):
 *         from .core_analysis import retrieve_membranes             # <<<<<<<<<<<<<<
 * 
 *         if self.membranes_cutoff == cutoff and self.membranes_retrieved and not update:
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_retrieve_membranes);
  __Pyx_GIVEREF(__pyx_n_s_retrieve_membranes);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_retrieve_membranes);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_core_analysis, __pyx_t_1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_retrieve_membranes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_retrieve_membranes = __pyx_t_1;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":1663
 *         from .core_analysis import retrieve_membranes
 * 
 *         if self.membranes_cutoff == cutoff and self.membranes_retrieved and not update:             # <<<<<<<<<<<<<<
 *             return self.membranes
 * 
 */
  __pyx_t_10 = ((__pyx_v_self->membranes_cutoff == __pyx_v_cutoff) != 0);
  if (__pyx_t_10) {
  } else {
    __pyx_t_9 = __pyx_t_10;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_10 = (__pyx_v_self->membranes_retrieved != 0);
  if (__pyx_t_10) {
  } else {
    __pyx_t_9 = __pyx_t_10;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_10 = ((!(__pyx_v_update != 0)) != 0);
  __pyx_t_9 = __pyx_t_10;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_9) {

    /* "fatslimlib/core_base.pyx":1664
 * 
 *         if self.membranes_cutoff == cutoff and self.membranes_retrieved and not update:
 *             return self.membranes             # <<<<<<<<<<<<<<
 * 
 *         self.proximity_cutoff = cutoff
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_self->membranes);
    __pyx_r = __pyx_v_self->membranes;
    goto __pyx_L0;

    /* "fatslimlib/core_base.pyx":1663
 *         from .core_analysis import retrieve_membranes
 * 
 *         if self.membranes_cutoff == cutoff and self.membranes_retrieved and not update:             # <<<<<<<<<<<<<<
 *             return self.membranes
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":1666
 *             return self.membranes
 * 
 *         self.proximity_cutoff = cutoff             # <<<<<<<<<<<<<<
 *         if not update and self.trajectory.membranes is not None:
 *             traj_membranes = self.trajectory.membranes
 */
  __pyx_v_self->proximity_cutoff = __pyx_v_cutoff;

  /* "fatslimlib/core_base.pyx":1667
 * 
 *         self.proximity_cutoff = cutoff
 *         if not update and self.trajectory.membranes is not None:             # <<<<<<<<<<<<<<
 *             traj_membranes = self.trajectory.membranes
 *             self.membranes = []
 */
  __pyx_t_10 = ((!(__pyx_v_update != 0)) != 0);
  if (__pyx_t_10) {
  } else {
    __pyx_t_9 = __pyx_t_10;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_10 = (__pyx_v_self->trajectory->membranes != ((PyObject*)Py_None));
  __pyx_t_11 = (__pyx_t_10 != 0);
  __pyx_t_9 = __pyx_t_11;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_9) {

    /* "fatslimlib/core_base.pyx":1668
 *         self.proximity_cutoff = cutoff
 *         if not update and self.trajectory.membranes is not None:
 *             traj_membranes = self.trajectory.membranes             # <<<<<<<<<<<<<<
 *             self.membranes = []
 *             for membrane in traj_membranes:
 */
    __pyx_t_2 = __pyx_v_self->trajectory->membranes;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_traj_membranes = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "fatslimlib/core_base.pyx":1669
 *         if not update and self.trajectory.membranes is not None:
 *             traj_membranes = self.trajectory.membranes
 *             self.membranes = []             # <<<<<<<<<<<<<<
 *             for membrane in traj_membranes:
 *                 l1 = Aggregate(self, membrane[0])
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->membranes);
    __Pyx_DECREF(__pyx_v_self->membranes);
    __pyx_v_self->membranes = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "fatslimlib/core_base.pyx":1670
 *             traj_membranes = self.trajectory.membranes
 *             self.membranes = []
 *             for membrane in traj_membranes:             # <<<<<<<<<<<<<<
 *                 l1 = Aggregate(self, membrane[0])
 *                 l2 = Aggregate(self, membrane[1])
 */
    if (unlikely(__pyx_v_traj_membranes == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 1670, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_v_traj_membranes; __Pyx_INCREF(__pyx_t_2); __pyx_t_12 = 0;
    for (;;) {
      if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_2)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_12); __Pyx_INCREF(__pyx_t_1); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 1670, __pyx_L1_error)
      #else
      __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1670, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_membrane, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "fatslimlib/core_base.pyx":1671
 *             self.membranes = []
 *             for membrane in traj_membranes:
 *                 l1 = Aggregate(self, membrane[0])             # <<<<<<<<<<<<<<
 *                 l2 = Aggregate(self, membrane[1])
 *                 self.membranes.append(Membrane(self, l1, l2))
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_membrane, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1671, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1671, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_10fatslimlib_13core_analysis_Aggregate), __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1671, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_l1, ((struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "fatslimlib/core_base.pyx":1672
 *             for membrane in traj_membranes:
 *                 l1 = Aggregate(self, membrane[0])
 *                 l2 = Aggregate(self, membrane[1])             # <<<<<<<<<<<<<<
 *                 self.membranes.append(Membrane(self, l1, l2))
 *         else:
 */
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_membrane, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1672, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1672, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_10fatslimlib_13core_analysis_Aggregate), __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1672, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_l2, ((struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate *)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "fatslimlib/core_base.pyx":1673
 *                 l1 = Aggregate(self, membrane[0])
 *                 l2 = Aggregate(self, membrane[1])
 *                 self.membranes.append(Membrane(self, l1, l2))             # <<<<<<<<<<<<<<
 *         else:
 *             self.membranes = retrieve_membranes(self, cutoff)
 */
      if (unlikely(__pyx_v_self->membranes == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
        __PYX_ERR(0, 1673, __pyx_L1_error)
      }
      __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1673, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_self));
      __Pyx_INCREF(((PyObject *)__pyx_v_l1));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_l1));
      PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_l1));
      __Pyx_INCREF(((PyObject *)__pyx_v_l2));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_l2));
      PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)__pyx_v_l2));
      __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_10fatslimlib_13core_analysis_Membrane), __pyx_t_1, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1673, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_self->membranes, __pyx_t_5); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 1673, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "fatslimlib/core_base.pyx":1670
 *             traj_membranes = self.trajectory.membranes
 *             self.membranes = []
 *             for membrane in traj_membranes:             # <<<<<<<<<<<<<<
 *                 l1 = Aggregate(self, membrane[0])
 *                 l2 = Aggregate(self, membrane[1])
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fatslimlib/core_base.pyx":1667
 * 
 *         self.proximity_cutoff = cutoff
 *         if not update and self.trajectory.membranes is not None:             # <<<<<<<<<<<<<<
 *             traj_membranes = self.trajectory.membranes
 *             self.membranes = []
 */
    goto __pyx_L7;
  }

  /* "fatslimlib/core_base.pyx":1675
 *                 self.membranes.append(Membrane(self, l1, l2))
 *         else:
 *             self.membranes = retrieve_membranes(self, cutoff)             # <<<<<<<<<<<<<<
 *             traj_membranes = []
 *             for membrane in self.membranes:
 */
  /*else*/ {
    __pyx_t_5 = PyFloat_FromDouble(__pyx_v_cutoff); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_retrieve_membranes);
    __pyx_t_1 = __pyx_v_retrieve_membranes; __pyx_t_8 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, ((PyObject *)__pyx_v_self), __pyx_t_5};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1675, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, ((PyObject *)__pyx_v_self), __pyx_t_5};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1675, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1675, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_7, ((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_7, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1675, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 1675, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->membranes);
    __Pyx_DECREF(__pyx_v_self->membranes);
    __pyx_v_self->membranes = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "fatslimlib/core_base.pyx":1676
 *         else:
 *             self.membranes = retrieve_membranes(self, cutoff)
 *             traj_membranes = []             # <<<<<<<<<<<<<<
 *             for membrane in self.membranes:
 *                 memb_beadids = [np.asarray(membrane.leaflet1.beadids),
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_traj_membranes = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "fatslimlib/core_base.pyx":1677
 *             self.membranes = retrieve_membranes(self, cutoff)
 *             traj_membranes = []
 *             for membrane in self.membranes:             # <<<<<<<<<<<<<<
 *                 memb_beadids = [np.asarray(membrane.leaflet1.beadids),
 *                                 np.asarray(membrane.leaflet2.beadids)]
 */
    if (unlikely(__pyx_v_self->membranes == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 1677, __pyx_L1_error)
    }
    __pyx_t_2 = __pyx_v_self->membranes; __Pyx_INCREF(__pyx_t_2); __pyx_t_12 = 0;
    for (;;) {
      if (__pyx_t_12 >= PyList_GET_SIZE(__pyx_t_2)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_12); __Pyx_INCREF(__pyx_t_1); __pyx_t_12++; if (unlikely(0 < 0)) __PYX_ERR(0, 1677, __pyx_L1_error)
      #else
      __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_12); __pyx_t_12++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1677, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_membrane, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "fatslimlib/core_base.pyx":1678
 *             traj_membranes = []
 *             for membrane in self.membranes:
 *                 memb_beadids = [np.asarray(membrane.leaflet1.beadids),             # <<<<<<<<<<<<<<
 *                                 np.asarray(membrane.leaflet2.beadids)]
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1678, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_asarray); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1678, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_membrane, __pyx_n_s_leaflet1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1678, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_beadids); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1678, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_8);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1678, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "fatslimlib/core_base.pyx":1679
 *             for membrane in self.membranes:
 *                 memb_beadids = [np.asarray(membrane.leaflet1.beadids),
 *                                 np.asarray(membrane.leaflet2.beadids)]             # <<<<<<<<<<<<<<
 * 
 *                 traj_membranes.append(memb_beadids)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1679, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_asarray); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1679, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_membrane, __pyx_n_s_leaflet2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1679, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_beadids); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1679, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_8, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1679, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "fatslimlib/core_base.pyx":1678
 *             traj_membranes = []
 *             for membrane in self.membranes:
 *                 memb_beadids = [np.asarray(membrane.leaflet1.beadids),             # <<<<<<<<<<<<<<
 *                                 np.asarray(membrane.leaflet2.beadids)]
 * 
 */
      __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1678, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_1);
      PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_5);
      PyList_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
      __pyx_t_1 = 0;
      __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_memb_beadids, ((PyObject*)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "fatslimlib/core_base.pyx":1681
 *                                 np.asarray(membrane.leaflet2.beadids)]
 * 
 *                 traj_membranes.append(memb_beadids)             # <<<<<<<<<<<<<<
 *             self.trajectory.membranes = traj_membranes
 *         self.membranes_retrieved = True
 */
      __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_traj_membranes, __pyx_v_memb_beadids); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 1681, __pyx_L1_error)

      /* "fatslimlib/core_base.pyx":1677
 *             self.membranes = retrieve_membranes(self, cutoff)
 *             traj_membranes = []
 *             for membrane in self.membranes:             # <<<<<<<<<<<<<<
 *                 memb_beadids = [np.asarray(membrane.leaflet1.beadids),
 *                                 np.asarray(membrane.leaflet2.beadids)]
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fatslimlib/core_base.pyx":1682
 * 
 *                 traj_membranes.append(memb_beadids)
 *             self.trajectory.membranes = traj_membranes             # <<<<<<<<<<<<<<
 *         self.membranes_retrieved = True
 * 
 */
    __Pyx_INCREF(__pyx_v_traj_membranes);
    __Pyx_GIVEREF(__pyx_v_traj_membranes);
    __Pyx_GOTREF(__pyx_v_self->trajectory->membranes);
    __Pyx_DECREF(__pyx_v_self->trajectory->membranes);
    __pyx_v_self->trajectory->membranes = __pyx_v_traj_membranes;
  }
  __pyx_L7:;

  /* "fatslimlib/core_base.pyx":1683
 *                 traj_membranes.append(memb_beadids)
 *             self.trajectory.membranes = traj_membranes
 *         self.membranes_retrieved = True             # <<<<<<<<<<<<<<
 * 
 *         return self.membranes
 */
  __pyx_v_self->membranes_retrieved = 1;

  /* "fatslimlib/core_base.pyx":1685
 *         self.membranes_retrieved = True
 * 
 *         return self.membranes             # <<<<<<<<<<<<<<
 * 
 *     # Additional python properties
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->membranes);
  __pyx_r = __pyx_v_self->membranes;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1660
 *         return self.aggregates
 * 
 *     cpdef list get_membranes(self, real cutoff=DEFAULT_PROXIMITY_CUTOFF, bint update=False):             # <<<<<<<<<<<<<<
 *         from .core_analysis import retrieve_membranes
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.get_membranes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_retrieve_membranes);
  __Pyx_XDECREF(__pyx_v_traj_membranes);
  __Pyx_XDECREF(__pyx_v_membrane);
  __Pyx_XDECREF((PyObject *)__pyx_v_l1);
  __Pyx_XDECREF((PyObject *)__pyx_v_l2);
  __Pyx_XDECREF(__pyx_v_memb_beadids);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_11get_membranes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_11get_membranes(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  real __pyx_v_cutoff;
  int __pyx_v_update;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_membranes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_cutoff,&__pyx_n_s_update,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cutoff);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_update);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "get_membranes") < 0)) __PYX_ERR(0, 1660, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_cutoff = __pyx_PyFloat_AsDouble(values[0]); if (unlikely((__pyx_v_cutoff == ((real)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1660, __pyx_L3_error)
    } else {
      __pyx_v_cutoff = ((real)2.0);
    }
    if (values[1]) {
      __pyx_v_update = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_update == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1660, __pyx_L3_error)
    } else {
      __pyx_v_update = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_membranes", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1660, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.get_membranes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_10get_membranes(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self), __pyx_v_cutoff, __pyx_v_update);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_10get_membranes(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self, real __pyx_v_cutoff, int __pyx_v_update) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_get_membranes __pyx_t_2;
  __Pyx_RefNannySetupContext("get_membranes", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.cutoff = __pyx_v_cutoff;
  __pyx_t_2.update = __pyx_v_update;
  __pyx_t_1 = __pyx_vtabptr_10fatslimlib_9core_base_Frame->get_membranes(__pyx_v_self, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.get_membranes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1689
 *     # Additional python properties
 *     property hg_atomids:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return np.asarray(self.trajectory.hg_group_atomids)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_10hg_atomids_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_10hg_atomids_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_10hg_atomids___get__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_10hg_atomids___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":1690
 *     property hg_atomids:
 *         def __get__(self):
 *             return np.asarray(self.trajectory.hg_group_atomids)             # <<<<<<<<<<<<<<
 * 
 *     property lipid_atomids:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_asarray); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_v_self->trajectory->hg_group_atomids.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1690, __pyx_L1_error)}
  __pyx_t_2 = __pyx_memoryview_fromslice(__pyx_v_self->trajectory->hg_group_atomids, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_fsl_int, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_fsl_int, 0);; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_2) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1689
 *     # Additional python properties
 *     property hg_atomids:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return np.asarray(self.trajectory.hg_group_atomids)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.hg_atomids.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1693
 * 
 *     property lipid_atomids:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef fsl_int i
 *             cdef fsl_int[:] offsets = self.trajectory.lipid_atomids_offsets
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_13lipid_atomids_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_13lipid_atomids_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_13lipid_atomids___get__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_13lipid_atomids___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  fsl_int __pyx_v_i;
  __Pyx_memviewslice __pyx_v_offsets = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_v_lipid_atomids = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  fsl_int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":1695
 *         def __get__(self):
 *             cdef fsl_int i
 *             cdef fsl_int[:] offsets = self.trajectory.lipid_atomids_offsets             # <<<<<<<<<<<<<<
 * 
 *             lipid_atomids = []
 */
  if (unlikely(!__pyx_v_self->trajectory->lipid_atomids_offsets.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1695, __pyx_L1_error)}
  __pyx_t_1 = __pyx_v_self->trajectory->lipid_atomids_offsets;
  __PYX_INC_MEMVIEW(&__pyx_t_1, 1);
  __pyx_v_offsets = __pyx_t_1;
  __pyx_t_1.memview = NULL;
  __pyx_t_1.data = NULL;

  /* "fatslimlib/core_base.pyx":1697
 *             cdef fsl_int[:] offsets = self.trajectory.lipid_atomids_offsets
 * 
 *             lipid_atomids = []             # <<<<<<<<<<<<<<
 *             for i in range(offsets.shape[0] - 1):
 *                 lipid_atomids.append(np.asarray(self.trajectory.
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_lipid_atomids = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":1698
 * 
 *             lipid_atomids = []
 *             for i in range(offsets.shape[0] - 1):             # <<<<<<<<<<<<<<
 *                 lipid_atomids.append(np.asarray(self.trajectory.
 *                                                 lipid_atomids[offsets[i]:offsets[i+1]]))
 */
  __pyx_t_3 = ((__pyx_v_offsets.shape[0]) - 1);
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "fatslimlib/core_base.pyx":1699
 *             lipid_atomids = []
 *             for i in range(offsets.shape[0] - 1):
 *                 lipid_atomids.append(np.asarray(self.trajectory.             # <<<<<<<<<<<<<<
 *                                                 lipid_atomids[offsets[i]:offsets[i+1]]))
 *             return lipid_atomids
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_asarray); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_v_self->trajectory->lipid_atomids.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1699, __pyx_L1_error)}

    /* "fatslimlib/core_base.pyx":1700
 *             for i in range(offsets.shape[0] - 1):
 *                 lipid_atomids.append(np.asarray(self.trajectory.
 *                                                 lipid_atomids[offsets[i]:offsets[i+1]]))             # <<<<<<<<<<<<<<
 *             return lipid_atomids
 * 
 */
    __pyx_t_8 = __pyx_v_i;
    __pyx_t_9 = (__pyx_v_i + 1);
    __pyx_t_1.data = __pyx_v_self->trajectory->lipid_atomids.data;
    __pyx_t_1.memview = __pyx_v_self->trajectory->lipid_atomids.memview;
    __PYX_INC_MEMVIEW(&__pyx_t_1, 0);
    __pyx_t_10 = -1;
    if (unlikely(__pyx_memoryview_slice_memviewslice(
    &__pyx_t_1,
    __pyx_v_self->trajectory->lipid_atomids.shape[0], __pyx_v_self->trajectory->lipid_atomids.strides[0], __pyx_v_self->trajectory->lipid_atomids.suboffsets[0],
    0,
    0,
    &__pyx_t_10,
    (*((fsl_int *) ( /* dim=0 */ (__pyx_v_offsets.data + __pyx_t_8 * __pyx_v_offsets.strides[0]) ))),
    (*((fsl_int *) ( /* dim=0 */ (__pyx_v_offsets.data + __pyx_t_9 * __pyx_v_offsets.strides[0]) ))),
    0,
    1,
    1,
    0,
    1) < 0))
{
    __PYX_ERR(0, 1700, __pyx_L1_error)
}

__pyx_t_6 = __pyx_memoryview_fromslice(__pyx_t_1, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_fsl_int, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_fsl_int, 0);; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __PYX_XDEC_MEMVIEW(&__pyx_t_1, 1);
    __pyx_t_1.memview = NULL;
    __pyx_t_1.data = NULL;
    __pyx_t_11 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_2 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_11, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "fatslimlib/core_base.pyx":1699
 *             lipid_atomids = []
 *             for i in range(offsets.shape[0] - 1):
 *                 lipid_atomids.append(np.asarray(self.trajectory.             # <<<<<<<<<<<<<<
 *                                                 lipid_atomids[offsets[i]:offsets[i+1]]))
 *             return lipid_atomids
 */
    __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_lipid_atomids, __pyx_t_2); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 1699, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "fatslimlib/core_base.pyx":1701
 *                 lipid_atomids.append(np.asarray(self.trajectory.
 *                                                 lipid_atomids[offsets[i]:offsets[i+1]]))
 *             return lipid_atomids             # <<<<<<<<<<<<<<
 * 
 *     property index:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_lipid_atomids);
  __pyx_r = __pyx_v_lipid_atomids;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1693
 * 
 *     property lipid_atomids:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             cdef fsl_int i
 *             cdef fsl_int[:] offsets = self.trajectory.lipid_atomids_offsets
 */

  /* function exit code */
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.lipid_atomids.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_offsets, 1);
  __Pyx_XDECREF(__pyx_v_lipid_atomids);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1704
 * 
 *     property index:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.index
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_5index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_5index_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_5index___get__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_5index___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "fatslimlib/core_base.pyx":1705
 *     property index:
 *         def __get__(self):
 *             return self.index             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_v_self->index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1704
 * 
 *     property index:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self.index
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pxd":216
 *     # Attributes
 *     cdef fsl_int index
 *     cdef readonly real timestep             # <<<<<<<<<<<<<<
 *     cdef readonly Trajectory trajectory
 *     cdef readonly Topology topology
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_8timestep_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_8timestep_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_8timestep___get__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_8timestep___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->timestep); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.timestep.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pxd":217
 *     cdef fsl_int index
 *     cdef readonly real timestep
 *     cdef readonly Trajectory trajectory             # <<<<<<<<<<<<<<
 *     cdef readonly Topology topology
 *     cdef CoordinateReader coords_reader
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_10trajectory_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_10trajectory_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_10trajectory___get__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_10trajectory___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->trajectory));
  __pyx_r = ((PyObject *)__pyx_v_self->trajectory);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pxd":218
 *     cdef readonly real timestep
 *     cdef readonly Trajectory trajectory
 *     cdef readonly Topology topology             # <<<<<<<<<<<<<<
 *     cdef CoordinateReader coords_reader
 *     cdef readonly PBCBox box
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_8topology_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_8topology_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_8topology___get__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_8topology___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->topology));
  __pyx_r = ((PyObject *)__pyx_v_self->topology);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pxd":220
 *     cdef readonly Topology topology
 *     cdef CoordinateReader coords_reader
 *     cdef readonly PBCBox box             # <<<<<<<<<<<<<<
 *     cdef readonly fsl_int size
 *     # Headgroups
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_3box_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_3box_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_3box___get__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_3box___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->box));
  __pyx_r = ((PyObject *)__pyx_v_self->box);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pxd":221
 *     cdef CoordinateReader coords_reader
 *     cdef readonly PBCBox box
 *     cdef readonly fsl_int size             # <<<<<<<<<<<<<<
 *     # Headgroups
 *     cdef real[:, ::1] hg_group_coords, hg_coords_bbox
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_4size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_4size___get__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_4size___get__(struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_v_self->size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.neighbors cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_13__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_13__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_12__reduce_cython__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.neighbors cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.neighbors cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.neighbors cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.neighbors cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.neighbors cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_15__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_5Frame_15__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_5Frame_14__setstate_cython__(((struct __pyx_obj_10fatslimlib_9core_base_Frame *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_5Frame_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10fatslimlib_9core_base_Frame *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.neighbors cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.neighbors cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.neighbors cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.neighbors cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Frame.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1715
 * 
 * cdef class Trajectory(object):
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  TopologyReader top_reader,
 *                  IndexReader index_reader,
 */

/* Python wrapper */
static int __pyx_pw_10fatslimlib_9core_base_10Trajectory_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10fatslimlib_9core_base_10Trajectory_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *__pyx_v_top_reader = 0;
  struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_index_reader = 0;
  struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_coords_reader = 0;
  int __pyx_v_be_verbose;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_top_reader,&__pyx_n_s_index_reader,&__pyx_n_s_coords_reader,&__pyx_n_s_be_verbose,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_top_reader)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_index_reader)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, 1); __PYX_ERR(0, 1715, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_coords_reader)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, 2); __PYX_ERR(0, 1715, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_be_verbose);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1715, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_top_reader = ((struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *)values[0]);
    __pyx_v_index_reader = ((struct __pyx_obj_10fatslimlib_9core_base_IndexReader *)values[1]);
    __pyx_v_coords_reader = ((struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *)values[2]);
    if (values[3]) {
      __pyx_v_be_verbose = __Pyx_PyObject_IsTrue(values[3]); if (unlikely((__pyx_v_be_verbose == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1719, __pyx_L3_error)
    } else {

      /* "fatslimlib/core_base.pyx":1719
 *                  IndexReader index_reader,
 *                  CoordinateReader coords_reader,
 *                  bint be_verbose=True):             # <<<<<<<<<<<<<<
 *         self.be_verbose = be_verbose
 *         self.bead_group_name = None
 */
      __pyx_v_be_verbose = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 3, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1715, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.Trajectory.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_top_reader), __pyx_ptype_10fatslimlib_9core_base_TopologyReader, 1, "top_reader", 0))) __PYX_ERR(0, 1716, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_index_reader), __pyx_ptype_10fatslimlib_9core_base_IndexReader, 1, "index_reader", 0))) __PYX_ERR(0, 1717, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_coords_reader), __pyx_ptype_10fatslimlib_9core_base_CoordinateReader, 1, "coords_reader", 0))) __PYX_ERR(0, 1718, __pyx_L1_error)
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_10Trajectory___init__(((struct __pyx_obj_10fatslimlib_9core_base_Trajectory *)__pyx_v_self), __pyx_v_top_reader, __pyx_v_index_reader, __pyx_v_coords_reader, __pyx_v_be_verbose);

  /* "fatslimlib/core_base.pyx":1715
 * 
 * cdef class Trajectory(object):
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  TopologyReader top_reader,
 *                  IndexReader index_reader,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10fatslimlib_9core_base_10Trajectory___init__(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self, struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *__pyx_v_top_reader, struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v_index_reader, struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v_coords_reader, int __pyx_v_be_verbose) {
  CYTHON_UNUSED long __pyx_v_cur_frame;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_memviewslice __pyx_t_2 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "fatslimlib/core_base.pyx":1720
 *                  CoordinateReader coords_reader,
 *                  bint be_verbose=True):
 *         self.be_verbose = be_verbose             # <<<<<<<<<<<<<<
 *         self.bead_group_name = None
 *         self.interacting_group_name = None
 */
  __pyx_v_self->be_verbose = __pyx_v_be_verbose;

  /* "fatslimlib/core_base.pyx":1721
 *                  bint be_verbose=True):
 *         self.be_verbose = be_verbose
 *         self.bead_group_name = None             # <<<<<<<<<<<<<<
 *         self.interacting_group_name = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->bead_group_name);
  __Pyx_DECREF(__pyx_v_self->bead_group_name);
  __pyx_v_self->bead_group_name = ((PyObject*)Py_None);

  /* "fatslimlib/core_base.pyx":1722
 *         self.be_verbose = be_verbose
 *         self.bead_group_name = None
 *         self.interacting_group_name = None             # <<<<<<<<<<<<<<
 * 
 *         assert isinstance(top_reader, TopologyReader)
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->interacting_group_name);
  __Pyx_DECREF(__pyx_v_self->interacting_group_name);
  __pyx_v_self->interacting_group_name = ((PyObject*)Py_None);

  /* "fatslimlib/core_base.pyx":1724
 *         self.interacting_group_name = None
 * 
 *         assert isinstance(top_reader, TopologyReader)             # <<<<<<<<<<<<<<
 *         self.topol_reader = top_reader
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_top_reader), __pyx_ptype_10fatslimlib_9core_base_TopologyReader); 
    if (unlikely(!(__pyx_t_1 != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 1724, __pyx_L1_error)
    }
  }
  #endif

  /* "fatslimlib/core_base.pyx":1725
 * 
 *         assert isinstance(top_reader, TopologyReader)
 *         self.topol_reader = top_reader             # <<<<<<<<<<<<<<
 * 
 *         assert isinstance(index_reader, IndexReader)
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_top_reader));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_top_reader));
  __Pyx_GOTREF(__pyx_v_self->topol_reader);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->topol_reader));
  __pyx_v_self->topol_reader = __pyx_v_top_reader;

  /* "fatslimlib/core_base.pyx":1727
 *         self.topol_reader = top_reader
 * 
 *         assert isinstance(index_reader, IndexReader)             # <<<<<<<<<<<<<<
 *         self.index_reader = index_reader
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_index_reader), __pyx_ptype_10fatslimlib_9core_base_IndexReader); 
    if (unlikely(!(__pyx_t_1 != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 1727, __pyx_L1_error)
    }
  }
  #endif

  /* "fatslimlib/core_base.pyx":1728
 * 
 *         assert isinstance(index_reader, IndexReader)
 *         self.index_reader = index_reader             # <<<<<<<<<<<<<<
 * 
 *         assert isinstance(coords_reader, CoordinateReader)
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_index_reader));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_index_reader));
  __Pyx_GOTREF(__pyx_v_self->index_reader);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->index_reader));
  __pyx_v_self->index_reader = __pyx_v_index_reader;

  /* "fatslimlib/core_base.pyx":1730
 *         self.index_reader = index_reader
 * 
 *         assert isinstance(coords_reader, CoordinateReader)             # <<<<<<<<<<<<<<
 *         self.coords_reader = coords_reader
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_coords_reader), __pyx_ptype_10fatslimlib_9core_base_CoordinateReader); 
    if (unlikely(!(__pyx_t_1 != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(0, 1730, __pyx_L1_error)
    }
  }
  #endif

  /* "fatslimlib/core_base.pyx":1731
 * 
 *         assert isinstance(coords_reader, CoordinateReader)
 *         self.coords_reader = coords_reader             # <<<<<<<<<<<<<<
 * 
 *         # Load atomids & group
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_coords_reader));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_coords_reader));
  __Pyx_GOTREF(__pyx_v_self->coords_reader);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->coords_reader));
  __pyx_v_self->coords_reader = __pyx_v_coords_reader;

  /* "fatslimlib/core_base.pyx":1734
 * 
 *         # Load atomids & group
 *         self.hg_group_atomids = None             # <<<<<<<<<<<<<<
 *         self.lipid_atomids = None
 *         self.lipid_atomids_offsets = None
 */
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1734, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->hg_group_atomids, 0);
  __pyx_v_self->hg_group_atomids = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "fatslimlib/core_base.pyx":1735
 *         # Load atomids & group
 *         self.hg_group_atomids = None
 *         self.lipid_atomids = None             # <<<<<<<<<<<<<<
 *         self.lipid_atomids_offsets = None
 *         self.interacting_atomids = None
 */
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1735, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->lipid_atomids, 0);
  __pyx_v_self->lipid_atomids = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "fatslimlib/core_base.pyx":1736
 *         self.hg_group_atomids = None
 *         self.lipid_atomids = None
 *         self.lipid_atomids_offsets = None             # <<<<<<<<<<<<<<
 *         self.interacting_atomids = None
 *         self.hg_bead_atomids = None
 */
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1736, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->lipid_atomids_offsets, 0);
  __pyx_v_self->lipid_atomids_offsets = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "fatslimlib/core_base.pyx":1737
 *         self.lipid_atomids = None
 *         self.lipid_atomids_offsets = None
 *         self.interacting_atomids = None             # <<<<<<<<<<<<<<
 *         self.hg_bead_atomids = None
 *         self.hg_bead_atomids_offsets = None
 */
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1737, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->interacting_atomids, 0);
  __pyx_v_self->interacting_atomids = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "fatslimlib/core_base.pyx":1738
 *         self.lipid_atomids_offsets = None
 *         self.interacting_atomids = None
 *         self.hg_bead_atomids = None             # <<<<<<<<<<<<<<
 *         self.hg_bead_atomids_offsets = None
 *         self.forcefield_type = fft_notset
 */
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1738, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->hg_bead_atomids, 0);
  __pyx_v_self->hg_bead_atomids = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "fatslimlib/core_base.pyx":1739
 *         self.interacting_atomids = None
 *         self.hg_bead_atomids = None
 *         self.hg_bead_atomids_offsets = None             # <<<<<<<<<<<<<<
 *         self.forcefield_type = fft_notset
 *         cur_frame = 0
 */
  __pyx_t_2 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(Py_None, PyBUF_WRITABLE); if (unlikely(!__pyx_t_2.memview)) __PYX_ERR(0, 1739, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->hg_bead_atomids_offsets, 0);
  __pyx_v_self->hg_bead_atomids_offsets = __pyx_t_2;
  __pyx_t_2.memview = NULL;
  __pyx_t_2.data = NULL;

  /* "fatslimlib/core_base.pyx":1740
 *         self.hg_bead_atomids = None
 *         self.hg_bead_atomids_offsets = None
 *         self.forcefield_type = fft_notset             # <<<<<<<<<<<<<<
 *         cur_frame = 0
 * 
 */
  __pyx_v_self->forcefield_type = __pyx_e_10fatslimlib_9core_base_fft_notset;

  /* "fatslimlib/core_base.pyx":1741
 *         self.hg_bead_atomids_offsets = None
 *         self.forcefield_type = fft_notset
 *         cur_frame = 0             # <<<<<<<<<<<<<<
 * 
 *         # Cache related
 */
  __pyx_v_cur_frame = 0;

  /* "fatslimlib/core_base.pyx":1744
 * 
 *         # Cache related
 *         self.normal_cutoff = DEFAULT_PROXIMITY_CUTOFF             # <<<<<<<<<<<<<<
 *         self.membranes = None
 *         self.aggregates = None
 */
  __pyx_v_self->normal_cutoff = 2.0;

  /* "fatslimlib/core_base.pyx":1745
 *         # Cache related
 *         self.normal_cutoff = DEFAULT_PROXIMITY_CUTOFF
 *         self.membranes = None             # <<<<<<<<<<<<<<
 *         self.aggregates = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->membranes);
  __Pyx_DECREF(__pyx_v_self->membranes);
  __pyx_v_self->membranes = ((PyObject*)Py_None);

  /* "fatslimlib/core_base.pyx":1746
 *         self.normal_cutoff = DEFAULT_PROXIMITY_CUTOFF
 *         self.membranes = None
 *         self.aggregates = None             # <<<<<<<<<<<<<<
 * 
 *         self.topology = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->aggregates);
  __Pyx_DECREF(__pyx_v_self->aggregates);
  __pyx_v_self->aggregates = ((PyObject*)Py_None);

  /* "fatslimlib/core_base.pyx":1748
 *         self.aggregates = None
 * 
 *         self.topology = None             # <<<<<<<<<<<<<<
 *         self.initialized = False
 *         self.timesteps = []
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->topology);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->topology));
  __pyx_v_self->topology = ((struct __pyx_obj_10fatslimlib_9core_base_Topology *)Py_None);

  /* "fatslimlib/core_base.pyx":1749
 * 
 *         self.topology = None
 *         self.initialized = False             # <<<<<<<<<<<<<<
 *         self.timesteps = []
 * 
 */
  __pyx_v_self->initialized = 0;

  /* "fatslimlib/core_base.pyx":1750
 *         self.topology = None
 *         self.initialized = False
 *         self.timesteps = []             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->timesteps);
  __Pyx_DECREF(__pyx_v_self->timesteps);
  __pyx_v_self->timesteps = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":1715
 * 
 * cdef class Trajectory(object):
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  TopologyReader top_reader,
 *                  IndexReader index_reader,
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_2, 1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("fatslimlib.core_base.Trajectory.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1753
 * 
 * 
 *     def initialize(self, str hg_group="headgroups", str interacting_group="protein"):             # <<<<<<<<<<<<<<
 *         cdef fsl_int size, i, resid_size
 *         cdef fsl_int last_resid, resid
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_10Trajectory_3initialize(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_10Trajectory_3initialize(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_hg_group = 0;
  PyObject *__pyx_v_interacting_group = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("initialize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_hg_group,&__pyx_n_s_interacting_group,0};
    PyObject* values[2] = {0,0};
    values[0] = ((PyObject*)__pyx_n_u_headgroups);
    values[1] = ((PyObject*)__pyx_n_u_protein);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hg_group);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_interacting_group);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "initialize") < 0)) __PYX_ERR(0, 1753, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_hg_group = ((PyObject*)values[0]);
    __pyx_v_interacting_group = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("initialize", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1753, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.Trajectory.initialize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hg_group), (&PyUnicode_Type), 1, "hg_group", 1))) __PYX_ERR(0, 1753, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_interacting_group), (&PyUnicode_Type), 1, "interacting_group", 1))) __PYX_ERR(0, 1753, __pyx_L1_error)
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_10Trajectory_2initialize(((struct __pyx_obj_10fatslimlib_9core_base_Trajectory *)__pyx_v_self), __pyx_v_hg_group, __pyx_v_interacting_group);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_10Trajectory_2initialize(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self, PyObject *__pyx_v_hg_group, PyObject *__pyx_v_interacting_group) {
  fsl_int __pyx_v_size;
  fsl_int __pyx_v_i;
  fsl_int __pyx_v_resid_size;
  fsl_int __pyx_v_last_resid;
  fsl_int __pyx_v_resid;
  __Pyx_memviewslice __pyx_v_hg_group_atomids = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_hg_bead_atomids_offsets = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lipid_atomids_offsets = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lipid_atomids = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_resids = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_v_lipid_hg_indices = { 0, 0, { 0 }, { 0 }, { 0 } };
  struct __pyx_obj_10fatslimlib_9core_base_Topology *__pyx_v_topology = 0;
  struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *__pyx_v_residue;
  struct __pyx_t_10fatslimlib_9core_base_topol_atom_t *__pyx_v_atom;
  PyObject *__pyx_v_residues_dict = 0;
  fsl_int __pyx_v_atom_index;
  PyObject *__pyx_v_begin = NULL;
  PyObject *__pyx_v_lipid_atomids_list = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_UCS4 __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_memviewslice __pyx_t_7 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  fsl_int __pyx_t_10;
  fsl_int __pyx_t_11;
  fsl_int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  fsl_int __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  int __pyx_t_21;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  int __pyx_t_25;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  PyObject *__pyx_t_29 = NULL;
  PyObject *__pyx_t_30 = NULL;
  PyObject *__pyx_t_31 = NULL;
  PyObject *__pyx_t_32 = NULL;
  Py_ssize_t __pyx_t_33;
  Py_ssize_t __pyx_t_34;
  Py_ssize_t __pyx_t_35;
  Py_ssize_t __pyx_t_36;
  Py_ssize_t __pyx_t_37;
  Py_ssize_t __pyx_t_38;
  __Pyx_RefNannySetupContext("initialize", 0);

  /* "fatslimlib/core_base.pyx":1767
 * 
 *         # Preload file
 *         verbose_print("Initializing trajectory using groups: '%s' and '%s'... "             # <<<<<<<<<<<<<<
 *                       % (hg_group, interacting_group),
 *                       self.be_verbose,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_verbose_print); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = 0;
  __pyx_t_4 = 127;
  __Pyx_INCREF(__pyx_kp_u_Initializing_trajectory_using_gr);
  __pyx_t_3 += 39;
  __Pyx_GIVEREF(__pyx_kp_u_Initializing_trajectory_using_gr);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_Initializing_trajectory_using_gr);

  /* "fatslimlib/core_base.pyx":1768
 *         # Preload file
 *         verbose_print("Initializing trajectory using groups: '%s' and '%s'... "
 *                       % (hg_group, interacting_group),             # <<<<<<<<<<<<<<
 *                       self.be_verbose,
 *                       end="")
 */
  __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_v_hg_group); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_4;
  __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u_and);
  __pyx_t_3 += 7;
  __Pyx_GIVEREF(__pyx_kp_u_and);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_and);
  __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_v_interacting_group); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_4;
  __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u__25);
  __pyx_t_3 += 5;
  __Pyx_GIVEREF(__pyx_kp_u__25);
  PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_kp_u__25);

  /* "fatslimlib/core_base.pyx":1767
 * 
 *         # Preload file
 *         verbose_print("Initializing trajectory using groups: '%s' and '%s'... "             # <<<<<<<<<<<<<<
 *                       % (hg_group, interacting_group),
 *                       self.be_verbose,
 */
  __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_2, 5, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":1769
 *         verbose_print("Initializing trajectory using groups: '%s' and '%s'... "
 *                       % (hg_group, interacting_group),
 *                       self.be_verbose,             # <<<<<<<<<<<<<<
 *                       end="")
 *         sys.stdout.flush()
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->be_verbose); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "fatslimlib/core_base.pyx":1767
 * 
 *         # Preload file
 *         verbose_print("Initializing trajectory using groups: '%s' and '%s'... "             # <<<<<<<<<<<<<<
 *                       % (hg_group, interacting_group),
 *                       self.be_verbose,
 */
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
  __pyx_t_5 = 0;
  __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":1770
 *                       % (hg_group, interacting_group),
 *                       self.be_verbose,
 *                       end="")             # <<<<<<<<<<<<<<
 *         sys.stdout.flush()
 *         begin = time()
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_end, __pyx_kp_u__22) < 0) __PYX_ERR(0, 1770, __pyx_L1_error)

  /* "fatslimlib/core_base.pyx":1767
 * 
 *         # Preload file
 *         verbose_print("Initializing trajectory using groups: '%s' and '%s'... "             # <<<<<<<<<<<<<<
 *                       % (hg_group, interacting_group),
 *                       self.be_verbose,
 */
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fatslimlib/core_base.pyx":1771
 *                       self.be_verbose,
 *                       end="")
 *         sys.stdout.flush()             # <<<<<<<<<<<<<<
 *         begin = time()
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_stdout); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_flush); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fatslimlib/core_base.pyx":1772
 *                       end="")
 *         sys.stdout.flush()
 *         begin = time()             # <<<<<<<<<<<<<<
 * 
 *         topology = self.topol_reader.topology
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_begin = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "fatslimlib/core_base.pyx":1774
 *         begin = time()
 * 
 *         topology = self.topol_reader.topology             # <<<<<<<<<<<<<<
 *         self.topology = topology
 * 
 */
  __pyx_t_5 = ((PyObject *)__pyx_v_self->topol_reader->topology);
  __Pyx_INCREF(__pyx_t_5);
  __pyx_v_topology = ((struct __pyx_obj_10fatslimlib_9core_base_Topology *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "fatslimlib/core_base.pyx":1775
 * 
 *         topology = self.topol_reader.topology
 *         self.topology = topology             # <<<<<<<<<<<<<<
 * 
 *         # Handle headgroups
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_topology));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_topology));
  __Pyx_GOTREF(__pyx_v_self->topology);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->topology));
  __pyx_v_self->topology = __pyx_v_topology;

  /* "fatslimlib/core_base.pyx":1778
 * 
 *         # Handle headgroups
 *         self.bead_group_name = hg_group.encode()             # <<<<<<<<<<<<<<
 *         hg_group_atomids = self.index_reader[hg_group]
 *         self.hg_group_atomids = hg_group_atomids
 */
  if (unlikely(__pyx_v_hg_group == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 1778, __pyx_L1_error)
  }
  __pyx_t_5 = PyUnicode_AsEncodedString(__pyx_v_hg_group, NULL, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF(__pyx_v_self->bead_group_name);
  __Pyx_DECREF(__pyx_v_self->bead_group_name);
  __pyx_v_self->bead_group_name = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "fatslimlib/core_base.pyx":1779
 *         # Handle headgroups
 *         self.bead_group_name = hg_group.encode()
 *         hg_group_atomids = self.index_reader[hg_group]             # <<<<<<<<<<<<<<
 *         self.hg_group_atomids = hg_group_atomids
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self->index_reader), __pyx_v_hg_group); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1779, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_hg_group_atomids = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "fatslimlib/core_base.pyx":1780
 *         self.bead_group_name = hg_group.encode()
 *         hg_group_atomids = self.index_reader[hg_group]
 *         self.hg_group_atomids = hg_group_atomids             # <<<<<<<<<<<<<<
 * 
 *         # Group by resid and retrieve offsets
 */
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->hg_group_atomids, 0);
  __PYX_INC_MEMVIEW(&__pyx_v_hg_group_atomids, 0);
  __pyx_v_self->hg_group_atomids = __pyx_v_hg_group_atomids;

  /* "fatslimlib/core_base.pyx":1783
 * 
 *         # Group by resid and retrieve offsets
 *         size = hg_group_atomids.shape[0]             # <<<<<<<<<<<<<<
 *         resids = np.empty(size, dtype=np.int64)
 *         hg_bead_atomids_offsets = np.empty(size+1, dtype=np.int64)
 */
  __pyx_v_size = (__pyx_v_hg_group_atomids.shape[0]);

  /* "fatslimlib/core_base.pyx":1784
 *         # Group by resid and retrieve offsets
 *         size = hg_group_atomids.shape[0]
 *         resids = np.empty(size, dtype=np.int64)             # <<<<<<<<<<<<<<
 *         hg_bead_atomids_offsets = np.empty(size+1, dtype=np.int64)
 *         lipid_atomids_offsets = np.empty(size+1, dtype=np.int64)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_empty); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_fsl_int(__pyx_v_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_int64); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 1784, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1784, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_v_resids = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "fatslimlib/core_base.pyx":1785
 *         size = hg_group_atomids.shape[0]
 *         resids = np.empty(size, dtype=np.int64)
 *         hg_bead_atomids_offsets = np.empty(size+1, dtype=np.int64)             # <<<<<<<<<<<<<<
 *         lipid_atomids_offsets = np.empty(size+1, dtype=np.int64)
 *         lipid_hg_indices = np.empty(size, dtype=np.int64)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_empty); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyInt_From_long((__pyx_v_size + 1)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_8);
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_int64); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_dtype, __pyx_t_1) < 0) __PYX_ERR(0, 1785, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(__pyx_t_1, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1785, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_hg_bead_atomids_offsets = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "fatslimlib/core_base.pyx":1786
 *         resids = np.empty(size, dtype=np.int64)
 *         hg_bead_atomids_offsets = np.empty(size+1, dtype=np.int64)
 *         lipid_atomids_offsets = np.empty(size+1, dtype=np.int64)             # <<<<<<<<<<<<<<
 *         lipid_hg_indices = np.empty(size, dtype=np.int64)
 *         lipid_atomids_list = []
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_empty); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_long((__pyx_v_size + 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_int64); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dtype, __pyx_t_2) < 0) __PYX_ERR(0, 1786, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(__pyx_t_2, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1786, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_lipid_atomids_offsets = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "fatslimlib/core_base.pyx":1787
 *         hg_bead_atomids_offsets = np.empty(size+1, dtype=np.int64)
 *         lipid_atomids_offsets = np.empty(size+1, dtype=np.int64)
 *         lipid_hg_indices = np.empty(size, dtype=np.int64)             # <<<<<<<<<<<<<<
 *         lipid_atomids_list = []
 *         last_resid = -1
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_empty); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_fsl_int(__pyx_v_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_int64); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_5) < 0) __PYX_ERR(0, 1787, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(__pyx_t_5, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1787, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_lipid_hg_indices = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "fatslimlib/core_base.pyx":1788
 *         lipid_atomids_offsets = np.empty(size+1, dtype=np.int64)
 *         lipid_hg_indices = np.empty(size, dtype=np.int64)
 *         lipid_atomids_list = []             # <<<<<<<<<<<<<<
 *         last_resid = -1
 *         lipid_atomids_offsets[0] = 0
 */
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_lipid_atomids_list = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "fatslimlib/core_base.pyx":1789
 *         lipid_hg_indices = np.empty(size, dtype=np.int64)
 *         lipid_atomids_list = []
 *         last_resid = -1             # <<<<<<<<<<<<<<
 *         lipid_atomids_offsets[0] = 0
 *         resid_size = 0
 */
  __pyx_v_last_resid = -1;

  /* "fatslimlib/core_base.pyx":1790
 *         lipid_atomids_list = []
 *         last_resid = -1
 *         lipid_atomids_offsets[0] = 0             # <<<<<<<<<<<<<<
 *         resid_size = 0
 * 
 */
  __pyx_t_9 = 0;
  *((fsl_int *) ( /* dim=0 */ (__pyx_v_lipid_atomids_offsets.data + __pyx_t_9 * __pyx_v_lipid_atomids_offsets.strides[0]) )) = 0;

  /* "fatslimlib/core_base.pyx":1791
 *         last_resid = -1
 *         lipid_atomids_offsets[0] = 0
 *         resid_size = 0             # <<<<<<<<<<<<<<
 * 
 *         residues_dict = {}
 */
  __pyx_v_resid_size = 0;

  /* "fatslimlib/core_base.pyx":1793
 *         resid_size = 0
 * 
 *         residues_dict = {}             # <<<<<<<<<<<<<<
 *         for i in range(size):
 *             atom = topology.fast_get_atom(hg_group_atomids[i])
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_residues_dict = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "fatslimlib/core_base.pyx":1794
 * 
 *         residues_dict = {}
 *         for i in range(size):             # <<<<<<<<<<<<<<
 *             atom = topology.fast_get_atom(hg_group_atomids[i])
 *             if atom == NULL:
 */
  __pyx_t_10 = __pyx_v_size;
  __pyx_t_11 = __pyx_t_10;
  for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
    __pyx_v_i = __pyx_t_12;

    /* "fatslimlib/core_base.pyx":1795
 *         residues_dict = {}
 *         for i in range(size):
 *             atom = topology.fast_get_atom(hg_group_atomids[i])             # <<<<<<<<<<<<<<
 *             if atom == NULL:
 *                 raise KeyError("Atomid not registered in topology: %i. Does group '%s' from '%s'"
 */
    __pyx_t_13 = __pyx_v_i;
    __pyx_v_atom = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Topology *)__pyx_v_topology->__pyx_vtab)->fast_get_atom(__pyx_v_topology, (*((fsl_int *) ( /* dim=0 */ (__pyx_v_hg_group_atomids.data + __pyx_t_13 * __pyx_v_hg_group_atomids.strides[0]) ))));

    /* "fatslimlib/core_base.pyx":1796
 *         for i in range(size):
 *             atom = topology.fast_get_atom(hg_group_atomids[i])
 *             if atom == NULL:             # <<<<<<<<<<<<<<
 *                 raise KeyError("Atomid not registered in topology: %i. Does group '%s' from '%s'"
 *                                    " file contains H?" % (hg_group_atomids[i],
 */
    __pyx_t_14 = ((__pyx_v_atom == NULL) != 0);
    if (unlikely(__pyx_t_14)) {

      /* "fatslimlib/core_base.pyx":1798
 *             if atom == NULL:
 *                 raise KeyError("Atomid not registered in topology: %i. Does group '%s' from '%s'"
 *                                    " file contains H?" % (hg_group_atomids[i],             # <<<<<<<<<<<<<<
 *                                                           hg_group,
 *                                                           self.index_reader.filename))
 */
      __pyx_t_15 = __pyx_v_i;
      __pyx_t_5 = __Pyx_PyInt_From_fsl_int((*((fsl_int *) ( /* dim=0 */ (__pyx_v_hg_group_atomids.data + __pyx_t_15 * __pyx_v_hg_group_atomids.strides[0]) )))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1798, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "fatslimlib/core_base.pyx":1800
 *                                    " file contains H?" % (hg_group_atomids[i],
 *                                                           hg_group,
 *                                                           self.index_reader.filename))             # <<<<<<<<<<<<<<
 *             residue = &topology.residues[atom.residue_internal_id]
 * 
 */
      __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1798, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
      __Pyx_INCREF(__pyx_v_hg_group);
      __Pyx_GIVEREF(__pyx_v_hg_group);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_hg_group);
      __Pyx_INCREF(__pyx_v_self->index_reader->filename);
      __Pyx_GIVEREF(__pyx_v_self->index_reader->filename);
      PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_self->index_reader->filename);
      __pyx_t_5 = 0;

      /* "fatslimlib/core_base.pyx":1798
 *             if atom == NULL:
 *                 raise KeyError("Atomid not registered in topology: %i. Does group '%s' from '%s'"
 *                                    " file contains H?" % (hg_group_atomids[i],             # <<<<<<<<<<<<<<
 *                                                           hg_group,
 *                                                           self.index_reader.filename))
 */
      __pyx_t_5 = PyUnicode_Format(__pyx_kp_u_Atomid_not_registered_in_topolog, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1798, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "fatslimlib/core_base.pyx":1797
 *             atom = topology.fast_get_atom(hg_group_atomids[i])
 *             if atom == NULL:
 *                 raise KeyError("Atomid not registered in topology: %i. Does group '%s' from '%s'"             # <<<<<<<<<<<<<<
 *                                    " file contains H?" % (hg_group_atomids[i],
 *                                                           hg_group,
 */
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1797, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 1797, __pyx_L1_error)

      /* "fatslimlib/core_base.pyx":1796
 *         for i in range(size):
 *             atom = topology.fast_get_atom(hg_group_atomids[i])
 *             if atom == NULL:             # <<<<<<<<<<<<<<
 *                 raise KeyError("Atomid not registered in topology: %i. Does group '%s' from '%s'"
 *                                    " file contains H?" % (hg_group_atomids[i],
 */
    }

    /* "fatslimlib/core_base.pyx":1801
 *                                                           hg_group,
 *                                                           self.index_reader.filename))
 *             residue = &topology.residues[atom.residue_internal_id]             # <<<<<<<<<<<<<<
 * 
 *             resid = residue.resid
 */
    __pyx_v_residue = (&(__pyx_v_topology->residues[__pyx_v_atom->residue_internal_id]));

    /* "fatslimlib/core_base.pyx":1803
 *             residue = &topology.residues[atom.residue_internal_id]
 * 
 *             resid = residue.resid             # <<<<<<<<<<<<<<
 * 
 *             if resid > last_resid: # WARNING: It is assumed that atomids (hence resids) are sorted!
 */
    __pyx_t_16 = __pyx_v_residue->resid;
    __pyx_v_resid = __pyx_t_16;

    /* "fatslimlib/core_base.pyx":1805
 *             resid = residue.resid
 * 
 *             if resid > last_resid: # WARNING: It is assumed that atomids (hence resids) are sorted!             # <<<<<<<<<<<<<<
 *                 resids[resid_size] = resid
 *                 hg_bead_atomids_offsets[resid_size] = i
 */
    __pyx_t_14 = ((__pyx_v_resid > __pyx_v_last_resid) != 0);
    if (__pyx_t_14) {

      /* "fatslimlib/core_base.pyx":1806
 * 
 *             if resid > last_resid: # WARNING: It is assumed that atomids (hence resids) are sorted!
 *                 resids[resid_size] = resid             # <<<<<<<<<<<<<<
 *                 hg_bead_atomids_offsets[resid_size] = i
 *                 lipid_atomids_offsets[resid_size + 1] = lipid_atomids_offsets[resid_size] \
 */
      __pyx_t_17 = __pyx_v_resid_size;
      *((fsl_int *) ( /* dim=0 */ (__pyx_v_resids.data + __pyx_t_17 * __pyx_v_resids.strides[0]) )) = __pyx_v_resid;

      /* "fatslimlib/core_base.pyx":1807
 *             if resid > last_resid: # WARNING: It is assumed that atomids (hence resids) are sorted!
 *                 resids[resid_size] = resid
 *                 hg_bead_atomids_offsets[resid_size] = i             # <<<<<<<<<<<<<<
 *                 lipid_atomids_offsets[resid_size + 1] = lipid_atomids_offsets[resid_size] \
 *                                                         + residue.size
 */
      __pyx_t_18 = __pyx_v_resid_size;
      *((fsl_int *) ( /* dim=0 */ (__pyx_v_hg_bead_atomids_offsets.data + __pyx_t_18 * __pyx_v_hg_bead_atomids_offsets.strides[0]) )) = __pyx_v_i;

      /* "fatslimlib/core_base.pyx":1808
 *                 resids[resid_size] = resid
 *                 hg_bead_atomids_offsets[resid_size] = i
 *                 lipid_atomids_offsets[resid_size + 1] = lipid_atomids_offsets[resid_size] \             # <<<<<<<<<<<<<<
 *                                                         + residue.size
 *                 lipid_atomids_list.extend(topol_residue_atomids_as_list(residue))
 */
      __pyx_t_19 = __pyx_v_resid_size;

      /* "fatslimlib/core_base.pyx":1809
 *                 hg_bead_atomids_offsets[resid_size] = i
 *                 lipid_atomids_offsets[resid_size + 1] = lipid_atomids_offsets[resid_size] \
 *                                                         + residue.size             # <<<<<<<<<<<<<<
 *                 lipid_atomids_list.extend(topol_residue_atomids_as_list(residue))
 *                 last_resid = resid
 */
      __pyx_t_20 = (__pyx_v_resid_size + 1);
      *((fsl_int *) ( /* dim=0 */ (__pyx_v_lipid_atomids_offsets.data + __pyx_t_20 * __pyx_v_lipid_atomids_offsets.strides[0]) )) = ((*((fsl_int *) ( /* dim=0 */ (__pyx_v_lipid_atomids_offsets.data + __pyx_t_19 * __pyx_v_lipid_atomids_offsets.strides[0]) ))) + __pyx_v_residue->size);

      /* "fatslimlib/core_base.pyx":1810
 *                 lipid_atomids_offsets[resid_size + 1] = lipid_atomids_offsets[resid_size] \
 *                                                         + residue.size
 *                 lipid_atomids_list.extend(topol_residue_atomids_as_list(residue))             # <<<<<<<<<<<<<<
 *                 last_resid = resid
 *                 resid_size += 1
 */
      __pyx_t_2 = __pyx_f_10fatslimlib_9core_base_topol_residue_atomids_as_list(__pyx_v_residue); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1810, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_21 = __Pyx_PyList_Extend(__pyx_v_lipid_atomids_list, __pyx_t_2); if (unlikely(__pyx_t_21 == ((int)-1))) __PYX_ERR(0, 1810, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "fatslimlib/core_base.pyx":1811
 *                                                         + residue.size
 *                 lipid_atomids_list.extend(topol_residue_atomids_as_list(residue))
 *                 last_resid = resid             # <<<<<<<<<<<<<<
 *                 resid_size += 1
 * 
 */
      __pyx_v_last_resid = __pyx_v_resid;

      /* "fatslimlib/core_base.pyx":1812
 *                 lipid_atomids_list.extend(topol_residue_atomids_as_list(residue))
 *                 last_resid = resid
 *                 resid_size += 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_resid_size = (__pyx_v_resid_size + 1);

      /* "fatslimlib/core_base.pyx":1805
 *             resid = residue.resid
 * 
 *             if resid > last_resid: # WARNING: It is assumed that atomids (hence resids) are sorted!             # <<<<<<<<<<<<<<
 *                 resids[resid_size] = resid
 *                 hg_bead_atomids_offsets[resid_size] = i
 */
    }

    /* "fatslimlib/core_base.pyx":1815
 * 
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 atom_index = residues_dict[residue.name_id][atom.name_id]
 *             except KeyError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_22, &__pyx_t_23, &__pyx_t_24);
      __Pyx_XGOTREF(__pyx_t_22);
      __Pyx_XGOTREF(__pyx_t_23);
      __Pyx_XGOTREF(__pyx_t_24);
      /*try:*/ {

        /* "fatslimlib/core_base.pyx":1816
 * 
 *             try:
 *                 atom_index = residues_dict[residue.name_id][atom.name_id]             # <<<<<<<<<<<<<<
 *             except KeyError:
 *                 atom_index = topol_residue_atomid_index(residue, atom.atomid)
 */
        __pyx_t_2 = __Pyx_PyInt_From_fsl_int(__pyx_v_residue->name_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1816, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_v_residues_dict, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1816, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_5, __pyx_v_atom->name_id, fsl_int, 1, __Pyx_PyInt_From_fsl_int, 0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1816, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_16 = __Pyx_PyInt_As_fsl_int(__pyx_t_2); if (unlikely((__pyx_t_16 == ((fsl_int)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1816, __pyx_L7_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_atom_index = __pyx_t_16;

        /* "fatslimlib/core_base.pyx":1815
 * 
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 atom_index = residues_dict[residue.name_id][atom.name_id]
 *             except KeyError:
 */
      }
      __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
      __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
      __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
      goto __pyx_L14_try_end;
      __pyx_L7_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "fatslimlib/core_base.pyx":1817
 *             try:
 *                 atom_index = residues_dict[residue.name_id][atom.name_id]
 *             except KeyError:             # <<<<<<<<<<<<<<
 *                 atom_index = topol_residue_atomid_index(residue, atom.atomid)
 *                 try:
 */
      __pyx_t_25 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
      if (__pyx_t_25) {
        __Pyx_AddTraceback("fatslimlib.core_base.Trajectory.initialize", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(0, 1817, __pyx_L9_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GOTREF(__pyx_t_6);

        /* "fatslimlib/core_base.pyx":1818
 *                 atom_index = residues_dict[residue.name_id][atom.name_id]
 *             except KeyError:
 *                 atom_index = topol_residue_atomid_index(residue, atom.atomid)             # <<<<<<<<<<<<<<
 *                 try:
 *                     residues_dict[residue.name_id][atom.name_id] = atom_index
 */
        __pyx_v_atom_index = __pyx_f_10fatslimlib_9core_base_topol_residue_atomid_index(__pyx_v_residue, __pyx_v_atom->atomid);

        /* "fatslimlib/core_base.pyx":1819
 *             except KeyError:
 *                 atom_index = topol_residue_atomid_index(residue, atom.atomid)
 *                 try:             # <<<<<<<<<<<<<<
 *                     residues_dict[residue.name_id][atom.name_id] = atom_index
 *                 except KeyError:
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_26, &__pyx_t_27, &__pyx_t_28);
          __Pyx_XGOTREF(__pyx_t_26);
          __Pyx_XGOTREF(__pyx_t_27);
          __Pyx_XGOTREF(__pyx_t_28);
          /*try:*/ {

            /* "fatslimlib/core_base.pyx":1820
 *                 atom_index = topol_residue_atomid_index(residue, atom.atomid)
 *                 try:
 *                     residues_dict[residue.name_id][atom.name_id] = atom_index             # <<<<<<<<<<<<<<
 *                 except KeyError:
 *                     residues_dict[residue.name_id] = {atom.name_id: atom_index}
 */
            __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_v_atom_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1820, __pyx_L17_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_8 = __Pyx_PyInt_From_fsl_int(__pyx_v_residue->name_id); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1820, __pyx_L17_error)
            __Pyx_GOTREF(__pyx_t_8);
            __pyx_t_29 = __Pyx_PyDict_GetItem(__pyx_v_residues_dict, __pyx_t_8); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1820, __pyx_L17_error)
            __Pyx_GOTREF(__pyx_t_29);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            if (unlikely(__Pyx_SetItemInt(__pyx_t_29, __pyx_v_atom->name_id, __pyx_t_1, fsl_int, 1, __Pyx_PyInt_From_fsl_int, 0, 0, 0) < 0)) __PYX_ERR(0, 1820, __pyx_L17_error)
            __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "fatslimlib/core_base.pyx":1819
 *             except KeyError:
 *                 atom_index = topol_residue_atomid_index(residue, atom.atomid)
 *                 try:             # <<<<<<<<<<<<<<
 *                     residues_dict[residue.name_id][atom.name_id] = atom_index
 *                 except KeyError:
 */
          }
          __Pyx_XDECREF(__pyx_t_26); __pyx_t_26 = 0;
          __Pyx_XDECREF(__pyx_t_27); __pyx_t_27 = 0;
          __Pyx_XDECREF(__pyx_t_28); __pyx_t_28 = 0;
          goto __pyx_L24_try_end;
          __pyx_L17_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_29); __pyx_t_29 = 0;
          __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

          /* "fatslimlib/core_base.pyx":1821
 *                 try:
 *                     residues_dict[residue.name_id][atom.name_id] = atom_index
 *                 except KeyError:             # <<<<<<<<<<<<<<
 *                     residues_dict[residue.name_id] = {atom.name_id: atom_index}
 * 
 */
          __pyx_t_25 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
          if (__pyx_t_25) {
            __Pyx_AddTraceback("fatslimlib.core_base.Trajectory.initialize", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_29, &__pyx_t_8) < 0) __PYX_ERR(0, 1821, __pyx_L19_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_GOTREF(__pyx_t_29);
            __Pyx_GOTREF(__pyx_t_8);

            /* "fatslimlib/core_base.pyx":1822
 *                     residues_dict[residue.name_id][atom.name_id] = atom_index
 *                 except KeyError:
 *                     residues_dict[residue.name_id] = {atom.name_id: atom_index}             # <<<<<<<<<<<<<<
 * 
 *             lipid_hg_indices[i] = atom_index + lipid_atomids_offsets[resid_size-1]
 */
            __pyx_t_30 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 1822, __pyx_L19_except_error)
            __Pyx_GOTREF(__pyx_t_30);
            __pyx_t_31 = __Pyx_PyInt_From_fsl_int(__pyx_v_atom->name_id); if (unlikely(!__pyx_t_31)) __PYX_ERR(0, 1822, __pyx_L19_except_error)
            __Pyx_GOTREF(__pyx_t_31);
            __pyx_t_32 = __Pyx_PyInt_From_fsl_int(__pyx_v_atom_index); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1822, __pyx_L19_except_error)
            __Pyx_GOTREF(__pyx_t_32);
            if (PyDict_SetItem(__pyx_t_30, __pyx_t_31, __pyx_t_32) < 0) __PYX_ERR(0, 1822, __pyx_L19_except_error)
            __Pyx_DECREF(__pyx_t_31); __pyx_t_31 = 0;
            __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
            __pyx_t_32 = __Pyx_PyInt_From_fsl_int(__pyx_v_residue->name_id); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1822, __pyx_L19_except_error)
            __Pyx_GOTREF(__pyx_t_32);
            if (unlikely(PyDict_SetItem(__pyx_v_residues_dict, __pyx_t_32, __pyx_t_30) < 0)) __PYX_ERR(0, 1822, __pyx_L19_except_error)
            __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
            __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF(__pyx_t_29); __pyx_t_29 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            goto __pyx_L18_exception_handled;
          }
          goto __pyx_L19_except_error;
          __pyx_L19_except_error:;

          /* "fatslimlib/core_base.pyx":1819
 *             except KeyError:
 *                 atom_index = topol_residue_atomid_index(residue, atom.atomid)
 *                 try:             # <<<<<<<<<<<<<<
 *                     residues_dict[residue.name_id][atom.name_id] = atom_index
 *                 except KeyError:
 */
          __Pyx_XGIVEREF(__pyx_t_26);
          __Pyx_XGIVEREF(__pyx_t_27);
          __Pyx_XGIVEREF(__pyx_t_28);
          __Pyx_ExceptionReset(__pyx_t_26, __pyx_t_27, __pyx_t_28);
          goto __pyx_L9_except_error;
          __pyx_L18_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_26);
          __Pyx_XGIVEREF(__pyx_t_27);
          __Pyx_XGIVEREF(__pyx_t_28);
          __Pyx_ExceptionReset(__pyx_t_26, __pyx_t_27, __pyx_t_28);
          __pyx_L24_try_end:;
        }
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L8_exception_handled;
      }
      goto __pyx_L9_except_error;
      __pyx_L9_except_error:;

      /* "fatslimlib/core_base.pyx":1815
 * 
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 atom_index = residues_dict[residue.name_id][atom.name_id]
 *             except KeyError:
 */
      __Pyx_XGIVEREF(__pyx_t_22);
      __Pyx_XGIVEREF(__pyx_t_23);
      __Pyx_XGIVEREF(__pyx_t_24);
      __Pyx_ExceptionReset(__pyx_t_22, __pyx_t_23, __pyx_t_24);
      goto __pyx_L1_error;
      __pyx_L8_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_22);
      __Pyx_XGIVEREF(__pyx_t_23);
      __Pyx_XGIVEREF(__pyx_t_24);
      __Pyx_ExceptionReset(__pyx_t_22, __pyx_t_23, __pyx_t_24);
      __pyx_L14_try_end:;
    }

    /* "fatslimlib/core_base.pyx":1824
 *                     residues_dict[residue.name_id] = {atom.name_id: atom_index}
 * 
 *             lipid_hg_indices[i] = atom_index + lipid_atomids_offsets[resid_size-1]             # <<<<<<<<<<<<<<
 * 
 *         hg_bead_atomids_offsets[resid_size] = size
 */
    __pyx_t_33 = (__pyx_v_resid_size - 1);
    __pyx_t_34 = __pyx_v_i;
    *((fsl_int *) ( /* dim=0 */ (__pyx_v_lipid_hg_indices.data + __pyx_t_34 * __pyx_v_lipid_hg_indices.strides[0]) )) = (__pyx_v_atom_index + (*((fsl_int *) ( /* dim=0 */ (__pyx_v_lipid_atomids_offsets.data + __pyx_t_33 * __pyx_v_lipid_atomids_offsets.strides[0]) ))));
  }

  /* "fatslimlib/core_base.pyx":1826
 *             lipid_hg_indices[i] = atom_index + lipid_atomids_offsets[resid_size-1]
 * 
 *         hg_bead_atomids_offsets[resid_size] = size             # <<<<<<<<<<<<<<
 *         lipid_atomids = np.array(lipid_atomids_list, dtype=np.int64)
 * 
 */
  __pyx_t_35 = __pyx_v_resid_size;
  *((fsl_int *) ( /* dim=0 */ (__pyx_v_hg_bead_atomids_offsets.data + __pyx_t_35 * __pyx_v_hg_bead_atomids_offsets.strides[0]) )) = __pyx_v_size;

  /* "fatslimlib/core_base.pyx":1827
 * 
 *         hg_bead_atomids_offsets[resid_size] = size
 *         lipid_atomids = np.array(lipid_atomids_list, dtype=np.int64)             # <<<<<<<<<<<<<<
 * 
 *         self.hg_bead_atomids = np.array(hg_group_atomids, dtype=np.int64)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_np); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_array); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_lipid_atomids_list);
  __Pyx_GIVEREF(__pyx_v_lipid_atomids_list);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_lipid_atomids_list);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_29 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_int64); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_29) < 0) __PYX_ERR(0, 1827, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __pyx_t_29 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, __pyx_t_2); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(__pyx_t_29, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1827, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __pyx_v_lipid_atomids = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "fatslimlib/core_base.pyx":1829
 *         lipid_atomids = np.array(lipid_atomids_list, dtype=np.int64)
 * 
 *         self.hg_bead_atomids = np.array(hg_group_atomids, dtype=np.int64)             # <<<<<<<<<<<<<<
 *         self.hg_bead_atomids_offsets = hg_bead_atomids_offsets[:resid_size + 1]
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_29, __pyx_n_s_np); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_29, __pyx_n_s_array); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __pyx_t_29 = __pyx_memoryview_fromslice(__pyx_v_hg_group_atomids, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_fsl_int, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_fsl_int, 0);; if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_29);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_29);
  __pyx_t_29 = 0;
  __pyx_t_29 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_int64); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_29, __pyx_n_s_dtype, __pyx_t_8) < 0) __PYX_ERR(0, 1829, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, __pyx_t_29); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
  __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1829, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->hg_bead_atomids, 0);
  __pyx_v_self->hg_bead_atomids = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "fatslimlib/core_base.pyx":1830
 * 
 *         self.hg_bead_atomids = np.array(hg_group_atomids, dtype=np.int64)
 *         self.hg_bead_atomids_offsets = hg_bead_atomids_offsets[:resid_size + 1]             # <<<<<<<<<<<<<<
 * 
 *         self.lipid_atomids = lipid_atomids
 */
  __pyx_t_7.data = __pyx_v_hg_bead_atomids_offsets.data;
  __pyx_t_7.memview = __pyx_v_hg_bead_atomids_offsets.memview;
  __PYX_INC_MEMVIEW(&__pyx_t_7, 0);
  __pyx_t_25 = -1;
  if (unlikely(__pyx_memoryview_slice_memviewslice(
    &__pyx_t_7,
    __pyx_v_hg_bead_atomids_offsets.shape[0], __pyx_v_hg_bead_atomids_offsets.strides[0], __pyx_v_hg_bead_atomids_offsets.suboffsets[0],
    0,
    0,
    &__pyx_t_25,
    0,
    (__pyx_v_resid_size + 1),
    0,
    0,
    1,
    0,
    1) < 0))
{
    __PYX_ERR(0, 1830, __pyx_L1_error)
}

__PYX_XDEC_MEMVIEW(&__pyx_v_self->hg_bead_atomids_offsets, 0);
  __pyx_v_self->hg_bead_atomids_offsets = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "fatslimlib/core_base.pyx":1832
 *         self.hg_bead_atomids_offsets = hg_bead_atomids_offsets[:resid_size + 1]
 * 
 *         self.lipid_atomids = lipid_atomids             # <<<<<<<<<<<<<<
 *         self.lipid_atomids_offsets = lipid_atomids_offsets[:resid_size + 1]
 *         self.lipid_hg_indices = lipid_hg_indices
 */
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->lipid_atomids, 0);
  __PYX_INC_MEMVIEW(&__pyx_v_lipid_atomids, 0);
  __pyx_v_self->lipid_atomids = __pyx_v_lipid_atomids;

  /* "fatslimlib/core_base.pyx":1833
 * 
 *         self.lipid_atomids = lipid_atomids
 *         self.lipid_atomids_offsets = lipid_atomids_offsets[:resid_size + 1]             # <<<<<<<<<<<<<<
 *         self.lipid_hg_indices = lipid_hg_indices
 * 
 */
  __pyx_t_7.data = __pyx_v_lipid_atomids_offsets.data;
  __pyx_t_7.memview = __pyx_v_lipid_atomids_offsets.memview;
  __PYX_INC_MEMVIEW(&__pyx_t_7, 0);
  __pyx_t_25 = -1;
  if (unlikely(__pyx_memoryview_slice_memviewslice(
    &__pyx_t_7,
    __pyx_v_lipid_atomids_offsets.shape[0], __pyx_v_lipid_atomids_offsets.strides[0], __pyx_v_lipid_atomids_offsets.suboffsets[0],
    0,
    0,
    &__pyx_t_25,
    0,
    (__pyx_v_resid_size + 1),
    0,
    0,
    1,
    0,
    1) < 0))
{
    __PYX_ERR(0, 1833, __pyx_L1_error)
}

__PYX_XDEC_MEMVIEW(&__pyx_v_self->lipid_atomids_offsets, 0);
  __pyx_v_self->lipid_atomids_offsets = __pyx_t_7;
  __pyx_t_7.memview = NULL;
  __pyx_t_7.data = NULL;

  /* "fatslimlib/core_base.pyx":1834
 *         self.lipid_atomids = lipid_atomids
 *         self.lipid_atomids_offsets = lipid_atomids_offsets[:resid_size + 1]
 *         self.lipid_hg_indices = lipid_hg_indices             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __PYX_XDEC_MEMVIEW(&__pyx_v_self->lipid_hg_indices, 0);
  __PYX_INC_MEMVIEW(&__pyx_v_lipid_hg_indices, 0);
  __pyx_v_self->lipid_hg_indices = __pyx_v_lipid_hg_indices;

  /* "fatslimlib/core_base.pyx":1838
 * 
 *         # Handle interacting group (if any)
 *         self.interacting_group_name = interacting_group.encode()             # <<<<<<<<<<<<<<
 *         try:
 *             self.interacting_atomids = self.index_reader[interacting_group]
 */
  if (unlikely(__pyx_v_interacting_group == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 1838, __pyx_L1_error)
  }
  __pyx_t_8 = PyUnicode_AsEncodedString(__pyx_v_interacting_group, NULL, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_8);
  __Pyx_GOTREF(__pyx_v_self->interacting_group_name);
  __Pyx_DECREF(__pyx_v_self->interacting_group_name);
  __pyx_v_self->interacting_group_name = ((PyObject*)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "fatslimlib/core_base.pyx":1839
 *         # Handle interacting group (if any)
 *         self.interacting_group_name = interacting_group.encode()
 *         try:             # <<<<<<<<<<<<<<
 *             self.interacting_atomids = self.index_reader[interacting_group]
 *         except KeyError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_24, &__pyx_t_23, &__pyx_t_22);
    __Pyx_XGOTREF(__pyx_t_24);
    __Pyx_XGOTREF(__pyx_t_23);
    __Pyx_XGOTREF(__pyx_t_22);
    /*try:*/ {

      /* "fatslimlib/core_base.pyx":1840
 *         self.interacting_group_name = interacting_group.encode()
 *         try:
 *             self.interacting_atomids = self.index_reader[interacting_group]             # <<<<<<<<<<<<<<
 *         except KeyError:
 *             self.interacting_atomids = np.array([], dtype=np.int64)
 */
      __pyx_t_8 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self->index_reader), __pyx_v_interacting_group); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1840, __pyx_L27_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(__pyx_t_8, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1840, __pyx_L27_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_XDEC_MEMVIEW(&__pyx_v_self->interacting_atomids, 0);
      __pyx_v_self->interacting_atomids = __pyx_t_7;
      __pyx_t_7.memview = NULL;
      __pyx_t_7.data = NULL;

      /* "fatslimlib/core_base.pyx":1839
 *         # Handle interacting group (if any)
 *         self.interacting_group_name = interacting_group.encode()
 *         try:             # <<<<<<<<<<<<<<
 *             self.interacting_atomids = self.index_reader[interacting_group]
 *         except KeyError:
 */
    }
    __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
    __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_XDECREF(__pyx_t_22); __pyx_t_22 = 0;
    goto __pyx_L32_try_end;
    __pyx_L27_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_29); __pyx_t_29 = 0;
    __Pyx_XDECREF(__pyx_t_30); __pyx_t_30 = 0;
    __Pyx_XDECREF(__pyx_t_31); __pyx_t_31 = 0;
    __Pyx_XDECREF(__pyx_t_32); __pyx_t_32 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "fatslimlib/core_base.pyx":1841
 *         try:
 *             self.interacting_atomids = self.index_reader[interacting_group]
 *         except KeyError:             # <<<<<<<<<<<<<<
 *             self.interacting_atomids = np.array([], dtype=np.int64)
 * 
 */
    __pyx_t_25 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError);
    if (__pyx_t_25) {
      __Pyx_AddTraceback("fatslimlib.core_base.Trajectory.initialize", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_29, &__pyx_t_6) < 0) __PYX_ERR(0, 1841, __pyx_L29_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_29);
      __Pyx_GOTREF(__pyx_t_6);

      /* "fatslimlib/core_base.pyx":1842
 *             self.interacting_atomids = self.index_reader[interacting_group]
 *         except KeyError:
 *             self.interacting_atomids = np.array([], dtype=np.int64)             # <<<<<<<<<<<<<<
 * 
 *         # Sanity check the interacting atomids
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_np); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1842, __pyx_L29_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_array); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1842, __pyx_L29_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1842, __pyx_L29_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1842, __pyx_L29_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1842, __pyx_L29_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GetModuleGlobalName(__pyx_t_30, __pyx_n_s_np); if (unlikely(!__pyx_t_30)) __PYX_ERR(0, 1842, __pyx_L29_except_error)
      __Pyx_GOTREF(__pyx_t_30);
      __pyx_t_32 = __Pyx_PyObject_GetAttrStr(__pyx_t_30, __pyx_n_s_int64); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1842, __pyx_L29_except_error)
      __Pyx_GOTREF(__pyx_t_32);
      __Pyx_DECREF(__pyx_t_30); __pyx_t_30 = 0;
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dtype, __pyx_t_32) < 0) __PYX_ERR(0, 1842, __pyx_L29_except_error)
      __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
      __pyx_t_32 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1842, __pyx_L29_except_error)
      __Pyx_GOTREF(__pyx_t_32);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_7 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(__pyx_t_32, PyBUF_WRITABLE); if (unlikely(!__pyx_t_7.memview)) __PYX_ERR(0, 1842, __pyx_L29_except_error)
      __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
      __PYX_XDEC_MEMVIEW(&__pyx_v_self->interacting_atomids, 0);
      __pyx_v_self->interacting_atomids = __pyx_t_7;
      __pyx_t_7.memview = NULL;
      __pyx_t_7.data = NULL;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_29); __pyx_t_29 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L28_exception_handled;
    }
    goto __pyx_L29_except_error;
    __pyx_L29_except_error:;

    /* "fatslimlib/core_base.pyx":1839
 *         # Handle interacting group (if any)
 *         self.interacting_group_name = interacting_group.encode()
 *         try:             # <<<<<<<<<<<<<<
 *             self.interacting_atomids = self.index_reader[interacting_group]
 *         except KeyError:
 */
    __Pyx_XGIVEREF(__pyx_t_24);
    __Pyx_XGIVEREF(__pyx_t_23);
    __Pyx_XGIVEREF(__pyx_t_22);
    __Pyx_ExceptionReset(__pyx_t_24, __pyx_t_23, __pyx_t_22);
    goto __pyx_L1_error;
    __pyx_L28_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_24);
    __Pyx_XGIVEREF(__pyx_t_23);
    __Pyx_XGIVEREF(__pyx_t_22);
    __Pyx_ExceptionReset(__pyx_t_24, __pyx_t_23, __pyx_t_22);
    __pyx_L32_try_end:;
  }

  /* "fatslimlib/core_base.pyx":1845
 * 
 *         # Sanity check the interacting atomids
 *         for i in range(self.interacting_atomids.shape[0]):             # <<<<<<<<<<<<<<
 *             if topology.fast_get_atom_internal_id(self.interacting_atomids[i]) < 0:
 *                 raise KeyError("Atomid not registered in topology: %i. Does group '%s' from '%s'"
 */
  if (unlikely(!__pyx_v_self->interacting_atomids.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1845, __pyx_L1_error)}
  __pyx_t_3 = (__pyx_v_self->interacting_atomids.shape[0]);
  __pyx_t_36 = __pyx_t_3;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_36; __pyx_t_10+=1) {
    __pyx_v_i = __pyx_t_10;

    /* "fatslimlib/core_base.pyx":1846
 *         # Sanity check the interacting atomids
 *         for i in range(self.interacting_atomids.shape[0]):
 *             if topology.fast_get_atom_internal_id(self.interacting_atomids[i]) < 0:             # <<<<<<<<<<<<<<
 *                 raise KeyError("Atomid not registered in topology: %i. Does group '%s' from '%s'"
 *                                " file contains H?" % (self.interacting_atomids[i],
 */
    if (unlikely(!__pyx_v_self->interacting_atomids.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1846, __pyx_L1_error)}
    __pyx_t_37 = __pyx_v_i;
    __pyx_t_14 = ((((struct __pyx_vtabstruct_10fatslimlib_9core_base_Topology *)__pyx_v_topology->__pyx_vtab)->fast_get_atom_internal_id(__pyx_v_topology, (*((fsl_int *) ( /* dim=0 */ (__pyx_v_self->interacting_atomids.data + __pyx_t_37 * __pyx_v_self->interacting_atomids.strides[0]) )))) < 0) != 0);
    if (unlikely(__pyx_t_14)) {

      /* "fatslimlib/core_base.pyx":1848
 *             if topology.fast_get_atom_internal_id(self.interacting_atomids[i]) < 0:
 *                 raise KeyError("Atomid not registered in topology: %i. Does group '%s' from '%s'"
 *                                " file contains H?" % (self.interacting_atomids[i],             # <<<<<<<<<<<<<<
 *                                                       interacting_group,
 *                                                       self.index_reader.filename))
 */
      if (unlikely(!__pyx_v_self->interacting_atomids.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1848, __pyx_L1_error)}
      __pyx_t_38 = __pyx_v_i;
      __pyx_t_6 = __Pyx_PyInt_From_fsl_int((*((fsl_int *) ( /* dim=0 */ (__pyx_v_self->interacting_atomids.data + __pyx_t_38 * __pyx_v_self->interacting_atomids.strides[0]) )))); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1848, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);

      /* "fatslimlib/core_base.pyx":1850
 *                                " file contains H?" % (self.interacting_atomids[i],
 *                                                       interacting_group,
 *                                                       self.index_reader.filename))             # <<<<<<<<<<<<<<
 * 
 *         # Identify the forcefield type
 */
      __pyx_t_29 = PyTuple_New(3); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1848, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_29);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_29, 0, __pyx_t_6);
      __Pyx_INCREF(__pyx_v_interacting_group);
      __Pyx_GIVEREF(__pyx_v_interacting_group);
      PyTuple_SET_ITEM(__pyx_t_29, 1, __pyx_v_interacting_group);
      __Pyx_INCREF(__pyx_v_self->index_reader->filename);
      __Pyx_GIVEREF(__pyx_v_self->index_reader->filename);
      PyTuple_SET_ITEM(__pyx_t_29, 2, __pyx_v_self->index_reader->filename);
      __pyx_t_6 = 0;

      /* "fatslimlib/core_base.pyx":1848
 *             if topology.fast_get_atom_internal_id(self.interacting_atomids[i]) < 0:
 *                 raise KeyError("Atomid not registered in topology: %i. Does group '%s' from '%s'"
 *                                " file contains H?" % (self.interacting_atomids[i],             # <<<<<<<<<<<<<<
 *                                                       interacting_group,
 *                                                       self.index_reader.filename))
 */
      __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_Atomid_not_registered_in_topolog, __pyx_t_29); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1848, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

      /* "fatslimlib/core_base.pyx":1847
 *         for i in range(self.interacting_atomids.shape[0]):
 *             if topology.fast_get_atom_internal_id(self.interacting_atomids[i]) < 0:
 *                 raise KeyError("Atomid not registered in topology: %i. Does group '%s' from '%s'"             # <<<<<<<<<<<<<<
 *                                " file contains H?" % (self.interacting_atomids[i],
 *                                                       interacting_group,
 */
      __pyx_t_29 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_t_6); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1847, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_29);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_Raise(__pyx_t_29, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;
      __PYX_ERR(0, 1847, __pyx_L1_error)

      /* "fatslimlib/core_base.pyx":1846
 *         # Sanity check the interacting atomids
 *         for i in range(self.interacting_atomids.shape[0]):
 *             if topology.fast_get_atom_internal_id(self.interacting_atomids[i]) < 0:             # <<<<<<<<<<<<<<
 *                 raise KeyError("Atomid not registered in topology: %i. Does group '%s' from '%s'"
 *                                " file contains H?" % (self.interacting_atomids[i],
 */
    }
  }

  /* "fatslimlib/core_base.pyx":1853
 * 
 *         # Identify the forcefield type
 *         self.get_forcefield_type()             # <<<<<<<<<<<<<<
 * 
 *         self.initialized = True
 */
  (void)(((struct __pyx_vtabstruct_10fatslimlib_9core_base_Trajectory *)__pyx_v_self->__pyx_vtab)->get_forcefield_type(__pyx_v_self, 0));

  /* "fatslimlib/core_base.pyx":1855
 *         self.get_forcefield_type()
 * 
 *         self.initialized = True             # <<<<<<<<<<<<<<
 * 
 *         self.timesteps = np.asarray(self.coords_reader.timesteps).tolist()
 */
  __pyx_v_self->initialized = 1;

  /* "fatslimlib/core_base.pyx":1857
 *         self.initialized = True
 * 
 *         self.timesteps = np.asarray(self.coords_reader.timesteps).tolist()             # <<<<<<<<<<<<<<
 * 
 *         verbose_print("Done in %s" % (pretty_delta(begin, time())),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_np); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_32 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_asarray); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_v_self->coords_reader->timesteps.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1857, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->coords_reader->timesteps, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_real, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_real, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_32))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_32);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_32);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_32, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_32, __pyx_t_2, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_32, __pyx_t_8);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __pyx_t_32 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_tolist); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_32))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_32);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_32);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_32, function);
    }
  }
  __pyx_t_29 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_32, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_32);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  if (!(likely(PyList_CheckExact(__pyx_t_29))||((__pyx_t_29) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_29)->tp_name), 0))) __PYX_ERR(0, 1857, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_29);
  __Pyx_GOTREF(__pyx_v_self->timesteps);
  __Pyx_DECREF(__pyx_v_self->timesteps);
  __pyx_v_self->timesteps = ((PyObject*)__pyx_t_29);
  __pyx_t_29 = 0;

  /* "fatslimlib/core_base.pyx":1859
 *         self.timesteps = np.asarray(self.coords_reader.timesteps).tolist()
 * 
 *         verbose_print("Done in %s" % (pretty_delta(begin, time())),             # <<<<<<<<<<<<<<
 *                       self.be_verbose)
 *         sys.stdout.flush()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_32, __pyx_n_s_verbose_print); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_time); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __pyx_f_10fatslimlib_9core_base_pretty_delta(__pyx_v_begin, __pyx_t_6, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Done_in_s, __pyx_t_8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "fatslimlib/core_base.pyx":1860
 * 
 *         verbose_print("Done in %s" % (pretty_delta(begin, time())),
 *                       self.be_verbose)             # <<<<<<<<<<<<<<
 *         sys.stdout.flush()
 * 
 */
  __pyx_t_8 = __Pyx_PyBool_FromLong(__pyx_v_self->be_verbose); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = NULL;
  __pyx_t_25 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_32))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_32);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_32);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_32, function);
      __pyx_t_25 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_32)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_t_6, __pyx_t_8};
    __pyx_t_29 = __Pyx_PyFunction_FastCall(__pyx_t_32, __pyx_temp+1-__pyx_t_25, 2+__pyx_t_25); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1859, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_29);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_32)) {
    PyObject *__pyx_temp[3] = {__pyx_t_2, __pyx_t_6, __pyx_t_8};
    __pyx_t_29 = __Pyx_PyCFunction_FastCall(__pyx_t_32, __pyx_temp+1-__pyx_t_25, 2+__pyx_t_25); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1859, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_29);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  } else
  #endif
  {
    __pyx_t_1 = PyTuple_New(2+__pyx_t_25); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__pyx_t_2) {
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2); __pyx_t_2 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_25, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_25, __pyx_t_8);
    __pyx_t_6 = 0;
    __pyx_t_8 = 0;
    __pyx_t_29 = __Pyx_PyObject_Call(__pyx_t_32, __pyx_t_1, NULL); if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_29);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

  /* "fatslimlib/core_base.pyx":1861
 *         verbose_print("Done in %s" % (pretty_delta(begin, time())),
 *                       self.be_verbose)
 *         sys.stdout.flush()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_32, __pyx_n_s_sys); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_32, __pyx_n_s_stdout); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __pyx_t_32 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_flush); if (unlikely(!__pyx_t_32)) __PYX_ERR(0, 1861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_32);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_32))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_32);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_32);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_32, function);
    }
  }
  __pyx_t_29 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_32, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_32);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_29)) __PYX_ERR(0, 1861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_29);
  __Pyx_DECREF(__pyx_t_32); __pyx_t_32 = 0;
  __Pyx_DECREF(__pyx_t_29); __pyx_t_29 = 0;

  /* "fatslimlib/core_base.pyx":1753
 * 
 * 
 *     def initialize(self, str hg_group="headgroups", str interacting_group="protein"):             # <<<<<<<<<<<<<<
 *         cdef fsl_int size, i, resid_size
 *         cdef fsl_int last_resid, resid
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __PYX_XDEC_MEMVIEW(&__pyx_t_7, 1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_29);
  __Pyx_XDECREF(__pyx_t_30);
  __Pyx_XDECREF(__pyx_t_31);
  __Pyx_XDECREF(__pyx_t_32);
  __Pyx_AddTraceback("fatslimlib.core_base.Trajectory.initialize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_hg_group_atomids, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_hg_bead_atomids_offsets, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lipid_atomids_offsets, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lipid_atomids, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_resids, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_v_lipid_hg_indices, 1);
  __Pyx_XDECREF((PyObject *)__pyx_v_topology);
  __Pyx_XDECREF(__pyx_v_residues_dict);
  __Pyx_XDECREF(__pyx_v_begin);
  __Pyx_XDECREF(__pyx_v_lipid_atomids_list);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1864
 * 
 * 
 *     cpdef fsl_int get_forcefield_type(self):             # <<<<<<<<<<<<<<
 *         cdef real rvdw
 *         if self.forcefield_type != fft_notset: # Already done
 */

static PyObject *__pyx_pw_10fatslimlib_9core_base_10Trajectory_5get_forcefield_type(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static fsl_int __pyx_f_10fatslimlib_9core_base_10Trajectory_get_forcefield_type(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self, int __pyx_skip_dispatch) {
  real __pyx_v_rvdw;
  fsl_int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  fsl_int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("get_forcefield_type", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_forcefield_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1864, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10fatslimlib_9core_base_10Trajectory_5get_forcefield_type)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1864, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_fsl_int(__pyx_t_2); if (unlikely((__pyx_t_5 == ((fsl_int)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1864, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "fatslimlib/core_base.pyx":1866
 *     cpdef fsl_int get_forcefield_type(self):
 *         cdef real rvdw
 *         if self.forcefield_type != fft_notset: # Already done             # <<<<<<<<<<<<<<
 *             return self.forcefield_type
 * 
 */
  __pyx_t_6 = ((__pyx_v_self->forcefield_type != __pyx_e_10fatslimlib_9core_base_fft_notset) != 0);
  if (__pyx_t_6) {

    /* "fatslimlib/core_base.pyx":1867
 *         cdef real rvdw
 *         if self.forcefield_type != fft_notset: # Already done
 *             return self.forcefield_type             # <<<<<<<<<<<<<<
 * 
 *         rvdw = self.get_rvdw()
 */
    __pyx_r = __pyx_v_self->forcefield_type;
    goto __pyx_L0;

    /* "fatslimlib/core_base.pyx":1866
 *     cpdef fsl_int get_forcefield_type(self):
 *         cdef real rvdw
 *         if self.forcefield_type != fft_notset: # Already done             # <<<<<<<<<<<<<<
 *             return self.forcefield_type
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":1869
 *             return self.forcefield_type
 * 
 *         rvdw = self.get_rvdw()             # <<<<<<<<<<<<<<
 * 
 *         if rvdw < FFT_ALLATOM_THRESHOLD:
 */
  __pyx_v_rvdw = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Trajectory *)__pyx_v_self->__pyx_vtab)->get_rvdw(__pyx_v_self, 0);

  /* "fatslimlib/core_base.pyx":1871
 *         rvdw = self.get_rvdw()
 * 
 *         if rvdw < FFT_ALLATOM_THRESHOLD:             # <<<<<<<<<<<<<<
 *             self.forcefield_type = fft_allatom
 *         elif rvdw < FFT_UNIFIED_THRESHOLD:
 */
  __pyx_t_6 = ((__pyx_v_rvdw < 0.015) != 0);
  if (__pyx_t_6) {

    /* "fatslimlib/core_base.pyx":1872
 * 
 *         if rvdw < FFT_ALLATOM_THRESHOLD:
 *             self.forcefield_type = fft_allatom             # <<<<<<<<<<<<<<
 *         elif rvdw < FFT_UNIFIED_THRESHOLD:
 *             self.forcefield_type = fft_unified
 */
    __pyx_v_self->forcefield_type = __pyx_e_10fatslimlib_9core_base_fft_allatom;

    /* "fatslimlib/core_base.pyx":1871
 *         rvdw = self.get_rvdw()
 * 
 *         if rvdw < FFT_ALLATOM_THRESHOLD:             # <<<<<<<<<<<<<<
 *             self.forcefield_type = fft_allatom
 *         elif rvdw < FFT_UNIFIED_THRESHOLD:
 */
    goto __pyx_L4;
  }

  /* "fatslimlib/core_base.pyx":1873
 *         if rvdw < FFT_ALLATOM_THRESHOLD:
 *             self.forcefield_type = fft_allatom
 *         elif rvdw < FFT_UNIFIED_THRESHOLD:             # <<<<<<<<<<<<<<
 *             self.forcefield_type = fft_unified
 *         elif rvdw < FFT_COARSE_THRESHOLD:
 */
  __pyx_t_6 = ((__pyx_v_rvdw < 0.025) != 0);
  if (__pyx_t_6) {

    /* "fatslimlib/core_base.pyx":1874
 *             self.forcefield_type = fft_allatom
 *         elif rvdw < FFT_UNIFIED_THRESHOLD:
 *             self.forcefield_type = fft_unified             # <<<<<<<<<<<<<<
 *         elif rvdw < FFT_COARSE_THRESHOLD:
 *             self.forcefield_type = fft_coarse
 */
    __pyx_v_self->forcefield_type = __pyx_e_10fatslimlib_9core_base_fft_unified;

    /* "fatslimlib/core_base.pyx":1873
 *         if rvdw < FFT_ALLATOM_THRESHOLD:
 *             self.forcefield_type = fft_allatom
 *         elif rvdw < FFT_UNIFIED_THRESHOLD:             # <<<<<<<<<<<<<<
 *             self.forcefield_type = fft_unified
 *         elif rvdw < FFT_COARSE_THRESHOLD:
 */
    goto __pyx_L4;
  }

  /* "fatslimlib/core_base.pyx":1875
 *         elif rvdw < FFT_UNIFIED_THRESHOLD:
 *             self.forcefield_type = fft_unified
 *         elif rvdw < FFT_COARSE_THRESHOLD:             # <<<<<<<<<<<<<<
 *             self.forcefield_type = fft_coarse
 *         else:
 */
  __pyx_t_6 = ((__pyx_v_rvdw < 0.25) != 0);
  if (__pyx_t_6) {

    /* "fatslimlib/core_base.pyx":1876
 *             self.forcefield_type = fft_unified
 *         elif rvdw < FFT_COARSE_THRESHOLD:
 *             self.forcefield_type = fft_coarse             # <<<<<<<<<<<<<<
 *         else:
 *             self.forcefield_type = fft_unknown
 */
    __pyx_v_self->forcefield_type = __pyx_e_10fatslimlib_9core_base_fft_coarse;

    /* "fatslimlib/core_base.pyx":1875
 *         elif rvdw < FFT_UNIFIED_THRESHOLD:
 *             self.forcefield_type = fft_unified
 *         elif rvdw < FFT_COARSE_THRESHOLD:             # <<<<<<<<<<<<<<
 *             self.forcefield_type = fft_coarse
 *         else:
 */
    goto __pyx_L4;
  }

  /* "fatslimlib/core_base.pyx":1878
 *             self.forcefield_type = fft_coarse
 *         else:
 *             self.forcefield_type = fft_unknown             # <<<<<<<<<<<<<<
 * 
 *     cpdef real get_rvdw(self):
 */
  /*else*/ {
    __pyx_v_self->forcefield_type = __pyx_e_10fatslimlib_9core_base_fft_unknown;
  }
  __pyx_L4:;

  /* "fatslimlib/core_base.pyx":1864
 * 
 * 
 *     cpdef fsl_int get_forcefield_type(self):             # <<<<<<<<<<<<<<
 *         cdef real rvdw
 *         if self.forcefield_type != fft_notset: # Already done
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("fatslimlib.core_base.Trajectory.get_forcefield_type", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_10Trajectory_5get_forcefield_type(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_10Trajectory_5get_forcefield_type(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_forcefield_type (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_10Trajectory_4get_forcefield_type(((struct __pyx_obj_10fatslimlib_9core_base_Trajectory *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_10Trajectory_4get_forcefield_type(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_forcefield_type", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_f_10fatslimlib_9core_base_10Trajectory_get_forcefield_type(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1864, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Trajectory.get_forcefield_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1880
 *             self.forcefield_type = fft_unknown
 * 
 *     cpdef real get_rvdw(self):             # <<<<<<<<<<<<<<
 *         cdef real[:, ::1] coords
 *         cdef fsl_int i
 */

static PyObject *__pyx_pw_10fatslimlib_9core_base_10Trajectory_7get_rvdw(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static real __pyx_f_10fatslimlib_9core_base_10Trajectory_get_rvdw(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self, int __pyx_skip_dispatch) {
  __Pyx_memviewslice __pyx_v_coords = { 0, 0, { 0 }, { 0 }, { 0 } };
  fsl_int __pyx_v_i;
  real __pyx_v_dist;
  real __pyx_v_min_dist;
  fsl_int __pyx_v_lastid;
  real __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  real __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_memviewslice __pyx_t_10 = { 0, 0, { 0 }, { 0 }, { 0 } };
  __Pyx_memviewslice __pyx_t_11 = { 0, 0, { 0 }, { 0 }, { 0 } };
  fsl_int __pyx_t_12;
  fsl_int __pyx_t_13;
  fsl_int __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  Py_ssize_t __pyx_t_19;
  Py_ssize_t __pyx_t_20;
  Py_ssize_t __pyx_t_21;
  Py_ssize_t __pyx_t_22;
  Py_ssize_t __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  Py_ssize_t __pyx_t_25;
  Py_ssize_t __pyx_t_26;
  Py_ssize_t __pyx_t_27;
  Py_ssize_t __pyx_t_28;
  Py_ssize_t __pyx_t_29;
  Py_ssize_t __pyx_t_30;
  Py_ssize_t __pyx_t_31;
  Py_ssize_t __pyx_t_32;
  Py_ssize_t __pyx_t_33;
  Py_ssize_t __pyx_t_34;
  Py_ssize_t __pyx_t_35;
  Py_ssize_t __pyx_t_36;
  Py_ssize_t __pyx_t_37;
  Py_ssize_t __pyx_t_38;
  __Pyx_RefNannySetupContext("get_rvdw", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_rvdw); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1880, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10fatslimlib_9core_base_10Trajectory_7get_rvdw)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1880, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == ((real)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1880, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "fatslimlib/core_base.pyx":1883
 *         cdef real[:, ::1] coords
 *         cdef fsl_int i
 *         cdef real dist, min_dist = 1.0e5             # <<<<<<<<<<<<<<
 *         cdef fsl_int lastid
 * 
 */
  __pyx_v_min_dist = 1.0e5;

  /* "fatslimlib/core_base.pyx":1886
 *         cdef fsl_int lastid
 * 
 *         if self.lipid_atomids is None:             # <<<<<<<<<<<<<<
 *             raise ValueError("Trajectory is not initialized correctly!")
 * 
 */
  if (unlikely(!__pyx_v_self->lipid_atomids.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1886, __pyx_L1_error)}
  __pyx_t_6 = ((((PyObject *) __pyx_v_self->lipid_atomids.memview) == Py_None) != 0);
  if (unlikely(__pyx_t_6)) {

    /* "fatslimlib/core_base.pyx":1887
 * 
 *         if self.lipid_atomids is None:
 *             raise ValueError("Trajectory is not initialized correctly!")             # <<<<<<<<<<<<<<
 * 
 *         # Load the first lipid from the first frame
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1887, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1887, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":1886
 *         cdef fsl_int lastid
 * 
 *         if self.lipid_atomids is None:             # <<<<<<<<<<<<<<
 *             raise ValueError("Trajectory is not initialized correctly!")
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":1890
 * 
 *         # Load the first lipid from the first frame
 *         coords = self.coords_reader.load_coords(0, np.arange(self.lipid_atomids_offsets[0]+1,             # <<<<<<<<<<<<<<
 *                                                              self.lipid_atomids_offsets[1]+1,
 *                                                              dtype=np.int64))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_arange); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_v_self->lipid_atomids_offsets.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1890, __pyx_L1_error)}
  __pyx_t_7 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_long(((*((fsl_int *) ( /* dim=0 */ (__pyx_v_self->lipid_atomids_offsets.data + __pyx_t_7 * __pyx_v_self->lipid_atomids_offsets.strides[0]) ))) + 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "fatslimlib/core_base.pyx":1891
 *         # Load the first lipid from the first frame
 *         coords = self.coords_reader.load_coords(0, np.arange(self.lipid_atomids_offsets[0]+1,
 *                                                              self.lipid_atomids_offsets[1]+1,             # <<<<<<<<<<<<<<
 *                                                              dtype=np.int64))
 * 
 */
  if (unlikely(!__pyx_v_self->lipid_atomids_offsets.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1891, __pyx_L1_error)}
  __pyx_t_8 = 1;
  __pyx_t_3 = __Pyx_PyInt_From_long(((*((fsl_int *) ( /* dim=0 */ (__pyx_v_self->lipid_atomids_offsets.data + __pyx_t_8 * __pyx_v_self->lipid_atomids_offsets.strides[0]) ))) + 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "fatslimlib/core_base.pyx":1890
 * 
 *         # Load the first lipid from the first frame
 *         coords = self.coords_reader.load_coords(0, np.arange(self.lipid_atomids_offsets[0]+1,             # <<<<<<<<<<<<<<
 *                                                              self.lipid_atomids_offsets[1]+1,
 *                                                              dtype=np.int64))
 */
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;

  /* "fatslimlib/core_base.pyx":1892
 *         coords = self.coords_reader.load_coords(0, np.arange(self.lipid_atomids_offsets[0]+1,
 *                                                              self.lipid_atomids_offsets[1]+1,
 *                                                              dtype=np.int64))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_int64); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_dtype, __pyx_t_9) < 0) __PYX_ERR(0, 1892, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "fatslimlib/core_base.pyx":1890
 * 
 *         # Load the first lipid from the first frame
 *         coords = self.coords_reader.load_coords(0, np.arange(self.lipid_atomids_offsets[0]+1,             # <<<<<<<<<<<<<<
 *                                                              self.lipid_atomids_offsets[1]+1,
 *                                                              dtype=np.int64))
 */
  __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_10 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(__pyx_t_9, PyBUF_WRITABLE); if (unlikely(!__pyx_t_10.memview)) __PYX_ERR(0, 1890, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_11 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_CoordinateReader *)__pyx_v_self->coords_reader->__pyx_vtab)->load_coords(__pyx_v_self->coords_reader, 0, __pyx_t_10); if (unlikely(!__pyx_t_11.memview)) __PYX_ERR(0, 1890, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __pyx_t_10.memview = NULL;
  __pyx_t_10.data = NULL;
  __pyx_v_coords = __pyx_t_11;
  __pyx_t_11.memview = NULL;
  __pyx_t_11.data = NULL;

  /* "fatslimlib/core_base.pyx":1900
 *         # corresponds to a C-H bond it means the FF is all-atom (because the H is non polar).
 *         # If the distance is C-C, wa have unified atom. If greater, it must be a coarse grained FF.
 *         lastid = coords.shape[0] - 1             # <<<<<<<<<<<<<<
 *         for i in range(lastid):
 *             dist = (coords[i, XX] - coords[lastid, XX]) * (coords[i, XX] - coords[lastid, XX]) +\
 */
  __pyx_v_lastid = ((__pyx_v_coords.shape[0]) - 1);

  /* "fatslimlib/core_base.pyx":1901
 *         # If the distance is C-C, wa have unified atom. If greater, it must be a coarse grained FF.
 *         lastid = coords.shape[0] - 1
 *         for i in range(lastid):             # <<<<<<<<<<<<<<
 *             dist = (coords[i, XX] - coords[lastid, XX]) * (coords[i, XX] - coords[lastid, XX]) +\
 *                 (coords[i, YY] - coords[lastid, YY]) * (coords[i, YY] - coords[lastid, YY]) +\
 */
  __pyx_t_12 = __pyx_v_lastid;
  __pyx_t_13 = __pyx_t_12;
  for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
    __pyx_v_i = __pyx_t_14;

    /* "fatslimlib/core_base.pyx":1902
 *         lastid = coords.shape[0] - 1
 *         for i in range(lastid):
 *             dist = (coords[i, XX] - coords[lastid, XX]) * (coords[i, XX] - coords[lastid, XX]) +\             # <<<<<<<<<<<<<<
 *                 (coords[i, YY] - coords[lastid, YY]) * (coords[i, YY] - coords[lastid, YY]) +\
 *                 (coords[i, ZZ] - coords[lastid, ZZ]) * (coords[i, ZZ] - coords[lastid, ZZ])
 */
    __pyx_t_15 = __pyx_v_i;
    __pyx_t_16 = 0;
    __pyx_t_17 = __pyx_v_lastid;
    __pyx_t_18 = 0;
    __pyx_t_19 = __pyx_v_i;
    __pyx_t_20 = 0;
    __pyx_t_21 = __pyx_v_lastid;
    __pyx_t_22 = 0;

    /* "fatslimlib/core_base.pyx":1903
 *         for i in range(lastid):
 *             dist = (coords[i, XX] - coords[lastid, XX]) * (coords[i, XX] - coords[lastid, XX]) +\
 *                 (coords[i, YY] - coords[lastid, YY]) * (coords[i, YY] - coords[lastid, YY]) +\             # <<<<<<<<<<<<<<
 *                 (coords[i, ZZ] - coords[lastid, ZZ]) * (coords[i, ZZ] - coords[lastid, ZZ])
 * 
 */
    __pyx_t_23 = __pyx_v_i;
    __pyx_t_24 = 1;
    __pyx_t_25 = __pyx_v_lastid;
    __pyx_t_26 = 1;
    __pyx_t_27 = __pyx_v_i;
    __pyx_t_28 = 1;
    __pyx_t_29 = __pyx_v_lastid;
    __pyx_t_30 = 1;

    /* "fatslimlib/core_base.pyx":1904
 *             dist = (coords[i, XX] - coords[lastid, XX]) * (coords[i, XX] - coords[lastid, XX]) +\
 *                 (coords[i, YY] - coords[lastid, YY]) * (coords[i, YY] - coords[lastid, YY]) +\
 *                 (coords[i, ZZ] - coords[lastid, ZZ]) * (coords[i, ZZ] - coords[lastid, ZZ])             # <<<<<<<<<<<<<<
 * 
 *             if dist < min_dist:
 */
    __pyx_t_31 = __pyx_v_i;
    __pyx_t_32 = 2;
    __pyx_t_33 = __pyx_v_lastid;
    __pyx_t_34 = 2;
    __pyx_t_35 = __pyx_v_i;
    __pyx_t_36 = 2;
    __pyx_t_37 = __pyx_v_lastid;
    __pyx_t_38 = 2;

    /* "fatslimlib/core_base.pyx":1903
 *         for i in range(lastid):
 *             dist = (coords[i, XX] - coords[lastid, XX]) * (coords[i, XX] - coords[lastid, XX]) +\
 *                 (coords[i, YY] - coords[lastid, YY]) * (coords[i, YY] - coords[lastid, YY]) +\             # <<<<<<<<<<<<<<
 *                 (coords[i, ZZ] - coords[lastid, ZZ]) * (coords[i, ZZ] - coords[lastid, ZZ])
 * 
 */
    __pyx_v_dist = (((((*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_15 * __pyx_v_coords.strides[0]) )) + __pyx_t_16)) ))) - (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_17 * __pyx_v_coords.strides[0]) )) + __pyx_t_18)) )))) * ((*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_19 * __pyx_v_coords.strides[0]) )) + __pyx_t_20)) ))) - (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_21 * __pyx_v_coords.strides[0]) )) + __pyx_t_22)) ))))) + (((*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_23 * __pyx_v_coords.strides[0]) )) + __pyx_t_24)) ))) - (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_25 * __pyx_v_coords.strides[0]) )) + __pyx_t_26)) )))) * ((*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_27 * __pyx_v_coords.strides[0]) )) + __pyx_t_28)) ))) - (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_29 * __pyx_v_coords.strides[0]) )) + __pyx_t_30)) )))))) + (((*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_31 * __pyx_v_coords.strides[0]) )) + __pyx_t_32)) ))) - (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_33 * __pyx_v_coords.strides[0]) )) + __pyx_t_34)) )))) * ((*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_35 * __pyx_v_coords.strides[0]) )) + __pyx_t_36)) ))) - (*((real *) ( /* dim=1 */ ((char *) (((real *) ( /* dim=0 */ (__pyx_v_coords.data + __pyx_t_37 * __pyx_v_coords.strides[0]) )) + __pyx_t_38)) ))))));

    /* "fatslimlib/core_base.pyx":1906
 *                 (coords[i, ZZ] - coords[lastid, ZZ]) * (coords[i, ZZ] - coords[lastid, ZZ])
 * 
 *             if dist < min_dist:             # <<<<<<<<<<<<<<
 *                 min_dist = dist
 * 
 */
    __pyx_t_6 = ((__pyx_v_dist < __pyx_v_min_dist) != 0);
    if (__pyx_t_6) {

      /* "fatslimlib/core_base.pyx":1907
 * 
 *             if dist < min_dist:
 *                 min_dist = dist             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_min_dist = __pyx_v_dist;

      /* "fatslimlib/core_base.pyx":1906
 *                 (coords[i, ZZ] - coords[lastid, ZZ]) * (coords[i, ZZ] - coords[lastid, ZZ])
 * 
 *             if dist < min_dist:             # <<<<<<<<<<<<<<
 *                 min_dist = dist
 * 
 */
    }
  }

  /* "fatslimlib/core_base.pyx":1910
 * 
 * 
 *         return min_dist             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_min_dist;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1880
 *             self.forcefield_type = fft_unknown
 * 
 *     cpdef real get_rvdw(self):             # <<<<<<<<<<<<<<
 *         cdef real[:, ::1] coords
 *         cdef fsl_int i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __PYX_XDEC_MEMVIEW(&__pyx_t_10, 1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_11, 1);
  __Pyx_WriteUnraisable("fatslimlib.core_base.Trajectory.get_rvdw", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __PYX_XDEC_MEMVIEW(&__pyx_v_coords, 1);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_10Trajectory_7get_rvdw(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_10Trajectory_7get_rvdw(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_rvdw (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_10Trajectory_6get_rvdw(((struct __pyx_obj_10fatslimlib_9core_base_Trajectory *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_10Trajectory_6get_rvdw(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_rvdw", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_10fatslimlib_9core_base_10Trajectory_get_rvdw(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Trajectory.get_rvdw", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1913
 * 
 * 
 *     cdef assert_initialization(self):             # <<<<<<<<<<<<<<
 *         if not self.initialized:
 *             raise ValueError("Trajectory not initialized")
 */

static PyObject *__pyx_f_10fatslimlib_9core_base_10Trajectory_assert_initialization(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("assert_initialization", 0);

  /* "fatslimlib/core_base.pyx":1914
 * 
 *     cdef assert_initialization(self):
 *         if not self.initialized:             # <<<<<<<<<<<<<<
 *             raise ValueError("Trajectory not initialized")
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_self->initialized != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "fatslimlib/core_base.pyx":1915
 *     cdef assert_initialization(self):
 *         if not self.initialized:
 *             raise ValueError("Trajectory not initialized")             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, fsl_int item):
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1915, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1915, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":1914
 * 
 *     cdef assert_initialization(self):
 *         if not self.initialized:             # <<<<<<<<<<<<<<
 *             raise ValueError("Trajectory not initialized")
 * 
 */
  }

  /* "fatslimlib/core_base.pyx":1913
 * 
 * 
 *     cdef assert_initialization(self):             # <<<<<<<<<<<<<<
 *         if not self.initialized:
 *             raise ValueError("Trajectory not initialized")
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("fatslimlib.core_base.Trajectory.assert_initialization", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1917
 *             raise ValueError("Trajectory not initialized")
 * 
 *     def __getitem__(self, fsl_int item):             # <<<<<<<<<<<<<<
 *         self.assert_initialization()
 *         cdef real timestep
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_10Trajectory_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_item); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_10Trajectory_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_item) {
  fsl_int __pyx_v_item;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_item); {
    __pyx_v_item = __Pyx_PyInt_As_fsl_int(__pyx_arg_item); if (unlikely((__pyx_v_item == ((fsl_int)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1917, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.Trajectory.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_10Trajectory_8__getitem__(((struct __pyx_obj_10fatslimlib_9core_base_Trajectory *)__pyx_v_self), ((fsl_int)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_10Trajectory_8__getitem__(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self, fsl_int __pyx_v_item) {
  real __pyx_v_timestep;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "fatslimlib/core_base.pyx":1918
 * 
 *     def __getitem__(self, fsl_int item):
 *         self.assert_initialization()             # <<<<<<<<<<<<<<
 *         cdef real timestep
 *         item = int(item)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10fatslimlib_9core_base_Trajectory *)__pyx_v_self->__pyx_vtab)->assert_initialization(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":1921
 *         cdef real timestep
 *         item = int(item)
 *         if not 0 <= item < self.coords_reader.nframes:             # <<<<<<<<<<<<<<
 *             raise IndexError("Frame index out of range (%i frames in trajectory)" %
 *                              self.coords_reader.nframes)
 */
  __pyx_t_2 = (0 <= __pyx_v_item);
  if (__pyx_t_2) {
    __pyx_t_2 = (__pyx_v_item < __pyx_v_self->coords_reader->nframes);
  }
  __pyx_t_3 = ((!(__pyx_t_2 != 0)) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "fatslimlib/core_base.pyx":1923
 *         if not 0 <= item < self.coords_reader.nframes:
 *             raise IndexError("Frame index out of range (%i frames in trajectory)" %
 *                              self.coords_reader.nframes)             # <<<<<<<<<<<<<<
 *         timestep = self.coords_reader.timesteps[item]
 *         return Frame(self, item, timestep)
 */
    __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_v_self->coords_reader->nframes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1923, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "fatslimlib/core_base.pyx":1922
 *         item = int(item)
 *         if not 0 <= item < self.coords_reader.nframes:
 *             raise IndexError("Frame index out of range (%i frames in trajectory)" %             # <<<<<<<<<<<<<<
 *                              self.coords_reader.nframes)
 *         timestep = self.coords_reader.timesteps[item]
 */
    __pyx_t_4 = PyUnicode_Format(__pyx_kp_u_Frame_index_out_of_range_i_frame, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1922, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1922, __pyx_L1_error)

    /* "fatslimlib/core_base.pyx":1921
 *         cdef real timestep
 *         item = int(item)
 *         if not 0 <= item < self.coords_reader.nframes:             # <<<<<<<<<<<<<<
 *             raise IndexError("Frame index out of range (%i frames in trajectory)" %
 *                              self.coords_reader.nframes)
 */
  }

  /* "fatslimlib/core_base.pyx":1924
 *             raise IndexError("Frame index out of range (%i frames in trajectory)" %
 *                              self.coords_reader.nframes)
 *         timestep = self.coords_reader.timesteps[item]             # <<<<<<<<<<<<<<
 *         return Frame(self, item, timestep)
 * 
 */
  if (unlikely(!__pyx_v_self->coords_reader->timesteps.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(0, 1924, __pyx_L1_error)}
  __pyx_t_5 = __pyx_v_item;
  __pyx_v_timestep = (*((real *) ( /* dim=0 */ (__pyx_v_self->coords_reader->timesteps.data + __pyx_t_5 * __pyx_v_self->coords_reader->timesteps.strides[0]) )));

  /* "fatslimlib/core_base.pyx":1925
 *                              self.coords_reader.nframes)
 *         timestep = self.coords_reader.timesteps[item]
 *         return Frame(self, item, timestep)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_fsl_int(__pyx_v_item); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_timestep); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_4);
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_10fatslimlib_9core_base_Frame), __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1917
 *             raise ValueError("Trajectory not initialized")
 * 
 *     def __getitem__(self, fsl_int item):             # <<<<<<<<<<<<<<
 *         self.assert_initialization()
 *         cdef real timestep
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("fatslimlib.core_base.Trajectory.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1927
 *         return Frame(self, item, timestep)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.coords_reader.nframes
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_10fatslimlib_9core_base_10Trajectory_11__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_10fatslimlib_9core_base_10Trajectory_11__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_10Trajectory_10__len__(((struct __pyx_obj_10fatslimlib_9core_base_Trajectory *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_10fatslimlib_9core_base_10Trajectory_10__len__(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "fatslimlib/core_base.pyx":1928
 * 
 *     def __len__(self):
 *         return self.coords_reader.nframes             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_self->coords_reader->nframes;
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1927
 *         return Frame(self, item, timestep)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.coords_reader.nframes
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pyx":1931
 * 
 * 
 *     def get_topology(self):             # <<<<<<<<<<<<<<
 *         return self.topology
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_10Trajectory_13get_topology(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_10Trajectory_13get_topology(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_topology (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_10Trajectory_12get_topology(((struct __pyx_obj_10fatslimlib_9core_base_Trajectory *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_10Trajectory_12get_topology(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_topology", 0);

  /* "fatslimlib/core_base.pyx":1932
 * 
 *     def get_topology(self):
 *         return self.topology             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->topology));
  __pyx_r = ((PyObject *)__pyx_v_self->topology);
  goto __pyx_L0;

  /* "fatslimlib/core_base.pyx":1931
 * 
 * 
 *     def get_topology(self):             # <<<<<<<<<<<<<<
 *         return self.topology
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fatslimlib/core_base.pxd":279
 *     cdef bint initialized
 * 
 *     cdef readonly list timesteps             # <<<<<<<<<<<<<<
 * 
 *     cdef fsl_int[:] hg_group_atomids, interacting_atomids, lipid_atomids, hg_bead_atomids
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_10Trajectory_9timesteps_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_10Trajectory_9timesteps_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_10Trajectory_9timesteps___get__(((struct __pyx_obj_10fatslimlib_9core_base_Trajectory *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_10Trajectory_9timesteps___get__(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->timesteps);
  __pyx_r = __pyx_v_self->timesteps;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_10Trajectory_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_10Trajectory_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_10Trajectory_14__reduce_cython__(((struct __pyx_obj_10fatslimlib_9core_base_Trajectory *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_10Trajectory_14__reduce_cython__(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_t_15;
  int __pyx_t_16;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.aggregates, self.be_verbose, self.bead_group_name, self.coords_reader, self.cur_frame, self.forcefield_type, self.hg_bead_atomids, self.hg_bead_atomids_offsets, self.hg_group_atomids, self.index_reader, self.initialized, self.interacting_atomids, self.interacting_group_name, self.lipid_atomids, self.lipid_atomids_offsets, self.lipid_hg_indices, self.membranes, self.normal_cutoff, self.timesteps, self.topol_reader, self.topology)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->be_verbose); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_fsl_int(__pyx_v_self->cur_frame); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_fsl_int(__pyx_v_self->forcefield_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (unlikely(!__pyx_v_self->hg_bead_atomids.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 5, __pyx_L1_error)}
  __pyx_t_4 = __pyx_memoryview_fromslice(__pyx_v_self->hg_bead_atomids, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_fsl_int, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_fsl_int, 0);; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (unlikely(!__pyx_v_self->hg_bead_atomids_offsets.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 5, __pyx_L1_error)}
  __pyx_t_5 = __pyx_memoryview_fromslice(__pyx_v_self->hg_bead_atomids_offsets, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_fsl_int, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_fsl_int, 0);; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (unlikely(!__pyx_v_self->hg_group_atomids.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 5, __pyx_L1_error)}
  __pyx_t_6 = __pyx_memoryview_fromslice(__pyx_v_self->hg_group_atomids, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_fsl_int, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_fsl_int, 0);; if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->initialized); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(!__pyx_v_self->interacting_atomids.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 5, __pyx_L1_error)}
  __pyx_t_8 = __pyx_memoryview_fromslice(__pyx_v_self->interacting_atomids, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_fsl_int, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_fsl_int, 0);; if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (unlikely(!__pyx_v_self->lipid_atomids.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 5, __pyx_L1_error)}
  __pyx_t_9 = __pyx_memoryview_fromslice(__pyx_v_self->lipid_atomids, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_fsl_int, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_fsl_int, 0);; if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  if (unlikely(!__pyx_v_self->lipid_atomids_offsets.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 5, __pyx_L1_error)}
  __pyx_t_10 = __pyx_memoryview_fromslice(__pyx_v_self->lipid_atomids_offsets, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_fsl_int, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_fsl_int, 0);; if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (unlikely(!__pyx_v_self->lipid_hg_indices.memview)) {PyErr_SetString(PyExc_AttributeError,"Memoryview is not initialized");__PYX_ERR(1, 5, __pyx_L1_error)}
  __pyx_t_11 = __pyx_memoryview_fromslice(__pyx_v_self->lipid_hg_indices, 1, (PyObject *(*)(char *)) __pyx_memview_get_nn_fsl_int, (int (*)(char *, PyObject *)) __pyx_memview_set_nn_fsl_int, 0);; if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = PyFloat_FromDouble(__pyx_v_self->normal_cutoff); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = PyTuple_New(21); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_INCREF(__pyx_v_self->aggregates);
  __Pyx_GIVEREF(__pyx_v_self->aggregates);
  PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_v_self->aggregates);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->bead_group_name);
  __Pyx_GIVEREF(__pyx_v_self->bead_group_name);
  PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_self->bead_group_name);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->coords_reader));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->coords_reader));
  PyTuple_SET_ITEM(__pyx_t_13, 3, ((PyObject *)__pyx_v_self->coords_reader));
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_13, 4, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_13, 5, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_13, 6, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_13, 7, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_13, 8, __pyx_t_6);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->index_reader));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->index_reader));
  PyTuple_SET_ITEM(__pyx_t_13, 9, ((PyObject *)__pyx_v_self->index_reader));
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_13, 10, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_13, 11, __pyx_t_8);
  __Pyx_INCREF(__pyx_v_self->interacting_group_name);
  __Pyx_GIVEREF(__pyx_v_self->interacting_group_name);
  PyTuple_SET_ITEM(__pyx_t_13, 12, __pyx_v_self->interacting_group_name);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_13, 13, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_13, 14, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_13, 15, __pyx_t_11);
  __Pyx_INCREF(__pyx_v_self->membranes);
  __Pyx_GIVEREF(__pyx_v_self->membranes);
  PyTuple_SET_ITEM(__pyx_t_13, 16, __pyx_v_self->membranes);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_13, 17, __pyx_t_12);
  __Pyx_INCREF(__pyx_v_self->timesteps);
  __Pyx_GIVEREF(__pyx_v_self->timesteps);
  PyTuple_SET_ITEM(__pyx_t_13, 18, __pyx_v_self->timesteps);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->topol_reader));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->topol_reader));
  PyTuple_SET_ITEM(__pyx_t_13, 19, ((PyObject *)__pyx_v_self->topol_reader));
  __Pyx_INCREF(((PyObject *)__pyx_v_self->topology));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->topology));
  PyTuple_SET_ITEM(__pyx_t_13, 20, ((PyObject *)__pyx_v_self->topology));
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_13);
  __pyx_t_13 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.aggregates, self.be_verbose, self.bead_group_name, self.coords_reader, self.cur_frame, self.forcefield_type, self.hg_bead_atomids, self.hg_bead_atomids_offsets, self.hg_group_atomids, self.index_reader, self.initialized, self.interacting_atomids, self.interacting_group_name, self.lipid_atomids, self.lipid_atomids_offsets, self.lipid_hg_indices, self.membranes, self.normal_cutoff, self.timesteps, self.topol_reader, self.topology)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_13 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_v__dict = __pyx_t_13;
  __pyx_t_13 = 0;

  /* "(tree fragment)":7
 *     state = (self.aggregates, self.be_verbose, self.bead_group_name, self.coords_reader, self.cur_frame, self.forcefield_type, self.hg_bead_atomids, self.hg_bead_atomids_offsets, self.hg_group_atomids, self.index_reader, self.initialized, self.interacting_atomids, self.interacting_group_name, self.lipid_atomids, self.lipid_atomids_offsets, self.lipid_hg_indices, self.membranes, self.normal_cutoff, self.timesteps, self.topol_reader, self.topology)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_14 = (__pyx_v__dict != Py_None);
  __pyx_t_15 = (__pyx_t_14 != 0);
  if (__pyx_t_15) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_13 = PyTuple_New(1); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_v__dict);
    __pyx_t_12 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_13); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_12));
    __pyx_t_12 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.aggregates is not None or self.bead_group_name is not None or self.coords_reader is not None or self.index_reader is not None or self.interacting_group_name is not None or self.membranes is not None or self.timesteps is not None or self.topol_reader is not None or self.topology is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.aggregates, self.be_verbose, self.bead_group_name, self.coords_reader, self.cur_frame, self.forcefield_type, self.hg_bead_atomids, self.hg_bead_atomids_offsets, self.hg_group_atomids, self.index_reader, self.initialized, self.interacting_atomids, self.interacting_group_name, self.lipid_atomids, self.lipid_atomids_offsets, self.lipid_hg_indices, self.membranes, self.normal_cutoff, self.timesteps, self.topol_reader, self.topology)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.aggregates is not None or self.bead_group_name is not None or self.coords_reader is not None or self.index_reader is not None or self.interacting_group_name is not None or self.membranes is not None or self.timesteps is not None or self.topol_reader is not None or self.topology is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Trajectory, (type(self), 0x6a20261, None), state
 */
  /*else*/ {
    __pyx_t_14 = (__pyx_v_self->aggregates != ((PyObject*)Py_None));
    __pyx_t_16 = (__pyx_t_14 != 0);
    if (!__pyx_t_16) {
    } else {
      __pyx_t_15 = __pyx_t_16;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_16 = (__pyx_v_self->bead_group_name != ((PyObject*)Py_None));
    __pyx_t_14 = (__pyx_t_16 != 0);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_15 = __pyx_t_14;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_14 = (((PyObject *)__pyx_v_self->coords_reader) != Py_None);
    __pyx_t_16 = (__pyx_t_14 != 0);
    if (!__pyx_t_16) {
    } else {
      __pyx_t_15 = __pyx_t_16;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_16 = (((PyObject *)__pyx_v_self->index_reader) != Py_None);
    __pyx_t_14 = (__pyx_t_16 != 0);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_15 = __pyx_t_14;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->interacting_group_name != ((PyObject*)Py_None));
    __pyx_t_16 = (__pyx_t_14 != 0);
    if (!__pyx_t_16) {
    } else {
      __pyx_t_15 = __pyx_t_16;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_16 = (__pyx_v_self->membranes != ((PyObject*)Py_None));
    __pyx_t_14 = (__pyx_t_16 != 0);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_15 = __pyx_t_14;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_14 = (__pyx_v_self->timesteps != ((PyObject*)Py_None));
    __pyx_t_16 = (__pyx_t_14 != 0);
    if (!__pyx_t_16) {
    } else {
      __pyx_t_15 = __pyx_t_16;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_16 = (((PyObject *)__pyx_v_self->topol_reader) != Py_None);
    __pyx_t_14 = (__pyx_t_16 != 0);
    if (!__pyx_t_14) {
    } else {
      __pyx_t_15 = __pyx_t_14;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_14 = (((PyObject *)__pyx_v_self->topology) != Py_None);
    __pyx_t_16 = (__pyx_t_14 != 0);
    __pyx_t_15 = __pyx_t_16;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_15;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.aggregates is not None or self.bead_group_name is not None or self.coords_reader is not None or self.index_reader is not None or self.interacting_group_name is not None or self.membranes is not None or self.timesteps is not None or self.topol_reader is not None or self.topology is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Trajectory, (type(self), 0x6a20261, None), state
 *     else:
 */
  __pyx_t_15 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_15) {

    /* "(tree fragment)":13
 *         use_setstate = self.aggregates is not None or self.bead_group_name is not None or self.coords_reader is not None or self.index_reader is not None or self.interacting_group_name is not None or self.membranes is not None or self.timesteps is not None or self.topol_reader is not None or self.topology is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Trajectory, (type(self), 0x6a20261, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Trajectory, (type(self), 0x6a20261, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_pyx_unpickle_Trajectory); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_13, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_111280737);
    __Pyx_GIVEREF(__pyx_int_111280737);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_int_111280737);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_13, 2, Py_None);
    __pyx_t_11 = PyTuple_New(3); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_12);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_13);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_v_state);
    __pyx_t_12 = 0;
    __pyx_t_13 = 0;
    __pyx_r = __pyx_t_11;
    __pyx_t_11 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.aggregates is not None or self.bead_group_name is not None or self.coords_reader is not None or self.index_reader is not None or self.interacting_group_name is not None or self.membranes is not None or self.timesteps is not None or self.topol_reader is not None or self.topology is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Trajectory, (type(self), 0x6a20261, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Trajectory, (type(self), 0x6a20261, None), state
 *     else:
 *         return __pyx_unpickle_Trajectory, (type(self), 0x6a20261, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Trajectory__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_pyx_unpickle_Trajectory); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = PyTuple_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_13, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_111280737);
    __Pyx_GIVEREF(__pyx_int_111280737);
    PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_int_111280737);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_v_state);
    __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_11);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_13);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_13);
    __pyx_t_11 = 0;
    __pyx_t_13 = 0;
    __pyx_r = __pyx_t_12;
    __pyx_t_12 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("fatslimlib.core_base.Trajectory.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Trajectory, (type(self), 0x6a20261, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Trajectory__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_10Trajectory_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_10fatslimlib_9core_base_10Trajectory_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_10Trajectory_16__setstate_cython__(((struct __pyx_obj_10fatslimlib_9core_base_Trajectory *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_10Trajectory_16__setstate_cython__(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Trajectory, (type(self), 0x6a20261, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Trajectory__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10fatslimlib_9core_base___pyx_unpickle_Trajectory__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Trajectory, (type(self), 0x6a20261, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Trajectory__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fatslimlib.core_base.Trajectory.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_TopologyReader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_17__pyx_unpickle_TopologyReader(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_10fatslimlib_9core_base_17__pyx_unpickle_TopologyReader = {"__pyx_unpickle_TopologyReader", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10fatslimlib_9core_base_17__pyx_unpickle_TopologyReader, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10fatslimlib_9core_base_17__pyx_unpickle_TopologyReader(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_TopologyReader (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TopologyReader", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TopologyReader", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_TopologyReader") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_TopologyReader", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.__pyx_unpickle_TopologyReader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_16__pyx_unpickle_TopologyReader(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_16__pyx_unpickle_TopologyReader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TopologyReader", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x8dafc6b:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x8dafc6b = (filename, topology))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x8dafc6b) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x8dafc6b:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x8dafc6b = (filename, topology))" % __pyx_checksum)
 *     __pyx_result = TopologyReader.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x8dafc6b:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x8dafc6b = (filename, topology))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = TopologyReader.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x8d, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x8dafc6b:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x8dafc6b = (filename, topology))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x8dafc6b = (filename, topology))" % __pyx_checksum)
 *     __pyx_result = TopologyReader.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TopologyReader__set_state(<TopologyReader> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10fatslimlib_9core_base_TopologyReader), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x8dafc6b = (filename, topology))" % __pyx_checksum)
 *     __pyx_result = TopologyReader.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TopologyReader__set_state(<TopologyReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = TopologyReader.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TopologyReader__set_state(<TopologyReader> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_TopologyReader__set_state(TopologyReader __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_10fatslimlib_9core_base___pyx_unpickle_TopologyReader__set_state(((struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x8dafc6b = (filename, topology))" % __pyx_checksum)
 *     __pyx_result = TopologyReader.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_TopologyReader__set_state(<TopologyReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_TopologyReader__set_state(<TopologyReader> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_TopologyReader__set_state(TopologyReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.filename = __pyx_state[0]; __pyx_result.topology = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_TopologyReader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.__pyx_unpickle_TopologyReader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_TopologyReader__set_state(<TopologyReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TopologyReader__set_state(TopologyReader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.filename = __pyx_state[0]; __pyx_result.topology = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10fatslimlib_9core_base___pyx_unpickle_TopologyReader__set_state(struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_TopologyReader__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_TopologyReader__set_state(TopologyReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.filename = __pyx_state[0]; __pyx_result.topology = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(PyBytes_CheckExact(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)))||((PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0))->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 0);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->filename);
  __Pyx_DECREF(__pyx_v___pyx_result->filename);
  __pyx_v___pyx_result->filename = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1), __pyx_ptype_10fatslimlib_9core_base_Topology))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 1);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->topology);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->topology));
  __pyx_v___pyx_result->topology = ((struct __pyx_obj_10fatslimlib_9core_base_Topology *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_TopologyReader__set_state(TopologyReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.filename = __pyx_state[0]; __pyx_result.topology = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 2) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.filename = __pyx_state[0]; __pyx_result.topology = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)) : __Pyx_PyObject_CallOneArg(__pyx_t_7, PyTuple_GET_ITEM(__pyx_v___pyx_state, 2));
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_TopologyReader__set_state(TopologyReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.filename = __pyx_state[0]; __pyx_result.topology = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_TopologyReader__set_state(<TopologyReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TopologyReader__set_state(TopologyReader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.filename = __pyx_state[0]; __pyx_result.topology = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("fatslimlib.core_base.__pyx_unpickle_TopologyReader__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_CoordinateReader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_19__pyx_unpickle_CoordinateReader(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_10fatslimlib_9core_base_19__pyx_unpickle_CoordinateReader = {"__pyx_unpickle_CoordinateReader", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10fatslimlib_9core_base_19__pyx_unpickle_CoordinateReader, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10fatslimlib_9core_base_19__pyx_unpickle_CoordinateReader(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_CoordinateReader (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CoordinateReader", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CoordinateReader", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_CoordinateReader") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_CoordinateReader", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.__pyx_unpickle_CoordinateReader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_18__pyx_unpickle_CoordinateReader(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_18__pyx_unpickle_CoordinateReader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CoordinateReader", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x500400c:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x500400c = (box_offsets, coordinate_offsets, filename, nframes, timesteps))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x500400c) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x500400c:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x500400c = (box_offsets, coordinate_offsets, filename, nframes, timesteps))" % __pyx_checksum)
 *     __pyx_result = CoordinateReader.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x500400c:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x500400c = (box_offsets, coordinate_offsets, filename, nframes, timesteps))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = CoordinateReader.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x50, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x500400c:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x500400c = (box_offsets, coordinate_offsets, filename, nframes, timesteps))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x500400c = (box_offsets, coordinate_offsets, filename, nframes, timesteps))" % __pyx_checksum)
 *     __pyx_result = CoordinateReader.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CoordinateReader__set_state(<CoordinateReader> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10fatslimlib_9core_base_CoordinateReader), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x500400c = (box_offsets, coordinate_offsets, filename, nframes, timesteps))" % __pyx_checksum)
 *     __pyx_result = CoordinateReader.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CoordinateReader__set_state(<CoordinateReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = CoordinateReader.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CoordinateReader__set_state(<CoordinateReader> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_CoordinateReader__set_state(CoordinateReader __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_10fatslimlib_9core_base___pyx_unpickle_CoordinateReader__set_state(((struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x500400c = (box_offsets, coordinate_offsets, filename, nframes, timesteps))" % __pyx_checksum)
 *     __pyx_result = CoordinateReader.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_CoordinateReader__set_state(<CoordinateReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_CoordinateReader__set_state(<CoordinateReader> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_CoordinateReader__set_state(CoordinateReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.box_offsets = __pyx_state[0]; __pyx_result.coordinate_offsets = __pyx_state[1]; __pyx_result.filename = __pyx_state[2]; __pyx_result.nframes = __pyx_state[3]; __pyx_result.timesteps = __pyx_state[4]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_CoordinateReader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.__pyx_unpickle_CoordinateReader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_CoordinateReader__set_state(<CoordinateReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CoordinateReader__set_state(CoordinateReader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.box_offsets = __pyx_state[0]; __pyx_result.coordinate_offsets = __pyx_state[1]; __pyx_result.filename = __pyx_state[2]; __pyx_result.nframes = __pyx_state[3]; __pyx_result.timesteps = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10fatslimlib_9core_base___pyx_unpickle_CoordinateReader__set_state(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1 = { 0, 0, { 0 }, { 0 }, { 0 } };
  PyObject *__pyx_t_2 = NULL;
  fsl_int __pyx_t_3;
  __Pyx_memviewslice __pyx_t_4 = { 0, 0, { 0 }, { 0 }, { 0 } };
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_CoordinateReader__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_CoordinateReader__set_state(CoordinateReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.box_offsets = __pyx_state[0]; __pyx_result.coordinate_offsets = __pyx_state[1]; __pyx_result.filename = __pyx_state[2]; __pyx_result.nframes = __pyx_state[3]; __pyx_result.timesteps = __pyx_state[4]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0), PyBUF_WRITABLE); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(1, 12, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->box_offsets, 0);
  __pyx_v___pyx_result->box_offsets = __pyx_t_1;
  __pyx_t_1.memview = NULL;
  __pyx_t_1.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1), PyBUF_WRITABLE); if (unlikely(!__pyx_t_1.memview)) __PYX_ERR(1, 12, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->coordinate_offsets, 0);
  __pyx_v___pyx_result->coordinate_offsets = __pyx_t_1;
  __pyx_t_1.memview = NULL;
  __pyx_t_1.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(PyBytes_CheckExact(PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)))||((PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(PyTuple_GET_ITEM(__pyx_v___pyx_state, 2))->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_2 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 2);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->filename);
  __Pyx_DECREF(__pyx_v___pyx_result->filename);
  __pyx_v___pyx_result->filename = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyInt_As_fsl_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)); if (unlikely((__pyx_t_3 == ((fsl_int)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->nframes = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(PyTuple_GET_ITEM(__pyx_v___pyx_state, 4), PyBUF_WRITABLE); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(1, 12, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->timesteps, 0);
  __pyx_v___pyx_result->timesteps = __pyx_t_4;
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_CoordinateReader__set_state(CoordinateReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.box_offsets = __pyx_state[0]; __pyx_result.coordinate_offsets = __pyx_state[1]; __pyx_result.filename = __pyx_state[2]; __pyx_result.nframes = __pyx_state[3]; __pyx_result.timesteps = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_6 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_7 = ((__pyx_t_6 > 5) != 0);
  if (__pyx_t_7) {
  } else {
    __pyx_t_5 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_8 = (__pyx_t_7 != 0);
  __pyx_t_5 = __pyx_t_8;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_5) {

    /* "(tree fragment)":14
 *     __pyx_result.box_offsets = __pyx_state[0]; __pyx_result.coordinate_offsets = __pyx_state[1]; __pyx_result.filename = __pyx_state[2]; __pyx_result.nframes = __pyx_state[3]; __pyx_result.timesteps = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[5])             # <<<<<<<<<<<<<<
 */
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_update); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_2 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_9, PyTuple_GET_ITEM(__pyx_v___pyx_state, 5)) : __Pyx_PyObject_CallOneArg(__pyx_t_10, PyTuple_GET_ITEM(__pyx_v___pyx_state, 5));
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_CoordinateReader__set_state(CoordinateReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.box_offsets = __pyx_state[0]; __pyx_result.coordinate_offsets = __pyx_state[1]; __pyx_result.filename = __pyx_state[2]; __pyx_result.nframes = __pyx_state[3]; __pyx_result.timesteps = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_CoordinateReader__set_state(<CoordinateReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_CoordinateReader__set_state(CoordinateReader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.box_offsets = __pyx_state[0]; __pyx_result.coordinate_offsets = __pyx_state[1]; __pyx_result.filename = __pyx_state[2]; __pyx_result.nframes = __pyx_state[3]; __pyx_result.timesteps = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __PYX_XDEC_MEMVIEW(&__pyx_t_1, 1);
  __Pyx_XDECREF(__pyx_t_2);
  __PYX_XDEC_MEMVIEW(&__pyx_t_4, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("fatslimlib.core_base.__pyx_unpickle_CoordinateReader__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_IndexReader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_21__pyx_unpickle_IndexReader(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_10fatslimlib_9core_base_21__pyx_unpickle_IndexReader = {"__pyx_unpickle_IndexReader", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10fatslimlib_9core_base_21__pyx_unpickle_IndexReader, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10fatslimlib_9core_base_21__pyx_unpickle_IndexReader(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_IndexReader (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IndexReader", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IndexReader", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_IndexReader") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IndexReader", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.__pyx_unpickle_IndexReader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_20__pyx_unpickle_IndexReader(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_20__pyx_unpickle_IndexReader(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IndexReader", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7ac6dcc:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7ac6dcc = (filename, groups, loaded))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x7ac6dcc) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7ac6dcc:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7ac6dcc = (filename, groups, loaded))" % __pyx_checksum)
 *     __pyx_result = IndexReader.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x7ac6dcc:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7ac6dcc = (filename, groups, loaded))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = IndexReader.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x7ac6dcc:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7ac6dcc = (filename, groups, loaded))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7ac6dcc = (filename, groups, loaded))" % __pyx_checksum)
 *     __pyx_result = IndexReader.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IndexReader__set_state(<IndexReader> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10fatslimlib_9core_base_IndexReader), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7ac6dcc = (filename, groups, loaded))" % __pyx_checksum)
 *     __pyx_result = IndexReader.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IndexReader__set_state(<IndexReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = IndexReader.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IndexReader__set_state(<IndexReader> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_IndexReader__set_state(IndexReader __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_10fatslimlib_9core_base___pyx_unpickle_IndexReader__set_state(((struct __pyx_obj_10fatslimlib_9core_base_IndexReader *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x7ac6dcc = (filename, groups, loaded))" % __pyx_checksum)
 *     __pyx_result = IndexReader.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IndexReader__set_state(<IndexReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IndexReader__set_state(<IndexReader> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_IndexReader__set_state(IndexReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.filename = __pyx_state[0]; __pyx_result.groups = __pyx_state[1]; __pyx_result.loaded = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IndexReader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.__pyx_unpickle_IndexReader", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_IndexReader__set_state(<IndexReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IndexReader__set_state(IndexReader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.filename = __pyx_state[0]; __pyx_result.groups = __pyx_state[1]; __pyx_result.loaded = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10fatslimlib_9core_base___pyx_unpickle_IndexReader__set_state(struct __pyx_obj_10fatslimlib_9core_base_IndexReader *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IndexReader__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_IndexReader__set_state(IndexReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.filename = __pyx_state[0]; __pyx_result.groups = __pyx_state[1]; __pyx_result.loaded = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(PyBytes_CheckExact(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)))||((PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0))->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 0);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->filename);
  __Pyx_DECREF(__pyx_v___pyx_result->filename);
  __pyx_v___pyx_result->filename = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 1);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->groups);
  __Pyx_DECREF(__pyx_v___pyx_result->groups);
  __pyx_v___pyx_result->groups = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->loaded = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_IndexReader__set_state(IndexReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.filename = __pyx_state[0]; __pyx_result.groups = __pyx_state[1]; __pyx_result.loaded = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 3) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.filename = __pyx_state[0]; __pyx_result.groups = __pyx_state[1]; __pyx_result.loaded = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)) : __Pyx_PyObject_CallOneArg(__pyx_t_7, PyTuple_GET_ITEM(__pyx_v___pyx_state, 3));
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_IndexReader__set_state(IndexReader __pyx_result, tuple __pyx_state):
 *     __pyx_result.filename = __pyx_state[0]; __pyx_result.groups = __pyx_state[1]; __pyx_result.loaded = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_IndexReader__set_state(<IndexReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IndexReader__set_state(IndexReader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.filename = __pyx_state[0]; __pyx_result.groups = __pyx_state[1]; __pyx_result.loaded = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("fatslimlib.core_base.__pyx_unpickle_IndexReader__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Trajectory(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10fatslimlib_9core_base_23__pyx_unpickle_Trajectory(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_10fatslimlib_9core_base_23__pyx_unpickle_Trajectory = {"__pyx_unpickle_Trajectory", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10fatslimlib_9core_base_23__pyx_unpickle_Trajectory, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10fatslimlib_9core_base_23__pyx_unpickle_Trajectory(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Trajectory (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Trajectory", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Trajectory", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Trajectory") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Trajectory", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fatslimlib.core_base.__pyx_unpickle_Trajectory", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10fatslimlib_9core_base_22__pyx_unpickle_Trajectory(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10fatslimlib_9core_base_22__pyx_unpickle_Trajectory(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Trajectory", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x6a20261:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x6a20261 = (aggregates, be_verbose, bead_group_name, coords_reader, cur_frame, forcefield_type, hg_bead_atomids, hg_bead_atomids_offsets, hg_group_atomids, index_reader, initialized, interacting_atomids, interacting_group_name, lipid_atomids, lipid_atomids_offsets, lipid_hg_indices, membranes, normal_cutoff, timesteps, topol_reader, topology))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x6a20261) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x6a20261:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x6a20261 = (aggregates, be_verbose, bead_group_name, coords_reader, cur_frame, forcefield_type, hg_bead_atomids, hg_bead_atomids_offsets, hg_group_atomids, index_reader, initialized, interacting_atomids, interacting_group_name, lipid_atomids, lipid_atomids_offsets, lipid_hg_indices, membranes, normal_cutoff, timesteps, topol_reader, topology))" % __pyx_checksum)
 *     __pyx_result = Trajectory.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x6a20261:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x6a20261 = (aggregates, be_verbose, bead_group_name, coords_reader, cur_frame, forcefield_type, hg_bead_atomids, hg_bead_atomids_offsets, hg_group_atomids, index_reader, initialized, interacting_atomids, interacting_group_name, lipid_atomids, lipid_atomids_offsets, lipid_hg_indices, membranes, normal_cutoff, timesteps, topol_reader, topology))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Trajectory.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x6a, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x6a20261:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x6a20261 = (aggregates, be_verbose, bead_group_name, coords_reader, cur_frame, forcefield_type, hg_bead_atomids, hg_bead_atomids_offsets, hg_group_atomids, index_reader, initialized, interacting_atomids, interacting_group_name, lipid_atomids, lipid_atomids_offsets, lipid_hg_indices, membranes, normal_cutoff, timesteps, topol_reader, topology))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x6a20261 = (aggregates, be_verbose, bead_group_name, coords_reader, cur_frame, forcefield_type, hg_bead_atomids, hg_bead_atomids_offsets, hg_group_atomids, index_reader, initialized, interacting_atomids, interacting_group_name, lipid_atomids, lipid_atomids_offsets, lipid_hg_indices, membranes, normal_cutoff, timesteps, topol_reader, topology))" % __pyx_checksum)
 *     __pyx_result = Trajectory.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Trajectory__set_state(<Trajectory> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10fatslimlib_9core_base_Trajectory), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x6a20261 = (aggregates, be_verbose, bead_group_name, coords_reader, cur_frame, forcefield_type, hg_bead_atomids, hg_bead_atomids_offsets, hg_group_atomids, index_reader, initialized, interacting_atomids, interacting_group_name, lipid_atomids, lipid_atomids_offsets, lipid_hg_indices, membranes, normal_cutoff, timesteps, topol_reader, topology))" % __pyx_checksum)
 *     __pyx_result = Trajectory.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Trajectory__set_state(<Trajectory> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Trajectory.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Trajectory__set_state(<Trajectory> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Trajectory__set_state(Trajectory __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_10fatslimlib_9core_base___pyx_unpickle_Trajectory__set_state(((struct __pyx_obj_10fatslimlib_9core_base_Trajectory *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x6a20261 = (aggregates, be_verbose, bead_group_name, coords_reader, cur_frame, forcefield_type, hg_bead_atomids, hg_bead_atomids_offsets, hg_group_atomids, index_reader, initialized, interacting_atomids, interacting_group_name, lipid_atomids, lipid_atomids_offsets, lipid_hg_indices, membranes, normal_cutoff, timesteps, topol_reader, topology))" % __pyx_checksum)
 *     __pyx_result = Trajectory.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Trajectory__set_state(<Trajectory> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Trajectory__set_state(<Trajectory> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Trajectory__set_state(Trajectory __pyx_result, tuple __pyx_state):
 *     __pyx_result.aggregates = __pyx_state[0]; __pyx_result.be_verbose = __pyx_state[1]; __pyx_result.bead_group_name = __pyx_state[2]; __pyx_result.coords_reader = __pyx_state[3]; __pyx_result.cur_frame = __pyx_state[4]; __pyx_result.forcefield_type = __pyx_state[5]; __pyx_result.hg_bead_atomids = __pyx_state[6]; __pyx_result.hg_bead_atomids_offsets = __pyx_state[7]; __pyx_result.hg_group_atomids = __pyx_state[8]; __pyx_result.index_reader = __pyx_state[9]; __pyx_result.initialized = __pyx_state[10]; __pyx_result.interacting_atomids = __pyx_state[11]; __pyx_result.interacting_group_name = __pyx_state[12]; __pyx_result.lipid_atomids = __pyx_state[13]; __pyx_result.lipid_atomids_offsets = __pyx_state[14]; __pyx_result.lipid_hg_indices = __pyx_state[15]; __pyx_result.membranes = __pyx_state[16]; __pyx_result.normal_cutoff = __pyx_state[17]; __pyx_result.timesteps = __pyx_state[18]; __pyx_result.topol_reader = __pyx_state[19]; __pyx_result.topology = __pyx_state[20]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Trajectory(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fatslimlib.core_base.__pyx_unpickle_Trajectory", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Trajectory__set_state(<Trajectory> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Trajectory__set_state(Trajectory __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.aggregates = __pyx_state[0]; __pyx_result.be_verbose = __pyx_state[1]; __pyx_result.bead_group_name = __pyx_state[2]; __pyx_result.coords_reader = __pyx_state[3]; __pyx_result.cur_frame = __pyx_state[4]; __pyx_result.forcefield_type = __pyx_state[5]; __pyx_result.hg_bead_atomids = __pyx_state[6]; __pyx_result.hg_bead_atomids_offsets = __pyx_state[7]; __pyx_result.hg_group_atomids = __pyx_state[8]; __pyx_result.index_reader = __pyx_state[9]; __pyx_result.initialized = __pyx_state[10]; __pyx_result.interacting_atomids = __pyx_state[11]; __pyx_result.interacting_group_name = __pyx_state[12]; __pyx_result.lipid_atomids = __pyx_state[13]; __pyx_result.lipid_atomids_offsets = __pyx_state[14]; __pyx_result.lipid_hg_indices = __pyx_state[15]; __pyx_result.membranes = __pyx_state[16]; __pyx_result.normal_cutoff = __pyx_state[17]; __pyx_result.timesteps = __pyx_state[18]; __pyx_result.topol_reader = __pyx_state[19]; __pyx_result.topology = __pyx_state[20]
 *     if len(__pyx_state) > 21 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10fatslimlib_9core_base___pyx_unpickle_Trajectory__set_state(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  fsl_int __pyx_t_3;
  __Pyx_memviewslice __pyx_t_4 = { 0, 0, { 0 }, { 0 }, { 0 } };
  real __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Trajectory__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Trajectory__set_state(Trajectory __pyx_result, tuple __pyx_state):
 *     __pyx_result.aggregates = __pyx_state[0]; __pyx_result.be_verbose = __pyx_state[1]; __pyx_result.bead_group_name = __pyx_state[2]; __pyx_result.coords_reader = __pyx_state[3]; __pyx_result.cur_frame = __pyx_state[4]; __pyx_result.forcefield_type = __pyx_state[5]; __pyx_result.hg_bead_atomids = __pyx_state[6]; __pyx_result.hg_bead_atomids_offsets = __pyx_state[7]; __pyx_result.hg_group_atomids = __pyx_state[8]; __pyx_result.index_reader = __pyx_state[9]; __pyx_result.initialized = __pyx_state[10]; __pyx_result.interacting_atomids = __pyx_state[11]; __pyx_result.interacting_group_name = __pyx_state[12]; __pyx_result.lipid_atomids = __pyx_state[13]; __pyx_result.lipid_atomids_offsets = __pyx_state[14]; __pyx_result.lipid_hg_indices = __pyx_state[15]; __pyx_result.membranes = __pyx_state[16]; __pyx_result.normal_cutoff = __pyx_state[17]; __pyx_result.timesteps = __pyx_state[18]; __pyx_result.topol_reader = __pyx_state[19]; __pyx_result.topology = __pyx_state[20]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 21 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[21])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(PyList_CheckExact(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)))||((PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0))->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 0);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->aggregates);
  __Pyx_DECREF(__pyx_v___pyx_result->aggregates);
  __pyx_v___pyx_result->aggregates = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->be_verbose = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(PyBytes_CheckExact(PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)))||((PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(PyTuple_GET_ITEM(__pyx_v___pyx_state, 2))->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 2);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->bead_group_name);
  __Pyx_DECREF(__pyx_v___pyx_result->bead_group_name);
  __pyx_v___pyx_result->bead_group_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 3), __pyx_ptype_10fatslimlib_9core_base_CoordinateReader))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 3);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->coords_reader);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->coords_reader));
  __pyx_v___pyx_result->coords_reader = ((struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyInt_As_fsl_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 4)); if (unlikely((__pyx_t_3 == ((fsl_int)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->cur_frame = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyInt_As_fsl_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 5)); if (unlikely((__pyx_t_3 == ((fsl_int)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->forcefield_type = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 6), PyBUF_WRITABLE); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(1, 12, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->hg_bead_atomids, 0);
  __pyx_v___pyx_result->hg_bead_atomids = __pyx_t_4;
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 7), PyBUF_WRITABLE); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(1, 12, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->hg_bead_atomids_offsets, 0);
  __pyx_v___pyx_result->hg_bead_atomids_offsets = __pyx_t_4;
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 8), PyBUF_WRITABLE); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(1, 12, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->hg_group_atomids, 0);
  __pyx_v___pyx_result->hg_group_atomids = __pyx_t_4;
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 9)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 9), __pyx_ptype_10fatslimlib_9core_base_IndexReader))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 9);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->index_reader);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->index_reader));
  __pyx_v___pyx_result->index_reader = ((struct __pyx_obj_10fatslimlib_9core_base_IndexReader *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(PyTuple_GET_ITEM(__pyx_v___pyx_state, 10)); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->initialized = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 11), PyBUF_WRITABLE); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(1, 12, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->interacting_atomids, 0);
  __pyx_v___pyx_result->interacting_atomids = __pyx_t_4;
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(PyBytes_CheckExact(PyTuple_GET_ITEM(__pyx_v___pyx_state, 12)))||((PyTuple_GET_ITEM(__pyx_v___pyx_state, 12)) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(PyTuple_GET_ITEM(__pyx_v___pyx_state, 12))->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 12);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->interacting_group_name);
  __Pyx_DECREF(__pyx_v___pyx_result->interacting_group_name);
  __pyx_v___pyx_result->interacting_group_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 13), PyBUF_WRITABLE); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(1, 12, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->lipid_atomids, 0);
  __pyx_v___pyx_result->lipid_atomids = __pyx_t_4;
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 14), PyBUF_WRITABLE); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(1, 12, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->lipid_atomids_offsets, 0);
  __pyx_v___pyx_result->lipid_atomids_offsets = __pyx_t_4;
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 15), PyBUF_WRITABLE); if (unlikely(!__pyx_t_4.memview)) __PYX_ERR(1, 12, __pyx_L1_error)
  __PYX_XDEC_MEMVIEW(&__pyx_v___pyx_result->lipid_hg_indices, 0);
  __pyx_v___pyx_result->lipid_hg_indices = __pyx_t_4;
  __pyx_t_4.memview = NULL;
  __pyx_t_4.data = NULL;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(PyList_CheckExact(PyTuple_GET_ITEM(__pyx_v___pyx_state, 16)))||((PyTuple_GET_ITEM(__pyx_v___pyx_state, 16)) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(PyTuple_GET_ITEM(__pyx_v___pyx_state, 16))->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 16);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->membranes);
  __Pyx_DECREF(__pyx_v___pyx_result->membranes);
  __pyx_v___pyx_result->membranes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_5 = __pyx_PyFloat_AsDouble(PyTuple_GET_ITEM(__pyx_v___pyx_state, 17)); if (unlikely((__pyx_t_5 == ((real)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->normal_cutoff = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(PyList_CheckExact(PyTuple_GET_ITEM(__pyx_v___pyx_state, 18)))||((PyTuple_GET_ITEM(__pyx_v___pyx_state, 18)) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(PyTuple_GET_ITEM(__pyx_v___pyx_state, 18))->tp_name), 0))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 18);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->timesteps);
  __Pyx_DECREF(__pyx_v___pyx_result->timesteps);
  __pyx_v___pyx_result->timesteps = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 19)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 19), __pyx_ptype_10fatslimlib_9core_base_TopologyReader))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 19);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->topol_reader);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->topol_reader));
  __pyx_v___pyx_result->topol_reader = ((struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 20)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 20), __pyx_ptype_10fatslimlib_9core_base_Topology))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 20);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->topology);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->topology));
  __pyx_v___pyx_result->topology = ((struct __pyx_obj_10fatslimlib_9core_base_Topology *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Trajectory__set_state(Trajectory __pyx_result, tuple __pyx_state):
 *     __pyx_result.aggregates = __pyx_state[0]; __pyx_result.be_verbose = __pyx_state[1]; __pyx_result.bead_group_name = __pyx_state[2]; __pyx_result.coords_reader = __pyx_state[3]; __pyx_result.cur_frame = __pyx_state[4]; __pyx_result.forcefield_type = __pyx_state[5]; __pyx_result.hg_bead_atomids = __pyx_state[6]; __pyx_result.hg_bead_atomids_offsets = __pyx_state[7]; __pyx_result.hg_group_atomids = __pyx_state[8]; __pyx_result.index_reader = __pyx_state[9]; __pyx_result.initialized = __pyx_state[10]; __pyx_result.interacting_atomids = __pyx_state[11]; __pyx_result.interacting_group_name = __pyx_state[12]; __pyx_result.lipid_atomids = __pyx_state[13]; __pyx_result.lipid_atomids_offsets = __pyx_state[14]; __pyx_result.lipid_hg_indices = __pyx_state[15]; __pyx_result.membranes = __pyx_state[16]; __pyx_result.normal_cutoff = __pyx_state[17]; __pyx_result.timesteps = __pyx_state[18]; __pyx_result.topol_reader = __pyx_state[19]; __pyx_result.topology = __pyx_state[20]
 *     if len(__pyx_state) > 21 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[21])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_6 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_7 = ((__pyx_t_6 > 21) != 0);
  if (__pyx_t_7) {
  } else {
    __pyx_t_2 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_8 = (__pyx_t_7 != 0);
  __pyx_t_2 = __pyx_t_8;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.aggregates = __pyx_state[0]; __pyx_result.be_verbose = __pyx_state[1]; __pyx_result.bead_group_name = __pyx_state[2]; __pyx_result.coords_reader = __pyx_state[3]; __pyx_result.cur_frame = __pyx_state[4]; __pyx_result.forcefield_type = __pyx_state[5]; __pyx_result.hg_bead_atomids = __pyx_state[6]; __pyx_result.hg_bead_atomids_offsets = __pyx_state[7]; __pyx_result.hg_group_atomids = __pyx_state[8]; __pyx_result.index_reader = __pyx_state[9]; __pyx_result.initialized = __pyx_state[10]; __pyx_result.interacting_atomids = __pyx_state[11]; __pyx_result.interacting_group_name = __pyx_state[12]; __pyx_result.lipid_atomids = __pyx_state[13]; __pyx_result.lipid_atomids_offsets = __pyx_state[14]; __pyx_result.lipid_hg_indices = __pyx_state[15]; __pyx_result.membranes = __pyx_state[16]; __pyx_result.normal_cutoff = __pyx_state[17]; __pyx_result.timesteps = __pyx_state[18]; __pyx_result.topol_reader = __pyx_state[19]; __pyx_result.topology = __pyx_state[20]
 *     if len(__pyx_state) > 21 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[21])             # <<<<<<<<<<<<<<
 */
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_update); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_9, PyTuple_GET_ITEM(__pyx_v___pyx_state, 21)) : __Pyx_PyObject_CallOneArg(__pyx_t_10, PyTuple_GET_ITEM(__pyx_v___pyx_state, 21));
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Trajectory__set_state(Trajectory __pyx_result, tuple __pyx_state):
 *     __pyx_result.aggregates = __pyx_state[0]; __pyx_result.be_verbose = __pyx_state[1]; __pyx_result.bead_group_name = __pyx_state[2]; __pyx_result.coords_reader = __pyx_state[3]; __pyx_result.cur_frame = __pyx_state[4]; __pyx_result.forcefield_type = __pyx_state[5]; __pyx_result.hg_bead_atomids = __pyx_state[6]; __pyx_result.hg_bead_atomids_offsets = __pyx_state[7]; __pyx_result.hg_group_atomids = __pyx_state[8]; __pyx_result.index_reader = __pyx_state[9]; __pyx_result.initialized = __pyx_state[10]; __pyx_result.interacting_atomids = __pyx_state[11]; __pyx_result.interacting_group_name = __pyx_state[12]; __pyx_result.lipid_atomids = __pyx_state[13]; __pyx_result.lipid_atomids_offsets = __pyx_state[14]; __pyx_result.lipid_hg_indices = __pyx_state[15]; __pyx_result.membranes = __pyx_state[16]; __pyx_result.normal_cutoff = __pyx_state[17]; __pyx_result.timesteps = __pyx_state[18]; __pyx_result.topol_reader = __pyx_state[19]; __pyx_result.topology = __pyx_state[20]
 *     if len(__pyx_state) > 21 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[21])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Trajectory__set_state(<Trajectory> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Trajectory__set_state(Trajectory __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.aggregates = __pyx_state[0]; __pyx_result.be_verbose = __pyx_state[1]; __pyx_result.bead_group_name = __pyx_state[2]; __pyx_result.coords_reader = __pyx_state[3]; __pyx_result.cur_frame = __pyx_state[4]; __pyx_result.forcefield_type = __pyx_state[5]; __pyx_result.hg_bead_atomids = __pyx_state[6]; __pyx_result.hg_bead_atomids_offsets = __pyx_state[7]; __pyx_result.hg_group_atomids = __pyx_state[8]; __pyx_result.index_reader = __pyx_state[9]; __pyx_result.initialized = __pyx_state[10]; __pyx_result.interacting_atomids = __pyx_state[11]; __pyx_result.interacting_group_name = __pyx_state[12]; __pyx_result.lipid_atomids = __pyx_state[13]; __pyx_result.lipid_atomids_offsets = __pyx_state[14]; __pyx_result.lipid_hg_indices = __pyx_state[15]; __pyx_result.membranes = __pyx_state[16]; __pyx_result.normal_cutoff = __pyx_state[17]; __pyx_result.timesteps = __pyx_state[18]; __pyx_result.topol_reader = __pyx_state[19]; __pyx_result.topology = __pyx_state[20]
 *     if len(__pyx_state) > 21 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __PYX_XDEC_MEMVIEW(&__pyx_t_4, 1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("fatslimlib.core_base.__pyx_unpickle_Trajectory__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

/* Python wrapper */
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_shape = 0;
  Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_format = 0;
  PyObject *__pyx_v_mode = 0;
  int __pyx_v_allocate_buffer;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_shape,&__pyx_n_s_itemsize,&__pyx_n_s_format,&__pyx_n_s_mode,&__pyx_n_s_allocate_buffer,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[3] = ((PyObject *)__pyx_n_s_c);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_itemsize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 1); __PYX_ERR(1, 122, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_format)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 2); __PYX_ERR(1, 122, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mode);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_allocate_buffer);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 122, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_shape = ((PyObject*)values[0]);
    __pyx_v_itemsize = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_itemsize == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 122, __pyx_L3_error)
    __pyx_v_format = values[2];
    __pyx_v_mode = values[3];
    if (values[4]) {
      __pyx_v_allocate_buffer = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_allocate_buffer == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 123, __pyx_L3_error)
    } else {

      /* "View.MemoryView":123
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,
 *                   mode="c", bint allocate_buffer=True):             # <<<<<<<<<<<<<<
 * 
 *         cdef int idx
 */
      __pyx_v_allocate_buffer = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 122, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shape), (&PyTuple_Type), 1, "shape", 1))) __PYX_ERR(1, 122, __pyx_L1_error)
  if (unlikely(((PyObject *)__pyx_v_format) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "format"); __PYX_ERR(1, 122, __pyx_L1_error)
  }
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(((struct __pyx_array_obj *)__pyx_v_self), __pyx_v_shape, __pyx_v_itemsize, __pyx_v_format, __pyx_v_mode, __pyx_v_allocate_buffer);

  /* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_dim;
  PyObject **__pyx_v_p;
  char __pyx_v_order;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  char *__pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_format);

  /* "View.MemoryView":129
 *         cdef PyObject **p
 * 
 *         self.ndim = <int> len(shape)             # <<<<<<<<<<<<<<
 *         self.itemsize = itemsize
 * 
 */
  if (unlikely(__pyx_v_shape == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 129, __pyx_L1_error)
  }
  __pyx_t_1 = PyTuple_GET_SIZE(__pyx_v_shape); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 129, __pyx_L1_error)
  __pyx_v_self->ndim = ((int)__pyx_t_1);

  /* "View.MemoryView":130
 * 
 *         self.ndim = <int> len(shape)
 *         self.itemsize = itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not self.ndim:
 */
  __pyx_v_self->itemsize = __pyx_v_itemsize;

  /* "View.MemoryView":132
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_self->ndim != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":133
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 133, __pyx_L1_error)

    /* "View.MemoryView":132
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */
  }

  /* "View.MemoryView":135
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */
  __pyx_t_2 = ((__pyx_v_itemsize <= 0) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":136
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 136, __pyx_L1_error)

    /* "View.MemoryView":135
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */
  }

  /* "View.MemoryView":138
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_format); 
  __pyx_t_4 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_4) {

    /* "View.MemoryView":139
 * 
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')             # <<<<<<<<<<<<<<
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_format, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_n_s_ASCII) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_s_ASCII);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_format, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":138
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  }

  /* "View.MemoryView":140
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string             # <<<<<<<<<<<<<<
 *         self.format = self._format
 * 
 */
  if (!(likely(PyBytes_CheckExact(__pyx_v_format))||((__pyx_v_format) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_format)->tp_name), 0))) __PYX_ERR(1, 140, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_format;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_format);
  __Pyx_DECREF(__pyx_v_self->_format);
  __pyx_v_self->_format = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":141
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_self->_format == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(1, 141, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyBytes_AsWritableString(__pyx_v_self->_format); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(1, 141, __pyx_L1_error)
  __pyx_v_self->format = __pyx_t_7;

  /* "View.MemoryView":144
 * 
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)             # <<<<<<<<<<<<<<
 *         self._strides = self._shape + self.ndim
 * 
 */
  __pyx_v_self->_shape = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * __pyx_v_self->ndim) * 2)));

  /* "View.MemoryView":145
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)
 *         self._strides = self._shape + self.ndim             # <<<<<<<<<<<<<<
 * 
 *         if not self._shape:
 */
  __pyx_v_self->_strides = (__pyx_v_self->_shape + __pyx_v_self->ndim);

  /* "View.MemoryView":147
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */
  __pyx_t_4 = ((!(__pyx_v_self->_shape != 0)) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "View.MemoryView":148
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 148, __pyx_L1_error)

    /* "View.MemoryView":147
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */
  }

  /* "View.MemoryView":151
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */
  __pyx_t_8 = 0;
  __pyx_t_3 = __pyx_v_shape; __Pyx_INCREF(__pyx_t_3); __pyx_t_1 = 0;
  for (;;) {
    if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_5); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(1, 151, __pyx_L1_error)
    #else
    __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 151, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_9;
    __pyx_v_idx = __pyx_t_8;
    __pyx_t_8 = (__pyx_t_8 + 1);

    /* "View.MemoryView":152
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */
    __pyx_t_4 = ((__pyx_v_dim <= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":153
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))             # <<<<<<<<<<<<<<
 *             self._shape[idx] = dim
 * 
 */
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_6);
      __pyx_t_5 = 0;
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(1, 153, __pyx_L1_error)

      /* "View.MemoryView":152
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */
    }

    /* "View.MemoryView":154
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim             # <<<<<<<<<<<<<<
 * 
 *         cdef char order
 */
    (__pyx_v_self->_shape[__pyx_v_idx]) = __pyx_v_dim;

    /* "View.MemoryView":151
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":157
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_fortran, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 157, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "View.MemoryView":158
 *         cdef char order
 *         if mode == 'fortran':
 *             order = b'F'             # <<<<<<<<<<<<<<
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 */
    __pyx_v_order = 'F';

    /* "View.MemoryView":159
 *         if mode == 'fortran':
 *             order = b'F'
 *             self.mode = u'fortran'             # <<<<<<<<<<<<<<
 *         elif mode == 'c':
 *             order = b'C'
 */
    __Pyx_INCREF(__pyx_n_u_fortran);
    __Pyx_GIVEREF(__pyx_n_u_fortran);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_fortran;

    /* "View.MemoryView":157
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":160
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_c, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 160, __pyx_L1_error)
  if (likely(__pyx_t_4)) {

    /* "View.MemoryView":161
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 *             order = b'C'             # <<<<<<<<<<<<<<
 *             self.mode = u'c'
 *         else:
 */
    __pyx_v_order = 'C';

    /* "View.MemoryView":162
 *         elif mode == 'c':
 *             order = b'C'
 *             self.mode = u'c'             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 */
    __Pyx_INCREF(__pyx_n_u_c);
    __Pyx_GIVEREF(__pyx_n_u_c);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_c;

    /* "View.MemoryView":160
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":164
 *             self.mode = u'c'
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)             # <<<<<<<<<<<<<<
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_v_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_10, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_ERR(1, 164, __pyx_L1_error)
  }
  __pyx_L10:;

  /* "View.MemoryView":166
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,             # <<<<<<<<<<<<<<
 *                                              itemsize, self.ndim, order)
 * 
 */
  __pyx_v_self->len = __pyx_fill_contig_strides_array(__pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_itemsize, __pyx_v_self->ndim, __pyx_v_order);

  /* "View.MemoryView":169
 *                                              itemsize, self.ndim, order)
 * 
 *         self.free_data = allocate_buffer             # <<<<<<<<<<<<<<
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:
 */
  __pyx_v_self->free_data = __pyx_v_allocate_buffer;

  /* "View.MemoryView":170
 * 
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'             # <<<<<<<<<<<<<<
 *         if allocate_buffer:
 * 
 */
  __pyx_t_10 = PyObject_RichCompare(__pyx_v_format, __pyx_n_b_O, Py_EQ); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 170, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_self->dtype_is_object = __pyx_t_4;

  /* "View.MemoryView":171
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = (__pyx_v_allocate_buffer != 0);
  if (__pyx_t_4) {

    /* "View.MemoryView":174
 * 
 * 
 *             self.data = <char *>malloc(self.len)             # <<<<<<<<<<<<<<
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")
 */
    __pyx_v_self->data = ((char *)malloc(__pyx_v_self->len));

    /* "View.MemoryView":175
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */
    __pyx_t_4 = ((!(__pyx_v_self->data != 0)) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":176
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(1, 176, __pyx_L1_error)

      /* "View.MemoryView":175
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */
    }

    /* "View.MemoryView":178
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */
    __pyx_t_4 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":179
 * 
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data             # <<<<<<<<<<<<<<
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 */
      __pyx_v_p = ((PyObject **)__pyx_v_self->data);

      /* "View.MemoryView":180
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):             # <<<<<<<<<<<<<<
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)
 */
      if (unlikely(__pyx_v_itemsize == 0)) {
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
        __PYX_ERR(1, 180, __pyx_L1_error)
      }
      else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_self->len))) {
        PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
        __PYX_ERR(1, 180, __pyx_L1_error)
      }
      __pyx_t_1 = (__pyx_v_self->len / __pyx_v_itemsize);
      __pyx_t_9 = __pyx_t_1;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_9; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "View.MemoryView":181
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None             # <<<<<<<<<<<<<<
 *                     Py_INCREF(Py_None)
 * 
 */
        (__pyx_v_p[__pyx_v_i]) = Py_None;

        /* "View.MemoryView":182
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
        Py_INCREF(Py_None);
      }

      /* "View.MemoryView":178
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */
    }

    /* "View.MemoryView":171
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_format);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":185
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(((struct __pyx_array_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_bufmode;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":186
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1             # <<<<<<<<<<<<<<
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
  __pyx_v_bufmode = -1;

  /* "View.MemoryView":187
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_c, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 187, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":188
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
    __pyx_v_bufmode = (PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":187
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":189
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_fortran, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 189, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":190
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 */
    __pyx_v_bufmode = (PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":189
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":191
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */
  __pyx_t_1 = ((!((__pyx_v_flags & __pyx_v_bufmode) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":192
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 192, __pyx_L1_error)

    /* "View.MemoryView":191
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */
  }

  /* "View.MemoryView":193
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data             # <<<<<<<<<<<<<<
 *         info.len = self.len
 *         info.ndim = self.ndim
 */
  __pyx_t_4 = __pyx_v_self->data;
  __pyx_v_info->buf = __pyx_t_4;

  /* "View.MemoryView":194
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 *         info.len = self.len             # <<<<<<<<<<<<<<
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 */
  __pyx_t_5 = __pyx_v_self->len;
  __pyx_v_info->len = __pyx_t_5;

  /* "View.MemoryView":195
 *         info.buf = self.data
 *         info.len = self.len
 *         info.ndim = self.ndim             # <<<<<<<<<<<<<<
 *         info.shape = self._shape
 *         info.strides = self._strides
 */
  __pyx_t_6 = __pyx_v_self->ndim;
  __pyx_v_info->ndim = __pyx_t_6;

  /* "View.MemoryView":196
 *         info.len = self.len
 *         info.ndim = self.ndim
 *         info.shape = self._shape             # <<<<<<<<<<<<<<
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 */
  __pyx_t_7 = __pyx_v_self->_shape;
  __pyx_v_info->shape = __pyx_t_7;

  /* "View.MemoryView":197
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 *         info.strides = self._strides             # <<<<<<<<<<<<<<
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 */
  __pyx_t_7 = __pyx_v_self->_strides;
  __pyx_v_info->strides = __pyx_t_7;

  /* "View.MemoryView":198
 *         info.shape = self._shape
 *         info.strides = self._strides
 *         info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *         info.itemsize = self.itemsize
 *         info.readonly = 0
 */
  __pyx_v_info->suboffsets = NULL;

  /* "View.MemoryView":199
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize             # <<<<<<<<<<<<<<
 *         info.readonly = 0
 * 
 */
  __pyx_t_5 = __pyx_v_self->itemsize;
  __pyx_v_info->itemsize = __pyx_t_5;

  /* "View.MemoryView":200
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 *         info.readonly = 0             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  __pyx_v_info->readonly = 0;

  /* "View.MemoryView":202
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":203
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_4 = __pyx_v_self->format;
    __pyx_v_info->format = __pyx_t_4;

    /* "View.MemoryView":202
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
    goto __pyx_L5;
  }

  /* "View.MemoryView":205
 *             info.format = self.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.obj = self
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L5:;

  /* "View.MemoryView":207
 *             info.format = NULL
 * 
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":185
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":211
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

/* Python wrapper */
static void __pyx_array___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_array___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":212
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */
  __pyx_t_1 = ((__pyx_v_self->callback_free_data != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":213
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)             # <<<<<<<<<<<<<<
 *         elif self.free_data:
 *             if self.dtype_is_object:
 */
    __pyx_v_self->callback_free_data(__pyx_v_self->data);

    /* "View.MemoryView":212
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":214
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */
  __pyx_t_1 = (__pyx_v_self->free_data != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":215
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */
    __pyx_t_1 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":216
 *         elif self.free_data:
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,             # <<<<<<<<<<<<<<
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_self->data, __pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_self->ndim, 0);

      /* "View.MemoryView":215
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */
    }

    /* "View.MemoryView":218
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 *             free(self.data)             # <<<<<<<<<<<<<<
 *         PyObject_Free(self._shape)
 * 
 */
    free(__pyx_v_self->data);

    /* "View.MemoryView":214
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":219
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 *         PyObject_Free(self._shape)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  PyObject_Free(__pyx_v_self->_shape);

  /* "View.MemoryView":211
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":222
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_5array_7memview___get__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":223
 *     @property
 *     def memview(self):
 *         return self.get_memview()             # <<<<<<<<<<<<<<
 * 
 *     @cname('get_memview')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_array *)__pyx_v_self->__pyx_vtab)->get_memview(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":222
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.memview.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":226
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self) {
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("get_memview", 0);

  /* "View.MemoryView":227
 *     @cname('get_memview')
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE             # <<<<<<<<<<<<<<
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 */
  __pyx_v_flags = ((PyBUF_ANY_CONTIGUOUS | PyBUF_FORMAT) | PyBUF_WRITABLE);

  /* "View.MemoryView":228
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":226
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.get_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":230
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":231
 * 
 *     def __len__(self):
 *         return self._shape[0]             # <<<<<<<<<<<<<<
 * 
 *     def __getattr__(self, attr):
 */
  __pyx_r = (__pyx_v_self->_shape[0]);
  goto __pyx_L0;

  /* "View.MemoryView":230
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":233
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr); /*proto*/
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattr__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_attr));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__getattr__", 0);

  /* "View.MemoryView":234
 * 
 *     def __getattr__(self, attr):
 *         return getattr(self.memview, attr)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, item):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_attr); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":233
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":236
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":237
 * 
 *     def __getitem__(self, item):
 *         return self.memview[item]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, item, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_item); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":236
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":239
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

/* Python wrapper */
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /*proto*/
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "View.MemoryView":240
 * 
 *     def __setitem__(self, item, value):
 *         self.memview[item] = value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_v_item, __pyx_v_value) < 0)) __PYX_ERR(1, 240, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":239
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array___reduce_cython__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array_2__setstate_cython__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":244
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

static struct __pyx_array_obj *__pyx_array_new(PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, char *__pyx_v_format, char *__pyx_v_mode, char *__pyx_v_buf) {
  struct __pyx_array_obj *__pyx_v_result = 0;
  struct __pyx_array_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("array_cwrapper", 0);

  /* "View.MemoryView":248
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_buf == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":249
 * 
 *     if buf == NULL:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))             # <<<<<<<<<<<<<<
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":248
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":251
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */
  /*else*/ {
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_3 = 0;

    /* "View.MemoryView":252
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)             # <<<<<<<<<<<<<<
 *         result.data = buf
 * 
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_allocate_buffer, Py_False) < 0) __PYX_ERR(1, 252, __pyx_L1_error)

    /* "View.MemoryView":251
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "View.MemoryView":253
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)
 *         result.data = buf             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
    __pyx_v_result->data = __pyx_v_buf;
  }
  __pyx_L3:;

  /* "View.MemoryView":255
 *         result.data = buf
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":244
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.array_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":281
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

/* Python wrapper */
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 281, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_name = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 281, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.Enum.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "View.MemoryView":282
 *     cdef object name
 *     def __init__(self, name):
 *         self.name = name             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return self.name
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "View.MemoryView":281
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":283
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":284
 *         self.name = name
 *     def __repr__(self):
 *         return self.name             # <<<<<<<<<<<<<<
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* "View.MemoryView":283
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum___reduce_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->name != Py_None);
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.Enum.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum_2__setstate_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_unpickle_Enum__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.Enum.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":298
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

static void *__pyx_align_pointer(void *__pyx_v_memory, size_t __pyx_v_alignment) {
  Py_intptr_t __pyx_v_aligned_p;
  size_t __pyx_v_offset;
  void *__pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":300
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory             # <<<<<<<<<<<<<<
 *     cdef size_t offset
 * 
 */
  __pyx_v_aligned_p = ((Py_intptr_t)__pyx_v_memory);

  /* "View.MemoryView":304
 * 
 *     with cython.cdivision(True):
 *         offset = aligned_p % alignment             # <<<<<<<<<<<<<<
 * 
 *     if offset > 0:
 */
  __pyx_v_offset = (__pyx_v_aligned_p % __pyx_v_alignment);

  /* "View.MemoryView":306
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  __pyx_t_1 = ((__pyx_v_offset > 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":307
 * 
 *     if offset > 0:
 *         aligned_p += alignment - offset             # <<<<<<<<<<<<<<
 * 
 *     return <void *> aligned_p
 */
    __pyx_v_aligned_p = (__pyx_v_aligned_p + (__pyx_v_alignment - __pyx_v_offset));

    /* "View.MemoryView":306
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  }

  /* "View.MemoryView":309
 *         aligned_p += alignment - offset
 * 
 *     return <void *> aligned_p             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((void *)__pyx_v_aligned_p);
  goto __pyx_L0;

  /* "View.MemoryView":298
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":345
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

/* Python wrapper */
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  int __pyx_v_flags;
  int __pyx_v_dtype_is_object;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_flags,&__pyx_n_s_dtype_is_object,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flags)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, 1); __PYX_ERR(1, 345, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dtype_is_object);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 345, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_obj = values[0];
    __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 345, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_dtype_is_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_dtype_is_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 345, __pyx_L3_error)
    } else {
      __pyx_v_dtype_is_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 345, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_obj, __pyx_v_flags, __pyx_v_dtype_is_object);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "View.MemoryView":346
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj             # <<<<<<<<<<<<<<
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 */
  __Pyx_INCREF(__pyx_v_obj);
  __Pyx_GIVEREF(__pyx_v_obj);
  __Pyx_GOTREF(__pyx_v_self->obj);
  __Pyx_DECREF(__pyx_v_self->obj);
  __pyx_v_self->obj = __pyx_v_obj;

  /* "View.MemoryView":347
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj
 *         self.flags = flags             # <<<<<<<<<<<<<<
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 */
  __pyx_v_self->flags = __pyx_v_flags;

  /* "View.MemoryView":348
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  __pyx_t_2 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))) == ((PyObject *)__pyx_memoryview_type));
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_obj != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "View.MemoryView":349
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)             # <<<<<<<<<<<<<<
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 */
    __pyx_t_4 = __Pyx_GetBuffer(__pyx_v_obj, (&__pyx_v_self->view), __pyx_v_flags); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 349, __pyx_L1_error)

    /* "View.MemoryView":350
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    __pyx_t_1 = ((((PyObject *)__pyx_v_self->view.obj) == NULL) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":351
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None             # <<<<<<<<<<<<<<
 *                 Py_INCREF(Py_None)
 * 
 */
      ((Py_buffer *)(&__pyx_v_self->view))->obj = Py_None;

      /* "View.MemoryView":352
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         global __pyx_memoryview_thread_locks_used
 */
      Py_INCREF(Py_None);

      /* "View.MemoryView":350
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    }

    /* "View.MemoryView":348
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  }

  /* "View.MemoryView":355
 * 
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 */
  __pyx_t_1 = ((__pyx_memoryview_thread_locks_used < 8) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":356
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]             # <<<<<<<<<<<<<<
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:
 */
    __pyx_v_self->lock = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);

    /* "View.MemoryView":357
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1             # <<<<<<<<<<<<<<
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 */
    __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used + 1);

    /* "View.MemoryView":355
 * 
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 */
  }

  /* "View.MemoryView":358
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:             # <<<<<<<<<<<<<<
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 */
  __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":359
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()             # <<<<<<<<<<<<<<
 *             if self.lock is NULL:
 *                 raise MemoryError
 */
    __pyx_v_self->lock = PyThread_allocate_lock();

    /* "View.MemoryView":360
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "View.MemoryView":361
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
      PyErr_NoMemory(); __PYX_ERR(1, 361, __pyx_L1_error)

      /* "View.MemoryView":360
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    }

    /* "View.MemoryView":358
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:             # <<<<<<<<<<<<<<
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 */
  }

  /* "View.MemoryView":363
 *                 raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":364
 * 
 *         if flags & PyBUF_FORMAT:
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')             # <<<<<<<<<<<<<<
 *         else:
 *             self.dtype_is_object = dtype_is_object
 */
    __pyx_t_2 = (((__pyx_v_self->view.format[0]) == 'O') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_2 = (((__pyx_v_self->view.format[1]) == '\x00') != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L11_bool_binop_done:;
    __pyx_v_self->dtype_is_object = __pyx_t_1;

    /* "View.MemoryView":363
 *                 raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":366
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 *             self.dtype_is_object = dtype_is_object             # <<<<<<<<<<<<<<
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 */
  /*else*/ {
    __pyx_v_self->dtype_is_object = __pyx_v_dtype_is_object;
  }
  __pyx_L10:;

  /* "View.MemoryView":368
 *             self.dtype_is_object = dtype_is_object
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(             # <<<<<<<<<<<<<<
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL
 */
  __pyx_v_self->acquisition_count_aligned_p = ((__pyx_atomic_int *)__pyx_align_pointer(((void *)(&(__pyx_v_self->acquisition_count[0]))), (sizeof(__pyx_atomic_int))));

  /* "View.MemoryView":370
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(memoryview self):
 */
  __pyx_v_self->typeinfo = NULL;

  /* "View.MemoryView":345
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":372
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

/* Python wrapper */
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self) {
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyThread_type_lock __pyx_t_6;
  PyThread_type_lock __pyx_t_7;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":373
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
  __pyx_t_1 = (__pyx_v_self->obj != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":374
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)             # <<<<<<<<<<<<<<
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 */
    __Pyx_ReleaseBuffer((&__pyx_v_self->view));

    /* "View.MemoryView":373
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":375
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  __pyx_t_2 = ((((Py_buffer *)(&__pyx_v_self->view))->obj == Py_None) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":377
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL             # <<<<<<<<<<<<<<
 *             Py_DECREF(Py_None)
 * 
 */
    ((Py_buffer *)(&__pyx_v_self->view))->obj = NULL;

    /* "View.MemoryView":378
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 *             Py_DECREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         cdef int i
 */
    Py_DECREF(Py_None);

    /* "View.MemoryView":375
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":382
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  __pyx_t_2 = ((__pyx_v_self->lock != NULL) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":383
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):             # <<<<<<<<<<<<<<
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 */
    __pyx_t_3 = __pyx_memoryview_thread_locks_used;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "View.MemoryView":384
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      __pyx_t_2 = (((__pyx_memoryview_thread_locks[__pyx_v_i]) == __pyx_v_self->lock) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":385
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1             # <<<<<<<<<<<<<<
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 */
        __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used - 1);

        /* "View.MemoryView":386
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        __pyx_t_2 = ((__pyx_v_i != __pyx_memoryview_thread_locks_used) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":388
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */
          __pyx_t_6 = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);
          __pyx_t_7 = (__pyx_memoryview_thread_locks[__pyx_v_i]);

          /* "View.MemoryView":387
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (             # <<<<<<<<<<<<<<
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break
 */
          (__pyx_memoryview_thread_locks[__pyx_v_i]) = __pyx_t_6;
          (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]) = __pyx_t_7;

          /* "View.MemoryView":386
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        }

        /* "View.MemoryView":389
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 PyThread_free_lock(self.lock)
 */
        goto __pyx_L6_break;

        /* "View.MemoryView":384
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      }
    }
    /*else*/ {

      /* "View.MemoryView":391
 *                     break
 *             else:
 *                 PyThread_free_lock(self.lock)             # <<<<<<<<<<<<<<
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 */
      PyThread_free_lock(__pyx_v_self->lock);
    }
    __pyx_L6_break:;

    /* "View.MemoryView":382
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  }

  /* "View.MemoryView":372
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":393
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  Py_ssize_t __pyx_v_dim;
  char *__pyx_v_itemp;
  PyObject *__pyx_v_idx = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  char *__pyx_t_7;
  __Pyx_RefNannySetupContext("get_item_pointer", 0);

  /* "View.MemoryView":395
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf             # <<<<<<<<<<<<<<
 * 
 *         for dim, idx in enumerate(index):
 */
  __pyx_v_itemp = ((char *)__pyx_v_self->view.buf);

  /* "View.MemoryView":397
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_v_index)) || PyTuple_CheckExact(__pyx_v_index)) {
    __pyx_t_2 = __pyx_v_index; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 397, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(1, 397, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(1, 397, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 397, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_1;
    __pyx_t_1 = (__pyx_t_1 + 1);

    /* "View.MemoryView":398
 * 
 *         for dim, idx in enumerate(index):
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)             # <<<<<<<<<<<<<<
 * 
 *         return itemp
 */
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_v_idx); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 398, __pyx_L1_error)
    __pyx_t_7 = __pyx_pybuffer_index((&__pyx_v_self->view), __pyx_v_itemp, __pyx_t_6, __pyx_v_dim); if (unlikely(__pyx_t_7 == ((char *)NULL))) __PYX_ERR(1, 398, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_7;

    /* "View.MemoryView":397
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":400
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 *         return itemp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_itemp;
  goto __pyx_L0;

  /* "View.MemoryView":393
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.get_item_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":403
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index); /*proto*/
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_indices = NULL;
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  char *__pyx_t_6;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":404
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  __pyx_t_1 = (__pyx_v_index == __pyx_builtin_Ellipsis);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":405
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:
 *             return self             # <<<<<<<<<<<<<<
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __pyx_r = ((PyObject *)__pyx_v_self);
    goto __pyx_L0;

    /* "View.MemoryView":404
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  }

  /* "View.MemoryView":407
 *             return self
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         cdef char *itemp
 */
  __pyx_t_3 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(__pyx_t_3 != Py_None)) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 407, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(1, 407, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_indices = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "View.MemoryView":410
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 410, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "View.MemoryView":411
 *         cdef char *itemp
 *         if have_slices:
 *             return memview_slice(self, indices)             # <<<<<<<<<<<<<<
 *         else:
 *             itemp = self.get_item_pointer(indices)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((PyObject *)__pyx_memview_slice(__pyx_v_self, __pyx_v_indices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":410
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  }

  /* "View.MemoryView":413
 *             return memview_slice(self, indices)
 *         else:
 *             itemp = self.get_item_pointer(indices)             # <<<<<<<<<<<<<<
 *             return self.convert_item_to_object(itemp)
 * 
 */
  /*else*/ {
    __pyx_t_6 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_indices); if (unlikely(__pyx_t_6 == ((char *)NULL))) __PYX_ERR(1, 413, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_6;

    /* "View.MemoryView":414
 *         else:
 *             itemp = self.get_item_pointer(indices)
 *             return self.convert_item_to_object(itemp)             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(memoryview self, object index, object value):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->convert_item_to_object(__pyx_v_self, __pyx_v_itemp); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":403
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":416
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

/* Python wrapper */
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /*proto*/
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_obj = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("__setitem__", 0);
  __Pyx_INCREF(__pyx_v_index);

  /* "View.MemoryView":417
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */
  __pyx_t_1 = (__pyx_v_self->view.readonly != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":418
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__35, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 418, __pyx_L1_error)

    /* "View.MemoryView":417
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */
  }

  /* "View.MemoryView":420
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         if have_slices:
 */
  __pyx_t_2 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(__pyx_t_2 != Py_None)) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 420, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(1, 420, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_3;
  __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_index, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "View.MemoryView":422
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 422, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "View.MemoryView":423
 * 
 *         if have_slices:
 *             obj = self.is_slice(value)             # <<<<<<<<<<<<<<
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->is_slice(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_obj = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "View.MemoryView":424
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_obj); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 424, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "View.MemoryView":425
 *             obj = self.is_slice(value)
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)
 */
      __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assignment(__pyx_v_self, __pyx_t_2, __pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "View.MemoryView":424
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":427
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)             # <<<<<<<<<<<<<<
 *         else:
 *             self.setitem_indexed(index, value)
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_memoryview_type))))) __PYX_ERR(1, 427, __pyx_L1_error)
      __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assign_scalar(__pyx_v_self, ((struct __pyx_memoryview_obj *)__pyx_t_4), __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L5:;

    /* "View.MemoryView":422
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":429
 *                 self.setitem_slice_assign_scalar(self[index], value)
 *         else:
 *             self.setitem_indexed(index, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef is_slice(self, obj):
 */
  /*else*/ {
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_indexed(__pyx_v_self, __pyx_v_index, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "View.MemoryView":416
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":431
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("is_slice", 0);
  __Pyx_INCREF(__pyx_v_obj);

  /* "View.MemoryView":432
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_memoryview_type); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "View.MemoryView":434
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_6 = __Pyx_PyInt_From_int(((__pyx_v_self->flags & (~PyBUF_WRITABLE)) | PyBUF_ANY_CONTIGUOUS)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":435
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)             # <<<<<<<<<<<<<<
 *             except TypeError:
 *                 return None
 */
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 435, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);

        /* "View.MemoryView":434
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_v_obj);
        __Pyx_GIVEREF(__pyx_v_obj);
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_obj);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_7);
        __pyx_t_6 = 0;
        __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "View.MemoryView":436
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 *             except TypeError:             # <<<<<<<<<<<<<<
 *                 return None
 * 
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_6) < 0) __PYX_ERR(1, 436, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":437
 *                                  self.dtype_is_object)
 *             except TypeError:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *         return obj
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L7_except_return;
      }
      goto __pyx_L6_except_error;
      __pyx_L6_except_error:;

      /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L7_except_return:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L0;
      __pyx_L9_try_end:;
    }

    /* "View.MemoryView":432
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  }

  /* "View.MemoryView":439
 *                 return None
 * 
 *         return obj             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assignment(self, dst, src):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_obj);
  __pyx_r = __pyx_v_obj;
  goto __pyx_L0;

  /* "View.MemoryView":431
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":441
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src) {
  __Pyx_memviewslice __pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_src_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  __Pyx_memviewslice *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("setitem_slice_assignment", 0);

  /* "View.MemoryView":445
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */
  if (!(likely(((__pyx_v_src) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_src, __pyx_memoryview_type))))) __PYX_ERR(1, 445, __pyx_L1_error)
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_src), (&__pyx_v_src_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 445, __pyx_L1_error)

  /* "View.MemoryView":446
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],             # <<<<<<<<<<<<<<
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 */
  if (!(likely(((__pyx_v_dst) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_dst, __pyx_memoryview_type))))) __PYX_ERR(1, 446, __pyx_L1_error)
  __pyx_t_2 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_dst), (&__pyx_v_dst_slice)); if (unlikely(__pyx_t_2 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 446, __pyx_L1_error)

  /* "View.MemoryView":447
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":445
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */
  __pyx_t_6 = __pyx_memoryview_copy_contents((__pyx_t_1[0]), (__pyx_t_2[0]), __pyx_t_4, __pyx_t_5, __pyx_v_self->dtype_is_object); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 445, __pyx_L1_error)

  /* "View.MemoryView":441
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assignment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":449
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value) {
  int __pyx_v_array[0x80];
  void *__pyx_v_tmp;
  void *__pyx_v_item;
  __Pyx_memviewslice *__pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_tmp_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  char const *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("setitem_slice_assign_scalar", 0);

  /* "View.MemoryView":451
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 *         cdef int array[128]
 *         cdef void *tmp = NULL             # <<<<<<<<<<<<<<
 *         cdef void *item
 * 
 */
  __pyx_v_tmp = NULL;

  /* "View.MemoryView":456
 *         cdef __Pyx_memviewslice *dst_slice
 *         cdef __Pyx_memviewslice tmp_slice
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)             # <<<<<<<<<<<<<<
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_dst, (&__pyx_v_tmp_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 456, __pyx_L1_error)
  __pyx_v_dst_slice = __pyx_t_1;

  /* "View.MemoryView":458
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
  __pyx_t_2 = ((((size_t)__pyx_v_self->view.itemsize) > (sizeof(__pyx_v_array))) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":459
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)             # <<<<<<<<<<<<<<
 *             if tmp == NULL:
 *                 raise MemoryError
 */
    __pyx_v_tmp = PyMem_Malloc(__pyx_v_self->view.itemsize);

    /* "View.MemoryView":460
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    __pyx_t_2 = ((__pyx_v_tmp == NULL) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":461
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 *             item = tmp
 *         else:
 */
      PyErr_NoMemory(); __PYX_ERR(1, 461, __pyx_L1_error)

      /* "View.MemoryView":460
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    }

    /* "View.MemoryView":462
 *             if tmp == NULL:
 *                 raise MemoryError
 *             item = tmp             # <<<<<<<<<<<<<<
 *         else:
 *             item = <void *> array
 */
    __pyx_v_item = __pyx_v_tmp;

    /* "View.MemoryView":458
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":464
 *             item = tmp
 *         else:
 *             item = <void *> array             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  /*else*/ {
    __pyx_v_item = ((void *)__pyx_v_array);
  }
  __pyx_L3:;

  /* "View.MemoryView":466
 *             item = <void *> array
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value
 */
  /*try:*/ {

    /* "View.MemoryView":467
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
    __pyx_t_2 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":468
 *         try:
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value             # <<<<<<<<<<<<<<
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)
 */
      (((PyObject **)__pyx_v_item)[0]) = ((PyObject *)__pyx_v_value);

      /* "View.MemoryView":467
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
      goto __pyx_L8;
    }

    /* "View.MemoryView":470
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else*/ {
      __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, ((char *)__pyx_v_item), __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 470, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_L8:;

    /* "View.MemoryView":474
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    __pyx_t_2 = ((__pyx_v_self->view.suboffsets != NULL) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":475
 * 
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)             # <<<<<<<<<<<<<<
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 *                                 item, self.dtype_is_object)
 */
      __pyx_t_3 = assert_direct_dimensions(__pyx_v_self->view.suboffsets, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 475, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "View.MemoryView":474
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    }

    /* "View.MemoryView":476
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,             # <<<<<<<<<<<<<<
 *                                 item, self.dtype_is_object)
 *         finally:
 */
    __pyx_memoryview_slice_assign_scalar(__pyx_v_dst_slice, __pyx_v_dst->view.ndim, __pyx_v_self->view.itemsize, __pyx_v_item, __pyx_v_self->dtype_is_object);
  }

  /* "View.MemoryView":479
 *                                 item, self.dtype_is_object)
 *         finally:
 *             PyMem_Free(tmp)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_indexed(self, index, value):
 */
  /*finally:*/ {
    /*normal exit:*/{
      PyMem_Free(__pyx_v_tmp);
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __pyx_t_4 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_6 = __pyx_filename;
      {
        PyMem_Free(__pyx_v_tmp);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __pyx_lineno = __pyx_t_4; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_6;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "View.MemoryView":449
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assign_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":481
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("setitem_indexed", 0);

  /* "View.MemoryView":482
 * 
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)             # <<<<<<<<<<<<<<
 *         self.assign_item_from_object(itemp, value)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_index); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(1, 482, __pyx_L1_error)
  __pyx_v_itemp = __pyx_t_1;

  /* "View.MemoryView":483
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":481
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_indexed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":485
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_bytesitem = 0;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":488
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef bytes bytesitem
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":491
 *         cdef bytes bytesitem
 * 
 *         bytesitem = itemp[:self.view.itemsize]             # <<<<<<<<<<<<<<
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 */
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_itemp + 0, __pyx_v_self->view.itemsize - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_bytesitem = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "View.MemoryView":493
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)             # <<<<<<<<<<<<<<
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_unpack); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 493, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 493, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 493, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 493, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 493, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_7) {
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_6);
        __Pyx_INCREF(__pyx_v_bytesitem);
        __Pyx_GIVEREF(__pyx_v_bytesitem);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_bytesitem);
        __pyx_t_6 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 493, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_result = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    }

    /* "View.MemoryView":497
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
    /*else:*/ {
      __pyx_t_10 = strlen(__pyx_v_self->view.format); 
      __pyx_t_11 = ((__pyx_t_10 == 1) != 0);
      if (__pyx_t_11) {

        /* "View.MemoryView":498
 *         else:
 *             if len(self.view.format) == 1:
 *                 return result[0]             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_result, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 498, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L6_except_return;

        /* "View.MemoryView":497
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
      }

      /* "View.MemoryView":499
 *             if len(self.view.format) == 1:
 *                 return result[0]
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_result);
      __pyx_r = __pyx_v_result;
      goto __pyx_L6_except_return;
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "View.MemoryView":494
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unable to convert item to object")
 *         else:
 */
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_5, &__pyx_t_9);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_error); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 494, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_5, __pyx_t_9);
    __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_9 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(1, 494, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_1);

      /* "View.MemoryView":495
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__36, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 495, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(1, 495, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
  }

  /* "View.MemoryView":485
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesitem);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":501
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_struct = NULL;
  char __pyx_v_c;
  PyObject *__pyx_v_bytesvalue = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  char *__pyx_t_11;
  char *__pyx_t_12;
  char *__pyx_t_13;
  char *__pyx_t_14;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":504
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef char c
 *         cdef bytes bytesvalue
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":509
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
  __pyx_t_2 = PyTuple_Check(__pyx_v_value); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "View.MemoryView":510
 * 
 *         if isinstance(value, tuple):
 *             bytesvalue = struct.pack(self.view.format, *value)             # <<<<<<<<<<<<<<
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyNumber_Add(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(1, 510, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":509
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":512
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)             # <<<<<<<<<<<<<<
 * 
 *         for i, c in enumerate(bytesvalue):
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 512, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 512, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 512, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_1);
      __Pyx_INCREF(__pyx_v_value);
      __Pyx_GIVEREF(__pyx_v_value);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_value);
      __pyx_t_1 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 512, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(1, 512, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
  __pyx_t_9 = 0;
  if (unlikely(__pyx_v_bytesvalue == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");
    __PYX_ERR(1, 514, __pyx_L1_error)
  }
  __Pyx_INCREF(__pyx_v_bytesvalue);
  __pyx_t_10 = __pyx_v_bytesvalue;
  __pyx_t_12 = PyBytes_AS_STRING(__pyx_t_10);
  __pyx_t_13 = (__pyx_t_12 + PyBytes_GET_SIZE(__pyx_t_10));
  for (__pyx_t_14 = __pyx_t_12; __pyx_t_14 < __pyx_t_13; __pyx_t_14++) {
    __pyx_t_11 = __pyx_t_14;
    __pyx_v_c = (__pyx_t_11[0]);

    /* "View.MemoryView":515
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    __pyx_v_i = __pyx_t_9;

    /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
    __pyx_t_9 = (__pyx_t_9 + 1);

    /* "View.MemoryView":515
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    (__pyx_v_itemp[__pyx_v_i]) = __pyx_v_c;
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "View.MemoryView":501
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("View.MemoryView.memoryview.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":518
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  char *__pyx_t_5;
  void *__pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":519
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */
  __pyx_t_2 = ((__pyx_v_flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->view.readonly != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":520
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__37, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 520, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 520, __pyx_L1_error)

    /* "View.MemoryView":519
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */
  }

  /* "View.MemoryView":522
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_ND) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":523
 * 
 *         if flags & PyBUF_ND:
 *             info.shape = self.view.shape             # <<<<<<<<<<<<<<
 *         else:
 *             info.shape = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.shape;
    __pyx_v_info->shape = __pyx_t_4;

    /* "View.MemoryView":522
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
    goto __pyx_L6;
  }

  /* "View.MemoryView":525
 *             info.shape = self.view.shape
 *         else:
 *             info.shape = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_STRIDES:
 */
  /*else*/ {
    __pyx_v_info->shape = NULL;
  }
  __pyx_L6:;

  /* "View.MemoryView":527
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_STRIDES) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":528
 * 
 *         if flags & PyBUF_STRIDES:
 *             info.strides = self.view.strides             # <<<<<<<<<<<<<<
 *         else:
 *             info.strides = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.strides;
    __pyx_v_info->strides = __pyx_t_4;

    /* "View.MemoryView":527
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
    goto __pyx_L7;
  }

  /* "View.MemoryView":530
 *             info.strides = self.view.strides
 *         else:
 *             info.strides = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_INDIRECT:
 */
  /*else*/ {
    __pyx_v_info->strides = NULL;
  }
  __pyx_L7:;

  /* "View.MemoryView":532
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_INDIRECT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":533
 * 
 *         if flags & PyBUF_INDIRECT:
 *             info.suboffsets = self.view.suboffsets             # <<<<<<<<<<<<<<
 *         else:
 *             info.suboffsets = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.suboffsets;
    __pyx_v_info->suboffsets = __pyx_t_4;

    /* "View.MemoryView":532
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
    goto __pyx_L8;
  }

  /* "View.MemoryView":535
 *             info.suboffsets = self.view.suboffsets
 *         else:
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  /*else*/ {
    __pyx_v_info->suboffsets = NULL;
  }
  __pyx_L8:;

  /* "View.MemoryView":537
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":538
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.view.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_5 = __pyx_v_self->view.format;
    __pyx_v_info->format = __pyx_t_5;

    /* "View.MemoryView":537
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":540
 *             info.format = self.view.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.buf = self.view.buf
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L9:;

  /* "View.MemoryView":542
 *             info.format = NULL
 * 
 *         info.buf = self.view.buf             # <<<<<<<<<<<<<<
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 */
  __pyx_t_6 = __pyx_v_self->view.buf;
  __pyx_v_info->buf = __pyx_t_6;

  /* "View.MemoryView":543
 * 
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim             # <<<<<<<<<<<<<<
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 */
  __pyx_t_7 = __pyx_v_self->view.ndim;
  __pyx_v_info->ndim = __pyx_t_7;

  /* "View.MemoryView":544
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize             # <<<<<<<<<<<<<<
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 */
  __pyx_t_8 = __pyx_v_self->view.itemsize;
  __pyx_v_info->itemsize = __pyx_t_8;

  /* "View.MemoryView":545
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len             # <<<<<<<<<<<<<<
 *         info.readonly = self.view.readonly
 *         info.obj = self
 */
  __pyx_t_8 = __pyx_v_self->view.len;
  __pyx_v_info->len = __pyx_t_8;

  /* "View.MemoryView":546
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly             # <<<<<<<<<<<<<<
 *         info.obj = self
 * 
 */
  __pyx_t_1 = __pyx_v_self->view.readonly;
  __pyx_v_info->readonly = __pyx_t_1;

  /* "View.MemoryView":547
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":518
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":553
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":554
 *     @property
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&result.from_slice)
 *         return result
 */
  __pyx_t_1 = __pyx_memoryview_copy_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_memoryviewslice_type))))) __PYX_ERR(1, 554, __pyx_L1_error)
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":555
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  __pyx_t_2 = __pyx_memslice_transpose((&__pyx_v_result->from_slice)); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(1, 555, __pyx_L1_error)

  /* "View.MemoryView":556
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":553
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.T.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":559
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":560
 *     @property
 *     def base(self):
 *         return self.obj             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->obj);
  __pyx_r = __pyx_v_self->obj;
  goto __pyx_L0;

  /* "View.MemoryView":559
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":563
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":564
 *     @property
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
  for (__pyx_t_4 = __pyx_v_self->view.shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_length = (__pyx_t_2[0]);
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(1, 564, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = PyList_AsTuple(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":563
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.shape.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":567
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_stride;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":568
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */
  __pyx_t_1 = ((__pyx_v_self->view.strides == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":570
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__38, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 570, __pyx_L1_error)

    /* "View.MemoryView":568
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */
  }

  /* "View.MemoryView":572
 *             raise ValueError("Buffer view does not expose strides")
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = (__pyx_v_self->view.strides + __pyx_v_self->view.ndim);
  for (__pyx_t_5 = __pyx_v_self->view.strides; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
    __pyx_t_3 = __pyx_t_5;
    __pyx_v_stride = (__pyx_t_3[0]);
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_stride); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(1, 572, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_t_6 = PyList_AsTuple(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":567
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.strides.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":575
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":576
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->view.suboffsets == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":577
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Multiply(__pyx_tuple__39, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":576
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  }

  /* "View.MemoryView":579
 *             return (-1,) * self.view.ndim
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = (__pyx_v_self->view.suboffsets + __pyx_v_self->view.ndim);
  for (__pyx_t_6 = __pyx_v_self->view.suboffsets; __pyx_t_6 < __pyx_t_5; __pyx_t_6++) {
    __pyx_t_4 = __pyx_t_6;
    __pyx_v_suboffset = (__pyx_t_4[0]);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_suboffset); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(1, 579, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = PyList_AsTuple(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":575
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.suboffsets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":582
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":583
 *     @property
 *     def ndim(self):
 *         return self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":582
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.ndim.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":586
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":587
 *     @property
 *     def itemsize(self):
 *         return self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":586
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.itemsize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":590
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":591
 *     @property
 *     def nbytes(self):
 *         return self.size * self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":590
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.nbytes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":594
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":595
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_size == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":596
 *     def size(self):
 *         if self._size is None:
 *             result = 1             # <<<<<<<<<<<<<<
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 */
    __Pyx_INCREF(__pyx_int_1);
    __pyx_v_result = __pyx_int_1;

    /* "View.MemoryView":598
 *             result = 1
 * 
 *             for length in self.view.shape[:self.view.ndim]:             # <<<<<<<<<<<<<<
 *                 result *= length
 * 
 */
    __pyx_t_4 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
    for (__pyx_t_5 = __pyx_v_self->view.shape; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
      __pyx_t_3 = __pyx_t_5;
      __pyx_t_6 = PyInt_FromSsize_t((__pyx_t_3[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "View.MemoryView":599
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 *                 result *= length             # <<<<<<<<<<<<<<
 * 
 *             self._size = result
 */
      __pyx_t_6 = PyNumber_InPlaceMultiply(__pyx_v_result, __pyx_v_length); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 599, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_6);
      __pyx_t_6 = 0;
    }

    /* "View.MemoryView":601
 *                 result *= length
 * 
 *             self._size = result             # <<<<<<<<<<<<<<
 * 
 *         return self._size
 */
    __Pyx_INCREF(__pyx_v_result);
    __Pyx_GIVEREF(__pyx_v_result);
    __Pyx_GOTREF(__pyx_v_self->_size);
    __Pyx_DECREF(__pyx_v_self->_size);
    __pyx_v_self->_size = __pyx_v_result;

    /* "View.MemoryView":595
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  }

  /* "View.MemoryView":603
 *             self._size = result
 * 
 *         return self._size             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_size);
  __pyx_r = __pyx_v_self->_size;
  goto __pyx_L0;

  /* "View.MemoryView":594
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":605
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

/* Python wrapper */
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":606
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->view.ndim >= 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":607
 *     def __len__(self):
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         return 0
 */
    __pyx_r = (__pyx_v_self->view.shape[0]);
    goto __pyx_L0;

    /* "View.MemoryView":606
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  }

  /* "View.MemoryView":609
 *             return self.view.shape[0]
 * 
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":605
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":611
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":612
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":613
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "View.MemoryView":612
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":611
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":615
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "View.MemoryView":616
 * 
 *     def __str__(self):
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_object, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":615
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":619
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_c_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("is_c_contig", 0);

  /* "View.MemoryView":622
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 622, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":623
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def is_f_contig(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'C', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":619
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_c_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":625
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_f_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("is_f_contig", 0);

  /* "View.MemoryView":628
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 628, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":629
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'F', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":625
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_f_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":631
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_mslice;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "View.MemoryView":633
 *     def copy(self):
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &mslice)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_F_CONTIGUOUS));

  /* "View.MemoryView":635
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 * 
 *         slice_copy(self, &mslice)             # <<<<<<<<<<<<<<
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,
 *                                    self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_mslice));

  /* "View.MemoryView":636
 * 
 *         slice_copy(self, &mslice)
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                    self.view.itemsize,
 *                                    flags|PyBUF_C_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_mslice), ((char *)"c"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_C_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 636, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":641
 *                                    self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &mslice)             # <<<<<<<<<<<<<<
 * 
 *     def copy_fortran(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_mslice)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":631
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":643
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy_fortran (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("copy_fortran", 0);

  /* "View.MemoryView":645
 *     def copy_fortran(self):
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &src)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_C_CONTIGUOUS));

  /* "View.MemoryView":647
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 * 
 *         slice_copy(self, &src)             # <<<<<<<<<<<<<<
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,
 *                                 self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_src));

  /* "View.MemoryView":648
 * 
 *         slice_copy(self, &src)
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                 self.view.itemsize,
 *                                 flags|PyBUF_F_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_src), ((char *)"fortran"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_F_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 648, __pyx_L1_error)
  __pyx_v_dst = __pyx_t_1;

  /* "View.MemoryView":653
 *                                 self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &dst)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_dst)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":643
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy_fortran", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview___reduce_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__40, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview_2__setstate_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__41, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":657
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

static PyObject *__pyx_memoryview_new(PyObject *__pyx_v_o, int __pyx_v_flags, int __pyx_v_dtype_is_object, __Pyx_TypeInfo *__pyx_v_typeinfo) {
  struct __pyx_memoryview_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("memoryview_cwrapper", 0);

  /* "View.MemoryView":658
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)             # <<<<<<<<<<<<<<
 *     result.typeinfo = typeinfo
 *     return result
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_o);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryview_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":659
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_v_result->typeinfo = __pyx_v_typeinfo;

  /* "View.MemoryView":660
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_check')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":657
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":663
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

static CYTHON_INLINE int __pyx_memoryview_check(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("memoryview_check", 0);

  /* "View.MemoryView":664
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):
 *     return isinstance(o, memoryview)             # <<<<<<<<<<<<<<
 * 
 * cdef tuple _unellipsify(object index, int ndim):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_memoryview_type); 
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "View.MemoryView":663
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":666
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

static PyObject *_unellipsify(PyObject *__pyx_v_index, int __pyx_v_ndim) {
  PyObject *__pyx_v_tup = NULL;
  PyObject *__pyx_v_result = NULL;
  int __pyx_v_have_slices;
  int __pyx_v_seen_ellipsis;
  CYTHON_UNUSED PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_item = NULL;
  Py_ssize_t __pyx_v_nslices;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  __Pyx_RefNannySetupContext("_unellipsify", 0);

  /* "View.MemoryView":671
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */
  __pyx_t_1 = PyTuple_Check(__pyx_v_index); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":672
 *     """
 *     if not isinstance(index, tuple):
 *         tup = (index,)             # <<<<<<<<<<<<<<
 *     else:
 *         tup = index
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 672, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_index);
    __Pyx_GIVEREF(__pyx_v_index);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_index);
    __pyx_v_tup = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "View.MemoryView":671
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":674
 *         tup = (index,)
 *     else:
 *         tup = index             # <<<<<<<<<<<<<<
 * 
 *     result = []
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_index);
    __pyx_v_tup = __pyx_v_index;
  }
  __pyx_L3:;

  /* "View.MemoryView":676
 *         tup = index
 * 
 *     result = []             # <<<<<<<<<<<<<<
 *     have_slices = False
 *     seen_ellipsis = False
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_result = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":677
 * 
 *     result = []
 *     have_slices = False             # <<<<<<<<<<<<<<
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 */
  __pyx_v_have_slices = 0;

  /* "View.MemoryView":678
 *     result = []
 *     have_slices = False
 *     seen_ellipsis = False             # <<<<<<<<<<<<<<
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 */
  __pyx_v_seen_ellipsis = 0;

  /* "View.MemoryView":679
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_3 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_tup)) || PyTuple_CheckExact(__pyx_v_tup)) {
    __pyx_t_4 = __pyx_v_tup; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_tup); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 679, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(1, 679, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(1, 679, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_6(__pyx_t_4);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 679, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_3);
    __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "View.MemoryView":680
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */
    __pyx_t_2 = (__pyx_v_item == __pyx_builtin_Ellipsis);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":681
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */
      __pyx_t_1 = ((!(__pyx_v_seen_ellipsis != 0)) != 0);
      if (__pyx_t_1) {

        /* "View.MemoryView":682
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */
        __pyx_t_8 = PyObject_Length(__pyx_v_tup); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(1, 682, __pyx_L1_error)
        __pyx_t_7 = PyList_New(1 * ((((__pyx_v_ndim - __pyx_t_8) + 1)<0) ? 0:((__pyx_v_ndim - __pyx_t_8) + 1))); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 682, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        { Py_ssize_t __pyx_temp;
          for (__pyx_temp=0; __pyx_temp < ((__pyx_v_ndim - __pyx_t_8) + 1); __pyx_temp++) {
            __Pyx_INCREF(__pyx_slice__42);
            __Pyx_GIVEREF(__pyx_slice__42);
            PyList_SET_ITEM(__pyx_t_7, __pyx_temp, __pyx_slice__42);
          }
        }
        __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_7); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 682, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "View.MemoryView":683
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True             # <<<<<<<<<<<<<<
 *             else:
 *                 result.append(slice(None))
 */
        __pyx_v_seen_ellipsis = 1;

        /* "View.MemoryView":681
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":685
 *                 seen_ellipsis = True
 *             else:
 *                 result.append(slice(None))             # <<<<<<<<<<<<<<
 *             have_slices = True
 *         else:
 */
      /*else*/ {
        __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_slice__42); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 685, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":686
 *             else:
 *                 result.append(slice(None))
 *             have_slices = True             # <<<<<<<<<<<<<<
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 */
      __pyx_v_have_slices = 1;

      /* "View.MemoryView":680
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":688
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */
    /*else*/ {
      __pyx_t_2 = PySlice_Check(__pyx_v_item); 
      __pyx_t_10 = ((!(__pyx_t_2 != 0)) != 0);
      if (__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = ((!(PyIndex_Check(__pyx_v_item) != 0)) != 0);
      __pyx_t_1 = __pyx_t_10;
      __pyx_L9_bool_binop_done:;
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":689
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 *                 raise TypeError("Cannot index with type '%s'" % type(item))             # <<<<<<<<<<<<<<
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 */
        __pyx_t_7 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_index_with_type_s, ((PyObject *)Py_TYPE(__pyx_v_item))); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 689, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 689, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_t_11, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __PYX_ERR(1, 689, __pyx_L1_error)

        /* "View.MemoryView":688
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */
      }

      /* "View.MemoryView":691
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 *             have_slices = have_slices or isinstance(item, slice)             # <<<<<<<<<<<<<<
 *             result.append(item)
 * 
 */
      __pyx_t_10 = (__pyx_v_have_slices != 0);
      if (!__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = PySlice_Check(__pyx_v_item); 
      __pyx_t_2 = (__pyx_t_10 != 0);
      __pyx_t_1 = __pyx_t_2;
      __pyx_L11_bool_binop_done:;
      __pyx_v_have_slices = __pyx_t_1;

      /* "View.MemoryView":692
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 *             result.append(item)             # <<<<<<<<<<<<<<
 * 
 *     nslices = ndim - len(result)
 */
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_v_item); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 692, __pyx_L1_error)
    }
    __pyx_L6:;

    /* "View.MemoryView":679
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":694
 *             result.append(item)
 * 
 *     nslices = ndim - len(result)             # <<<<<<<<<<<<<<
 *     if nslices:
 *         result.extend([slice(None)] * nslices)
 */
  __pyx_t_5 = PyList_GET_SIZE(__pyx_v_result); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 694, __pyx_L1_error)
  __pyx_v_nslices = (__pyx_v_ndim - __pyx_t_5);

  /* "View.MemoryView":695
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */
  __pyx_t_1 = (__pyx_v_nslices != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":696
 *     nslices = ndim - len(result)
 *     if nslices:
 *         result.extend([slice(None)] * nslices)             # <<<<<<<<<<<<<<
 * 
 *     return have_slices or nslices, tuple(result)
 */
    __pyx_t_3 = PyList_New(1 * ((__pyx_v_nslices<0) ? 0:__pyx_v_nslices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    { Py_ssize_t __pyx_temp;
      for (__pyx_temp=0; __pyx_temp < __pyx_v_nslices; __pyx_temp++) {
        __Pyx_INCREF(__pyx_slice__42);
        __Pyx_GIVEREF(__pyx_slice__42);
        PyList_SET_ITEM(__pyx_t_3, __pyx_temp, __pyx_slice__42);
      }
    }
    __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_3); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 696, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "View.MemoryView":695
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */
  }

  /* "View.MemoryView":698
 *         result.extend([slice(None)] * nslices)
 * 
 *     return have_slices or nslices, tuple(result)             # <<<<<<<<<<<<<<
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 */
  __Pyx_XDECREF(__pyx_r);
  if (!__pyx_v_have_slices) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_have_slices); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L14_bool_binop_done;
  }
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_nslices); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_L14_bool_binop_done:;
  __pyx_t_4 = PyList_AsTuple(__pyx_v_result); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = ((PyObject*)__pyx_t_11);
  __pyx_t_11 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":666
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("View.MemoryView._unellipsify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tup);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":700
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

static PyObject *assert_direct_dimensions(Py_ssize_t *__pyx_v_suboffsets, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("assert_direct_dimensions", 0);

  /* "View.MemoryView":701
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")
 */
  __pyx_t_2 = (__pyx_v_suboffsets + __pyx_v_ndim);
  for (__pyx_t_3 = __pyx_v_suboffsets; __pyx_t_3 < __pyx_t_2; __pyx_t_3++) {
    __pyx_t_1 = __pyx_t_3;
    __pyx_v_suboffset = (__pyx_t_1[0]);

    /* "View.MemoryView":702
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */
    __pyx_t_4 = ((__pyx_v_suboffset >= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":703
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__43, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(1, 703, __pyx_L1_error)

      /* "View.MemoryView":702
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */
    }
  }

  /* "View.MemoryView":700
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.assert_direct_dimensions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":710
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *__pyx_v_memview, PyObject *__pyx_v_indices) {
  int __pyx_v_new_ndim;
  int __pyx_v_suboffset_dim;
  int __pyx_v_dim;
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  __Pyx_memviewslice *__pyx_v_p_src;
  struct __pyx_memoryviewslice_obj *__pyx_v_memviewsliceobj = 0;
  __Pyx_memviewslice *__pyx_v_p_dst;
  int *__pyx_v_p_suboffset_dim;
  Py_ssize_t __pyx_v_start;
  Py_ssize_t __pyx_v_stop;
  Py_ssize_t __pyx_v_step;
  int __pyx_v_have_start;
  int __pyx_v_have_stop;
  int __pyx_v_have_step;
  PyObject *__pyx_v_index = NULL;
  struct __pyx_memoryview_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_memoryview_obj *__pyx_t_4;
  char *__pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  __Pyx_RefNannySetupContext("memview_slice", 0);

  /* "View.MemoryView":711
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim             # <<<<<<<<<<<<<<
 *     cdef bint negative_step
 *     cdef __Pyx_memviewslice src, dst
 */
  __pyx_v_new_ndim = 0;
  __pyx_v_suboffset_dim = -1;

  /* "View.MemoryView":718
 * 
 * 
 *     memset(&dst, 0, sizeof(dst))             # <<<<<<<<<<<<<<
 * 
 *     cdef _memoryviewslice memviewsliceobj
 */
  (void)(memset((&__pyx_v_dst), 0, (sizeof(__pyx_v_dst))));

  /* "View.MemoryView":722
 *     cdef _memoryviewslice memviewsliceobj
 * 
 *     assert memview.view.ndim > 0             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((__pyx_v_memview->view.ndim > 0) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(1, 722, __pyx_L1_error)
    }
  }
  #endif

  /* "View.MemoryView":724
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":725
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview             # <<<<<<<<<<<<<<
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(1, 725, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_memviewsliceobj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":726
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, &src)
 */
    __pyx_v_p_src = (&__pyx_v_memviewsliceobj->from_slice);

    /* "View.MemoryView":724
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":728
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 *         slice_copy(memview, &src)             # <<<<<<<<<<<<<<
 *         p_src = &src
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_src));

    /* "View.MemoryView":729
 *     else:
 *         slice_copy(memview, &src)
 *         p_src = &src             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_p_src = (&__pyx_v_src);
  }
  __pyx_L3:;

  /* "View.MemoryView":735
 * 
 * 
 *     dst.memview = p_src.memview             # <<<<<<<<<<<<<<
 *     dst.data = p_src.data
 * 
 */
  __pyx_t_4 = __pyx_v_p_src->memview;
  __pyx_v_dst.memview = __pyx_t_4;

  /* "View.MemoryView":736
 * 
 *     dst.memview = p_src.memview
 *     dst.data = p_src.data             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __pyx_v_p_src->data;
  __pyx_v_dst.data = __pyx_t_5;

  /* "View.MemoryView":741
 * 
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst             # <<<<<<<<<<<<<<
 *     cdef int *p_suboffset_dim = &suboffset_dim
 *     cdef Py_ssize_t start, stop, step
 */
  __pyx_v_p_dst = (&__pyx_v_dst);

  /* "View.MemoryView":742
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst
 *     cdef int *p_suboffset_dim = &suboffset_dim             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t start, stop, step
 *     cdef bint have_start, have_stop, have_step
 */
  __pyx_v_p_suboffset_dim = (&__pyx_v_suboffset_dim);

  /* "View.MemoryView":746
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */
  __pyx_t_6 = 0;
  if (likely(PyList_CheckExact(__pyx_v_indices)) || PyTuple_CheckExact(__pyx_v_indices)) {
    __pyx_t_3 = __pyx_v_indices; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_indices); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 746, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 746, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(1, 746, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(1, 746, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      }
    } else {
      __pyx_t_9 = __pyx_t_8(__pyx_t_3);
      if (unlikely(!__pyx_t_9)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 746, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_9);
    }
    __Pyx_XDECREF_SET(__pyx_v_index, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_v_dim = __pyx_t_6;
    __pyx_t_6 = (__pyx_t_6 + 1);

    /* "View.MemoryView":747
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
    __pyx_t_2 = (PyIndex_Check(__pyx_v_index) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":751
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 *                 index, 0, 0, # start, stop, step             # <<<<<<<<<<<<<<
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 */
      __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_index); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 751, __pyx_L1_error)

      /* "View.MemoryView":748
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_t_10, 0, 0, 0, 0, 0, 0); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(1, 748, __pyx_L1_error)

      /* "View.MemoryView":747
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":754
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
    __pyx_t_2 = (__pyx_v_index == Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":755
 *                 False)
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1             # <<<<<<<<<<<<<<
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 */
      (__pyx_v_p_dst->shape[__pyx_v_new_ndim]) = 1;

      /* "View.MemoryView":756
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0             # <<<<<<<<<<<<<<
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1
 */
      (__pyx_v_p_dst->strides[__pyx_v_new_ndim]) = 0;

      /* "View.MemoryView":757
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1             # <<<<<<<<<<<<<<
 *             new_ndim += 1
 *         else:
 */
      (__pyx_v_p_dst->suboffsets[__pyx_v_new_ndim]) = -1L;

      /* "View.MemoryView":758
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 *         else:
 *             start = index.start or 0
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);

      /* "View.MemoryView":754
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":760
 *             new_ndim += 1
 *         else:
 *             start = index.start or 0             # <<<<<<<<<<<<<<
 *             stop = index.stop or 0
 *             step = index.step or 0
 */
    /*else*/ {
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 760, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 760, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 760, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L7_bool_binop_done:;
      __pyx_v_start = __pyx_t_10;

      /* "View.MemoryView":761
 *         else:
 *             start = index.start or 0
 *             stop = index.stop or 0             # <<<<<<<<<<<<<<
 *             step = index.step or 0
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 761, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 761, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 761, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L9_bool_binop_done:;
      __pyx_v_stop = __pyx_t_10;

      /* "View.MemoryView":762
 *             start = index.start or 0
 *             stop = index.stop or 0
 *             step = index.step or 0             # <<<<<<<<<<<<<<
 * 
 *             have_start = index.start is not None
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 762, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 762, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 762, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L11_bool_binop_done:;
      __pyx_v_step = __pyx_t_10;

      /* "View.MemoryView":764
 *             step = index.step or 0
 * 
 *             have_start = index.start is not None             # <<<<<<<<<<<<<<
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 764, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_start = __pyx_t_1;

      /* "View.MemoryView":765
 * 
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None             # <<<<<<<<<<<<<<
 *             have_step = index.step is not None
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 765, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_stop = __pyx_t_1;

      /* "View.MemoryView":766
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None             # <<<<<<<<<<<<<<
 * 
 *             slice_memviewslice(
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_step = __pyx_t_1;

      /* "View.MemoryView":768
 *             have_step = index.step is not None
 * 
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_v_start, __pyx_v_stop, __pyx_v_step, __pyx_v_have_start, __pyx_v_have_stop, __pyx_v_have_step, 1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(1, 768, __pyx_L1_error)

      /* "View.MemoryView":774
 *                 have_start, have_stop, have_step,
 *                 True)
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);
    }
    __pyx_L6:;

    /* "View.MemoryView":746
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":776
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":777
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":778
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_dtype_func,
 *                                     memview.dtype_is_object)
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(1, 778, __pyx_L1_error) }

    /* "View.MemoryView":779
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 *     else:
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(1, 779, __pyx_L1_error) }

    /* "View.MemoryView":777
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, __pyx_v_memviewsliceobj->to_object_func, __pyx_v_memviewsliceobj->to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(1, 777, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":776
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  }

  /* "View.MemoryView":782
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":783
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,
 *                                     memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, NULL, NULL, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 782, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "View.MemoryView":782
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(1, 782, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":710
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memview_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_memviewsliceobj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":807
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *__pyx_v_dst, Py_ssize_t __pyx_v_shape, Py_ssize_t __pyx_v_stride, Py_ssize_t __pyx_v_suboffset, int __pyx_v_dim, int __pyx_v_new_ndim, int *__pyx_v_suboffset_dim, Py_ssize_t __pyx_v_start, Py_ssize_t __pyx_v_stop, Py_ssize_t __pyx_v_step, int __pyx_v_have_start, int __pyx_v_have_stop, int __pyx_v_have_step, int __pyx_v_is_slice) {
  Py_ssize_t __pyx_v_new_shape;
  int __pyx_v_negative_step;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "View.MemoryView":827
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
  __pyx_t_1 = ((!(__pyx_v_is_slice != 0)) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":829
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    __pyx_t_1 = ((__pyx_v_start < 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":830
 * 
 *         if start < 0:
 *             start += shape             # <<<<<<<<<<<<<<
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 */
      __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

      /* "View.MemoryView":829
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    }

    /* "View.MemoryView":831
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    __pyx_t_1 = (0 <= __pyx_v_start);
    if (__pyx_t_1) {
      __pyx_t_1 = (__pyx_v_start < __pyx_v_shape);
    }
    __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":832
 *             start += shape
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)             # <<<<<<<<<<<<<<
 *     else:
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"Index out of bounds (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 832, __pyx_L1_error)

      /* "View.MemoryView":831
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    }

    /* "View.MemoryView":827
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":835
 *     else:
 * 
 *         negative_step = have_step != 0 and step < 0             # <<<<<<<<<<<<<<
 * 
 *         if have_step and step == 0:
 */
  /*else*/ {
    __pyx_t_1 = ((__pyx_v_have_step != 0) != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_step < 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L6_bool_binop_done:;
    __pyx_v_negative_step = __pyx_t_2;

    /* "View.MemoryView":837
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */
    __pyx_t_1 = (__pyx_v_have_step != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_step == 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_2) {

      /* "View.MemoryView":838
 * 
 *         if have_step and step == 0:
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Step may not be zero (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 838, __pyx_L1_error)

      /* "View.MemoryView":837
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */
    }

    /* "View.MemoryView":841
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
    __pyx_t_2 = (__pyx_v_have_start != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":842
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
      __pyx_t_2 = ((__pyx_v_start < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":843
 *         if have_start:
 *             if start < 0:
 *                 start += shape             # <<<<<<<<<<<<<<
 *                 if start < 0:
 *                     start = 0
 */
        __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

        /* "View.MemoryView":844
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        __pyx_t_2 = ((__pyx_v_start < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":845
 *                 start += shape
 *                 if start < 0:
 *                     start = 0             # <<<<<<<<<<<<<<
 *             elif start >= shape:
 *                 if negative_step:
 */
          __pyx_v_start = 0;

          /* "View.MemoryView":844
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        }

        /* "View.MemoryView":842
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
        goto __pyx_L12;
      }

      /* "View.MemoryView":846
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      __pyx_t_2 = ((__pyx_v_start >= __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":847
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
        __pyx_t_2 = (__pyx_v_negative_step != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":848
 *             elif start >= shape:
 *                 if negative_step:
 *                     start = shape - 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     start = shape
 */
          __pyx_v_start = (__pyx_v_shape - 1);

          /* "View.MemoryView":847
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
          goto __pyx_L14;
        }

        /* "View.MemoryView":850
 *                     start = shape - 1
 *                 else:
 *                     start = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        /*else*/ {
          __pyx_v_start = __pyx_v_shape;
        }
        __pyx_L14:;

        /* "View.MemoryView":846
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      }
      __pyx_L12:;

      /* "View.MemoryView":841
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
      goto __pyx_L11;
    }

    /* "View.MemoryView":852
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":853
 *         else:
 *             if negative_step:
 *                 start = shape - 1             # <<<<<<<<<<<<<<
 *             else:
 *                 start = 0
 */
        __pyx_v_start = (__pyx_v_shape - 1);

        /* "View.MemoryView":852
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
        goto __pyx_L15;
      }

      /* "View.MemoryView":855
 *                 start = shape - 1
 *             else:
 *                 start = 0             # <<<<<<<<<<<<<<
 * 
 *         if have_stop:
 */
      /*else*/ {
        __pyx_v_start = 0;
      }
      __pyx_L15:;
    }
    __pyx_L11:;

    /* "View.MemoryView":857
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
    __pyx_t_2 = (__pyx_v_have_stop != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":858
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
      __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":859
 *         if have_stop:
 *             if stop < 0:
 *                 stop += shape             # <<<<<<<<<<<<<<
 *                 if stop < 0:
 *                     stop = 0
 */
        __pyx_v_stop = (__pyx_v_stop + __pyx_v_shape);

        /* "View.MemoryView":860
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":861
 *                 stop += shape
 *                 if stop < 0:
 *                     stop = 0             # <<<<<<<<<<<<<<
 *             elif stop > shape:
 *                 stop = shape
 */
          __pyx_v_stop = 0;

          /* "View.MemoryView":860
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        }

        /* "View.MemoryView":858
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
        goto __pyx_L17;
      }

      /* "View.MemoryView":862
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      __pyx_t_2 = ((__pyx_v_stop > __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":863
 *                     stop = 0
 *             elif stop > shape:
 *                 stop = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        __pyx_v_stop = __pyx_v_shape;

        /* "View.MemoryView":862
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      }
      __pyx_L17:;

      /* "View.MemoryView":857
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
      goto __pyx_L16;
    }

    /* "View.MemoryView":865
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":866
 *         else:
 *             if negative_step:
 *                 stop = -1             # <<<<<<<<<<<<<<
 *             else:
 *                 stop = shape
 */
        __pyx_v_stop = -1L;

        /* "View.MemoryView":865
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
        goto __pyx_L19;
      }

      /* "View.MemoryView":868
 *                 stop = -1
 *             else:
 *                 stop = shape             # <<<<<<<<<<<<<<
 * 
 *         if not have_step:
 */
      /*else*/ {
        __pyx_v_stop = __pyx_v_shape;
      }
      __pyx_L19:;
    }
    __pyx_L16:;

    /* "View.MemoryView":870
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    __pyx_t_2 = ((!(__pyx_v_have_step != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":871
 * 
 *         if not have_step:
 *             step = 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_step = 1;

      /* "View.MemoryView":870
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    }

    /* "View.MemoryView":875
 * 
 *         with cython.cdivision(True):
 *             new_shape = (stop - start) // step             # <<<<<<<<<<<<<<
 * 
 *             if (stop - start) - step * new_shape:
 */
    __pyx_v_new_shape = ((__pyx_v_stop - __pyx_v_start) / __pyx_v_step);

    /* "View.MemoryView":877
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    __pyx_t_2 = (((__pyx_v_stop - __pyx_v_start) - (__pyx_v_step * __pyx_v_new_shape)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":878
 * 
 *             if (stop - start) - step * new_shape:
 *                 new_shape += 1             # <<<<<<<<<<<<<<
 * 
 *         if new_shape < 0:
 */
      __pyx_v_new_shape = (__pyx_v_new_shape + 1);

      /* "View.MemoryView":877
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    }

    /* "View.MemoryView":880
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    __pyx_t_2 = ((__pyx_v_new_shape < 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":881
 * 
 *         if new_shape < 0:
 *             new_shape = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_new_shape = 0;

      /* "View.MemoryView":880
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    }

    /* "View.MemoryView":884
 * 
 * 
 *         dst.strides[new_ndim] = stride * step             # <<<<<<<<<<<<<<
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset
 */
    (__pyx_v_dst->strides[__pyx_v_new_ndim]) = (__pyx_v_stride * __pyx_v_step);

    /* "View.MemoryView":885
 * 
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape             # <<<<<<<<<<<<<<
 *         dst.suboffsets[new_ndim] = suboffset
 * 
 */
    (__pyx_v_dst->shape[__pyx_v_new_ndim]) = __pyx_v_new_shape;

    /* "View.MemoryView":886
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_dst->suboffsets[__pyx_v_new_ndim]) = __pyx_v_suboffset;
  }
  __pyx_L3:;

  /* "View.MemoryView":889
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
  __pyx_t_2 = (((__pyx_v_suboffset_dim[0]) < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":890
 * 
 *     if suboffset_dim[0] < 0:
 *         dst.data += start * stride             # <<<<<<<<<<<<<<
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 */
    __pyx_v_dst->data = (__pyx_v_dst->data + (__pyx_v_start * __pyx_v_stride));

    /* "View.MemoryView":889
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
    goto __pyx_L23;
  }

  /* "View.MemoryView":892
 *         dst.data += start * stride
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride             # <<<<<<<<<<<<<<
 * 
 *     if suboffset >= 0:
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_suboffset_dim[0]);
    (__pyx_v_dst->suboffsets[__pyx_t_3]) = ((__pyx_v_dst->suboffsets[__pyx_t_3]) + (__pyx_v_start * __pyx_v_stride));
  }
  __pyx_L23:;

  /* "View.MemoryView":894
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":895
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
    __pyx_t_2 = ((!(__pyx_v_is_slice != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":896
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
      __pyx_t_2 = ((__pyx_v_new_ndim == 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":897
 *         if not is_slice:
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 */
        __pyx_v_dst->data = ((((char **)__pyx_v_dst->data)[0]) + __pyx_v_suboffset);

        /* "View.MemoryView":896
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
        goto __pyx_L26;
      }

      /* "View.MemoryView":899
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "             # <<<<<<<<<<<<<<
 *                                      "must be indexed and not sliced", dim)
 *         else:
 */
      /*else*/ {

        /* "View.MemoryView":900
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 *                                      "must be indexed and not sliced", dim)             # <<<<<<<<<<<<<<
 *         else:
 *             suboffset_dim[0] = new_ndim
 */
        __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"All dimensions preceding dimension %d must be indexed and not sliced"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 899, __pyx_L1_error)
      }
      __pyx_L26:;

      /* "View.MemoryView":895
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
      goto __pyx_L25;
    }

    /* "View.MemoryView":902
 *                                      "must be indexed and not sliced", dim)
 *         else:
 *             suboffset_dim[0] = new_ndim             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
    /*else*/ {
      (__pyx_v_suboffset_dim[0]) = __pyx_v_new_ndim;
    }
    __pyx_L25:;

    /* "View.MemoryView":894
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  }

  /* "View.MemoryView":904
 *             suboffset_dim[0] = new_ndim
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":807
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.slice_memviewslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":910
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

static char *__pyx_pybuffer_index(Py_buffer *__pyx_v_view, char *__pyx_v_bufp, Py_ssize_t __pyx_v_index, Py_ssize_t __pyx_v_dim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_suboffset;
  Py_ssize_t __pyx_v_itemsize;
  char *__pyx_v_resultp;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("pybuffer_index", 0);

  /* "View.MemoryView":912
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t itemsize = view.itemsize
 *     cdef char *resultp
 */
  __pyx_v_suboffset = -1L;

  /* "View.MemoryView":913
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 *     cdef Py_ssize_t itemsize = view.itemsize             # <<<<<<<<<<<<<<
 *     cdef char *resultp
 * 
 */
  __pyx_t_1 = __pyx_v_view->itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":916
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */
  __pyx_t_2 = ((__pyx_v_view->ndim == 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":917
 * 
 *     if view.ndim == 0:
 *         shape = view.len / itemsize             # <<<<<<<<<<<<<<
 *         stride = itemsize
 *     else:
 */
    if (unlikely(__pyx_v_itemsize == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(1, 917, __pyx_L1_error)
    }
    else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_view->len))) {
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
      __PYX_ERR(1, 917, __pyx_L1_error)
    }
    __pyx_v_shape = (__pyx_v_view->len / __pyx_v_itemsize);

    /* "View.MemoryView":918
 *     if view.ndim == 0:
 *         shape = view.len / itemsize
 *         stride = itemsize             # <<<<<<<<<<<<<<
 *     else:
 *         shape = view.shape[dim]
 */
    __pyx_v_stride = __pyx_v_itemsize;

    /* "View.MemoryView":916
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":920
 *         stride = itemsize
 *     else:
 *         shape = view.shape[dim]             # <<<<<<<<<<<<<<
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 */
  /*else*/ {
    __pyx_v_shape = (__pyx_v_view->shape[__pyx_v_dim]);

    /* "View.MemoryView":921
 *     else:
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]             # <<<<<<<<<<<<<<
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]
 */
    __pyx_v_stride = (__pyx_v_view->strides[__pyx_v_dim]);

    /* "View.MemoryView":922
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    __pyx_t_2 = ((__pyx_v_view->suboffsets != NULL) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":923
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]             # <<<<<<<<<<<<<<
 * 
 *     if index < 0:
 */
      __pyx_v_suboffset = (__pyx_v_view->suboffsets[__pyx_v_dim]);

      /* "View.MemoryView":922
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":925
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  __pyx_t_2 = ((__pyx_v_index < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":926
 * 
 *     if index < 0:
 *         index += view.shape[dim]             # <<<<<<<<<<<<<<
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 */
    __pyx_v_index = (__pyx_v_index + (__pyx_v_view->shape[__pyx_v_dim]));

    /* "View.MemoryView":927
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
    __pyx_t_2 = ((__pyx_v_index < 0) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":928
 *         index += view.shape[dim]
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     if index >= shape:
 */
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(1, 928, __pyx_L1_error)

      /* "View.MemoryView":927
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
    }

    /* "View.MemoryView":925
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  }

  /* "View.MemoryView":930
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
  __pyx_t_2 = ((__pyx_v_index >= __pyx_v_shape) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":931
 * 
 *     if index >= shape:
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     resultp = bufp + index * stride
 */
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 931, __pyx_L1_error)

    /* "View.MemoryView":930
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
  }

  /* "View.MemoryView":933
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     resultp = bufp + index * stride             # <<<<<<<<<<<<<<
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset
 */
  __pyx_v_resultp = (__pyx_v_bufp + (__pyx_v_index * __pyx_v_stride));

  /* "View.MemoryView":934
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":935
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset             # <<<<<<<<<<<<<<
 * 
 *     return resultp
 */
    __pyx_v_resultp = ((((char **)__pyx_v_resultp)[0]) + __pyx_v_suboffset);

    /* "View.MemoryView":934
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  }

  /* "View.MemoryView":937
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 *     return resultp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_resultp;
  goto __pyx_L0;

  /* "View.MemoryView":910
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.pybuffer_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":943
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

static int __pyx_memslice_transpose(__Pyx_memviewslice *__pyx_v_memslice) {
  int __pyx_v_ndim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_r;
  int __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  long __pyx_t_3;
  long __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;

  /* "View.MemoryView":944
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:
 *     cdef int ndim = memslice.memview.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 */
  __pyx_t_1 = __pyx_v_memslice->memview->view.ndim;
  __pyx_v_ndim = __pyx_t_1;

  /* "View.MemoryView":946
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 *     cdef Py_ssize_t *shape = memslice.shape             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t *strides = memslice.strides
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->shape;
  __pyx_v_shape = __pyx_t_2;

  /* "View.MemoryView":947
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 *     cdef Py_ssize_t *strides = memslice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->strides;
  __pyx_v_strides = __pyx_t_2;

  /* "View.MemoryView":951
 * 
 *     cdef int i, j
 *     for i in range(ndim / 2):             # <<<<<<<<<<<<<<
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 */
  __pyx_t_3 = (__pyx_v_ndim / 2);
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_4; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":952
 *     cdef int i, j
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i             # <<<<<<<<<<<<<<
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]
 */
    __pyx_v_j = ((__pyx_v_ndim - 1) - __pyx_v_i);

    /* "View.MemoryView":953
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]             # <<<<<<<<<<<<<<
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 */
    __pyx_t_5 = (__pyx_v_strides[__pyx_v_j]);
    __pyx_t_6 = (__pyx_v_strides[__pyx_v_i]);
    (__pyx_v_strides[__pyx_v_i]) = __pyx_t_5;
    (__pyx_v_strides[__pyx_v_j]) = __pyx_t_6;

    /* "View.MemoryView":954
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]             # <<<<<<<<<<<<<<
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 */
    __pyx_t_6 = (__pyx_v_shape[__pyx_v_j]);
    __pyx_t_5 = (__pyx_v_shape[__pyx_v_i]);
    (__pyx_v_shape[__pyx_v_i]) = __pyx_t_6;
    (__pyx_v_shape[__pyx_v_j]) = __pyx_t_5;

    /* "View.MemoryView":956
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_i]) >= 0) != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_j]) >= 0) != 0);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_7) {

      /* "View.MemoryView":957
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")             # <<<<<<<<<<<<<<
 * 
 *     return 1
 */
      __pyx_t_9 = __pyx_memoryview_err(__pyx_builtin_ValueError, ((char *)"Cannot transpose memoryview with indirect dimensions")); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 957, __pyx_L1_error)

      /* "View.MemoryView":956
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    }
  }

  /* "View.MemoryView":959
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "View.MemoryView":943
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.transpose_memslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":976
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

/* Python wrapper */
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":977
 * 
 *     def __dealloc__(self):
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __PYX_XDEC_MEMVIEW((&__pyx_v_self->from_slice), 1);

  /* "View.MemoryView":976
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":979
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":980
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->to_object_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":981
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)             # <<<<<<<<<<<<<<
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_self->to_object_func(__pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 981, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":980
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  }

  /* "View.MemoryView":983
 *             return self.to_object_func(itemp)
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_memoryview_convert_item_to_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":979
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":985
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":986
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->to_dtype_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":987
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)             # <<<<<<<<<<<<<<
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)
 */
    __pyx_t_2 = __pyx_v_self->to_dtype_func(__pyx_v_itemp, __pyx_v_value); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(1, 987, __pyx_L1_error)

    /* "View.MemoryView":986
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":989
 *             self.to_dtype_func(itemp, value)
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  /*else*/ {
    __pyx_t_3 = __pyx_memoryview_assign_item_from_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 989, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":985
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":992
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":993
 *     @property
 *     def base(self):
 *         return self.from_object             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->from_object);
  __pyx_r = __pyx_v_self->from_object;
  goto __pyx_L0;

  /* "View.MemoryView":992
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice___reduce_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__44, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice_2__setstate_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__45, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":999
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice __pyx_v_memviewslice, int __pyx_v_ndim, PyObject *(*__pyx_v_to_object_func)(char *), int (*__pyx_v_to_dtype_func)(char *, PyObject *), int __pyx_v_dtype_is_object) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_TypeInfo *__pyx_t_4;
  Py_buffer __pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  Py_ssize_t *__pyx_t_8;
  Py_ssize_t __pyx_t_9;
  __Pyx_RefNannySetupContext("memoryview_fromslice", 0);

  /* "View.MemoryView":1007
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __pyx_t_1 = ((((PyObject *)__pyx_v_memviewslice.memview) == Py_None) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1008
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:
 *         return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "View.MemoryView":1007
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  }

  /* "View.MemoryView":1013
 * 
 * 
 *     result = _memoryviewslice(None, 0, dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     result.from_slice = memviewslice
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_3, 0, Py_None);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryviewslice_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":1015
 *     result = _memoryviewslice(None, 0, dtype_is_object)
 * 
 *     result.from_slice = memviewslice             # <<<<<<<<<<<<<<
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 */
  __pyx_v_result->from_slice = __pyx_v_memviewslice;

  /* "View.MemoryView":1016
 * 
 *     result.from_slice = memviewslice
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)             # <<<<<<<<<<<<<<
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 */
  __PYX_INC_MEMVIEW((&__pyx_v_memviewslice), 1);

  /* "View.MemoryView":1018
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base             # <<<<<<<<<<<<<<
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_memviewslice.memview), __pyx_n_s_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_result->from_object);
  __Pyx_DECREF(__pyx_v_result->from_object);
  __pyx_v_result->from_object = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "View.MemoryView":1019
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 *     result.typeinfo = memviewslice.memview.typeinfo             # <<<<<<<<<<<<<<
 * 
 *     result.view = memviewslice.memview.view
 */
  __pyx_t_4 = __pyx_v_memviewslice.memview->typeinfo;
  __pyx_v_result->__pyx_base.typeinfo = __pyx_t_4;

  /* "View.MemoryView":1021
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 *     result.view = memviewslice.memview.view             # <<<<<<<<<<<<<<
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 */
  __pyx_t_5 = __pyx_v_memviewslice.memview->view;
  __pyx_v_result->__pyx_base.view = __pyx_t_5;

  /* "View.MemoryView":1022
 * 
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data             # <<<<<<<<<<<<<<
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 */
  __pyx_v_result->__pyx_base.view.buf = ((void *)__pyx_v_memviewslice.data);

  /* "View.MemoryView":1023
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim             # <<<<<<<<<<<<<<
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)
 */
  __pyx_v_result->__pyx_base.view.ndim = __pyx_v_ndim;

  /* "View.MemoryView":1024
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None             # <<<<<<<<<<<<<<
 *     Py_INCREF(Py_None)
 * 
 */
  ((Py_buffer *)(&__pyx_v_result->__pyx_base.view))->obj = Py_None;

  /* "View.MemoryView":1025
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 */
  Py_INCREF(Py_None);

  /* "View.MemoryView":1027
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
  __pyx_t_1 = ((((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)->flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1028
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 *         result.flags = PyBUF_RECORDS             # <<<<<<<<<<<<<<
 *     else:
 *         result.flags = PyBUF_RECORDS_RO
 */
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS;

    /* "View.MemoryView":1027
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":1030
 *         result.flags = PyBUF_RECORDS
 *     else:
 *         result.flags = PyBUF_RECORDS_RO             # <<<<<<<<<<<<<<
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 */
  /*else*/ {
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS_RO;
  }
  __pyx_L4:;

  /* "View.MemoryView":1032
 *         result.flags = PyBUF_RECORDS_RO
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape             # <<<<<<<<<<<<<<
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides
 * 
 */
  __pyx_v_result->__pyx_base.view.shape = ((Py_ssize_t *)__pyx_v_result->from_slice.shape);

  /* "View.MemoryView":1033
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_result->__pyx_base.view.strides = ((Py_ssize_t *)__pyx_v_result->from_slice.strides);

  /* "View.MemoryView":1036
 * 
 * 
 *     result.view.suboffsets = NULL             # <<<<<<<<<<<<<<
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 */
  __pyx_v_result->__pyx_base.view.suboffsets = NULL;

  /* "View.MemoryView":1037
 * 
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 */
  __pyx_t_7 = (__pyx_v_result->from_slice.suboffsets + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->from_slice.suboffsets; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_v_suboffset = (__pyx_t_6[0]);

    /* "View.MemoryView":1038
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    __pyx_t_1 = ((__pyx_v_suboffset >= 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":1039
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_result->__pyx_base.view.suboffsets = ((Py_ssize_t *)__pyx_v_result->from_slice.suboffsets);

      /* "View.MemoryView":1040
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break             # <<<<<<<<<<<<<<
 * 
 *     result.view.len = result.view.itemsize
 */
      goto __pyx_L6_break;

      /* "View.MemoryView":1038
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    }
  }
  __pyx_L6_break:;

  /* "View.MemoryView":1042
 *             break
 * 
 *     result.view.len = result.view.itemsize             # <<<<<<<<<<<<<<
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length
 */
  __pyx_t_9 = __pyx_v_result->__pyx_base.view.itemsize;
  __pyx_v_result->__pyx_base.view.len = __pyx_t_9;

  /* "View.MemoryView":1043
 * 
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         result.view.len *= length
 * 
 */
  __pyx_t_7 = (__pyx_v_result->__pyx_base.view.shape + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->__pyx_base.view.shape; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_t_2 = PyInt_FromSsize_t((__pyx_t_6[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "View.MemoryView":1044
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length             # <<<<<<<<<<<<<<
 * 
 *     result.to_object_func = to_object_func
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_result->__pyx_base.view.len); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_InPlaceMultiply(__pyx_t_2, __pyx_v_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 1044, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result->__pyx_base.view.len = __pyx_t_9;
  }

  /* "View.MemoryView":1046
 *         result.view.len *= length
 * 
 *     result.to_object_func = to_object_func             # <<<<<<<<<<<<<<
 *     result.to_dtype_func = to_dtype_func
 * 
 */
  __pyx_v_result->to_object_func = __pyx_v_to_object_func;

  /* "View.MemoryView":1047
 * 
 *     result.to_object_func = to_object_func
 *     result.to_dtype_func = to_dtype_func             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  __pyx_v_result->to_dtype_func = __pyx_v_to_dtype_func;

  /* "View.MemoryView":1049
 *     result.to_dtype_func = to_dtype_func
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":999
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_fromslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1052
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_mslice) {
  struct __pyx_memoryviewslice_obj *__pyx_v_obj = 0;
  __Pyx_memviewslice *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("get_slice_from_memview", 0);

  /* "View.MemoryView":1055
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1056
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview             # <<<<<<<<<<<<<<
 *         return &obj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(1, 1056, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_obj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":1057
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview
 *         return &obj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, mslice)
 */
    __pyx_r = (&__pyx_v_obj->from_slice);
    goto __pyx_L0;

    /* "View.MemoryView":1055
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  }

  /* "View.MemoryView":1059
 *         return &obj.from_slice
 *     else:
 *         slice_copy(memview, mslice)             # <<<<<<<<<<<<<<
 *         return mslice
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, __pyx_v_mslice);

    /* "View.MemoryView":1060
 *     else:
 *         slice_copy(memview, mslice)
 *         return mslice             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_slice_copy')
 */
    __pyx_r = __pyx_v_mslice;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1052
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.get_slice_from_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_obj);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1063
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_dst) {
  int __pyx_v_dim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  Py_ssize_t *__pyx_v_suboffsets;
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  __Pyx_RefNannySetupContext("slice_copy", 0);

  /* "View.MemoryView":1067
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 * 
 *     shape = memview.view.shape             # <<<<<<<<<<<<<<
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets
 */
  __pyx_t_1 = __pyx_v_memview->view.shape;
  __pyx_v_shape = __pyx_t_1;

  /* "View.MemoryView":1068
 * 
 *     shape = memview.view.shape
 *     strides = memview.view.strides             # <<<<<<<<<<<<<<
 *     suboffsets = memview.view.suboffsets
 * 
 */
  __pyx_t_1 = __pyx_v_memview->view.strides;
  __pyx_v_strides = __pyx_t_1;

  /* "View.MemoryView":1069
 *     shape = memview.view.shape
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets             # <<<<<<<<<<<<<<
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 */
  __pyx_t_1 = __pyx_v_memview->view.suboffsets;
  __pyx_v_suboffsets = __pyx_t_1;

  /* "View.MemoryView":1071
 *     suboffsets = memview.view.suboffsets
 * 
 *     dst.memview = <__pyx_memoryview *> memview             # <<<<<<<<<<<<<<
 *     dst.data = <char *> memview.view.buf
 * 
 */
  __pyx_v_dst->memview = ((struct __pyx_memoryview_obj *)__pyx_v_memview);

  /* "View.MemoryView":1072
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 *     dst.data = <char *> memview.view.buf             # <<<<<<<<<<<<<<
 * 
 *     for dim in range(memview.view.ndim):
 */
  __pyx_v_dst->data = ((char *)__pyx_v_memview->view.buf);

  /* "View.MemoryView":1074
 *     dst.data = <char *> memview.view.buf
 * 
 *     for dim in range(memview.view.ndim):             # <<<<<<<<<<<<<<
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 */
  __pyx_t_2 = __pyx_v_memview->view.ndim;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_dim = __pyx_t_4;

    /* "View.MemoryView":1075
 * 
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]             # <<<<<<<<<<<<<<
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 */
    (__pyx_v_dst->shape[__pyx_v_dim]) = (__pyx_v_shape[__pyx_v_dim]);

    /* "View.MemoryView":1076
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]             # <<<<<<<<<<<<<<
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 * 
 */
    (__pyx_v_dst->strides[__pyx_v_dim]) = (__pyx_v_strides[__pyx_v_dim]);

    /* "View.MemoryView":1077
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object')
 */
    if ((__pyx_v_suboffsets != 0)) {
      __pyx_t_5 = (__pyx_v_suboffsets[__pyx_v_dim]);
    } else {
      __pyx_t_5 = -1L;
    }
    (__pyx_v_dst->suboffsets[__pyx_v_dim]) = __pyx_t_5;
  }

  /* "View.MemoryView":1063
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1080
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *__pyx_v_memview) {
  __Pyx_memviewslice __pyx_v_memviewslice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("memoryview_copy", 0);

  /* "View.MemoryView":1083
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)             # <<<<<<<<<<<<<<
 *     return memoryview_copy_from_slice(memview, &memviewslice)
 * 
 */
  __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_memviewslice));

  /* "View.MemoryView":1084
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)
 *     return memoryview_copy_from_slice(memview, &memviewslice)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_memoryview_copy_object_from_slice(__pyx_v_memview, (&__pyx_v_memviewslice)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1080
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1087
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_memviewslice) {
  PyObject *(*__pyx_v_to_object_func)(char *);
  int (*__pyx_v_to_dtype_func)(char *, PyObject *);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *(*__pyx_t_3)(char *);
  int (*__pyx_t_4)(char *, PyObject *);
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("memoryview_copy_from_slice", 0);

  /* "View.MemoryView":1094
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1095
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func             # <<<<<<<<<<<<<<
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 */
    __pyx_t_3 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_object_func;
    __pyx_v_to_object_func = __pyx_t_3;

    /* "View.MemoryView":1096
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func             # <<<<<<<<<<<<<<
 *     else:
 *         to_object_func = NULL
 */
    __pyx_t_4 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_dtype_func;
    __pyx_v_to_dtype_func = __pyx_t_4;

    /* "View.MemoryView":1094
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1098
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 *         to_object_func = NULL             # <<<<<<<<<<<<<<
 *         to_dtype_func = NULL
 * 
 */
  /*else*/ {
    __pyx_v_to_object_func = NULL;

    /* "View.MemoryView":1099
 *     else:
 *         to_object_func = NULL
 *         to_dtype_func = NULL             # <<<<<<<<<<<<<<
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 */
    __pyx_v_to_dtype_func = NULL;
  }
  __pyx_L3:;

  /* "View.MemoryView":1101
 *         to_dtype_func = NULL
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,             # <<<<<<<<<<<<<<
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "View.MemoryView":1103
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __pyx_memoryview_fromslice((__pyx_v_memviewslice[0]), __pyx_v_memview->view.ndim, __pyx_v_to_object_func, __pyx_v_to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1087
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy_from_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1109
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

static Py_ssize_t abs_py_ssize_t(Py_ssize_t __pyx_v_arg) {
  Py_ssize_t __pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":1110
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_arg < 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1111
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:
 *         return -arg             # <<<<<<<<<<<<<<
 *     else:
 *         return arg
 */
    __pyx_r = (-__pyx_v_arg);
    goto __pyx_L0;

    /* "View.MemoryView":1110
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */
  }

  /* "View.MemoryView":1113
 *         return -arg
 *     else:
 *         return arg             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_get_best_slice_order')
 */
  /*else*/ {
    __pyx_r = __pyx_v_arg;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1109
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1116
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

static char __pyx_get_best_slice_order(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim) {
  int __pyx_v_i;
  Py_ssize_t __pyx_v_c_stride;
  Py_ssize_t __pyx_v_f_stride;
  char __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1121
 *     """
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t f_stride = 0
 * 
 */
  __pyx_v_c_stride = 0;

  /* "View.MemoryView":1122
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0
 *     cdef Py_ssize_t f_stride = 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_f_stride = 0;

  /* "View.MemoryView":1124
 *     cdef Py_ssize_t f_stride = 0
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1125
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1126
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_c_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1127
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
      goto __pyx_L4_break;

      /* "View.MemoryView":1125
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L4_break:;

  /* "View.MemoryView":1129
 *             break
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 */
  __pyx_t_1 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1130
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1131
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_f_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1132
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 */
      goto __pyx_L7_break;

      /* "View.MemoryView":1130
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L7_break:;

  /* "View.MemoryView":1134
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  __pyx_t_2 = ((abs_py_ssize_t(__pyx_v_c_stride) <= abs_py_ssize_t(__pyx_v_f_stride)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1135
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 *         return 'C'             # <<<<<<<<<<<<<<
 *     else:
 *         return 'F'
 */
    __pyx_r = 'C';
    goto __pyx_L0;

    /* "View.MemoryView":1134
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  }

  /* "View.MemoryView":1137
 *         return 'C'
 *     else:
 *         return 'F'             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  /*else*/ {
    __pyx_r = 'F';
    goto __pyx_L0;
  }

  /* "View.MemoryView":1116
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1140
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

static void _copy_strided_to_strided(char *__pyx_v_src_data, Py_ssize_t *__pyx_v_src_strides, char *__pyx_v_dst_data, Py_ssize_t *__pyx_v_dst_strides, Py_ssize_t *__pyx_v_src_shape, Py_ssize_t *__pyx_v_dst_shape, int __pyx_v_ndim, size_t __pyx_v_itemsize) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED Py_ssize_t __pyx_v_src_extent;
  Py_ssize_t __pyx_v_dst_extent;
  Py_ssize_t __pyx_v_src_stride;
  Py_ssize_t __pyx_v_dst_stride;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;

  /* "View.MemoryView":1147
 * 
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 */
  __pyx_v_src_extent = (__pyx_v_src_shape[0]);

  /* "View.MemoryView":1148
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 */
  __pyx_v_dst_extent = (__pyx_v_dst_shape[0]);

  /* "View.MemoryView":1149
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 */
  __pyx_v_src_stride = (__pyx_v_src_strides[0]);

  /* "View.MemoryView":1150
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_dst_stride = (__pyx_v_dst_strides[0]);

  /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    __pyx_t_2 = ((__pyx_v_src_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_dst_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }

    /* "View.MemoryView":1154
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):             # <<<<<<<<<<<<<<
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 */
    __pyx_t_2 = (((size_t)__pyx_v_src_stride) == __pyx_v_itemsize);
    if (__pyx_t_2) {
      __pyx_t_2 = (__pyx_v_itemsize == ((size_t)__pyx_v_dst_stride));
    }
    __pyx_t_3 = (__pyx_t_2 != 0);
    __pyx_t_1 = __pyx_t_3;
    __pyx_L5_bool_binop_done:;

    /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    if (__pyx_t_1) {

      /* "View.MemoryView":1155
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)             # <<<<<<<<<<<<<<
 *        else:
 *            for i in range(dst_extent):
 */
      (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, (__pyx_v_itemsize * __pyx_v_dst_extent)));

      /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
      goto __pyx_L4;
    }

    /* "View.MemoryView":1157
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 *            for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 */
    /*else*/ {
      __pyx_t_4 = __pyx_v_dst_extent;
      __pyx_t_5 = __pyx_t_4;
      for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
        __pyx_v_i = __pyx_t_6;

        /* "View.MemoryView":1158
 *        else:
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)             # <<<<<<<<<<<<<<
 *                src_data += src_stride
 *                dst_data += dst_stride
 */
        (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, __pyx_v_itemsize));

        /* "View.MemoryView":1159
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride             # <<<<<<<<<<<<<<
 *                dst_data += dst_stride
 *     else:
 */
        __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

        /* "View.MemoryView":1160
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 *                dst_data += dst_stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(dst_extent):
 */
        __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
      }
    }
    __pyx_L4:;

    /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1162
 *                dst_data += dst_stride
 *     else:
 *         for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *             _copy_strided_to_strided(src_data, src_strides + 1,
 *                                      dst_data, dst_strides + 1,
 */
  /*else*/ {
    __pyx_t_4 = __pyx_v_dst_extent;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "View.MemoryView":1163
 *     else:
 *         for i in range(dst_extent):
 *             _copy_strided_to_strided(src_data, src_strides + 1,             # <<<<<<<<<<<<<<
 *                                      dst_data, dst_strides + 1,
 *                                      src_shape + 1, dst_shape + 1,
 */
      _copy_strided_to_strided(__pyx_v_src_data, (__pyx_v_src_strides + 1), __pyx_v_dst_data, (__pyx_v_dst_strides + 1), (__pyx_v_src_shape + 1), (__pyx_v_dst_shape + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize);

      /* "View.MemoryView":1167
 *                                      src_shape + 1, dst_shape + 1,
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride             # <<<<<<<<<<<<<<
 *             dst_data += dst_stride
 * 
 */
      __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

      /* "View.MemoryView":1168
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride
 *             dst_data += dst_stride             # <<<<<<<<<<<<<<
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,
 */
      __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1140
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

  /* function exit code */
}

/* "View.MemoryView":1170
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

static void copy_strided_to_strided(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize) {

  /* "View.MemoryView":1173
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 *     _copy_strided_to_strided(src.data, src.strides, dst.data, dst.strides,             # <<<<<<<<<<<<<<
 *                              src.shape, dst.shape, ndim, itemsize)
 * 
 */
  _copy_strided_to_strided(__pyx_v_src->data, __pyx_v_src->strides, __pyx_v_dst->data, __pyx_v_dst->strides, __pyx_v_src->shape, __pyx_v_dst->shape, __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1170
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1177
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *__pyx_v_src, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_size;
  Py_ssize_t __pyx_r;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;

  /* "View.MemoryView":1179
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     for shape in src.shape[:ndim]:
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_size = __pyx_t_1;

  /* "View.MemoryView":1181
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 * 
 *     for shape in src.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         size *= shape
 * 
 */
  __pyx_t_3 = (__pyx_v_src->shape + __pyx_v_ndim);
  for (__pyx_t_4 = __pyx_v_src->shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_shape = (__pyx_t_2[0]);

    /* "View.MemoryView":1182
 * 
 *     for shape in src.shape[:ndim]:
 *         size *= shape             # <<<<<<<<<<<<<<
 * 
 *     return size
 */
    __pyx_v_size = (__pyx_v_size * __pyx_v_shape);
  }

  /* "View.MemoryView":1184
 *         size *= shape
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_fill_contig_strides_array')
 */
  __pyx_r = __pyx_v_size;
  goto __pyx_L0;

  /* "View.MemoryView":1177
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1187
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, Py_ssize_t __pyx_v_stride, int __pyx_v_ndim, char __pyx_v_order) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1196
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
  __pyx_t_1 = ((__pyx_v_order == 'F') != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1197
 * 
 *     if order == 'F':
 *         for idx in range(ndim):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
    __pyx_t_2 = __pyx_v_ndim;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_idx = __pyx_t_4;

      /* "View.MemoryView":1198
 *     if order == 'F':
 *         for idx in range(ndim):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 *     else:
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1199
 *         for idx in range(ndim):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }

    /* "View.MemoryView":1196
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1201
 *             stride *= shape[idx]
 *     else:
 *         for idx in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
  /*else*/ {
    for (__pyx_t_2 = (__pyx_v_ndim - 1); __pyx_t_2 > -1; __pyx_t_2-=1) {
      __pyx_v_idx = __pyx_t_2;

      /* "View.MemoryView":1202
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 * 
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1203
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 * 
 *     return stride
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1205
 *             stride *= shape[idx]
 * 
 *     return stride             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 */
  __pyx_r = __pyx_v_stride;
  goto __pyx_L0;

  /* "View.MemoryView":1187
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1208
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_tmpslice, char __pyx_v_order, int __pyx_v_ndim) {
  int __pyx_v_i;
  void *__pyx_v_result;
  size_t __pyx_v_itemsize;
  size_t __pyx_v_size;
  void *__pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  struct __pyx_memoryview_obj *__pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;

  /* "View.MemoryView":1219
 *     cdef void *result
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1220
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef size_t size = slice_get_size(src, ndim)             # <<<<<<<<<<<<<<
 * 
 *     result = malloc(size)
 */
  __pyx_v_size = __pyx_memoryview_slice_get_size(__pyx_v_src, __pyx_v_ndim);

  /* "View.MemoryView":1222
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 *     result = malloc(size)             # <<<<<<<<<<<<<<
 *     if not result:
 *         _err(MemoryError, NULL)
 */
  __pyx_v_result = malloc(__pyx_v_size);

  /* "View.MemoryView":1223
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_result != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1224
 *     result = malloc(size)
 *     if not result:
 *         _err(MemoryError, NULL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_err(__pyx_builtin_MemoryError, NULL); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 1224, __pyx_L1_error)

    /* "View.MemoryView":1223
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */
  }

  /* "View.MemoryView":1227
 * 
 * 
 *     tmpslice.data = <char *> result             # <<<<<<<<<<<<<<
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 */
  __pyx_v_tmpslice->data = ((char *)__pyx_v_result);

  /* "View.MemoryView":1228
 * 
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview             # <<<<<<<<<<<<<<
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 */
  __pyx_t_4 = __pyx_v_src->memview;
  __pyx_v_tmpslice->memview = __pyx_t_4;

  /* "View.MemoryView":1229
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1230
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]             # <<<<<<<<<<<<<<
 *         tmpslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_tmpslice->shape[__pyx_v_i]) = (__pyx_v_src->shape[__pyx_v_i]);

    /* "View.MemoryView":1231
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,
 */
    (__pyx_v_tmpslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1233
 *         tmpslice.suboffsets[i] = -1
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,             # <<<<<<<<<<<<<<
 *                               ndim, order)
 * 
 */
  (void)(__pyx_fill_contig_strides_array((&(__pyx_v_tmpslice->shape[0])), (&(__pyx_v_tmpslice->strides[0])), __pyx_v_itemsize, __pyx_v_ndim, __pyx_v_order));

  /* "View.MemoryView":1237
 * 
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1238
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    __pyx_t_2 = (((__pyx_v_tmpslice->shape[__pyx_v_i]) == 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1239
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     if slice_is_contig(src[0], order, ndim):
 */
      (__pyx_v_tmpslice->strides[__pyx_v_i]) = 0;

      /* "View.MemoryView":1238
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    }
  }

  /* "View.MemoryView":1241
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
  __pyx_t_2 = (__pyx_memviewslice_is_contig((__pyx_v_src[0]), __pyx_v_order, __pyx_v_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1242
 * 
 *     if slice_is_contig(src[0], order, ndim):
 *         memcpy(result, src.data, size)             # <<<<<<<<<<<<<<
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 */
    (void)(memcpy(__pyx_v_result, __pyx_v_src->data, __pyx_v_size));

    /* "View.MemoryView":1241
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":1244
 *         memcpy(result, src.data, size)
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  /*else*/ {
    copy_strided_to_strided(__pyx_v_src, __pyx_v_tmpslice, __pyx_v_ndim, __pyx_v_itemsize);
  }
  __pyx_L9:;

  /* "View.MemoryView":1246
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":1208
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.copy_data_to_temp", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1251
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

static int __pyx_memoryview_err_extents(int __pyx_v_i, Py_ssize_t __pyx_v_extent1, Py_ssize_t __pyx_v_extent2) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_extents", 0);

  /* "View.MemoryView":1254
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 *                                                         (i, extent1, extent2))             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err_dim')
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_extent1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_extent2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;

  /* "View.MemoryView":1253
 * cdef int _err_extents(int i, Py_ssize_t extent1,
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %             # <<<<<<<<<<<<<<
 *                                                         (i, extent1, extent2))
 * 
 */
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_4, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __PYX_ERR(1, 1253, __pyx_L1_error)

  /* "View.MemoryView":1251
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1257
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

static int __pyx_memoryview_err_dim(PyObject *__pyx_v_error, char *__pyx_v_msg, int __pyx_v_dim) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_dim", 0);
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1258
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:
 *     raise error(msg.decode('ascii') % dim)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err')
 */
  __pyx_t_2 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyUnicode_Format(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_v_error);
  __pyx_t_3 = __pyx_v_error; __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 1258, __pyx_L1_error)

  /* "View.MemoryView":1257
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_dim", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1261
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

static int __pyx_memoryview_err(PyObject *__pyx_v_error, char *__pyx_v_msg) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err", 0);
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1262
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_msg != NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":1263
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))             # <<<<<<<<<<<<<<
 *     else:
 *         raise error
 */
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_error);
    __pyx_t_4 = __pyx_v_error; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 1263, __pyx_L1_error)

    /* "View.MemoryView":1262
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */
  }

  /* "View.MemoryView":1265
 *         raise error(msg.decode('ascii'))
 *     else:
 *         raise error             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_contents')
 */
  /*else*/ {
    __Pyx_Raise(__pyx_v_error, 0, 0, 0);
    __PYX_ERR(1, 1265, __pyx_L1_error)
  }

  /* "View.MemoryView":1261
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView._err", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1268
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

static int __pyx_memoryview_copy_contents(__Pyx_memviewslice __pyx_v_src, __Pyx_memviewslice __pyx_v_dst, int __pyx_v_src_ndim, int __pyx_v_dst_ndim, int __pyx_v_dtype_is_object) {
  void *__pyx_v_tmpdata;
  size_t __pyx_v_itemsize;
  int __pyx_v_i;
  char __pyx_v_order;
  int __pyx_v_broadcasting;
  int __pyx_v_direct_copy;
  __Pyx_memviewslice __pyx_v_tmp;
  int __pyx_v_ndim;
  int __pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_t_8;

  /* "View.MemoryView":1276
 *     Check for overlapping memory and verify the shapes.
 *     """
 *     cdef void *tmpdata = NULL             # <<<<<<<<<<<<<<
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 */
  __pyx_v_tmpdata = NULL;

  /* "View.MemoryView":1277
 *     """
 *     cdef void *tmpdata = NULL
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 */
  __pyx_t_1 = __pyx_v_src.memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1279
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)             # <<<<<<<<<<<<<<
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False
 */
  __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_src), __pyx_v_src_ndim);

  /* "View.MemoryView":1280
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False             # <<<<<<<<<<<<<<
 *     cdef bint direct_copy = False
 *     cdef __Pyx_memviewslice tmp
 */
  __pyx_v_broadcasting = 0;

  /* "View.MemoryView":1281
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False             # <<<<<<<<<<<<<<
 *     cdef __Pyx_memviewslice tmp
 * 
 */
  __pyx_v_direct_copy = 0;

  /* "View.MemoryView":1284
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
  __pyx_t_2 = ((__pyx_v_src_ndim < __pyx_v_dst_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1285
 * 
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_src), __pyx_v_src_ndim, __pyx_v_dst_ndim);

    /* "View.MemoryView":1284
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1286
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  __pyx_t_2 = ((__pyx_v_dst_ndim < __pyx_v_src_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1287
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)             # <<<<<<<<<<<<<<
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_dst), __pyx_v_dst_ndim, __pyx_v_src_ndim);

    /* "View.MemoryView":1286
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":1289
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
  __pyx_t_3 = __pyx_v_dst_ndim;
  __pyx_t_4 = __pyx_v_src_ndim;
  if (((__pyx_t_3 > __pyx_t_4) != 0)) {
    __pyx_t_5 = __pyx_t_3;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  __pyx_v_ndim = __pyx_t_5;

  /* "View.MemoryView":1291
 *     cdef int ndim = max(src_ndim, dst_ndim)
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 */
  __pyx_t_5 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_5;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1292
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) != (__pyx_v_dst.shape[__pyx_v_i])) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1293
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
      __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) == 1) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":1294
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 *                 broadcasting = True             # <<<<<<<<<<<<<<
 *                 src.strides[i] = 0
 *             else:
 */
        __pyx_v_broadcasting = 1;

        /* "View.MemoryView":1295
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 *                 src.strides[i] = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 */
        (__pyx_v_src.strides[__pyx_v_i]) = 0;

        /* "View.MemoryView":1293
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":1297
 *                 src.strides[i] = 0
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])             # <<<<<<<<<<<<<<
 * 
 *         if src.suboffsets[i] >= 0:
 */
      /*else*/ {
        __pyx_t_6 = __pyx_memoryview_err_extents(__pyx_v_i, (__pyx_v_dst.shape[__pyx_v_i]), (__pyx_v_src.shape[__pyx_v_i])); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 1297, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":1292
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    }

    /* "View.MemoryView":1299
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */
    __pyx_t_2 = (((__pyx_v_src.suboffsets[__pyx_v_i]) >= 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1300
 * 
 *         if src.suboffsets[i] >= 0:
 *             _err_dim(ValueError, "Dimension %d is not direct", i)             # <<<<<<<<<<<<<<
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 */
      __pyx_t_6 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Dimension %d is not direct"), __pyx_v_i); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 1300, __pyx_L1_error)

      /* "View.MemoryView":1299
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */
    }
  }

  /* "View.MemoryView":1302
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  __pyx_t_2 = (__pyx_slices_overlap((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1304
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    __pyx_t_2 = ((!(__pyx_memviewslice_is_contig(__pyx_v_src, __pyx_v_order, __pyx_v_ndim) != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1305
 * 
 *         if not slice_is_contig(src, order, ndim):
 *             order = get_best_order(&dst, ndim)             # <<<<<<<<<<<<<<
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 */
      __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim);

      /* "View.MemoryView":1304
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    }

    /* "View.MemoryView":1307
 *             order = get_best_order(&dst, ndim)
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)             # <<<<<<<<<<<<<<
 *         src = tmp
 * 
 */
    __pyx_t_7 = __pyx_memoryview_copy_data_to_temp((&__pyx_v_src), (&__pyx_v_tmp), __pyx_v_order, __pyx_v_ndim); if (unlikely(__pyx_t_7 == ((void *)NULL))) __PYX_ERR(1, 1307, __pyx_L1_error)
    __pyx_v_tmpdata = __pyx_t_7;

    /* "View.MemoryView":1308
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 *         src = tmp             # <<<<<<<<<<<<<<
 * 
 *     if not broadcasting:
 */
    __pyx_v_src = __pyx_v_tmp;

    /* "View.MemoryView":1302
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  }

  /* "View.MemoryView":1310
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_broadcasting != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1313
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'C', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1314
 * 
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)             # <<<<<<<<<<<<<<
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'C', __pyx_v_ndim);

      /* "View.MemoryView":1313
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
      goto __pyx_L12;
    }

    /* "View.MemoryView":1315
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'F', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1316
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)             # <<<<<<<<<<<<<<
 * 
 *         if direct_copy:
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'F', __pyx_v_ndim);

      /* "View.MemoryView":1315
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    }
    __pyx_L12:;

    /* "View.MemoryView":1318
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    __pyx_t_2 = (__pyx_v_direct_copy != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1320
 *         if direct_copy:
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

      /* "View.MemoryView":1321
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))             # <<<<<<<<<<<<<<
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 */
      (void)(memcpy(__pyx_v_dst.data, __pyx_v_src.data, __pyx_memoryview_slice_get_size((&__pyx_v_src), __pyx_v_ndim)));

      /* "View.MemoryView":1322
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 *             free(tmpdata)
 *             return 0
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

      /* "View.MemoryView":1323
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
      free(__pyx_v_tmpdata);

      /* "View.MemoryView":1324
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "View.MemoryView":1318
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    }

    /* "View.MemoryView":1310
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1326
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = (__pyx_v_order == 'F');
  if (__pyx_t_2) {
    __pyx_t_2 = ('F' == __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim));
  }
  __pyx_t_8 = (__pyx_t_2 != 0);
  if (__pyx_t_8) {

    /* "View.MemoryView":1329
 * 
 * 
 *         transpose_memslice(&src)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&dst)
 * 
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_src)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(1, 1329, __pyx_L1_error)

    /* "View.MemoryView":1330
 * 
 *         transpose_memslice(&src)
 *         transpose_memslice(&dst)             # <<<<<<<<<<<<<<
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_dst)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(1, 1330, __pyx_L1_error)

    /* "View.MemoryView":1326
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1332
 *         transpose_memslice(&dst)
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1333
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)             # <<<<<<<<<<<<<<
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 */
  copy_strided_to_strided((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1334
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 *     free(tmpdata)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1336
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 *     free(tmpdata)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  free(__pyx_v_tmpdata);

  /* "View.MemoryView":1337
 * 
 *     free(tmpdata)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1268
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.memoryview_copy_contents", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1340
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim, int __pyx_v_ndim_other) {
  int __pyx_v_i;
  int __pyx_v_offset;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "View.MemoryView":1344
 *                             int ndim_other) nogil:
 *     cdef int i
 *     cdef int offset = ndim_other - ndim             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_offset = (__pyx_v_ndim_other - __pyx_v_ndim);

  /* "View.MemoryView":1346
 *     cdef int offset = ndim_other - ndim
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1347
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]             # <<<<<<<<<<<<<<
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 */
    (__pyx_v_mslice->shape[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->shape[__pyx_v_i]);

    /* "View.MemoryView":1348
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 */
    (__pyx_v_mslice->strides[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->strides[__pyx_v_i]);

    /* "View.MemoryView":1349
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(offset):
 */
    (__pyx_v_mslice->suboffsets[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->suboffsets[__pyx_v_i]);
  }

  /* "View.MemoryView":1351
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 *     for i in range(offset):             # <<<<<<<<<<<<<<
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 */
  __pyx_t_1 = __pyx_v_offset;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1352
 * 
 *     for i in range(offset):
 *         mslice.shape[i] = 1             # <<<<<<<<<<<<<<
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1
 */
    (__pyx_v_mslice->shape[__pyx_v_i]) = 1;

    /* "View.MemoryView":1353
 *     for i in range(offset):
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_mslice->strides[__pyx_v_i]) = (__pyx_v_mslice->strides[0]);

    /* "View.MemoryView":1354
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_mslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1340
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1362
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_dtype_is_object, int __pyx_v_ndim, int __pyx_v_inc) {
  int __pyx_t_1;

  /* "View.MemoryView":1366
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */
  __pyx_t_1 = (__pyx_v_dtype_is_object != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1367
 * 
 *     if dtype_is_object:
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,             # <<<<<<<<<<<<<<
 *                                            dst.strides, ndim, inc)
 * 
 */
    __pyx_memoryview_refcount_objects_in_slice_with_gil(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_inc);

    /* "View.MemoryView":1366
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */
  }

  /* "View.MemoryView":1362
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

  /* function exit code */
}

/* "View.MemoryView":1371
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  __Pyx_RefNannyDeclarations
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("refcount_objects_in_slice_with_gil", 0);

  /* "View.MemoryView":1374
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 *     refcount_objects_in_slice(data, shape, strides, ndim, inc)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 */
  __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, __pyx_v_shape, __pyx_v_strides, __pyx_v_ndim, __pyx_v_inc);

  /* "View.MemoryView":1371
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "View.MemoryView":1377
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

static void __pyx_memoryview_refcount_objects_in_slice(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("refcount_objects_in_slice", 0);

  /* "View.MemoryView":1381
 *     cdef Py_ssize_t i
 * 
 *     for i in range(shape[0]):             # <<<<<<<<<<<<<<
 *         if ndim == 1:
 *             if inc:
 */
  __pyx_t_1 = (__pyx_v_shape[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1382
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
    __pyx_t_4 = ((__pyx_v_ndim == 1) != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":1383
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
      __pyx_t_4 = (__pyx_v_inc != 0);
      if (__pyx_t_4) {

        /* "View.MemoryView":1384
 *         if ndim == 1:
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])
 */
        Py_INCREF((((PyObject **)__pyx_v_data)[0]));

        /* "View.MemoryView":1383
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
        goto __pyx_L6;
      }

      /* "View.MemoryView":1386
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 */
      /*else*/ {
        Py_DECREF((((PyObject **)__pyx_v_data)[0]));
      }
      __pyx_L6:;

      /* "View.MemoryView":1382
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":1388
 *                 Py_DECREF((<PyObject **> data)[0])
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                       ndim - 1, inc)
 * 
 */
    /*else*/ {

      /* "View.MemoryView":1389
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 *                                       ndim - 1, inc)             # <<<<<<<<<<<<<<
 * 
 *         data += strides[0]
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_inc);
    }
    __pyx_L5:;

    /* "View.MemoryView":1391
 *                                       ndim - 1, inc)
 * 
 *         data += strides[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_data = (__pyx_v_data + (__pyx_v_strides[0]));
  }

  /* "View.MemoryView":1377
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1397
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item, int __pyx_v_dtype_is_object) {

  /* "View.MemoryView":1400
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1401
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,             # <<<<<<<<<<<<<<
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)
 */
  __pyx_memoryview__slice_assign_scalar(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_itemsize, __pyx_v_item);

  /* "View.MemoryView":1403
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1397
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1407
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

static void __pyx_memoryview__slice_assign_scalar(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_extent;
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;

  /* "View.MemoryView":1411
 *                               size_t itemsize, void *item) nogil:
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t extent = shape[0]
 * 
 */
  __pyx_v_stride = (__pyx_v_strides[0]);

  /* "View.MemoryView":1412
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]
 *     cdef Py_ssize_t extent = shape[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_extent = (__pyx_v_shape[0]);

  /* "View.MemoryView":1414
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1415
 * 
 *     if ndim == 1:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             memcpy(data, item, itemsize)
 *             data += stride
 */
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1416
 *     if ndim == 1:
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)             # <<<<<<<<<<<<<<
 *             data += stride
 *     else:
 */
      (void)(memcpy(__pyx_v_data, __pyx_v_item, __pyx_v_itemsize));

      /* "View.MemoryView":1417
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 *             data += stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(extent):
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }

    /* "View.MemoryView":1414
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1419
 *             data += stride
 *     else:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 */
  /*else*/ {
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1420
 *     else:
 *         for i in range(extent):
 *             _slice_assign_scalar(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                 ndim - 1, itemsize, item)
 *             data += stride
 */
      __pyx_memoryview__slice_assign_scalar(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize, __pyx_v_item);

      /* "View.MemoryView":1422
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 *             data += stride             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1407
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum = {"__pyx_unpickle_Enum", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Enum") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xb068931) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xb068931:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xb0, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_MemviewEnum_type), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_unpickle_Enum__set_state(((struct __pyx_MemviewEnum_obj *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 1) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_10fatslimlib_9core_base_Atom(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10fatslimlib_9core_base_Atom *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10fatslimlib_9core_base_Atom *)o);
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10fatslimlib_9core_base_Atom(PyObject *o) {
  struct __pyx_obj_10fatslimlib_9core_base_Atom *p = (struct __pyx_obj_10fatslimlib_9core_base_Atom *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  if (p->__weakref__) PyObject_ClearWeakRefs(o);
  Py_CLEAR(p->name);
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_4Atom_x(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_4Atom_1x_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_4Atom_y(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_4Atom_1y_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_4Atom_z(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_4Atom_1z_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_4Atom_atomid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_4Atom_6atomid_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_4Atom_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_4Atom_4name_1__get__(o);
}

static PyMethodDef __pyx_methods_10fatslimlib_9core_base_Atom[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_4Atom_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_4Atom_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10fatslimlib_9core_base_Atom[] = {
  {(char *)"x", __pyx_getprop_10fatslimlib_9core_base_4Atom_x, 0, (char *)0, 0},
  {(char *)"y", __pyx_getprop_10fatslimlib_9core_base_4Atom_y, 0, (char *)0, 0},
  {(char *)"z", __pyx_getprop_10fatslimlib_9core_base_4Atom_z, 0, (char *)0, 0},
  {(char *)"atomid", __pyx_getprop_10fatslimlib_9core_base_4Atom_atomid, 0, (char *)0, 0},
  {(char *)"name", __pyx_getprop_10fatslimlib_9core_base_4Atom_name, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_10fatslimlib_9core_base_Atom = {
  PyVarObject_HEAD_INIT(0, 0)
  "fatslimlib.core_base.Atom", /*tp_name*/
  sizeof(struct __pyx_obj_10fatslimlib_9core_base_Atom), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10fatslimlib_9core_base_Atom, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10fatslimlib_9core_base_Atom, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10fatslimlib_9core_base_Atom, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_10fatslimlib_9core_base_4Atom_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10fatslimlib_9core_base_Atom, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_10fatslimlib_9core_base_AtomGroup __pyx_vtable_10fatslimlib_9core_base_AtomGroup;

static PyObject *__pyx_tp_new_10fatslimlib_9core_base_AtomGroup(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *)o);
  p->__pyx_vtab = __pyx_vtabptr_10fatslimlib_9core_base_AtomGroup;
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_10fatslimlib_9core_base_9AtomGroup_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_10fatslimlib_9core_base_AtomGroup(PyObject *o) {
  struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *p = (struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_10fatslimlib_9core_base_9AtomGroup_7__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  if (p->__weakref__) PyObject_ClearWeakRefs(o);
  Py_CLEAR(p->name);
  (*Py_TYPE(o)->tp_free)(o);
}

static PyMethodDef __pyx_methods_10fatslimlib_9core_base_AtomGroup[] = {
  {"append", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_9AtomGroup_3append, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_9AtomGroup_9__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_9AtomGroup_11__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_AtomGroup = {
  __pyx_pw_10fatslimlib_9core_base_9AtomGroup_5__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_AtomGroup = {
  __pyx_pw_10fatslimlib_9core_base_9AtomGroup_5__len__, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_10fatslimlib_9core_base_AtomGroup = {
  PyVarObject_HEAD_INIT(0, 0)
  "fatslimlib.core_base.AtomGroup", /*tp_name*/
  sizeof(struct __pyx_obj_10fatslimlib_9core_base_AtomGroup), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10fatslimlib_9core_base_AtomGroup, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_AtomGroup, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_AtomGroup, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10fatslimlib_9core_base_AtomGroup, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10fatslimlib_9core_base_AtomGroup, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_10fatslimlib_9core_base_PBCBox __pyx_vtable_10fatslimlib_9core_base_PBCBox;

static PyObject *__pyx_tp_new_10fatslimlib_9core_base_PBCBox(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10fatslimlib_9core_base_PBCBox *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)o);
  p->__pyx_vtab = __pyx_vtabptr_10fatslimlib_9core_base_PBCBox;
  return o;
}

static void __pyx_tp_dealloc_10fatslimlib_9core_base_PBCBox(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_6PBCBox_box(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_6PBCBox_3box_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_6PBCBox_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_6PBCBox_6center_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_6PBCBox_bbox_center(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_6PBCBox_11bbox_center_1__get__(o);
}

static PyMethodDef __pyx_methods_10fatslimlib_9core_base_PBCBox[] = {
  {"update", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_6PBCBox_3update, METH_O, 0},
  {"to_gro", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_6PBCBox_5to_gro, METH_NOARGS, 0},
  {"dx", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10fatslimlib_9core_base_6PBCBox_9dx, METH_VARARGS|METH_KEYWORDS, 0},
  {"dx_leaflet", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10fatslimlib_9core_base_6PBCBox_11dx_leaflet, METH_VARARGS|METH_KEYWORDS, 0},
  {"distance2", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10fatslimlib_9core_base_6PBCBox_13distance2, METH_VARARGS|METH_KEYWORDS, 0},
  {"distance", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10fatslimlib_9core_base_6PBCBox_15distance, METH_VARARGS|METH_KEYWORDS, 0},
  {"leaflet_distance2", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10fatslimlib_9core_base_6PBCBox_17leaflet_distance2, METH_VARARGS|METH_KEYWORDS, 0},
  {"leaflet_distance", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10fatslimlib_9core_base_6PBCBox_19leaflet_distance, METH_VARARGS|METH_KEYWORDS, 0},
  {"get_xcm", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_6PBCBox_21get_xcm, METH_O, 0},
  {"asarray", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_6PBCBox_23asarray, METH_NOARGS, 0},
  {"put_atoms_in_bbox", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_6PBCBox_25put_atoms_in_bbox, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_6PBCBox_27__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_6PBCBox_29__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10fatslimlib_9core_base_PBCBox[] = {
  {(char *)"box", __pyx_getprop_10fatslimlib_9core_base_6PBCBox_box, 0, (char *)0, 0},
  {(char *)"center", __pyx_getprop_10fatslimlib_9core_base_6PBCBox_center, 0, (char *)0, 0},
  {(char *)"bbox_center", __pyx_getprop_10fatslimlib_9core_base_6PBCBox_bbox_center, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_10fatslimlib_9core_base_PBCBox = {
  PyVarObject_HEAD_INIT(0, 0)
  "fatslimlib.core_base.PBCBox", /*tp_name*/
  sizeof(struct __pyx_obj_10fatslimlib_9core_base_PBCBox), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10fatslimlib_9core_base_PBCBox, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_10fatslimlib_9core_base_6PBCBox_7__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10fatslimlib_9core_base_PBCBox, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10fatslimlib_9core_base_PBCBox, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_10fatslimlib_9core_base_6PBCBox_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10fatslimlib_9core_base_PBCBox, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_10fatslimlib_9core_base_TopologyGroup __pyx_vtable_10fatslimlib_9core_base_TopologyGroup;

static PyObject *__pyx_tp_new_10fatslimlib_9core_base_TopologyGroup(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *)o);
  p->__pyx_vtab = __pyx_vtabptr_10fatslimlib_9core_base_TopologyGroup;
  p->parent = ((struct __pyx_obj_10fatslimlib_9core_base_Topology *)Py_None); Py_INCREF(Py_None);
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->resname = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10fatslimlib_9core_base_TopologyGroup(PyObject *o) {
  struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *p = (struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_10fatslimlib_9core_base_13TopologyGroup_5__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->parent);
  Py_CLEAR(p->name);
  Py_CLEAR(p->resname);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_10fatslimlib_9core_base_TopologyGroup(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *p = (struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *)o;
  if (p->parent) {
    e = (*v)(((PyObject *)p->parent), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10fatslimlib_9core_base_TopologyGroup(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *p = (struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *)o;
  tmp = ((PyObject*)p->parent);
  p->parent = ((struct __pyx_obj_10fatslimlib_9core_base_Topology *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_13TopologyGroup_atomids(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_13TopologyGroup_7atomids_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_13TopologyGroup_residue_atomids(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_13TopologyGroup_15residue_atomids_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_13TopologyGroup_parent(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_13TopologyGroup_6parent_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_13TopologyGroup_resid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_13TopologyGroup_5resid_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_13TopologyGroup_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_13TopologyGroup_4size_1__get__(o);
}

static PyMethodDef __pyx_methods_10fatslimlib_9core_base_TopologyGroup[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_13TopologyGroup_9__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_13TopologyGroup_11__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10fatslimlib_9core_base_TopologyGroup[] = {
  {(char *)"atomids", __pyx_getprop_10fatslimlib_9core_base_13TopologyGroup_atomids, 0, (char *)0, 0},
  {(char *)"residue_atomids", __pyx_getprop_10fatslimlib_9core_base_13TopologyGroup_residue_atomids, 0, (char *)0, 0},
  {(char *)"parent", __pyx_getprop_10fatslimlib_9core_base_13TopologyGroup_parent, 0, (char *)0, 0},
  {(char *)"resid", __pyx_getprop_10fatslimlib_9core_base_13TopologyGroup_resid, 0, (char *)0, 0},
  {(char *)"size", __pyx_getprop_10fatslimlib_9core_base_13TopologyGroup_size, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_TopologyGroup = {
  __pyx_pw_10fatslimlib_9core_base_13TopologyGroup_3__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_TopologyGroup = {
  __pyx_pw_10fatslimlib_9core_base_13TopologyGroup_3__len__, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_10fatslimlib_9core_base_TopologyGroup = {
  PyVarObject_HEAD_INIT(0, 0)
  "fatslimlib.core_base.TopologyGroup", /*tp_name*/
  sizeof(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10fatslimlib_9core_base_TopologyGroup, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_TopologyGroup, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_TopologyGroup, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_10fatslimlib_9core_base_13TopologyGroup_7__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10fatslimlib_9core_base_TopologyGroup, /*tp_traverse*/
  __pyx_tp_clear_10fatslimlib_9core_base_TopologyGroup, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10fatslimlib_9core_base_TopologyGroup, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10fatslimlib_9core_base_TopologyGroup, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_10fatslimlib_9core_base_13TopologyGroup_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10fatslimlib_9core_base_TopologyGroup, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_10fatslimlib_9core_base_Topology __pyx_vtable_10fatslimlib_9core_base_Topology;

static PyObject *__pyx_tp_new_10fatslimlib_9core_base_Topology(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10fatslimlib_9core_base_Topology *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10fatslimlib_9core_base_Topology *)o);
  p->__pyx_vtab = __pyx_vtabptr_10fatslimlib_9core_base_Topology;
  return o;
}

static void __pyx_tp_dealloc_10fatslimlib_9core_base_Topology(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_10fatslimlib_9core_base_8Topology_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_8Topology_nresidues(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_8Topology_9nresidues_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_8Topology_natoms(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_8Topology_6natoms_1__get__(o);
}

static PyMethodDef __pyx_methods_10fatslimlib_9core_base_Topology[] = {
  {"append", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10fatslimlib_9core_base_8Topology_5append, METH_VARARGS|METH_KEYWORDS, 0},
  {"get_atom", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_8Topology_7get_atom, METH_O, 0},
  {"get_residue_from_atomid", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_8Topology_9get_residue_from_atomid, METH_O, 0},
  {"get_residue", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_8Topology_11get_residue, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_8Topology_15__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_8Topology_17__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10fatslimlib_9core_base_Topology[] = {
  {(char *)"nresidues", __pyx_getprop_10fatslimlib_9core_base_8Topology_nresidues, 0, (char *)0, 0},
  {(char *)"natoms", __pyx_getprop_10fatslimlib_9core_base_8Topology_natoms, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_Topology = {
  __pyx_pw_10fatslimlib_9core_base_8Topology_13__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Topology = {
  __pyx_pw_10fatslimlib_9core_base_8Topology_13__len__, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_10fatslimlib_9core_base_Topology = {
  PyVarObject_HEAD_INIT(0, 0)
  "fatslimlib.core_base.Topology", /*tp_name*/
  sizeof(struct __pyx_obj_10fatslimlib_9core_base_Topology), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10fatslimlib_9core_base_Topology, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Topology, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Topology, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10fatslimlib_9core_base_Topology, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10fatslimlib_9core_base_Topology, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_10fatslimlib_9core_base_8Topology_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10fatslimlib_9core_base_Topology, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_10fatslimlib_9core_base_TopologyReader __pyx_vtable_10fatslimlib_9core_base_TopologyReader;

static PyObject *__pyx_tp_new_10fatslimlib_9core_base_TopologyReader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *)o);
  p->__pyx_vtab = __pyx_vtabptr_10fatslimlib_9core_base_TopologyReader;
  p->filename = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->topology = ((struct __pyx_obj_10fatslimlib_9core_base_Topology *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10fatslimlib_9core_base_TopologyReader(PyObject *o) {
  struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *p = (struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->filename);
  Py_CLEAR(p->topology);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_10fatslimlib_9core_base_TopologyReader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *p = (struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *)o;
  if (p->topology) {
    e = (*v)(((PyObject *)p->topology), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10fatslimlib_9core_base_TopologyReader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *p = (struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *)o;
  tmp = ((PyObject*)p->topology);
  p->topology = ((struct __pyx_obj_10fatslimlib_9core_base_Topology *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_14TopologyReader_topology(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_14TopologyReader_8topology_1__get__(o);
}

static PyMethodDef __pyx_methods_10fatslimlib_9core_base_TopologyReader[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_14TopologyReader_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_14TopologyReader_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10fatslimlib_9core_base_TopologyReader[] = {
  {(char *)"topology", __pyx_getprop_10fatslimlib_9core_base_14TopologyReader_topology, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_10fatslimlib_9core_base_TopologyReader = {
  PyVarObject_HEAD_INIT(0, 0)
  "fatslimlib.core_base.TopologyReader", /*tp_name*/
  sizeof(struct __pyx_obj_10fatslimlib_9core_base_TopologyReader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10fatslimlib_9core_base_TopologyReader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10fatslimlib_9core_base_TopologyReader, /*tp_traverse*/
  __pyx_tp_clear_10fatslimlib_9core_base_TopologyReader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10fatslimlib_9core_base_TopologyReader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10fatslimlib_9core_base_TopologyReader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_10fatslimlib_9core_base_14TopologyReader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10fatslimlib_9core_base_TopologyReader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_10fatslimlib_9core_base_CoordinateReader __pyx_vtable_10fatslimlib_9core_base_CoordinateReader;

static PyObject *__pyx_tp_new_10fatslimlib_9core_base_CoordinateReader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *)o);
  p->__pyx_vtab = __pyx_vtabptr_10fatslimlib_9core_base_CoordinateReader;
  p->filename = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->coordinate_offsets.data = NULL;
  p->coordinate_offsets.memview = NULL;
  p->box_offsets.data = NULL;
  p->box_offsets.memview = NULL;
  p->timesteps.data = NULL;
  p->timesteps.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_10fatslimlib_9core_base_CoordinateReader(PyObject *o) {
  struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *p = (struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  Py_CLEAR(p->filename);
  __PYX_XDEC_MEMVIEW(&p->coordinate_offsets, 1);
  __PYX_XDEC_MEMVIEW(&p->box_offsets, 1);
  __PYX_XDEC_MEMVIEW(&p->timesteps, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_16CoordinateReader_nframes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_16CoordinateReader_7nframes_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_16CoordinateReader_timesteps(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_16CoordinateReader_9timesteps_1__get__(o);
}

static PyMethodDef __pyx_methods_10fatslimlib_9core_base_CoordinateReader[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_16CoordinateReader_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_16CoordinateReader_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10fatslimlib_9core_base_CoordinateReader[] = {
  {(char *)"nframes", __pyx_getprop_10fatslimlib_9core_base_16CoordinateReader_nframes, 0, (char *)0, 0},
  {(char *)"timesteps", __pyx_getprop_10fatslimlib_9core_base_16CoordinateReader_timesteps, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_10fatslimlib_9core_base_CoordinateReader = {
  PyVarObject_HEAD_INIT(0, 0)
  "fatslimlib.core_base.CoordinateReader", /*tp_name*/
  sizeof(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10fatslimlib_9core_base_CoordinateReader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10fatslimlib_9core_base_CoordinateReader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10fatslimlib_9core_base_CoordinateReader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_10fatslimlib_9core_base_16CoordinateReader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10fatslimlib_9core_base_CoordinateReader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_10fatslimlib_9core_base_IndexReader __pyx_vtable_10fatslimlib_9core_base_IndexReader;

static PyObject *__pyx_tp_new_10fatslimlib_9core_base_IndexReader(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10fatslimlib_9core_base_IndexReader *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10fatslimlib_9core_base_IndexReader *)o);
  p->__pyx_vtab = __pyx_vtabptr_10fatslimlib_9core_base_IndexReader;
  p->filename = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->groups = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10fatslimlib_9core_base_IndexReader(PyObject *o) {
  struct __pyx_obj_10fatslimlib_9core_base_IndexReader *p = (struct __pyx_obj_10fatslimlib_9core_base_IndexReader *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->filename);
  Py_CLEAR(p->groups);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_10fatslimlib_9core_base_IndexReader(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10fatslimlib_9core_base_IndexReader *p = (struct __pyx_obj_10fatslimlib_9core_base_IndexReader *)o;
  if (p->groups) {
    e = (*v)(p->groups, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10fatslimlib_9core_base_IndexReader(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10fatslimlib_9core_base_IndexReader *p = (struct __pyx_obj_10fatslimlib_9core_base_IndexReader *)o;
  tmp = ((PyObject*)p->groups);
  p->groups = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_10fatslimlib_9core_base_IndexReader(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_11IndexReader_groups(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_11IndexReader_6groups_1__get__(o);
}

static PyMethodDef __pyx_methods_10fatslimlib_9core_base_IndexReader[] = {
  {"load", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_11IndexReader_3load, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_11IndexReader_9__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_11IndexReader_11__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10fatslimlib_9core_base_IndexReader[] = {
  {(char *)"groups", __pyx_getprop_10fatslimlib_9core_base_11IndexReader_groups, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_IndexReader = {
  __pyx_pw_10fatslimlib_9core_base_11IndexReader_5__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_10fatslimlib_9core_base_IndexReader, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_IndexReader = {
  __pyx_pw_10fatslimlib_9core_base_11IndexReader_5__len__, /*mp_length*/
  __pyx_pw_10fatslimlib_9core_base_11IndexReader_7__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_10fatslimlib_9core_base_IndexReader = {
  PyVarObject_HEAD_INIT(0, 0)
  "fatslimlib.core_base.IndexReader", /*tp_name*/
  sizeof(struct __pyx_obj_10fatslimlib_9core_base_IndexReader), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10fatslimlib_9core_base_IndexReader, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_IndexReader, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_IndexReader, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10fatslimlib_9core_base_IndexReader, /*tp_traverse*/
  __pyx_tp_clear_10fatslimlib_9core_base_IndexReader, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10fatslimlib_9core_base_IndexReader, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10fatslimlib_9core_base_IndexReader, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_10fatslimlib_9core_base_11IndexReader_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10fatslimlib_9core_base_IndexReader, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_10fatslimlib_9core_base_Frame __pyx_vtable_10fatslimlib_9core_base_Frame;

static PyObject *__pyx_tp_new_10fatslimlib_9core_base_Frame(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10fatslimlib_9core_base_Frame *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10fatslimlib_9core_base_Frame *)o);
  p->__pyx_vtab = __pyx_vtabptr_10fatslimlib_9core_base_Frame;
  p->trajectory = ((struct __pyx_obj_10fatslimlib_9core_base_Trajectory *)Py_None); Py_INCREF(Py_None);
  p->topology = ((struct __pyx_obj_10fatslimlib_9core_base_Topology *)Py_None); Py_INCREF(Py_None);
  p->coords_reader = ((struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *)Py_None); Py_INCREF(Py_None);
  p->box = ((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)Py_None); Py_INCREF(Py_None);
  p->lipid_coords_aslist = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->lipids_coords_bbox_aslist = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->aggregates = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->membranes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->hg_group_coords.data = NULL;
  p->hg_group_coords.memview = NULL;
  p->hg_coords_bbox.data = NULL;
  p->hg_coords_bbox.memview = NULL;
  p->lipid_coords.data = NULL;
  p->lipid_coords.memview = NULL;
  p->lipid_coords_bbox.data = NULL;
  p->lipid_coords_bbox.memview = NULL;
  p->lipid_atomids_offsets.data = NULL;
  p->lipid_atomids_offsets.memview = NULL;
  p->bead_coords.data = NULL;
  p->bead_coords.memview = NULL;
  p->bead_coords_bbox.data = NULL;
  p->bead_coords_bbox.memview = NULL;
  p->interacting_group_coords.data = NULL;
  p->interacting_group_coords.memview = NULL;
  p->interacting_group_coords_bbox.data = NULL;
  p->interacting_group_coords_bbox.memview = NULL;
  p->directions.data = NULL;
  p->directions.memview = NULL;
  p->normals.data = NULL;
  p->normals.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_10fatslimlib_9core_base_Frame(PyObject *o) {
  struct __pyx_obj_10fatslimlib_9core_base_Frame *p = (struct __pyx_obj_10fatslimlib_9core_base_Frame *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_10fatslimlib_9core_base_5Frame_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->trajectory);
  Py_CLEAR(p->topology);
  Py_CLEAR(p->coords_reader);
  Py_CLEAR(p->box);
  Py_CLEAR(p->lipid_coords_aslist);
  Py_CLEAR(p->lipids_coords_bbox_aslist);
  Py_CLEAR(p->aggregates);
  Py_CLEAR(p->membranes);
  __PYX_XDEC_MEMVIEW(&p->hg_group_coords, 1);
  __PYX_XDEC_MEMVIEW(&p->hg_coords_bbox, 1);
  __PYX_XDEC_MEMVIEW(&p->lipid_coords, 1);
  __PYX_XDEC_MEMVIEW(&p->lipid_coords_bbox, 1);
  __PYX_XDEC_MEMVIEW(&p->lipid_atomids_offsets, 1);
  __PYX_XDEC_MEMVIEW(&p->bead_coords, 1);
  __PYX_XDEC_MEMVIEW(&p->bead_coords_bbox, 1);
  __PYX_XDEC_MEMVIEW(&p->interacting_group_coords, 1);
  __PYX_XDEC_MEMVIEW(&p->interacting_group_coords_bbox, 1);
  __PYX_XDEC_MEMVIEW(&p->directions, 1);
  __PYX_XDEC_MEMVIEW(&p->normals, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_10fatslimlib_9core_base_Frame(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10fatslimlib_9core_base_Frame *p = (struct __pyx_obj_10fatslimlib_9core_base_Frame *)o;
  if (p->trajectory) {
    e = (*v)(((PyObject *)p->trajectory), a); if (e) return e;
  }
  if (p->topology) {
    e = (*v)(((PyObject *)p->topology), a); if (e) return e;
  }
  if (p->coords_reader) {
    e = (*v)(((PyObject *)p->coords_reader), a); if (e) return e;
  }
  if (p->box) {
    e = (*v)(((PyObject *)p->box), a); if (e) return e;
  }
  if (p->lipid_coords_aslist) {
    e = (*v)(p->lipid_coords_aslist, a); if (e) return e;
  }
  if (p->lipids_coords_bbox_aslist) {
    e = (*v)(p->lipids_coords_bbox_aslist, a); if (e) return e;
  }
  if (p->aggregates) {
    e = (*v)(p->aggregates, a); if (e) return e;
  }
  if (p->membranes) {
    e = (*v)(p->membranes, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10fatslimlib_9core_base_Frame(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10fatslimlib_9core_base_Frame *p = (struct __pyx_obj_10fatslimlib_9core_base_Frame *)o;
  tmp = ((PyObject*)p->trajectory);
  p->trajectory = ((struct __pyx_obj_10fatslimlib_9core_base_Trajectory *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->topology);
  p->topology = ((struct __pyx_obj_10fatslimlib_9core_base_Topology *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->coords_reader);
  p->coords_reader = ((struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->box);
  p->box = ((struct __pyx_obj_10fatslimlib_9core_base_PBCBox *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->lipid_coords_aslist);
  p->lipid_coords_aslist = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->lipids_coords_bbox_aslist);
  p->lipids_coords_bbox_aslist = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->aggregates);
  p->aggregates = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->membranes);
  p->membranes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_5Frame_hg_group_coords(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_5Frame_15hg_group_coords_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_5Frame_hg_group_coords_bbox(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_5Frame_20hg_group_coords_bbox_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_5Frame_lipid_coords(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_5Frame_12lipid_coords_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_5Frame_lipid_coords_bbox(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_5Frame_17lipid_coords_bbox_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_5Frame_interacting_group_coords(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_5Frame_24interacting_group_coords_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_5Frame_interacting_group_coords_bbox(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_5Frame_29interacting_group_coords_bbox_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_5Frame_bead_coords(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_5Frame_11bead_coords_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_5Frame_bead_coords_bbox(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_5Frame_16bead_coords_bbox_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_5Frame_directions(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_5Frame_10directions_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_5Frame_normals(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_5Frame_7normals_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_5Frame_hg_atomids(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_5Frame_10hg_atomids_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_5Frame_lipid_atomids(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_5Frame_13lipid_atomids_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_5Frame_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_5Frame_5index_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_5Frame_timestep(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_5Frame_8timestep_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_5Frame_trajectory(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_5Frame_10trajectory_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_5Frame_topology(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_5Frame_8topology_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_5Frame_box(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_5Frame_3box_1__get__(o);
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_5Frame_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_5Frame_4size_1__get__(o);
}

static PyMethodDef __pyx_methods_10fatslimlib_9core_base_Frame[] = {
  {"get_forcefield_type", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_5Frame_7get_forcefield_type, METH_NOARGS, 0},
  {"get_aggregates", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10fatslimlib_9core_base_5Frame_9get_aggregates, METH_VARARGS|METH_KEYWORDS, 0},
  {"get_membranes", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10fatslimlib_9core_base_5Frame_11get_membranes, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_5Frame_13__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_5Frame_15__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10fatslimlib_9core_base_Frame[] = {
  {(char *)"hg_group_coords", __pyx_getprop_10fatslimlib_9core_base_5Frame_hg_group_coords, 0, (char *)0, 0},
  {(char *)"hg_group_coords_bbox", __pyx_getprop_10fatslimlib_9core_base_5Frame_hg_group_coords_bbox, 0, (char *)0, 0},
  {(char *)"lipid_coords", __pyx_getprop_10fatslimlib_9core_base_5Frame_lipid_coords, 0, (char *)0, 0},
  {(char *)"lipid_coords_bbox", __pyx_getprop_10fatslimlib_9core_base_5Frame_lipid_coords_bbox, 0, (char *)0, 0},
  {(char *)"interacting_group_coords", __pyx_getprop_10fatslimlib_9core_base_5Frame_interacting_group_coords, 0, (char *)0, 0},
  {(char *)"interacting_group_coords_bbox", __pyx_getprop_10fatslimlib_9core_base_5Frame_interacting_group_coords_bbox, 0, (char *)0, 0},
  {(char *)"bead_coords", __pyx_getprop_10fatslimlib_9core_base_5Frame_bead_coords, 0, (char *)0, 0},
  {(char *)"bead_coords_bbox", __pyx_getprop_10fatslimlib_9core_base_5Frame_bead_coords_bbox, 0, (char *)0, 0},
  {(char *)"directions", __pyx_getprop_10fatslimlib_9core_base_5Frame_directions, 0, (char *)0, 0},
  {(char *)"normals", __pyx_getprop_10fatslimlib_9core_base_5Frame_normals, 0, (char *)0, 0},
  {(char *)"hg_atomids", __pyx_getprop_10fatslimlib_9core_base_5Frame_hg_atomids, 0, (char *)0, 0},
  {(char *)"lipid_atomids", __pyx_getprop_10fatslimlib_9core_base_5Frame_lipid_atomids, 0, (char *)0, 0},
  {(char *)"index", __pyx_getprop_10fatslimlib_9core_base_5Frame_index, 0, (char *)0, 0},
  {(char *)"timestep", __pyx_getprop_10fatslimlib_9core_base_5Frame_timestep, 0, (char *)0, 0},
  {(char *)"trajectory", __pyx_getprop_10fatslimlib_9core_base_5Frame_trajectory, 0, (char *)0, 0},
  {(char *)"topology", __pyx_getprop_10fatslimlib_9core_base_5Frame_topology, 0, (char *)0, 0},
  {(char *)"box", __pyx_getprop_10fatslimlib_9core_base_5Frame_box, 0, (char *)0, 0},
  {(char *)"size", __pyx_getprop_10fatslimlib_9core_base_5Frame_size, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_Frame = {
  __pyx_pw_10fatslimlib_9core_base_5Frame_5__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Frame = {
  __pyx_pw_10fatslimlib_9core_base_5Frame_5__len__, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_10fatslimlib_9core_base_Frame = {
  PyVarObject_HEAD_INIT(0, 0)
  "fatslimlib.core_base.Frame", /*tp_name*/
  sizeof(struct __pyx_obj_10fatslimlib_9core_base_Frame), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10fatslimlib_9core_base_Frame, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Frame, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Frame, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10fatslimlib_9core_base_Frame, /*tp_traverse*/
  __pyx_tp_clear_10fatslimlib_9core_base_Frame, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10fatslimlib_9core_base_Frame, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10fatslimlib_9core_base_Frame, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_10fatslimlib_9core_base_5Frame_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10fatslimlib_9core_base_Frame, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_10fatslimlib_9core_base_Trajectory __pyx_vtable_10fatslimlib_9core_base_Trajectory;

static PyObject *__pyx_tp_new_10fatslimlib_9core_base_Trajectory(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10fatslimlib_9core_base_Trajectory *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10fatslimlib_9core_base_Trajectory *)o);
  p->__pyx_vtab = __pyx_vtabptr_10fatslimlib_9core_base_Trajectory;
  p->bead_group_name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->interacting_group_name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->topology = ((struct __pyx_obj_10fatslimlib_9core_base_Topology *)Py_None); Py_INCREF(Py_None);
  p->topol_reader = ((struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *)Py_None); Py_INCREF(Py_None);
  p->index_reader = ((struct __pyx_obj_10fatslimlib_9core_base_IndexReader *)Py_None); Py_INCREF(Py_None);
  p->coords_reader = ((struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *)Py_None); Py_INCREF(Py_None);
  p->timesteps = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->membranes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->aggregates = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->hg_group_atomids.data = NULL;
  p->hg_group_atomids.memview = NULL;
  p->interacting_atomids.data = NULL;
  p->interacting_atomids.memview = NULL;
  p->lipid_atomids.data = NULL;
  p->lipid_atomids.memview = NULL;
  p->hg_bead_atomids.data = NULL;
  p->hg_bead_atomids.memview = NULL;
  p->lipid_atomids_offsets.data = NULL;
  p->lipid_atomids_offsets.memview = NULL;
  p->hg_bead_atomids_offsets.data = NULL;
  p->hg_bead_atomids_offsets.memview = NULL;
  p->lipid_hg_indices.data = NULL;
  p->lipid_hg_indices.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc_10fatslimlib_9core_base_Trajectory(PyObject *o) {
  struct __pyx_obj_10fatslimlib_9core_base_Trajectory *p = (struct __pyx_obj_10fatslimlib_9core_base_Trajectory *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->bead_group_name);
  Py_CLEAR(p->interacting_group_name);
  Py_CLEAR(p->topology);
  Py_CLEAR(p->topol_reader);
  Py_CLEAR(p->index_reader);
  Py_CLEAR(p->coords_reader);
  Py_CLEAR(p->timesteps);
  Py_CLEAR(p->membranes);
  Py_CLEAR(p->aggregates);
  __PYX_XDEC_MEMVIEW(&p->hg_group_atomids, 1);
  __PYX_XDEC_MEMVIEW(&p->interacting_atomids, 1);
  __PYX_XDEC_MEMVIEW(&p->lipid_atomids, 1);
  __PYX_XDEC_MEMVIEW(&p->hg_bead_atomids, 1);
  __PYX_XDEC_MEMVIEW(&p->lipid_atomids_offsets, 1);
  __PYX_XDEC_MEMVIEW(&p->hg_bead_atomids_offsets, 1);
  __PYX_XDEC_MEMVIEW(&p->lipid_hg_indices, 1);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_10fatslimlib_9core_base_Trajectory(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10fatslimlib_9core_base_Trajectory *p = (struct __pyx_obj_10fatslimlib_9core_base_Trajectory *)o;
  if (p->topology) {
    e = (*v)(((PyObject *)p->topology), a); if (e) return e;
  }
  if (p->topol_reader) {
    e = (*v)(((PyObject *)p->topol_reader), a); if (e) return e;
  }
  if (p->index_reader) {
    e = (*v)(((PyObject *)p->index_reader), a); if (e) return e;
  }
  if (p->coords_reader) {
    e = (*v)(((PyObject *)p->coords_reader), a); if (e) return e;
  }
  if (p->timesteps) {
    e = (*v)(p->timesteps, a); if (e) return e;
  }
  if (p->membranes) {
    e = (*v)(p->membranes, a); if (e) return e;
  }
  if (p->aggregates) {
    e = (*v)(p->aggregates, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10fatslimlib_9core_base_Trajectory(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10fatslimlib_9core_base_Trajectory *p = (struct __pyx_obj_10fatslimlib_9core_base_Trajectory *)o;
  tmp = ((PyObject*)p->topology);
  p->topology = ((struct __pyx_obj_10fatslimlib_9core_base_Topology *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->topol_reader);
  p->topol_reader = ((struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->index_reader);
  p->index_reader = ((struct __pyx_obj_10fatslimlib_9core_base_IndexReader *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->coords_reader);
  p->coords_reader = ((struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->timesteps);
  p->timesteps = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->membranes);
  p->membranes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->aggregates);
  p->aggregates = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_10fatslimlib_9core_base_Trajectory(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_getprop_10fatslimlib_9core_base_10Trajectory_timesteps(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10fatslimlib_9core_base_10Trajectory_9timesteps_1__get__(o);
}

static PyMethodDef __pyx_methods_10fatslimlib_9core_base_Trajectory[] = {
  {"initialize", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10fatslimlib_9core_base_10Trajectory_3initialize, METH_VARARGS|METH_KEYWORDS, 0},
  {"get_forcefield_type", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_10Trajectory_5get_forcefield_type, METH_NOARGS, 0},
  {"get_rvdw", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_10Trajectory_7get_rvdw, METH_NOARGS, 0},
  {"get_topology", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_10Trajectory_13get_topology, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_10Trajectory_15__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_10Trajectory_17__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10fatslimlib_9core_base_Trajectory[] = {
  {(char *)"timesteps", __pyx_getprop_10fatslimlib_9core_base_10Trajectory_timesteps, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_Trajectory = {
  __pyx_pw_10fatslimlib_9core_base_10Trajectory_11__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_10fatslimlib_9core_base_Trajectory, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Trajectory = {
  __pyx_pw_10fatslimlib_9core_base_10Trajectory_11__len__, /*mp_length*/
  __pyx_pw_10fatslimlib_9core_base_10Trajectory_9__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_10fatslimlib_9core_base_Trajectory = {
  PyVarObject_HEAD_INIT(0, 0)
  "fatslimlib.core_base.Trajectory", /*tp_name*/
  sizeof(struct __pyx_obj_10fatslimlib_9core_base_Trajectory), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10fatslimlib_9core_base_Trajectory, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Trajectory, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Trajectory, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10fatslimlib_9core_base_Trajectory, /*tp_traverse*/
  __pyx_tp_clear_10fatslimlib_9core_base_Trajectory, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10fatslimlib_9core_base_Trajectory, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10fatslimlib_9core_base_Trajectory, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_10fatslimlib_9core_base_10Trajectory_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10fatslimlib_9core_base_Trajectory, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_array __pyx_vtable_array;

static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_array_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_array_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_array;
  p->mode = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_format = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_array___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_array(PyObject *o) {
  struct __pyx_array_obj *p = (struct __pyx_array_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_array___dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->mode);
  Py_CLEAR(p->_format);
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_array(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_array(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_array___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_tp_getattro_array(PyObject *o, PyObject *n) {
  PyObject *v = __Pyx_PyObject_GenericGetAttr(o, n);
  if (!v && PyErr_ExceptionMatches(PyExc_AttributeError)) {
    PyErr_Clear();
    v = __pyx_array___getattr__(o, n);
  }
  return v;
}

static PyObject *__pyx_getprop___pyx_array_memview(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(o);
}

static PyMethodDef __pyx_methods_array[] = {
  {"__getattr__", (PyCFunction)__pyx_array___getattr__, METH_O|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_array_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_array_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_array[] = {
  {(char *)"memview", __pyx_getprop___pyx_array_memview, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_array = {
  __pyx_array___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_array, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_array = {
  __pyx_array___len__, /*mp_length*/
  __pyx_array___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_array, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_array = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_array_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_array = {
  PyVarObject_HEAD_INIT(0, 0)
  "fatslimlib.core_base.array", /*tp_name*/
  sizeof(struct __pyx_array_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_array, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_array, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_array, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  __pyx_tp_getattro_array, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_array, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_array, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_array, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_array, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_MemviewEnum_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_MemviewEnum_obj *)o);
  p->name = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_Enum(PyObject *o) {
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_Enum(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  if (p->name) {
    e = (*v)(p->name, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_Enum(PyObject *o) {
  PyObject* tmp;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  tmp = ((PyObject*)p->name);
  p->name = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_Enum[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type___pyx_MemviewEnum = {
  PyVarObject_HEAD_INIT(0, 0)
  "fatslimlib.core_base.Enum", /*tp_name*/
  sizeof(struct __pyx_MemviewEnum_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_Enum, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_MemviewEnum___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_Enum, /*tp_traverse*/
  __pyx_tp_clear_Enum, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_Enum, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_MemviewEnum___init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_Enum, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_memoryview __pyx_vtable_memoryview;

static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryview_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryview_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_memoryview;
  p->obj = Py_None; Py_INCREF(Py_None);
  p->_size = Py_None; Py_INCREF(Py_None);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  p->view.obj = NULL;
  if (unlikely(__pyx_memoryview___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_memoryview(PyObject *o) {
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_memoryview___dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->obj);
  Py_CLEAR(p->_size);
  Py_CLEAR(p->_array_interface);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_memoryview(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  if (p->obj) {
    e = (*v)(p->obj, a); if (e) return e;
  }
  if (p->_size) {
    e = (*v)(p->_size, a); if (e) return e;
  }
  if (p->_array_interface) {
    e = (*v)(p->_array_interface, a); if (e) return e;
  }
  if (p->view.obj) {
    e = (*v)(p->view.obj, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_memoryview(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  tmp = ((PyObject*)p->obj);
  p->obj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_size);
  p->_size = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_array_interface);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  Py_CLEAR(p->view.obj);
  return 0;
}
static PyObject *__pyx_sq_item_memoryview(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_memoryview(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_memoryview___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_getprop___pyx_memoryview_T(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_shape(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_strides(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_suboffsets(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_ndim(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_itemsize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_nbytes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(o);
}

static PyMethodDef __pyx_methods_memoryview[] = {
  {"is_c_contig", (PyCFunction)__pyx_memoryview_is_c_contig, METH_NOARGS, 0},
  {"is_f_contig", (PyCFunction)__pyx_memoryview_is_f_contig, METH_NOARGS, 0},
  {"copy", (PyCFunction)__pyx_memoryview_copy, METH_NOARGS, 0},
  {"copy_fortran", (PyCFunction)__pyx_memoryview_copy_fortran, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_memoryview[] = {
  {(char *)"T", __pyx_getprop___pyx_memoryview_T, 0, (char *)0, 0},
  {(char *)"base", __pyx_getprop___pyx_memoryview_base, 0, (char *)0, 0},
  {(char *)"shape", __pyx_getprop___pyx_memoryview_shape, 0, (char *)0, 0},
  {(char *)"strides", __pyx_getprop___pyx_memoryview_strides, 0, (char *)0, 0},
  {(char *)"suboffsets", __pyx_getprop___pyx_memoryview_suboffsets, 0, (char *)0, 0},
  {(char *)"ndim", __pyx_getprop___pyx_memoryview_ndim, 0, (char *)0, 0},
  {(char *)"itemsize", __pyx_getprop___pyx_memoryview_itemsize, 0, (char *)0, 0},
  {(char *)"nbytes", __pyx_getprop___pyx_memoryview_nbytes, 0, (char *)0, 0},
  {(char *)"size", __pyx_getprop___pyx_memoryview_size, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_memoryview = {
  __pyx_memoryview___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_memoryview, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_memoryview = {
  __pyx_memoryview___len__, /*mp_length*/
  __pyx_memoryview___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_memoryview, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_memoryview = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_memoryview_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_memoryview = {
  PyVarObject_HEAD_INIT(0, 0)
  "fatslimlib.core_base.memoryview", /*tp_name*/
  sizeof(struct __pyx_memoryview_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_memoryview, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_memoryview___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_memoryview, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_memoryview, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_memoryview___str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_memoryview, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_memoryview, /*tp_traverse*/
  __pyx_tp_clear_memoryview, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_memoryview, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_memoryview, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_memoryview, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct__memoryviewslice __pyx_vtable__memoryviewslice;

static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryviewslice_obj *p;
  PyObject *o = __pyx_tp_new_memoryview(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryviewslice_obj *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_memoryview*)__pyx_vtabptr__memoryviewslice;
  p->from_object = Py_None; Py_INCREF(Py_None);
  p->from_slice.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc__memoryviewslice(PyObject *o) {
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_memoryviewslice___dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->from_object);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_memoryview(o);
}

static int __pyx_tp_traverse__memoryviewslice(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  e = __pyx_tp_traverse_memoryview(o, v, a); if (e) return e;
  if (p->from_object) {
    e = (*v)(p->from_object, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear__memoryviewslice(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  __pyx_tp_clear_memoryview(o);
  tmp = ((PyObject*)p->from_object);
  p->from_object = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  __PYX_XDEC_MEMVIEW(&p->from_slice, 1);
  return 0;
}

static PyObject *__pyx_getprop___pyx_memoryviewslice_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(o);
}

static PyMethodDef __pyx_methods__memoryviewslice[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets__memoryviewslice[] = {
  {(char *)"base", __pyx_getprop___pyx_memoryviewslice_base, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type___pyx_memoryviewslice = {
  PyVarObject_HEAD_INIT(0, 0)
  "fatslimlib.core_base._memoryviewslice", /*tp_name*/
  sizeof(struct __pyx_memoryviewslice_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc__memoryviewslice, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Internal class for passing memoryview slices to Python", /*tp_doc*/
  __pyx_tp_traverse__memoryviewslice, /*tp_traverse*/
  __pyx_tp_clear__memoryviewslice, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods__memoryviewslice, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets__memoryviewslice, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new__memoryviewslice, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"get_num_threads", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_1get_num_threads, METH_NOARGS, 0},
  {"get_max_threads", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_3get_max_threads, METH_NOARGS, 0},
  {"get_num_procs", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_5get_num_procs, METH_NOARGS, 0},
  {"set_num_threads", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_7set_num_threads, METH_O, 0},
  {"pretty_delta", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10fatslimlib_9core_base_9pretty_delta, METH_VARARGS|METH_KEYWORDS, 0},
  {"pretty_duration", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_11pretty_duration, METH_O, 0},
  {"get_memory_usage", (PyCFunction)__pyx_pw_10fatslimlib_9core_base_13get_memory_usage, METH_NOARGS, 0},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_core_base(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_core_base},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "core_base",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_0_29_14, __pyx_k_0_29_14, sizeof(__pyx_k_0_29_14), 0, 0, 1, 0},
  {&__pyx_kp_u_0f, __pyx_k_0f, sizeof(__pyx_k_0f), 0, 1, 0, 0},
  {&__pyx_kp_u_5f, __pyx_k_5f, sizeof(__pyx_k_5f), 0, 1, 0, 0},
  {&__pyx_n_s_ASCII, __pyx_k_ASCII, sizeof(__pyx_k_ASCII), 0, 0, 1, 1},
  {&__pyx_n_s_Atom, __pyx_k_Atom, sizeof(__pyx_k_Atom), 0, 0, 1, 1},
  {&__pyx_n_s_AtomGroup, __pyx_k_AtomGroup, sizeof(__pyx_k_AtomGroup), 0, 0, 1, 1},
  {&__pyx_kp_u_Atom_with_atomid_i_is_already_re, __pyx_k_Atom_with_atomid_i_is_already_re, sizeof(__pyx_k_Atom_with_atomid_i_is_already_re), 0, 1, 0, 0},
  {&__pyx_kp_u_Atomid_i_internal_id_i, __pyx_k_Atomid_i_internal_id_i, sizeof(__pyx_k_Atomid_i_internal_id_i), 0, 1, 0, 0},
  {&__pyx_kp_u_Atomid_not_registered_in_topolog, __pyx_k_Atomid_not_registered_in_topolog, sizeof(__pyx_k_Atomid_not_registered_in_topolog), 0, 1, 0, 0},
  {&__pyx_kp_u_Box_does_not_correspond_to_PBC_x, __pyx_k_Box_does_not_correspond_to_PBC_x, sizeof(__pyx_k_Box_does_not_correspond_to_PBC_x), 0, 1, 0, 0},
  {&__pyx_kp_u_Box_must_be_a_i_x_i_matrix_shape, __pyx_k_Box_must_be_a_i_x_i_matrix_shape, sizeof(__pyx_k_Box_must_be_a_i_x_i_matrix_shape), 0, 1, 0, 0},
  {&__pyx_kp_s_Buffer_view_does_not_expose_stri, __pyx_k_Buffer_view_does_not_expose_stri, sizeof(__pyx_k_Buffer_view_does_not_expose_stri), 0, 0, 1, 0},
  {&__pyx_kp_s_Can_only_create_a_buffer_that_is, __pyx_k_Can_only_create_a_buffer_that_is, sizeof(__pyx_k_Can_only_create_a_buffer_that_is), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_assign_to_read_only_memor, __pyx_k_Cannot_assign_to_read_only_memor, sizeof(__pyx_k_Cannot_assign_to_read_only_memor), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_create_writable_memory_vi, __pyx_k_Cannot_create_writable_memory_vi, sizeof(__pyx_k_Cannot_create_writable_memory_vi), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_index_with_type_s, __pyx_k_Cannot_index_with_type_s, sizeof(__pyx_k_Cannot_index_with_type_s), 0, 0, 1, 0},
  {&__pyx_n_s_CoordinateReader, __pyx_k_CoordinateReader, sizeof(__pyx_k_CoordinateReader), 0, 0, 1, 1},
  {&__pyx_kp_u_Could_not_change_number_of_OpenM, __pyx_k_Could_not_change_number_of_OpenM, sizeof(__pyx_k_Could_not_change_number_of_OpenM), 0, 1, 0, 0},
  {&__pyx_kp_u_Done_in_s, __pyx_k_Done_in_s, sizeof(__pyx_k_Done_in_s), 0, 1, 0, 0},
  {&__pyx_n_s_Ellipsis, __pyx_k_Ellipsis, sizeof(__pyx_k_Ellipsis), 0, 0, 1, 1},
  {&__pyx_kp_s_Empty_shape_tuple_for_cython_arr, __pyx_k_Empty_shape_tuple_for_cython_arr, sizeof(__pyx_k_Empty_shape_tuple_for_cython_arr), 0, 0, 1, 0},
  {&__pyx_n_s_FORCEFIELDTYPE_ALLATOM, __pyx_k_FORCEFIELDTYPE_ALLATOM, sizeof(__pyx_k_FORCEFIELDTYPE_ALLATOM), 0, 0, 1, 1},
  {&__pyx_n_s_FORCEFIELDTYPE_COARSE, __pyx_k_FORCEFIELDTYPE_COARSE, sizeof(__pyx_k_FORCEFIELDTYPE_COARSE), 0, 0, 1, 1},
  {&__pyx_n_s_FORCEFIELDTYPE_NOTSET, __pyx_k_FORCEFIELDTYPE_NOTSET, sizeof(__pyx_k_FORCEFIELDTYPE_NOTSET), 0, 0, 1, 1},
  {&__pyx_n_s_FORCEFIELDTYPE_UNIFIED, __pyx_k_FORCEFIELDTYPE_UNIFIED, sizeof(__pyx_k_FORCEFIELDTYPE_UNIFIED), 0, 0, 1, 1},
  {&__pyx_n_s_FORCEFIELDTYPE_UNKNOWN, __pyx_k_FORCEFIELDTYPE_UNKNOWN, sizeof(__pyx_k_FORCEFIELDTYPE_UNKNOWN), 0, 0, 1, 1},
  {&__pyx_n_s_Frame, __pyx_k_Frame, sizeof(__pyx_k_Frame), 0, 0, 1, 1},
  {&__pyx_kp_u_Frame_index_error_i_i_frames, __pyx_k_Frame_index_error_i_i_frames, sizeof(__pyx_k_Frame_index_error_i_i_frames), 0, 1, 0, 0},
  {&__pyx_kp_u_Frame_index_out_of_range_i_frame, __pyx_k_Frame_index_out_of_range_i_frame, sizeof(__pyx_k_Frame_index_out_of_range_i_frame), 0, 1, 0, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x50, __pyx_k_Incompatible_checksums_s_vs_0x50, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x50), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x6a, __pyx_k_Incompatible_checksums_s_vs_0x6a, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x6a), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x7a, __pyx_k_Incompatible_checksums_s_vs_0x7a, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x7a), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x8d, __pyx_k_Incompatible_checksums_s_vs_0x8d, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x8d), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xb0, __pyx_k_Incompatible_checksums_s_vs_0xb0, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xb0), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_n_s_IndexReader, __pyx_k_IndexReader, sizeof(__pyx_k_IndexReader), 0, 0, 1, 1},
  {&__pyx_kp_s_Indirect_dimensions_not_supporte, __pyx_k_Indirect_dimensions_not_supporte, sizeof(__pyx_k_Indirect_dimensions_not_supporte), 0, 0, 1, 0},
  {&__pyx_kp_u_Initializing_trajectory_using_gr, __pyx_k_Initializing_trajectory_using_gr, sizeof(__pyx_k_Initializing_trajectory_using_gr), 0, 1, 0, 0},
  {&__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_k_Invalid_mode_expected_c_or_fortr, sizeof(__pyx_k_Invalid_mode_expected_c_or_fortr), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_k_Invalid_shape_in_axis_d_d, sizeof(__pyx_k_Invalid_shape_in_axis_d_d), 0, 0, 1, 0},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_kp_u_Loading_groups_from_s, __pyx_k_Loading_groups_from_s, sizeof(__pyx_k_Loading_groups_from_s), 0, 1, 0, 0},
  {&__pyx_kp_u_Loading_topology_from_s, __pyx_k_Loading_topology_from_s, sizeof(__pyx_k_Loading_topology_from_s), 0, 1, 0, 0},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_k_MemoryView_of_r_at_0x_x, sizeof(__pyx_k_MemoryView_of_r_at_0x_x), 0, 0, 1, 0},
  {&__pyx_kp_s_MemoryView_of_r_object, __pyx_k_MemoryView_of_r_object, sizeof(__pyx_k_MemoryView_of_r_object), 0, 0, 1, 0},
  {&__pyx_kp_u_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 1, 0, 0},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_n_b_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 0, 1},
  {&__pyx_n_s_OPENMP_MAX_THREADS, __pyx_k_OPENMP_MAX_THREADS, sizeof(__pyx_k_OPENMP_MAX_THREADS), 0, 0, 1, 1},
  {&__pyx_n_s_OPENMP_NUM_THREADS, __pyx_k_OPENMP_NUM_THREADS, sizeof(__pyx_k_OPENMP_NUM_THREADS), 0, 0, 1, 1},
  {&__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_k_Out_of_bounds_on_buffer_access_a, sizeof(__pyx_k_Out_of_bounds_on_buffer_access_a), 0, 0, 1, 0},
  {&__pyx_n_s_PBCBox, __pyx_k_PBCBox, sizeof(__pyx_k_PBCBox), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_kp_s_Pickling_of_struct_members_such, __pyx_k_Pickling_of_struct_members_such, sizeof(__pyx_k_Pickling_of_struct_members_such), 0, 0, 1, 0},
  {&__pyx_kp_s_Pickling_of_struct_members_such_2, __pyx_k_Pickling_of_struct_members_such_2, sizeof(__pyx_k_Pickling_of_struct_members_such_2), 0, 0, 1, 0},
  {&__pyx_kp_u_Preloading_coordinates_from_s, __pyx_k_Preloading_coordinates_from_s, sizeof(__pyx_k_Preloading_coordinates_from_s), 0, 1, 0, 0},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_kp_u_There_is_probably_something_wro, __pyx_k_There_is_probably_something_wro, sizeof(__pyx_k_There_is_probably_something_wro), 0, 1, 0, 0},
  {&__pyx_n_s_Topology, __pyx_k_Topology, sizeof(__pyx_k_Topology), 0, 0, 1, 1},
  {&__pyx_n_s_TopologyGroup, __pyx_k_TopologyGroup, sizeof(__pyx_k_TopologyGroup), 0, 0, 1, 1},
  {&__pyx_kp_u_TopologyGroup_s_i_atoms, __pyx_k_TopologyGroup_s_i_atoms, sizeof(__pyx_k_TopologyGroup_s_i_atoms), 0, 1, 0, 0},
  {&__pyx_n_s_TopologyReader, __pyx_k_TopologyReader, sizeof(__pyx_k_TopologyReader), 0, 0, 1, 1},
  {&__pyx_n_s_Trajectory, __pyx_k_Trajectory, sizeof(__pyx_k_Trajectory), 0, 0, 1, 1},
  {&__pyx_kp_u_Trajectory_is_not_initialized_co, __pyx_k_Trajectory_is_not_initialized_co, sizeof(__pyx_k_Trajectory_is_not_initialized_co), 0, 1, 0, 0},
  {&__pyx_kp_u_Trajectory_not_initialized, __pyx_k_Trajectory_not_initialized, sizeof(__pyx_k_Trajectory_not_initialized), 0, 1, 0, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Unable_to_convert_item_to_object, __pyx_k_Unable_to_convert_item_to_object, sizeof(__pyx_k_Unable_to_convert_item_to_object), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_View_MemoryView, __pyx_k_View_MemoryView, sizeof(__pyx_k_View_MemoryView), 0, 0, 1, 1},
  {&__pyx_kp_u_VmRSS, __pyx_k_VmRSS, sizeof(__pyx_k_VmRSS), 0, 1, 0, 0},
  {&__pyx_kp_u_WARNING_Found_more_than_d_tricl, __pyx_k_WARNING_Found_more_than_d_tricl, sizeof(__pyx_k_WARNING_Found_more_than_d_tricl), 0, 1, 0, 0},
  {&__pyx_kp_u__22, __pyx_k__22, sizeof(__pyx_k__22), 0, 1, 0, 0},
  {&__pyx_kp_u__25, __pyx_k__25, sizeof(__pyx_k__25), 0, 1, 0, 0},
  {&__pyx_kp_u__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 1, 0, 0},
  {&__pyx_kp_u__7, __pyx_k__7, sizeof(__pyx_k__7), 0, 1, 0, 0},
  {&__pyx_n_s_a_memview, __pyx_k_a_memview, sizeof(__pyx_k_a_memview), 0, 0, 1, 1},
  {&__pyx_n_s_allocate_buffer, __pyx_k_allocate_buffer, sizeof(__pyx_k_allocate_buffer), 0, 0, 1, 1},
  {&__pyx_kp_u_and, __pyx_k_and, sizeof(__pyx_k_and), 0, 1, 0, 0},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_arange, __pyx_k_arange, sizeof(__pyx_k_arange), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_asarray, __pyx_k_asarray, sizeof(__pyx_k_asarray), 0, 0, 1, 1},
  {&__pyx_n_s_atomid, __pyx_k_atomid, sizeof(__pyx_k_atomid), 0, 0, 1, 1},
  {&__pyx_kp_u_atomid_not_valid_i, __pyx_k_atomid_not_valid_i, sizeof(__pyx_k_atomid_not_valid_i), 0, 1, 0, 0},
  {&__pyx_kp_u_atomid_not_valid_i_max_atomid_i, __pyx_k_atomid_not_valid_i_max_atomid_i, sizeof(__pyx_k_atomid_not_valid_i_max_atomid_i), 0, 1, 0, 0},
  {&__pyx_n_s_atomids, __pyx_k_atomids, sizeof(__pyx_k_atomids), 0, 0, 1, 1},
  {&__pyx_n_u_atomids, __pyx_k_atomids, sizeof(__pyx_k_atomids), 0, 1, 0, 1},
  {&__pyx_n_s_atomname_str, __pyx_k_atomname_str, sizeof(__pyx_k_atomname_str), 0, 0, 1, 1},
  {&__pyx_n_s_atoms, __pyx_k_atoms, sizeof(__pyx_k_atoms), 0, 0, 1, 1},
  {&__pyx_n_s_b_memview, __pyx_k_b_memview, sizeof(__pyx_k_b_memview), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_basename, __pyx_k_basename, sizeof(__pyx_k_basename), 0, 0, 1, 1},
  {&__pyx_n_s_be_verbose, __pyx_k_be_verbose, sizeof(__pyx_k_be_verbose), 0, 0, 1, 1},
  {&__pyx_n_s_beadids, __pyx_k_beadids, sizeof(__pyx_k_beadids), 0, 0, 1, 1},
  {&__pyx_n_s_begin, __pyx_k_begin, sizeof(__pyx_k_begin), 0, 0, 1, 1},
  {&__pyx_n_s_box, __pyx_k_box, sizeof(__pyx_k_box), 0, 0, 1, 1},
  {&__pyx_n_u_box, __pyx_k_box, sizeof(__pyx_k_box), 0, 1, 0, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_u_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_kp_s_contiguous_and_direct, __pyx_k_contiguous_and_direct, sizeof(__pyx_k_contiguous_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_contiguous_and_indirect, __pyx_k_contiguous_and_indirect, sizeof(__pyx_k_contiguous_and_indirect), 0, 0, 1, 0},
  {&__pyx_n_s_coords_reader, __pyx_k_coords_reader, sizeof(__pyx_k_coords_reader), 0, 0, 1, 1},
  {&__pyx_n_s_core_analysis, __pyx_k_core_analysis, sizeof(__pyx_k_core_analysis), 0, 0, 1, 1},
  {&__pyx_n_s_cutoff, __pyx_k_cutoff, sizeof(__pyx_k_cutoff), 0, 0, 1, 1},
  {&__pyx_n_s_cython_version, __pyx_k_cython_version, sizeof(__pyx_k_cython_version), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_distance2, __pyx_k_distance2, sizeof(__pyx_k_distance2), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dtype_is_object, __pyx_k_dtype_is_object, sizeof(__pyx_k_dtype_is_object), 0, 0, 1, 1},
  {&__pyx_n_s_empty, __pyx_k_empty, sizeof(__pyx_k_empty), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_s_fatslimlib_core_base, __pyx_k_fatslimlib_core_base, sizeof(__pyx_k_fatslimlib_core_base), 0, 0, 1, 1},
  {&__pyx_kp_s_fatslimlib_core_base_pyx, __pyx_k_fatslimlib_core_base_pyx, sizeof(__pyx_k_fatslimlib_core_base_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_filename, __pyx_k_filename, sizeof(__pyx_k_filename), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_float64, __pyx_k_float64, sizeof(__pyx_k_float64), 0, 0, 1, 1},
  {&__pyx_n_s_flush, __pyx_k_flush, sizeof(__pyx_k_flush), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 0, 1, 1},
  {&__pyx_n_u_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 1, 0, 1},
  {&__pyx_n_s_get_aggregates, __pyx_k_get_aggregates, sizeof(__pyx_k_get_aggregates), 0, 0, 1, 1},
  {&__pyx_n_s_get_forcefield_type, __pyx_k_get_forcefield_type, sizeof(__pyx_k_get_forcefield_type), 0, 0, 1, 1},
  {&__pyx_n_s_get_membranes, __pyx_k_get_membranes, sizeof(__pyx_k_get_membranes), 0, 0, 1, 1},
  {&__pyx_n_s_get_rvdw, __pyx_k_get_rvdw, sizeof(__pyx_k_get_rvdw), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_k_got_differing_extents_in_dimensi, sizeof(__pyx_k_got_differing_extents_in_dimensi), 0, 0, 1, 0},
  {&__pyx_n_u_group, __pyx_k_group, sizeof(__pyx_k_group), 0, 1, 0, 1},
  {&__pyx_n_s_groupid, __pyx_k_groupid, sizeof(__pyx_k_groupid), 0, 0, 1, 1},
  {&__pyx_n_u_headgroups, __pyx_k_headgroups, sizeof(__pyx_k_headgroups), 0, 1, 0, 1},
  {&__pyx_n_s_hg_group, __pyx_k_hg_group, sizeof(__pyx_k_hg_group), 0, 0, 1, 1},
  {&__pyx_kp_b_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 0, 0},
  {&__pyx_kp_u_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 1, 0, 0},
  {&__pyx_kp_u_i_frames_preloaded_in_s, __pyx_k_i_frames_preloaded_in_s, sizeof(__pyx_k_i_frames_preloaded_in_s), 0, 1, 0, 0},
  {&__pyx_kp_u_i_groups_loaded_in_s, __pyx_k_i_groups_loaded_in_s, sizeof(__pyx_k_i_groups_loaded_in_s), 0, 1, 0, 0},
  {&__pyx_kp_u_i_topology_atoms_from_i_residue, __pyx_k_i_topology_atoms_from_i_residue, sizeof(__pyx_k_i_topology_atoms_from_i_residue), 0, 1, 0, 0},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
  {&__pyx_n_s_index_reader, __pyx_k_index_reader, sizeof(__pyx_k_index_reader), 0, 0, 1, 1},
  {&__pyx_kp_u_instance, __pyx_k_instance, sizeof(__pyx_k_instance), 0, 1, 0, 0},
  {&__pyx_n_s_int64, __pyx_k_int64, sizeof(__pyx_k_int64), 0, 0, 1, 1},
  {&__pyx_n_s_interacting_group, __pyx_k_interacting_group, sizeof(__pyx_k_interacting_group), 0, 0, 1, 1},
  {&__pyx_kp_u_is_not_a, __pyx_k_is_not_a, sizeof(__pyx_k_is_not_a), 0, 1, 0, 0},
  {&__pyx_n_s_itemsize, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
  {&__pyx_kp_s_itemsize_0_for_cython_array, __pyx_k_itemsize_0_for_cython_array, sizeof(__pyx_k_itemsize_0_for_cython_array), 0, 0, 1, 0},
  {&__pyx_n_s_leaflet1, __pyx_k_leaflet1, sizeof(__pyx_k_leaflet1), 0, 0, 1, 1},
  {&__pyx_n_s_leaflet2, __pyx_k_leaflet2, sizeof(__pyx_k_leaflet2), 0, 0, 1, 1},
  {&__pyx_n_s_leaflet_distance2, __pyx_k_leaflet_distance2, sizeof(__pyx_k_leaflet_distance2), 0, 0, 1, 1},
  {&__pyx_n_s_load, __pyx_k_load, sizeof(__pyx_k_load), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_memview, __pyx_k_memview, sizeof(__pyx_k_memview), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_u_ms, __pyx_k_ms, sizeof(__pyx_k_ms), 0, 1, 0, 1},
  {&__pyx_n_s_msg, __pyx_k_msg, sizeof(__pyx_k_msg), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_u_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 1, 0, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_natoms, __pyx_k_natoms, sizeof(__pyx_k_natoms), 0, 0, 1, 1},
  {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_normal_memview, __pyx_k_normal_memview, sizeof(__pyx_k_normal_memview), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_nresidues, __pyx_k_nresidues, sizeof(__pyx_k_nresidues), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
  {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
  {&__pyx_n_s_parent, __pyx_k_parent, sizeof(__pyx_k_parent), 0, 0, 1, 1},
  {&__pyx_n_s_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_kp_u_proc_self_status, __pyx_k_proc_self_status, sizeof(__pyx_k_proc_self_status), 0, 1, 0, 0},
  {&__pyx_n_u_protein, __pyx_k_protein, sizeof(__pyx_k_protein), 0, 1, 0, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_capi, __pyx_k_pyx_capi, sizeof(__pyx_k_pyx_capi), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_getbuffer, __pyx_k_pyx_getbuffer, sizeof(__pyx_k_pyx_getbuffer), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_CoordinateReader, __pyx_k_pyx_unpickle_CoordinateReader, sizeof(__pyx_k_pyx_unpickle_CoordinateReader), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Enum, __pyx_k_pyx_unpickle_Enum, sizeof(__pyx_k_pyx_unpickle_Enum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_IndexReader, __pyx_k_pyx_unpickle_IndexReader, sizeof(__pyx_k_pyx_unpickle_IndexReader), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_TopologyReader, __pyx_k_pyx_unpickle_TopologyReader, sizeof(__pyx_k_pyx_unpickle_TopologyReader), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Trajectory, __pyx_k_pyx_unpickle_Trajectory, sizeof(__pyx_k_pyx_unpickle_Trajectory), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_u_rb, __pyx_k_rb, sizeof(__pyx_k_rb), 0, 1, 0, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_resid, __pyx_k_resid, sizeof(__pyx_k_resid), 0, 0, 1, 1},
  {&__pyx_n_u_resid, __pyx_k_resid, sizeof(__pyx_k_resid), 0, 1, 0, 1},
  {&__pyx_kp_u_resid_not_valid_i, __pyx_k_resid_not_valid_i, sizeof(__pyx_k_resid_not_valid_i), 0, 1, 0, 0},
  {&__pyx_n_s_residues, __pyx_k_residues, sizeof(__pyx_k_residues), 0, 0, 1, 1},
  {&__pyx_n_s_resname, __pyx_k_resname, sizeof(__pyx_k_resname), 0, 0, 1, 1},
  {&__pyx_n_u_resname, __pyx_k_resname, sizeof(__pyx_k_resname), 0, 1, 0, 1},
  {&__pyx_n_s_resname_str, __pyx_k_resname_str, sizeof(__pyx_k_resname_str), 0, 0, 1, 1},
  {&__pyx_n_s_retrieve_aggregates, __pyx_k_retrieve_aggregates, sizeof(__pyx_k_retrieve_aggregates), 0, 0, 1, 1},
  {&__pyx_n_s_retrieve_membranes, __pyx_k_retrieve_membranes, sizeof(__pyx_k_retrieve_membranes), 0, 0, 1, 1},
  {&__pyx_n_u_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 1, 0, 1},
  {&__pyx_kp_s_self_atomids_to_internalids_self, __pyx_k_self_atomids_to_internalids_self, sizeof(__pyx_k_self_atomids_to_internalids_self), 0, 0, 1, 0},
  {&__pyx_kp_s_self_atoms_cannot_be_converted_t, __pyx_k_self_atoms_cannot_be_converted_t, sizeof(__pyx_k_self_atoms_cannot_be_converted_t), 0, 0, 1, 0},
  {&__pyx_kp_s_self_neighbors_cannot_be_convert, __pyx_k_self_neighbors_cannot_be_convert, sizeof(__pyx_k_self_neighbors_cannot_be_convert), 0, 0, 1, 0},
  {&__pyx_n_s_setdefault, __pyx_k_setdefault, sizeof(__pyx_k_setdefault), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_u_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 1, 0, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_startswith, __pyx_k_startswith, sizeof(__pyx_k_startswith), 0, 0, 1, 1},
  {&__pyx_n_s_stdout, __pyx_k_stdout, sizeof(__pyx_k_stdout), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_kp_s_strided_and_direct, __pyx_k_strided_and_direct, sizeof(__pyx_k_strided_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_direct_or_indirect, __pyx_k_strided_and_direct_or_indirect, sizeof(__pyx_k_strided_and_direct_or_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_indirect, __pyx_k_strided_and_indirect, sizeof(__pyx_k_strided_and_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
  {&__pyx_n_s_strip_name, __pyx_k_strip_name, sizeof(__pyx_k_strip_name), 0, 0, 1, 1},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
  {&__pyx_n_s_timestep, __pyx_k_timestep, sizeof(__pyx_k_timestep), 0, 0, 1, 1},
  {&__pyx_n_s_tolist, __pyx_k_tolist, sizeof(__pyx_k_tolist), 0, 0, 1, 1},
  {&__pyx_n_s_top_reader, __pyx_k_top_reader, sizeof(__pyx_k_top_reader), 0, 0, 1, 1},
  {&__pyx_n_s_trajectory, __pyx_k_trajectory, sizeof(__pyx_k_trajectory), 0, 0, 1, 1},
  {&__pyx_kp_s_unable_to_allocate_array_data, __pyx_k_unable_to_allocate_array_data, sizeof(__pyx_k_unable_to_allocate_array_data), 0, 0, 1, 0},
  {&__pyx_kp_s_unable_to_allocate_shape_and_str, __pyx_k_unable_to_allocate_shape_and_str, sizeof(__pyx_k_unable_to_allocate_shape_and_str), 0, 0, 1, 0},
  {&__pyx_n_u_unk, __pyx_k_unk, sizeof(__pyx_k_unk), 0, 1, 0, 1},
  {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_u_us, __pyx_k_us, sizeof(__pyx_k_us), 0, 1, 0, 1},
  {&__pyx_n_s_verbose, __pyx_k_verbose, sizeof(__pyx_k_verbose), 0, 0, 1, 1},
  {&__pyx_n_s_verbose_print, __pyx_k_verbose_print, sizeof(__pyx_k_verbose_print), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
  {&__pyx_n_s_z, __pyx_k_z, sizeof(__pyx_k_z), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 109, __pyx_L1_error)
  __pyx_builtin_open = __Pyx_GetBuiltinName(__pyx_n_s_open); if (!__pyx_builtin_open) __PYX_ERR(0, 151, __pyx_L1_error)
  __pyx_builtin_print = __Pyx_GetBuiltinName(__pyx_n_s_print); if (!__pyx_builtin_print) __PYX_ERR(0, 160, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 252, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(0, 353, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 542, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 1090, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 1140, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 1231, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(1, 151, __pyx_L1_error)
  __pyx_builtin_Ellipsis = __Pyx_GetBuiltinName(__pyx_n_s_Ellipsis); if (!__pyx_builtin_Ellipsis) __PYX_ERR(1, 404, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(1, 613, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "fatslimlib/core_base.pyx":141
 *     if duration != 0:
 *         expo = int(floor(log10(duration)))
 *         unit = {-1 : (3, "ms"), -2:(6, "us")}             # <<<<<<<<<<<<<<
 *         ex, unit_str = unit.setdefault(expo // 3, (0, "s"))
 *     else:
 */
  __pyx_tuple_ = PyTuple_Pack(2, __pyx_int_3, __pyx_n_u_ms); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);
  __pyx_tuple__2 = PyTuple_Pack(2, __pyx_int_6, __pyx_n_u_us); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "fatslimlib/core_base.pyx":142
 *         expo = int(floor(log10(duration)))
 *         unit = {-1 : (3, "ms"), -2:(6, "us")}
 *         ex, unit_str = unit.setdefault(expo // 3, (0, "s"))             # <<<<<<<<<<<<<<
 *     else:
 *         ex = 0.0
 */
  __pyx_tuple__3 = PyTuple_Pack(2, __pyx_int_0, __pyx_n_u_s); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "fatslimlib/core_base.pyx":151
 * cpdef real get_memory_usage():
 *     memory = -1
 *     with open('/proc/self/status', "rb") as fp:             # <<<<<<<<<<<<<<
 *         for line in fp:
 *             if line.startswith("VmRSS:"):
 */
  __pyx_tuple__5 = PyTuple_Pack(2, __pyx_kp_u_proc_self_status, __pyx_n_u_rb); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);
  __pyx_tuple__6 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("Pickling of struct members such as self.c_atom must be explicitly requested with @auto_pickle(True)")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("Pickling of struct members such as self.c_atom must be explicitly requested with @auto_pickle(True)")
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_Pickling_of_struct_members_such); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "(tree fragment)":4
 *     raise TypeError("Pickling of struct members such as self.c_atom must be explicitly requested with @auto_pickle(True)")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("Pickling of struct members such as self.c_atom must be explicitly requested with @auto_pickle(True)")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_Pickling_of_struct_members_such); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "fatslimlib/core_base.pyx":527
 *                                                   "correction vectors, ignoring some."
 *                                                   % MAX_NTRICVEC)
 *                                             print("  There is probably something wrong with "             # <<<<<<<<<<<<<<
 *                                                   "your box.")
 *                                             print(box)
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_u_There_is_probably_something_wro); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "fatslimlib/core_base.pyx":545
 *                              (DIM, DIM, box.shape[0], box.shape[1]))
 *         if (box[XX, XX] == 0) or (box[YY, YY] == 0) or (box[ZZ, ZZ] == 0):
 *             raise ValueError("Box does not correspond to PBC=xyz")             # <<<<<<<<<<<<<<
 *         self.fast_update(box)
 * 
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_u_Box_does_not_correspond_to_PBC_x); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "fatslimlib/core_base.pyx":550
 *     property box:
 *         def __get__(self):
 *             box_py = np.empty((DIM, DIM))             # <<<<<<<<<<<<<<
 *             cdef real[:,::1] box_memview = box_py
 * 
 */
  __pyx_tuple__14 = PyTuple_Pack(2, __pyx_int_3, __pyx_int_3); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "fatslimlib/core_base.pyx":696
 *         with gil:
 *             if natoms == 0:
 *                 bbox_coords = np.empty((0, DIM))             # <<<<<<<<<<<<<<
 *             else:
 *                 bbox_coords = coords.copy()
 */
  __pyx_tuple__15 = PyTuple_Pack(2, __pyx_int_0, __pyx_int_3); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("Pickling of struct members such as self.c_pbcbox must be explicitly requested with @auto_pickle(True)")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("Pickling of struct members such as self.c_pbcbox must be explicitly requested with @auto_pickle(True)")
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_Pickling_of_struct_members_such_2); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "(tree fragment)":4
 *     raise TypeError("Pickling of struct members such as self.c_pbcbox must be explicitly requested with @auto_pickle(True)")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("Pickling of struct members such as self.c_pbcbox must be explicitly requested with @auto_pickle(True)")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_Pickling_of_struct_members_such_2); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.atoms cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.atoms cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_self_atoms_cannot_be_converted_t); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "(tree fragment)":4
 *     raise TypeError("self.atoms cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.atoms cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_self_atoms_cannot_be_converted_t); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.atomids_to_internalids,self.atomnames,self.atoms,self.resids_to_internalids,self.residues,self.resnames cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.atomids_to_internalids,self.atomnames,self.atoms,self.resids_to_internalids,self.residues,self.resnames cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_s_self_atomids_to_internalids_self); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "(tree fragment)":4
 *     raise TypeError("self.atomids_to_internalids,self.atomnames,self.atoms,self.resids_to_internalids,self.residues,self.resnames cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.atomids_to_internalids,self.atomnames,self.atoms,self.resids_to_internalids,self.residues,self.resnames cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_s_self_atomids_to_internalids_self); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.neighbors cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.neighbors cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_s_self_neighbors_cannot_be_convert); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "(tree fragment)":4
 *     raise TypeError("self.neighbors cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.neighbors cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_kp_s_self_neighbors_cannot_be_convert); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "fatslimlib/core_base.pyx":1887
 * 
 *         if self.lipid_atomids is None:
 *             raise ValueError("Trajectory is not initialized correctly!")             # <<<<<<<<<<<<<<
 * 
 *         # Load the first lipid from the first frame
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_u_Trajectory_is_not_initialized_co); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 1887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "fatslimlib/core_base.pyx":1915
 *     cdef assert_initialization(self):
 *         if not self.initialized:
 *             raise ValueError("Trajectory not initialized")             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, fsl_int item):
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_u_Trajectory_not_initialized); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 1915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "View.MemoryView":133
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_s_Empty_shape_tuple_for_cython_arr); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(1, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "View.MemoryView":136
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_s_itemsize_0_for_cython_array); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(1, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "View.MemoryView":148
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_shape_and_str); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(1, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "View.MemoryView":176
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_array_data); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "View.MemoryView":192
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_kp_s_Can_only_create_a_buffer_that_is); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(1, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "View.MemoryView":418
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
  __pyx_tuple__35 = PyTuple_Pack(1, __pyx_kp_s_Cannot_assign_to_read_only_memor); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(1, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "View.MemoryView":495
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_kp_s_Unable_to_convert_item_to_object); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(1, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "View.MemoryView":520
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
  __pyx_tuple__37 = PyTuple_Pack(1, __pyx_kp_s_Cannot_create_writable_memory_vi); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(1, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "View.MemoryView":570
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
  __pyx_tuple__38 = PyTuple_Pack(1, __pyx_kp_s_Buffer_view_does_not_expose_stri); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(1, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "View.MemoryView":577
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
  __pyx_tuple__39 = PyTuple_New(1); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(1, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_tuple__39, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_tuple__39);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__40 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__41 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);

  /* "View.MemoryView":682
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */
  __pyx_slice__42 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__42)) __PYX_ERR(1, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__42);
  __Pyx_GIVEREF(__pyx_slice__42);

  /* "View.MemoryView":703
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__43 = PyTuple_Pack(1, __pyx_kp_s_Indirect_dimensions_not_supporte); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(1, 703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__44 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__45 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);

  /* "fatslimlib/core_base.pyx":158
 *     return memory
 * 
 * def verbose_print(msg, bint verbose, end="\n"):             # <<<<<<<<<<<<<<
 *     if verbose:
 *         print(msg, end=end)
 */
  __pyx_tuple__46 = PyTuple_Pack(3, __pyx_n_s_msg, __pyx_n_s_verbose, __pyx_n_s_end); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fatslimlib_core_base_pyx, __pyx_n_s_verbose_print, 158, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(0, 158, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_TopologyReader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__48 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);
  __pyx_codeobj__49 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_TopologyReader, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__49)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_tuple__50 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__50, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_CoordinateReader, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_tuple__52 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);
  __pyx_codeobj__53 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__52, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_IndexReader, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__53)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_tuple__54 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);
  __pyx_codeobj__55 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__54, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Trajectory, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__55)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "View.MemoryView":286
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_tuple__56 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct_or_indirect); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(1, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);

  /* "View.MemoryView":287
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_tuple__57 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(1, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);

  /* "View.MemoryView":288
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__58 = PyTuple_Pack(1, __pyx_kp_s_strided_and_indirect); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(1, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);

  /* "View.MemoryView":291
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_tuple__59 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_direct); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(1, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);

  /* "View.MemoryView":292
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__60 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_indirect); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(1, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__61 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__61, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Enum, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* InitThreads.init */
  #ifdef WITH_THREAD
PyEval_InitThreads();
#endif

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  __pyx_umethod_PyDict_Type_setdefault.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyUnicode_Type_strip.type = (PyObject*)&PyUnicode_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_0_0 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_float_0_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_6 = PyInt_FromLong(6); if (unlikely(!__pyx_int_6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_12 = PyInt_FromLong(12); if (unlikely(!__pyx_int_12)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_83902476 = PyInt_FromLong(83902476L); if (unlikely(!__pyx_int_83902476)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_111280737 = PyInt_FromLong(111280737L); if (unlikely(!__pyx_int_111280737)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_128740812 = PyInt_FromLong(128740812L); if (unlikely(!__pyx_int_128740812)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_148569195 = PyInt_FromLong(148569195L); if (unlikely(!__pyx_int_148569195)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_184977713 = PyInt_FromLong(184977713L); if (unlikely(!__pyx_int_184977713)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_2 = PyInt_FromLong(-2); if (unlikely(!__pyx_int_neg_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  generic = Py_None; Py_INCREF(Py_None);
  strided = Py_None; Py_INCREF(Py_None);
  indirect = Py_None; Py_INCREF(Py_None);
  contiguous = Py_None; Py_INCREF(Py_None);
  indirect_contiguous = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  if (__Pyx_ExportVoidPtr(__pyx_n_s_OPENMP_MAX_THREADS, (void *)&__pyx_v_10fatslimlib_9core_base_OPENMP_MAX_THREADS, "fsl_int") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_OPENMP_NUM_THREADS, (void *)&__pyx_v_10fatslimlib_9core_base_OPENMP_NUM_THREADS, "fsl_int") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("get_num_threads", (void (*)(void))__pyx_f_10fatslimlib_9core_base_get_num_threads, "fsl_int (int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("get_max_threads", (void (*)(void))__pyx_f_10fatslimlib_9core_base_get_max_threads, "fsl_int (int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("get_num_procs", (void (*)(void))__pyx_f_10fatslimlib_9core_base_get_num_procs, "fsl_int (int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("set_num_threads", (void (*)(void))__pyx_f_10fatslimlib_9core_base_set_num_threads, "PyObject *(fsl_int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("norm2", (void (*)(void))__pyx_f_10fatslimlib_9core_base_norm2, "real (__Pyx_memviewslice)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("mat_to_memview", (void (*)(void))__pyx_f_10fatslimlib_9core_base_mat_to_memview, "void (real (*)[3], __Pyx_memviewslice)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("rvec_to_memview", (void (*)(void))__pyx_f_10fatslimlib_9core_base_rvec_to_memview, "void (real *, __Pyx_memviewslice)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("pretty_delta", (void (*)(void))__pyx_f_10fatslimlib_9core_base_pretty_delta, "PyObject *(PyObject *, PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("pretty_duration", (void (*)(void))__pyx_f_10fatslimlib_9core_base_pretty_duration, "PyObject *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_10fatslimlib_9core_base_Atom) < 0) __PYX_ERR(0, 163, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10fatslimlib_9core_base_Atom.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10fatslimlib_9core_base_Atom.tp_dictoffset && __pyx_type_10fatslimlib_9core_base_Atom.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10fatslimlib_9core_base_Atom.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Atom, (PyObject *)&__pyx_type_10fatslimlib_9core_base_Atom) < 0) __PYX_ERR(0, 163, __pyx_L1_error)
  if (__pyx_type_10fatslimlib_9core_base_Atom.tp_weaklistoffset == 0) __pyx_type_10fatslimlib_9core_base_Atom.tp_weaklistoffset = offsetof(struct __pyx_obj_10fatslimlib_9core_base_Atom, __weakref__);
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10fatslimlib_9core_base_Atom) < 0) __PYX_ERR(0, 163, __pyx_L1_error)
  __pyx_ptype_10fatslimlib_9core_base_Atom = &__pyx_type_10fatslimlib_9core_base_Atom;
  __pyx_vtabptr_10fatslimlib_9core_base_AtomGroup = &__pyx_vtable_10fatslimlib_9core_base_AtomGroup;
  __pyx_vtable_10fatslimlib_9core_base_AtomGroup.fast_append = (PyObject *(*)(struct __pyx_obj_10fatslimlib_9core_base_AtomGroup *, struct __pyx_obj_10fatslimlib_9core_base_Atom *))__pyx_f_10fatslimlib_9core_base_9AtomGroup_fast_append;
  if (PyType_Ready(&__pyx_type_10fatslimlib_9core_base_AtomGroup) < 0) __PYX_ERR(0, 323, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10fatslimlib_9core_base_AtomGroup.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10fatslimlib_9core_base_AtomGroup.tp_dictoffset && __pyx_type_10fatslimlib_9core_base_AtomGroup.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10fatslimlib_9core_base_AtomGroup.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10fatslimlib_9core_base_AtomGroup.tp_dict, __pyx_vtabptr_10fatslimlib_9core_base_AtomGroup) < 0) __PYX_ERR(0, 323, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_AtomGroup, (PyObject *)&__pyx_type_10fatslimlib_9core_base_AtomGroup) < 0) __PYX_ERR(0, 323, __pyx_L1_error)
  if (__pyx_type_10fatslimlib_9core_base_AtomGroup.tp_weaklistoffset == 0) __pyx_type_10fatslimlib_9core_base_AtomGroup.tp_weaklistoffset = offsetof(struct __pyx_obj_10fatslimlib_9core_base_AtomGroup, __weakref__);
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10fatslimlib_9core_base_AtomGroup) < 0) __PYX_ERR(0, 323, __pyx_L1_error)
  __pyx_ptype_10fatslimlib_9core_base_AtomGroup = &__pyx_type_10fatslimlib_9core_base_AtomGroup;
  __pyx_vtabptr_10fatslimlib_9core_base_PBCBox = &__pyx_vtable_10fatslimlib_9core_base_PBCBox;
  __pyx_vtable_10fatslimlib_9core_base_PBCBox.fast_update = (void (*)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, __Pyx_memviewslice))__pyx_f_10fatslimlib_9core_base_6PBCBox_fast_update;
  __pyx_vtable_10fatslimlib_9core_base_PBCBox.fast_pbc_dx = (void (*)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, real *, real *, real *))__pyx_f_10fatslimlib_9core_base_6PBCBox_fast_pbc_dx;
  __pyx_vtable_10fatslimlib_9core_base_PBCBox.fast_pbc_dx_leaflet = (void (*)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, real *, real *, real *, real *))__pyx_f_10fatslimlib_9core_base_6PBCBox_fast_pbc_dx_leaflet;
  __pyx_vtable_10fatslimlib_9core_base_PBCBox.fast_distance2 = (real (*)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, real *, real *))__pyx_f_10fatslimlib_9core_base_6PBCBox_fast_distance2;
  __pyx_vtable_10fatslimlib_9core_base_PBCBox.fast_distance = (real (*)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, real *, real *))__pyx_f_10fatslimlib_9core_base_6PBCBox_fast_distance;
  __pyx_vtable_10fatslimlib_9core_base_PBCBox.fast_leaflet_distance2 = (real (*)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, real *, real *, real *))__pyx_f_10fatslimlib_9core_base_6PBCBox_fast_leaflet_distance2;
  __pyx_vtable_10fatslimlib_9core_base_PBCBox.fast_leaflet_distance = (real (*)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, real *, real *, real *))__pyx_f_10fatslimlib_9core_base_6PBCBox_fast_leaflet_distance;
  __pyx_vtable_10fatslimlib_9core_base_PBCBox.fast_pbc_xcm = (void (*)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, __Pyx_memviewslice, real *, struct __pyx_opt_args_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm *__pyx_optional_args))__pyx_f_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm;
  __pyx_vtable_10fatslimlib_9core_base_PBCBox.fast_pbc_xcm_from_ref = (void (*)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, __Pyx_memviewslice, real *, real *, struct __pyx_opt_args_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm_from_ref *__pyx_optional_args))__pyx_f_10fatslimlib_9core_base_6PBCBox_fast_pbc_xcm_from_ref;
  __pyx_vtable_10fatslimlib_9core_base_PBCBox.fast_put_atoms_in_bbox = (__Pyx_memviewslice (*)(struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, __Pyx_memviewslice))__pyx_f_10fatslimlib_9core_base_6PBCBox_fast_put_atoms_in_bbox;
  if (PyType_Ready(&__pyx_type_10fatslimlib_9core_base_PBCBox) < 0) __PYX_ERR(0, 420, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10fatslimlib_9core_base_PBCBox.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10fatslimlib_9core_base_PBCBox.tp_dictoffset && __pyx_type_10fatslimlib_9core_base_PBCBox.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10fatslimlib_9core_base_PBCBox.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10fatslimlib_9core_base_PBCBox.tp_dict, __pyx_vtabptr_10fatslimlib_9core_base_PBCBox) < 0) __PYX_ERR(0, 420, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PBCBox, (PyObject *)&__pyx_type_10fatslimlib_9core_base_PBCBox) < 0) __PYX_ERR(0, 420, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10fatslimlib_9core_base_PBCBox) < 0) __PYX_ERR(0, 420, __pyx_L1_error)
  __pyx_ptype_10fatslimlib_9core_base_PBCBox = &__pyx_type_10fatslimlib_9core_base_PBCBox;
  __pyx_vtabptr_10fatslimlib_9core_base_TopologyGroup = &__pyx_vtable_10fatslimlib_9core_base_TopologyGroup;
  __pyx_vtable_10fatslimlib_9core_base_TopologyGroup.fast_contains = (int (*)(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *, struct __pyx_t_10fatslimlib_9core_base_topol_atom_t))__pyx_f_10fatslimlib_9core_base_13TopologyGroup_fast_contains;
  __pyx_vtable_10fatslimlib_9core_base_TopologyGroup.add_atom = (void (*)(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *, struct __pyx_t_10fatslimlib_9core_base_topol_atom_t))__pyx_f_10fatslimlib_9core_base_13TopologyGroup_add_atom;
  __pyx_vtable_10fatslimlib_9core_base_TopologyGroup.merge_other_group = (void (*)(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *, struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *))__pyx_f_10fatslimlib_9core_base_13TopologyGroup_merge_other_group;
  __pyx_vtable_10fatslimlib_9core_base_TopologyGroup.fast_get_atomids = (__Pyx_memviewslice (*)(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *))__pyx_f_10fatslimlib_9core_base_13TopologyGroup_fast_get_atomids;
  __pyx_vtable_10fatslimlib_9core_base_TopologyGroup.group_by_resid = (PyObject *(*)(struct __pyx_obj_10fatslimlib_9core_base_TopologyGroup *))__pyx_f_10fatslimlib_9core_base_13TopologyGroup_group_by_resid;
  if (PyType_Ready(&__pyx_type_10fatslimlib_9core_base_TopologyGroup) < 0) __PYX_ERR(0, 714, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10fatslimlib_9core_base_TopologyGroup.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10fatslimlib_9core_base_TopologyGroup.tp_dictoffset && __pyx_type_10fatslimlib_9core_base_TopologyGroup.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10fatslimlib_9core_base_TopologyGroup.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10fatslimlib_9core_base_TopologyGroup.tp_dict, __pyx_vtabptr_10fatslimlib_9core_base_TopologyGroup) < 0) __PYX_ERR(0, 714, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_TopologyGroup, (PyObject *)&__pyx_type_10fatslimlib_9core_base_TopologyGroup) < 0) __PYX_ERR(0, 714, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10fatslimlib_9core_base_TopologyGroup) < 0) __PYX_ERR(0, 714, __pyx_L1_error)
  __pyx_ptype_10fatslimlib_9core_base_TopologyGroup = &__pyx_type_10fatslimlib_9core_base_TopologyGroup;
  __pyx_vtabptr_10fatslimlib_9core_base_Topology = &__pyx_vtable_10fatslimlib_9core_base_Topology;
  __pyx_vtable_10fatslimlib_9core_base_Topology.set_residues_allocation = (void (*)(struct __pyx_obj_10fatslimlib_9core_base_Topology *, fsl_int))__pyx_f_10fatslimlib_9core_base_8Topology_set_residues_allocation;
  __pyx_vtable_10fatslimlib_9core_base_Topology.set_atoms_allocation = (void (*)(struct __pyx_obj_10fatslimlib_9core_base_Topology *, fsl_int))__pyx_f_10fatslimlib_9core_base_8Topology_set_atoms_allocation;
  __pyx_vtable_10fatslimlib_9core_base_Topology.internal_append = (void (*)(struct __pyx_obj_10fatslimlib_9core_base_Topology *, fsl_int, PyObject *, PyObject *, fsl_int))__pyx_f_10fatslimlib_9core_base_8Topology_internal_append;
  __pyx_vtable_10fatslimlib_9core_base_Topology.fast_get_atom_internal_id = (fsl_int (*)(struct __pyx_obj_10fatslimlib_9core_base_Topology *, fsl_int))__pyx_f_10fatslimlib_9core_base_8Topology_fast_get_atom_internal_id;
  __pyx_vtable_10fatslimlib_9core_base_Topology.fast_get_atom = (struct __pyx_t_10fatslimlib_9core_base_topol_atom_t *(*)(struct __pyx_obj_10fatslimlib_9core_base_Topology *, fsl_int))__pyx_f_10fatslimlib_9core_base_8Topology_fast_get_atom;
  __pyx_vtable_10fatslimlib_9core_base_Topology.fast_get_resid_from_atomid = (fsl_int (*)(struct __pyx_obj_10fatslimlib_9core_base_Topology *, fsl_int))__pyx_f_10fatslimlib_9core_base_8Topology_fast_get_resid_from_atomid;
  __pyx_vtable_10fatslimlib_9core_base_Topology.fast_get_residue_internal_id = (fsl_int (*)(struct __pyx_obj_10fatslimlib_9core_base_Topology *, fsl_int))__pyx_f_10fatslimlib_9core_base_8Topology_fast_get_residue_internal_id;
  __pyx_vtable_10fatslimlib_9core_base_Topology.fast_get_residue = (struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *(*)(struct __pyx_obj_10fatslimlib_9core_base_Topology *, fsl_int))__pyx_f_10fatslimlib_9core_base_8Topology_fast_get_residue;
  __pyx_vtable_10fatslimlib_9core_base_Topology.fast_get_residue_from_atomid = (struct __pyx_t_10fatslimlib_9core_base_topol_residue_t *(*)(struct __pyx_obj_10fatslimlib_9core_base_Topology *, fsl_int))__pyx_f_10fatslimlib_9core_base_8Topology_fast_get_residue_from_atomid;
  if (PyType_Ready(&__pyx_type_10fatslimlib_9core_base_Topology) < 0) __PYX_ERR(0, 871, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10fatslimlib_9core_base_Topology.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10fatslimlib_9core_base_Topology.tp_dictoffset && __pyx_type_10fatslimlib_9core_base_Topology.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10fatslimlib_9core_base_Topology.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10fatslimlib_9core_base_Topology.tp_dict, __pyx_vtabptr_10fatslimlib_9core_base_Topology) < 0) __PYX_ERR(0, 871, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Topology, (PyObject *)&__pyx_type_10fatslimlib_9core_base_Topology) < 0) __PYX_ERR(0, 871, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10fatslimlib_9core_base_Topology) < 0) __PYX_ERR(0, 871, __pyx_L1_error)
  __pyx_ptype_10fatslimlib_9core_base_Topology = &__pyx_type_10fatslimlib_9core_base_Topology;
  __pyx_vtabptr_10fatslimlib_9core_base_TopologyReader = &__pyx_vtable_10fatslimlib_9core_base_TopologyReader;
  __pyx_vtable_10fatslimlib_9core_base_TopologyReader.load = (PyObject *(*)(struct __pyx_obj_10fatslimlib_9core_base_TopologyReader *))__pyx_f_10fatslimlib_9core_base_14TopologyReader_load;
  if (PyType_Ready(&__pyx_type_10fatslimlib_9core_base_TopologyReader) < 0) __PYX_ERR(0, 1211, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10fatslimlib_9core_base_TopologyReader.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10fatslimlib_9core_base_TopologyReader.tp_dictoffset && __pyx_type_10fatslimlib_9core_base_TopologyReader.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10fatslimlib_9core_base_TopologyReader.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10fatslimlib_9core_base_TopologyReader.tp_dict, __pyx_vtabptr_10fatslimlib_9core_base_TopologyReader) < 0) __PYX_ERR(0, 1211, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_TopologyReader, (PyObject *)&__pyx_type_10fatslimlib_9core_base_TopologyReader) < 0) __PYX_ERR(0, 1211, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10fatslimlib_9core_base_TopologyReader) < 0) __PYX_ERR(0, 1211, __pyx_L1_error)
  __pyx_ptype_10fatslimlib_9core_base_TopologyReader = &__pyx_type_10fatslimlib_9core_base_TopologyReader;
  __pyx_vtabptr_10fatslimlib_9core_base_CoordinateReader = &__pyx_vtable_10fatslimlib_9core_base_CoordinateReader;
  __pyx_vtable_10fatslimlib_9core_base_CoordinateReader.preload = (PyObject *(*)(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *))__pyx_f_10fatslimlib_9core_base_16CoordinateReader_preload;
  __pyx_vtable_10fatslimlib_9core_base_CoordinateReader.load_box = (struct __pyx_obj_10fatslimlib_9core_base_PBCBox *(*)(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *, fsl_int))__pyx_f_10fatslimlib_9core_base_16CoordinateReader_load_box;
  __pyx_vtable_10fatslimlib_9core_base_CoordinateReader.assert_frame_id = (PyObject *(*)(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *, fsl_int))__pyx_f_10fatslimlib_9core_base_16CoordinateReader_assert_frame_id;
  __pyx_vtable_10fatslimlib_9core_base_CoordinateReader.load_coords = (__Pyx_memviewslice (*)(struct __pyx_obj_10fatslimlib_9core_base_CoordinateReader *, fsl_int, __Pyx_memviewslice))__pyx_f_10fatslimlib_9core_base_16CoordinateReader_load_coords;
  if (PyType_Ready(&__pyx_type_10fatslimlib_9core_base_CoordinateReader) < 0) __PYX_ERR(0, 1235, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10fatslimlib_9core_base_CoordinateReader.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10fatslimlib_9core_base_CoordinateReader.tp_dictoffset && __pyx_type_10fatslimlib_9core_base_CoordinateReader.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10fatslimlib_9core_base_CoordinateReader.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10fatslimlib_9core_base_CoordinateReader.tp_dict, __pyx_vtabptr_10fatslimlib_9core_base_CoordinateReader) < 0) __PYX_ERR(0, 1235, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CoordinateReader, (PyObject *)&__pyx_type_10fatslimlib_9core_base_CoordinateReader) < 0) __PYX_ERR(0, 1235, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10fatslimlib_9core_base_CoordinateReader) < 0) __PYX_ERR(0, 1235, __pyx_L1_error)
  __pyx_ptype_10fatslimlib_9core_base_CoordinateReader = &__pyx_type_10fatslimlib_9core_base_CoordinateReader;
  __pyx_vtabptr_10fatslimlib_9core_base_IndexReader = &__pyx_vtable_10fatslimlib_9core_base_IndexReader;
  __pyx_vtable_10fatslimlib_9core_base_IndexReader.fast_load = (PyObject *(*)(struct __pyx_obj_10fatslimlib_9core_base_IndexReader *))__pyx_f_10fatslimlib_9core_base_11IndexReader_fast_load;
  if (PyType_Ready(&__pyx_type_10fatslimlib_9core_base_IndexReader) < 0) __PYX_ERR(0, 1265, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10fatslimlib_9core_base_IndexReader.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10fatslimlib_9core_base_IndexReader.tp_dictoffset && __pyx_type_10fatslimlib_9core_base_IndexReader.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10fatslimlib_9core_base_IndexReader.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10fatslimlib_9core_base_IndexReader.tp_dict, __pyx_vtabptr_10fatslimlib_9core_base_IndexReader) < 0) __PYX_ERR(0, 1265, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IndexReader, (PyObject *)&__pyx_type_10fatslimlib_9core_base_IndexReader) < 0) __PYX_ERR(0, 1265, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10fatslimlib_9core_base_IndexReader) < 0) __PYX_ERR(0, 1265, __pyx_L1_error)
  __pyx_ptype_10fatslimlib_9core_base_IndexReader = &__pyx_type_10fatslimlib_9core_base_IndexReader;
  __pyx_vtabptr_10fatslimlib_9core_base_Frame = &__pyx_vtable_10fatslimlib_9core_base_Frame;
  __pyx_vtable_10fatslimlib_9core_base_Frame.get_forcefield_type = (fsl_int (*)(struct __pyx_obj_10fatslimlib_9core_base_Frame *, int __pyx_skip_dispatch))__pyx_f_10fatslimlib_9core_base_5Frame_get_forcefield_type;
  __pyx_vtable_10fatslimlib_9core_base_Frame.fast_get_hg_group_coords = (__Pyx_memviewslice (*)(struct __pyx_obj_10fatslimlib_9core_base_Frame *))__pyx_f_10fatslimlib_9core_base_5Frame_fast_get_hg_group_coords;
  __pyx_vtable_10fatslimlib_9core_base_Frame.fast_get_hg_group_coords_bbox = (__Pyx_memviewslice (*)(struct __pyx_obj_10fatslimlib_9core_base_Frame *))__pyx_f_10fatslimlib_9core_base_5Frame_fast_get_hg_group_coords_bbox;
  __pyx_vtable_10fatslimlib_9core_base_Frame.fast_get_interacting_group_coords = (__Pyx_memviewslice (*)(struct __pyx_obj_10fatslimlib_9core_base_Frame *))__pyx_f_10fatslimlib_9core_base_5Frame_fast_get_interacting_group_coords;
  __pyx_vtable_10fatslimlib_9core_base_Frame.fast_get_interacting_group_coords_bbox = (__Pyx_memviewslice (*)(struct __pyx_obj_10fatslimlib_9core_base_Frame *))__pyx_f_10fatslimlib_9core_base_5Frame_fast_get_interacting_group_coords_bbox;
  __pyx_vtable_10fatslimlib_9core_base_Frame.fast_get_bead_coords = (__Pyx_memviewslice (*)(struct __pyx_obj_10fatslimlib_9core_base_Frame *))__pyx_f_10fatslimlib_9core_base_5Frame_fast_get_bead_coords;
  __pyx_vtable_10fatslimlib_9core_base_Frame.fast_get_bead_coords_bbox = (__Pyx_memviewslice (*)(struct __pyx_obj_10fatslimlib_9core_base_Frame *))__pyx_f_10fatslimlib_9core_base_5Frame_fast_get_bead_coords_bbox;
  __pyx_vtable_10fatslimlib_9core_base_Frame.fast_get_lipid_coords = (__Pyx_memviewslice (*)(struct __pyx_obj_10fatslimlib_9core_base_Frame *))__pyx_f_10fatslimlib_9core_base_5Frame_fast_get_lipid_coords;
  __pyx_vtable_10fatslimlib_9core_base_Frame.get_lipid_coords_aslist = (PyObject *(*)(struct __pyx_obj_10fatslimlib_9core_base_Frame *))__pyx_f_10fatslimlib_9core_base_5Frame_get_lipid_coords_aslist;
  __pyx_vtable_10fatslimlib_9core_base_Frame.fast_get_lipid_coords_bbox = (__Pyx_memviewslice (*)(struct __pyx_obj_10fatslimlib_9core_base_Frame *))__pyx_f_10fatslimlib_9core_base_5Frame_fast_get_lipid_coords_bbox;
  __pyx_vtable_10fatslimlib_9core_base_Frame.get_lipid_coords_bbox_aslist = (PyObject *(*)(struct __pyx_obj_10fatslimlib_9core_base_Frame *))__pyx_f_10fatslimlib_9core_base_5Frame_get_lipid_coords_bbox_aslist;
  __pyx_vtable_10fatslimlib_9core_base_Frame.fast_get_directions = (__Pyx_memviewslice (*)(struct __pyx_obj_10fatslimlib_9core_base_Frame *))__pyx_f_10fatslimlib_9core_base_5Frame_fast_get_directions;
  __pyx_vtable_10fatslimlib_9core_base_Frame.fast_get_normals = (__Pyx_memviewslice (*)(struct __pyx_obj_10fatslimlib_9core_base_Frame *, struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_fast_get_normals *__pyx_optional_args))__pyx_f_10fatslimlib_9core_base_5Frame_fast_get_normals;
  __pyx_vtable_10fatslimlib_9core_base_Frame.get_aggregates = (PyObject *(*)(struct __pyx_obj_10fatslimlib_9core_base_Frame *, int __pyx_skip_dispatch, struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_get_aggregates *__pyx_optional_args))__pyx_f_10fatslimlib_9core_base_5Frame_get_aggregates;
  __pyx_vtable_10fatslimlib_9core_base_Frame.get_membranes = (PyObject *(*)(struct __pyx_obj_10fatslimlib_9core_base_Frame *, int __pyx_skip_dispatch, struct __pyx_opt_args_10fatslimlib_9core_base_5Frame_get_membranes *__pyx_optional_args))__pyx_f_10fatslimlib_9core_base_5Frame_get_membranes;
  if (PyType_Ready(&__pyx_type_10fatslimlib_9core_base_Frame) < 0) __PYX_ERR(0, 1299, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10fatslimlib_9core_base_Frame.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10fatslimlib_9core_base_Frame.tp_dictoffset && __pyx_type_10fatslimlib_9core_base_Frame.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10fatslimlib_9core_base_Frame.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10fatslimlib_9core_base_Frame.tp_dict, __pyx_vtabptr_10fatslimlib_9core_base_Frame) < 0) __PYX_ERR(0, 1299, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Frame, (PyObject *)&__pyx_type_10fatslimlib_9core_base_Frame) < 0) __PYX_ERR(0, 1299, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10fatslimlib_9core_base_Frame) < 0) __PYX_ERR(0, 1299, __pyx_L1_error)
  __pyx_ptype_10fatslimlib_9core_base_Frame = &__pyx_type_10fatslimlib_9core_base_Frame;
  __pyx_vtabptr_10fatslimlib_9core_base_Trajectory = &__pyx_vtable_10fatslimlib_9core_base_Trajectory;
  __pyx_vtable_10fatslimlib_9core_base_Trajectory.assert_initialization = (PyObject *(*)(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *))__pyx_f_10fatslimlib_9core_base_10Trajectory_assert_initialization;
  __pyx_vtable_10fatslimlib_9core_base_Trajectory.get_forcefield_type = (fsl_int (*)(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *, int __pyx_skip_dispatch))__pyx_f_10fatslimlib_9core_base_10Trajectory_get_forcefield_type;
  __pyx_vtable_10fatslimlib_9core_base_Trajectory.get_rvdw = (real (*)(struct __pyx_obj_10fatslimlib_9core_base_Trajectory *, int __pyx_skip_dispatch))__pyx_f_10fatslimlib_9core_base_10Trajectory_get_rvdw;
  if (PyType_Ready(&__pyx_type_10fatslimlib_9core_base_Trajectory) < 0) __PYX_ERR(0, 1714, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10fatslimlib_9core_base_Trajectory.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10fatslimlib_9core_base_Trajectory.tp_dictoffset && __pyx_type_10fatslimlib_9core_base_Trajectory.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10fatslimlib_9core_base_Trajectory.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10fatslimlib_9core_base_Trajectory.tp_dict, __pyx_vtabptr_10fatslimlib_9core_base_Trajectory) < 0) __PYX_ERR(0, 1714, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Trajectory, (PyObject *)&__pyx_type_10fatslimlib_9core_base_Trajectory) < 0) __PYX_ERR(0, 1714, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10fatslimlib_9core_base_Trajectory) < 0) __PYX_ERR(0, 1714, __pyx_L1_error)
  __pyx_ptype_10fatslimlib_9core_base_Trajectory = &__pyx_type_10fatslimlib_9core_base_Trajectory;
  __pyx_vtabptr_array = &__pyx_vtable_array;
  __pyx_vtable_array.get_memview = (PyObject *(*)(struct __pyx_array_obj *))__pyx_array_get_memview;
  if (PyType_Ready(&__pyx_type___pyx_array) < 0) __PYX_ERR(1, 105, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_array.tp_print = 0;
  #endif
  if (__Pyx_SetVtable(__pyx_type___pyx_array.tp_dict, __pyx_vtabptr_array) < 0) __PYX_ERR(1, 105, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_array) < 0) __PYX_ERR(1, 105, __pyx_L1_error)
  __pyx_array_type = &__pyx_type___pyx_array;
  if (PyType_Ready(&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(1, 279, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_MemviewEnum.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_MemviewEnum.tp_dictoffset && __pyx_type___pyx_MemviewEnum.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_MemviewEnum.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(1, 279, __pyx_L1_error)
  __pyx_MemviewEnum_type = &__pyx_type___pyx_MemviewEnum;
  __pyx_vtabptr_memoryview = &__pyx_vtable_memoryview;
  __pyx_vtable_memoryview.get_item_pointer = (char *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_get_item_pointer;
  __pyx_vtable_memoryview.is_slice = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_is_slice;
  __pyx_vtable_memoryview.setitem_slice_assignment = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_slice_assignment;
  __pyx_vtable_memoryview.setitem_slice_assign_scalar = (PyObject *(*)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_setitem_slice_assign_scalar;
  __pyx_vtable_memoryview.setitem_indexed = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_indexed;
  __pyx_vtable_memoryview.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryview_convert_item_to_object;
  __pyx_vtable_memoryview.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryview_assign_item_from_object;
  if (PyType_Ready(&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(1, 330, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryview.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryview.tp_dictoffset && __pyx_type___pyx_memoryview.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryview.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryview.tp_dict, __pyx_vtabptr_memoryview) < 0) __PYX_ERR(1, 330, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(1, 330, __pyx_L1_error)
  __pyx_memoryview_type = &__pyx_type___pyx_memoryview;
  __pyx_vtabptr__memoryviewslice = &__pyx_vtable__memoryviewslice;
  __pyx_vtable__memoryviewslice.__pyx_base = *__pyx_vtabptr_memoryview;
  __pyx_vtable__memoryviewslice.__pyx_base.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryviewslice_convert_item_to_object;
  __pyx_vtable__memoryviewslice.__pyx_base.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryviewslice_assign_item_from_object;
  __pyx_type___pyx_memoryviewslice.tp_base = __pyx_memoryview_type;
  if (PyType_Ready(&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(1, 965, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryviewslice.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryviewslice.tp_dictoffset && __pyx_type___pyx_memoryviewslice.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryviewslice.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryviewslice.tp_dict, __pyx_vtabptr__memoryviewslice) < 0) __PYX_ERR(1, 965, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(1, 965, __pyx_L1_error)
  __pyx_memoryviewslice_type = &__pyx_type___pyx_memoryviewslice;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("fatslimlib.core_analysis"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10fatslimlib_13core_analysis_Aggregate = __Pyx_ImportType(__pyx_t_1, "fatslimlib.core_analysis", "Aggregate", sizeof(struct __pyx_obj_10fatslimlib_13core_analysis_Aggregate), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10fatslimlib_13core_analysis_Aggregate) __PYX_ERR(3, 29, __pyx_L1_error)
  __pyx_vtabptr_10fatslimlib_13core_analysis_Aggregate = (struct __pyx_vtabstruct_10fatslimlib_13core_analysis_Aggregate*)__Pyx_GetVtable(__pyx_ptype_10fatslimlib_13core_analysis_Aggregate->tp_dict); if (unlikely(!__pyx_vtabptr_10fatslimlib_13core_analysis_Aggregate)) __PYX_ERR(3, 29, __pyx_L1_error)
  __pyx_ptype_10fatslimlib_13core_analysis_Membrane = __Pyx_ImportType(__pyx_t_1, "fatslimlib.core_analysis", "Membrane", sizeof(struct __pyx_obj_10fatslimlib_13core_analysis_Membrane), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10fatslimlib_13core_analysis_Membrane) __PYX_ERR(3, 85, __pyx_L1_error)
  __pyx_vtabptr_10fatslimlib_13core_analysis_Membrane = (struct __pyx_vtabstruct_10fatslimlib_13core_analysis_Membrane*)__Pyx_GetVtable(__pyx_ptype_10fatslimlib_13core_analysis_Membrane->tp_dict); if (unlikely(!__pyx_vtabptr_10fatslimlib_13core_analysis_Membrane)) __PYX_ERR(3, 85, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("fatslimlib.core_ns"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "fast_neighbor_search", (void (**)(void))&__pyx_f_10fatslimlib_7core_ns_fast_neighbor_search, "struct __pyx_t_10fatslimlib_7core_ns_ns_neighborhood_holder *(__Pyx_memviewslice, __Pyx_memviewslice, struct __pyx_obj_10fatslimlib_9core_base_PBCBox *, real)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "free_neighborhood_holder", (void (**)(void))&__pyx_f_10fatslimlib_7core_ns_free_neighborhood_holder, "void (struct __pyx_t_10fatslimlib_7core_ns_ns_neighborhood_holder *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initcore_base(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initcore_base(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_core_base(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_core_base(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_core_base(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  static PyThread_type_lock __pyx_t_3[8];
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'core_base' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_core_base(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("core_base", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_fatslimlib__core_base) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "fatslimlib.core_base")) {
      if (unlikely(PyDict_SetItemString(modules, "fatslimlib.core_base", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) goto __pyx_L1_error;
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) goto __pyx_L1_error;
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  if (unlikely(__Pyx_modinit_variable_export_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_function_export_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() != 0)) goto __pyx_L1_error;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "fatslimlib/core_base.pyx":79
 * from .core_analysis cimport Aggregate, Membrane
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * from time import time
 * import os
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fatslimlib/core_base.pyx":80
 * 
 * import numpy as np
 * from time import time             # <<<<<<<<<<<<<<
 * import os
 * import sys
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_time);
  __Pyx_GIVEREF(__pyx_n_s_time);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_time);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_time, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_time, __pyx_t_1) < 0) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":81
 * import numpy as np
 * from time import time
 * import os             # <<<<<<<<<<<<<<
 * import sys
 * 
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_os, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_2) < 0) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":82
 * from time import time
 * import os
 * import sys             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_sys, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_2) < 0) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":86
 * 
 * # Constants
 * __cython_version__ = cython.__version__             # <<<<<<<<<<<<<<
 * 
 * # OpenMP-related stuff
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cython_version, __pyx_kp_s_0_29_14) < 0) __PYX_ERR(0, 86, __pyx_L1_error)

  /* "fatslimlib/core_base.pyx":89
 * 
 * # OpenMP-related stuff
 * cdef fsl_int OPENMP_MAX_THREADS = omp_get_max_threads()             # <<<<<<<<<<<<<<
 * cdef fsl_int OPENMP_NUM_THREADS = OPENMP_MAX_THREADS
 * 
 */
  __pyx_v_10fatslimlib_9core_base_OPENMP_MAX_THREADS = omp_get_max_threads();

  /* "fatslimlib/core_base.pyx":90
 * # OpenMP-related stuff
 * cdef fsl_int OPENMP_MAX_THREADS = omp_get_max_threads()
 * cdef fsl_int OPENMP_NUM_THREADS = OPENMP_MAX_THREADS             # <<<<<<<<<<<<<<
 * 
 * cpdef fsl_int get_num_threads():
 */
  __pyx_v_10fatslimlib_9core_base_OPENMP_NUM_THREADS = __pyx_v_10fatslimlib_9core_base_OPENMP_MAX_THREADS;

  /* "fatslimlib/core_base.pyx":158
 *     return memory
 * 
 * def verbose_print(msg, bint verbose, end="\n"):             # <<<<<<<<<<<<<<
 *     if verbose:
 *         print(msg, end=end)
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_10fatslimlib_9core_base_15verbose_print, NULL, __pyx_n_s_fatslimlib_core_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_verbose_print, __pyx_t_2) < 0) __PYX_ERR(0, 158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":1708
 * 
 * 
 * FORCEFIELDTYPE_NOTSET = fft_notset             # <<<<<<<<<<<<<<
 * FORCEFIELDTYPE_ALLATOM = fft_allatom
 * FORCEFIELDTYPE_UNIFIED = fft_unified
 */
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_10fatslimlib_9core_base_ForcefieldType(__pyx_e_10fatslimlib_9core_base_fft_notset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1708, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FORCEFIELDTYPE_NOTSET, __pyx_t_2) < 0) __PYX_ERR(0, 1708, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":1709
 * 
 * FORCEFIELDTYPE_NOTSET = fft_notset
 * FORCEFIELDTYPE_ALLATOM = fft_allatom             # <<<<<<<<<<<<<<
 * FORCEFIELDTYPE_UNIFIED = fft_unified
 * FORCEFIELDTYPE_COARSE = fft_coarse
 */
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_10fatslimlib_9core_base_ForcefieldType(__pyx_e_10fatslimlib_9core_base_fft_allatom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FORCEFIELDTYPE_ALLATOM, __pyx_t_2) < 0) __PYX_ERR(0, 1709, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":1710
 * FORCEFIELDTYPE_NOTSET = fft_notset
 * FORCEFIELDTYPE_ALLATOM = fft_allatom
 * FORCEFIELDTYPE_UNIFIED = fft_unified             # <<<<<<<<<<<<<<
 * FORCEFIELDTYPE_COARSE = fft_coarse
 * FORCEFIELDTYPE_UNKNOWN = fft_unknown
 */
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_10fatslimlib_9core_base_ForcefieldType(__pyx_e_10fatslimlib_9core_base_fft_unified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FORCEFIELDTYPE_UNIFIED, __pyx_t_2) < 0) __PYX_ERR(0, 1710, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":1711
 * FORCEFIELDTYPE_ALLATOM = fft_allatom
 * FORCEFIELDTYPE_UNIFIED = fft_unified
 * FORCEFIELDTYPE_COARSE = fft_coarse             # <<<<<<<<<<<<<<
 * FORCEFIELDTYPE_UNKNOWN = fft_unknown
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_10fatslimlib_9core_base_ForcefieldType(__pyx_e_10fatslimlib_9core_base_fft_coarse); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FORCEFIELDTYPE_COARSE, __pyx_t_2) < 0) __PYX_ERR(0, 1711, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":1712
 * FORCEFIELDTYPE_UNIFIED = fft_unified
 * FORCEFIELDTYPE_COARSE = fft_coarse
 * FORCEFIELDTYPE_UNKNOWN = fft_unknown             # <<<<<<<<<<<<<<
 * 
 * cdef class Trajectory(object):
 */
  __pyx_t_2 = __Pyx_PyInt_From_enum____pyx_t_10fatslimlib_9core_base_ForcefieldType(__pyx_e_10fatslimlib_9core_base_fft_unknown); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FORCEFIELDTYPE_UNKNOWN, __pyx_t_2) < 0) __PYX_ERR(0, 1712, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_TopologyReader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_10fatslimlib_9core_base_17__pyx_unpickle_TopologyReader, NULL, __pyx_n_s_fatslimlib_core_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_TopologyReader, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_TopologyReader__set_state(<TopologyReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_TopologyReader__set_state(TopologyReader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.filename = __pyx_state[0]; __pyx_result.topology = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_10fatslimlib_9core_base_19__pyx_unpickle_CoordinateReader, NULL, __pyx_n_s_fatslimlib_core_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_CoordinateReader, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IndexReader(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_10fatslimlib_9core_base_21__pyx_unpickle_IndexReader, NULL, __pyx_n_s_fatslimlib_core_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_IndexReader, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_IndexReader__set_state(<IndexReader> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IndexReader__set_state(IndexReader __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.filename = __pyx_state[0]; __pyx_result.groups = __pyx_state[1]; __pyx_result.loaded = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_10fatslimlib_9core_base_23__pyx_unpickle_Trajectory, NULL, __pyx_n_s_fatslimlib_core_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Trajectory, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fatslimlib/core_base.pyx":1
 * # -*- coding: utf-8; Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-             # <<<<<<<<<<<<<<
 * # vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
 * #
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":209
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(array self):
 */
  __pyx_t_2 = __pyx_capsule_create(((void *)(&__pyx_array_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_array_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_2) < 0) __PYX_ERR(1, 209, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_array_type);

  /* "View.MemoryView":286
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__56, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(generic);
  __Pyx_DECREF_SET(generic, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":287
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__57, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(strided);
  __Pyx_DECREF_SET(strided, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":288
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__58, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(indirect);
  __Pyx_DECREF_SET(indirect, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":291
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__59, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(contiguous);
  __Pyx_DECREF_SET(contiguous, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":292
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__60, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(indirect_contiguous);
  __Pyx_DECREF_SET(indirect_contiguous, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":316
 * 
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0             # <<<<<<<<<<<<<<
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [
 *     PyThread_allocate_lock(),
 */
  __pyx_memoryview_thread_locks_used = 0;

  /* "View.MemoryView":317
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [             # <<<<<<<<<<<<<<
 *     PyThread_allocate_lock(),
 *     PyThread_allocate_lock(),
 */
  __pyx_t_3[0] = PyThread_allocate_lock();
  __pyx_t_3[1] = PyThread_allocate_lock();
  __pyx_t_3[2] = PyThread_allocate_lock();
  __pyx_t_3[3] = PyThread_allocate_lock();
  __pyx_t_3[4] = PyThread_allocate_lock();
  __pyx_t_3[5] = PyThread_allocate_lock();
  __pyx_t_3[6] = PyThread_allocate_lock();
  __pyx_t_3[7] = PyThread_allocate_lock();
  memcpy(&(__pyx_memoryview_thread_locks[0]), __pyx_t_3, sizeof(__pyx_memoryview_thread_locks[0]) * (8));

  /* "View.MemoryView":549
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_memoryview_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_2) < 0) __PYX_ERR(1, 549, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_memoryview_type);

  /* "View.MemoryView":995
 *         return self.from_object
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_memoryviewslice_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_2) < 0) __PYX_ERR(1, 995, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_memoryviewslice_type);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum, NULL, __pyx_n_s_View_MemoryView); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Enum, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init fatslimlib.core_base", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init fatslimlib.core_base");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_NeObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_FALSE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a != b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_TRUE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_TRUE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal != 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a != (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* PyIntFromDouble */
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE PyObject* __Pyx_PyInt_FromDouble(double value) {
    if (value >= (double)LONG_MIN && value <= (double)LONG_MAX) {
        return PyInt_FromLong((long)value);
    }
    return PyLong_FromDouble(value);
}
#endif

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* dict_setdefault */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value,
                                                       CYTHON_UNUSED int is_safe_type) {
    PyObject* value;
#if PY_VERSION_HEX >= 0x030400A0
    if ((1)) {
        value = PyDict_SetDefault(d, key, default_value);
        if (unlikely(!value)) return NULL;
        Py_INCREF(value);
#else
    if (is_safe_type == 1 || (is_safe_type == -1 &&
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
            (PyUnicode_CheckExact(key) || PyString_CheckExact(key) || PyLong_CheckExact(key)))) {
        value = PyDict_GetItemWithError(d, key);
        if (unlikely(!value)) {
            if (unlikely(PyErr_Occurred()))
                return NULL;
            if (unlikely(PyDict_SetItem(d, key, default_value) == -1))
                return NULL;
            value = default_value;
        }
        Py_INCREF(value);
#else
            (PyString_CheckExact(key) || PyUnicode_CheckExact(key) || PyInt_CheckExact(key) || PyLong_CheckExact(key)))) {
        value = PyDict_GetItem(d, key);
        if (unlikely(!value)) {
            if (unlikely(PyDict_SetItem(d, key, default_value) == -1))
                return NULL;
            value = default_value;
        }
        Py_INCREF(value);
#endif
#endif
    } else {
        value = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_setdefault, d, key, default_value);
    }
    return value;
}

/* PyIntBinop */
    #if !CYTHON_COMPILING_IN_PYPY
#if PY_MAJOR_VERSION < 3 || CYTHON_USE_PYLONG_INTERNALS
#define __Pyx_PyInt_FloorDivideObjC_ZeroDivisionError(operand)\
    if (unlikely(zerodivision_check && ((operand) == 0))) {\
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division by zero");\
        return NULL;\
    }
#endif
static PyObject* __Pyx_PyInt_FloorDivideObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            __Pyx_PyInt_FloorDivideObjC_ZeroDivisionError(b)
            if (unlikely(b == -1 && ((unsigned long)a) == 0-(unsigned long)a))
                return PyInt_Type.tp_as_number->nb_floor_divide(op1, op2);
            else {
                long q, r;
                q = a / b;
                r = a - q*b;
                q -= ((r != 0) & ((r ^ b) < 0));
                x = q;
            }
            return PyInt_FromLong(x);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_floor_divide(op1, op2);
            }
        }
                __Pyx_PyInt_FloorDivideObjC_ZeroDivisionError(b)
                {
                    long q, r;
                    q = a / b;
                    r = a - q*b;
                    q -= ((r != 0) & ((r ^ b) < 0));
                    x = q;
                }
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                {
                    PY_LONG_LONG q, r;
                    q = lla / llb;
                    r = lla - q*llb;
                    q -= ((r != 0) & ((r ^ llb) < 0));
                    llx = q;
                }
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    return (inplace ? PyNumber_InPlaceFloorDivide : PyNumber_FloorDivide)(op1, op2);
}
#endif

/* RaiseTooManyValuesToUnpack */
    static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
    static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
    static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
    static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* PyObjectFormat */
    #if CYTHON_USE_UNICODE_WRITER
static PyObject* __Pyx_PyObject_Format(PyObject* obj, PyObject* format_spec) {
    int ret;
    _PyUnicodeWriter writer;
    if (likely(PyFloat_CheckExact(obj))) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x03040000
        _PyUnicodeWriter_Init(&writer, 0);
#else
        _PyUnicodeWriter_Init(&writer);
#endif
        ret = _PyFloat_FormatAdvancedWriter(
            &writer,
            obj,
            format_spec, 0, PyUnicode_GET_LENGTH(format_spec));
    } else if (likely(PyLong_CheckExact(obj))) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x03040000
        _PyUnicodeWriter_Init(&writer, 0);
#else
        _PyUnicodeWriter_Init(&writer);
#endif
        ret = _PyLong_FormatAdvancedWriter(
            &writer,
            obj,
            format_spec, 0, PyUnicode_GET_LENGTH(format_spec));
    } else {
        return PyObject_Format(obj, format_spec);
    }
    if (unlikely(ret == -1)) {
        _PyUnicodeWriter_Dealloc(&writer);
        return NULL;
    }
    return _PyUnicodeWriter_Finish(&writer);
}
#endif

/* PyObjectFormatAndDecref */
    static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyString_CheckExact(s))) {
        PyObject *result = PyUnicode_FromEncodedObject(s, NULL, "strict");
        Py_DECREF(s);
        return result;
    }
    #endif
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* JoinPyUnicode */
    static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      CYTHON_UNUSED Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind;
    Py_ssize_t i, char_pos;
    void *result_udata;
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely(char_pos + ulength < 0))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + char_pos * result_ukind, udata, (size_t) (ulength * result_ukind));
        } else {
            #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    result_ulength++;
    value_count++;
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* PyObjectCallNoArg */
    #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyObjectCall2Args */
    static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* GetItemInt */
    static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* GetTopmostException */
    #if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
    #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
    #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* WriteUnraisableException */
    static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyDictVersioning */
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
    #if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* ArgTypeTest */
    static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* decode_c_bytes */
    static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    if (unlikely((start < 0) | (stop < 0))) {
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (stop > length)
        stop = length;
    length = stop - start;
    if (unlikely(length <= 0))
        return PyUnicode_FromUnicode(NULL, 0);
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* PyObjectGetMethod */
    static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
    static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* append */
    static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* MemviewSliceInit */
    static int
__Pyx_init_memviewslice(struct __pyx_memoryview_obj *memview,
                        int ndim,
                        __Pyx_memviewslice *memviewslice,
                        int memview_is_new_reference)
{
    __Pyx_RefNannyDeclarations
    int i, retval=-1;
    Py_buffer *buf = &memview->view;
    __Pyx_RefNannySetupContext("init_memviewslice", 0);
    if (memviewslice->memview || memviewslice->data) {
        PyErr_SetString(PyExc_ValueError,
            "memviewslice is already initialized!");
        goto fail;
    }
    if (buf->strides) {
        for (i = 0; i < ndim; i++) {
            memviewslice->strides[i] = buf->strides[i];
        }
    } else {
        Py_ssize_t stride = buf->itemsize;
        for (i = ndim - 1; i >= 0; i--) {
            memviewslice->strides[i] = stride;
            stride *= buf->shape[i];
        }
    }
    for (i = 0; i < ndim; i++) {
        memviewslice->shape[i]   = buf->shape[i];
        if (buf->suboffsets) {
            memviewslice->suboffsets[i] = buf->suboffsets[i];
        } else {
            memviewslice->suboffsets[i] = -1;
        }
    }
    memviewslice->memview = memview;
    memviewslice->data = (char *)buf->buf;
    if (__pyx_add_acquisition_count(memview) == 0 && !memview_is_new_reference) {
        Py_INCREF(memview);
    }
    retval = 0;
    goto no_fail;
fail:
    memviewslice->memview = 0;
    memviewslice->data = 0;
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}
#ifndef Py_NO_RETURN
#define Py_NO_RETURN
#endif
static void __pyx_fatalerror(const char *fmt, ...) Py_NO_RETURN {
    va_list vargs;
    char msg[200];
#ifdef HAVE_STDARG_PROTOTYPES
    va_start(vargs, fmt);
#else
    va_start(vargs);
#endif
    vsnprintf(msg, 200, fmt, vargs);
    va_end(vargs);
    Py_FatalError(msg);
}
static CYTHON_INLINE int
__pyx_add_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)++;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE int
__pyx_sub_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)--;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE void
__Pyx_INC_MEMVIEW(__Pyx_memviewslice *memslice, int have_gil, int lineno)
{
    int first_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (!memview || (PyObject *) memview == Py_None)
        return;
    if (__pyx_get_slice_count(memview) < 0)
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    first_time = __pyx_add_acquisition_count(memview) == 0;
    if (first_time) {
        if (have_gil) {
            Py_INCREF((PyObject *) memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_INCREF((PyObject *) memview);
            PyGILState_Release(_gilstate);
        }
    }
}
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *memslice,
                                             int have_gil, int lineno) {
    int last_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (!memview ) {
        return;
    } else if ((PyObject *) memview == Py_None) {
        memslice->memview = NULL;
        return;
    }
    if (__pyx_get_slice_count(memview) <= 0)
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    last_time = __pyx_sub_acquisition_count(memview) == 1;
    memslice->data = NULL;
    if (last_time) {
        if (have_gil) {
            Py_CLEAR(memslice->memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_CLEAR(memslice->memview);
            PyGILState_Release(_gilstate);
        }
    } else {
        memslice->memview = NULL;
    }
}

/* GetAttr */
    static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* SetItemInt */
    static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* None */
    static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* KeywordStringCheck */
    static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* CallUnboundCMethod0 */
    static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* PyIntBinop */
    #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* decode_c_string */
    static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    length = stop - start;
    if (unlikely(length <= 0))
        return PyUnicode_FromUnicode(NULL, 0);
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* PyErrExceptionMatches */
    #if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr3 */
    static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* ObjectGetItem */
    #if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* Import */
    static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
    static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* PyUnicode_Unicode */
    static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj) {
    if (unlikely(obj == Py_None))
        obj = __pyx_kp_u_None;
    return __Pyx_NewRef(obj);
}

/* DictGetItem */
    #if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* ExtTypeTest */
    static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* HasAttr */
    static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* BytesEquals */
    static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
    static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* RaiseNoneIterError */
    static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* SwapException */
    #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* FastTypeChecks */
    #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* PyObject_GenericGetAttrNoDict */
    #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
    #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetupReduce */
    static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* SetVTable */
    static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* TypeImport */
    #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
    static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* CLineInTraceback */
    #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
    static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
    #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_array_type)) return __pyx_array_getbuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_memoryview_type)) return __pyx_memoryview_getbuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


    /* MemviewSliceIsContig */
    static int
__pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim)
{
    int i, index, step, start;
    Py_ssize_t itemsize = mvs.memview->view.itemsize;
    if (order == 'F') {
        step = 1;
        start = 0;
    } else {
        step = -1;
        start = ndim - 1;
    }
    for (i = 0; i < ndim; i++) {
        index = start + step * i;
        if (mvs.suboffsets[index] >= 0 || mvs.strides[index] != itemsize)
            return 0;
        itemsize *= mvs.shape[index];
    }
    return 1;
}

/* OverlappingSlices */
    static void
__pyx_get_array_memory_extents(__Pyx_memviewslice *slice,
                               void **out_start, void **out_end,
                               int ndim, size_t itemsize)
{
    char *start, *end;
    int i;
    start = end = slice->data;
    for (i = 0; i < ndim; i++) {
        Py_ssize_t stride = slice->strides[i];
        Py_ssize_t extent = slice->shape[i];
        if (extent == 0) {
            *out_start = *out_end = start;
            return;
        } else {
            if (stride > 0)
                end += stride * (extent - 1);
            else
                start += stride * (extent - 1);
        }
    }
    *out_start = start;
    *out_end = end + itemsize;
}
static int
__pyx_slices_overlap(__Pyx_memviewslice *slice1,
                     __Pyx_memviewslice *slice2,
                     int ndim, size_t itemsize)
{
    void *start1, *end1, *start2, *end2;
    __pyx_get_array_memory_extents(slice1, &start1, &end1, ndim, itemsize);
    __pyx_get_array_memory_extents(slice2, &start2, &end2, ndim, itemsize);
    return (start1 < end2) && (start2 < end1);
}

/* Capsule */
    static CYTHON_INLINE PyObject *
__pyx_capsule_create(void *p, CYTHON_UNUSED const char *sig)
{
    PyObject *cobj;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, NULL);
#else
    cobj = PyCObject_FromVoidPtr(p, NULL);
#endif
    return cobj;
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_fsl_int(fsl_int value) {
    const fsl_int neg_one = (fsl_int) ((fsl_int) 0 - (fsl_int) 1), const_zero = (fsl_int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(fsl_int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(fsl_int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(fsl_int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(fsl_int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(fsl_int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(fsl_int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
    #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* MemviewDtypeToObject */
    static CYTHON_INLINE PyObject *__pyx_memview_get_nn_fsl_int(const char *itemp) {
    return (PyObject *) __Pyx_PyInt_From_fsl_int(*(fsl_int *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_nn_fsl_int(const char *itemp, PyObject *obj) {
    fsl_int value = __Pyx_PyInt_As_fsl_int(obj);
    if ((value == ((fsl_int)-1)) && PyErr_Occurred())
        return 0;
    *(fsl_int *) itemp = value;
    return 1;
}

/* IsLittleEndian */
    static CYTHON_INLINE int __Pyx_Is_Little_Endian(void)
{
  union {
    uint32_t u32;
    uint8_t u8[4];
  } S;
  S.u32 = 0x01020304;
  return S.u8[0] == 4;
}

/* BufferFormatCheck */
    static void __Pyx_BufFmt_Init(__Pyx_BufFmt_Context* ctx,
                              __Pyx_BufFmt_StackElem* stack,
                              __Pyx_TypeInfo* type) {
  stack[0].field = &ctx->root;
  stack[0].parent_offset = 0;
  ctx->root.type = type;
  ctx->root.name = "buffer dtype";
  ctx->root.offset = 0;
  ctx->head = stack;
  ctx->head->field = &ctx->root;
  ctx->fmt_offset = 0;
  ctx->head->parent_offset = 0;
  ctx->new_packmode = '@';
  ctx->enc_packmode = '@';
  ctx->new_count = 1;
  ctx->enc_count = 0;
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  ctx->is_valid_array = 0;
  ctx->struct_alignment = 0;
  while (type->typegroup == 'S') {
    ++ctx->head;
    ctx->head->field = type->fields;
    ctx->head->parent_offset = 0;
    type = type->fields->type;
  }
}
static int __Pyx_BufFmt_ParseNumber(const char** ts) {
    int count;
    const char* t = *ts;
    if (*t < '0' || *t > '9') {
      return -1;
    } else {
        count = *t++ - '0';
        while (*t >= '0' && *t <= '9') {
            count *= 10;
            count += *t++ - '0';
        }
    }
    *ts = t;
    return count;
}
static int __Pyx_BufFmt_ExpectNumber(const char **ts) {
    int number = __Pyx_BufFmt_ParseNumber(ts);
    if (number == -1)
        PyErr_Format(PyExc_ValueError,\
                     "Does not understand character buffer dtype format string ('%c')", **ts);
    return number;
}
static void __Pyx_BufFmt_RaiseUnexpectedChar(char ch) {
  PyErr_Format(PyExc_ValueError,
               "Unexpected format string character: '%c'", ch);
}
static const char* __Pyx_BufFmt_DescribeTypeChar(char ch, int is_complex) {
  switch (ch) {
    case '?': return "'bool'";
    case 'c': return "'char'";
    case 'b': return "'signed char'";
    case 'B': return "'unsigned char'";
    case 'h': return "'short'";
    case 'H': return "'unsigned short'";
    case 'i': return "'int'";
    case 'I': return "'unsigned int'";
    case 'l': return "'long'";
    case 'L': return "'unsigned long'";
    case 'q': return "'long long'";
    case 'Q': return "'unsigned long long'";
    case 'f': return (is_complex ? "'complex float'" : "'float'");
    case 'd': return (is_complex ? "'complex double'" : "'double'");
    case 'g': return (is_complex ? "'complex long double'" : "'long double'");
    case 'T': return "a struct";
    case 'O': return "Python object";
    case 'P': return "a pointer";
    case 's': case 'p': return "a string";
    case 0: return "end";
    default: return "unparseable format string";
  }
}
static size_t __Pyx_BufFmt_TypeCharToStandardSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return 2;
    case 'i': case 'I': case 'l': case 'L': return 4;
    case 'q': case 'Q': return 8;
    case 'f': return (is_complex ? 8 : 4);
    case 'd': return (is_complex ? 16 : 8);
    case 'g': {
      PyErr_SetString(PyExc_ValueError, "Python does not define a standard format string size for long double ('g')..");
      return 0;
    }
    case 'O': case 'P': return sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static size_t __Pyx_BufFmt_TypeCharToNativeSize(char ch, int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(short);
    case 'i': case 'I': return sizeof(int);
    case 'l': case 'L': return sizeof(long);
    #ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(PY_LONG_LONG);
    #endif
    case 'f': return sizeof(float) * (is_complex ? 2 : 1);
    case 'd': return sizeof(double) * (is_complex ? 2 : 1);
    case 'g': return sizeof(long double) * (is_complex ? 2 : 1);
    case 'O': case 'P': return sizeof(void*);
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
typedef struct { char c; short x; } __Pyx_st_short;
typedef struct { char c; int x; } __Pyx_st_int;
typedef struct { char c; long x; } __Pyx_st_long;
typedef struct { char c; float x; } __Pyx_st_float;
typedef struct { char c; double x; } __Pyx_st_double;
typedef struct { char c; long double x; } __Pyx_st_longdouble;
typedef struct { char c; void *x; } __Pyx_st_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { char c; PY_LONG_LONG x; } __Pyx_st_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToAlignment(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_st_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_st_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_st_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_st_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_st_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_st_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_st_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_st_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
/* These are for computing the padding at the end of the struct to align
   on the first member of the struct. This will probably the same as above,
   but we don't have any guarantees.
 */
typedef struct { short x; char c; } __Pyx_pad_short;
typedef struct { int x; char c; } __Pyx_pad_int;
typedef struct { long x; char c; } __Pyx_pad_long;
typedef struct { float x; char c; } __Pyx_pad_float;
typedef struct { double x; char c; } __Pyx_pad_double;
typedef struct { long double x; char c; } __Pyx_pad_longdouble;
typedef struct { void *x; char c; } __Pyx_pad_void_p;
#ifdef HAVE_LONG_LONG
typedef struct { PY_LONG_LONG x; char c; } __Pyx_pad_longlong;
#endif
static size_t __Pyx_BufFmt_TypeCharToPadding(char ch, CYTHON_UNUSED int is_complex) {
  switch (ch) {
    case '?': case 'c': case 'b': case 'B': case 's': case 'p': return 1;
    case 'h': case 'H': return sizeof(__Pyx_pad_short) - sizeof(short);
    case 'i': case 'I': return sizeof(__Pyx_pad_int) - sizeof(int);
    case 'l': case 'L': return sizeof(__Pyx_pad_long) - sizeof(long);
#ifdef HAVE_LONG_LONG
    case 'q': case 'Q': return sizeof(__Pyx_pad_longlong) - sizeof(PY_LONG_LONG);
#endif
    case 'f': return sizeof(__Pyx_pad_float) - sizeof(float);
    case 'd': return sizeof(__Pyx_pad_double) - sizeof(double);
    case 'g': return sizeof(__Pyx_pad_longdouble) - sizeof(long double);
    case 'P': case 'O': return sizeof(__Pyx_pad_void_p) - sizeof(void*);
    default:
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
}
static char __Pyx_BufFmt_TypeCharToGroup(char ch, int is_complex) {
  switch (ch) {
    case 'c':
        return 'H';
    case 'b': case 'h': case 'i':
    case 'l': case 'q': case 's': case 'p':
        return 'I';
    case '?': case 'B': case 'H': case 'I': case 'L': case 'Q':
        return 'U';
    case 'f': case 'd': case 'g':
        return (is_complex ? 'C' : 'R');
    case 'O':
        return 'O';
    case 'P':
        return 'P';
    default: {
      __Pyx_BufFmt_RaiseUnexpectedChar(ch);
      return 0;
    }
  }
}
static void __Pyx_BufFmt_RaiseExpected(__Pyx_BufFmt_Context* ctx) {
  if (ctx->head == NULL || ctx->head->field == &ctx->root) {
    const char* expected;
    const char* quote;
    if (ctx->head == NULL) {
      expected = "end";
      quote = "";
    } else {
      expected = ctx->head->field->type->name;
      quote = "'";
    }
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected %s%s%s but got %s",
                 quote, expected, quote,
                 __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex));
  } else {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_StructField* parent = (ctx->head - 1)->field;
    PyErr_Format(PyExc_ValueError,
                 "Buffer dtype mismatch, expected '%s' but got %s in '%s.%s'",
                 field->type->name, __Pyx_BufFmt_DescribeTypeChar(ctx->enc_type, ctx->is_complex),
                 parent->type->name, field->name);
  }
}
static int __Pyx_BufFmt_ProcessTypeChunk(__Pyx_BufFmt_Context* ctx) {
  char group;
  size_t size, offset, arraysize = 1;
  if (ctx->enc_type == 0) return 0;
  if (ctx->head->field->type->arraysize[0]) {
    int i, ndim = 0;
    if (ctx->enc_type == 's' || ctx->enc_type == 'p') {
        ctx->is_valid_array = ctx->head->field->type->ndim == 1;
        ndim = 1;
        if (ctx->enc_count != ctx->head->field->type->arraysize[0]) {
            PyErr_Format(PyExc_ValueError,
                         "Expected a dimension of size %zu, got %zu",
                         ctx->head->field->type->arraysize[0], ctx->enc_count);
            return -1;
        }
    }
    if (!ctx->is_valid_array) {
      PyErr_Format(PyExc_ValueError, "Expected %d dimensions, got %d",
                   ctx->head->field->type->ndim, ndim);
      return -1;
    }
    for (i = 0; i < ctx->head->field->type->ndim; i++) {
      arraysize *= ctx->head->field->type->arraysize[i];
    }
    ctx->is_valid_array = 0;
    ctx->enc_count = 1;
  }
  group = __Pyx_BufFmt_TypeCharToGroup(ctx->enc_type, ctx->is_complex);
  do {
    __Pyx_StructField* field = ctx->head->field;
    __Pyx_TypeInfo* type = field->type;
    if (ctx->enc_packmode == '@' || ctx->enc_packmode == '^') {
      size = __Pyx_BufFmt_TypeCharToNativeSize(ctx->enc_type, ctx->is_complex);
    } else {
      size = __Pyx_BufFmt_TypeCharToStandardSize(ctx->enc_type, ctx->is_complex);
    }
    if (ctx->enc_packmode == '@') {
      size_t align_at = __Pyx_BufFmt_TypeCharToAlignment(ctx->enc_type, ctx->is_complex);
      size_t align_mod_offset;
      if (align_at == 0) return -1;
      align_mod_offset = ctx->fmt_offset % align_at;
      if (align_mod_offset > 0) ctx->fmt_offset += align_at - align_mod_offset;
      if (ctx->struct_alignment == 0)
          ctx->struct_alignment = __Pyx_BufFmt_TypeCharToPadding(ctx->enc_type,
                                                                 ctx->is_complex);
    }
    if (type->size != size || type->typegroup != group) {
      if (type->typegroup == 'C' && type->fields != NULL) {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        ++ctx->head;
        ctx->head->field = type->fields;
        ctx->head->parent_offset = parent_offset;
        continue;
      }
      if ((type->typegroup == 'H' || group == 'H') && type->size == size) {
      } else {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
      }
    }
    offset = ctx->head->parent_offset + field->offset;
    if (ctx->fmt_offset != offset) {
      PyErr_Format(PyExc_ValueError,
                   "Buffer dtype mismatch; next field is at offset %" CYTHON_FORMAT_SSIZE_T "d but %" CYTHON_FORMAT_SSIZE_T "d expected",
                   (Py_ssize_t)ctx->fmt_offset, (Py_ssize_t)offset);
      return -1;
    }
    ctx->fmt_offset += size;
    if (arraysize)
      ctx->fmt_offset += (arraysize - 1) * size;
    --ctx->enc_count;
    while (1) {
      if (field == &ctx->root) {
        ctx->head = NULL;
        if (ctx->enc_count != 0) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return -1;
        }
        break;
      }
      ctx->head->field = ++field;
      if (field->type == NULL) {
        --ctx->head;
        field = ctx->head->field;
        continue;
      } else if (field->type->typegroup == 'S') {
        size_t parent_offset = ctx->head->parent_offset + field->offset;
        if (field->type->fields->type == NULL) continue;
        field = field->type->fields;
        ++ctx->head;
        ctx->head->field = field;
        ctx->head->parent_offset = parent_offset;
        break;
      } else {
        break;
      }
    }
  } while (ctx->enc_count);
  ctx->enc_type = 0;
  ctx->is_complex = 0;
  return 0;
}
static PyObject *
__pyx_buffmt_parse_array(__Pyx_BufFmt_Context* ctx, const char** tsp)
{
    const char *ts = *tsp;
    int i = 0, number;
    int ndim = ctx->head->field->type->ndim;
;
    ++ts;
    if (ctx->new_count != 1) {
        PyErr_SetString(PyExc_ValueError,
                        "Cannot handle repeated arrays in format string");
        return NULL;
    }
    if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
    while (*ts && *ts != ')') {
        switch (*ts) {
            case ' ': case '\f': case '\r': case '\n': case '\t': case '\v':  continue;
            default:  break;
        }
        number = __Pyx_BufFmt_ExpectNumber(&ts);
        if (number == -1) return NULL;
        if (i < ndim && (size_t) number != ctx->head->field->type->arraysize[i])
            return PyErr_Format(PyExc_ValueError,
                        "Expected a dimension of size %zu, got %d",
                        ctx->head->field->type->arraysize[i], number);
        if (*ts != ',' && *ts != ')')
            return PyErr_Format(PyExc_ValueError,
                                "Expected a comma in format string, got '%c'", *ts);
        if (*ts == ',') ts++;
        i++;
    }
    if (i != ndim)
        return PyErr_Format(PyExc_ValueError, "Expected %d dimension(s), got %d",
                            ctx->head->field->type->ndim, i);
    if (!*ts) {
        PyErr_SetString(PyExc_ValueError,
                        "Unexpected end of format string, expected ')'");
        return NULL;
    }
    ctx->is_valid_array = 1;
    ctx->new_count = 1;
    *tsp = ++ts;
    return Py_None;
}
static const char* __Pyx_BufFmt_CheckString(__Pyx_BufFmt_Context* ctx, const char* ts) {
  int got_Z = 0;
  while (1) {
    switch(*ts) {
      case 0:
        if (ctx->enc_type != 0 && ctx->head == NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        if (ctx->head != NULL) {
          __Pyx_BufFmt_RaiseExpected(ctx);
          return NULL;
        }
        return ts;
      case ' ':
      case '\r':
      case '\n':
        ++ts;
        break;
      case '<':
        if (!__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Little-endian buffer not supported on big-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '>':
      case '!':
        if (__Pyx_Is_Little_Endian()) {
          PyErr_SetString(PyExc_ValueError, "Big-endian buffer not supported on little-endian compiler");
          return NULL;
        }
        ctx->new_packmode = '=';
        ++ts;
        break;
      case '=':
      case '@':
      case '^':
        ctx->new_packmode = *ts++;
        break;
      case 'T':
        {
          const char* ts_after_sub;
          size_t i, struct_count = ctx->new_count;
          size_t struct_alignment = ctx->struct_alignment;
          ctx->new_count = 1;
          ++ts;
          if (*ts != '{') {
            PyErr_SetString(PyExc_ValueError, "Buffer acquisition: Expected '{' after 'T'");
            return NULL;
          }
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          ctx->enc_count = 0;
          ctx->struct_alignment = 0;
          ++ts;
          ts_after_sub = ts;
          for (i = 0; i != struct_count; ++i) {
            ts_after_sub = __Pyx_BufFmt_CheckString(ctx, ts);
            if (!ts_after_sub) return NULL;
          }
          ts = ts_after_sub;
          if (struct_alignment) ctx->struct_alignment = struct_alignment;
        }
        break;
      case '}':
        {
          size_t alignment = ctx->struct_alignment;
          ++ts;
          if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
          ctx->enc_type = 0;
          if (alignment && ctx->fmt_offset % alignment) {
            ctx->fmt_offset += alignment - (ctx->fmt_offset % alignment);
          }
        }
        return ts;
      case 'x':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->fmt_offset += ctx->new_count;
        ctx->new_count = 1;
        ctx->enc_count = 0;
        ctx->enc_type = 0;
        ctx->enc_packmode = ctx->new_packmode;
        ++ts;
        break;
      case 'Z':
        got_Z = 1;
        ++ts;
        if (*ts != 'f' && *ts != 'd' && *ts != 'g') {
          __Pyx_BufFmt_RaiseUnexpectedChar('Z');
          return NULL;
        }
        CYTHON_FALLTHROUGH;
      case '?': case 'c': case 'b': case 'B': case 'h': case 'H': case 'i': case 'I':
      case 'l': case 'L': case 'q': case 'Q':
      case 'f': case 'd': case 'g':
      case 'O': case 'p':
        if (ctx->enc_type == *ts && got_Z == ctx->is_complex &&
            ctx->enc_packmode == ctx->new_packmode) {
          ctx->enc_count += ctx->new_count;
          ctx->new_count = 1;
          got_Z = 0;
          ++ts;
          break;
        }
        CYTHON_FALLTHROUGH;
      case 's':
        if (__Pyx_BufFmt_ProcessTypeChunk(ctx) == -1) return NULL;
        ctx->enc_count = ctx->new_count;
        ctx->enc_packmode = ctx->new_packmode;
        ctx->enc_type = *ts;
        ctx->is_complex = got_Z;
        ++ts;
        ctx->new_count = 1;
        got_Z = 0;
        break;
      case ':':
        ++ts;
        while(*ts != ':') ++ts;
        ++ts;
        break;
      case '(':
        if (!__pyx_buffmt_parse_array(ctx, &ts)) return NULL;
        break;
      default:
        {
          int number = __Pyx_BufFmt_ExpectNumber(&ts);
          if (number == -1) return NULL;
          ctx->new_count = (size_t)number;
        }
    }
  }
}

/* TypeInfoCompare */
      static int
__pyx_typeinfo_cmp(__Pyx_TypeInfo *a, __Pyx_TypeInfo *b)
{
    int i;
    if (!a || !b)
        return 0;
    if (a == b)
        return 1;
    if (a->size != b->size || a->typegroup != b->typegroup ||
            a->is_unsigned != b->is_unsigned || a->ndim != b->ndim) {
        if (a->typegroup == 'H' || b->typegroup == 'H') {
            return a->size == b->size;
        } else {
            return 0;
        }
    }
    if (a->ndim) {
        for (i = 0; i < a->ndim; i++)
            if (a->arraysize[i] != b->arraysize[i])
                return 0;
    }
    if (a->typegroup == 'S') {
        if (a->flags != b->flags)
            return 0;
        if (a->fields || b->fields) {
            if (!(a->fields && b->fields))
                return 0;
            for (i = 0; a->fields[i].type && b->fields[i].type; i++) {
                __Pyx_StructField *field_a = a->fields + i;
                __Pyx_StructField *field_b = b->fields + i;
                if (field_a->offset != field_b->offset ||
                    !__pyx_typeinfo_cmp(field_a->type, field_b->type))
                    return 0;
            }
            return !a->fields[i].type && !b->fields[i].type;
        }
    }
    return 1;
}

/* MemviewSliceValidateAndInit */
      static int
__pyx_check_strides(Py_buffer *buf, int dim, int ndim, int spec)
{
    if (buf->shape[dim] <= 1)
        return 1;
    if (buf->strides) {
        if (spec & __Pyx_MEMVIEW_CONTIG) {
            if (spec & (__Pyx_MEMVIEW_PTR|__Pyx_MEMVIEW_FULL)) {
                if (buf->strides[dim] != sizeof(void *)) {
                    PyErr_Format(PyExc_ValueError,
                                 "Buffer is not indirectly contiguous "
                                 "in dimension %d.", dim);
                    goto fail;
                }
            } else if (buf->strides[dim] != buf->itemsize) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
        if (spec & __Pyx_MEMVIEW_FOLLOW) {
            Py_ssize_t stride = buf->strides[dim];
            if (stride < 0)
                stride = -stride;
            if (stride < buf->itemsize) {
                PyErr_SetString(PyExc_ValueError,
                                "Buffer and memoryview are not contiguous "
                                "in the same dimension.");
                goto fail;
            }
        }
    } else {
        if (spec & __Pyx_MEMVIEW_CONTIG && dim != ndim - 1) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not contiguous in "
                         "dimension %d", dim);
            goto fail;
        } else if (spec & (__Pyx_MEMVIEW_PTR)) {
            PyErr_Format(PyExc_ValueError,
                         "C-contiguous buffer is not indirect in "
                         "dimension %d", dim);
            goto fail;
        } else if (buf->suboffsets) {
            PyErr_SetString(PyExc_ValueError,
                            "Buffer exposes suboffsets but no strides");
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_check_suboffsets(Py_buffer *buf, int dim, CYTHON_UNUSED int ndim, int spec)
{
    if (spec & __Pyx_MEMVIEW_DIRECT) {
        if (buf->suboffsets && buf->suboffsets[dim] >= 0) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer not compatible with direct access "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    if (spec & __Pyx_MEMVIEW_PTR) {
        if (!buf->suboffsets || (buf->suboffsets[dim] < 0)) {
            PyErr_Format(PyExc_ValueError,
                         "Buffer is not indirectly accessible "
                         "in dimension %d.", dim);
            goto fail;
        }
    }
    return 1;
fail:
    return 0;
}
static int
__pyx_verify_contig(Py_buffer *buf, int ndim, int c_or_f_flag)
{
    int i;
    if (c_or_f_flag & __Pyx_IS_F_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = 0; i < ndim; i++) {
            if (stride * buf->itemsize != buf->strides[i] &&
                    buf->shape[i] > 1)
            {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not fortran contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    } else if (c_or_f_flag & __Pyx_IS_C_CONTIG) {
        Py_ssize_t stride = 1;
        for (i = ndim - 1; i >- 1; i--) {
            if (stride * buf->itemsize != buf->strides[i] &&
                    buf->shape[i] > 1) {
                PyErr_SetString(PyExc_ValueError,
                    "Buffer not C contiguous.");
                goto fail;
            }
            stride = stride * buf->shape[i];
        }
    }
    return 1;
fail:
    return 0;
}
static int __Pyx_ValidateAndInit_memviewslice(
                int *axes_specs,
                int c_or_f_flag,
                int buf_flags,
                int ndim,
                __Pyx_TypeInfo *dtype,
                __Pyx_BufFmt_StackElem stack[],
                __Pyx_memviewslice *memviewslice,
                PyObject *original_obj)
{
    struct __pyx_memoryview_obj *memview, *new_memview;
    __Pyx_RefNannyDeclarations
    Py_buffer *buf;
    int i, spec = 0, retval = -1;
    __Pyx_BufFmt_Context ctx;
    int from_memoryview = __pyx_memoryview_check(original_obj);
    __Pyx_RefNannySetupContext("ValidateAndInit_memviewslice", 0);
    if (from_memoryview && __pyx_typeinfo_cmp(dtype, ((struct __pyx_memoryview_obj *)
                                                            original_obj)->typeinfo)) {
        memview = (struct __pyx_memoryview_obj *) original_obj;
        new_memview = NULL;
    } else {
        memview = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                            original_obj, buf_flags, 0, dtype);
        new_memview = memview;
        if (unlikely(!memview))
            goto fail;
    }
    buf = &memview->view;
    if (buf->ndim != ndim) {
        PyErr_Format(PyExc_ValueError,
                "Buffer has wrong number of dimensions (expected %d, got %d)",
                ndim, buf->ndim);
        goto fail;
    }
    if (new_memview) {
        __Pyx_BufFmt_Init(&ctx, stack, dtype);
        if (!__Pyx_BufFmt_CheckString(&ctx, buf->format)) goto fail;
    }
    if ((unsigned) buf->itemsize != dtype->size) {
        PyErr_Format(PyExc_ValueError,
                     "Item size of buffer (%" CYTHON_FORMAT_SSIZE_T "u byte%s) "
                     "does not match size of '%s' (%" CYTHON_FORMAT_SSIZE_T "u byte%s)",
                     buf->itemsize,
                     (buf->itemsize > 1) ? "s" : "",
                     dtype->name,
                     dtype->size,
                     (dtype->size > 1) ? "s" : "");
        goto fail;
    }
    for (i = 0; i < ndim; i++) {
        spec = axes_specs[i];
        if (!__pyx_check_strides(buf, i, ndim, spec))
            goto fail;
        if (!__pyx_check_suboffsets(buf, i, ndim, spec))
            goto fail;
    }
    if (buf->strides && !__pyx_verify_contig(buf, ndim, c_or_f_flag))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview, ndim, memviewslice,
                                         new_memview != NULL) == -1)) {
        goto fail;
    }
    retval = 0;
    goto no_fail;
fail:
    Py_XDECREF(new_memview);
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}

/* ObjectToMemviewSlice */
      static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn_fsl_int(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_nn_fsl_int, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* MemviewDtypeToObject */
      static CYTHON_INLINE PyObject *__pyx_memview_get_nn_real(const char *itemp) {
    return (PyObject *) PyFloat_FromDouble(*(real *) itemp);
}
static CYTHON_INLINE int __pyx_memview_set_nn_real(const char *itemp, PyObject *obj) {
    real value = __pyx_PyFloat_AsDouble(obj);
    if ((value == ((real)-1)) && PyErr_Occurred())
        return 0;
    *(real *) itemp = value;
    return 1;
}

/* ObjectToMemviewSlice */
      static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_ds_nn_real(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_STRIDED) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, 0,
                                                 PyBUF_RECORDS_RO | writable_flag, 1,
                                                 &__Pyx_TypeInfo_nn_real, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* ObjectToMemviewSlice */
      static CYTHON_INLINE __Pyx_memviewslice __Pyx_PyObject_to_MemoryviewSlice_d_dc_nn_real(PyObject *obj, int writable_flag) {
    __Pyx_memviewslice result = { 0, 0, { 0 }, { 0 }, { 0 } };
    __Pyx_BufFmt_StackElem stack[1];
    int axes_specs[] = { (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_FOLLOW), (__Pyx_MEMVIEW_DIRECT | __Pyx_MEMVIEW_CONTIG) };
    int retcode;
    if (obj == Py_None) {
        result.memview = (struct __pyx_memoryview_obj *) Py_None;
        return result;
    }
    retcode = __Pyx_ValidateAndInit_memviewslice(axes_specs, __Pyx_IS_C_CONTIG,
                                                 (PyBUF_C_CONTIGUOUS | PyBUF_FORMAT) | writable_flag, 2,
                                                 &__Pyx_TypeInfo_nn_real, stack,
                                                 &result, obj);
    if (unlikely(retcode == -1))
        goto __pyx_fail;
    return result;
__pyx_fail:
    result.memview = NULL;
    result.data = NULL;
    return result;
}

/* CIntToPy */
      static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum____pyx_t_10fatslimlib_9core_base_ForcefieldType(enum __pyx_t_10fatslimlib_9core_base_ForcefieldType value) {
    const enum __pyx_t_10fatslimlib_9core_base_ForcefieldType neg_one = (enum __pyx_t_10fatslimlib_9core_base_ForcefieldType) ((enum __pyx_t_10fatslimlib_9core_base_ForcefieldType) 0 - (enum __pyx_t_10fatslimlib_9core_base_ForcefieldType) 1), const_zero = (enum __pyx_t_10fatslimlib_9core_base_ForcefieldType) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_10fatslimlib_9core_base_ForcefieldType) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_10fatslimlib_9core_base_ForcefieldType) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_10fatslimlib_9core_base_ForcefieldType) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_10fatslimlib_9core_base_ForcefieldType) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_10fatslimlib_9core_base_ForcefieldType) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_10fatslimlib_9core_base_ForcefieldType),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
      static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* None */
      static CYTHON_INLINE int __Pyx_ErrOccurredWithGIL(void) {
  int err;
  #ifdef WITH_THREAD
  PyGILState_STATE _save = PyGILState_Ensure();
  #endif
  err = !!PyErr_Occurred();
  #ifdef WITH_THREAD
  PyGILState_Release(_save);
  #endif
  return err;
}

/* MemviewSliceCopyTemplate */
      static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object)
{
    __Pyx_RefNannyDeclarations
    int i;
    __Pyx_memviewslice new_mvs = { 0, 0, { 0 }, { 0 }, { 0 } };
    struct __pyx_memoryview_obj *from_memview = from_mvs->memview;
    Py_buffer *buf = &from_memview->view;
    PyObject *shape_tuple = NULL;
    PyObject *temp_int = NULL;
    struct __pyx_array_obj *array_obj = NULL;
    struct __pyx_memoryview_obj *memview_obj = NULL;
    __Pyx_RefNannySetupContext("__pyx_memoryview_copy_new_contig", 0);
    for (i = 0; i < ndim; i++) {
        if (from_mvs->suboffsets[i] >= 0) {
            PyErr_Format(PyExc_ValueError, "Cannot copy memoryview slice with "
                                           "indirect dimensions (axis %d)", i);
            goto fail;
        }
    }
    shape_tuple = PyTuple_New(ndim);
    if (unlikely(!shape_tuple)) {
        goto fail;
    }
    __Pyx_GOTREF(shape_tuple);
    for(i = 0; i < ndim; i++) {
        temp_int = PyInt_FromSsize_t(from_mvs->shape[i]);
        if(unlikely(!temp_int)) {
            goto fail;
        } else {
            PyTuple_SET_ITEM(shape_tuple, i, temp_int);
            temp_int = NULL;
        }
    }
    array_obj = __pyx_array_new(shape_tuple, sizeof_dtype, buf->format, (char *) mode, NULL);
    if (unlikely(!array_obj)) {
        goto fail;
    }
    __Pyx_GOTREF(array_obj);
    memview_obj = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                    (PyObject *) array_obj, contig_flag,
                                    dtype_is_object,
                                    from_mvs->memview->typeinfo);
    if (unlikely(!memview_obj))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview_obj, ndim, &new_mvs, 1) < 0))
        goto fail;
    if (unlikely(__pyx_memoryview_copy_contents(*from_mvs, new_mvs, ndim, ndim,
                                                dtype_is_object) < 0))
        goto fail;
    goto no_fail;
fail:
    __Pyx_XDECREF(new_mvs.memview);
    new_mvs.memview = NULL;
    new_mvs.data = NULL;
no_fail:
    __Pyx_XDECREF(shape_tuple);
    __Pyx_XDECREF(temp_int);
    __Pyx_XDECREF(array_obj);
    __Pyx_RefNannyFinishContext();
    return new_mvs;
}

/* CIntFromPy */
      static CYTHON_INLINE fsl_int __Pyx_PyInt_As_fsl_int(PyObject *x) {
    const fsl_int neg_one = (fsl_int) ((fsl_int) 0 - (fsl_int) 1), const_zero = (fsl_int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(fsl_int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(fsl_int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (fsl_int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (fsl_int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(fsl_int, digit, digits[0])
                case 2:
                    if (8 * sizeof(fsl_int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(fsl_int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(fsl_int) >= 2 * PyLong_SHIFT) {
                            return (fsl_int) (((((fsl_int)digits[1]) << PyLong_SHIFT) | (fsl_int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(fsl_int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(fsl_int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(fsl_int) >= 3 * PyLong_SHIFT) {
                            return (fsl_int) (((((((fsl_int)digits[2]) << PyLong_SHIFT) | (fsl_int)digits[1]) << PyLong_SHIFT) | (fsl_int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(fsl_int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(fsl_int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(fsl_int) >= 4 * PyLong_SHIFT) {
                            return (fsl_int) (((((((((fsl_int)digits[3]) << PyLong_SHIFT) | (fsl_int)digits[2]) << PyLong_SHIFT) | (fsl_int)digits[1]) << PyLong_SHIFT) | (fsl_int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (fsl_int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(fsl_int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(fsl_int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(fsl_int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(fsl_int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (fsl_int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(fsl_int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(fsl_int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(fsl_int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(fsl_int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(fsl_int) - 1 > 2 * PyLong_SHIFT) {
                            return (fsl_int) (((fsl_int)-1)*(((((fsl_int)digits[1]) << PyLong_SHIFT) | (fsl_int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(fsl_int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(fsl_int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(fsl_int) - 1 > 2 * PyLong_SHIFT) {
                            return (fsl_int) ((((((fsl_int)digits[1]) << PyLong_SHIFT) | (fsl_int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(fsl_int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(fsl_int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(fsl_int) - 1 > 3 * PyLong_SHIFT) {
                            return (fsl_int) (((fsl_int)-1)*(((((((fsl_int)digits[2]) << PyLong_SHIFT) | (fsl_int)digits[1]) << PyLong_SHIFT) | (fsl_int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(fsl_int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(fsl_int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(fsl_int) - 1 > 3 * PyLong_SHIFT) {
                            return (fsl_int) ((((((((fsl_int)digits[2]) << PyLong_SHIFT) | (fsl_int)digits[1]) << PyLong_SHIFT) | (fsl_int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(fsl_int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(fsl_int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(fsl_int) - 1 > 4 * PyLong_SHIFT) {
                            return (fsl_int) (((fsl_int)-1)*(((((((((fsl_int)digits[3]) << PyLong_SHIFT) | (fsl_int)digits[2]) << PyLong_SHIFT) | (fsl_int)digits[1]) << PyLong_SHIFT) | (fsl_int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(fsl_int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(fsl_int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(fsl_int) - 1 > 4 * PyLong_SHIFT) {
                            return (fsl_int) ((((((((((fsl_int)digits[3]) << PyLong_SHIFT) | (fsl_int)digits[2]) << PyLong_SHIFT) | (fsl_int)digits[1]) << PyLong_SHIFT) | (fsl_int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(fsl_int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(fsl_int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(fsl_int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(fsl_int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            fsl_int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (fsl_int) -1;
        }
    } else {
        fsl_int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (fsl_int) -1;
        val = __Pyx_PyInt_As_fsl_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to fsl_int");
    return (fsl_int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to fsl_int");
    return (fsl_int) -1;
}

/* CIntFromPy */
      static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
      static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
      static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
      static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
    const char neg_one = (char) ((char) 0 - (char) 1), const_zero = (char) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(char, digit, digits[0])
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 2 * PyLong_SHIFT) {
                            return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 3 * PyLong_SHIFT) {
                            return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 4 * PyLong_SHIFT) {
                            return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (char) -1;
        }
    } else {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* CheckBinaryVersion */
      static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* PyObjectSetAttrStr */
      #if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* VoidPtrExport */
      static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig) {
    PyObject *d;
    PyObject *cobj = 0;
    d = PyDict_GetItem(__pyx_d, __pyx_n_s_pyx_capi);
    Py_XINCREF(d);
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (__Pyx_PyObject_SetAttrStr(__pyx_m, __pyx_n_s_pyx_capi, d) < 0)
            goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItem(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionExport */
      static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionImport */
      #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
      static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
